// ---------------------------------------------------------------
// function     : l4c_DAILY.c
// purpose      : main() for l4c_DAILY.ex, NASA SMAP mission L4_C suite
// author       : joe glassy, Lupine Logic Inc.
// version      : see L4C_VERS_TAG , as 5.0 or later
// revised      : v7.0-2015-09-06T14:31:00-07:00
// revised      : v6.0-2015-07-28T09:00:00 --intermediate version, only sync with other L4_C components.
// revised      : v5.0.1-2014-11-10T15:13:00, updated for ORT2, see l4c_DAILY.h
//                change-log for specific changes made.
// history      : v5.0.1-2014-11-03T11:14, added day loop for future UM only
//                multi-day modeling capability.
// history      : v5.0.1-2014-08-21T16:34:00:00, for D5.0.1 and beyond.
//                Integrated FparClimatology and NDVI option-adds, and
//                updated BPLUT to 2/18/2014 instance, and implements
//                Lucas 6/4/2014 and 6/5/2014 changes noted in HISTORY.txt
// history      : v5.0.1-2014-04-30T12:54:00-07:00, for D5.0.1
// history      : v5.0.1 summary of changes and minor fixes/enhancements in progress:
//                a) deprecate WRITE of surface_flags (allow read),
//                b) populate bitflags subfields,
//                c) address rootzone SM scaling, d) fix fpar_qa bit handling,
//                e) fix InCell function, f) implement Fpar Climatology I/O

// invocation   : ./l4c_DAILY.ex <l4c_DAILY_ddd.rip>
// dependencies : HDF4, HDF5 libraries and L4C FRAMEWORK API.
// history             : v5.0.1 changes consist of minor fixes/enhancements: a) deprecate surface_flags
//                       b) populate bitflags subfields, c) address rootzone SM scaling, d) TBD
// history      : 2014-02-05T11:23:00-07:00, integration checkpoint
// history      : 2014-01-16T10:03:00-07:00, integration checkpoint
// history      : 2014-01-07T09:48:00-07:00, integration checkpoint
// history      : 2014-01-01T11:23:00-07:00, monthly checkpoint
// history      : 2013-12-01T09:27:00-07:00, monthly checkpoint
// history      : 2013-11-22T16:02:00-07:00, re-factoring review
//              : 2013-11-01, monthly checkpoint
//              : 2013-10-01, monthly checkpoint
//              : 2013-09-01, monthly checkpoint
//              : 2013-08-01, monthly checkpoint
//              : 2013-07-02, D4.0 delivered to NCCS
//
//  RECAP of Calendar Protocols for SOC and F/T Rolling now used:
//   opMode =0 adjusts JUST the curr_doy slot to 1. (default Ops mode behavior)
//   opMode =1 forces a ZERO OUT of all day slots in calendar.
//   opMode =2 adjusts all slots to set these to 1 from beginning of year
//             up to and including the current model day-of-year
//
// notes        : 2014-11-03, Since l4c_DAILY.ex was originally designed to run for
//                exactly one "model day" the l4c_InitSession() logic
//                performs session-wise initialization for the whole session
//                itself, as well as initialization for exactly ONE day's run,
//                so to enhance this to work with multi-day runs, some logic
//                in l4c_InitSession must be analyzed and migrated to
//                trigger within the placeholder daily loop shown below/
// notes        : L4_C Output filename form:
//                SMAP_L4_C_MDL_20140216T221400_R01100_001.h5
//                SMAP_L4_C_MDL_20140216T221400_R01100_001.qa

// ---------------------------------------------------------------

#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"

// ---------------------------------------------------------------
int main(int argc,char *argv[])
{

 SesCtl *sesCtl_p = l4c_GetSesCtl();
 char fxName[256] = {0} ;
 int status = 0 ;
 int runDay = 0 ;


 strcpy(fxName,__func__);

 // Initialize a single daily l4c_DAILY.ex session
 l4c_InitSession(argc,argv);

    if(sesCtl_p->runDayOfYear > 366 || sesCtl_p->runToStopDay > 366)
    {
        xLogMsg("@F In(%s) Unexpected doy(%d) or run-to-stop-day(%d)~\n",
             fxName, sesCtl_p->runDayOfYear, sesCtl_p->runToStopDay);
        exit(-1);
    }

    // <JMG> 2014-11-03, added for future multi-day modeling capability
    // TBD: to enable multi-day runs, some (not all) logic now in
    // l4c_InitSession() will need to be moved INSIDE this loop.
    for(runDay =  sesCtl_p->runDayOfYear;
        runDay <= sesCtl_p->runToStopDay ;
        runDay += sesCtl_p->runDayStepInterval)
    {
        xLogMsg("\n");
        xLogMsg("@A In(%s) RUN DAY %5d STOP DAY %5d using DAYSTEPINTERVAL %d (relevant only for multi-day mode)~ \n",
            fxName, runDay, sesCtl_p->runToStopDay, sesCtl_p->runDayStepInterval);

        // implement PRE-PROCESSSING operations
        l4c_DailyDriver(runDay, DoPreProcessing) ;

        // implement primary TCF DAILY Forward Step logic for current day
        l4c_Action_forwardStep(); // <KAE> TODO

        // implement POST-PROCESSOR (Inflate)
        l4c_DailyDriver(runDay,DoPostProcessing);

    } // end:: multi-day loop


 // Close out session
 l4c_CloseSession();

 status= EXIT_SUCCESS ;

 return status;

} // end::main()
