# -----------------------------------------------------------
# file         : l4c_DAILY.mak
# purpose      : Makefile for ./L4C_EVALSYS l4c_DAILY.ex executable, that also now
#                freshens the underlying UM Framework library build as necessary
# revision     : 2020-07-09T18:509, fixed this makefile to work in the unfortunate case where
#                HDF4 libraries are in /usr/lib64, and where the HDF4_LIB def points to /usr/lib64
#                and where HDF5 library was installed via copy which obliterates the required
#                soft links, making hash soup out of all the .so cascading references.
# revision     : 2020-07-08T12:49, updated for use with HDF5 v1.10.4. in progress.
# revision     : 2016-02-18T08:54, check for consistency with GMAO build l4c_DAILY_NCCS.mak
# revision     : 2016-02-17T15:20:00, update for v7.1.0 delivery, consolidated more
#                source names to match preface, l4c_DAILY_{xxxx}.c
# history      : 2015-09-07T20:27:00, minor update for v7.0.x delivery
# history      : 2015-09-07T20:27:00, minor update for v7.0 delivery
# history      : 2014-10-30T14:45:00, added l4c_DAILY_Diagnostics.c module to split out
#                all "diagnostic only" code into one module file; prev in l4c_DAILY_sci.c
# history      : 2014-09-04T14:26, renamed l4c_GridManager.c to l4c_DAILY_DailyDriver.c
#               to distinghish it from future l4c_DAILY_AnnualDriver.c
# history      : 2014-08-14T15:32, added l4c_DAILY_PrepOutput.c on refactoring
# history      : 2014-07-25T17:13, merged in all recent LAJ and JMG changes and
#                split out RIP nvp logic into l4c_DAILY_rip.c module file.
# history      : 2014-06-24T12:33:00, added l4c_DAILY_WriteMetadata.c to
#                address need for writing "group-of-groups" metadata
# history      : 2014-05-19T12:28:00, consolidated l4c_DAILY_PkgMgrInit_Themes.c into
#                l4c_DAILY_PkgMgrInit.c to simplify.
# history      : 2014-03-14T15:48:000, modified to also freshen the underlying UM Framework API
#                library, static  "libfrm.a"
# history      : 2014-03-03T15:42:00, refactored pre proc code into
#                l4c_PreProc1km_Driver.c and added this to Makefile.
# history      : 2014-02-21T15:25:00, further runtimetesting
# history      : 2014-02-20T10:10,now runtime testing, working through early segfaults
# history      : 2014-02-17T20:29:00,improved Makefile from missouri clean ups
# history      : 2014-02-16T16:02:00, lead fork, final integration
# history      : 2014-02-15T16:02:00, lead fork PASS 1,2,3
# history      : 2014-02-14T13:43:00, lead fork review
# history      : 2014-02-05T17:24:00
# history      : 2013-12-16T16:57:00-07:00,wheat,jmg
# history      : 2013-11-11T09:45:00-07:00,ruby,jmg
# author       : joe glassy, Lupine Logic Inc.
#
# platform     : wheat.ntsg.umt.edu
# dependencies : HDF4, HDF5 and FRM_API (libfrm.a) libraries and the following
#                predefined environment variables: FRM_API_ROOT, FRM_API_LIB, FRM_API_INC
#                HDF4_INC,HDF4_LIB, HDF5_INC,HDF5_LIB
# dependencies : First define L4CAPI_ROOT to point to the base directory
#                where the UM Framework API code resides, as in:
#                FRM_API_ROOT=/pilv2/SMAP/L4CSYS/L4CAPI
# invocation   : make -f l4c_DAILY.mak |& tee -a BuildLog-v7.4.1-YYYY-mm-ddThhmm.log
# notes        : The following naming conventions apply:
#			   : --module files starting with "l4c_" indicate they are part
#              :   of the overall L4_C namespace, not particular to just the
#              :   l4c_DAILY code set.
#              : --module files starting with l4c_Day_" indicate they are
#              : specifically part of just the l4c_Day_* algorithm code set.
# -----------------------------------------------------------

CC = gcc

# CFLAGS =  -g -Wall -std=c99 -fPIC -D_GNU_SOURCE -DHAVE_UUID
# Optimized version of build
# CFLAGS =  -O3 -fPIC -D_GNU_SOURCE -DHAVE_UUID -DHAVE_HDF4 -DHAVE_HDF5

# Debug version of build. Use c99 to allow use of fpclassify() floating point exception trap
# for production use, remove -g and add -O
CFLAGS =  -std=gnu99 -g -Wall -Wextra -fpic -D_GNU_SOURCE -DHAVE_UUID -DHAVE_HDF4 -DHAVE_HDF5

# <JMG> 2014-09-30, changed to use FRM_API_BASE on /pilv2/SMAP/L4CSYS/v7/L4CAPI
# Inheriting L4CAPI_ROOT env var set in .bashrc
FRM_API_SRC = $(L4CAPI_ROOT)/src
# In case these aren't explicitly defined we define these here:
FRM_API_INC = $(L4CAPI_ROOT)/inc
FRM_API_LIB = $(L4CAPI_ROOT)/lib
# <JMG>
FRM_LIB = $(FRM_API_LIB)/libfrm.a

INCDIR =.

INCS = -I$(FRM_API_INC) -I$(HDF4INC) -I$(HDF5_INC) -I$(INCDIR) -I/usr/local/include

# Re-ordered these to put HDF4 at the end, after the HDF5_LIB so the correct libraries
#   are used. 2020-07-09 <JMG>
LIBS = -L$(HDF5_LIB) -L$(FRM_API_LIB) -L$(HDF4_LIB) -L/usr/local/lib

# <JMG> 2015-05-18,test inclusion of -lsz to see if found ok in /usr/local/lib
LDFLAGS = -ldl -lfrm -ldf -lmfhdf -lhdf5 -lhdf5_hl -lm -lsz -luuid -lz

all: $(FRM_LIB)  $(TARGET)

# added...
ARTOOL  = ar
ARFLAGS = rv
# added...
.SUFFIXES : .o .c .a .h

# target for building L4_C executable itself
TARGET = l4c_DAILY.ex

# PLEASE Note: some of these modules are retained and linked in but are not
#  actively used on l4c_DAILY.ex, as they pertain to spinup processes or
#  other processes not directly involving the daily L4_C run.

# <JMG> 10-30-2014, added l4c_DAILY_Diagnostics.c module
#
# took out  uuta.c, array_io.c soon..  update intermediates...

SRCS = l4c_DAILY_Action.c\
	l4c_DAILY_BindToPkgMgr.c\
	l4c_DAILY_bplut.c\
	l4c_DAILY_DailyDriver.c\
	l4c_DAILY_DatasetProperties.c\
	l4c_DAILY_Diagnostics.c\
	l4c_DAILY_EASEGRID_Transforms.c\
	l4c_DAILY_ForwardStep.c\
	l4c_DAILY_FparQaBin.c\
	l4c_DAILY_GridPost_Inflate.c\
	l4c_DAILY_GridPre_Deflate.c\
	l4c_DAILY_hdf5.c\
	l4c_DAILY_Hdf5_Read_and_Updates.c\
	l4c_DAILY_ModFunc.c \
	l4c_DAILY_ModMult.c \
	l4c_DAILY_PkgMgrInit.c\
	l4c_DAILY_PostProcessor.c\
	l4c_DAILY_PrepOutput.c \
	l4c_DAILY_PreProcessor.c\
	l4c_DAILY_rip.c \
	l4c_DAILY_sci.c \
	l4c_DAILY_ses.c \
	l4c_DAILY_StringUtilTcf.c\
	l4c_DAILY_tcfMem.c \
	l4c_DAILY_tcfModDriverSubset.c\
	l4c_DAILY_tcfModQc.c\
	l4c_DAILY_WriteMetadata.c \
	l4c_DAILY.c

SRCS_LIB = $(FRM_API_SRC)/frm_core.c \
	$(FRM_API_SRC)/frm_mem_obj.c \
	$(FRM_API_SRC)/frm_log_msg.c \
	$(FRM_API_SRC)/frm_str_fx.c  \
	$(FRM_API_SRC)/frm_tiv_api.c \
	$(FRM_API_SRC)/frm_io_fx.c \
	$(FRM_API_SRC)/frm_util_fx.c \
	$(FRM_API_SRC)/frm_hdf_io.c\
	$(FRM_API_SRC)/nsb_api.c \
	$(FRM_API_SRC)/nsb_dbsys.c\
	$(FRM_API_SRC)/nsb_accessors.c \
	$(FRM_API_SRC)/nsb_parsers.c \
	$(FRM_API_SRC)/pkg_mgr_objects.c \
	$(FRM_API_SRC)/pkg_mgr_themes.c \
	$(FRM_API_SRC)/pkg_mgr.c

# Create one object file per .c source file, from the application sources
OBJS = $(SRCS:.c=.o)

# Create one object file per .c source file from the library c sources
OBJS_LIB = $(SRCS_LIB:.c=.o)

# Global default rule, update the library, then the executable application
all: $(FRM_LIB) $(TARGET)

# rule for building the application
$(TARGET): $(OBJS)
	$(CC) $(CFLAGS) -o $(TARGET) $(OBJS) $(INCS) $(LIBS) $(LDFLAGS)

# Rule for building or updating the UM Framework library
# from each .o combine to construct the static library, .a
$(FRM_LIB): ${OBJS_LIB}
	$(ARTOOL) $(ARFLAGS) $@ $?


# Rule to update the .a library whenever a newer .c file is encountered
.c.a :
	$(CC) $(CFLAGS) $(ADDFLAGS) $(INCS) $<
	$(ARTOOL) $(ARFLAGS) $@ $*.o
	rm -f $*.o

# Rule to update the .a library whenever a newer .o file is found
.o.a :
	$(ARTOOL) $(ARFLAGS) $<

clean:
	rm -f *.o *~ *.ex
	rm -f $(FRM_API_SRC)/*.o $(FRM_API_SRC)/*.a $(FRM_API_LIB)/*.a

install:
	mv $(TARGET) ..


# L4_C standard test stanzas...
# omit for now: l4c_DAILY_VALID.keys
test:
	./$(TARGET)  l4c_DAILY.rip
	# Extract observed min,max,mean for each dataset, from latest log file
	grep OVERVIEW SMAP_L4_C_MDL.log | cut -d' ' -f4,8,10,12 | less

testoverview:
	grep OVERVIEW SMAP_L4_C_MDL.log | cut -d' ' -f4,8,10,12 | less

# Update the .o files whenever a newer .c file is encountered
.c.o:
	$(CC) $(CFLAGS) $(INCS) -c $< -o $@
