// ---------------------------------------------------------------
// function     : l4c_DAILY_Action.c
// purpose      : Encapsulates tcfAction.c logic for just the
//                TCF "step_forward" action
// orig author  : Lucas Jones (tcfAction.c)
// version      : see L4C_VERS_TAG , as v4.1.5 or later
// revised      : 2014-07-23T120:00:00, modified to do aggregation of 1km to 9km
// within forward step, laj revised      : 2014-03-03T09:49:00, now that the
// pre-processor is
//                allocating all major eco and met arrays and passing
//                them to the "core", all the alloc's of these need to
//                be removed from this and TCF Mem..
// history      : 2014-02-21T15:26:00, integrated with PkgMgr et al
// history      : 2014-02-05T11:242:00-07:00,wheat,jmg"
// dependencies : HDF4, HDF5 libraries and L4C FRAMEWORK API.
// author       : joe glassy, Lupine Logic Inc.
// history      : n/a
// notes        : Based on and adapted from tcf_spin v3.x, this
//                module contains logic for just the DAILY "step_forward"
//                whereas the l4c_DAILY_ActionSpinup.c and
//                l4c_DAILY_spinup_logic.c contain the
//                corresponding functions for analytical, numerical
//                spin up as well as the seasonal-adjustment modes
// ---------------------------------------------------------------

// Additional documentation and  notes:

//  purpose      : High level functions for managing model simulation
//                 and initialization modes (one function for each mode).
//                 These functions allocate/free array memory,
//                 load static inputs, and call driver functions.
//  project      : SMAP L4C science development.
//  dependencies : tcfModDriverGrid.c, tcfMem.c
//  revision     : 2014-02-15T13:16:00, for tcf_spin v3.2 lead fork ADT
//  namespace updates revision     : Initial reformating -- laj 09/09/13 author
//  : Lucas A. Jones; lucas@ntsg.umt.edu notes        :

// #include "tcf_inc.h" now re-factored into l4c_DAILY.h
#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"

// SUITES of MASTER OBJECT (namespace xxxxMasterObj)..bundles of images
static AncImagePkgAdt ancImagePkgMasterObj;
// <JMG> 2014-11-03, container holding various pre-and-post phase image
// buffers that previously haven't been well organized.
static PrePostImgPkgAdt prePostImagePkgMasterObj;

static PftAdt pftAdtMasterObj;

// PRIMARY CONTAINERS:
static MetArrayAdt metArrayMasterObj;
static EcoStateArrayAdt ecoStateArrayMasterObj;
// per-pixel containers
static EcoStateAdt ecoStateMasterObj;
static EcoStateQaAdt ecoStateQaMasterObj;
static MultCoefAdt multCoefMasterObj;

// ---------------------------------------------------------------
// function : l4c_GetAncImagePkgAdt()
// purpose  : return a pointer to ancillary image package
// revision : 2014-02-23T14:36
// returns  : AncImagePkgAdt * ptr to MasterObj
// notes    : This container organizes various ancillary image buffers
//            such as the primary Plant Function Type (PFT) 1km image
//            and pixel-wise filter images etc.
AncImagePkgAdt *l4c_GetAncImagePkgAdt(void) {
  return &ancImagePkgMasterObj;

} // end:: l4c_GetAncImagePkgAdt()

// ---------------------------------------------------------------
// function : l4c_GetPrePostImagePkgAdt()
// purpose  : return a pointer to master instance of the pre-and-post
//            image package (container).
// revision : 2014-11-03T17:55:00, added to collect all prev. unorganized
//            special case image buffers into one container.
// returns  : PrePostImgPkgAdt *p -- ptr to pre-and-post image container
// notes    : This container organizes various large "re-usable" image
//            buffers used in pre and post phase processing, as well as
//            any other "special case" variable buffers
PrePostImgPkgAdt *l4c_GetPrePostImagePkgAdt(void) {
  return &prePostImagePkgMasterObj;

} // end:: l4c_GetPrePostImagePkgAdt()

// ---------------------------------------------------------------
// function : l4c_GetEcoStateAdt()
// purpose  : return a pointer to EcoStateAdt package
// revision : 2014-02-23T14:36
// returns  : EcoStateAdt* ptr to MasterObj
EcoStateAdt *l4c_GetEcoStateAdt(void) {
  return &ecoStateMasterObj;
} // end:: l4c_GetEcoStateAdt()

// ---------------------------------------------------------------
// function : l4c_GetMultCoefAdt()
// purpose  : return a pointer to MultCoefAdt package
// revision : 2014-02-23T14:36
// returns  : MultCoefAdt* ptr to MasterObj
MultCoefAdt *l4c_GetMultCoefAdt(void) {
  return &multCoefMasterObj;
} // end:: l4c_GetMultCoefAdt();

// ---------------------------------------------------------------
// function : l4c_GetEcoStateQaAdt()
// purpose  : return a pointer to EcoStateQaAdt package
// revision : 2014-02-23T14:36
// returns  : EcoStateQaAdt* ptr to QA MasterObj
EcoStateQaAdt *l4c_GetEcoStateQaAdt(void) {
  return &ecoStateQaMasterObj;

} // end:: l4c_GetEcoStateQaAdt()

// ---------------------------------------------------------------
// function : l4c_GetEcoStateArrayAdt()
// purpose  : return a pointer to EcoStateArrayAdt package
// revision : 2014-02-23T14:36
// returns  : EcoStateArrayAdt* ptr to MasterObj
EcoStateArrayAdt *l4c_GetEcoStateArrayAdt(void) {
  return &ecoStateArrayMasterObj;

} // end:: l4c_GetEcoStateArrayAdt()

// ---------------------------------------------------------------
// function : l4c_GetMetArrayAdt()
// purpose  : return a pointer to MetArrayAdt image package
// revision : 2014-02-23T14:36
// returns  : MetArrayAdt *mPtr_p to MasterObj
MetArrayAdt *l4c_GetMetArrayAdt(void) {
  return &metArrayMasterObj;

} // end:: l4c_GetMetArrayAdt()

// ---------------------------------------------------------------
// function : l4c_GetPftAdt()
// purpose  : return a pointer to PftAdt image package
// revision : 2014-02-23T14:36
// returns  : PftAdt *mPtr_p to MasterObj
PftAdt *l4c_GetPftAdt(void) {
  return &pftAdtMasterObj;
} // end:: l4c_GetPftAdt();

// ------------------------------------------------------------------
// function : l4c_Action_forwardStep()   --replaces tcfAction_forwardStep()
// Purpose  : Does forward step for one specified simulation day.
//            Output state is comprehensive with all
//            product-specification-document (PSD) states tracked,
//            including QA fields.
//            --This logic drives forward simulation for a specified
//            single time-step. Produces all variables specified in the
//            L4C ATBD, including QA fields.
// Inputs   : SesCtl *sesCtl_p
// Inputs:   Initial C pools, passed in the STATE_ARRAY structure, dynamic daily
// met data (MetArrayObj),
//           static 1D sparse vector, pft map (PFT_ARRAY), and static model
//           parameters (GLOB_PARAM, BPLUT).
//
// Outputs: Daily states, including final C pools, in 1km grid arrays returned
//          in the STATE_ARRAY structure.
// revision : 2014-02-16T12:44:00, from LEAD FORK PASS 3.
// Returns  : void
// Notes:   Increments the 't8day' counter variable to determine when to read in
// available fpar 8-day composites. author   : Lucas Jones, joe Glassy notes :
// n/a
//
void l4c_Action_forwardStep(void) {
  SesCtl *sesCtl_p = l4c_GetSesCtl();
  MetArrayAdt *metArrayObj_p = &metArrayMasterObj;
  EcoStateArrayAdt *ecoStateArrayObj_p = &ecoStateArrayMasterObj;
  PftAdt *pftObj_p = &pftAdtMasterObj;
  AncImagePkgAdt *ancImagePkgObj_p = &ancImagePkgMasterObj;

  // size_t nElem1km_M01land =  EGv2_LDOM1KM ; // 134787240  ;
  uint32 nelem = 0;

  xLogMsg("@A In(%s) task (%s) executing daily TCF model (FORWARD STEP) for "
          "GRID data~\n",
          __func__, sesCtl_p->subtaskIdStr);

  // l4c_SesInitAnc() allocated and read "MCD12Q1_M01land_lc_dom_uint8")
  // so set a convenience pointer to this existing memory
  pftObj_p->dom_1km = ancImagePkgObj_p->imgPftM01land_dom_p;

  l4c_BindPkg_AncillaryStatic(ancImagePkgObj_p, metArrayObj_p);

  // BIND new MET arrays to PkgMgr nodes. This also includes
  // fpar, fpar_qa, ndvi as viirsNdvi, ndvi_qc as viirsNdviQc, smrz_mn,smrz_mx
  // etc.
  l4c_BindPkg_MetArrays(metArrayObj_p);

  // ATTACH TO 1D-Sparse Vector OUTPUT arrays already done!
  // BIND new SOC and Npp arrays just allocated
  l4c_BindPkg_SOC_and_Npp(ecoStateArrayObj_p);

  // <JMG> 2014-08-07, now we also include
  //  binding of SMAP NatureRun soil rootzone saturation min,max:
  l4c_BindPkg_SOC_Intermediate(ancImagePkgObj_p, metArrayObj_p);

  // for a serial task, nElem1kmPerTask should be EGv2_LDOM1KM =134787240
  nelem = sesCtl_p->nElem1kmPerTask;

  // Allocate Pft nested 1km array (81 elements) to hold information
  // required by post-processor 9km summarization functions
  pftObj_p->nested_1km =
      (uint8 *)memCalloc(NESTED_1KM_IN_9KM_SQ, sizeof(uint8));

  //  ALLOCATE OUTPUT STATE GROUP ARRAYS in 1D Sparse vector, as documented in
  //  L4_C Product Spec Doc.
  // <JMG> 2014-10-01,    varFlt32_Obj_p->carbon_model_bitflag[] array
  //  is potentially getting corrupted sometime after this call
  //  where it is allocated. Note: Lucas l4c_PostProc9km_stepVarAlloc()
  //  function returns a container variable containing VarFloat32Adt
  //  that itself contains a number of very large allocated arrays - this is
  //  a very unusal idiom for how this is done (e.g. it (may?) be blowing the
  //  stack to have a returned variable be a container structure that
  //  contains large amounts of heap allocated memory..
  //  This (may?) be implicated in a transient FATAL SEGVIOLATION
  //  where carbon_model_bitflag[] array becomes NULL or corrupt

  ecoStateArrayObj_p->nee_qa = l4c_PostProc9km_stepVarAlloc("QA");

  ecoStateArrayObj_p->nee = l4c_PostProc9km_stepVarAlloc("STATE");
  ecoStateArrayObj_p->gpp = l4c_PostProc9km_stepVarAlloc("STATE");
  ecoStateArrayObj_p->rhtot = l4c_PostProc9km_stepVarAlloc("STATE");
  ecoStateArrayObj_p->Ctot = l4c_PostProc9km_stepVarAlloc("STATE");
  ecoStateArrayObj_p->Emult = l4c_PostProc9km_stepVarAlloc("STATE");
  ecoStateArrayObj_p->Tmult = l4c_PostProc9km_stepVarAlloc("STATE");
  ecoStateArrayObj_p->Wmult = l4c_PostProc9km_stepVarAlloc("STATE");

  // NOTE: name 'FrozenArea_9km' not sufficiently qualified, and should probably
  //  have followed the convention of "M09" vs "M09land" (1D-sparse vector) to
  //  distinghish this from the frozenArea_M09_2DFullGrid_p =
  //  memCalloc(EGv2_GLOBAL_9KM_PIXELS,sizeof(float32)); array declared in
  //  l4c_PostProc9km_func.c...
  ecoStateArrayObj_p->FrozenArea_9km =
      (float32 *)memCalloc(sesCtl_p->nElem9kmPerTask, sizeof(float32));
  memObjSetName(ecoStateArrayObj_p->FrozenArea_9km,
                "ecoStateArrayObj_p->FrozenArea_9km");

  // Provide explicit names for FT_ROLLING_xxx 3KM arrays allocated in
  // pre-processor
  memObjSetName(ecoStateArrayObj_p->ftRolling_3km,
                "ecoStateArrayObj_p->ftRolling_3km");
  memObjSetName(ecoStateArrayObj_p->ftRollingAge_3km,
                "ecoStateArrayObj_p->ftRollingAge_3km");

  // BIND ecoStateArrayObj suite of arrays to PkgMgr nodes
  l4c_BindPkg_EcoStateArrays(ecoStateArrayObj_p);

  // Note: prior 'loading' of variables done here is now performed by the Pre
  // Processor
  //   since these all generally require going through a DEFLATE transformation.

  // =========================================================================
  // PERFORM FORWARD MODEL STEP FOR SPECIFIED DAY (i.e. partial year forward
  // run)
  l4c_ModDriverGrid_forwardStep(sesCtl_p->runDayOfYear, pftObj_p, metArrayObj_p,
                                ecoStateArrayObj_p);

  // Free Pft nested 1km array (81 elements), not required after forward step
  memFree(pftObj_p->nested_1km);

  // Adjust properties of SOC to reflect their 2D full grid form,
  // just prior to INFLATE operation on these, prior to writing them back out
  l4c_PrepInflate_HDF5_SOC_Datasets();

  // =========================================================================
  // In GMAO ops, always enable SOC updates.
  // <JMG> 2014-09-27, added, option to switch off for devel  mode
  // diag and repetitive tests
  if (sesCtl_p->enableSOCUpdate)
    l4c_SOC_Update_Intermediates(ecoStateArrayObj_p);

  // =========================================================================
  // CALL POST-PROCESSOR from within l4c_DAILY_DailyDriver.c:l4c_DailyDriver()

  l4c_EventCompletionMsg((char *)__func__);

} // end :: l4c_Action_forwardStep()
