// ---------------------------------------------------------------
// function     : l4c_DAILY_ActionSpinup.c
// purpose      : Stores tcfAction.c logic for all non daily (spinup) steps
// author       : joe glassy, Lupine Logic Inc.
// version      : see L4C_VERS_TAG , as 4.1.5 or later
// revised      : v4.1.6-2014-02-05T11:242:00-07:00,wheat,jmg"
// dependencies : HDF4, HDF5 libraries and L4C FRAMEWORK API.
// history      : n/a
// notes        : Based on and adapted from tcf_spin v3.2, this
//                module contains setup logic for just the NON DAILY 
//                SPINUP (actions) 
// ---------------------------------------------------------------

// #include "tcf_inc.h" now re-factored into l4c_DAILY.h
#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"


// ----------------------------------------------------
// tcfAction_anaSpin()

// Purpose: Analytical preconditioning for SOC initialization and annual litterfall calculation.

void tcfAction_anaSpin(void)
{
SesCtl *sesCtl_p = l4c_GetSesCtl() ;
  int        status;

  uint32     nelem;

  char       *fpath = NULL;

  PftAdt                   PFT_ARRAY;
  MetArrayAdt             MetArrayObj;
  EcoStateArrayAdt           STATE_ARRAY;

  xLogMsg( "@A Task %s: DOING ANALYTICAL SOC PRECOND for GRID data...\n", sesCtl_p->subtaskIdStr);


  /*ALLOCATE INPUT ARRAYS*/
  MetArrayObj = tcfMem_metArrayAlloc(sesCtl_p->nElem9kmPerTask);

  // LOAD STATIC MET (Min/max for SMRZ rescaling)
  tcfMem_loadStaticMet(&MetArrayObj);

  PFT_ARRAY = tcfMem_pftArrayAllocInit(sesCtl_p->nElem1kmPerTask, sesCtl_p->elem1kmOffset);

  // ALLOCATE REQUIRED OUTPUT ARRAYS
  STATE_ARRAY = tcfMem_stateArrayAlloc(sesCtl_p->nElem1kmPerTask);
 
  // ALLOCATE OPTIONAL OUTPUT ARRAYS
  // NOTE: The only required field to save is C[NPOOL] ~ LDOM1KM, the remaining fields are just
  //  whatever you want to write out to see. State structures for doing model computations are available inside
  //  the respective driver functions.
  STATE_ARRAY.kmult_sum.dom_1km = (float32*)memCalloc(sesCtl_p->nElem1kmPerTask, sizeof(float32));
    
  // DO ANALYTICAL SOC PRECONDITIONING, ALSO RETURNS NPP SUM FOR LITTERFALL
  tcfModDriverGrid_anaPrecond(PFT_ARRAY, MetArrayObj, &STATE_ARRAY);
 
  // SAVE ANALYTIC RESULTS
  tcfMem_saveCstate("Cana", "", sesCtl_p->nElem1kmPerTask, STATE_ARRAY);

  nelem = sesCtl_p->nElem1kmPerTask;
  fpath = fpath_cat(3, GLOB_PARAMS.output_basepath, "tcf_npp_sum_M01land_task", sesCtl_p->subtaskIdStr);
  status = write_flt32(fpath, STATE_ARRAY.npp_sum.dom_1km, nelem);

  // Optional:
  nelem = sesCtl_p->nElem1kmPerTask;
  fpath = fpath_cat(3, GLOB_PARAMS.output_basepath, "tcf_kmult_sum_M01land_task", sesCtl_p->subtaskIdStr);
  status = write_flt32(fpath, STATE_ARRAY.kmult_sum.dom_1km, nelem);
  
  // FREE ALLOCATED MEMORY
  tcfMem_metArrayFree(&MetArrayObj);
  
  // Free input arrays:
  memFree(PFT_ARRAY.dom_1km);
  
  // Free output arrays:
  tcfMem_stateArrayFree( &STATE_ARRAY );

  memFree( STATE_ARRAY.kmult_sum.dom_1km );

} // end :: tcfAction_anaSpin()


// ---------------------------------------------------------------
// Function: tcfAction_numSpin()
// Purpose : perform Numerical spin-up for SOC initialization.  SOC pools always 
//           initialized to Jan. 1 state. 
// Inputs  : void
// Returns : n/a (void)
//
void tcfAction_numSpin(void)
{
  SesCtl *sesCtl_p = l4c_GetSesCtl() ;
  PftAdt              PFT_ARRAY;
  MetArrayAdt         MetArrayObj;
  EcoStateArrayAdt    STATE_ARRAY;
	
  int        status;
  uint32     nelem;
  char       *fpath = NULL;

  xLogMsg( "@A Task %s:  NUMERICAL SPIN-UP for GRID data...\n", sesCtl_p->subtaskIdStr);


  // LLOCATE INPUT ARRAYS
  MetArrayObj = tcfMem_metArrayAlloc(sesCtl_p->nElem9kmPerTask);

  // LOAD STATIC MET (Min/max for SMRZ rescaling)
  tcfMem_loadStaticMet(&MetArrayObj);

  PFT_ARRAY = tcfMem_pftArrayAllocInit(sesCtl_p->nElem1kmPerTask, sesCtl_p->elem1kmOffset);

  // ALLOCATE REQUIRED OUTPUT ARRAYS
  STATE_ARRAY = tcfMem_stateArrayAlloc(sesCtl_p->nElem1kmPerTask);
 
 
  // ALLOCATE OPTIONAL OUTPUT ARRAYS
  
  // NOTE: The only required field to save is C[NPOOL] ~ LDOM1KM, the remaining fields are just
  //  whatever you want to write out to see. State structures for doing model computations are available inside
  //  the respective driver functions.
  
  nelem = sesCtl_p->nElem1kmPerTask;
  STATE_ARRAY.nee_sum.dom_1km = (float32*)memCalloc(nelem, sizeof(float32));
  STATE_ARRAY.spinup_iter = (uint32*)memCalloc(nelem, sizeof(uint32));
  STATE_ARRAY.spinup_stop = (uint8*)memCalloc(nelem, sizeof(uint8));

  //  Load previous SOC and npp sums (for litterfall) to initialize
  tcfMem_loadCstate("Cana", "", &STATE_ARRAY);
   
  //  DO NUMERICAL SPINUP FOR SOC 
  tcfModDriverGrid_numSpin(PFT_ARRAY, MetArrayObj, &STATE_ARRAY);
   
  //  SAVE RESULTS
  tcfMem_saveCstate("Cnum", "", sesCtl_p->nElem1kmPerTask, STATE_ARRAY);

  /*Optional:*/
  nelem = sesCtl_p->nElem1kmPerTask;
  fpath = fpath_cat(3, GLOB_PARAMS.output_basepath, "tcf_nee_sum_M01land_task", sesCtl_p->subtaskIdStr);
  status = write_flt32(fpath, STATE_ARRAY.nee_sum.dom_1km, nelem);

  nelem = sesCtl_p->nElem1kmPerTask;
  fpath = fpath_cat(3, GLOB_PARAMS.output_basepath, "tcf_numspin_iter_M01land_task", sesCtl_p->subtaskIdStr);
  status = write_uint32(fpath, STATE_ARRAY.spinup_iter, nelem);
  
  
  // FREE ALLOCATED MEMORY
  
  //  Free input arrays
  tcfMem_metArrayFree(&MetArrayObj);
  free(PFT_ARRAY.dom_1km);
  
  // Free required output arrays
  tcfMem_stateArrayFree(&STATE_ARRAY);

  //  Free optional output arrays
  memFree(STATE_ARRAY.nee_sum.dom_1km);
  memFree(STATE_ARRAY.spinup_iter);
  memFree(STATE_ARRAY.spinup_stop);


} // end :: tcfAction_numSpin()


// ---------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// tcfAction_seasAdj()

// Purpose: Seasonal adjustment initializating SOC for a model step occuring for a specified day mid-year. 
//          Minimal state tracking.

void tcfAction_seasAdj(void)
{
SesCtl *sesCtl_p = l4c_GetSesCtl() ;
  uint32     nelem;

  char       doyStr[MXSTRLEN] = {0};

  PftAdt                   PFT_ARRAY;
  MetArrayAdt             MetArrayObj;
  EcoStateArrayAdt           STATE_ARRAY;


  xLogMsg( "@A Task %s: DOING FORWARD RUN for GRID data, day of year to stop: %d.\n",
	sesCtl_p->subtaskIdStr, sesCtl_p->dayOfYearToStop);


  /*ALLOCATE INPUT ARRAYS*/
  MetArrayObj = tcfMem_metArrayAlloc(sesCtl_p->nElem9kmPerTask);

  // LOAD STATIC MET (Min/max for SMRZ rescaling)
  tcfMem_loadStaticMet(&MetArrayObj);

  PFT_ARRAY = tcfMem_pftArrayAllocInit(sesCtl_p->nElem1kmPerTask, sesCtl_p->elem1kmOffset);

  /*ALLOCATE REQUIRED OUTPUT ARRAYS*/
  STATE_ARRAY = tcfMem_stateArrayAlloc(sesCtl_p->nElem1kmPerTask);
 
  /*ALLOCATE OPTIONAL OUTPUT ARRAYS*/
  /*NOTE: The only required field to save is C[NPOOL] ~ LDOM1KM, the remaining fields are just
    whatever you want to write out to see. State structures for doing model computations are available inside
    the respective driver functions.*/
  nelem = sesCtl_p->nElem1kmPerTask;
  STATE_ARRAY.nee_sum.dom_1km = (float32*)memCalloc(nelem, sizeof(float32));
  STATE_ARRAY.spinup_stop = (uint8*)memCalloc(nelem, sizeof(uint8)); //required as zeros to keep run going


  // Load previous SOC and npp sums (for litterfall) to initialize
  tcfMem_loadCstate("Cnum", "", &STATE_ARRAY);
   
  /* DO SPIN-UP SEASONAL ADJUSTMENT FOR SOC    
   (i.e. partial year forward run)           
  */
   
  tcfModDriverGrid_forwardSpin(sesCtl_p->dayOfYearToStop, 0, PFT_ARRAY, MetArrayObj, &STATE_ARRAY);

  // SAVE REQUIRED STATE RESULTS
  
  sprintf(sesCtl_p->doy_Str, "_day%-3.3d", sesCtl_p->dayOfYearToStop);
  tcfMem_saveCstate("C", sesCtl_p->doy_Str, sesCtl_p->nElem1kmPerTask, STATE_ARRAY);

  /*Optional:*/
  /*<nothing>*/
  
  /*FREE ALLOCATED MEMORY*/
  
  //  Free input arrays
  tcfMem_metArrayFree(&MetArrayObj);
  memFree(PFT_ARRAY.dom_1km);
  
  // Free required output arrays
  tcfMem_stateArrayFree(&STATE_ARRAY);

  // Free optional output arrays
  memFree(STATE_ARRAY.nee_sum.dom_1km);
  memFree(STATE_ARRAY.spinup_stop);

} // end :: tcfAction_seasAdj()


// ---------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// tcfAction_forwardRun()

// Purpose: Does forward run for one entire simulation year.  Output state reporting is left flexible and not
//          all PSD states tracked.
 

void tcfAction_forwardRun(void)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl() ;

  uint32     nelem;

  PftAdt                   PFT_ARRAY;
  MetArrayAdt             MetArrayObj;
  EcoStateArrayAdt           STATE_ARRAY;


  xLogMsg( "@A Task %s: DOING FORWARD RUN for GRID data.\n", sesCtl_p->subtaskIdStr);

  /*ALLOCATE INPUT ARRAYS*/
  MetArrayObj = tcfMem_metArrayAlloc(sesCtl_p->nElem9kmPerTask);

  // LOAD STATIC MET (Min/max for SMRZ rescaling)
  tcfMem_loadStaticMet(&MetArrayObj);

  PFT_ARRAY = tcfMem_pftArrayAllocInit(sesCtl_p->nElem1kmPerTask, sesCtl_p->elem1kmOffset);

  /*ALLOCATE REQUIRED OUTPUT ARRAYS*/
  STATE_ARRAY = tcfMem_stateArrayAlloc(sesCtl_p->nElem1kmPerTask);
 
  /*ALLOCATE OPTIONAL OUTPUT ARRAYS*/
  /*NOTE: The only required field to save is C[NPOOL] ~ LDOM1KM, the remaining fields are just
    whatever you want to write out to see. State structures for doing model computations are available inside
    the respective driver functions.*/
  nelem = sesCtl_p->nElem1kmPerTask;
  
  STATE_ARRAY.nee.dom_1km   = (float32*)memCalloc(nelem, sizeof(float32));
  STATE_ARRAY.gpp.dom_1km   = (float32*)memCalloc(nelem, sizeof(float32));
  STATE_ARRAY.rhtot.dom_1km = (float32*)memCalloc(nelem, sizeof(float32));

  STATE_ARRAY.nee_sum.dom_1km = (float32*)memCalloc(nelem, sizeof(float32));
  STATE_ARRAY.gpp_sum.dom_1km = (float32*)memCalloc(nelem, sizeof(float32));
  STATE_ARRAY.rhtot_sum.dom_1km = (float32*)memCalloc(nelem, sizeof(float32));

  /*Load previous SOC and npp sums (for litterfall) to initialize*/
  tcfMem_loadCstate("Cnum", "", &STATE_ARRAY);

  // --------------------------------------------------------------  
  // DO FULL YEAR SIMULATIION, (i.e. partial year forward run)     
  tcfModDriverGrid_forwardRun(PFT_ARRAY, MetArrayObj, &STATE_ARRAY);
  // --------------------------------------------------------------   

  // SAVE REQUIRED STATE RESULTS
  // verify if this is finally converted right..
  sprintf(Cname,"%s%3d%s","Cdoy",SesCtl_p->dayOfYearToStop,"p");
  save_C_state(Cname, sesCtl_p->nElem1kmPerTask, STATE_ARRAY); 

  // Optional:  <nothing>
  
  // FREE ALLOCATED MEMORY
  
  // Free input arrays
  tcfMem_metArrayFree(&MetArrayObj);
  memFree(PFT_ARRAY.dom_1km);
  
  // Free required output arrays
  tcfMem_stateArrayFree(&STATE_ARRAY);

  // Free optional output arrays
  memFree(STATE_ARRAY.nee.dom_1km);
  memFree(STATE_ARRAY.gpp.dom_1km);
  memFree(STATE_ARRAY.rhtot.dom_1km);

  memFree(STATE_ARRAY.nee_sum.dom_1km);
  memFree(STATE_ARRAY.gpp_sum.dom_1km);
  memFree(STATE_ARRAY.rhtot_sum.dom_1km);

} // end :: tcfAction_forwardRun()
