// ----------------------------------------------------------------
// module       : l4c_DAILY_BindToPkgMgr.c
// purpose      : Implement all discrete PkgMgr "BIND" actions for the
//                internal tcf_spin model logic, where
//                suites of pre-existing arrays are BOUND to their
//                predefined PkgMgr --> PkgTheme --> PkgObj nodes
//                Included here thus far are:
//                  EcoStateArrayAdt
//                  MetArrayAdt
// revision     : 2014-08-07 <JMG> removed surflag_3km[] replaced with
// ft_state_um_3km[] revision     : 2014-03-11T13:22, changed FPAR_8DAY to use
// underscore.. history      : 2014-02-28T13:29, corrected bug where the binding
// of
//                "fpar" was omitted, and mistakenly connected to ndvi
// author       : Joe Glassy
// project      : SMAP L4C science development.
// dependencies : tcfModDriverGrid.c, tcfMem.c
// revision     : Initial reformatting -- laj 09/09/13
// notes        : Reminder: a "bind" here refers to how an data structure
//                (often a heap array) is associated with a particular
//                Package Manager's "theme" (PkgTheme) child instance
// ----------------------------------------------------------------

#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"

// ---------------------------------------------------------------
// function : l4c_BindPkg_Ancillary()
// purpose  : Bind various SMAP_L4_C_Ancillary.h5 arrays
// inputs   :
// notes    : TBD, may need to ADD 'latitude','longitude' and
//            M09land_lc_mask_uint8 to this here..
// notes    : x pixels per 1km 1D-sparse vector
// notes    : 1664040 pixels per 9km 1D-sparse vector
//
int l4c_BindPkg_AncillaryStatic(AncImagePkgAdt *ancImagePkgObj_p,
                                MetArrayAdt *metArrayObj_p) {
  SesCtl *sesCtl_p = l4c_GetSesCtl();
  // declare the PkgMgr mechanism parts we need
  PkgMgr *pkgMgr_p = NULL;
  PkgTheme *pThme_p = NULL;

  // PERFORM PkgMgr tasks
  pkgMgr_p = sesCtl_p->l4cPkgMgr_p;

  // The smrz_mn, smrz_mx are store in "SMAP_L4_C_Ancillary" since they
  // are static layers, but their "core" model variables are members of
  // metArrayAdt
  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "SMAP_L4_C_Ancillary");

  pkgMgr_BindObjByName(pThme_p, "MCD12Q1_M01land_lc_dom_uint8",
                       ancImagePkgObj_p->imgPftM01land_dom_p);

  // <JMG> 2014-05-16, added for consistency and to enable use of
  pkgMgr_BindObjByName(pThme_p, "L4C_M01land_PixFilter_uint8",
                       ancImagePkgObj_p->imgM01land_PixFilter_p);

  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "MetArrayAdt");
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.smrz_mn", metArrayObj_p->smrz_mn);
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.smrz_mx", metArrayObj_p->smrz_mx);

  return 0;

} // end::l4c_BindPkg_Ancillary()

// ---------------------------------------------------------------
// function : l4c_BindPkg_SOC_Intermediate()
// purpose  : Bind various SMAP_L4_C_SOC.h5 arrays
// inputs   :
// revised  : 2015-05-27T16:34, <JMG> changed to use RIP dynamic lookup on
//            smrz min,max sat scaling layers
// notes    : x pixels per 1km 1D-sparse vector
// notes    : 1664040 pixels per 9km 1D-sparse vector
//
int l4c_BindPkg_SOC_Intermediate(AncImagePkgAdt *ancImagePkgObj_p,
                                 MetArrayAdt *metArrayObj_p) {
  SesCtl *sesCtl_p = l4c_GetSesCtl();
  NsbObj *nsbObj_p = l4c_GetNsbOverseer();
  // declare the PkgMgr mechanism parts we need
  PkgMgr *pkgMgr_p = NULL;
  PkgTheme *pThme_p = NULL;

  char nvpBuf[FILENAME_MAX] = {0}; // reusable buffer for NSB "Get" retrievals
  char *dsetName_p = NULL;         // future: uses nsbObj_p to supply all these!

  // PERFORM PkgMgr tasks
  pkgMgr_p = sesCtl_p->l4cPkgMgr_p;

  // The smrz_mn, smrz_mx are store in "SMAP_L4_C_Ancillary" since they
  // are static layers, but their "core" model variables are members of
  // metArrayAdt
  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "SMAP_L4_C_SOC");

  // <JMG> 2015-05-27 TBD, avoid hard coding, read from RIP file instead!!
  dsetName_p = nsbGetValueByKey(
      nsbObj_p, "L4C_INPUT_DATASET_SOILROOTZONE_M09_MIN", nvpBuf);
  pkgMgr_BindObjByName(pThme_p, dsetName_p, metArrayObj_p->smrz_mn);

  dsetName_p = nsbGetValueByKey(
      nsbObj_p, "L4C_INPUT_DATASET_SOILROOTZONE_M09_MAX", nvpBuf);
  pkgMgr_BindObjByName(pThme_p, dsetName_p, metArrayObj_p->smrz_mx);

  return 0;

} // end::l4c_BindPkg_SOC_Intermediate()

// ---------------------------------------------------------------
// function : l4c_BindPkg_EcoStateArrays()
// purpose  : Bind EcoStateArrayAdt arrays to PkgMgr
// inputs   : EcoStateArrayAdt ecoStateArrayObj
// notes    : TBD, may need to ADD 'latitude','longitude' and
//            M09land_lc_mask_uint8 to this here..
// notes    : x pixels per 1km 1D-sparse vector
// notes    : 1664040 pixels per 9km 1D-sparse vector
//
int l4c_BindPkg_EcoStateArrays(EcoStateArrayAdt *ecoStateArrayObj_p) {
  SesCtl *sesCtl_p = l4c_GetSesCtl();

  // declare the PkgMgr mechanism parts we need
  PkgMgr *pkgMgr_p = NULL;
  PkgTheme *pThme_p = NULL;

  // PERFORM PkgMgr tasks
  pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "EcoStateArrayAdt");

  // 2) BIND each array variable that has just been allocated
  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj_p->nee",
                       ecoStateArrayObj_p->nee.dom_1km);
  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj_p->gpp",
                       ecoStateArrayObj_p->gpp.dom_1km);
  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj_p->npp",
                       ecoStateArrayObj_p->npp.dom_1km);

  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj_p->rhtot",
                       ecoStateArrayObj_p->rhtot.dom_1km);
  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj_p->Ctot",
                       ecoStateArrayObj_p->Ctot.dom_1km);

  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj_p->Emult",
                       ecoStateArrayObj_p->Emult.dom_1km);
  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj_p->Tmult",
                       ecoStateArrayObj_p->Tmult.dom_1km);
  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj_p->Wmult",
                       ecoStateArrayObj_p->Wmult.dom_1km);

  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj_p->FrozenArea_9km",
                       ecoStateArrayObj_p->FrozenArea_9km);
  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj_p->ftRolling_3km",
                       ecoStateArrayObj_p->ftRolling_3km);
  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj_p->ftRollingAge_3km",
                       ecoStateArrayObj_p->ftRollingAge_3km);

  // <LAJ> Need new Qa binds?

  // verify where reco is allocated?
  // pkgMgr_BindObjByName(pThme_p,"ecoStateArrayObj_p->reco" ,
  // ecoStateArrayObj_p->reco.dom_1km);

  return 0;

} // end:: l4c_BindPkg_EcoStateArrays()

// ---------------------------------------------------------------
// function : lc4c_BindPkg_SOC_and_Npp()
// purpose  : Bind EcoStateArrayAdt SOC and npp_sum
// inputs   : EcoStateArrayAdt ecoStateArrayObj
// notes    : This is done separately here to match the way
//            the corresponding allocator, tcfMem_stateArrayAlloc()
//            is setup
//
// notes    : All entries BOUND here MUST have had their names first
//            registered (earlier) using a call in l4c_DAILY_PkgMgrInit.c like:
//            pkgMgr_SetThemes_ChildName(pkgTheme_p,41,"ecoStateArrayObj.ftRollingAge_3km");
//
// notes    : Remember: just because we bind up an existing variable does
//            NOT mean we are necessarily using it. This is just a
//            catalog or dictionary, e.g.index.
//
int l4c_BindPkg_SOC_and_Npp(EcoStateArrayAdt *ecoStateArrayObj_p) {
  SesCtl *sesCtl_p = l4c_GetSesCtl();

  // set pointer to primary PkgMgr
  PkgMgr *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
  PkgTheme *pThme_p = NULL; // ptr to a PkgTheme

  // PERFORM PkgMgr tasks: BIND this batch
  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "EcoStateArrayAdt");

  // bind up NPP SUM
  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj_p->npp_sum",
                       ecoStateArrayObj_p->npp_sum.dom_1km);
  // bind up the (3) SOC pools
  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj.C0",
                       ecoStateArrayObj_p->C[0].dom_1km);
  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj.C1",
                       ecoStateArrayObj_p->C[1].dom_1km);
  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj.C2",
                       ecoStateArrayObj_p->C[2].dom_1km);
  // also bind up the (3) Freeze Thaw layers
  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj.ftRolling_3km",
                       ecoStateArrayObj_p->ftRolling_3km);
  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj.ftRollingAge_3km",
                       ecoStateArrayObj_p->ftRollingAge_3km);

  // <JMG> 2014-08-07, re-verify if this is still used and deprecate if not
  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj.ftRollingQA_3km",
                       ecoStateArrayObj_p->ftRollingQA_3km);

  // don't forget to bind up the frozen area percent vector too!
  pkgMgr_BindObjByName(pThme_p, "ecoStateArrayObj.FrozenArea_9km",
                       ecoStateArrayObj_p->FrozenArea_9km);

  return 0;

} // end:: l4c_BindPkg_SOC_and_Npp()

// ---------------------------------------------------------------
// function : l4c_BindPkg_MetArrays(()
// purpose  : Bind MET MetArrayAdt arrays to PkgMgr
// inputs   : MetArrayAdt metArrayObj
// notes    : 1664040 pixels per 9km 1D-sparse vector
// revised  : 2014-05-16T13:30, viirsNdvi_qc added...
int l4c_BindPkg_MetArrays(MetArrayAdt *metArrayObj_p) {
  SesCtl *sesCtl_p = l4c_GetSesCtl();

  // set pointer to primary PkgMgr
  PkgMgr *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
  PkgTheme *pThme_p = NULL; // ptr to a PkgTheme

  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "MetArrayAdt");

  // 2) BIND each metArrayAdt (instance: metArrayObj) array variable
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.viirsNdvi",
                       metArrayObj_p->viirsNdvi);
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.viirsNdvi_qc",
                       metArrayObj_p->viirsNdvi_qc);

  // <JMG> 2/28/2014, corrected bug where 'fpar' omitted and bound to virrsNdvi
  // by mistake. TBD: yet-to-change, ALL 'fpar_qa' to 'fpar_qc' to match MODIS
  // HDF4 usage after all dependencies are documented...
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.fpar", metArrayObj_p->fpar);
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.fpar_qa", metArrayObj_p->fpar_qa);

  // <JMG> 04/29/2014, corrected to match logic where new fpar_clim added
  // <LAJ> 5/4/2014, moved line to here from bottom of the function
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.fpar_clim",
                       metArrayObj_p->fpar_clim);

  // <JMG> 4/29/2014, NOTE: surflag_3km is now used in a READ-ONLY context, and
  // is
  //    NOT written out as a L4_C 9km output variable
  // <JMG> 2014-08-07, replaced surflag_3km[] with ft_state_um_3km[] array
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.ft_state_um_3km",
                       metArrayObj_p->ft_state_um_3km);

  // <JMG> added 'par' binding, 2014-02-28T13:53, was omitted earlier
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.par", metArrayObj_p->par);

  pkgMgr_BindObjByName(pThme_p, "metArrayObj.tmin", metArrayObj_p->tmin);
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.tmax", metArrayObj_p->tmax);
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.vpd", metArrayObj_p->vpd);
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.tsoil", metArrayObj_p->tsoil);
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.smtop", metArrayObj_p->smtop);
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.smrz", metArrayObj_p->smrz);
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.smrz_mn", metArrayObj_p->smrz_mn);
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.smrz_mx", metArrayObj_p->smrz_mx);

  // <JMG> 2/28/2014, corrected-- qv2m was originally misspelled qvwm
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.qv2wm", metArrayObj_p->qv2m);
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.t2m_avg", metArrayObj_p->t2m_avg);
  pkgMgr_BindObjByName(pThme_p, "metArrayObj.ps_avg", metArrayObj_p->ps_avg);

  return 0;

} // end:: l4c_BindPkg_MetArrays();

// ---------------------------------------------------------------
// function : l4c_BindPkg_PostProc9km()
// purpose  : Bind PostProc 9km (tcf_proc9k) related variables
// inputs   : void
// notes    : Review this theme to either streamline to re-dedicate it,
//            or deprecate it if its role has been superceded elsewhere.
int l4c_BindPkg_PostProc9km(void) {
  SesCtl *sesCtl_p = l4c_GetSesCtl();
  // declare the PkgMgr mechanism parts we need
  PkgMgr *pkgMgr_p = NULL;
  PkgTheme *pThme_p = NULL;

  // PERFORM PkgMgr tasks
  pkgMgr_p = sesCtl_p->l4cPkgMgr_p;

  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "POST_PROC_INFLATE_9KM");

  // Future Change suggestion: these operations are currently being performed
  //  in the Post Processor although for consistency and readabililty,
  //  these 'bind' ops should also probably be included here

  return 0;

} // end:: l4c_BindPkg_PostProc9km()
