// ---------------------------------------------------------------
// module       : l4c_DAILY_DailyDriver.c
// purpose      : Serves as the entry driver for all (pre) process grid
//                transformations for l4c_DAILY.ex. TBD: may use for
//                post-proc. This module includes
//                all general grid support logic that is not directly
//                involved in the actual dataset-level Inflate|Deflate
//                transformations, whose logic is in:
//                  l4c_GridPre_Deflate.c
//                  l4c_Gridost_Inflate.c
// author       : joe glassy, Lupine Logic Inc.
// version      : see L4C_VERS_TAG , as 4.1.5 or later
// revised      : 2016-02-17, updated more module names to use prefix
// l4c_DAILY_xxx revised      : 2014-09-04, renamed to l4c_DAILY_DailyDriver.c
// to
//                to distinghish this from future l4c_DAILY_AnnualDriver.c
//                pending.
// history      : 2014-07-23T120:00:00, modified to do aggregation of 1km to 9km
// within forward step, laj history      : 2014-04-28, incremented to 20 the
// count of layers to DEFLATE
//                to include the (2) SMRZ (MERRA) layers in Ancillary and allow
//                READ only of surface_flag
// revised      : 2014-02-25T15:25:00,jmg
// dependencies : HDF4, HDF5 libraries and L4C FRAMEWORK API.
// history      : n/a
// notes        : Based on and adapted from mkgrid spland.c logic
//
// Development Technique Note: when logic is tested to a certain
//  incremental point, I place a call to mark the HIGH WATER MARK of
//  tested logic development
//  runtime build & run test HIGH WATER MARK
//		l4c_EventHighWaterMarkMsg("@A ~l4c_GridManager INTENTIONAL HWM
//SHUT DOWN\n");

// ---------------------------------------------------------------

#include "l4c_DAILY.h"
// orig was tcf_struct.h
#include "l4c_DAILY_tcf.h"

#include "l4c_DAILY_prot.h"

// ---------------------------------------------------------------
// function     : l4c_DailyDriver()  (prior, was l4c_GridManager)
// purpose      : l4c_DailyDriver() serves as the high level entry point
//                for all L4C daily model processes, as used in GMAO production.
// inputs       : const int runDay -- index of day currently being run
//                int appProcessingPhase  {DoPreProcessing,DoPostProcessing}
// returns      :
// author       : joe glassy, Lupine Logic Inc.
// revised      : 2014-11-03, for future multi-day runs, added enclosing
//                day loop. Note this primitive multi-day logic will need
//                refinement on the memory-mgmt, to take advantage of
//                arrays we can allocate once only and re-use from day-to-day
// revised      : 2014-09-16, renamed to l4c_DailyDriver() from
// l4c_GridManager() revised      : 2014-04-28T17:59, allow READ-ONLY of
// surface_flag,
//                add (2) SMRZ layers to list of target
//                variables to DEFLATE!
// history      : 2014-03-02, refactored all the pre-processor specific
//                support code into l4c_PreProc1km_Driver.c to match the
//                organization of post processor l4c_PostProc9km_Driver.c
// history      : n/a
// warning      : when multi-day mode is enabled this WILL need refinement
//                to optimize memory mgmt and coordination of pre and post
//                processing operations relevant to multi-day mode.
// notes        : The DEFLATE sequence is summarized:
//                a) construct a list of HDF5 dataset names to be DEFLATED,
//                   along with their key properties, storing these in
//                   PkgTheme's collection of PkgObj's, where the Theme's
//                   name corresponds to the HDF5 they are sourced from.
//                   By convention, we store the HDF5 file's name in
//                   PkgObj using frm_PkgMgr_SetObjGenericPath( path ).
//                b) create a re-usable 1KM M01 'full grid' raster. so
//    that as each native 2D layer is read, it is read into this.
//                c) For each INPUT layer, we will create a 1D-sparse vector
//                   array and direct-attach is the variable in
//                   initially empty MetArrayObj_p or EcoStateObj_p 's.
//                c) Implement a flow where for each dataset, the 2D
//                   (native) instance of the data is read into the common
//                   2D buffer
int l4c_DailyDriver(const int runDay, int appProcessingPhase) {
  SesCtl *sesCtl_p = l4c_GetSesCtl();
  // fetch pointer to the "core" model 1D-sparse vector instance of these
  // objects
  EcoStateArrayAdt *ecoStateArrayObj_p = l4c_GetEcoStateArrayAdt();
  MetArrayAdt *metArrayObj_p = l4c_GetMetArrayAdt();
  char fxName[256] = {0};

  // <JMG> 2014-05-16, added ndvi, ndvi_qc (verify total now as 23 ?)
  // <JMG> 2014-03-03, added 2 SMRZ NatureRun static layers
  // Remember: only a subset (23) of the (29) members of the PRE_PROC theme are
  // actually image layers to deflate.
  // TBD: Ideally the other elements should be
  // stored in their own theme..
  int nDatasets_to_Deflate = 23;

  strcpy(fxName, __func__);

  // branch on if Pre or Post Processing
  if (appProcessingPhase == DoPreProcessing) {
    l4c_PreProc_Setup(sesCtl_p->l4cPkgMgr_p, nDatasets_to_Deflate);
    // ===============================================================
    // call the DEFLATE driver to iterate over N=(18) datasets
    l4c_PreProc_ByLayerDriver(nDatasets_to_Deflate);

    // ALLOCATE and CALCULATE VPD ..in 1D-sparse vector space
    l4c_PreProc_Calculate_VPD(metArrayObj_p);
    // end:: block if DoPostProcessing

  } else if (appProcessingPhase == DoPostProcessing) {

    // =========================================================================
    // CALL POST-PROCESSOR, feeding all 1D sparse vector results from tcf_spin
    // to tcf_proc9km

    // Iterate through list of output 9KM 1D sparse vectors
    //    to perform the INFLATE action on, then write out results to file
    l4c_PostProc9km_driverConvert(ecoStateArrayObj_p);

    tcfMem_metArrayFree(metArrayObj_p);
    // TBD: rename to reflect exactly which subset of arrays are released
    tcfMem_stateArrayFree(ecoStateArrayObj_p);

    // Free other l4c_DAILY primary OUTPUT arrays
    // <LAJ> 7/22/14 : Standard output state variable arrays freed here.

    l4c_PostProc9km_stepVarFree("STATE", &ecoStateArrayObj_p->nee);
    l4c_PostProc9km_stepVarFree("QA", &ecoStateArrayObj_p->nee_qa);
    l4c_PostProc9km_stepVarFree("STATE", &ecoStateArrayObj_p->gpp);
    l4c_PostProc9km_stepVarFree("STATE", &ecoStateArrayObj_p->rhtot);
    l4c_PostProc9km_stepVarFree("STATE", &ecoStateArrayObj_p->Ctot);
    l4c_PostProc9km_stepVarFree("STATE", &ecoStateArrayObj_p->Emult);
    l4c_PostProc9km_stepVarFree("STATE", &ecoStateArrayObj_p->Tmult);
    l4c_PostProc9km_stepVarFree("STATE", &ecoStateArrayObj_p->Wmult);

    // Free FT-specific output arrays
    memFree(ecoStateArrayObj_p->FrozenArea_9km);
    memFree(ecoStateArrayObj_p->ftRolling_3km);
    memFree(ecoStateArrayObj_p->ftRollingAge_3km);

    // <KAE> 2024-01-11 Free the litterfall allocation rate
    // memFree(ecoStateArrayObj_p->litter_rate);
    memObjFree("litter_rate", ecoStateArrayObj_p->litter_rate);

  } // end:: block if DoPostProcessing

  l4c_EventCompletionMsg(__func__);

  return 0;

} // end:: l4c_DailyDriver(void)
