// ---------------------------------------------------------------------
// module       : l4c_DAILY_DatasetProperties.c
// purpose      : Implement definition of all HDF5 dataset properties
// project      : SMAP L4C science development.
// dependencies : HDF4 4.2.x, HDF5 v1.8.x or later libraries, L4C FRAMEWORK APIs
// revised      : 2016-02-18, changed all EC names,  *mult_av to mult_mean
// revised      : 2014-11-14T16:08, mods to support new QA RMSE variables
// history      : 2014-09-03T12:34, split out from l4c_DAILY_hdf5.c
// author       : jglassy
// notes        : Namespace signature: :"l4c_Hdf5_XXXXX"
// ---------------------------------------------------------------------

#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"


// ---------------------------------------------------------------------
// function     : l4c_AssignOutputDatasetProperties()
// purpose      : Retrieves dataset properties from the PkgMgr theme "L4C_OUTPUTS"
//                in order to store these properties as per PSD in 
//                theme "L4C_OUTPUTS" child objects named after the HDF5 dataset
//                Note that while default values are provided, most properties
//                are overridden by (keyword) parameters in the RIP file when
//                appropriate keywords are enabled.
// inputs       : void (tbd)
// dependencies : HDF4 4.2.x, HDF5 v1.8.x or later libraries, L4C FRAMEWORK APIs
// revised      : 2014-09-10T12:34, split out from l4c_DAILY_hdf5.c
// author       : jglassy
// notes        : CALLED BY: l4c_DAILY_Hdf5.c in int l4c_Hdf5_CreateDatasets(hid_t file_id)
//                at approx line 632
// notes        : This code assigns values into just the in-memory instance of
//                the PkgMgr child (PkgObj) instance associated with each HDF5 output
//                dataset. The datasets themselves are written out in a separate step later
// notes        : Any new set of variables, like RMSE, must be defined in the
//                PkgMgr theme named "L4C_OUTPUTS", and be placed within the
//                first ("NEE/nee_mean") and last ("GEO/longitude") named PkgObj entries
//                in order for them to be automatically included in these
//                transfer operations.
//
// example      : A typical fully annotated dataset property attribute, shown here
//                using RIP parameter syntax is:
//              : # L4C_MET_OUTPUT_DATASET_name =
//               {GROUP}/{DsName},{datatype:float32},{unitstring},{_FillValue},
//               {valid_min},{valid_max},{rank=2},{dim1,dim2}
//
// The following ENUM constant symbols may be used to identifty these dataset properties:
//
// IsPropertyName,
// IsPropertyDatatype, 
// IsPropertyUnits,
// IsPropertyFillvalue,
// IsPropertyValidMin,
// IsPropertyValidMax,
// IsPropertyLongname
//
//  NOTE: we supply 6 (of 8) tokens at minimum in a valid RIP statement that lists 
//      dataset properties. If "rank,dims" at end are omitted, defaults are supplied
// L4C_MET_OUTPUT_DATASET_QV2M_M09_AVG           = /QV2M_M09_AVG,float32,Kg/Kg,-9999.0,min,max,longname
// L4C_MET_OUTPUT_DATASET_RADIATION_SHORTWAVE_DOWNWARD_FLUX = /RADIATION_SHORTWAVE_DOWNWARD_FLUX,float32,W/m^2,-9999.0,min,max,longname
// L4C_MET_OUTPUT_DATASET_SM_ROOTZONE_WETNESS    = /SM_ROOTZONE_WETNESS,float32,dimensionless,-9999.0,min,max,longname

// ---------------------------------------------------------------------
int l4c_AssignOutputDatasetProperties(void)
{
 SesCtl          *sesCtl_p = l4c_GetSesCtl();
 Hdf5GridPropAdt *hGridProp_p = l4c_GetHdf5GridProp();
 NsbObj   *nsb_p    = l4c_GetNsbOverseer();
 PkgMgr   *pkgMgr_p = l4c_GetPkgMgrOverseer();
 PkgTheme *pThme_p  = NULL;
 PkgObj   *pObj_p   = NULL;

 char *datasetName_p = NULL;
 char fxName[256] = {0} ;
 char shortnameBuffer[256] = {0} ;
 char longnameBuffer[1024] = {0};
 
 char valueStr[FILENAME_MAX] = {0};
 char nvpBuf[FILENAME_MAX]= {0} ;
 char outVarPropNvpKey[FILENAME_MAX] = {0}; // L4_C_OUTPUT_VARIABLE_{name} keyword...
 
 char tivTypeAsStr[32] = {0};
 // names of first and last _mdl_ OUTPUT DATASETS in the PkgMgr database 
 char firstOutEntry[256] = {0}; // name of FIRST _mdl_ GRANULE output dataset name
 char lastOutEntry[256]  = {0}; // name of LAST _mdl_ GRANULE output dataset name
 char *currentKeyword_p = NULL;
 
 int  tivDatatype = 0 ; 
 int tokenCount = 0 ;
 int noOp =  0, idx =0;
 int startIdx = -1;
 int endingIdx= -1;
 int nOutputVariables = 0 ;
 int sizOfOne =0;
 
 hid_t dset_id = -1 ;
 hid_t curr_id = -1 ;
 
 size_t dimsAry9km[2] = {0} ;
 
 float64 dblValu = 0.0,
         validRange =0.0 ,
         validMinimum =0.0,
         validMaximum =0.0;
 float32 fltValu = 0.0;
 
 
    strcpy(fxName,__func__);
    
    dimsAry9km[0] = EGv2_GLOBAL_9KM_NS_PIXELS ;
    dimsAry9km[1] = EGv2_GLOBAL_9KM_EW_PIXELS ;

    pThme_p  = pkgMgr_GetThemeByName(pkgMgr_p,"L4C_OUTPUTS");
    
    xLogMsg("@A In(%s) using RIP values to assign dataset properties to PkgMgr children~\n",
            fxName);

    // ASSERT: this operation MUST be performed AFTER the dataset's have
    // been created, but PRIOR to when we write out the dataset attributes.
    
    // TBD: Consider if more robust to fetch direct from RIP, in case their
    //  name changes later
    //  Note that this approach here does NOT yet really guanantee these are the
    //  first and last, but just that they exist in the collection.
    //  There isn't a perfect solution for all cases... Keep refining this.
    strcpy(firstOutEntry,"NEE/nee_mean");
    strcpy(lastOutEntry ,"GEO/longitude");
    
    // First verify that the FIRST and LAST entries actually exist as indicated
    // Next, establish the starting and ending base0 indices in the OUTPUT PkgMgr theme
    // to iterate through

    if(pkgMgr_DoesThemes_ChildNameExist(pThme_p,firstOutEntry))
    {
        startIdx  = pkgMgr_GetThemes_ChildIndexByName(pThme_p,firstOutEntry) ;
    } else {
        xLogMsg("@F(%s) Required FIRST PkgMgr output dataset(%s) definition does NOT exist! ~\n",
            fxName, firstOutEntry);
    }

    if(pkgMgr_DoesThemes_ChildNameExist(pThme_p, lastOutEntry))
    {
        endingIdx = pkgMgr_GetThemes_ChildIndexByName(pThme_p,lastOutEntry);
    } else {
        xLogMsg("@F(%s) Required LAST PkgMgr output dataset(%s) definition does NOT exist! ~\n",
            fxName, lastOutEntry);
    }
    
    nOutputVariables  = (endingIdx - startIdx)+1;

    // xLogMsg("@A In(%s) L4C_OUTPUTS Theme PkgObj startIdx %d endingIdx %d~ \n\n",fxName, startIdx,endingIdx);

    // Iterate in PkgMgr Theme "L4C_OUTPUTS", through list of all output variables
    for(idx=startIdx; idx <= endingIdx; idx++)
    {
        // Fetch the PkgObj (detailed) container object for THIS HDF5 output variable
        pObj_p = pkgMgr_GetObjByIndex(pThme_p,idx);
        datasetName_p = pkgMgr_GetObjName(pObj_p);
        
        // REMOVED diagnostic   xLogMsg("@A In(%s) Preparing dataset (%s)~\n",fxName, datasetName_p);
        
        // SKIP over any deprecated dataset entries!
        // <JMG> strstr() call  below was missing a trailing ")" symbol!!
        if(strstr(datasetName_p,"carbon_model_qual_pft") != NULL ||
           strstr(datasetName_p,"SMAP_L4_C_PKGOBJDEFAULTKEY") != NULL)
        {
            // xLogMsg("@W In(%s) SKIPPING DEPRECATED dataset (%s)~\n",fxName, datasetName_p);
            continue;
        }
                        
        // Fetch name (label) as original RIP name-value-pair KEYWORD
        // example: "NEE/nee_mean"
        currentKeyword_p = pkgMgr_GetObjName(pObj_p);
        
        // First assign DEFAULT properties, based on greatest-common-denom approach
        // This assures the attribute names are at least present even if their
        // values are incorrect. We use common defaults first, to be selectively overridden later
        
        pkgMgr_SetObjOwner(pObj_p, -1 ); // set sentinel, the modify to file_id as soon as known
        
        // pkgMgr_SetObjId(pObj_p,    -1 ); // set to dataset, or group, or attribute ID when known
        curr_id = pkgMgr_GetObjId(pObj_p);
        
        pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->output_L4C_DAILY_Path);
        
        // once defined and read from RIP this work around is no longer needed or used
        pkgMgr_SetObjMetadata(pObj_p, "GEO/latitude GEO/longitude");

        pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32 );

        // fetch TIV datatype string equivalent like "float32" given DFNT_FLOAT32
        strcpy(tivTypeAsStr, tiv_NameofTivType(DFNT_FLOAT32,&sizOfOne));
        
        pkgMgr_SetObjResolution(pObj_p, 9 );    // 9KM
        pkgMgr_SetObjRank(pObj_p, 2 );          // 2D
        pkgMgr_SetObjDims(pObj_p, dimsAry9km);  // load grid dimensions 1624 x 3856
        
        // Note: this is -9999.0 usually
        pkgMgr_SetObjFillvalue(pObj_p , (float32)FLOAT_FILL_VALUE);
        
        pkgMgr_SetObjUnits(pObj_p     ,"NO-UNITS-SPECIFIED-YET");
        pkgMgr_SetObjValidMin(pObj_p  , 0.0);
        pkgMgr_SetObjValidMax(pObj_p  , 0.0);
        pkgMgr_SetObjValidRange(pObj_p, 0.0,0.0);

        // SHORTNAME (shortname) assignment:
        //   if from a 'group-qualified' HDF5 dataset name like "NEE/nee_mean" ,
        //   extract just the "nee_mean" part
        if(strchr(currentKeyword_p,'/') != NULL)
            frm_strGetToken(shortnameBuffer,currentKeyword_p,1,"/");
        else 
            // otherwise, pass a non-group-qualified datasetname through
            strcpy(shortnameBuffer,currentKeyword_p); 
        
        // By default, set longname the same as shortname, in case a separate longname
        // is not yet defined
        pkgMgr_SetObjShortname(pObj_p, shortnameBuffer);
        pkgMgr_SetObjNelem(pObj_p    , EGv2_GLOBAL_9KM_PIXELS);     

        // pkgMgr_SetObjH5DatasetId(pObj_p,-1); 
        pkgMgr_SetObjH5Datatype(pObj_p, H5T_NATIVE_FLOAT);
        // TBD may need to branch on datatype to select correct dcpl and dspace here
        pkgMgr_SetObjH5Dcpl(pObj_p    , hGridProp_p->dcpl_9km );
        pkgMgr_SetObjH5Dspace(pObj_p  , hGridProp_p->dspace_9km);

        // First, using simple HDF5 dataset name as the base, 
        //  construct the L4C_OUTPUT_VARIABLE_{name} nvp keyword, remembering to strip off
        //  the initial {GROUP}/ part, so use prior assigned shortname which is "nee_mean" not "NEE/nee_mean"
        strcpy(valueStr,shortnameBuffer); // protect orig shortname, don't uppercase it..

        // The resulting query key needed to fetch equiv RIP name value pair 
        // must resemble a final form like valueStr is NEE_MEAN, from "L4C_OUTPUT_VARIABLE_NEE_MEAN"
        sprintf(outVarPropNvpKey,"L4C_OUTPUT_VARIABLE_%s",frm_strToUpper(valueStr));

        // For each given variable, search for its keyword in the RIP's NSB database
        // where outVarPropNvpKey is like "L4C_OUTPUT_VARIABLE_NEE_MEAN" etc
        if( nsbDoesKeyExist(nsb_p, outVarPropNvpKey) == 1)
        {
            // Next search the RIP NSB database using key, which returns
            // return its value as 
            // a comma delimited string like:
            // "NEE/nee_mean,gCm^2day^1,-9999.0,-20.0,100.0,mean daily 
            //  global Net ecosystem exchange at 9km"
            strcpy(valueStr, nsbGetValueByKey(nsb_p,outVarPropNvpKey,nvpBuf));
            
            // LONGNAME -- value string requires this special post-processing
            //  step to convert embedded "back-tick" delimiters BACK to 
            //  embedded blank spaces... Sorry, this hack is required since the
            //  standard NSB parser cannot yet support embedded blanks within tokens
            //  when a collection of tokens are comma delimited.
            
            // Restrict translation of backticks to just LONGNAME?
            if(strchr(valueStr,'`') !=NULL)
            {
                // char *frm_strReplaceChr(const char *targetChSet_p,
                //    char *srcAndResult_p,const char *newDest_char)
                
                // Replace ALL embedded back-tick "`" symbols with a space
                frm_strReplaceChr("`",valueStr," ");
            } // end:: translate strings with backtick to represent originally embedded blanks
            
          
            frm_strGetToken(longnameBuffer,valueStr,6,",");
            pkgMgr_SetObjLongname(pObj_p, longnameBuffer);
            
            // Perform a token pre-count to avoid un-necessary handling when properties are NOT defined
            tokenCount = frm_strCountTokens(valueStr,",");
            if(tokenCount <= 6)
            {
                xLogMsg("@A In(%s) skipping (%s) shortname[%s] No valid properties yet found in(%s)=[%s]~\n",
                    fxName, currentKeyword_p, shortnameBuffer, outVarPropNvpKey,valueStr);
                continue;
            } else {
                noOp = 0 ;
                // xLogMsg("@A In(%s) PROPERTIES (%s) shortname[%s] in(%s)=[%s]~\n",
                //      fxName, currentKeyword_p, shortnameBuffer, outVarPropNvpKey,valueStr);
            }
            
            // Example: use this method when supplying a RIP keyword for a dataset name to
            //    retrieve the literal HDF5 dataset's fully qualified name
            //    itself (includes GROUP if present)
            // nsbGetPropertyValueByKey(nsb_p, outVarPropNvpKey, IsPropertyName,nvpResult);
            
            // Assign datatype: fetch string format of TIV datatype identifier like "float32"
            nsbGetPropertyValueByKey(nsb_p, outVarPropNvpKey, IsPropertyDatatype,tivTypeAsStr);
            
            // DATATYPE Convert string form "float32" in RIP file, to integer code like IsFloat32 to store
            tivDatatype = l4c_DatatypeStringTo_Hdf4TypeId(tivTypeAsStr);
            
            pkgMgr_SetObjDatatype(pObj_p,tivDatatype); // load datatype
            
            // FILLVALUE Fetch fillvalue as a string first
            nsbGetPropertyValueByKey(nsb_p, outVarPropNvpKey, IsPropertyFillvalue,valueStr);
            
            pkgMgr_SetObjFillvalue(pObj_p,
                (float64)nsbGetPropertyNumericByKey(nsb_p,outVarPropNvpKey,IsPropertyFillvalue,&dblValu));

            // VALID MIN and MAX fetch then load "valid_min, valid_max aliased for NetCDF using vmin,vmax
            pkgMgr_SetObjValidMin(pObj_p,
                (float64)nsbGetPropertyNumericByKey(nsb_p,outVarPropNvpKey,IsPropertyValidMin,&dblValu));
                
            pkgMgr_SetObjValidMax(pObj_p,
                (float64)nsbGetPropertyNumericByKey(nsb_p,outVarPropNvpKey,IsPropertyValidMax,&dblValu));
                
            // UNITS -- fetch then load UNITS
            nsbGetPropertyValueByKey(nsb_p, outVarPropNvpKey, IsPropertyUnits,valueStr);
            pkgMgr_SetObjUnits(pObj_p,valueStr);
                
        } else {
            xLogMsg("@A In(%s) Dataset(%s) PropertyKEY(%s) does not yet exist or support RIP-assigned-properties ~\n",
                    fxName, currentKeyword_p, outVarPropNvpKey);
        }

        // Assuming this section is called AFTER l4c_hdf5_CreateDatasets() 
        //  was called, as it should be, we fetch the active HDF5 dataset_id 
        //  from each dataset's pObj_p instance.  Representative HDF5 dataset 
        //  ID values expected are like 83,886,142 and above...
        hGridProp_p->nHdf5OutputVariableCount = 0 ;

        pkgMgr_SetObjH5DatasetId(pObj_p, 0);
        // <JMG> 2016-01-07, BUG found and FIXED here, where fixed array
        // dset_L4C_IdAry[64] WAS sized at 64 (legacy), triggered corruption of
        // sesCtl_p->runDay when when the RMSE addition required an increase of
        // OUTPUT HDF5 datasets to 73. Now this array is sized to match number of
        // output datasets, defined as a safety macro in l4c_DAILY.h
        hGridProp_p->dset_L4C_IdAry[idx] = 0;
        
        // fetch ID...
        dset_id = pkgMgr_GetObjId(pObj_p);

        // For each valid ID, assign the hGridProp_p entry for this dataset
        if( dset_id > 80000000)
        {
            pkgMgr_SetObjH5DatasetId(pObj_p, dset_id); // set to HDF5 dataset ID only when known
            // ASSERT: if dataset's have already been created, 
            // dset_id may already be loaded in inside pObj_p. so fetch it
            hGridProp_p->dset_L4C_IdAry[idx] = pkgMgr_GetObjId(pObj_p);
            // increment verification count of HDF5 output variables correctly processed
            hGridProp_p->nHdf5OutputVariableCount++ ;
        } 
        
        // RESET local buffers...
        memset(outVarPropNvpKey, '\0', FILENAME_MAX);
        memset(valueStr,'\0',FILENAME_MAX);
        
    } // end:: idx -- loop through all output variables setting fallback defaults

    l4c_EventCompletionMsg(fxName);

 return 0 ;
} // end:: l4c_AssignOutputDatasetProperties()


#ifdef REFERENCE_ONLY
int REFERENCE_ONLY(void)
{
        // for REFERENCE ONLY, enumerate ALL critical "child" (e.g. variable level0 properties
    pObj_p = pkgMgr_GetObjByName(pThme_p, "NEE/nee_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "NEE/nee_std_dev");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "NEE/nee_pft1_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "NEE/nee_pft2_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "NEE/nee_pft3_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "NEE/nee_pft4_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "NEE/nee_pft5_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "NEE/nee_pft6_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "NEE/nee_pft7_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "NEE/nee_pft8_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "GPP/gpp_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "GPP/gpp_std_dev");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "GPP/gpp_pft1_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "GPP/gpp_pft2_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "GPP/gpp_pft3_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "GPP/gpp_pft4_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "GPP/gpp_pft5_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "GPP/gpp_pft6_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "GPP/gpp_pft7_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "GPP/gpp_pft8_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "RH/rh_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "RH/rh_std_dev");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "RH/rh_pft1_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "RH/rh_pft2_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "RH/rh_pft3_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "RH/rh_pft4_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "RH/rh_pft5_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "RH/rh_pft6_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "RH/rh_pft7_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "RH/rh_pft8_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "SOC/soc_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "SOC/soc_std_dev");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "SOC/soc_pft1_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "SOC/soc_pft2_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "SOC/soc_pft3_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "SOC/soc_pft4_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "SOC/soc_pft5_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "SOC/soc_pft6_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "SOC/soc_pft7_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "SOC/soc_pft8_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "EC/frozen_area");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "EC/emult_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "EC/tmult_mean");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "EC/wmult_mean");
    
    
    // pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/carbon_model_bitflag");
    // pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/carbon_model_qual_pft1");
    // pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/carbon_model_qual_pft2");
    // pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/carbon_model_qual_pft3");
    // pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/carbon_model_qual_pft4");
    // pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/carbon_model_qual_pft5");
    // pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/carbon_model_qual_pft6");
    // pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/carbon_model_qual_pft7");
    // pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/carbon_model_qual_pft8");
    
    pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/qa_count_pft1");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/qa_count_pft2");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/qa_count_pft3");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/qa_count_pft4");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/qa_count_pft5");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/qa_count_pft6");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/qa_count_pft7");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/qa_count_pft8");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/qa_count");

    pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/qa_rmse");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/qa_pft1_rmse");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/qa_pft2_rmse");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/qa_pft3_rmse");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/qa_pft4_rmse");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/qa_pft5_rmse");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/qa_pft6_rmse");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/qa_pft7_rmse");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "QA/qa_pft8_rmse");

    pObj_p = pkgMgr_GetObjByName(pThme_p, "GEO/latitude");
    pObj_p = pkgMgr_GetObjByName(pThme_p, "GEO/longitude");

} // end:: REFERENCE_ONLY
#endif
