// ---------------------------------------------------------------------
// module       : l4c_DAILY_Diagnostics.c
// purpose      : Implement DIAG_ONLY diagnostic outputs to facilitate
//                various output troubleshooting and issue resolution
//                efforts. Previously many of these functions were in
//                l4c_DAILY_sci.c, but we now try to consolidate these
//                into this (optional) module. Note that use of these
//                diagnostics may depend on external scripts (in Python
//                and/or R so the use of functions in this module ARE NOT
//                supported in GSFC GMAO OPS context
// project      : SMAP L4C science development.
// dependencies : genDiagImage.py, and HDF4 4.2.x, HDF5 v1.8.x 
//                or later libraries, L4C FRAMEWORK API (libfrm.a)
// revised      : 2014-10-30T14:39:00, added orig.
// history      : These functions previously resided in l4c_DAILY_sci.c
// author       : joe glassy, Lupine Logic Inc.
// notes        : n/a
// ---------------------------------------------------------------------

#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"


// ---------------------------------------------------------------------
// function : l4c_DiagSparseArrayToImage()
// inputs   :   char *imgLabel_p -- text label identifying context of diag info
//              char *outPath_p -- -- the frozen_area plot marked 'good' actually looks 'bad'path to (optional) output file
//              const int    outFormatType -- TBD, not yet used, fixed at PGM for now
//              const size_t resKmPixels EGv2_LDOM1KM or EGv2_LDOM9KM etc
//              int const tivDatatype, --TIV type like DFNT_UINT8 etc
//              void *sciArray_p -- ptr to raw 1D-sparse science data itself
// returns  : status, 0 if ok
// revision : 2014-10-30T15:13:00-07:00
// author   : joe glassy, Lupine Logic Inc.
// purpose  : implement diag only combined diagnostic operations
//            of inflating a 1D-sparse vector to 2D full global and
//            outputing this to a PGM uint8 image. Initially only
//            uint8 datasets are supported but later will extend to float32 etc
// dependencies: HDF4, HDF5 libs, and UM FRM API, and others to be added
// notes    : n/a
//
int l4c_DiagSparseArrayToImage(char *imgLabel_p,char *outPath_p,
    const float32 missingValue,const size_t resKmPixels,
    int const tivDatatype,void *sciArray_p)
{
 char fxName[256] = {0};
 char  pgmHdrBuffer[64] = {0} ;
 uint8 *origGridAry_p = (uint8 *)sciArray_p; // provide with a basic type
 uint8 *destGridImg_p = NULL;
 int   status = 0 ;
 size_t nPixels =0 ;
 size_t nOrigPixels =0 ;
 size_t ithPix =0;
 size_t nFullGridBytes = 0;
 size_t pgmDims[2] = {0} ;
 size_t checkSum =0 ;
 size_t diagAbsFreq[1024] = {0} ;
 
 float32 pix    = 0.0, pixCount=0.0, pixMin =0.0;
 float32 pixMax = 0.0, pixSum =0.0, pixMean =0.0;
 float32 pixMissingValue = 254.0 ; // for now, only uint8 supported
 
 FILE *pgmf = NULL; // optional PGM image header file handle
 
    strcpy(fxName,__func__);

    xLogMsg("@D In(%s) DIAG on (%s) tivDatatype %d~\n",
                fxName,imgLabel_p, tivDatatype);
                
    if( tivDatatype != DFNT_UINT8)
    {
        xLogMsg("@E In(%s) DIAG on %s only DFNT_UINT8 currently supported! ~\n",
                fxName, imgLabel_p);
        return 0 ;
    }

    // given a 1-D sparse vector resolution, set 2D full global grid equiv
    if(resKmPixels == EGv2_LDOM1KM)        { //   134787240  
        nOrigPixels = EGv2_LDOM1KM ;
        nPixels    = EGv2_GLOBAL_1KM_PIXELS ;
        nFullGridBytes = DFKNTsize(tivDatatype) * nPixels;
        pgmDims[0] = EGv2_GLOBAL_1KM_NS_PIXELS ;
        pgmDims[1] = EGv2_GLOBAL_1KM_EW_PIXELS ;
    } else if(resKmPixels == EGv2_LDOM3KM) { //  14976360
        nPixels = EGv2_GLOBAL_3KM_PIXELS;
        nOrigPixels = EGv2_LDOM3KM ;        
        nFullGridBytes = DFKNTsize(tivDatatype) * nPixels;      
        pgmDims[0] = EGv2_GLOBAL_3KM_NS_PIXELS ;
        pgmDims[1] = EGv2_GLOBAL_3KM_EW_PIXELS ;
    } else if(resKmPixels == EGv2_LDOM9KM) { // aaaaa
        nPixels = EGv2_GLOBAL_9KM_PIXELS ;
        nOrigPixels = EGv2_LDOM9KM ;        
        nFullGridBytes = DFKNTsize(tivDatatype) * nPixels;      
        pgmDims[0] = EGv2_GLOBAL_9KM_NS_PIXELS ;
        pgmDims[1] = EGv2_GLOBAL_9KM_EW_PIXELS ;
    }

    // Generate simple frequency distribution if uint8
    if(tivDatatype == DFNT_UINT8) // later extend to other datatypes
    {
        xLogMsg("@D In(%s) scanning 1D vector (%s) for abs. frequency counts~\n",
                fxName,imgLabel_p);
                
        // scan Original 1D-sparse vector array, accumulate absolute frequency counts
        for(ithPix=0; ithPix < nOrigPixels; ithPix++)
        {
            pix = (float32)origGridAry_p[ithPix];
            // skip over any missing value or fillvalue pixels
            if( pix >= pixMissingValue)
                    continue;
                    
            pixCount += 1.0 ;
            pixSum   += pix ;
            if( pixCount <= 1.0)
            {
                pixMin = pixMax = pix;
            } else {
                if(pix < pixMin) pixMin = pix;
                if(pix > pixMax) pixMax = pix;
            }
            // Increment freqbin count via pix value as bin itself
            diagAbsFreq[ origGridAry_p[ithPix] ]++ ;
            // <JMG> 2014-10-31T14:02, added correspondence check sum
            checkSum += diagAbsFreq[origGridAry_p[ithPix]];
        } // end: ithPix loop iterating through entire array 
        
        pixMean = pixSum / pixCount;
            
        // report results of frequency distribution calculation
        for(ithPix =0; ithPix < 256; ithPix++)
        {
            if(diagAbsFreq[ithPix] > 0)
            {
                xLogMsg("@D In(%s) DIAG for %s freq_bin %zu abs_freq %zu~\n",
                    fxName,imgLabel_p, ithPix, diagAbsFreq[ithPix] );
            }
        }
    }
    
    xLogMsg("@D In(%s) %s DIAG-SUMMARY count %.0f min %.3f max %.3f mean %.3f ~\n",
                fxName, imgLabel_p, pixCount, pixMin, pixMax,pixMean);
    
    xLogMsg("@D In(%s) INFLATING (%s) to 2D full grid ~\n",
                fxName, imgLabel_p);
                
    // full 2D grid -- generalizes any datatype to a uint8  reference type
    destGridImg_p = (uint8 *)memCalloc(nFullGridBytes,sizeof(uint8));
    
    // INFLATE to full 2D grid to resulting 'destGridImg_p'
    l4c_inflate_init_1km(destGridImg_p, tivDatatype);
    l4c_inflate_1km(origGridAry_p, destGridImg_p, tivDatatype);
    
    // Next, generate PGM output image, writing out 'destGridImg_p'
    if((pgmf = fopen(outPath_p,"wb")) == NULL)
    {
        xLogMsg("@F In(%s)FAILED to OPEN file ~\n",fxName,outPath_p);
    }
    xLogMsg("@D In(%s) DIAG on %s Creating 2D PGM image %s ~\n",
                fxName, imgLabel_p, outPath_p);
    // WRITE PGM HEADER out, listing nColumns THEN nRows in that order..
    fprintf(pgmf,"P5\n%zu %zu\n255\n",pgmDims[1],pgmDims[0]);
    fflush(pgmf);
    xLogMsg("@D In(%s) DIAG-PGM writing file %s\n",fxName,outPath_p);
    status = fwrite(destGridImg_p,sizeof(uint8),nFullGridBytes,pgmf);
    fclose(pgmf);

    memFree(destGridImg_p);
    
    xLogMsg("@D In(%s) DIAG completed for %s ~ \n",fxName, imgLabel_p);

 return 0 ; 
 
} // end:: l4c_DiagSparseArrayToImage()


// ---------------------------------------------------------------------
// function : l4c_SummarizePixels(PostProcMetrics *postMetrics_p,float32 *sciAry_p)
// inputs   : PostProcMetrics *postMetrics_p,
//          float32 *sciAry_pj -- 9km array of science pixel values
//            const char *dsetName_p -- name of dataset summarized
//            hid_t hdf5Id -- if valid, triggers WRITEs of metrics as dataset attributes
// returns  : size_t nQualified (non-fill) pixels
// revision : 2014-09-12T13:19:00
// author   : joe glassy, Lupine Logic Inc.
// purpose  : Optionally rescales output array, and, derives and returns
//            a set of very basic univariable statistics: observed min,max,
//            range,mean, std deviation, coefficient of variation etc.
//             for non-fillvalue pixels. Added to help supply granule level
//            ISO metadata terms.
float32 l4c_SummarizePixels(PostProcMetrics *postMetrics_p,
        const char *dsetName_p,const hid_t hdf5Id,
        const size_t pixelCount,const float32 fillValue,float32 *sciAry_p)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();
 
 char fxName[256] = {0} ;
 char finalMsg[256] = {0};
 char timeBuf[64] = {0};
 herr_t ret = -1;
 size_t  ithPix = 0;
 size_t  nPixels = pixelCount ;
 size_t  nUnqualified = 0 ; // count of dis-qualified pixels
 float32 sciPix   = 0.0 ; // pixel DN for post-proc transform
 float32 minAllowValue = 0.0 ; // later allow to parameterize
 float32 maxAllowValue = 100.0;// later allow to parameterize
 float64 sum     = 0.0 ;  
 float64 sumCount= 0.0 ;
 float64 sumSq   = 0.0 ;
 float64 stdError= 0.0   ; // standard error
 float64 pctStdError=0.0 ; // percent standard error
 float64 coef_var =0.0 ; // coefficient of variation 
 
 int     doesQualify = 0 ; // boolean true when within valid_min,valid_max
 // boolean true when valid_min,valid_max have been established
 int     userValidRangeAvailable  = 0; 
 // boolean, sep switch to enable HDF5 attrib writes
 int     doWriteMetrics = sesCtl_p->enableGranuleMetrics ;
 
 time_t calcStrtSec = time(0);
 time_t calcEndiSec = time(0);
 double calcElapSec = calcEndiSec - calcStrtSec;
 
    strcpy(fxName,__func__);
    
    // always initialize back at 0
    postMetrics_p->obs_CountQualified =0;
    postMetrics_p->minAllow = minAllowValue;
    postMetrics_p->maxAllow = maxAllowValue;
    
    // allow BLOCK on everything..
    if(sesCtl_p->enableGranuleMetrics == 0 )
    {
        nPixels =0 ;
    }
    
    // NOTE: native value scaling is {0.0 <= pixel <= 1.0} but PSD requires
    // value scaling of {0.0 <= pixel <= 100.0} as a percentage, so transform these
    // by multiplying each (non-fillvalue) pixel by 100
    for(ithPix=0; ithPix < nPixels; ithPix++)
    {
        sciPix = sciAry_p[ithPix];
        
        // constrain to a sensible sub-range as available
        doesQualify = (sciPix >= postMetrics_p->minAllow && postMetrics_p->maxAllow);
        
        if( userValidRangeAvailable && doesQualify )
        {   
            // if supplied, multiply by scale factor, like 100.0
            if(postMetrics_p->optionScaleFactor > 0.0)
                sciPix *= postMetrics_p->optionScaleFactor;
                
            // accumulate basic granule level metrics
            if(postMetrics_p->obs_CountQualified == 0)
            {
                // initialize observed min,max
                postMetrics_p->obs_min = postMetrics_p->obs_max = sciPix;
            } else {
                // conditionally update observed min,max
                if(sciPix < postMetrics_p->obs_min) postMetrics_p->obs_min = sciPix;
                if(sciPix > postMetrics_p->obs_max) postMetrics_p->obs_max = sciPix;
            }
            // accumulate sums, sum-of-squares
            sum += (float64)sciPix ; 
            sumSq += (float64)(sciPix * sciPix);
            // re-assign with transformed value
            sciAry_p[ithPix] = sciPix;
            // increment count of qualified pixels
            postMetrics_p->obs_CountQualified++ ;
        } else {
            nUnqualified++ ;
        }
    } // end:: ithPix loop to scale up to range {0.0<=pixel<=100.0}

    if( postMetrics_p->obs_CountQualified > 0)
    {
        // derive basic univariate summary statisics
        postMetrics_p->obs_range = (postMetrics_p->obs_max -postMetrics_p->obs_min);
        postMetrics_p->obs_mean  = (float32)sum / (float32)postMetrics_p->obs_CountQualified;
        sumCount =(float64)postMetrics_p->obs_CountQualified ;
        // Done in float64, but demoted to float32 after all calcs
        // uni_p->stddev   = sqrt( (uni_p->sum_sq-(uni_p->sum * uni_p->sum)/n)/(n-1.0));
        postMetrics_p->obs_stddev= (float32)sqrt((sumSq-(sum * sum)/sumCount)/(sumCount-1.0));

        // unpublished statistics
        if(sumCount >0)
        {
            stdError    = postMetrics_p->obs_stddev / sqrt(sumCount);
            if(postMetrics_p->obs_mean > 0.0)
            {
                pctStdError = (stdError / postMetrics_p->obs_mean)* 100.0;
                coef_var    = postMetrics_p->obs_stddev / postMetrics_p->obs_mean;
                // added later
                postMetrics_p->obs_stderror = stdError; // observed percent std error (non-fillvalue)
                postMetrics_p->obs_pctstderror = pctStdError; // observed percent std error (non-fillvalue)
                postMetrics_p->obs_coef_var    = coef_var; // observed percent std error (non-fillvalue)
            }
            xLogMsg("@A In(%s) %24s Observed minimum        %.4f ~ \n",fxName,dsetName_p,postMetrics_p->obs_min);
            xLogMsg("@A In(%s) %24s Observed maximum        %.4f ~ \n",fxName,dsetName_p,postMetrics_p->obs_max);
            xLogMsg("@A In(%s) %24s Observed range          %.4f ~ \n",fxName,dsetName_p,postMetrics_p->obs_range);
            xLogMsg("@A In(%s) %24s Observed mean           %.4f ~ \n",fxName,dsetName_p,postMetrics_p->obs_mean);
            xLogMsg("@A In(%s) %24s Observed std_deviation  %.4f ~ \n",fxName,dsetName_p,postMetrics_p->obs_stddev);
            xLogMsg("@A In(%s) %24s Observed coef. of var   %.4f ~ \n",fxName,dsetName_p,coef_var);
            xLogMsg("@A In(%s) %24s Observed std_error      %.4f ~ \n",fxName,dsetName_p,stdError);
            xLogMsg("@A In(%s) %24s Observed pct_std_error  %.4f ~ \n",fxName,dsetName_p,pctStdError);
            xLogMsg("@A In(%s) %24s served N. Unqualified %zu  ~ \n",fxName,dsetName_p,nUnqualified);
        } // end:: sumCount is positive...

        // write out as dataset level attributes if a valid dataset_id is supplied
        // attributes: observed_count,observed_min,observed_max,observed_mean
        
        // for now suppress writing out to HDF5 as metrics
        if(hdf5Id >0  && doWriteMetrics)
        {
            ret = l4c_Hdf5_CreateAttribute(hdf5Id,IsUint32,".","observed_count" ,&postMetrics_p->obs_CountQualified);
            sciPix = (float32)postMetrics_p->obs_min;
            ret = l4c_Hdf5_CreateAttribute(hdf5Id,IsFloat32,".","observed_min" ,&sciPix);
            sciPix = (float32)postMetrics_p->obs_max;
            ret = l4c_Hdf5_CreateAttribute(hdf5Id,IsFloat32,".","observed_max" ,&sciPix);
            sciPix = (float32)postMetrics_p->obs_range;
            ret = l4c_Hdf5_CreateAttribute(hdf5Id,IsFloat32,".","observed_range" ,&sciPix);
            sciPix = (float32)postMetrics_p->obs_mean;
            ret = l4c_Hdf5_CreateAttribute(hdf5Id,IsFloat32,".","observed_mean" ,&sciPix);
            sciPix = (float32)postMetrics_p->obs_stddev;
            ret = l4c_Hdf5_CreateAttribute(hdf5Id,IsFloat32,".","observed_std_deviation" ,&sciPix);
        }

    } // end:: CountQualified block tests true...
    
    calcEndiSec = time(0);
    calcElapSec = calcEndiSec - calcStrtSec;
    // char *frm_utilGetCtime(char *timBuf,const time_t usersTime)
    sprintf(finalMsg,"@A In(%s) for %s %zu qualified %zu disqualified started %s ended %s requiring %.2f sec",
        fxName, dsetName_p, 
        postMetrics_p->obs_CountQualified, nUnqualified,
        frm_utilGetCtime(timeBuf,calcStrtSec),
        frm_utilGetCtime(timeBuf,calcEndiSec),
        calcElapSec);
        
    xLogMsg("@A %s ~\n",finalMsg);
    
    l4c_EventCompletionMsg(fxName);

 // return count of qualified pixels
 return postMetrics_p->obs_CountQualified ;
 
} // end:: l4c_SummarizePixels()


// ---------------------------------------------------------------------
// function: l4c_DiagImage()
// purpose : accept a raw binary (2D) numeric array of known TIV type and 
//           generate a simple diagnostic image (pgm if uint8, or
//           jpeg if float32) using a python support utility that
//           calls numpy,scipy to produce the actual image)
//
// inputs  : char *imgPrefix, 
//           char *outputPath,
//           int tivDatatype,
//           size_t nPixels,
//           int *dims,
//           void *imgData_p
// returns      : status 0
// dependencies : genDiagImg.py python script in current directory
//
int l4c_DiagImage(char *imgPrefix, char *outputPath,int tivDatatype,size_t nPixels, size_t *dims,void *imgData_p)
{
 char osCmd[FILENAME_MAX]       = {0} ;
 char inpAryPath[FILENAME_MAX]  = {0} ;
 char tivTypeStr[32]            = {0} ;
 char typeBuffer[32]            = {0} ;
 char fxName[256]               = {0} ;
 char pyDiagPath[FILENAME_MAX]  = {0} ; // full unambiguous path to genDiagImg.py script
 char utilDirPath[FILENAME_MAX] = {0} ; // prefix directory where genDiagImg.py located 
 
 size_t nImgBytes = DFKNTsize(tivDatatype) * nPixels;
 uint8 *imgBuf_p = (uint8 *)imgData_p;
 int status = -1 ;
 int sizOfOne = 0 ;
 
 strcpy(fxName,__func__);
 
 // first, verify that genDiagImg.py exists
 strcpy(utilDirPath,"."); // default for now...
 sprintf(pyDiagPath,"%s/%s",utilDirPath,"genDiagImg.py");
 
 if(!frm_ioDoesFileExist(pyDiagPath))
 {
    xLogMsg("@E In(%s) utilDirPath(%s). DiagnosticImage capability requires python support utility [%s] ~\n",
        fxName,  utilDirPath,pyDiagPath);
    return -1 ;
 }
 
 // convert numeric TIV (HDF4) datatype identifier to its string equivalent
 strcpy(tivTypeStr,*tiv_NameofTivType(tivDatatype,&sizOfOne));
 
 
 // write out the raw numeric array as a file, "{imgLabel}.{datatype}"
 sprintf(inpAryPath,"%s.%s",imgPrefix,tivTypeStr);
 frm_ioWritebinary(inpAryPath,tivDatatype, nPixels, (uint8*)imgBuf_p);
 
 // the genDiagImg.py utility generates an image as well as 
 // displaying it immediately.  
 // python ../../l4c_DAILY/src/genDiagImg.py float32 1624 3856 smrz_mx.bin smrz_TEST.png 
 
 // convert the numeric TIV datatype to its string representation
 // like "float32" etc
 sprintf(osCmd,"python %s %s %zu %zu %s %s \n",
        pyDiagPath, tivTypeStr, dims[0], dims[1], inpAryPath, outputPath);
 xLogMsg("@A In(%s) osCmd[%s] ~\n",fxName,osCmd);

 system(osCmd); 

 return status ;
} // end:: l4c_DiagImage()


// ---------------------------------------------------------------------
// function  : l4c_GenFreqDistrib()
// purpose   : Accepts a raw binary (2D) numeric array of known TIV type
//             and from this, generates a simple diagnostic frequency 
//             distribution and (when available) produces an image of it. Uses numpy and
//             scipy.
// inputs       : char *imgPrefix -- user supplied prefix,
//              : char *outputPath -- user supplied output file path,
//              : int tivDatatype --  Type Independent 
//                    Variable (TIV) identifier like DFNT_UINT8
//              :  size_t nPixels,
//              : void *imgArray_p
// returns      : int, status
// dependencies : Python with numpy, matplotlib, later scipy as well
int l4c_GenFreqDistrib(char *imgPrefix, char *outputPath,int tivDatatype,
            size_t nPixels, void *imgArray_p)
{
 char osCmd[FILENAME_MAX] = {0};    
 char inpAryPath[FILENAME_MAX] = {0} ;
 char tivTypeStr[32]      = {0} ;
 char typeBuffer[32] = {0};
 char fxName[256] = {0} ;
 size_t nImgBytes   = DFKNTsize(tivDatatype) * nPixels;
 uint8 *imgBuf_p = (uint8 *)imgArray_p;
 int status = -1 ;
 int sizOfOne = 0 ;
 FILE *outf = NULL;
 
 strcpy(fxName,__func__);
 
 // first, verify that genDiagImg.py exists
 if(!frm_ioDoesFileExist("genFreqDistrib.py"))
 {
    xLogMsg("@E In(%s) requires python utility [%s] ~\n",
        fxName,"genFreqDistrib.py" );    
    return -1 ;
 }
 
 // Convert numeric TIV (HDF4) datatype identifier to its string equivalent
 strcpy(tivTypeStr,*tiv_NameofTivType(tivDatatype,&sizOfOne));
 
 // Write out the raw numeric array as a file, "{imgLabel}.{datatype}"
 sprintf(inpAryPath,"%s.%s",imgPrefix,tivTypeStr);
 outf   =fopen(inpAryPath,"wb");
 status = fwrite(imgBuf_p,sizeof(uint8),nImgBytes,outf);
 fclose(outf);
 
 // the genFreqDistrib.py utility uses numpy to generate a 
 // frequency distribution.  Invocation syntax is:
 // python ../../l4c_DAILY/src/genFreqDistrib.py float32 smrz_mx_1624x3856.flt32 smrz_TEST.png 
 
 // Convert the numeric TIV datatype to its string representation
 // like "float32" etc
 sprintf(osCmd,"python genFreqDistrib.py %s %s %s\n",
        tivTypeStr, inpAryPath, outputPath);
 xLogMsg("@A In(%s) osCmd[%s] ~\n",fxName,osCmd);

 system(osCmd); 

 return 0 ;
} // end:: l4c_GenFreqDistrib()


// Other Misc Statistics Utility Functions

#ifdef WHEN_COMPLETED_TBD
// ---------------------------------------------------------------------
// function: l4c_DataSummaryFreq()
// inputs  : const char *callerLbl_p,const char *dataObjLabel_p,
//              const char *rptLbl_p,int tivType,size_t nCount,void *rawData_p
// returns : int, status 0 when OK
// revised : 2016-01-08T12:45:00
// author  : joe glassy
// notes   : Initially supports byte data only.
int l4c_DataSummaryFreq(const char *callerLbl_p,const char *dataObjLabel_p,
              const char *rptLbl_p,int tivType,size_t nCount,void *rawData_p)
{
 char fxName[256] = {0} ; 
 int status = 0 ;   
 size_t  t =0;   
 size_t  nUnique = 0 ;
 long   nSum = 0 ;
 long   valuData = 0;
 long   vMin =0 ;
 long   vMax =0 ;
 uint8 *rdataUint8_p = (uint8 *)rawData_p;
  
    strcpy(fxName,__func__);
 
    // report context, who requested this summary, and on which data object
    xLogMsg("@A In(%s) %s caller %s for %s ~\n",fxName,rptLbl_p, callerLbl_p,dataObjLabel_p);
    
    if(tivType == DFNT_UINT8)
    {
        for(t=0; t < nCount; t++)
        {
            valuData = (long)rdataUint8_p[i] ;
            if( valuData > 0)
            {
               nSum += valuData;
               nUnique++ ; // tally n. of unique values with positive counts
               
               if(t ==0)
               {
                   vMin = vMax = valuData;
               } else {
                   if(valuData < vMin) vMin = valuData;
                   if(valuData > vMax) vMax = valuData;
               }
               xLogMsg("@A In(%s) caller: %s dataname %s index(%3d) absfreq(%ld)~\n",
                  fxName, callerLbl_p, dataObjLabel_p, t, valuData;
            } // end:: if value > 0

    } else {
        xLogMsg("@A In(%s) caller: %s dataname: %s index(%3d) -- TIV datatype %d (non byte) not yet supported here ~\n",
                  fxName, callerLbl_p, dataObjLabel_p, tivType );        
    }

#ifdef USE_CASE_MODEL
    for(nUnique=0,t=0; t < 256; t++)
        if( pfvmPftDom.pfvm_Uint8freq[t] > 0)
        {
            xLogMsg("@D In(%s) PftDom index(%3d) absfreq(%zu)~\n",
                fxName, t, pfvmPftDom.pfvm_Uint8freq[t]);
            nUnique++ ; // tally n. of unique values with positive counts
            nSum += pfvmPftDom.pfvm_Uint8freq[t] ;
        }
    xLogMsg("@D In(%s) %s POST-PROC BITFIELD PftDom tracked %d unique values. Sum(%zu)~\n",
        nUnique, nSum);     
#endif

    xLogMsg("@D In(%s) %s (caller %s) tracked %zu unique values. Sum %ld Min %ld Max %ld Range %ld ~\n",
        fxName, rptLbl_p, callerLbl_p, nUnique, nSum, (vMax-vMin) );     
        
    return 0 ;
    
} // end:: l4c_DataSummaryFreq()
#endif

// ---------------------------------------------------------------------
// function : l4c_SesCtl_Display()
// inputs   : void
// returns  : int , status 0
// purpose  : Display a selected diagnostic subset of sesCtl member variables
// revised  : 2016-01-07T14:04:00
// author   : Joe Glassy
// notes    : 
int l4c_SesCtl_Display(SesCtl *sesCtl_p,FILE *diagf)
{
 char fxName[256] = {0} ;
 
   strcpy(fxName,__func__);

  fprintf(diagf,"@D In(%s) runDayOfYear       %d ~\n",fxName,sesCtl_p->runDayOfYear);
  fprintf(diagf,"@D In(%s) runToStopDay       %d ~\n",fxName,sesCtl_p->runToStopDay);
  fprintf(diagf,"@D In(%s) runDayStepInterval %d ~\n",fxName,sesCtl_p->runDayStepInterval);
  fprintf(diagf,"@D In(%s) runCompositePeriod %d ~\n",fxName,sesCtl_p->runCompositePeriod);
  fprintf(diagf,"@D In(%s) runModelYear       %d ~\n",fxName,sesCtl_p->runModelYear);

  fprintf(diagf,"@D In(%s) sesNvpCount        %d ~\n",fxName,sesCtl_p->sesNvpCount);
  fprintf(diagf,"@D In(%s) dayOfYearToStop    %d ~\n",fxName,sesCtl_p->dayOfYearToStop);
  fprintf(diagf,"@D In(%s) hdf5ErrStackState  %d ~\n",fxName,sesCtl_p->hdf5ErrStackState);
  
  fprintf(diagf,"@D In(%s) runPixStepInterval %d ~\n",fxName,(int)sesCtl_p->runPixStepInterval);
  fprintf(diagf,"@D In(%s) hdf5ErrStackState  %d ~\n",fxName,sesCtl_p->hdf5ErrStackState);    

  fprintf(diagf,"@D In(%s) nSubTasks          %d ~\n",fxName,sesCtl_p->nSubtasks);
  fprintf(diagf,"@D In(%s) subtaskId          %ld ~\n",fxName,sesCtl_p->subtaskId);
  fprintf(diagf,"@D In(%s) subtaskIdStr       %s ~\n",fxName,sesCtl_p->subtaskIdStr);
  
  fprintf(diagf,"@D In(%s) nElem9kmPerTask %d ~\n",fxName,sesCtl_p->nElem9kmPerTask);
  fprintf(diagf,"@D In(%s) nElem3kmPerTask %d ~\n",fxName,sesCtl_p->nElem3kmPerTask);
  fprintf(diagf,"@D In(%s) nElem1kmPerTask %d ~\n",fxName,sesCtl_p->nElem1kmPerTask);
  fprintf(diagf,"@D In(%s) elem9kmOffset   %d ~\n",fxName,sesCtl_p->elem9kmOffset  );    
  fprintf(diagf,"@D In(%s) elem3kmOffset   %d ~\n",fxName,sesCtl_p->elem3kmOffset  );
  fprintf(diagf,"@D In(%s) elem1kmOffset   %d ~\n",fxName,sesCtl_p->elem1kmOffset  ) ;    
  
  fprintf(diagf,"@D In(%s) enableRMSE           %d ~\n",fxName,sesCtl_p->enableRMSE          );
  fprintf(diagf,"@D In(%s) enableSOCUpdate      %d ~\n",fxName,sesCtl_p->enableSOCUpdate     );
  fprintf(diagf,"@D In(%s) enableGranuleMetrics %d ~\n",fxName,sesCtl_p->enableGranuleMetrics);
  fprintf(diagf,"@D In(%s) enableCompleteMsg    %d ~\n",fxName,sesCtl_p->enableCompleteMsg   );    
  fprintf(diagf,"@D In(%s) compressLev          %d ~\n",fxName,sesCtl_p->compressLev         );
  fprintf(diagf,"@D In(%s) fparState            %d ~\n",fxName,sesCtl_p->fparState           ) ;
  fprintf(diagf,"@D In(%s) ftState              %d ~\n",fxName,sesCtl_p->ftState) ; 
  
  fprintf(diagf,"@D In(%s) socOpMode            %d ~\n",fxName,sesCtl_p->socOpMode) ;
  fprintf(diagf,"@D In(%s) ftOpMode             %d ~\n",fxName,sesCtl_p->ftOpMode) ;

  fprintf(diagf,"@D In(%s) appAction            %s ~\n",fxName,sesCtl_p->appAction) ;
  fprintf(diagf,"@D In(%s) gppMethodStr         %s ~\n",fxName,sesCtl_p->gppMethodStr    ) ;  
  fprintf(diagf,"@D In(%s) thisHost             %s ~\n",fxName,sesCtl_p->thisHost        ) ;  
  fprintf(diagf,"@D In(%s) thisUsername         %s ~\n",fxName,sesCtl_p->thisUsername    ) ;  
  fprintf(diagf,"@D In(%s) productionLevel      %s ~\n",fxName,sesCtl_p->productionLevel ) ;  
  fprintf(diagf,"@D In(%s) doi_assigned         %s ~\n",fxName,sesCtl_p->doi_assigned    ) ;
  fprintf(diagf,"@D In(%s) ftMethodString       %s ~\n",fxName,sesCtl_p->ftMethodString  ) ;  
  fprintf(diagf,"@D In(%s) SubsysReleaseTag     %s ~\n",fxName,sesCtl_p->SubsysReleaseTag) ;  
  fprintf(diagf,"@D In(%s) swVersion            %s ~\n",fxName,sesCtl_p->swVersion       ) ;  
  fprintf(diagf,"@D In(%s) swRelease            %s ~\n",fxName,sesCtl_p->swRelease       ) ;  
  fprintf(diagf,"@D In(%s) l4c_Task_Label        %s ~\n",fxName,sesCtl_p->l4c_Task_Label   ) ;
  fprintf(diagf,"@D In(%s) l4c_Task_UUID         %s ~\n",fxName,sesCtl_p->l4c_Task_UUID    ) ;  
  fprintf(diagf,"@D In(%s) l4c_Platform          %s ~\n",fxName,sesCtl_p->l4c_Platform     ) ;  
  fprintf(diagf,"@D In(%s) granuleUUID          %s ~\n",fxName,sesCtl_p->granuleUUID     ) ;  
  fprintf(diagf,"@D In(%s) ScienceVersID        %s ~\n",fxName,sesCtl_p->ScienceVersID   ) ;  
  fprintf(diagf,"@D In(%s) CRID                 %s ~\n",fxName,sesCtl_p->CRID            ) ;
  fprintf(diagf,"@D In(%s) frmApiVersion        %s ~\n",fxName,sesCtl_p->frmApiVersion   ) ;  
  fprintf(diagf,"@D In(%s) pkgApiVersion        %s ~\n",fxName,sesCtl_p->pkgApiVersion   ) ;  
  fprintf(diagf,"@D In(%s) nsbApiVersion        %s ~\n",fxName,sesCtl_p->nsbApiVersion   ) ;  
  fprintf(diagf,"@D In(%s) datasetVersion       %s ~\n",fxName,sesCtl_p->datasetVersion  ) ;  
                           
  fprintf(diagf,"@D In(%s) govATBD_Vers         %s ~\n",fxName,sesCtl_p->govATBD_Vers   ) ;  
  fprintf(diagf,"@D In(%s) govATBD_Date         %s ~\n",fxName,sesCtl_p->govATBD_Date   ) ;  
  fprintf(diagf,"@D In(%s) govPSD_Edition       %s ~\n",fxName,sesCtl_p->govPSD_Edition ) ;  
  fprintf(diagf,"@D In(%s) govPSD_Date          %s ~\n",fxName,sesCtl_p->govPSD_Date    ) ;  

  fprintf(diagf,"@D In(%s) doy_Str                  %s ~\n",fxName,sesCtl_p->doy_Str                 ) ;  
  fprintf(diagf,"@D In(%s) year_Str                 %s ~\n",fxName,sesCtl_p->year_Str                ) ;  
  fprintf(diagf,"@D In(%s) period_Str               %s ~\n",fxName,sesCtl_p->period_Str              ) ;  
  fprintf(diagf,"@D In(%s) period_Str               %s ~\n",fxName,sesCtl_p->period_Str              ) ;  
  fprintf(diagf,"@D In(%s) inputRipPath             %s ~\n",fxName,sesCtl_p->inputRipPath            ) ;  
  fprintf(diagf,"@D In(%s) inputRootDir             %s ~\n",fxName,sesCtl_p->inputRootDir            ) ;  
  fprintf(diagf,"@D In(%s) inputAncPath             %s ~\n",fxName,sesCtl_p->inputAncPath            ) ;  
  fprintf(diagf,"@D In(%s) inputSOCPath             %s ~\n",fxName,sesCtl_p->inputSOCPath            ) ;  
  fprintf(diagf,"@D In(%s) inputMetPath             %s ~\n",fxName,sesCtl_p->inputMetPath            ) ;  
  fprintf(diagf,"@D In(%s) inputFparPath            %s ~\n",fxName,sesCtl_p->inputFparPath           ) ;  
  fprintf(diagf,"@D In(%s) inputFparClimatologyPath %s ~\n",fxName,sesCtl_p->inputFparClimatologyPath) ;  
  fprintf(diagf,"@D In(%s) inputMetClimatology      %s ~\n",fxName,sesCtl_p->inputMetClimatology     ) ;  
  fprintf(diagf,"@D In(%s) inputBplutPath           %s ~\n",fxName,sesCtl_p->inputBplutPath          ) ;  
  fprintf(diagf,"@D In(%s) outputRootDir            %s ~\n",fxName,sesCtl_p->outputRootDir           ) ;  
  fprintf(diagf,"@D In(%s) output_L4C_DAILY_Path    %s ~\n",fxName,sesCtl_p->output_L4C_DAILY_Path    ) ;  
  fprintf(diagf,"@D In(%s) logPath                  %s ~\n",fxName,sesCtl_p->logPath                 ) ;  
  fprintf(diagf,"@D In(%s) logRootDir               %s ~\n",fxName,sesCtl_p->logRootDir              ) ;  
  fprintf(diagf,"@D In(%s) logInitialPath           %s ~\n",fxName,sesCtl_p->logInitialPath          ) ;  
  fprintf(diagf,"@D In(%s) logFinalPath             %s ~\n",fxName,sesCtl_p->logFinalPath            ) ;  
  fprintf(diagf,"@D In(%s) fparClimDatasetname      %s ~\n",fxName,sesCtl_p->fparClimDatasetname     ) ;  
  fprintf(diagf,"@D In(%s) latCoordVarname          %s ~\n",fxName,sesCtl_p->latCoordVarname         ) ;  
  fprintf(diagf,"@D In(%s) lonCoordVarname          %s ~\n",fxName,sesCtl_p->lonCoordVarname         ) ;  

  fprintf(diagf,"@D In(%s) sesStrtSec      %zu ~\n",fxName,sesCtl_p->sesStrtSec) ;
  fprintf(diagf,"@D In(%s) sesEndiSec      %zu ~\n",fxName,sesCtl_p->sesEndiSec) ; 
  fprintf(diagf,"@D In(%s) sesEndiSec      %lf ~\n",fxName,sesCtl_p->sesElapSec) ;

  fprintf(diagf,"@D In(%s) anc_file_id       %ld ~\n",fxName,sesCtl_p->anc_file_id  ) ;  
  fprintf(diagf,"@D In(%s) met_file_id       %ld ~\n",fxName,sesCtl_p->met_file_id  ) ;
  fprintf(diagf,"@D In(%s) soc_file_id       %ld ~\n",fxName,sesCtl_p->soc_file_id  ) ;
  fprintf(diagf,"@D In(%s) soc_group_id      %d ~\n",fxName,sesCtl_p->soc_group_id ) ;
  fprintf(diagf,"@D In(%s) fpar_file_id      %ld ~\n",fxName,sesCtl_p->fpar_file_id ) ;

  fprintf(diagf,"@D In(%s) l4cOutput_file_id  %ld ~\n",fxName,sesCtl_p->l4cOutput_file_id ) ;  
  fprintf(diagf,"@D In(%s) fparClima_file_id  %ld ~\n",fxName,sesCtl_p->fparClima_file_id ) ;  
  fprintf(diagf,"@D In(%s) viirs_file_id      %ld ~\n",fxName,sesCtl_p->viirs_file_id     ) ;  
  fprintf(diagf,"@D In(%s) nPkgThemes         %d ~\n",fxName,sesCtl_p->nPkgThemes) ;

  // runParamObj          scalar of (legacy) L4_C runtime parameters
  // *sesNsbDb_p          ptr to NVP stringblock database OVERSEER mgr
  //  *l4cPkgMgr_p         ptr to PkgMgr OVERSEER mgr

 return 0;
} // end:: l4c_SesCtl_Display()



