// ---------------------------------------------------------------
// function     : l4c_DAILY_ForwardStep.c
// purpose      : Serves as the daily TCF SPIN driver
// author       : Lucas Jones, NTSG
// author       : joe glassy, Lupine Logic Inc.
// inputs       : int doy,
//              : PftAdt      *pftObj_p,
//              : MetArrayAdt *metArrayObj_p,
//              : EcoStateArrayAdt *ecoStateArrayObj_p
//
// version      : see L4C_VERS_TAG , as 4.1.5 or later
// revised      : 2015-09-07, Currently Tsurf and GEOS5 "TS" work-around
//                are handled equivalently due to the way both write
//                the F/T status boolean to bit0 of FT_STATE_UM_M03
//              : 2015-08-26T14:00 <LAJ> Uncommented rescalSmrz() function
//                call for rescaling root zone sm
// revised      : v6.0-2015-06-10T16:57, correct 'pct' progress indicator for v6
// revised      : 2014-10-28T13337:00, enable: BPLUT 1-deep cache, avoids
//                redundant lookups for adajacent pixels sharing same PFT
// history      : 2014-10-27T11:37:00, enable: PERFORM ONE-TIME-ONLY PkgMgr
// theme Lookups! history      : 2014-07-23T120:00:00, modified to do
// aggregation of 1km to 9km within forward step, laj history      :
// 2014-03-12t14:39:00, removed legacy block at end and
//                set L4C_DIAG to output raw binary "nee" only when requested
// history      : 2014-02-05T11:242:00-07:00,wheat,jmg"
// dependencies : HDF4, HDF5 libraries and L4C FRAMEWORK API.
// history      : n/a
// notes        : Based on and adapted from tcf_spin v3.1/3.2
//
// QUICK R code to verify NEE values
// inPath = "ecoStateArrayObj_p_nee.dom_1km.flt32"
// neeRaw <- readBin(inPath,what='double',size=4,n=134787240)
//  summary(neeRaw[neeRaw > -9999.0] )
//   Min. 1st Qu.  Median    Mean 3rd Qu.    Max.
// 0.0000  0.9205  2.1300  2.5400  3.6060 18.2000
//  hist( neeRaw[neeRaw > -9999.0] )
// ------------------------------------------------------------------

#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"

// ------------------------------------------------------------------
// function : l4c_ModDriverGrid_forwardStep()
// Purpose  : Drives forward simulation for a specified single time-step.
//            Produces all variables specified in the
//            L4C ATBD, including QA fields.
// Inputs   : Initial C pools, passed in the STATE_ARRAY structure,
//             dynamic daily met data (MetArrayObj),
//            static pft map (PFT_ARRAY), and static model parameters
//            (GLOB_PARAM, BPLUT).
// Outputs  : Daily states, including final C pools, in 1km grid arrays returned
//            in the STATE_ARRAY structure.
// Revised  : 2014-02-19T21:15, introduced RunParamAdt to split out NVPs from
// SesCtl.. Notes    : Increments the 't8day' counter variable to determine
//            when to read in available fpar 8-day composites.
//
void l4c_ModDriverGrid_forwardStep(int doy, PftAdt *pftObj_p,
                                   MetArrayAdt *metArrayObj_p,
                                   EcoStateArrayAdt *ecoStateArrayObj_p) {
  SesCtl *sesCtl_p = l4c_GetSesCtl(); // ptr to session controls
  // <JMG> 2014-09-04, removed unused AncImagePkgAdt here
  // AncImagePkgAdt     *ancImagePkgObj_p   = l4c_GetAncImagePkgAdt();
  PkgMgr *pkgMgr_p = sesCtl_p->l4cPkgMgr_p; // set ptr to OVERSEER

  // perform one-time lookup...
  PkgTheme *pStateThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "L4C_OUTPUTS");

  // fetch ptrs to PftDom and QaCount trackers
  PostProcFreqVerifyMetrics *pfvmPftDom_p = l4c_GetPostProcFreqVerify_PftDom();
  PostProcFreqVerifyMetrics *pfvmQaScore_p =
      l4c_GetPostProcFreqVerify_QaScore();

  // RunParamAdt   *runParamObj_p = sesCtl_p->runParamObj_p; ... ptr to runtime
  // inputs for local, inner-most-loop state calcs, not requiring full 2D grid
  // arrays
  MetAdt metObj;
  // scratchpad instance, for inner-most-loop state calcs, not requiring full 2D
  // grid arrays
  EcoStateAdt ecoStateObj;

  BPLUT *bp_p = NULL; // ptr to current BPLUT record indexed via base1 pft_index
  BPLUT *prev_BPLUT_p = NULL; // ptr, facilitates 1-deep cache
  BPLUT prev_BPLUT_Recd;      // facilitates 1-deep cache

  uint8 cnt_3km_ftvalid_9km = 0;
  uint8 *fparQaDiag_M01_p =
      NULL; // <JMG> 2014-10-30T13:06, added for diag of fpar_qa

  uint8 pft_index = 0;
  uint8 prevPftIndex = 255; // LEAVE INIT at 255, implements 1-deep cache

  size_t nBplutCandidates = 0; // count of total candidate fetches
  size_t nBplutCacheHits = 0;  // tally of BPLUT fetch-cache hits
  size_t nBplutCacheRefreshes = 0;

  // size_t    t8day =0; TBD,...
  size_t l = 0;
  size_t rn3km, cn3km, rn1km, cn1km;
  size_t s3km_nested9km, s1km_nested3km, s1km_nested9km;
  size_t s1km, s3km, s9km;
  // size_t    nelem;
  size_t nQualifyingCorePix = 0;

  char fxName[256] = {0};
  char timeStr[32] = {0};
  char testOutPath[FILENAME_MAX] = {0}; // DIAG ONLY, binary output path if used

  // FILE *diagf = NULL ;   DIAG ONLY binary output handle if used

  time_t localStrtSec = time(0);
  time_t localEndiSec = time(0);
  double localElapSec = 0.0;
  double cumElapSec = 0.0;
  float32 pctQualified = 0.0;

  int t = 0;
  int z = 0; // count for diagnostics only
  int startNameIdx = 0;

  // <JMG> 2014-07-23T17:29, added for storing valid_min,valid_max and
  // thresholds
  QcStateLimitAdt *qcStateLim_p = l4c_GetQcStateLimits();
  // Fetch valid min,max and thresholds assigned to QcStateLimitAdt qaLim
  // structure
  qcStateLim_p = l4c_GetQcStateLimits();

  // -------------------------------------------------------------------
  // <JMG> 2014-10-27T11:37:00, PERFORM ONE-TIME-ONLY PkgMgr theme Lookups!
  // Define a set of unique, fixed, pre-set (once) starting indices
  //  into pStateThme_p in PkgTheme "L4C_OUTPUTS", for each science
  //  variable, to avoid massively [((6,262,144 * 7)-7) = 43,835,001]*2
  //  REDUNDANT lookups!
  int startPkgIdx_nee_mean = 0;
  int startPkgIdx_gpp_mean = 0;
  int startPkgIdx_rh_mean = 0;
  int startPkgIdx_soc_mean = 0;
  // define for EC terms: emult, tmult, wmult
  int startPkgIdx_ec_emult_av = 0;
  int startPkgIdx_ec_tmult_av = 0;
  int startPkgIdx_ec_wmult_av = 0;
  // From indices, define pointers to names stored in theme "L4C_OUTPUTS"
  char *nee_meanChildName_p = NULL;
  char *gpp_meanChildName_p = NULL;
  char *rh_meanChildName_p = NULL;
  char *soc_meanChildName_p = NULL;
  char *ec_emult_avChildName_p = NULL;
  char *ec_tmult_avChildName_p = NULL;
  char *ec_wmult_avChildName_p = NULL;
  // -------------------------------------------------------------------

  strcpy(fxName, __func__);

  // -----------------------------------------------------------------
  // sesCtl_p->sesStrtSec ;       // session starting time, in sec
  sesCtl_p->sesEndiSec = time(0); // snapshot cumulative time thus far
  cumElapSec = sesCtl_p->sesEndiSec - sesCtl_p->sesStrtSec;

  xLogMsg("@A In(%s) started %s at cumul global elap sec %.2f~\n", fxName,
          frm_utilGetCtime(timeStr, 0), cumElapSec);

  // guard against unexpected input parameters on multiple day runs
  if (sesCtl_p->runDayOfYear > 366 || sesCtl_p->runToStopDay > 366) {
    xLogMsg("@F In(%s) Unexpected Starting DOY %d or Ending DOY %d~\n", fxName,
            sesCtl_p->runDayOfYear, sesCtl_p->runToStopDay);
    exit(-1);
  }

  // Perform ONE-TIME-ONLY lookups to cache all required keys used in the 9km
  // loop below, to eliminate the redundant 86 million lookups into PkgMgr!
  // Note: the whole purpose of these 'round-trip' style lookups is to
  // protect and insulate source code from any possible future change
  // made to names in the PkgMgr theme. This way, any such changes propogate
  // through here without needing to change this code!

  // First, retrieve pointer to the theme table for state and QA output
  // variables
  pStateThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "L4C_OUTPUTS");
  xLogMsg("@A In(%s) initializing one-time-only PkgMgr lookups~\n", fxName);
  // Establish starting offsets by science var name
  startPkgIdx_nee_mean = pkgMgr_GetObjIndexByName(pStateThme_p, "NEE/nee_mean");
  startPkgIdx_gpp_mean = pkgMgr_GetObjIndexByName(pStateThme_p, "GPP/gpp_mean");
  startPkgIdx_rh_mean = pkgMgr_GetObjIndexByName(pStateThme_p, "RH/rh_mean");
  startPkgIdx_soc_mean = pkgMgr_GetObjIndexByName(pStateThme_p, "SOC/soc_mean");
  // <JMG> 2016-02-18T13:51 changed all emult_av to emult_mean, etc
  startPkgIdx_ec_emult_av =
      pkgMgr_GetObjIndexByName(pStateThme_p, "EC/emult_mean");
  startPkgIdx_ec_tmult_av =
      pkgMgr_GetObjIndexByName(pStateThme_p, "EC/tmult_mean");
  startPkgIdx_ec_wmult_av =
      pkgMgr_GetObjIndexByName(pStateThme_p, "EC/wmult_mean");

  // Next, using starting indices, set pointers to associated NAMES stored in
  // the PkgTheme "L4C_OUTPUTS"
  nee_meanChildName_p =
      pkgMgr_GetThemes_ChildName(pStateThme_p, startPkgIdx_nee_mean);
  gpp_meanChildName_p =
      pkgMgr_GetThemes_ChildName(pStateThme_p, startPkgIdx_gpp_mean);
  rh_meanChildName_p =
      pkgMgr_GetThemes_ChildName(pStateThme_p, startPkgIdx_rh_mean);
  soc_meanChildName_p =
      pkgMgr_GetThemes_ChildName(pStateThme_p, startPkgIdx_soc_mean);
  ec_emult_avChildName_p =
      pkgMgr_GetThemes_ChildName(pStateThme_p, startPkgIdx_ec_emult_av);
  ec_tmult_avChildName_p =
      pkgMgr_GetThemes_ChildName(pStateThme_p, startPkgIdx_ec_tmult_av);
  ec_wmult_avChildName_p =
      pkgMgr_GetThemes_ChildName(pStateThme_p, startPkgIdx_ec_wmult_av);

  // Transfer expected state min/max bound limits, from RIP file
  ecoStateArrayObj_p->nee.limitMin = qcStateLim_p->neeLimitMin;
  ecoStateArrayObj_p->nee.limitMax = qcStateLim_p->neeLimitMax;
  ecoStateArrayObj_p->gpp.limitMin = qcStateLim_p->gppLimitMin;
  ecoStateArrayObj_p->gpp.limitMax = qcStateLim_p->gppLimitMax;
  ecoStateArrayObj_p->rhtot.limitMin = qcStateLim_p->rhLimitMin;
  ecoStateArrayObj_p->rhtot.limitMax = qcStateLim_p->rhLimitMax;
  ecoStateArrayObj_p->Ctot.limitMin = qcStateLim_p->socLimitMin;
  ecoStateArrayObj_p->Ctot.limitMax = qcStateLim_p->socLimitMax;
  // these low,high endpoints are fixed since they don't vary
  ecoStateArrayObj_p->Emult.limitMin = 0.0;
  ecoStateArrayObj_p->Emult.limitMax = 1.0;
  ecoStateArrayObj_p->Tmult.limitMin = 0.0;
  ecoStateArrayObj_p->Tmult.limitMax = 1.0;
  ecoStateArrayObj_p->Wmult.limitMin = 0.0;
  ecoStateArrayObj_p->Wmult.limitMax = 1.0;

#ifdef DIAG_FPAR_QA
  // <JMG> 2014-10-30 DIAG ONLY, added to create a 1KM 1D-sparse diagnostics
  // array of "metObj_p->fpar_qa", storing scalar in array for inspection only
  fparQaDiag_M01_p = (uint8 *)memCalloc((size_t)EGv2_LDOM1KM, sizeof(uint8));
#endif

  // <JMG> 2014-10-16, initialize the local devel diag freq verifications
  // of PftDom and QaScore to validate QA bitfield assignments
  strcpy(pfvmPftDom_p->pfvm_Uint8Varname, "PftDom");
  strcpy(pfvmQaScore_p->pfvm_Uint8Varname, "QaScore");

  for (t = 0; t < 256; t++) {
    pfvmPftDom_p->pfvm_Uint8freq[t] = 0;
    pfvmQaScore_p->pfvm_Uint8freq[t] = 0;
  } // end:: t loop initializing tracker counts for selected uint8 vars

  // -----------------------------------------------------------------
  // MAIN 9KM LOOP -- OUTER 9KM 1D Sparse Vector LAND loop.
  //  Now with parameterized pixel step inteval (defaults to 1),
  //  allowing systematic skip-sampling for testing only.
  //  Remember: count of nElem9kmPerTask iterates through the
  //  1D-sparse 9km extent, and NOT through a 2D 9km extent.

  xLogMsg("@A In(%s) main 9KM model process for %zu pixels started. ~\n",
          fxName, (size_t)sesCtl_p->nElem9kmPerTask);

  // ITERATE through 1D sparse vector 9KM model space
  // NOTE: the limit count (sesCtl_p->nElem9kmPerTask) = LDOM9KM or 1664040

  for (s9km = 0; s9km < (size_t)sesCtl_p->nElem9kmPerTask;
       s9km += sesCtl_p->runPixStepInterval) {
    // periodically report out the fraction of overall task completed
    if (s9km % 200000 == 0) {
      pctQualified = (float32)s9km / (float32)EGv2_LDOM9KM;
      xLogMsg(
          "@A In(%s) s9km PIXEL %9zu of %9d (%7.3f pct done) at DOY %-3.3d ~\n",
          fxName, s9km, (size_t)EGv2_LDOM9KM, (pctQualified * 100.0), doy);
    }

    // TBD: when implementing 9km pixel-level filter mask, insert
    // test here.  l4c_DoFilter9kmPixel() != 0)  continue;

    // Unpack input 9-km MET arrays for step call interface
    metObj.tmin = (float64)metArrayObj_p->tmin[s9km]; // [K]
    metObj.vpd = (float64)metArrayObj_p->vpd[s9km];   // [Pa]
    metObj.par = (float64)metArrayObj_p->par[s9km] *
                 RAD2PAR; // convert from [W m-2] rad to [MJ m-2] par

    // Rescale root zone soil moisture.
    // <LAJ> 8/25/2015, Now recalibrated with min/max from Natv4, use again
    // <JMG> 2015-08-05T17:29, Re-enabled the call to tcfModMult_rescalSmrz()
    // which was previously commented out for further diagnosis and correction.
    // <JMG> 3/4/3014, now corrected: SIGSEGV, Segmentation fault was here, now
    // fixed
    metObj.smrz = tcfModMult_rescalSmrz(metArrayObj_p->smrz[s9km],
                                        metArrayObj_p->smrz_mn[s9km],
                                        metArrayObj_p->smrz_mx[s9km]);

    // <LAJ> 8/25/2015, Now recalibrated with min/max from Natv4, do not use
    // unscaled line below <LAJ> 5/4/2014, Use this line until better min/max
    // estimates available from L4SM Native Run <JMG> 2015-08-05T17:30,
    // commented next line out as per discussion with Lucas metObj.smrz =
    // (float64) metArrayObj_p->smrz[s9km] * 100.0;
    metObj.tsoil = (float64)metArrayObj_p->tsoil[s9km]; // [K], soil temperature
    metObj.smtop = (float64)metArrayObj_p->smtop[s9km] *
                   100.0; // soil moisture, convert from [dim] to [%]

    // Initialize frozen area sum for 9km grid cell
    ecoStateObj.ft_sum_9km = 0.0;
    cnt_3km_ftvalid_9km = 0;

    //  Begin nested 3km grid loop
    for (rn3km = 0; rn3km < NESTED_3KM_IN_9KM; rn3km++) {
      for (cn3km = 0; cn3km < NESTED_3KM_IN_9KM; cn3km++) {
        // Unpack input 3-km FT arrays for step call interface

        // serial array index nesting for 3km grid
        s3km_nested9km =
            M_2D_B0(rn3km, cn3km, NESTED_3KM_IN_9KM); // range {0..9}

        // Absolute serial reference into 3km global array
        s3km = (uint32)M_2D_B0(s9km, s3km_nested9km,
                               NESTED_3KM_IN_9KM_SQ); // range {0..LDOM3KM}

        // Switch FT method between
        // Populates ftRolling and ftAge 3km arrays
        metObj.ftMethod = tcfModFunc_switchFTmethod(
            s3km, &metObj, metArrayObj_p, ecoStateArrayObj_p);

        // Unpack ftRolling and Age arrays for model step call interface
        metObj.ft =
            (uint8)ecoStateArrayObj_p
                ->ftRolling_3km[s3km]; // (1=Thawed; 0=Frozen) units: [bool]
        metObj.ftAge =
            (uint8)ecoStateArrayObj_p->ftRollingAge_3km[s3km]; // units: [days]

        // tally Radar FT method
        // <JMG> 2015-09-07, current Tsurf and GEOS5 "TS" work-around are
        // handled equivalently here due to the way both write the F/T status
        // boolean to bit0 of FT_STATE_UM_M03 enum FreezeThawInputSourceChoices
        // -->
        //  DoUseFT_ActiveRadar, DoUseFT_ActiveTsurf, DoUseFT_GEOS5,
        //  DoUseFT_PassiveRadiom, DoUseFT_PassiveTsurf, DoUse_NO_FT
        //

        if (metObj.ftMethod == DoUseFT_ActiveRadar)
          ecoStateArrayObj_p->nee_qa.cnt_useRadarFT_9km[s9km]++;
        else if (metObj.ftMethod == DoUseFT_ActiveTsurf ||
                 metObj.ftMethod ==
                     DoUseFT_GEOS5) // DoUseFT_ActiveTsurf or new DoUseFT_GEOS5
          ecoStateArrayObj_p->nee_qa
              .cnt_useTsurfFT_9km[s9km]++; // Tally Tsurf FT method

        // else...NOTE: NO "Use-NO-FT" branch is supported here yet

        // Increment frozen area sum and count for 9km grid cell
        ecoStateObj.ft_sum_9km +=
            (float64)ecoStateArrayObj_p->ftRolling_3km[s3km];
        cnt_3km_ftvalid_9km++;

        // ==============================================================
        // 1KM PROCESSING LOOP SECTION
        // ==============================================================

        // Begin nested 1km 2D grid loops
        for (rn1km = 0; rn1km < NESTED_1KM_IN_3KM; rn1km++) {
          for (cn1km = 0; cn1km < NESTED_1KM_IN_3KM; cn1km++) {
            // Derive the SERIAL array indices for nested 1km grid into 3km,9km
            // grids
            s1km_nested3km =
                M_2D_B0(rn1km, cn1km, NESTED_1KM_IN_3KM); // range {0..9}
            s1km_nested9km = M_2D_B0(s3km_nested9km, s1km_nested3km,
                                     NESTED_1KM_IN_9KM); // range {0..81}

            // Absolute serial reference into 1km global (1D-sparse vector)
            // array
            s1km = (uint32)M_2D_B0(s9km, s1km_nested9km,
                                   NESTED_1KM_IN_9KM_SQ); // range {0..LDOM1KM}

            // Unpack PFT class (integer) at this 1km 1D-sparse vector location.
            // Assert: the PFT index here is base1, {1..8} and NOT a base0
            // integer index value
            pft_index = pftObj_p->dom_1km[s1km];

            // initialize SCALAR...
            metObj.fpar_qa = 254;

            // Unpack pft into nested array for 9km post processor
            pftObj_p->nested_1km[s1km_nested9km] = pftObj_p->dom_1km[s1km];

            // TBD: Implement a std 1-deep cache, so that when adjacent pixels
            // share same PFT
            //      we avoid redundant lookups
            nBplutCandidates++; // count total candidates

            if (pft_index == prevPftIndex) {
              bp_p = prev_BPLUT_p; // cache hit, so use it!
              nBplutCacheHits++;
            } else {
              bp_p = l4c_GetBPLUT_ByIndex(pft_index,
                                          NPFT); // actual lookup required
              prev_BPLUT_p = bp_p;               // refresh the 1-deep cache
              nBplutCacheRefreshes++;            // count times cache refreshed
            }
            // update BPLUT 1-deep cache PFT index
            prevPftIndex = pft_index;

            // Unpack 1-km fpar array for step call interface
            // Scale and offset can be set to estimate FPAR from VI data using a
            // linear regression model

            // Or FPAR is used directly with original gain and offset (as is
            // currently done) Also, unpack and reclassify fpar qa/qc data Now
            // qa has integer values 1 to 4, with 1 = best quality

            // Note coding of 4-level gppMethod here that differs from similar
            // gppMethod indicator switch: (-1 = default; 0 = used daily FPAR; 1
            // = used FPAR climatology; 2 = used NDVI)
            metObj.gppInputUsed =
                tcfModFunc_switchFparVi(bp_p, s1km, &metObj, metArrayObj_p);

#ifdef DIAG_FPAR_QA
            // <JMG> 2014-10-30T13:24:00, added for DIAG ONLY of fpar_qa scalar
            // value in metObj
            fparQaDiag_M01_p[s1km] = metObj.fpar_qa;
#endif

            // <KAE> 2024-01-05 Updated to support two litterfall allocation schemes
            if (strstr(sesCtl_p->litterfallScheme, "CONSTANT") != NULL)
            {
              // NPP sum must always be initialized for state estimates to be
              // available
              ecoStateObj.lit =
                  (float64)ecoStateArrayObj_p->npp_sum.dom_1km[s1km] / NDAYS;
            } else if (strstr(sesCtl_p->litterfallScheme, "SCHEDULED") != NULL) {
              // <KAE> 2024-01-05 NOTE that we divide the litter rate by 8 days
              //  because it is expressed as the fraction of annual litterfall over
              //  a MODIS/VIIRS 8-day composite period; fmax() ensures it is non-negative
              ecoStateObj.lit =
                  (float64)ecoStateArrayObj_p->npp_sum.dom_1km[s1km] * (
                    (float64)fmax(0.0f, ecoStateArrayObj_p->litter_rate[s1km] / 8.0f));
            }

            //  Check if cell is in 1-km processing domain
            //  Process ONLY the vegetated PFTs with non-missing input
            if (tcfModFunc_isInCell(pft_index, metObj, ecoStateObj) != 0) {

              // Initialize SOC,  SOC NPOOL=3, so {0,1,2} from whatever previous
              // state (from model run or spin-up)
              for (l = 0; l < NPOOL; l++)
                ecoStateObj.C[l] =
                    (float64)ecoStateArrayObj_p->C[l].dom_1km[s1km];

              // Do tcf forward step to propagate SOC states from last
              // timestep to current timestep
              tcfModFunc_forwardStep(bp_p, metObj, &ecoStateObj);

              // Compute state variable qa classification for the timestep.
              // Weights calculated internally.  Fpar qa is the only input qa
              // information considered at this time.
              tcfModQa_step(bp_p, metObj, &ecoStateObj);

              // Required: prepare SOC compartments {0,1,2} for next timestep
              for (l = 0; l < NPOOL; l++)
                ecoStateArrayObj_p->C[l].dom_1km[s1km] =
                    (float32)ecoStateObj.C[l];

              // UPDATE State variable (array) fields from scalar EcoStateAdt
              // instances
              ecoStateArrayObj_p->nee.nested_1km[s1km_nested9km] =
                  (float32)ecoStateObj.nee;
              ecoStateArrayObj_p->gpp.nested_1km[s1km_nested9km] =
                  (float32)ecoStateObj.gpp;
              ecoStateArrayObj_p->rhtot.nested_1km[s1km_nested9km] =
                  (float32)ecoStateObj.rhtot;
              ecoStateArrayObj_p->Ctot.nested_1km[s1km_nested9km] =
                  (float32)ecoStateObj.Ctot;
              ecoStateArrayObj_p->Emult.nested_1km[s1km_nested9km] =
                  (float32)ecoStateObj.gpp_mult.k;
              ecoStateArrayObj_p->Tmult.nested_1km[s1km_nested9km] =
                  (float32)ecoStateObj.rh_mult.tsoil;
              ecoStateArrayObj_p->Wmult.nested_1km[s1km_nested9km] =
                  (float32)ecoStateObj.rh_mult.smtop;

              // State variable qa field
              // <LAJ> 2014-03-12 NOTE: This is the fundamental 1km floating
              // point psuedo error. QA integer
              //  rank will be computed from this 1km psuedo-error 9km x PFT
              //  averages in the Post Processor.
              // RMSE? Verify if this supplies the new RMSE metrics as it
              // appears to.
              ecoStateArrayObj_p->nee_qa.nested_1km[s1km_nested9km] =
                  (float32)ecoStateObj.ecoStateQaObj.nee_error;

              // Increment 9KM pixel wise count of FPAR climatology GPP method
              // (1) used
              if (metObj.gppInputUsed == 1)
                ecoStateArrayObj_p->nee_qa.cnt_useFparClim_9km[s9km]++;

              // Tally NDVI GPP method...
              if (metObj.gppInputUsed == 2)
                ecoStateArrayObj_p->nee_qa.cnt_useNdvi_9km[s9km]++;

              // double check by tracking counts of qualifying pixels
              nQualifyingCorePix++;

              // end:: block tcfModFunc_isInCell() filter
            } else {
              // Non-veg pft grid cells: State variables get fill value

              // initialize (3) SOC 1km pools with fill value
              for (l = 0; l < NPOOL; l++)
                ecoStateArrayObj_p->C[l].dom_1km[s1km] = FLOAT_FILL_VALUE;

              // Set FILLVALUES to State variable fields:
              ecoStateArrayObj_p->nee.nested_1km[s1km_nested9km] =
                  FLOAT_FILL_VALUE;
              ecoStateArrayObj_p->gpp.nested_1km[s1km_nested9km] =
                  FLOAT_FILL_VALUE;
              ecoStateArrayObj_p->rhtot.nested_1km[s1km_nested9km] =
                  FLOAT_FILL_VALUE;
              ecoStateArrayObj_p->Ctot.nested_1km[s1km_nested9km] =
                  FLOAT_FILL_VALUE;
              ecoStateArrayObj_p->Emult.nested_1km[s1km_nested9km] =
                  FLOAT_FILL_VALUE;
              ecoStateArrayObj_p->Tmult.nested_1km[s1km_nested9km] =
                  FLOAT_FILL_VALUE;
              ecoStateArrayObj_p->Wmult.nested_1km[s1km_nested9km] =
                  FLOAT_FILL_VALUE;
              // Assign default (fill) for state variable qa fields. RMSE ?
              ecoStateArrayObj_p->nee_qa.nested_1km[s1km_nested9km] =
                  FLOAT_FILL_VALUE; // RMSE ?

            } // end :: 1km in-domain (PFT and filter check)

          } // end :: 1km COLUMN nested 3km loop

        } // end :: 1km ROW nested 3km loop

        // (Next level loops are for 3KM processing)

      } // end :: 3km row nested 9km loop

    } // end :: 3km col nested 9km loop

    // <JMG> 2014-10-01, check integrity of carbon_model_bitflag. At pixel
    // s9km=42444 a transient FATAL SEGVIOLATION is encountered...
    assert(ecoStateArrayObj_p->nee_qa.carbon_model_bitflag != NULL);

    // Complete frozen area [%] calculation for 9km gridcell
    ecoStateArrayObj_p->FrozenArea_9km[s9km] = (float32)ecoStateObj.ft_sum_9km /
                                               ((float32)cnt_3km_ftvalid_9km) *
                                               100.0;

    // <LAJ> 07/18/2014, The processing from 1km to 9km and 9km x Pft is now
    // done
    //  HERE instead of in post-processor. This allows us to operate on just the
    //  81 element array of 1km grid cells, rather than the entire global array

    // Aggregate from 1km to 9km here, and place results in 9km grid
    // Pass in 1km nested in 9km (81 element) state variable arrays
    // Populates 9km summary fields within STATE_ARRAY structure

    // NOTE on STARTING INDEX: "nee_mean" start index is 0, nee_pft{X}_mean
    // starts at index 2..
    l4c_PostProc_fltVar9km(s9km, pftObj_p, nee_meanChildName_p,
                           &ecoStateArrayObj_p->nee);

    // <JMG> 2014-10-01, Investigating FATAL SEGFAULT in
    // l4c_PostProc_fltVar9km() in SetBits call for NEE Since NEE is first and
    // this is where error occurs, see if placing GPP first causes GPP to blow
    // out
    l4c_PostProc_fltVar9km(s9km, pftObj_p, gpp_meanChildName_p,
                           &ecoStateArrayObj_p->gpp);

    // RH rhtot: starting index in master PkgMgr theme table is 20
    l4c_PostProc_fltVar9km(s9km, pftObj_p, rh_meanChildName_p,
                           &ecoStateArrayObj_p->rhtot);

    // SOC Ctot: starting index in master PkgMgr theme table is 30
    l4c_PostProc_fltVar9km(s9km, pftObj_p, soc_meanChildName_p,
                           &ecoStateArrayObj_p->Ctot);

    l4c_PostProc_fltVar9km(s9km, pftObj_p, ec_emult_avChildName_p,
                           &ecoStateArrayObj_p->Emult);

    // EC/Tmult_av: starting index in master PkgMgr theme table for Tmult is 42
    l4c_PostProc_fltVar9km(s9km, pftObj_p, ec_tmult_avChildName_p,
                           &ecoStateArrayObj_p->Tmult);

    // EC/Wmult_av: starting index in PkgMgr theme table for Wmult is 43
    l4c_PostProc_fltVar9km(s9km, pftObj_p, ec_wmult_avChildName_p,
                           &ecoStateArrayObj_p->Wmult);

    // Perform 9km and 9km x PFT aggregation for NEE QA. Does integer rank
    // classification after averaging. QA/carbon_model_qual_pft{1..8}. Last,
    // update QA for this 's9km' pixel in 9km space

    // <LAJ> 01/05/2016
    // The ecoStateArrayObj_p->nee_qa POINTER CONTAINS A FIELD WHICH POINTS TO
    // THE COMPUTED AVERAGE 'RMSE', ALL WE NEED TO DO IS TO WRITE OUT THIS FLOAT
    // ARRAY (RATHER THAN THE RANK ARRAY)
    l4c_PostProc_qaRank9km(s9km, pftObj_p, &ecoStateArrayObj_p->nee_qa);

  } // end :: 9km land vector loop

#ifdef DIAG_FPAR_QA
  // DIAG: can access populated diag array storing metObj.fpar_qa scalar as a
  // diag array
  l4c_DiagSparseArrayToImage("Diag_fpar_qa_scalar", "fpar_qa.pgm", 254,
                             EGv2_LDOM1KM, DFNT_UINT8, fparQaDiag_M01_p);
  l4c_GenFreqDistrib("fpar_qaCheck", "fpar_qaCheck", DFNT_UINT8, EGv2_LDOM1KM,
                     fparQaDiag_M01_p);
  // DIAG ONLY release memory for diag of "metObj.fpar_qa" in
  // fparQaDiag_M01_p[s1km]...
  memFree(fparQaDiag_M01_p);
#endif

  xLogMsg("\n");

  // pctQualified = (float32)nQualifyingCorePix / (float32)EGv2_LDOM9KM ;
  // use total of pixels referenced at innermost 1km loop as denominator
  pctQualified =
      ((float32)nQualifyingCorePix / (float32)nBplutCandidates) * 100.0;

  xLogMsg("@A In(%s) %zu of %zu PIXELS QUALIFIED (%.3f pct) AS TERRESTRIAL~\n",
          fxName, nQualifyingCorePix, nBplutCandidates, pctQualified);

  xLogMsg("@A In(%s) CONCLUDING DayOfYear (%s) TCF MODEL PROCESS ~\n", fxName,
          sesCtl_p->doy_Str);

  // <JMG> 2014-10-28T13:24:00, added, as part of main loop performance
  // optimizations
  xLogMsg("@A In(%s) BPLUT Size(%zu) %zu cache hits %zu cache refreshes from "
          "%zu candidates ~\n",
          fxName, sizeof(BPLUT), nBplutCacheHits, nBplutCacheRefreshes,
          nBplutCandidates);

#ifdef L4C_DIAG
  // DIAG ONLY: evaluate raw binary of nee...
  strcpy(testOutPath, "ecoStateArrayObj_p_nee.dom_1km.flt32");
  frm_ioWritebinary(testOutPath, DFNT_FLOAT32, sesCtl_p->nElem1kmPerTask,
                    ecoStateArrayObj_p->nee.dom_1km);
#endif
  // snapshot local process ending time and elapse time in seconds
  localEndiSec = time(0);
  localElapSec = difftime(localEndiSec, localStrtSec);

  xLogMsg("@A In(%s) FORWARD-STEP FINISHED (%s) requiring %.2lf elapsed sec "
          "(local)~ \n",
          fxName, frm_utilGetCtime(timeStr, 0), localElapSec);

  // update session-wise (globa) times
  sesCtl_p->sesEndiSec = time(0); // snapshot cumulative time thus far
  cumElapSec = sesCtl_p->sesEndiSec - sesCtl_p->sesStrtSec;
  xLogMsg("@A In(%s) session cumulative time required %.2lf elapsed sec "
          "(global)~ \n\n",
          fxName, cumElapSec);

  l4c_EventCompletionMsg(fxName);

} // end :: l4c_ModDriverGrid_forwardStep()
