// ---------------------------------------------------------------------
// module       : l4c_DAILY_FparQaBin.c , was orig fpar_qa_bin_util.c
// purpose      : Implement utility functions for modifying MODIS fpar qc bit fields
//                e.g MOD15A2 FPAR QA/QC and similar style bitfields
// project      : SMAP L4C science development.
// dependencies : HDF4, HDF5, FRM API libfrm.a
// revision     : Initial reformating -- laj 09/10/13
// author       : Lucas A. Jones; lucas@ntsg.umt.edu
// maint. author: Joe Glassy
// notes        : n/a


#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"


// ---------------------------------------------------------------------
// fpar_qa_bin_reclas()
// Purpose: High-level function to reclassify MOD15A2 FPAR QA/QC bit-field into four, ranked integer classes (0-3).
// Inputs: MOD15A2 FPAR QA/QC one byte bit-field.
// Outputs: Ranked integer FPAR quality class.
// Revised: date undocumented
// Notes  : Calls low-level function 'set_qa_state()'. The ranked integer class primarily depends on if clouds 
//        were present, and if clouds were not present, which algorithm was used for FPAR retrieval.
uint8 fpar_qa_bin_reclas(uint8 fpar_qa)
{
  uint8     rank;
  QcBitAdt   QCBIT;

  // Extract bits from QC field
  QCBIT.qc = fpar_qa;

  // break into the various QC fields
  bitword_parse(&QCBIT);

  // Re-classify qc conditions
  set_qa_state(QCBIT,&rank);

  return(rank);

} // end :: fpar_qa_bin_reclas()


// --------------------------------------------------------------------
// getBitsFromUint32()
// Adapted from: MODIS code.
//
// Purpose: this is an example of bit retrieval logic
//   using uint32 arguments. To work with uint8 (pixel) variables,
//   first load the uint8 value into a uint32 variable and pass in the
//   uint32 variable.

/*----------------------------------------------------------------------
!M  module  : l4c_FparQaBinUtil.c
!F  function: getBitsFromUint32(startBits,endBits,dataPixel)
!R  returns : uint32 -- up to 32 bits extracted, as an unsigned 32 bit int
!I  inputs  :int32 strt_at --starting base0 bit index, like 0.
             int32 end_at  --ending base0 bit index, like 1
             uint32 dta    -- a single raw pixel, expressed as a uint32 integer
!P  purpose : getBitsFromUint32() uses standard K&R ANSI C bit 
    extraction logic. The function extracts an embedded
    bitfield extending from base-0 bit position "strt_at"
    to "end_at" inclusive, from the variable "dta" and returns
    this result as a right-adjusted decimal value. On Intel architectures
    we assume a little-endian byte-ordering.
    The standard word layout we assume is depicted below, where
    MSB is 'most-significant-bit' and LSB is 'least-significant-bit) 
    regardless of whether the input word "dta" is a uint8, uint16 or
    uint32:
       
         MSB end                                 LSB end
         15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
          0  0  0  0  0  0  0  0  0  0  0  0  1  1  0  1
          
        Example: we begin with a (uint32) word whose 
        value in binary is '1101'. Assuming we wish to extract 
        bits {0,1}, e.g. the pattern '01', valued at decimal =1, we call
         modland_qc = getBitsFromUint32(0,1,thisPixelData)
!X--------------------------------------------------------------------*/
uint32 getBitsFromUint32(const int32 strt_at,const int32 end_at,uint32 dta)
{
 int32 p = end_at,             // anchor the start here 
       n = (end_at-strt_at)+1; // sum of bits (width) to retrieve 

 return (dta >> (p+1-n)) & ~(~0 << n);
 
} // getBitsFromUint32() 


// ---------------------------------------------------------------------
// function : set_qa_state()
// purpose  : Low-level function to reclassiy MOD15A2 FPAR QA/QC 
//            bit-field into four, ranked integer classes,valued at (0-3).
// Inputs   : QcBitAdt qcBitObj, 
//            uint8 fpar_rank, a (5-level) rank {0,1,2,3,4}
// Outputs  : via call argument pointer, Ranked integer FPAR quality class.
// Notes    : The ranked integer class primarily depends on 
//            if clouds were present, and if clouds were not present,
//            which algorithm was used for FPAR retrieval.
// 
void set_qa_state(QcBitAdt qcBitObj, uint8 *fpar_rank)
{

  // Worst conditions (QA=4): gapfill with climatology
  if (qcBitObj.fillval == 1 || qcBitObj.cloud != 0)
  {

    *fpar_rank = 4;

  } else if (qcBitObj.scf == 2 || qcBitObj.scf == 3) {
    // Poor conditions (QA=3): fpar backup alg.
    *fpar_rank = 3;

  } else if (qcBitObj.scf == 1)  {
    // Fair conditions (QA=2): fpar main RT with saturation
    *fpar_rank = 2;

  } else if (qcBitObj.scf == 0) {
    // Good conditions (QA=1): fpar main RT with saturation
    *fpar_rank = 1;
  }  else {
    *fpar_rank = 0;
  }

} // end :: set_qa_state()



// ---------------------------------------------------------------------
// function : bitword_parse()
// purpose  : Low-level function splits bit string into sub-units 
//            ('bitwords') aligned with FPAR QC fields
// inputs   : QcBitAdt *qcBitObj
// Outputs  : populated FPAR bitword structure.
// notes    : Bitword structure fields contain descriptive names for
//            each QC situation. 
// notes    : MOD15A2 FPAR QA/QC one byte bit-field.
// notes    : other L4_C functions use l4c_GetBits(),l4c_AssignBits()
//            l4c_SetBits() etc.
void bitword_parse(QcBitAdt *qcBitObj)
{

  // MODLAND_QC field
  qcBitObj->modland = (uint8)getBitsFromUint32(0,0,qcBitObj->qc); // <laj 06/02/2014> updated fpar qa bit read 

  // Sensor (Terra/Aqua) field
  qcBitObj->sensor = (uint8)getBitsFromUint32(1,1,qcBitObj->qc); // <laj 06/02/2014> updated qa bit read 

  // DeadDetector field
  qcBitObj->detector = (uint8)getBitsFromUint32(2,2,qcBitObj->qc); // <laj 06/02/2014> updated qa bit read 

  // CloudState field
  qcBitObj->cloud = (uint8)getBitsFromUint32(3,4,qcBitObj->qc); // <laj 06/02/2014> updated qa bit read 

  // SCF_QC field
  qcBitObj->scf = (uint8)getBitsFromUint32(5,6,qcBitObj->qc); // <laj 06/02/2014> updated qa bit read 

  // Fill values (from SFC_QC leftmost bit)
  qcBitObj->fillval = (uint8)getBitsFromUint32(7,7,qcBitObj->qc); // <laj 06/02/2014> updated qa bit read 

} // end :: set_qa_state()



// -------------------------------------------------------------------------------------------------------------
// printf_qcBitObjs()

// Purpose: Prints QCBIT structure fields.
//
// Notes: 
//
// Inputs: QCBIT structure fields.
//
// Outputs:
//

void printf_qcBitObjs(QcBitAdt qcBitObj)
{

  /*print complete binary representation to string
   *(for viewing) */
  printbit(qcBitObj.qc, qcBitObj.bits);

  printf("\n");
  printf("Entire QC: %s\n", qcBitObj.bits);
  printf("Modland:   %d\n", qcBitObj.modland);
  printf("Sensor:    %d\n", qcBitObj.sensor);
  printf("Detector:  %d\n", qcBitObj.detector);
  printf("Cloud:     %d\n", qcBitObj.cloud);
  printf("Scf:       %d\n", qcBitObj.scf);
  printf("Fillval:   %d\n", qcBitObj.fillval);
  printf("\n");

} // end :: printf_qcBitObjs()


/*Retrieves 'n' bits from the bit string buffer 'srcbuff' with location 
 *determined by 'offset' and puts them in string 'desbuff'  */ 

// -------------------------------------------------------------------------------------------------------------
// getbits()

// Purpose: Retrieves bits from a bit string buffer.
//
// Notes: 
//
// Inputs: Number of bits (n) to retrieve from the bit string buffer ('srcbuff') with array location ('offset').
//
// Outputs: Retrieved bits returned in string ('desbuff')
//

void getbits(char srcbuff[], char desbuff[], int offset, int n)
{

  strcpy(desbuff, "0\0");
  strncpy(desbuff,srcbuff+offset,n);
  desbuff[n+1] = '\0';

} // end :: getbits()

/*Extracts a string of bits from a single byte integer*/

// -------------------------------------------------------------------------------------------------------------
// printbits()

// Purpose: Extracts a string of bits from a single byte integer

void printbit(uint8 a, char buff[])
{

  int   i;
  int   n = CHAR_BIT;
  int   mask = 1 << (n-1);

  char  cbuff[MXSTRLEN] = {0};

  strcpy(buff, "\0");

  for (i = 1; i<=n; ++i){

    sprintf(cbuff,"%c",((a & mask) == 0) ? '0' : '1');
    
    strcat(buff,cbuff);
 
    a <<= 1;
      
  }
  strcat(buff,"\0");

} // end :: printbits()
