// ---------------------------------------------------------------
// function     : l4c_GridPost_Inflate.c
// purpose      : Perform all OUTPUT grid transformations for l4c_DAILY
//                which are primarily INFLATE operations transforming
//                FROM 1D sparse vector TO 2D full EASEGRID grids
// author       : joe glassy, Lupine Logic Inc.
// version      : see L4C_VERS_TAG , as 4.1.5 or later
// revised      : 2014-08-14, removed prep functions, moved to new
//                module file, l4c_DAILY_PrepOutput.c 
// revised      : 2014-05-19T14:26:00, jmg, renamed 
//                l4c_PrepOutput_HDF5_Datasets to "l4c_PrepOutput_HDF5_SOC_Datasets()
// revised      : v4.1.7-2014-02-05T11:242:00-07:00,wheat,jmg"
// dependencies : HDF4, HDF5 libraries and L4C FRAMEWORK API.
// history      : n/a
// notes        : n/a
// ---------------------------------------------------------------

#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"


// --------------------------------------------------------------- 
// Function : l4c_inflate_9km
// Purpose  : INFLATE function for 9km grid
// inputs   : SplandRefAdt SPLAND, void *src_p, void *dest_p, const uint32 dataType
// notes    : n/a
void l4c_inflate_9km(void *src_p, void *dest_p, const int hdf4Datatype)
{
  AncImagePkgAdt *ancImagePkgObj_p = l4c_GetAncImagePkgAdt();  	
  SplandRefAdt SPLAND  ;	
  
  uint32  l;
  size_t  srcSlot;
  size_t  destSlot;
  size_t  nelem = 1;

  size_t sOffset = 0;
  size_t dOffset = 0;
  size_t sSiz = DFKNTsize(hdf4Datatype); 
  size_t dSiz = DFKNTsize(hdf4Datatype);
  size_t nBytes = sSiz * nelem;
  uint8 *uSrc_p = (uint8 *)src_p ; 
  uint8 *uDest_p= (uint8 *)dest_p; 

	// ----------------------------------------------------------
	// STEP 1: Attach pre-existing {row,col} LUTs to local SPLAND
	SPLAND.row = ancImagePkgObj_p->imgM09land_row_uint16_p;
	SPLAND.col = ancImagePkgObj_p->imgM09land_col_uint16_p;

	// Transfer cells from 2D 9km grid to 9km sparse land vector*/
	for (l=0; l < LLAND9KM; l++)
	{
		destSlot = (size_t) MKGRID_2D_B0(SPLAND.row[l], SPLAND.col[l], NCOL9KM);
		srcSlot = (size_t)l;

		// Transfer data element from land vector to global 2D grid
		sOffset = sSiz * srcSlot ; 
		dOffset = dSiz * destSlot;
		memcpy(uDest_p+dOffset, uSrc_p+sOffset, nBytes);
		// copyUUTA(src_p, dest_p, srcSlot, destSlot, dataType, nelem);

	} //end:: land vec loop

	l4c_EventCompletionMsg( (char *)__func__);

} // end:: l4c_inflate_9km()


// --------------------------------------------------------------- 
// Function : l4c_inflate_3km()
// Purpose  : INFLATE function for 9km grid to convert FROM a 1D sparse
//            vector TO a full 2D grid
// inputs   : void *src_p, 
//            void *dest_p, 
//            const uint32 dataType
// 
// notes    : this is really just the same as the deflate functions 
// with the source and dest (and their respective indicies) swapped
//
void l4c_inflate_3km(void *src_p, void *dest_p, const int hdf4Datatype)
{
  AncImagePkgAdt *ancImagePkgObj_p   = l4c_GetAncImagePkgAdt();  	
  SplandRefAdt SPLAND  ;	

  uint32  l;
  uint8   rn, cn;
  uint8   serialIdx_nested;
  size_t  srcSlot;
  size_t  destSlot;
  size_t  nelem = 1;

  size_t sOffset = 0;
  size_t dOffset = 0;
  size_t sSiz = DFKNTsize(hdf4Datatype); 
  size_t dSiz = DFKNTsize(hdf4Datatype);
  size_t nBytes = sSiz * nelem;
  uint8 *uSrc_p = (uint8 *)src_p ; 
  uint8 *uDest_p= (uint8 *)dest_p; 

	// ----------------------------------------------------------
	// STEP 1: Attach pre-existing {row,col} LUTs to local SPLAND
	SPLAND.row = ancImagePkgObj_p->imgM09land_row_uint16_p;
	SPLAND.col = ancImagePkgObj_p->imgM09land_col_uint16_p;

	//  Transfer cells from 2D 9km grid to 9km sparse land vector*/
	for (l=0; l < LLAND9KM; l++)
	{
		for (rn=0; rn < NESTED_3KM_IN_9KM; rn++)
		{
			for (cn=0; cn < NESTED_3KM_IN_9KM; cn++)
			{
			// Reference 3km nested grid
			destSlot = (size_t) MKGRID_2D_N9_B0(SPLAND.row[l], SPLAND.col[l], rn, cn, NESTED_3KM_IN_9KM, NCOL3KM);
	      
			// Keep 3km nested cells in same relative orientation
			serialIdx_nested = MKGRID_2D_B0(rn, cn, NESTED_3KM_IN_9KM);
			srcSlot = (size_t) MKGRID_2D_B0(l, serialIdx_nested, NESTED_3KM_IN_9KM_SQ);

			// Transfer data element from land vector to global 2D grid
			// copyUUTA(src_p, dest_p, srcSlot, destSlot, hdf4Datatype, nelem);
			// try manual call..
			sOffset = sSiz * srcSlot ; 
			dOffset = dSiz * destSlot;
			memcpy(uDest_p+dOffset, uSrc_p+sOffset, nBytes);
			// copyUUTA(src_p, dest_p, srcSlot, destSlot, dataType, nelem);

			} //end:: col loop nested 3km
		} // end:: row loop nested 3km
    } // end:: land vec 9km loop
    
	l4c_EventCompletionMsg( (char *)__func__);
	
} // end::l4c_inflate_3km()


// --------------------------------------------------------------- 
// purpose:  INFLATE function for nested 1km grid
// inputs : SplandRefAdt SPLAND, void *src_p, void *dest_p, const uint32 dataType
// notes  :  this is really just the same as the deflate 
//           functions with the source and dest 
//           (and their respective indicies) swapped
//
void l4c_inflate_1km(void *src_p, void *dest_p, const int hdf4Datatype)
{
  AncImagePkgAdt *ancImagePkgObj_p   = l4c_GetAncImagePkgAdt();  	
  SplandRefAdt SPLAND  ;	

  uint32  l;
  uint8   rn, cn;
  uint8   serialIdx_nested;

  size_t  srcSlot;
  size_t  destSlot;
  size_t  nelem = 1;

  size_t sOffset = 0;
  size_t dOffset = 0;
  size_t sSiz = DFKNTsize(hdf4Datatype); 
  size_t dSiz = DFKNTsize(hdf4Datatype);
  size_t nBytes = sSiz * nelem;
  uint8 *uSrc_p = (uint8 *)src_p ; 
  uint8 *uDest_p= (uint8 *)dest_p; 

	// ----------------------------------------------------------
	// STEP 1: Attach pre-existing {row,col} LUTs to local SPLAND
	SPLAND.row = ancImagePkgObj_p->imgM09land_row_uint16_p;
	SPLAND.col = ancImagePkgObj_p->imgM09land_col_uint16_p;

	// Transfer cells from 2D 9km grid to 9km sparse land vector
	for (l=0; l < LLAND9KM; l++)
    {
		for (rn=0; rn < NESTED_1KM_IN_9KM; rn++)
		{
			for (cn=0; cn < NESTED_1KM_IN_9KM; cn++)
			{

			// Reference 1km nested grid
			destSlot = (size_t) MKGRID_2D_N9_B0(SPLAND.row[l], SPLAND.col[l], rn, cn, NESTED_1KM_IN_9KM, NCOL1KM);
	      
			//keep 1km nested cells in same relative orientation
			serialIdx_nested = MKGRID_2D_B0(rn, cn, NESTED_1KM_IN_9KM);
			srcSlot = (size_t) MKGRID_2D_B0(l, serialIdx_nested, NESTED_1KM_IN_9KM_SQ);
			// Transfer data element from land vector to global 2D grid
			sOffset = sSiz * srcSlot ; 
			dOffset = dSiz * destSlot;
			memcpy(uDest_p+dOffset, uSrc_p+sOffset, nBytes);
			// copyUUTA(src_p, dest_p, srcSlot, destSlot, dataType, nelem);
			
			} //end:: col loop nested 1km
		} //end:: row loop nested 1km
    } //end:: land vec 9km loop
    
	l4c_EventCompletionMsg( (char *)__func__);
	
} // end:: l4c_inflate_1km()



// --------------------------------------------------------------- 
// function : l4c_inflate_init_9km
// inputs   : void *dest_p, const uint32 dataType
// returns  :
// notes    : Verify: 3/8/2014T09:15, there appears to be a 
//            fatal bug upon entry here,
//            where atSlot==0, causing a seg fault on first
//            dest_p[] variable submitted.. investigation in 
//            progress. dest_p DOES appear to have been 
//            allocated with enough space and of correct datatype.
void l4c_inflate_init_9km(void *dest_p, const int hdf4Datatype)
{
  uint32  r, c;
  size_t  destSlot;

  	for (r=0; r < NROW9KM; r++)
    {
		for (c=0; c < NCOL9KM; c++)
		{
			destSlot = (size_t) MKGRID_2D_B0(r, c, NCOL9KM);
			// set_fillval_UUTA(dest_p, hdf4Datatype, destSlot);
			l4c_AssignFillvalue(hdf4Datatype, destSlot, dest_p);
		} //end:: col 9km loop
		
	} //end:: row 9km loop
	
  l4c_EventCompletionMsg( (char *)__func__);
  	
} // end::l4c_inflate_init_9km()


// --------------------------------------------------------------- 
// function : l4c_inflate_init_3km
// purpose  : INFLATE a 3km grid FROM a 1d sparse vector TO a 
//            full 2D grid
// inputs   : void *dest_p, const uint32 dataType
// returns  :
// notes    : n/a
void l4c_inflate_init_3km(void *dest_p, const int hdf4Datatype)
{

  uint32  r, c, rn, cn;

  size_t  destSlot;

  
	for (r=0; r < NROW9KM; r++)
    {
		for (c=0; c < NCOL9KM; c++)
		{
			for (rn=0; rn < NESTED_3KM_IN_9KM; rn++)
			{
				for (cn=0; cn < NESTED_3KM_IN_9KM; cn++)
				{
				//  Reference 3km nested grid 
				destSlot = (size_t)MKGRID_2D_N9_B0(r, c, rn, cn, NESTED_3KM_IN_9KM, NCOL3KM);
				// set_fillval_UUTA(dest_p, hdf4Datatype, destSlot);				
				l4c_AssignFillvalue(hdf4Datatype, destSlot, dest_p);
				} //end:: col 3km nested loop
			}//end:: row 3km nested loop
	  
		} //end:: col 9km loop
		
    } //end:: row 9km loop
    
  	l4c_EventCompletionMsg( (char *)__func__);
  	
} // end:: l4c_inflate_init_3km()


// --------------------------------------------------------------- 
// function : l4c_inflate_init_1km
// purpose  : initialize a 1km grid
// inputs   : void *dest_p, 
//          : const uint32 HDF4 style dataType, like DFNT_FLOAT32 etc
// returns  : void
// notes    : n/a
void l4c_inflate_init_1km(void *dest_p,const int hdf4Datatype)
{
  uint32  r, c, rn, cn;
  size_t  destSlot;

	for (r=0; r < NROW9KM; r++)
    {
		for (c=0; c < NCOL9KM; c++)
		{
			for (rn=0; rn < NESTED_1KM_IN_9KM; rn++)
			{
				for (cn=0; cn < NESTED_1KM_IN_9KM; cn++)
				{
		  
					// Reference 1km nested grid 
					destSlot = (size_t)MKGRID_2D_N9_B0(r, c, rn, cn, NESTED_1KM_IN_9KM, NCOL1KM);
					// set_fillval_UUTA(dest_p, hdf4Datatype, destSlot);
					l4c_AssignFillvalue(hdf4Datatype, destSlot, dest_p);	
				} // end:: col 1km nested loop
			}  //  end:: row 1km nested loop
		} //end:: col 9km loop
    } //end:: row 9km loop
    
  	l4c_EventCompletionMsg( (char *)__func__);
  	
} // end::l4c_inflate_init_1km()


// ------------------------------------------------------------
// function : l4c_AssignFillvalue()
// inputs   : const int hdf4Datatype, const size_t atSlot,void *vDest_p
// returns  : size_t, size in bytes of scalar instance of indicated type
// purpose  : Returns the size in bytes of one scalar of indicated type. 
//            This is a generic substitute for HDF4's DFKNTsize() to allow 
//            for small apps to not have to link against HDF4 libs
// notes    : Moved here from l4c_DAILY_sci.c on 2014-11-03 so it belongs
//            with functions most often using it.
void l4c_AssignFillvalue(const int hdf4Datatype, const size_t atSlot,void *vDest_p)
{
  unsigned char *destByte_p = (unsigned char *)vDest_p;
  
  // size_t sizOfOne   = l4c_sizOfDatatype(hdf4Datatype);
  size_t byteOffset = atSlot * l4c_sizOfDatatype(hdf4Datatype) ;

   int8    chrFill = -127 ;
   uint8   bytFill = BYTE_FILL_VALUE   ; // 254 from l4c_DAILY.h
   uint16  u16Fill = USHORT_FILL_VALUE ; // 65534 from l4c_DAILY.h
   int16   i16Fill = -32766     ;
   int32   i32Fill = 0 ; // TBD refine, get from limits.h ?
   uint32  u32Fill = 0 ; // TBD regine, from limits.h ?
   int64   i64Fill = 0 ; // TBD refine...
   uint64  u64Fill = 0 ; // TBD refine...
   float32 fltFill = FLOAT_FILL_VALUE; // -9999.0 from l4c_DAILY.h
   float64 dblFill = FLOAT_FILL_VALUE; // same as for float32 for now
  
	// Set fill value based on sync'd HDF4 <--> TIV data type(s) 
	switch(hdf4Datatype)
    {
    case IsChar    : memcpy(destByte_p + byteOffset, &chrFill, sizeof(char)   ); break;
    case IsUint8   : memcpy(destByte_p + byteOffset, &bytFill, sizeof(char)   ); break;
    case IsInt16   : memcpy(destByte_p + byteOffset, &i16Fill, sizeof(int16)  ); break;
    case IsUint16  : memcpy(destByte_p + byteOffset, &u16Fill, sizeof(uint16) ); break;
    case IsInt32   : memcpy(destByte_p + byteOffset, &i32Fill, sizeof(int32)  ); break;
    case IsUint32  : memcpy(destByte_p + byteOffset, &u32Fill, sizeof(uint32) ); break;
    case IsInt64   : memcpy(destByte_p + byteOffset, &i64Fill, sizeof(int64)  ); break;
    case IsUint64  : memcpy(destByte_p + byteOffset, &u64Fill, sizeof(uint64) ); break;
    case IsFloat32 : memcpy(destByte_p + byteOffset, &fltFill, sizeof(float32)); break;
    case IsFloat64 : memcpy(destByte_p + byteOffset, &dblFill, sizeof(float64)); break;
    default :
		xLogMsg("@F In(%s) TIV or HDF4 datatype %d is NOT yet supported!\n",
			(char *)__func__, hdf4Datatype);
		break;
    } // end switch on datatype
	
} // end:: l4c_AssignFillvalue()


// --------------------------------------------------------------------
// function : set_fillval_UUTA()
// purpose  : Set fill value for an array of a given data type at the 
//            specified location in the array
// inputs   :  void *vDest_p, const int32 dataType, const size_t atSlot
// outputs  :
// returns  :
// notes    : Assert: the TIV datatypes should now ALL be sync'd with the
//          : HDF4 datatypes.
// notes    : TBD: Ideally, these should not be hard-coded here but refer to 
//            header file macro values
//
// bugs     : Verify: 3/8/2014T09:15, there appears to be a 
//            fatal bug upon entry here, where atSlot==0, causing a 
//            seg fault on first access to load a pixel into  dest_p[] 
//            Investigation in progress. dest_p DOES appear to have been 
//            allocated with enough space and of correct datatype, and it
//            can be fully traversed (and test populated) prior to
//            entering this function. 
//
void set_fillval_UUTA(void *vDest_p, const int hdf4Datatype, const size_t atSlot)
{
  TI_ATOM_PTR *dst_p = (TI_ATOM_PTR *)vDest_p;
  
  size_t i = atSlot;

	// Set fill value based on sync'd HDF4 <--> TIV data type(s) 
	switch(hdf4Datatype)
    {
    case IsChar    : dst_p->aInt8_p[   i ] = (int8)         -127; break;
    case IsUint8   : dst_p->aUint8_p[  i ] = (uint8)         254; break;
    case IsInt16   : dst_p->aInt16_p[  i ] = (int16)      -32767; break;
    case IsUint16  : dst_p->aUint16_p[ i ] = (uint16)      65534; break;
    case IsInt32   : dst_p->aInt32_p[  i ] = (int32)   -2.1474e9; break;
    case IsUint32  : dst_p->aUint32_p[ i ] = (uint32)  -4.2950e9; break;
    case IsInt64   : dst_p->aInt64_p[  i ] = (int64)  -9.2233e18; break;
    case IsUint64  : dst_p->aUint64_p[ i ] = (uint64)  1.8446e19; break;
    case IsFloat32 : dst_p->aFlt32_p[  i ] = (float32)  -9.999e9; break;
    case IsFloat64 : dst_p->aFlt64_p[  i ] = (float64)  -9.999e9; break;
    } // end switch on datatype
	
} // end:: set_fillval_UUTA()

