// ---------------------------------------------------------------
// function     : l4c_GridPre_Deflate.c
// purpose      : perform all INPUTT grid transformations for l4c_DAILY
//                which are primarily DEFLATE operations transforming
//                FROM 2D full EASEGRID grids TO 1D sparse vectors.
// author       : joe glassy, Lupine Logic Inc.
// version      : see L4C_VERS_TAG , as 4.1.5 or later
// revised      : 2014-02-28T13:10, hit fatal error using COPY_UUTA so
//                 coded a manual version of this in-line that works. Will
//                 investigate later..but meanwhile, changed all the DEFLATE
//                 {1km,3km,9km} routines to do the UUTA step manually.

// history      : 2014-02-27T12:31:00, with new re-org of PkgInits..
// history      : 2014-02-26T11:56:00, replaced core spland code 
//                with l4c_ prefaced instances. 
// dependencies : HDF4, HDF5 libraries and L4C FRAMEWORK API.
// history      : n/a
// ---------------------------------------------------------------

#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"


// ---------------------------------------------------------------
// function  : l4c_deflate_1km
// purpose   : Perform DEFLATE on a EASEGRIDv2 1k full grid 2D image, of HDF4 datatype 'h4Datatype
// inputs    : uint16 *row_1D_LUT_p,uint16 *col_1D_LUT_p, void *src_p,void *dest_p,const int h4Datatype
// returns   : size_t, count of pixels processed
// notes     : This is an L4_C adaptation of spland_deflate_1km(). Note 
//             that the n. of pixels to process here is currently hard
//             wired using LLAND9KM, NESTED_1KM_IN_9KM, etc
size_t l4c_deflate_1km(uint16 *row_1D_LUT_p,uint16 *col_1D_LUT_p, void *src_p,void *dest_p,const int hdf4Datatype)
{
  size_t  ith_1D_sparse =0;
  size_t  srcSlot;
  size_t  destSlot;
  // size_t  p;
  size_t  nPixProcessed = 0;    
  int   rn =0, cn=0;
  int   serialIdx_nested=0;
  
  time_t secNow = time(0);
  char timeStr[32] = {0};
  
  uint8 *uSrc_p = (uint8 *)src_p ; 
  uint8 *uDest_p= (uint8 *)dest_p; 
  
  size_t nelem   = 1;
  size_t sOffset = 0;
  size_t dOffset = 0;
  size_t sSiz = DFKNTsize(hdf4Datatype); 
  size_t dSiz = DFKNTsize(hdf4Datatype);
  size_t nBytes = sSiz * nelem;

    // xLogMsg("@A DEFLATE_%s started %s\n", __func__, frm_utilGetCtime(timeStr,secNow) );
    
    // stop early if something in the config setup failed.
    assert(sSiz == dSiz && nBytes >= 1 && (sSiz >= 1 && sSiz <= sizeof(float64)) );
    
    //  Transfer cells from 2D 9km grid to 9km sparse land vector
    for (ith_1D_sparse=0; ith_1D_sparse < LLAND9KM; ith_1D_sparse++)
    {
        for (rn=0; rn < NESTED_1KM_IN_9KM; rn++)
        {
            for (cn=0; cn < NESTED_1KM_IN_9KM; cn++)
            {
                // Reference 1km nested grid 
                srcSlot = (size_t) MKGRID_2D_N9_B0(row_1D_LUT_p[ith_1D_sparse], col_1D_LUT_p[ith_1D_sparse],
                        rn, cn, NESTED_1KM_IN_9KM, NCOL1KM);
          
                // Keep 1km nested cells in same relative orientation
                serialIdx_nested = MKGRID_2D_B0(rn, cn, NESTED_1KM_IN_9KM);
                
                destSlot = (size_t) MKGRID_2D_B0(ith_1D_sparse,
                        serialIdx_nested, NESTED_1KM_IN_9KM_SQ);
                
                // Transfer data element from global grid to land vector
                // copyUUTA(src_p, dest_p, srcSlot, destSlot, hdf4Datatype, nelem);
                // try manual call..
                sOffset = sSiz * srcSlot ; 
                dOffset = dSiz * destSlot;
                memcpy(uDest_p+dOffset, uSrc_p+sOffset, nBytes);
                nPixProcessed++ ; // increment check count on pixels processed
                
            } //end:: col loop nested 1km
            
        } //end:: row loop nested 1km
    } //end:: land vec loop
    
    l4c_EventCompletionMsg(__func__);
    
    return nPixProcessed ;
        
} // end:: l4c_deflate_1km()


// ---------------------------------------------------------------
// function  : l4c_deflate_3km
// purpose   : Perform 3KM DEFLATE on a EASEGRIDv2 3km full grid 2D image, of HDF4 datatype 'h4Datatype
// inputs    : uint16 *row_1D_LUT_p,uint16 *col_1D_LUT_p, void *src_p,void *dest_p,const int h4Datatype
// returns   : size_t, total n. of pixels processed
// notes     : this is an L4_C adaptation of spland_deflate_3km()
size_t l4c_deflate_3km(uint16 *row_1D_LUT_p,uint16 *col_1D_LUT_p, 
                    void *src_p,void *dest_p,const int hdf4Datatype)
{
  int  rn, cn;
  size_t  serialIdx_nested;
  size_t  srcSlot;
  size_t  destSlot;
  size_t  nelem = 1;
  size_t  nPixProcessed = 0;
  size_t  ith_1D_sparse =0;

  uint8 *uSrc_p = (uint8 *)src_p ; 
  uint8 *uDest_p= (uint8 *)dest_p; 
  
  size_t sOffset = 0;
  size_t dOffset = 0;
  size_t sSiz = DFKNTsize(hdf4Datatype); 
  size_t dSiz = DFKNTsize(hdf4Datatype);
  size_t nBytes = sSiz * nelem;
  time_t secNow = time(0);
  char timeStr[32] = {0};
  
  
    xLogMsg("@A DEFLATE_%s started %s\n", __func__, frm_utilGetCtime(timeStr,secNow));
    
    // stop early if something in the config setup failed.
    assert(sSiz == dSiz && nBytes >= 1 && (sSiz >= 1 && sSiz <= sizeof(float64)) );

  // Transfer cells from 2D 9km grid to 9km sparse land vector
    for (ith_1D_sparse =0; ith_1D_sparse < LLAND9KM; ith_1D_sparse++)
    {
        for (rn=0; rn < NESTED_3KM_IN_9KM; rn++)
        {
            for (cn=0; cn < NESTED_3KM_IN_9KM; cn++)
            {

                // Reference 3km nested grid 
                srcSlot = (size_t) MKGRID_2D_N9_B0(row_1D_LUT_p[ith_1D_sparse],
                        col_1D_LUT_p[ith_1D_sparse],    rn, cn, NESTED_3KM_IN_9KM, NCOL3KM);
          
                // Keep 3km nested cells in same relative orientation
                serialIdx_nested = MKGRID_2D_B0(rn, cn, NESTED_3KM_IN_9KM);
                
                destSlot = (size_t) MKGRID_2D_B0(ith_1D_sparse,
                        serialIdx_nested, NESTED_3KM_IN_9KM_SQ);
        
                // Transfer data element from global grid to land vector
                // copyUUTA(src_p, dest_p, srcSlot, destSlot, hdf4Datatype, nelem);
                sOffset = sSiz * srcSlot ; 
                dOffset = dSiz * destSlot;
                memcpy(uDest_p+dOffset, uSrc_p+sOffset, nBytes);
                nPixProcessed++ ; // increment check count on pixels processed
                
            } // end:: col loop nested 3km
        } // end:: row loop nested 3km
    } // end:: land vec loop

    l4c_EventCompletionMsg(__func__);

    return nPixProcessed ;
    
} // end:: l4c_deflate_3km()


// ---------------------------------------------------------------
// function : l4c_deflate_9km()
// purpose  : DEFLATE a 9k grid
// inputs   : uint16 *row_1D_LUT_p,
//            uint16 *col_1D_LUT_p, 
//            void *src_p, -- source, 2D FULLGRID 9KM raster
//            void *dest_p -- destination 1D-sparse vector array
//           : const int hdf4Datatype, HDF4 style datatype identifier
// returns   : size_t, count of pixels processed
// notes    : n/a
size_t l4c_deflate_9km(uint16 *row_1D_LUT_p,uint16 *col_1D_LUT_p, void *src_p,void *dest_p,const int hdf4Datatype)
{
  size_t  srcSlot;
  size_t  destSlot;
  size_t  nelem = 1;
  size_t  nPixProcessed = 0;  
  size_t  ith_1D_sparse = 0; //base0 index of successive 1D sparse pixels in land domain

  time_t secNow = time(0);
  char timeStr[32] = {0} ;
  
  uint8 *uSrc_p = (uint8 *)src_p ; 
  uint8 *uDest_p= (uint8 *)dest_p; 
  
  size_t sOffset = 0;
  size_t dOffset = 0;
  size_t sSiz = DFKNTsize(hdf4Datatype); 
  size_t dSiz = DFKNTsize(hdf4Datatype);
  size_t nBytes = sSiz * nelem;

    xLogMsg("@A DEFLATE_%s started %s\n", __func__, frm_utilGetCtime(timeStr,secNow));
    // stop early if something in the config setup failed.
    assert(sSiz == dSiz && nBytes >= 1 && (sSiz >= 1 && sSiz <= sizeof(float64)) );

  // Transfer cells from 2D 9km grid to 9km sparse land vector
  for (ith_1D_sparse =0; ith_1D_sparse < LLAND9KM; ith_1D_sparse++)
    {
      srcSlot = (size_t) MKGRID_2D_B0(row_1D_LUT_p[ith_1D_sparse],
                                col_1D_LUT_p[ith_1D_sparse], NCOL9KM);
                                
      destSlot = (size_t) ith_1D_sparse;
      
        // Transfer data element from global grid to land vector
        // copyUUTA(src_p, dest_p, srcSlot, destSlot, hdf4Datatype, nelem);
        sOffset = sSiz * srcSlot ; 
        dOffset = dSiz * destSlot;
        memcpy(uDest_p+dOffset, uSrc_p+sOffset, nBytes);
        nPixProcessed++ ; // increment check count on pixels processed

    } //end:: land vec loop

    l4c_EventCompletionMsg(__func__);
    
    return nPixProcessed ;

} // end::l4c_deflate_9km
