// ----------------------------------------------------------------
// module       : l4c_DAILY_Hdf5_Read_and_Updates.c
// purpose      : Implement (container for) all HDF5 input reads
// author       : Joe Glassy
// project      : SMAP L4C science development.
// dependencies : HDF5 lib, FRM API lib
// revision     : 2016-01-11, <JMG> added logic to AUTO-RESET the FT and SOC 
//                annual Day-of-Year {1..366} calendar slots to 0 when January 1 encountered
//                since otherwise they essentially overwrite(lose track) of the current
//                annual periods state. This logic still isn't perfect for multi-year
//                running, but..its an improvement.
// revision     : 2014-11-11, added support for near future enabling of
//                ft_calendar_cumulative[], soc_calendar_cumulative[]
//                cross-check arrays to warn when a given day has been
//                run multiple times (should be run just once). Note that
//                enabling this cross-check logic is only possible when
//                upgrade to the newest SMAP_L4_C_SOC.h5 that contains
//                these long-lived attributes (ft_calendar_cumulative and
//                soc_calendar_cumulative). A hard-wired boolean switch named
//                "forceStrict_SOC_PoolUpdatePolicy" (permissive=0,strict=1)
//                now set to permissive, controls this behavior.
// history      : 2014-09-30, refactored to split out each category of
//                SOC variable into its own driver and handler etc.
// history      : 2014-05-09T10:36:00, added extra calendar handling of
//                soc_updated_datetime and "ft_updated_datetime" to match
//                v5.0.1 SMAP_L4_C_SOC.h5 daily intermediate file.
// revision     : 2014-03-11T13:20, leave SOC and other files open now
// revision     : 2014-03-05T16:00, moved l4c_READ_M09land_row_col_9km() to 
//                this 'neutral' module since it is used in BOTH the
//                pre and post processors.
// revision     : 2014-02-21T14:36, refactored out of l4c_DAILY_Hdf5.c etc
// notes        : n/a

#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"

// ---------------------------------------------------------------
// Function: l4c_ReadAnc_M09land_row_col_9km()
// Purpose : Load ancil row/col data for sparse format
// inputs  : void
// notes   : This is used in both DEFLATE and INFLATE processes, 
//           as the
//         : 1D-sparse vector (2) LUTs stored in 
//            SMAP_L4_C_Ancillary.h5 :
//           DATASET "MCD12Q1_M09land_col_uint16" 
//           DATASET "MCD12Q1_M09land_row_uint16" 
// revised : 2014-03-05T16:00, moved from l4c_PreProc_func.c to
//           here in l4c_DAILY_Hdf5_Read_and_Updates() since it is
//           used in BOTH pre and post processing phases
// returns : int, status
// notes   : the NSB database is a name-value-pair stringblock in memory
//           database of NVP's parsed out of the RIP file.
int l4c_ReadAnc_M09land_row_col_9km(void)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();
 
 PkgMgr *pkgMgr_p = l4c_GetPkgMgrOverseer() ; // fetch ptr to overseer PkgMgr
 // fetch ptr to Ancillary Dataset package
 AncImagePkgAdt *ancImg_p = l4c_GetAncImagePkgAdt(); 
 
 PkgTheme *pThme_p= NULL;
 
 size_t nPix9km_Sparse = EGv2_LDOM9KM ; // n. 9km pix in Sparse (land) domain
 hid_t file_id = -1;
 herr_t ret    = -1; 
 char fxName[64];
 
 
    // fetch handle to previsouly Open SMAP_L4_C_Ancillary.h5
    file_id = sesCtl_p->anc_file_id ;
    
    // Allocate (2) uint16 M09land sized arrays
    ancImg_p->imgM09land_row_uint16_p = (uint16*)memCalloc(nPix9km_Sparse, sizeof(uint16));
    ancImg_p->imgM09land_col_uint16_p = (uint16*)memCalloc(nPix9km_Sparse, sizeof(uint16));
    
    // READ in these (2) 1D sparse {row,col} LUTs
    strcpy(fxName, (char *)__func__);

    ret = frm_hdf5_ReadData(file_id, H5T_NATIVE_USHORT,"MCD12Q1_M09land_row_uint16",
        ancImg_p->imgM09land_row_uint16_p , fxName);
        
    ret = frm_hdf5_ReadData(file_id, H5T_NATIVE_USHORT,"MCD12Q1_M09land_col_uint16",
        ancImg_p->imgM09land_col_uint16_p , fxName );

    // BIND the row,col LUTs to their (child) PkgObj instances right away
    //  This allows us to access the arrays from anywhere, via calls to:
    //   pkgMgr_GetThemes_ChildDataByName(pThme, "MCD12Q1_M09land_row_uint16");

    // Set pointer into the theme called "SMAP_L4_C_Ancillary" 
    pThme_p  = pkgMgr_GetThemeByName(pkgMgr_p,"SMAP_L4_C_Ancillary");
    pkgMgr_BindObjByName(pThme_p,"MCD12Q1_M09land_row_uint16",ancImg_p->imgM09land_row_uint16_p);
    pkgMgr_BindObjByName(pThme_p,"MCD12Q1_M09land_col_uint16",ancImg_p->imgM09land_col_uint16_p);
    
    // leave anc file open...
    l4c_EventCompletionMsg( fxName);
    
    return 0 ;

} // end:: l4c_ReadAnc_M09land_row_col_9km()


// ----------------------------------------------------------------
// function     : l4c_SOC_or_FT_CalendarUpdate()
// purpose      : service function to RESET(when opMode=0) or
//                "update-to-doy" if 1, for either Freeze Thaw FT or
//                SOC check off calendar array (attribute /soc_calendar
//                or /ft_calendar of type uint8
// inputs       : const int opMode,
//              : const int curr_doy {1..366} as base1 day of year,
//              :  uint8 *soc_CalendarArray_p -- ptr to calendar
// revised      : 2016-01-11, <JMG> found+fix minor bug in calendar assignment 
//                where index 'curr_day' should have ben 'd' in loop below, and
//                added logic to auto-reset FT and SOC calenders when Jan 1
//                is encountered.
//
// notes        : REMEMBER that ALL calendar "checkoff" slot usage is base1, 
//                not base0 and note this same service function is used for EITHER
//                the FT calendar and the SOC calendars, stored in
//                AncImagePkgAdt as FT_calendar
//
// notes: L4C_N_CALENDAR_SLOTS is 370 typically (allows for base1 usage with extra slots)
//
//  PROCESSING CALENDAR PROTOCOLS USED:
//  opMode =0 adjusts JUST the curr_doy slot to 1 (this is the default behavior in operational mode)
//  opMode =1 forces a ZERO OUT of all day slots in calendar.
//  opMode =2 adjusts all slots to 1 from beginning of year to curr_doy, inclusive.

// ADDED calendarCumulative[] dayofyear slot arrays, where for each time we 
// execute the model on a given dayOfYear, we INCREMENT that slot's count,
// such that at any time later, we can test when a given slots count is >1 we 
// then know this dayOfYear has been repetitively executed that number of times
// when ideally it SHOULD be only executed ONCE. When the SOC pools for 
// a given dayOfyear are executed >1, this gives a false build-up that pools
// contribution, so we want to DETECT & WARN user if this happens using this
// cross-check.
int l4c_SOC_or_FT_CalendarUpdate(const int opMode,
    const char *calendarLabel_p,const int curr_doy)
{
 AncImagePkgAdt  *ancImagePkgObj_p = l4c_GetAncImagePkgAdt();
 SesCtl *sesCtl_p = l4c_GetSesCtl();
 
 int d =0;
 char fxName[256]= {0};
 hid_t attr_id = -1 ;
 hid_t soc_id = sesCtl_p->soc_file_id ;
 herr_t ret = -1;
 
    strcpy(fxName,(char *)__func__);

    // enforce, since curr_doy should never exceed 366 anyway
    assert( curr_doy <  L4C_N_CALENDAR_SLOTS) ; 

    xLogMsg( "@A In(%s) updating calendar (%s) using policy opMode %d~ \n",
        fxName,  calendarLabel_p, opMode);

    // <JMG> 2016-01-11, added logic to pre-emptively AUTO-RESET all FT and SOC 
    // annual Day-of-Year {1..366} calendar slots to 0 when January 1 encountered,
    // since otherwise they essentially overwrite(lose track) of the current
    // annual period relative to last annual period state. 
    // This logic still isn't perfect for multi-year running, (e.g. what
    // happens if Jan 1 is skipped) but, its an improvement.
    // TBD: store a special 'year-turn-over' attribute in long lived attribute
    //  file like SMAP_L4C_Ancillary.h5 or in the daily intermediate SOC file to
    //  indicate that this "turn-over reset" logic has been accomplished and when
    if( curr_doy == 1 )
    {
      xLogMsg( "@A In(%s) Annual period threshold encountered (DOY %d) resetting all annual calendars~ \n",
        fxName,  curr_doy);

        for(d=0; d < L4C_N_CALENDAR_SLOTS; d++)
        {
            ancImagePkgObj_p->SOC_CalendarCheck[d] =0 ; 
            ancImagePkgObj_p->FT_CalendarCheck[d] =0  ;
            ancImagePkgObj_p->SOC_CalendarCumulative[d] = 0 ; 
            ancImagePkgObj_p->FT_CalendarCumulative[d] = 0  ; 
        } // end:: d loop through dayOfYear indices
    } // end:: block when JANUARY 1 (year turn-over) encountered

            
    if(opMode == 0)
    {
        // ADJUST completion status ONLY this current DOY to '1',
        // to explicitly mark it as processed this day
        
        ancImagePkgObj_p->SOC_CalendarCheck[curr_doy] =1 ; // set 1 in slot 1..366 when day processed
        ancImagePkgObj_p->FT_CalendarCheck[curr_doy] =1  ; // set 1 in slot 1..366 when day processed
        
        // JMG> 2014-11-11, added cross-check, cumulative form, where for each model run instance
        // we INCREMENT current value of the given day slot instead of assigning a 1 to it
        ancImagePkgObj_p->SOC_CalendarCumulative[curr_doy]++ ; 
        ancImagePkgObj_p->FT_CalendarCumulative[curr_doy]++  ; 

    } else if(opMode == 1) { // FORCE RESET of just current day

        for(d=0; d < L4C_N_CALENDAR_SLOTS; d++)
        {
            ancImagePkgObj_p->SOC_CalendarCheck[curr_doy] =0 ; 
            ancImagePkgObj_p->FT_CalendarCheck[curr_doy ] =0  ;
            ancImagePkgObj_p->SOC_CalendarCumulative[curr_doy] = 0 ; 
            ancImagePkgObj_p->FT_CalendarCumulative[curr_doy ] = 0  ; 
        } // end:: d loop, resetting to 0 all slots

    } else {
        // INITIALIZE calendars, first assign'0' to ALL slots from beginning
        // of year up to and including the indicated days slot

        // <JMG> 2016-01-11, found+fix minor bug in calendar assignment 
        // where index 'curr_day' should have ben 'd' in loop below
        for(d=0; d < L4C_N_CALENDAR_SLOTS; d++)
        {
            ancImagePkgObj_p->SOC_CalendarCheck[d] =0 ; 
            ancImagePkgObj_p->FT_CalendarCheck[d] =0  ;
            ancImagePkgObj_p->SOC_CalendarCumulative[d] = 0 ; 
            ancImagePkgObj_p->FT_CalendarCumulative[d] = 0  ; 
        } // end:: d loop through dayOfYear indices
        
        // Next, ADJUST calendar, assigning '1' to ALL slots from beginning
        // of year up to and including the indicated days slot
        
        for(d=0; d <= curr_doy; d++)
        {
            ancImagePkgObj_p->SOC_CalendarCheck[curr_doy] =1 ; 
            ancImagePkgObj_p->FT_CalendarCheck[curr_doy ] =1  ;
            ancImagePkgObj_p->SOC_CalendarCumulative[curr_doy] =1 ;
            ancImagePkgObj_p->FT_CalendarCumulative[curr_doy ] =1 ; 
        } // end:: d loop
         
    }

    // WRITE RESULTs, adjusting either the ft_calendar or the 
    // soc_calendar as indicated
    if(strcmp(calendarLabel_p,"ft_calendar")==0)
    {
        // ADJUST F/T calendar "check-off" flag for this day of year
        if((attr_id = H5Aopen(soc_id, "ft_calendar", H5P_DEFAULT)) >= 0)
        {
            ret = H5Awrite(attr_id, H5T_NATIVE_UINT8, ancImagePkgObj_p->FT_CalendarCheck); 
        } else {
            xLogMsg( "@E In(%s) ERROR writing attrib ft_calendar back to SOC~\n",fxName);
        }
        
        // ADJUST F/T calendar_cumulative "check-off" flag for this day of year
        if((attr_id = H5Aopen(soc_id, "ft_calendar_cumulative", H5P_DEFAULT)) >= 0)
        {
            ret = H5Awrite(attr_id, H5T_NATIVE_INT16, ancImagePkgObj_p->FT_CalendarCumulative); 
        } else {
            xLogMsg( "@E In(%s) ERROR writing attrib ft_calendar_cumulative back to SOC~\n",fxName);
        }
        
        // end:: block for ft_calendar
    } else {

        if((attr_id = H5Aopen(soc_id, "soc_calendar", H5P_DEFAULT)) >= 0)
        {
            ret = H5Awrite(attr_id, H5T_NATIVE_UINT8, ancImagePkgObj_p->SOC_CalendarCheck); 
        } else {
            xLogMsg( "@E In(%s) ERROR writing attrib soc_calendar back to SOC~\n",fxName);      
        }
        
        if((attr_id = H5Aopen(soc_id, "soc_calendar_cumulative", H5P_DEFAULT)) >= 0)
        {
            ret = H5Awrite(attr_id, H5T_NATIVE_INT16, ancImagePkgObj_p->SOC_CalendarCumulative); 
        } else {
            xLogMsg( "@E In(%s) ERROR writing attrib soc_calendar_cumulative back to SOC~\n",fxName);
        }
        
        
    } // end:: block for soc_calendar
    
    // finally, close out the attributes handle
    H5Aclose(attr_id);      
    
    xLogMsg( "@A In(%s) updated L4_C calendar (%s) successfully written under policy opMode %d~ \n",
        fxName,  calendarLabel_p, opMode);
    
    l4c_EventCompletionMsg( fxName);
    
 return 0;  
} // end:: l4c_SOC_or_FT_CalendarUpdate()


// ----------------------------------------------------------------
// function     : l4c_HDF5_ReadStatic_SMRZ_MinMax(), Min/max for SMRZ
// purpose      : READ (2) SMRZ min,max daily-intermediate (updatable)
//                9km images from the SOC file SMAP_L4_C_SOC.h5.
// author       : Joe Glassy
// project      : SMAP L4C science development.
// dependencies : HDF5 lib, FRM API lib
// revision     : 2014-08-07T17:39, refactored to read the full 2D NatureRun
//                soil rootzone min,max images out of SOC instead of ANC file
//                since these are now updated daily
// revision     : 2014-02-21T14:36, refactored out of l4c_DAILY_Hdf5.c etc
// notes        :  in Anc.h5 --> "MERRA_climate_M09_gwetroot_max_flt32"
//                 in Anc.h5 --> "MERRA_climate_M09_gwetroot_min_flt32"
//  SMRZ rescaling min/max are ingested at pre-processor time and stored in:
//   metArrayObj_p->smrz_mn
//   metArrayObj_p->smrz_mx
//
int l4c_HDF5_ReadStatic_SMRZ_MinMax(MetArrayAdt *metArrayObj_p)
{
  SesCtl *sesCtl_p = l4c_GetSesCtl();
  NsbObj *nsbObj_p = sesCtl_p->sesNsbDb_p  ;  
  int status =0;
  
  hid_t soc_id = sesCtl_p->soc_file_id; // fetch prior opened soc file 
  char fxName[256] = {0};
  char nvpBuf[256] = {0} ;
  char szDatasetName[64] = {0} ;

    strcpy(fxName,(char *)__func__);

    xLogMsg("@A In(%s): reading smrz_mn, smrz_mx~\n",fxName);


    // 2) READ the (2) SMRZ vectors layers
    // first, MINIMUM
    strcpy(szDatasetName, nsbGetValueByKey(nsbObj_p,"L4C_INPUT_DATASET_SOILROOTZONE_M09_MIN", nvpBuf));
    
    status = frm_hdf5_ReadData(soc_id, H5T_NATIVE_FLOAT, szDatasetName,
                metArrayObj_p->smrz_mn,  fxName);

    // next, MAXIMUM
    strcpy(szDatasetName, nsbGetValueByKey(nsbObj_p,"L4C_INPUT_DATASET_SOILROOTZONE_M09_MAX", nvpBuf)); 
    status = frm_hdf5_ReadData(soc_id, H5T_NATIVE_FLOAT,szDatasetName,
                metArrayObj_p->smrz_mx,  fxName);

    // intentionally leave SMAP_L4_C_SOC open until session end ...
    l4c_EventCompletionMsg( fxName);

 return 0 ;
    
} // end:: l4c_HDF5_ReadStatic_SMRZ_MinMax()


// ----------------------------------------------------------------
// function     : l4c_SoilrootzoneMinMaxUpdate()
// purpose      : Numerically update SMAP soilrootzone min,max 9km 
//                daily intermediates stored in SMAP_L4_C_SOC.h5 
// inputs       : MetArrayAdt *metArrayObj_p -- ptr to MET array container object
//                size_t *nMinChanged,size_t *nMaxChanged)
// returns      : size_t, sum of either min or max pixels adjusted. If
//                none are adjusted, we can suppress the UPDATE-WRITE of
//                of the indicated min or max image
// author       : Joe Glassy
// project      : SMAP L4C science development.
// dependencies : HDF5 lib, FRM API lib
// revision     : 2014-08-07T17:33, added writes of (2) SMAP NatureRun 
//                soil rootzone min,max layers
//
// notes        : <JMG> 2015-12-09, Re-verify that firstAdjust logic is working correctly
//
size_t l4c_SoilrootzoneMinMaxUpdate(MetArrayAdt *metArrayObj_p,size_t *nMinChanged,size_t *nMaxChanged)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();
 char fxName[256] = {0} ;
 
 float32 smrz_pix = 0.0 ;
 size_t ithPix =0;
 size_t nMinUpdates = 0 ;
 size_t nMaxUpdates = 0 ;
 size_t nTestAdjusts = 0 ;
 int    firstAdjust = 0 ;

    strcpy(fxName,__func__);

    xLogMsg("\n@A In(%s) checking to adjust %zu smrz min and max 1D sparse vector pixels~ \n",
            fxName,sesCtl_p->nElem9kmPerTask);

    // iterate through 9KM (sparse) vector space
    for(ithPix = 0; ithPix < (size_t)sesCtl_p->nElem9kmPerTask ; ithPix++)
    {
        smrz_pix = metArrayObj_p->smrz[ithPix];
        
        // SKIP Exclude any fill value pixels, where FLOAT_FILL_VALUE 
        // MUST represent a value LOWER than all other valid-values.
        if( smrz_pix <= FLOAT_FILL_VALUE )
            continue;
            
        if( firstAdjust == 0)
        {
                metArrayObj_p->smrz_mx[ithPix] = smrz_pix;
                metArrayObj_p->smrz_mn[ithPix] = smrz_pix;
                firstAdjust = 1 ;
        } else {
            
            nTestAdjusts++ ; // count no. of potential adjustments
            
            if( smrz_pix < metArrayObj_p->smrz_mn[ithPix])
            {
                    metArrayObj_p->smrz_mn[ithPix] = smrz_pix;
                    nMinUpdates++; // count n. times minimum updated
            }
                    
            if( smrz_pix > metArrayObj_p->smrz_mx[ithPix])
            {
                    metArrayObj_p->smrz_mx[ithPix] = smrz_pix;
                    nMaxUpdates++; // counts n. times maximum updated
            }
        } // end:: else block to potentially adjust a min,max
        
    } // end:: ithPix loop through 9km space

    *nMinChanged = nMinUpdates ;
    *nMaxChanged = nMaxUpdates ;
    
    xLogMsg("@A In(%s) 1D-sparse (%zu pix) updated %zu smrz_mn and %zu smrz_mx of %zu potential changes~\n",
        fxName, ithPix, sesCtl_p->nElem9kmPerTask, nMinUpdates, nMaxUpdates, nTestAdjusts);
    
    // Returns total sum of updates made, whether min or max
    return (nMinUpdates + nMaxUpdates) ;
    
} // end:: l4c_SoilrootzoneMinMaxUpdate()


// ----------------------------------------------------------------
// function     : l4c_FT_RollingDriver()
// purpose      : Process the FT_ROLLING and FT_ROLLING_AGE 
//                daily intermediates stored in SMAP_L4_C_SOC.h5 
// inputs       : EcoStateArrayAdt *ecoStateArrayObj_p,
//                char *ftDatasetName_p,
//                uint8 *post2D_M01_p -- ptr to generic 2D (1KM or 3KM) buffer
               
// returns      : size_t, sum of either min or max pixels adjusted. If
//                none are adjusted, we can suppress the UPDATE-WRITE of
//                of the indicated min or max image
// author       : Joe Glassy
// project      : SMAP L4C science development.
// dependencies : HDF5 lib, FRM API lib
// revision     : 2014-08-07T17:33, added writes of (2) SMAP NatureRun 
//                soil rootzone min,max layers
// notes        : n/a
int l4c_FT_RollingDriver(EcoStateArrayAdt *ecoStateArrayObj_p,char *ftDatasetName_p,uint8 *imageBuffer2D_M03_p)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();
 NsbObj *nsbObj_p = l4c_GetNsbOverseer();
 
 char fxName[256] = {0};
 char thisFtDataset[128] = {0};
 char nvpBuf[256] ;
 hid_t dset_id = -1;
 hid_t soc_id = sesCtl_p->soc_file_id;
 herr_t ret = -1;
 uint8 *ftData_p = NULL; // ptr set to whichever FT variable is processed
 
 // set n. of 3KM pixels
 size_t nPix3km =  EGv2_GLOBAL_3KM_NS_PIXELS * EGv2_GLOBAL_3KM_EW_PIXELS;
 

    strcpy(fxName,__func__);

    // RESET buffer prior to re-using it, this time as a 3KM global buffer
    memset(imageBuffer2D_M03_p, '\0', (nPix3km * sizeof(uint8)) );

    l4c_inflate_init_3km(imageBuffer2D_M03_p , DFNT_UINT8);
        
    // BRANCH on which FT variable to process
    if(strcmp(ftDatasetName_p,"FT_ROLLING")==0)
    {
        strcpy(thisFtDataset, nsbGetValueByKey(nsbObj_p,"L4C_INPUT_DATASET_FT_ROLLING",nvpBuf));
        ftData_p = ecoStateArrayObj_p->ftRolling_3km ;
    } else if(strcmp(ftDatasetName_p,"FT_ROLLING_AGE")==0){
        strcpy(thisFtDataset, nsbGetValueByKey(nsbObj_p,"L4C_INPUT_DATASET_FT_ROLLING_AGE",nvpBuf));
        ftData_p = ecoStateArrayObj_p->ftRollingAge_3km ;       
    }

    xLogMsg( "@A In(%s) Inflating 2D 3KM FT component (%s) ~\n",fxName,thisFtDataset);

    l4c_inflate_3km(ftData_p, imageBuffer2D_M03_p,DFNT_UINT8);
    
    // RETRIEVE prior HDF5 dataset ID for this FT variable processed
    dset_id = frm_hdf5_GetDatasetId(soc_id,thisFtDataset);
    
    // WRITE out the updated FT result
    if((ret = H5Dwrite(dset_id,  H5T_NATIVE_UINT8, H5S_ALL, H5S_ALL, H5P_DEFAULT, imageBuffer2D_M03_p))<0)
    {
        xLogMsg( "@E In(%s) ERROR writing 2D (%s) ~\n",fxName,thisFtDataset);
    } else {
        xLogMsg( "@A In(%s) Wrote 2D 3KM FT component (%s) ~\n",fxName,thisFtDataset);
    }
        // <JMG> 2014-11-03, Add explicit FT basis SOC file-level I/O flush to be conservative
    ret = H5Fflush(soc_id, H5F_SCOPE_LOCAL); 

 // return status of write ops
 return (int)ret ;
    
} // end:: l4c_FT_RollingDriver()


// ----------------------------------------------------------------
// function     : l4c_SoilrootzoneMinMaxDriver()
// purpose      : Driver for writing the updated soil rootzone min,max 
//                daily-intermediate terms back to the SMAP_L4_C_SOC.h5
// inputs       : MetArrayAdt *metArrayObj_p
// returns      : status 
// 
int l4c_SoilrootzoneMinMaxDriver(MetArrayAdt *metArrayObj_p)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();
 NsbObj *nsbObj_p = l4c_GetNsbOverseer();
 
 char   fxName[256]= {0};   
 // build-in a quick string table for storing the (2) smrz_m{K} names
 char   szDatasetNamesTable[2][128]= {{0}} ;
 
 char   szDatasetName[128] ={0} ;
 char   nvpBuf[256] = {0} ;
 size_t nMinChanged = 0 ;
 size_t nMaxChanged = 0 ;
 size_t nChanges = 0 ; // total of nMin , nMax changes to smzr_mn,mx made
 size_t nBytes = 0 ;
 
 herr_t ret = -1;
 hid_t  dset_id = -1;
 hid_t  soc_id = -1;
 int s = 0 ;
 
 float32 *imgGenericBuffer2D_M09_p = NULL; // reusable buffer for 9KM transforms
 
 
    strcpy(fxName,__func__);

    // fetch prior opened SMAP_L4_C_SOC.h5 file id...
    soc_id = sesCtl_p->soc_file_id ;
        
    // Retrieve, then store the (2) dataset names from the NSB database, 
    // to be used below in for loop operations
    strcpy(szDatasetNamesTable[0],nsbGetValueByKey(nsbObj_p,"L4C_INPUT_DATASET_SOILROOTZONE_M09_MIN", nvpBuf));
    strcpy(szDatasetNamesTable[1],nsbGetValueByKey(nsbObj_p,"L4C_INPUT_DATASET_SOILROOTZONE_M09_MAX", nvpBuf));
    
    // <JMG> 2014-09-27, added, to clarify smrz processing
    imgGenericBuffer2D_M09_p =(float32 *)memCalloc(EGv2_GLOBAL_9KM_PIXELS,sizeof(float32));

    // First UPDATE the (2) soil rootzone min,max daily intermediate
    // terms, and then inflate them both, then write back to SOC HDF5 file
    nChanges = l4c_SoilrootzoneMinMaxUpdate(metArrayObj_p, &nMinChanged,&nMaxChanged);
    
    // FUTURE OPTIMIZATION: when nChanges is 0, we (could) omit
    //    the WRITE-UPDATE of either the min and max layers, as per rules:
    //       if nMinChanged > 0, we UPDATE just the (Minimum) smrz_mn term
    //       if nMaxChanged >0, we UPDATE just the (Maximum) smrz_mx term
    nBytes = EGv2_GLOBAL_9KM_PIXELS * sizeof(float32);
        
    // LOOP through (2) soil moisture root zone min,max layers
    for(s=0; s < 2 ; s++)
    {
        strcpy(szDatasetName, szDatasetNamesTable[s]);
        // Retrieve dset_id of smrz_mn as stored earlier!
        dset_id = frm_hdf5_GetDatasetId(soc_id,szDatasetName);
        xLogMsg( "@A In(%s) processing (%s) dataset ID %d ~\n",fxName,szDatasetName,dset_id);
        
        // DIAG ONLY: generate and view a diagnostic 9KM image of smrz_mn 
        // l4c_DiagImage("smrz_mn_postadjust", "smrz_mn.flt32",DFNT_FLOAT32, 
        //      EGv2_GLOBAL_9KM_PIXELS,localDims, imgGenericBuffer2D_M09_p);
        
        l4c_inflate_init_9km(imgGenericBuffer2D_M09_p , DFNT_FLOAT32);
        
        if( s == 0 )
        {   // process MINIMUM
            l4c_inflate_9km(metArrayObj_p->smrz_mn, imgGenericBuffer2D_M09_p, DFNT_FLOAT32);
        } else {
            // process MAXIMUM
            l4c_inflate_9km(metArrayObj_p->smrz_mx, imgGenericBuffer2D_M09_p, DFNT_FLOAT32);
        }
        // WRITE RESULT. Note that dset_id must have been opened in append or r+w mode of course.
        if((ret = H5Dwrite(dset_id,  H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, imgGenericBuffer2D_M09_p))<0)
        {
            xLogMsg( "@E In(%s) ERROR writing 2D (%s) back to SOC~\n",fxName,szDatasetName);
        } else {
            xLogMsg( "@A In(%s) wrote inflated 2D (%s) back to SOC~\n",fxName,szDatasetName);
        }
        // RESET -- perform a hard reset to flush, prior to diagnosing smrz_mn,mx issues
        memset(imgGenericBuffer2D_M09_p,'\0',nBytes);
        // CLOSE dataset
        frm_hdf5_CloseDataset(dset_id);
        
        // Generate and view a diagnostic 9KM image of smrz_mn 
        // l4c_DiagImage("smrz_mn_postadjust", "smrz_mn.flt32",DFNT_FLOAT32, 
        //      EGv2_GLOBAL_9KM_PIXELS,localDims, imgGenericBuffer2D_M09_p);

    } // end:: s loop through (2) soil moisture rootzone min,max images
    
    memFree(imgGenericBuffer2D_M09_p);
    
    l4c_EventCompletionMsg(fxName);
    
    return 0;
    
}  // end:: l4c_SoilrootzoneMinMaxDriver()


// ----------------------------------------------------------------
// function     : l4c_SOC_UpdatePoolDriver()
// purpose      : Update a single given SOC pool (C0,C1, or C2) 
// author       : Joe Glassy
// project      : SMAP L4C science development.
// dependencies : HDF5 lib, FRM API lib
// revision     : 2015-06-24T13:41, <JMG> added dset_id to error message for more context
// revision     : 2014-09-30T17:26, split out from monolithic code
int l4c_SOC_UpdatePoolDriver(char *socPoolLabel_p,EcoStateArrayAdt *ecoStateArrayObj_p,float32 *post2D_M01_p)
{
  SesCtl *sesCtl_p = l4c_GetSesCtl();   
  NsbObj *nsbObj_p = l4c_GetNsbOverseer();
  
  char  fxName[256] = {0} ;
  char  nvpBuf[256] = {0} ;
  char  timeStr[64] = {0} ;
  char  targetDatasetName[128] = {0};
  char  socPoolValueBuf[8] = {0};
  char  socDatasetName[64] = {0};
  // for diag
  char osCmd[FILENAME_MAX]= {0} ;

  herr_t ret = -1;
  hid_t soc_id = sesCtl_p->soc_file_id  ; // set file_id to OPEN SOC...
  hid_t dset_id = -1;
  hid_t hDtype = H5T_NATIVE_FLOAT; // fix for all SOC pool arrays
  
  int rank = 2;  // for 2D
  int socPoolIndex = -1 ;

  float32 *socData_p = NULL ; // ptr to either C0,C1, or C2 data

    strcpy(fxName,(char *)__func__);
    
    rank = 2;  // for M01 full 2D instance of the SOC C[0],C[1],C[2]...
    
    
    // Retrieve fully qualified dataset names from the RIP file's NSB 
    strcpy(targetDatasetName,"L4C_INPUT_DATASET_SOC_");
    // construct fully qualified TARGET name as C0,C1,C2 etc
    strcat(targetDatasetName, socPoolLabel_p ) ; // add the "C0", or "C1" or "C2" to form full name

    // full target-name complete, so look it up in NSB database and store it
    strcpy(socDatasetName,nsbGetValueByKey(nsbObj_p,targetDatasetName, nvpBuf));
    
    // next, fetch its HDF5 dataset id
    dset_id = frm_hdf5_GetDatasetId(soc_id,socDatasetName);
    // TBDL: test to verify dset_id is actually valid here...
    
    // Derive index {0,1,2} from SOC pool label "C0","C1","C2", using offset of 1
    socPoolIndex = atoi(socPoolLabel_p+1);
    // ENFORCE with ASSERT
    assert(socPoolIndex >= 0 && socPoolIndex <= 2);
    
    // Based on index {0,1,2} set data pointer to this SOC pool's 1KM data
    socData_p    = ecoStateArrayObj_p->C[socPoolIndex].dom_1km ;
    // ASSERT: we're pointing to good data
    assert(socData_p != NULL);
    
    xLogMsg( "@A In(%s) SOC POOL %d Update (%s) started at (%s)~\n",
       fxName, socPoolIndex, socPoolLabel_p, frm_utilGetCtime(timeStr,0));

    // Remember: we must inflate (C0,C1,C2).dom_1km from 1D sparse form
    // to full 2D form prior to writing back out to the SOC file
    l4c_inflate_init_1km(post2D_M01_p , DFNT_FLOAT32);
    
    // transfer 1D-sparse data to full 2D EASEGRID 1km grid prior to writing out
    l4c_inflate_1km(socData_p, post2D_M01_p,DFNT_FLOAT32);

    // WRITE OUT resulting updated pool C0, or C1, or C2
    if((ret = H5Dwrite(dset_id, hDtype, H5S_ALL, H5S_ALL, H5P_DEFAULT, post2D_M01_p))<0)
    {
        xLogMsg( "@E In(%s) ERROR writing 2D datasetID %d (%s) ~\n",
                      fxName, dset_id, socDatasetName);
    } 
    
    // <JMG> 2014-11-03, Add explicit SOC file-level I/O flush to be conservative
    ret = H5Fflush(soc_id, H5F_SCOPE_LOCAL); 

    // RESET re-usable full 2D global EASEGRIDv2 buffer for next use...
    memset(post2D_M01_p, '\0', (sizeof(float32)* EGv2_GLOBAL_1KM_PIXELS));

 // return valid SOC pool index {0,1,2} or -1 if failed.
 return socPoolIndex ;
 
} // end:: l4c_SOC_UpdatePoolDriver()


// ----------------------------------------------------------------
// function     : l4c_SOC_Update_Intermediates()
// purpose      : Implement (container for) all HDF5 UPDATEs to intermediates
//                Included here thus far are: FT_ROLLING_M03 and SOC, etc.
// author       : Joe Glassy
// project      : SMAP L4C science development.
// dependencies : HDF5 lib, FRM API lib
// revision     : 2016-01-11, added note to review annual end-of-year turnover
//                production rule.
// history      : 2014-08-07T17:33, added writes of (2) SMAP NatureRun 
//                soil rootzone min,max layers
// history      : 2014-03-11T13:20, leave SOC file open until end of session
// history      : 2014-02-21T14:36, refactored out of l4c_DAILY_Hdf5.c etc
// notes        : Remember to run (3) SOC,npp_sum, "M01land" layers BACK through 
//                 the INFLATE process PRIOR to writing these out!! as
//                 2D "M01" full 2D grid layers.
//                For 3km FT_ROLLING_M03, do the same, go thru INFLATE 
//                 first, then write out as full 2D 3km layers..
// notes        : PRIOR: tcfMem_saveCstate(),tcfMem_saveRollingCstate()
// notes        : 2016-01-11, TBD best to re-confirm if (or how) the SOC
//                pool (3) state variables may need adjustment when an annual
//                year boundary (e.g. hit January 1) is encountered. For now,
//                logic assumes Jan 1 is just another day in a multi-year sequence
//                where no special handling is performed on the (3) pool variables
//                upon hitting a January 1 day-of-year.
int l4c_SOC_Update_Intermediates(EcoStateArrayAdt *ecoStateArrayObj_p)
{
  SesCtl *sesCtl_p = l4c_GetSesCtl();   
  
  // SOC and FT calendars are store
  AncImagePkgAdt  *ancImagePkgObj_p = l4c_GetAncImagePkgAdt();
  MetArrayAdt      *metArrayObj_p   = l4c_GetMetArrayAdt();
  
  time_t socUpdate_strtSec = time(0);
  time_t socUpdate_endiSec = time(0);
  double socUpdate_elapSec = 0.0;
  double cumElapSec = 0.0 ;

  char  fxName[256] = {0};
  char  timeStr[32] = {0};
  
  // for diag
  char osCmd[FILENAME_MAX]= {0} ;
  
  hid_t soc_id = sesCtl_p->soc_file_id  ; // set file_id to OPEN SOC...
  hsize_t dims1km[2] = {0} ;
  hsize_t dims3km[2] = {0} ;
  hsize_t dims9km[2] = {0} ;
  
  size_t  localDims[2] = {0}; // for l4c_Diag ...
  
  int rank = 2;  // for 2D
  // boolean, if True=1,offers chance to AVOID a re-run on this day of year!
  int avoidRerun =0 ; 
  int forceStrict_SOC_PoolUpdatePolicy = 0; // set to permissive=0 until tested
  
  float32 *post2D_M01_p = NULL; // reusable buffer for 1KM transforms

    strcpy(fxName,(char *)__func__);
    
    xLogMsg( "@A In(%s) SOC-POOL-UPDATE started at (%s) ~\n",
            fxName,frm_utilGetCtime(timeStr,0));
    
    rank = 2;  // for M01 full 2D instance of the SOC C[0],C[1],C[2]...
    
    dims1km[0] = EGv2_GLOBAL_1KM_NS_PIXELS ;
    dims1km[1] = EGv2_GLOBAL_1KM_EW_PIXELS ;
    // set up for 3KM Freeze Thaw as well
    dims3km[0] = EGv2_GLOBAL_3KM_NS_PIXELS ;
    dims3km[1] = EGv2_GLOBAL_3KM_EW_PIXELS ;
    // set up for the (2) 9KM SMAP_NatureRun layers
    dims9km[0] = EGv2_GLOBAL_9KM_NS_PIXELS ;
    dims9km[1] = EGv2_GLOBAL_9KM_EW_PIXELS ;
    // set 9km dims for local diag use as size_t's...
    localDims[0] = dims9km[0];
    localDims[1] = dims9km[1];

    // <JMG> 2016-01-11, these warnings will trigger until the new code that
    // auto-resets all calendar slots to 0 when a January 1 is hit is deployed.
    
    // <JMG> 2014-11-11 added live CHECK on calendar status (finally, was missing before)
    // We intentionally perform this check BEFORE the big SOC and FT rolling updates
    // to give correction logic a chance to avoid this!
    // This warns the operator that for this SMAP_L4_C_SOC.h5 on this dayofyear,
    //  the same dayOfYear has been run multiple times, where it should ONLY be
    //  run once in normal usage.
    if(ancImagePkgObj_p->FT_CalendarCumulative[sesCtl_p->runDayOfYear] > 1)
    {
        xLogMsg( "@W In(%s) FT calendar slot value (%d)for dayofyear (%d) exceeds expected count of 1 ~ \n",
            fxName, ancImagePkgObj_p->FT_CalendarCumulative[sesCtl_p->runDayOfYear],
             sesCtl_p->runDayOfYear);
        // setting this to 1 indicates we know that an update of SOC following below
        // will essentially corrupt or contaminate an ideal SOC pool update production rule
        // where, as intended, each SOC pool update should occur ONCE and ONCE ONLY
        // for a given day-of-year
        avoidRerun = 1;
    }

    if(ancImagePkgObj_p->SOC_CalendarCumulative[sesCtl_p->runDayOfYear] > 1)
    {
        xLogMsg( "@W In(%s) SOC calendar slot value (%d)for dayofyear (%d) exceeds expected count of 1 ~ \n",
            fxName, ancImagePkgObj_p->SOC_CalendarCumulative[sesCtl_p->runDayOfYear],
            sesCtl_p->runDayOfYear);
        avoidRerun = 1;
    }

    // Next, create (or inherit) the SOC and FT_ROLLING_M03 dataset objects
    // Note: post2D_M01_p works here as a single, reusable 1KM full grid array for all SOC
    //       1km full grid rasters processed here, regardless of datatype
    post2D_M01_p =(float32 *)memCalloc(EGv2_GLOBAL_1KM_PIXELS,sizeof(float32)); 

    // <JMG> SOC file opened ONCE now, in append mode, in l4c_SesInit()
    // and its unique file_id was stored in soc_file_id
    soc_id = sesCtl_p->soc_file_id ;

    xLogMsg( "@W In(%s) SOC pool update policy now set at %d (0=permissive,1=strict)~ \n",
            fxName, forceStrict_SOC_PoolUpdatePolicy);

    // Hard wired temporarily as permissive; once tested we can make this more strict
    if( avoidRerun >= 0 || forceStrict_SOC_PoolUpdatePolicy >= 0)
    {
        // SOC udpate each of (3) SOC pools, writing back to SMAP_L4_C_SOC.h5
        l4c_SOC_UpdatePoolDriver("C0", ecoStateArrayObj_p,post2D_M01_p) ;
        l4c_SOC_UpdatePoolDriver("C1", ecoStateArrayObj_p,post2D_M01_p) ;
        l4c_SOC_UpdatePoolDriver("C2", ecoStateArrayObj_p,post2D_M01_p) ;

        // UPDATE the FT_ROLLING_M03 and FT_ROLLING_AGE_M03 next...
        l4c_FT_RollingDriver(ecoStateArrayObj_p,"FT_ROLLING"    ,(uint8 *)post2D_M01_p);
        l4c_FT_RollingDriver(ecoStateArrayObj_p,"FT_ROLLING_AGE",(uint8 *)post2D_M01_p);

        // SOIL MOISTURE ROOT ZONE MIN,MAX 9KM adjustments performed here
        l4c_SoilrootzoneMinMaxDriver(metArrayObj_p) ;
    
        // RIP file NVP parameter "L4C_CALENDAR_MODE" defines both 
        //   these FT,SOC calendar state arrays
        //   uint8 ancImagePkgObj_p->SOC_CalendarCheck[L4C_N_CALENDAR_SLOTS]
        //   uint8 ancImagePkgObj_p->FT_CalendarCheck[L4C_N_CALENDAR_SLOTS] 
        //   UPDATE the int calendar flag array for DOY runDayOfYear
        
        // <JMG> 2014-11-11 changed interface tpo l4c_SOC_or_FT_CalendarUpdate()
        // in order to pass in just the label "ft_calendar"|"soc_calendar" and
        // perform the rest of the calendar array retrieval (and WRITE) logic inside (here)
        l4c_SOC_or_FT_CalendarUpdate(sesCtl_p->ftOpMode ,"ft_calendar" ,sesCtl_p->runDayOfYear); 
        l4c_SOC_or_FT_CalendarUpdate(sesCtl_p->socOpMode,"soc_calendar",sesCtl_p->runDayOfYear); 

    } // end:: block if a strict SOC pool update policy is in force
    
    // release memory for "re-usable" full 2D 1km  global buffer 
    memFree(post2D_M01_p);
     
    socUpdate_endiSec = time(0);
    socUpdate_elapSec = difftime(socUpdate_endiSec,socUpdate_strtSec);
        
    sesCtl_p->sesEndiSec = time(0); // snapshot cumulative time thus far
    cumElapSec = sesCtl_p->sesEndiSec - sesCtl_p->sesStrtSec;
    
    xLogMsg("@A In(%s) session cumulative time required %.2lf elapsed sec (global)~ \n",
        fxName,  cumElapSec);

    xLogMsg("@A In(%s) SOC-POOL-UPDATE FINISHED (%s) requiring %.2lf sec (local) ~\n\n",
        fxName, frm_utilGetCtime(timeStr,0),socUpdate_elapSec);
        
        
    l4c_EventCompletionMsg( fxName);

 return 0 ;
    
} // end:: l4c_SOC_Update_Intermediates()
