// ---------------------------------------------------------------
// function     : l4c_DAILY_ModFunc.c
// purpose      : Encapsulates tcfModFunc.c
// author       : Lucas A. Jones; lucas@ntsg.umt.edu
// author       : joe glassy, Lupine Logic Inc.
// version      : see L4C_VERS_TAG , as 4.1.5 or later
// revised      : 2015-09-07T21:31, reclarified equivalance of DoUseFT_GEOS5 and
//                DoUseFT_ActiveTsurf since both use bit0 for F/T state
// revised      : 2014-04-16, Lucas Jones
//                   tcfModFunc_isInCell() updated
// revised      : v4.1.6-2014-02-05T11:242:00-07:00,wheat,jmg"
// dependencies : HDF4, HDF5 libraries and L4C FRAMEWORK API.
// history      : n/a
// notes        : Based on and adapted from tcf_spin v3.1/3.2
// ---------------------------------------------------------------

//  "tcf_inc.h" is now folded into l4c_DAILY.h
#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"


// --------------------------------------------------------------
/*
 module       : tcfModFunc.c
 purpose      : Core model process logic.  These functions implement model dynamics for the basic 
                computational unit, which is the daily timestep for one 1km grid cell.  
 project      : SMAP L4C science development.
 dependencies : tcfModMult.c
 revision     : 2014-02-15T12:54, for LEAD FORK with new ADT namenames
 revision     : Initial reformating -- laj 09/10/13
 author       : Lucas A. Jones; lucas@ntsg.umt.edu
 notes        :
*/

// --------------------------------------------------------------
// function: tcfModFunc_isInCell()
// Purpose : Checks for whether or not to process a 1km cell based 
//           on pft and availability of input data. 
// Inputs  : Current 1-km pft and input meteorology.
// Outputs : 1=process cell; 0=do not process cell.
// revised : 2034
// revised ; 2014-08-07, re-verified encoding of swath bit as below...
// Notes   : Changes in 1-km domain criteria not handled in land-domain 
//           vectorization should be implemented here,e.g. any pixel
//           level filter such as imposed by "L4C_M01land_PixFilter.uint8"
//           should be implemented HERE.
int tcfModFunc_isInCell(uint8 pft_index, MetAdt metObj, EcoStateAdt ecoStateObj)
{

  int   isIn = 0;
  
  // <LAJ> date 04/16/14 changed litterfall switch from >= 0 to > FLT_TOL
  if ((pft_index > 0) & (pft_index <= 8) & (metObj.tsoil > FLT_TOL) & (ecoStateObj.lit > FLT_TOL))
    isIn = 1;


  return(isIn);

} // end :: tcfModFunc_isInCell()


// ---------------------------------------------------------------
// function: tcfModFunc_switchFTmethod()
// purpose : Transfers L3_SM_A FT from global to local array, and provides switching mechanism to choose between Radar-derive FT and 
//          Tsurf-derived FT.
// inputs  : Serial index for global 1km grid (s1km), pointer to local MET, and global grid MetArrayObj structures.
// outputs : Sets FT Rolling and FT Age fields in the local MET structure, returns indicator for which FTmethod was used (Radar or Tsurf).
// returns : int ft_method, defined as: -1=used FT rolling 0=uses FT RADAR, 1=uses FT Tsurf
// revised : 2014-10-21T15:35, enabled RIP supplied ftState instead of hardwired macro prior
// history : 2014-07-25, <LAJ> 7/25/14, added this function  
// notes   : Extract all (4) available bits at positions 0,1,2,3 respectively
//           bits(0,1,2,3) --> ft_tsurf, ft_radar, ft_radar_qc, ft_in_swath
//           REMINDER for swath bit: 0 means IS-IN-SWATH, and 1=NOT IN SWATH as per enum: {IsInSwath, IsNotInSwath}
//           noting this exactly follows Barry Weiss' convention for now.
// 
int tcfModFunc_switchFTmethod(uint32 s3km, MetAdt *metObj_p, MetArrayAdt *metArrayObj_p, EcoStateArrayAdt *ecoStateArrayObj_p)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();
 
  int  ftMethod = -1; // indicator for switch status (-1 = default(FT ROLLING);  0=used Radar FT; 1=used Tsurf FT)

    // Unpack L3_SM_A surflag from (new) uint8 ft_state_um_3km array 
    metObj_p->ft_state_um_3km =  metArrayObj_p->ft_state_um_3km[s3km];
    
    // Tsurf FT
    metObj_p->ft_Tsurf_3km = (uint8)getBitsFromUint32(0,0, metObj_p->ft_state_um_3km);  
    // Radar FT 
    metObj_p->ft_Radar_3km = (uint8)getBitsFromUint32(1,1, metObj_p->ft_state_um_3km);
    // Radar FT QC, ft_Radar_Qc_3km, extracted but is NOT yet used
    metObj_p->ft_Radar_Qc_3km = (uint8)getBitsFromUint32(2,2, metObj_p->ft_state_um_3km);
    
    // Last, Determine if data within swath 
    // <LAJ> 7/25/14, Bit 15 works for now, but is not officially documented, 
    // so its use here this might change)
    // REMINDER: our L4_C team defines 0 as IS-IN-SWATH, and 1=NOT IN SWATH as per enum: {IsInSwath, IsNotInSwath}
    metObj_p->ft_in_swath_3km = (uint8)getBitsFromUint32(3,3, metObj_p->ft_state_um_3km);

    // <LAJ> 7/25/14, currently no fill values are assigned to L3_SM_A surflag
  
   // <JMG> 2014-08-07, simplified since surflag_3km no longer is defined
   // if (metObj_p->ft_in_swath_3km != 1 && metObj_p->surflag_3km != FILVAL_UINT16)
  
  // <JMG> compound test judged as not needed...
  // if (metObj_p->ft_in_swath_3km != 1 && metObj_p->ft_state_um_3km != FILVAL_UINT8)
  
    // first, pre-qualify as IN swath or not, matching on enum: {IsInSwath,IsNotInSwath}
    if (metObj_p->ft_in_swath_3km == IsInSwath )
    {
      
        // Radar FT available, use it
        // <LAJ> 7/25/25, USE_RADAR_FT is a macro defined in the l4c_DAILY.h, 
        // but alternatively could be read in from RIP or L3_SM_A QA fields
        
        // <JMG> 2014-10-21T15:33, replaced use of hard wired macro to use
        // RIP supplied switch sesCtl_p->ftState 
        // enum: {DoUseFT_ActiveRadar, DoUseFT_ActiveTsurf, DoUse_NO_FT}
        // if( sesCtl_p->ftState ==DoUseFT_ActiveTsurf )  ..or.. DoUseFT_ActiveRadar 
        
        if (sesCtl_p->ftState == DoUseFT_ActiveRadar)
        {
            // Assign current Radar freeze-thaw status to rolling FT status grid-cell
            ecoStateArrayObj_p->ftRolling_3km[s3km] = metObj_p->ft_Radar_3km;
            ftMethod = DoUseFT_ActiveRadar;  // assign FT method
            // Otherwise use Tsurf FT
        } else if(sesCtl_p->ftState == DoUseFT_ActiveTsurf || sesCtl_p->ftState ==DoUseFT_GEOS5) {

            // Assign current Lst freeze-thaw status to rolling grid-cell
            ecoStateArrayObj_p->ftRolling_3km[s3km] = metObj_p->ft_Tsurf_3km;
            // <JMG> 2015-09-07, ftMethod of DoUseFT_ActiveTsurf is also assigned
            // here when sesCtl_p->ftState ==DoUseFT_GEOS5, since the GEOS5 method
            // writes the freeze thaw state into the same (bit0) of FT_STATE_UM_M03
            // that DoUseFt_ActiveTsurf uses (so they're the functionally identical)
            ftMethod = DoUseFT_ActiveTsurf;  // assign FT method
            
            // NOTE: no provision exists yet for DoUse_NO_FT at this time

        } // end :: use Radar FT vs. Tsurf FT conditional

        // Reset age information
        ecoStateArrayObj_p->ftRollingAge_3km[s3km] = 0;

     // Data out of swath
    } else {

        // Increment age information    
        ecoStateArrayObj_p->ftRollingAge_3km[s3km]++;
        ftMethod = -1; // assign FT method as using FT rolling as source of FT

  } // end :: in-swath vs. out-of-swath conditional

  return(ftMethod);

} // end :: tcfModFunc_switchFTmethod()


// ---------------------------------------------------------------
// function : tcfModFunc_switchFparVi()
// Purpose  : Transfers Fpar from global to local array, casts as float, 
//            and provides switching mechanism to estimate Fpar using 
//            linear regression adjustment with VIIRS VI. 
// inputs   : BPLUT *bp_p, 
//          : uint32 s1km,
//          : MetAdt *metObj_p,
//          : MetArrayAdt *metArrayObj_p
// notes    : On inputs, BPLUT pointer, serial index for global 1km
//            grid (s1km), pointer to local MET, and global grid
//            MetArrayObj structures.
// output   : Assigns Fpar and Fpar QA fields in the local MET structure.
// revised  : 2014-11-10T11:27, <JMG> Renamed 4-state'gppMethod' (here and
//            in tcf_struct ADT metObj also) to gppInputUsed to facilitate distinguishing
//            this from RIP file's 3-state gppMethodStr {FPAR,NDVI,CLIM}
// history  : 2014-10-31T13:40, added explicit assignment of fpar_qa in
//            branch cases prior where only implicit logic was used.
// notes    : 3/5/2014, control logic here has been converted to rely on
//            sesCtl_p->fparState boolean switch, which is assigned solely
//            from the l4c_DAILY.rip name-value-pair token "L4C_GPP_SOURCE". 
// notes    : <LAJ> 6/4/2014, Modified version to offer more options 
//            for Fpar subsitution.  
// notes    : <JMG> 2014-10-24, now also provides a "forced FparClimatology" 
//            option via use of RIP keyword option L4C_GPP_SOURCE = "CLIM"|"FPARCLIM"
//          
int tcfModFunc_switchFparVi(BPLUT *bp_p, uint32 s1km, MetAdt *metObj_p, MetArrayAdt *metArrayObj_p)
{
  SesCtl      *sesCtl_p = l4c_GetSesCtl();
  RunParamAdt *runtimeParamObj_p = &sesCtl_p->runParamObj;
  
  // NDVI's analytical representation as a float is {-1.0 <= ndvi<= 1.0}
  float64 ndviSci =0.0 ;  
  uint8 tmp;
  
  // WARNING: gppInputUsed coding here is different from that used 
  // elsewhere in l4c_MODIS where e.g. gppInputUsedStr="NDVI" --> 0 , and "FPAR" --> 1
  
  // gppInputUsed indicator: // (-1 = default; 0 = used daily FPAR; 1 = used FPAR climatology; 2 = used NDVI)
  int  gppInputUsed    = -1  ;
    
  // Use VIIRS VI (NDVI or EVI) if specified in global parameter file.
  // Linear regression used to translate VI to Fpar.
  
  // <JMG> 2014-07-25T16:49, corrected prior hard-wired assignment
  // to now use dynamic setting from RIP nvp "L4C_GPP_SOURCE" w/
  // choices: {"FPAR"|"NDVI"}.  This switch is correctly assigned now
  // in l4c_DAILY_rip.c:l4c_SesCollectNVPs() function

    // Use original VIIRS NVDI with gain and offset, as below...  
    if( sesCtl_p->fparState == DoUseNDVI )
     {
        // viirsNdvi as a uint8 originates as {0..255}, actually as
        // {0..200}, so to unpack into an analytical science value ranging
        // from {-1.0 <= ndvi <= +1.0}, apply the gain and offset as
        // follows: {0..200} --> convert to float, subtract 100.0 for
        // --> {-100...+100} --> then divide by 100.0 --> result {-1.0...+1}
        ndviSci = metArrayObj_p->viirsNdvi[s1km] ;
        // convert from {0..200} to {-100...+100}, then to {-1.0 < ndvi <1.0}
        ndviSci = (ndviSci - 100.0) / 100.0 ; 
        
        // Adjust VIIRS VI to Fpar using linear regression coefficients
        metObj_p->fpar = (float64)ndviSci * bp_p->NDVItoFPAR_scale + bp_p->NDVItoFPAR_offset;
        
        // <JMG> 2014-10-31T13:35, to be explicit, assigned lowest quality Fpar flag (4)
        // NOTE: can change this later when regression quality better known)
        metObj_p->fpar_qa = bp_p->Fpar_QA_Rank_max;
    
        gppInputUsed = 2; // indicate NDVI usage

    } else if(sesCtl_p->fparState  == DoUseFparClim ) {
        // <JMG> 2014-10-21, enabled optional FORCED USE of fpar_clim[]
        // of FparClimatology
        
        // Forced, so  fill with climatology (unpack climatology, apply gain and offset)
        // Unpack 1-km fpar climatology array for model step call interface
        metObj_p->fpar = (float64) metArrayObj_p->fpar_clim[s1km] * runtimeParamObj_p->FPAR_scale + runtimeParamObj_p->FPAR_offset;
        // <JMG> 2014-10-31T13:35, fix this at 4 for forced FparClimatology
        metObj_p->fpar_qa = bp_p->Fpar_QA_Rank_max; // usually 4 here
        gppInputUsed = 1; // indicate that FparClimatology was used (fpar climatology usage)
        
    }  else   {
        // Otherwise use either MODIS FPAR or FPAR climatology
        
        // If MODIS FPAR file not available, assign QA class to use climatology
        // <LAJ> 6/5/2014 This state currently not set during file reading, but
        //                it must be set for this feature to fully function.
        
        if ( sesCtl_p->fparState == msngFparFile )
        {

            metObj_p->fpar_qa = (uint8)FPAR_QA_USE_CLIM_THRES;
            gppInputUsed = -1;  // indicate both fpar and ndvi missing
            // Otherwise if MODIS FPAR file available, read QA and THEN decide whether to use climatology
            
        } else {

            // Unpack and reclassify fpar qa/qc data. All MODIS and related algorithms use
            // a base0 encoding for qa, ranging from {0,1,2,3} so note the departure 
            // in this convention here, where fpar_rank is encoded {0,1,2,3,4} and where
            // {0|1} are treated equivalently as 'best'.
            
            // Assert: fpar_qa returned from fpar_qa_bin_reclas() has 
            // integer values 1 to 4, with 1 = best quality and 4 = use climatology
          
          metObj_p->fpar_qa = fpar_qa_bin_reclas(metArrayObj_p->fpar_qa[s1km]);

        } // end :: Fpar file available vs. not available decision 

        // Check if less then (4) e.g. (FPAR_QA_USE_CLIM_THRES=4)
        if ( metObj_p->fpar_qa < FPAR_QA_USE_CLIM_THRES)
        {
            // Good quality FPAR e.g. {0,1,2,3} was encountered, so apply gain,offset for use
            
            // Unpack 1-km fpar array for model step call interface
            // FPAR gain now in sesCtl  float32 FPAR_scale ;  // used to be in BPLUT, can't find new one so now in here
            metObj_p->fpar = (float64) metArrayObj_p->fpar[s1km] * runtimeParamObj_p->FPAR_scale + runtimeParamObj_p->FPAR_offset;
            gppInputUsed = 0; // finally, indicate daily fpar usage

          } else {
            // Here, the fpar_qa >= FPAR_QA_USE_CLIM_THRES so use FparClimatology
            // as source for fpar.
            // Poor quality FPAR, fill with climatology (unpack climatology, apply gain and offset)
            // Unpack 1-km fpar climatology array for model step call interface
            metObj_p->fpar = (float64) metArrayObj_p->fpar_clim[s1km] * runtimeParamObj_p->FPAR_scale + runtimeParamObj_p->FPAR_offset;
            gppInputUsed = 1; // indicate that FparClimatology was used (fpar climatology usage)

        }  // end :: Fpar vs. Fpar Clim decision

    } // end :: VI vs. Fpar decision

    // gppInputUsed can take on these values {0,1,2} in normal use, -1 if not overridden
    
    // Return: gppInputUsed, a 4-state algorithm level flag indicating which input was used to drive GPP
    return(gppInputUsed);

} // end :: tcfModFunc_switchFparVi()




// ---------------------------------------------------------------
// function : tcfModFunc_forwardStep()
// Purpose  : Fundamental model core.  Propagates model states for 
//            one computational unit (daily and 1km grid cell).
// Inputs  :  Previous state scalars (STATE), Dynamic daily met 
//            scalar data (MET), and static model parameters (BPLUT).    
// Outputs: States updated to current timestep (STATE).
// Notes: 
//
void tcfModFunc_forwardStep(BPLUT *bp_p, MetAdt MET, EcoStateAdt *STATE)
{

  int i = 0;

  /* <LAJ> I wrapped some sub-calculations into these two functions because
     these quantities are needed by TCF_ANASTST(), and I didn't want to
     have model caculations freely floating around in my driver functions. */
  tcfModFunc_gpp(bp_p, MET, STATE);
  tcfModFunc_kmult(bp_p, MET, STATE);

  // Heterotrophic respiration fluxes
  STATE->rh[0] = bp_p->kopt * STATE->rh_mult.k * STATE->C[0];
  STATE->rh[1] = bp_p->kopt * bp_p->kstr * STATE->rh_mult.k * STATE->C[1];
  STATE->rh[2] = bp_p->kopt * bp_p->kslw * STATE->rh_mult.k * STATE->C[2];

  // Net Cpool change
  STATE->dC[0] = STATE->lit * bp_p->fmet - STATE->rh[0];
  STATE->dC[1] = STATE->lit * (1.0 - bp_p->fmet) - STATE->rh[1];
  STATE->dC[2] = bp_p->fstr * STATE->rh[1] - STATE->rh[2];

  //  Adjust rhstr for material transfered to slow pool
  STATE->rh[1] *= (1.0 - bp_p->fstr);

  // Initialize sums
  STATE->rhtot = 0.0;
  STATE->Ctot = 0.0;
  STATE->dCtot = 0.0;

  // Loop through SOC (3) pool STATEs
  for (i=0;i < NPOOL;i++)
  {

    // sum rh fluxes for the SOC pools
    STATE->rhtot += STATE->rh[i];

    // sum dC
    STATE->dCtot += STATE->dC[i];

    // Update SOC pools for next timestep
    STATE->C[i] += STATE->dC[i];

    // Check lower bounds: if a C pool is below zero, then set to zero
    STATE->C[i] = (STATE->C[i]<0.0)? 0.0 : STATE->C[i];

    // sum C fluxes for the SOC pools
    STATE->Ctot += STATE->C[i];

    // ALSO NEED TO ADJUST THE DC and RH fluxes IF C IS ADJUSTED TO ZERO

  }

  // Get current reco flux
  STATE->reco = STATE->rhtot + STATE->gpp * bp_p->fraut;

  // Nee
  STATE->nee = STATE->reco - STATE->gpp;

} // end :: tcfModFunc_forwardStep()


// ---------------------------------------------------------------
// function : tcfModFunc_gpp()
// Purpose  : Does fundamental GPP and NPP flux calculations. 
// Inputs   : Dynamic daily met data scalars (MET), and static model parameters (BPLUT).
// Outputs  : Daily GPP and NPP states returned in STATE structure.
// Notes   : 
//
void tcfModFunc_gpp(BPLUT *bp_p, MetAdt MET, EcoStateAdt *STATE)
{
  float64  *resArray = NULL;

  // Get gpp multipliers and derivatives (for later QA calculation)

  // Gpp multipliers
 
  // Tmin multiplier
  resArray = tcfModMult_ramp(MET.tmin, bp_p->Tmin_min, bp_p->Tmin_max);
  STATE->gpp_mult.tmin = resArray[0];
  STATE->gpp_mult.d_tmin = resArray[1];

  // Vpd multiplier
  // NOTE: vpd has negative response slope, so reverse min/max in the arg list 
  resArray = tcfModMult_ramp(MET.vpd, bp_p->VPD_max, bp_p->VPD_min);
  STATE->gpp_mult.vpd = resArray[0];
  STATE->gpp_mult.d_vpd = resArray[1];

  // Root-zone soil moisture multiplier
  resArray = tcfModMult_ramp(MET.smrz, bp_p->SMrz_min, bp_p->SMrz_max);
  STATE->gpp_mult.smrz = resArray[0];
  STATE->gpp_mult.d_smrz = resArray[1];

  // Freeze-Thaw multiplier
  resArray = tcfModMult_gppFt(bp_p, MET.ft);
  STATE->gpp_mult.ft = resArray[0];
  STATE->gpp_mult.d_ft = resArray[1];

  // Bulk GPP met constraint multiplier
  STATE->gpp_mult.k = STATE->gpp_mult.tmin * STATE->gpp_mult.vpd * STATE->gpp_mult.smrz * STATE->gpp_mult.ft; 

  // Perform the GPP calc
  STATE->gpp = bp_p->LUEmax * MET.par * MET.fpar * STATE->gpp_mult.k;

  // Perform the NPP calculation
  STATE->npp = STATE->gpp * (1.0 - bp_p->fraut);

} // end :: tcfModFunc_gpp()


// ---------------------------------------------------------------
// function : tcfModFunc_anaStst()
// Purpose  : Fundamental model analytical steady-state calculation
//            to initialize SOC states.
// Inputs: Kmult and NPP annual sum scalars (STATE) and static model parameters (BPLUT).    
// Outputs: Approximate (average annual) steady-state SOC returned in STATE.
// Notes    : Logic MUST parallel that of the update equation 
//            in 'tcfModFunc_forwardStep()'. The initialized SOC 
//            pools for this method are for ANNUAL conditions and 
//            therefore will not be the exactly initialized 
//            values, because all the pools vary seasonally.
//
void tcfModFunc_anaStst(BPLUT *bp_p, EcoStateAdt *STATE)
{

      //  Metabolic C pool [g C m-2]
      STATE->C[0] = STATE->npp_sum * bp_p->fmet / (bp_p->kopt * STATE->kmult_sum);

      //  Structual C pool [g C m-2]
      STATE->C[1] = STATE->npp_sum * (1.0 - bp_p->fmet) / (bp_p->kopt * bp_p->kstr * STATE->kmult_sum);

      //  Slow C pool [g C m-2]
      STATE->C[2] = bp_p->fstr * bp_p->kstr * STATE->C[1] / bp_p->kslw; /*kmult_sum and kopt cancel*/

} // end :: tcfModFunc_anaStst()


// ---------------------------------------------------------------
// function : tcfModFunc_kmult()
// Purpose  : Does fundamental Rh-related meteorological rate
//            constraint multiplier sub-calculations.   
// Inputs: Daily met data scalars (MET), and static model parameters (BPLUT)
// Outputs: Kmult, the bulk meteorological rate constraint multiplier for Rh, is returned in the STATE structure.
// Notes: This subroutine is seperated from the model step logic to facilitate the match-up between model state 
//        dynamics and analytical steady-state solution.  The separation reduces the possiblity of an inadvertant 
//        dynamics mismatch (i.e. "a mistake").  
//
void tcfModFunc_kmult(BPLUT *bp_p, MetAdt MET, EcoStateAdt *STATE)
{

  float64  *resArray;

  // Get rh multipliers and derivatives (for later QA calculation)

  // Temperature multiplier
  resArray = tcfModMult_rhTsoil(bp_p, MET.tsoil);
  STATE->rh_mult.tsoil = resArray[0];
  STATE->rh_mult.d_tsoil = resArray[1];

  //  Moisture multiplier
  resArray = tcfModMult_ramp(MET.smtop, bp_p->SMtop_min, bp_p->SMtop_max);
  STATE->rh_mult.smtop = resArray[0];
  STATE->rh_mult.d_smtop = resArray[1];

  // Product of all multipliers
  STATE->rh_mult.k = STATE->rh_mult.tsoil * STATE->rh_mult.smtop;

} // end :: tcfModFunc_kmult()
