// ---------------------------------------------------------------
// function     : l4c_DAILY_ModFunc.c
// purpose      : Encapsulates tcfModFunc.c
// author       : Lucas A. Jones; lucas@ntsg.umt.edu
// author       : joe glassy, Lupine Logic Inc.
// version      : see L4C_VERS_TAG , as 4.1.5 or later
// revised      : 2016-03-03T09:30 <LAJ> revised all functions with derivative discontinuities
//              :  such that the 'd_mult' is set to zero at the discontinuity 
//              :  (e.g.  '<' and '>', changed to '<=' and '>=' to agree with science code)
//              :  additional minor changes to ensure agreement with science code.
//              : 2015-08-26T14:00, <LAJ> revised rescalSmrz() to agree
//              :  with updated tcf science code the two code versions 
//              :  were verified in agreement for simulation day April 13, 2015.
// history      : 2014-08-07T11:32, <JMG> added missing local intermediate terms (mult,dmult)
// history      : 2014-02-05T11:242:00-07:00,wheat,jmg"
// dependencies : HDF4, HDF5 libraries and L4C FRAMEWORK API.
// history      : n/a
// notes        : Based on and adapted from tcf_spin v3.1/3.2
// ---------------------------------------------------------------

// #include "tcf_inc.h"
#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"

// --------------------------------------------------------------
/*
 module       : tcfModMult.c
 purpose      : Low-level rate constraint multiplier functions required by model core logic and derivatives for 
                QA error propagation. These functions compute only meteorological rate constraints and associated 
        derivatives.  
 project      : SMAP L4C science development.
 dependencies : 
 revision     : 2014-02-15T12:52:00, for LEAD FORK with new ADT namespaces
 history      : Updated root zone soil moisture scaling function -- laj 08/25/15
              : Initial reformating -- laj 09/10/13
  
 author       : Lucas A. Jones; lucas@ntsg.umt.edu
 notes        : This module could be combined with 'tcfModFunc.c', but I wanted to keep rate functions together
                since there are many and they have similar purpose. 
*/


// --------------------------------------------------------------
// tcfModMult_ramp()

// Purpose: Generic linear ramp multiplier and derivative function. Mainly used in GPP flux calculation.
//
// Notes: Can be applied for any function with a linear increase/decrease between 0 and 1 and otherwise 
//        held constant (at 0 or 1).
//
// Inputs: Indpendent variable (x) and its extrema (which determine both the slope and where the rate multiplier
//         reach its constant 0 or 1 values).  
//
// Outputs: Rate multiplier constrained on the unit interval [0, 1] and derivative.


float64 *tcfModMult_ramp(float64 x, float32 xmin, float32 xmax)
{

  static float64   resArray[2];

  float64  mult  = 0.0;
  float64  dmult = 0.0;

  // Domain check to avoid dividing by zero
  if (fabs(xmin - xmax) < FLT_TOL)
    {
    
      xLogMsg("@E Ramp constraint function parameter xmin > xmax. Check BPLUT.\n");
      // <LAJ> 2016-03-03T09:30  These three lines required to agree with science code.
      // Otherwise this function falls through without assigning values to 'resArray'
      // return results 
      resArray[0] = 1.0;
      resArray[1] = 0.0;
      return(resArray);
    }

  // multiplier
  mult = (float64) (x - xmin) / (xmax - xmin);

  // set multiplier to constant outside of ramp
  mult = (mult>1.0) ? 1.0 : mult;
  mult = (mult<0.0) ? 0.0 : mult;

  // derivative
  dmult = (float64) 1.0 / (xmax - xmin);

  // Set derivative to zero outside of ramp
  // <LAJ> 2016-03-03T09:30  These were '<' and '>', changed to '<=' and '>=' to agree with science code
  dmult = (mult >= 1.0) ? 0.0 : dmult;
  dmult = (mult <= 0.0) ? 0.0 : dmult;

  // package return results 
  resArray[0] = mult;
  resArray[1] = dmult;
 
  return(resArray);

} // end :: tcfModMult_ramp()


// -------------------------------------------------------------------------------------------------------------
// function: tcfModMult_gppFt()
// Purpose :  Freeze-thaw constraint multiplier for GPP based on freeze-thaw state.  
// Inputs  : BPLUT *bp_p -- ptr to BPLUT record for this PFT
//           uint8 ft    -- freeze thaw value for this pixel
//           prior: Freeze-thaw state (uint8, 0 or 1) and model parameters (GLOB_PARAMS). 
// Outputs : Freeze-thaw rate multiplier constrained on the unit interval [0, 1] and derivative. 
// notes   : Since freeze-thaw is a boolean value, this uses a simple either-or function. 
//         : <LAJ> 2016-03-03T09:30  This and other FT-related error codes should be more thoroughly checked
//         :       prior to use in operational codes, since it currently does not have a counterpart in the 
//         :       science prototype code.
float64 *tcfModMult_gppFt(BPLUT *bp_p, uint8 ft)
{
 char   fxName[256] = {0} ;
 
 // <JMG> 2014-08-07T11:41, added missing long-lived gppResArray[] 
 static float64   gppResArray[2] = {0} ;
 
 float64 mult = 0.0 ; // <JMG> 2014-08-07T11:32, added local intermediate term 
 float64 dmult= 0.0 ; // <JMG> 2014-08-07T11:32, local intermediate term 
 

    strcpy(fxName,__func__);
  // Determine multiplier and derivative 
  
  // Frozen conditions (ft = 1)
    if (ft == 1)
    {
        mult = (float64) bp_p -> FT_min;
        dmult = (float64) bp_p -> FT_min;

    // Thawed conditions (ft = 0)
    // NOTE: FT_max should always = 1, but I added this condition because we have an FT_max field in the ATBD's BPLUT 
    } else if (ft == 0) {

        mult = (float64) bp_p -> FT_max;
        dmult = (float64) 1.0 - bp_p -> FT_max;
    
    } else {
        // trip fatal error if F/T contains an invalid value...
        xLogMsg("@F In(%s) Freeze-thaw contains (%d) invalid value, must be either 0 or 1. Check source data~ \n",
                ft, fxName);
    }

  // package return results 
  gppResArray[0] = mult;
  gppResArray[1] = dmult;
 
  return(gppResArray);

} // end :: tcfModMult_gppFt() 


// -------------------------------------------------------------------------------------------------------------
// tcfModMult_rhTsoil()

// Purpose: Soil temperature Arrhenius (exponential) multiplier function and derivative for heterotrophic 
//          respiration (Rh).
//
// Notes: 
//
// Inputs: Soil temperature in Kelvin [K] and model parameters (GLOB_PARAMS). 
//
// Outputs: Rh soil temperature rate multiplier constrained on the unit interval [0, 1] and derivative. 

float64 *tcfModMult_rhTsoil(BPLUT *bp_p, float64 tsoil)
{
  static float64   resArray[2] = {0} ;

  float64   theta, theta2;
  float64   mult, dmult;

  // multiplier
  theta = bp_p->Tsoil_beta0*(1.0/bp_p->Tsoil_beta1 - 1.0/(tsoil - bp_p->Tsoil_beta2));
  mult = exp(theta);

  mult = (mult>1.0) ? 1.0 : mult;
  mult = (mult<0.0) ? 0.0 : mult;

  // derivative
  theta = (tsoil - bp_p->Tsoil_beta2);
  theta2 = pow(theta, 2.0); 
  dmult = mult * bp_p->Tsoil_beta0 / theta2;

  // <LAJ> 2016-03-03T09:30  These were '<' and '>', changed to '<=' and '>=' to agree with science code
  dmult = (mult>=1.0) ? 0.0 : dmult;
  dmult = (mult<=0.0) ? 0.0 : dmult;

  //package return results
  resArray[0] = mult;
  resArray[1] = dmult;

  return(resArray);

} // end :: tcfModMult_rhTsoil() 



// -------------------------------------------------------------------------------------------------------------
// function :tcfModMult_rescalSmrz()
// purpose  :  Rescale root zone soil moisture using min/max for gridcell and log transform. 
// inputs   : float32 smrz, float32 smrz_mn, float32 smrz_mx
// notes    : on Inputs, Daily root zone soil moisture and time record min/max in fractional saturation [dim.]. 
// Outputs  : float64, res -- Rescaled daily root zone soil moisture in [% Sat.].
// revised  :2015-10-06T17:00, <LAJ> added catch which forces min/max to be at least equal to current sm. 
//           2015-08-26T14:00, <LAJ> revised rescalSmrz() to agree with updated tcf science code
//          :                         the two code versions were verified in agreement for simulation day April 13, 2015.
// notes: n/a 
float64 tcfModMult_rescalSmrz(float32 smrz, float32 smrz_mn, float32 smrz_mx)
{

  float64   res = 0;

  float64   sm64, sm64_mn, sm64_mx; 
  float64   smnorm, smnormlog;
  float64   log101;

  // Rescast inputs as double precision and rescale to percent saturation
  // Offsets help prevent taking log of 0.0
  sm64 = (float64) smrz * 100.0; 
  sm64_mn = (float64) smrz_mn * 100.0;
  sm64_mx = (float64) smrz_mx * 100.0;

  // Ensure min/max are not exceeded
  // <LAJ> 12/08/15  added catch for min/max exceedence 
  sm64_mn = (sm64 < sm64_mn) ?  sm64 : sm64_mn;
  sm64_mx = (sm64 > sm64_mx) ?  sm64 : sm64_mx;

  // Normalize
  smnorm = (sm64 - sm64_mn) / (sm64_mx - sm64_mn) * 100.0 + 1.0;

  // Log transform normalized data and rescale to range between 5.0 and 100 [% Sat.]
  smnormlog = log(smnorm);
  log101 = log(101.0);
  
  res = smnormlog / log101 * 95.0 + 5.0;

  return(res);


} // end :: tcfModMult_rescalSmrz()  

