// ---------------------------------------------------------------
// function     : l4c_DAILY_PkgMgrInit.c
// purpose      : Initializes PkgMgr mechanism for session. Prior to any
//                binding to actual heap variables, it is permissible to
//                set the names of themes and the child nodes owned by each
//                theme. Later, the variables may be bound, and after that,
//                may be accessed anywhere the PkgMgr is available.
// version      : see L4C_VERS_TAG , as v4.1.5 or later
// revision     : 2016-02-18, changed all mult_av to mult_mean
// revision     : 2016-01-04, continued RMSE integration and additions.
// revision     : <JMG> 2015-07-30, add support for GEOS5 FP TS derived F/T
// revision     : <JMG> 2014-08-07, deprecated MERRA climate min,max layers
//                and added ft_state_um_m03, and support for new
//                SMAP NatureRun soil rootzone min,max daily-update images.
// history      : 2014-05-19T12:13, consolidated ALL functions originally in
//                "l4c_DAILY_PkgMgrInit_Themes.c" into this single module
//                file l4c_DAILY_PkgMgrInit.c, and deprecated a number of
//                un-used, obsolete THEMES, as shown below.
// history      : 2014-02-21T15:50:00, added themes for QA, 9km Post processing
// history      : v4.1.6-2014-02-05T11:242:00-07:00,wheat,jmg"
// dependencies : HDF4, HDF5 libraries and L4C FRAMEWORK API.
// author       : joe glassy, Lupine Logic Inc.
// history      : n/a
// notes        : THEME CONFIGURATION For L4_C: one per major ADT
//                in tcf_struct.h :
// ---------------------------------------------------------------

// #include "tcf_inc.h" now re-factored into l4c_DAILY.h
#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"

// ---------------------------------------------------------------
// function   : l4c_PkgMgr_Init()
// Purpose    : Initializes PkgMgr (master scoping tool index containers) and
//              a collection of child PkgThemes, where 'many' PkgThemes bind to
//              each PkgMgr instance.
// Inputs     : int nThemes (currently 20) - count of at least as many
//               PkgMgr themes we'll use, with one per major ADT
//               in tcf_struct.h, and others used for pre,post processing
// Revised    : 2016-01-04, continued RMSE integration and additions.
// History    : 2014-05-19, updated nThemes count to 20, and note that
//               all pObjCounts slots with a value of '0' have been
//               deprecated due to not being used or having been superceded
//               via consolidation into others.
// History    : 2014-05-16T11:46:00, set '25' for PRE_PROC_DEFLATE_1KM
//              theme, to accomodate fpar_clim and ndvi_M01, ndvi_qc_M01
// History    : 2014-04-29T11:32:00, MetArrayAdt owns 18 child PkgObj,to fit new
// fpar_clim History    : 2014-02-24T15:41:00, added qv2m,t2m_avg,ps_avg
//              to MetArrayAdt and MetAdt, to consolidate, avoiding another
//              speciality ADT
// History    : 2014-02-21T15:30, added a theme for SMAP_L4_C_Ancillary
//              datasets, so now using (20) themes of the (24) declared
//              candidate themes
// History    : 2014-02-20T13:47:00
// Returns    : int, status
// Notes      : CALLED BY: l4c_SesInit()
// Author     : joe glassy
int l4c_PkgMgr_Init(const int nThemes) {
  SesCtl *sesCtl_p = l4c_GetSesCtl();
  PkgMgr *pkgMgr_p = NULL;     // convenience ptr to OVERSEER
  PkgTheme *pkgTheme_p = NULL; // convenience ptr to middle "parent"
  int status = 0;

  // PkgMgr: individual counts of PkgObj's that theme {0..19} owns, now 20,
  // ASSERT: we always size 'L4C_PKGMGR_THEME_COUNT' as >= the number of
  //         these we actually use. More is OK, less is fatal!
  int pObjCounts[L4C_PKGMGR_THEME_COUNT] = {0};

  // CREATE master PkgMgr instance as a scoping tool.
  //  We initially create only the empty (parent) instance, declaring
  //  0 themes here, since we put off the creation of the actual
  //  themes using pkgMgr_CreateThemes() below
  sesCtl_p->l4cPkgMgr_p =
      pkgMgr_CreateMgr("L4_C_MAIN_PkgMgr", 0, "L4_C_DAILY_Datasets");

  // assign a local convenience pointer to PkgMgr overseer
  pkgMgr_p = sesCtl_p->l4cPkgMgr_p;

  // Initialize the integer array (pObjCounts[]) that tracks the
  // number of (child) PkgObj objects that each theme, identified by
  // its slot index, will own.

  // <JMG> 2014-05-19, all pObjCounts slots with a value of '0' have been
  // deprecated due to not being used or having been superceded via
  // consolidation into others.

  // DEFINE in pObjCounts[] array, for each PkgTheme, the count of
  // child PkgObj instances this PkgTheme ("L4C_OUTPUTS") owns.

  pObjCounts[0] = 22; // Theme: SMAP_L4_C_Ancillary, now with file-level info
  pObjCounts[1] =
      10; // Theme: SMAP_L4_C_SOC.h5 datasets, now with file-level info
  // <JMG> 2015-07-29 added 1, total now 12, to support new 9km F/T
  pObjCounts[2] =
      12; // Theme: SMAP_L4_C_MET.h5 datasets, now with file-level info

  pObjCounts[3] = 6; // Theme: SMAP_L4_C_FPAR_8DAY (now 6, added (2) more for
                     // ndvi, ndvi_qc, file info

  // for now, includes both: fpar_M01_dayDDD_perPP and fpar_qc_M01_dayDDD..
  pObjCounts[4] =
      3; // Theme: SMAP_L4_C_FparClimatology datasets and file level info
  pObjCounts[5] =
      17; // Theme: SMAP_L4_C_LongtermClimatology (was L4_C_MerraClimatology)

  // <JMG> 2014-05-16, changed to 25 to allow for (3) new fields,
  // viirsNdvi_M01,viirsNdvi_qc_M01 and fpar_clim_dayDDD_perPP
  // <JMG> 2014-05-19, added (2) PkgObj so ALL file-level entries are included

  // <JMG> 2015-07-30 added support for new GEOS5 FP TS derived 3KM Freeze/Thaw
  // when L3_SM_A is NOT available. This is another fail-over mechanism.
  pObjCounts[6] = 29; // Theme: "PRE_PROC_DEFLATE_1KM"

  pObjCounts[7] = 12; // Theme: "POST_PROC_INFLATE_9KM"

  pObjCounts[8] = 64; // Theme: "L4C_OUTPUTS" merged ALL, now 64 child PkgObj

  // <JMG> 2014-11-14, added RMSE as switched option, not yet implemented
  pObjCounts[8] = 73; // Theme: "L4C_OUTPUTS" now {0..72}

  // <JMG> 2015-07-29, added support for GEOS5 FP TS derived F/T within existing
  // theme layout
  pObjCounts[9] = 19; // Theme: MetArrayAdt owns 19 child PkgObj, for:
                      // fpar_clim, ndvi,ndvi_qc

  // <KAE> 2024-07-16 Added litter_rate
  pObjCounts[10] = 37; // Theme: EcoStateArrayAdt owns N child PkgObj

  pObjCounts[11] = 1;  // Theme: "Metagroup_CRID"
  pObjCounts[12] = 10; // Theme: "Metagroup_Config"
  pObjCounts[13] = 22; // Theme: "Metagroup_DatasetIdentification"
  pObjCounts[14] = 7;  // Theme: "Metagroup_Extent"
  pObjCounts[15] = 7;  // Theme: "Metagroup_GridSpatialRepresentation"
  pObjCounts[16] = 1;  // Theme: "Metagroup_ProcessStep"
  pObjCounts[17] = 1;  // Theme: "Metagroup_SeriesIdentification"
  // <JMG> 2014-09-03T13:17, added as per Joe Ardizzone's suggestion
  pObjCounts[18] = 1; // Theme: "Metagroup_AcquisitionInformation"

  // <JMG> 7/23/2014, added for Jeff Lee style externalization of
  // HDF5 dataset output properties.
  // TBD: could eventually use prior theme L4C_OUTPUTS but until its tested
  // we use this dedicated theme. One PkgObj per OUTPUT dataset.
  pObjCounts[19] = 46; // Theme: "L4C_OUTPUT_VARIABLE_PROPERTIES"

  // Reminder: pkgMgrCreateThemes uses the above counts, in the
  //   pObjCounts[] array, to assign each of nThemes with this number of
  //   child (PkgObj) entries carried in the pObjCounts[] slot
  pkgMgr_CreateThemes(pkgMgr_p, nThemes, pObjCounts, "SMAP_L4_C");

  pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 0);
  l4c_PkgInit_SMAP_L4_C_Ancillary(pkgTheme_p, "SMAP_L4_C_Ancillary");

  pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 1);
  l4c_PkgInit_SMAP_L4_C_SOC(pkgTheme_p, "SMAP_L4_C_SOC");

  pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 2);
  l4c_PkgInit_SMAP_L4_C_MET(pkgTheme_p, "SMAP_L4_C_MET");

  pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 3);
  l4c_PkgInit_SMAP_L4_C_FPAR_8DAY(pkgTheme_p, "SMAP_L4_C_FPAR_8DAY");

  pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 4);
  l4c_PkgInit_SMAP_FparClimatology(pkgTheme_p, "SMAP_L4_C_FparClimatology");

  pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 5);
  l4c_PkgInit_SMAP_LongtermClimatology(pkgTheme_p,
                                       "SMAP_L4_C_LongtermClimatology");

  pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 6);
  l4c_PkgInit_PRE_PROC_DEFLATE_1KM(pkgTheme_p, "PRE_PROC_DEFLATE_1KM");

  pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 7);
  l4c_PkgInit_POST_PROC_INFLATE_9KM(pkgTheme_p, "POST_PROC_INFLATE_9KM");

  pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 8);
  l4c_PkgInit_L4C_OUTPUTS(pkgTheme_p, "L4C_OUTPUTS");

  pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 9);
  l4c_PkgInit_MetArrayAdt(pkgTheme_p, "MetArrayAdt");

  pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 10);
  l4c_PkgInit_EcoStateArrayAdt(pkgTheme_p, "EcoStateArrayAdt");

  // <JMG> 2014-06-24, added a new set of themes loosely grouped
  // here in the Metagroups function, to init formal
  // 'group-of-groups' metadata by group. Here we depart from the
  // 1-function-per-theme initialization pattern above, and init
  // all the metagroup themes in one function.
  pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 11);

  l4c_PkgInit_Metagroups(pkgMgr_p);

  // <JMG> 7/23/2014 added, but not yet fully tested
  pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 19);
  pkgMgr_SetThemeName(pkgTheme_p, "L4C_OUTPUT_VARIABLE_PROPERTIES");

  l4c_EventCompletionMsg((char *)__func__);

  return status;

} // end:: l4c_PkgMgr_Init()

// ---------------------------------------------------------------
// function : l4c_PkgInit_SetFileProperties()
// purpose  : Assign relevant INPUT file properties, for theme
//            "SMAP_INPUT_HDF5_FILES", as including:
//            {Name,Owner,GenericPath,Datatype}
// inputs   : void
// notes    : CRITICAL: must wait to call THIS function until AFTER all
//            themes have been named, and until after all HDF5 files
//            have been opened, so their file_id's are available
// notes    : CALLED BY: l4c_SesInit()
// notes    : As long as the file_id's don't change this function may
//             safely be called multiple times, where perhaps not all the
//             files were opened the first time, but were opened by
//             subsequent calls...
int l4c_PkgInit_SetFileProperties(void) {
  SesCtl *sesCtl_p = l4c_GetSesCtl(); // ptr to session controls structure

  // declare the PkgMgr mechanism parts we need
  PkgMgr *pkgMgr_p = NULL;  // ptr to overseer
  PkgTheme *pThme_p = NULL; // ptr to theme (parent) instance
  PkgObj *pObj_p = NULL;    //  ptr to child instances

  // First, set local pointer to master instance of the PkgMgr overseer
  pkgMgr_p = sesCtl_p->l4cPkgMgr_p;

  // Retrieve themes pointer using its name "SMAP_L4_C_Ancillary"
  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "SMAP_L4_C_Ancillary");
  // <JMG> 2014-05-19, set SELF-REFERENTIAL file info into its own theme
  pObj_p = pkgMgr_GetObjByName(pThme_p, "SMAP_L4_C_Ancillary");

  // Uses owner id member to store this HDF5 file's file_id. when it is
  // available
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->anc_file_id);
  // Uses GenericPath to store filesystem path to indicated file
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputAncPath);
  // Uses Datatype property to store its TIV datatype identifier
  pkgMgr_SetObjDatatype(pObj_p, IsString);

  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "SMAP_L4_C_FPAR_8DAY");
  pObj_p = pkgMgr_GetObjByName(pThme_p, "SMAP_L4_C_FPAR_8DAY");
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->fpar_file_id);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputFparPath);

  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "SMAP_L4_C_MET");
  pObj_p = pkgMgr_GetObjByName(pThme_p, "SMAP_L4_C_MET");
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);

  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "SMAP_L4_C_SOC");
  pObj_p = pkgMgr_GetObjByName(pThme_p, "SMAP_L4_C_SOC");
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->soc_file_id);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputSOCPath);

  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "SMAP_L4_C_FparClimatology");
  pObj_p = pkgMgr_GetObjByName(pThme_p, "SMAP_L4_C_FparClimatology");
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->fparClima_file_id);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputFparClimatologyPath);

  l4c_EventCompletionMsg((char *)__func__);

  return 0;

} // end:: l4c_PkgInit_SetFileProperties()

// ---------------------------------------------------------------
// function  : l4c_PkgInit_L4C_OUTPUTS()
// purpose   : Define names of PkgObj's owned by theme "L4C_OUTPUTS"
// inputs    : PkgTheme *pkgTheme_p,char *themeName_p
// notes     : called by l4c_PkgMgr_Init()
// notes     : TBD: Ideally (eventually) all these unique names should be
//             retrieved via fetch from RIP file keywords using
//             nsbGetValueByKey() semantics
// revised   : 2014-09-03T13:12, merged all QA info into ONE theme
//             of "L4C_OUTPUTS"
void l4c_PkgInit_L4C_OUTPUTS(PkgTheme *pThme_p, char *themeName_p) {
  NsbObj *nsb_p = l4c_GetNsbOverseer();
  char nvpBuf[256] = {0};
  char nvpResult[512] = {0};
  char tivTypeAsStr[32] = {0};
  char fxName[256] = {0};
  int nPkgObj = 0;

  strcpy(fxName, __func__);

  nPkgObj = pkgMgr_GetThemeObjectCount(pThme_p);
  xLogMsg("@A In(%s) Assigning %d %s child names ~ \n", fxName, nPkgObj,
          themeName_p);

  // assign the theme name here "L4C_OUTPUTS"
  pkgMgr_SetThemeName(pThme_p, themeName_p);

  pkgMgr_SetThemes_ChildName(pThme_p, 0, "NEE/nee_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 1, "NEE/nee_std_dev");
  pkgMgr_SetThemes_ChildName(pThme_p, 2, "NEE/nee_pft1_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 3, "NEE/nee_pft2_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 4, "NEE/nee_pft3_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 5, "NEE/nee_pft4_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 6, "NEE/nee_pft5_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 7, "NEE/nee_pft6_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 8, "NEE/nee_pft7_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 9, "NEE/nee_pft8_mean");

  pkgMgr_SetThemes_ChildName(pThme_p, 10, "GPP/gpp_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 11, "GPP/gpp_std_dev");
  pkgMgr_SetThemes_ChildName(pThme_p, 12, "GPP/gpp_pft1_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 13, "GPP/gpp_pft2_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 14, "GPP/gpp_pft3_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 15, "GPP/gpp_pft4_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 16, "GPP/gpp_pft5_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 17, "GPP/gpp_pft6_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 18, "GPP/gpp_pft7_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 19, "GPP/gpp_pft8_mean");

  pkgMgr_SetThemes_ChildName(pThme_p, 20, "RH/rh_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 21, "RH/rh_std_dev");
  pkgMgr_SetThemes_ChildName(pThme_p, 22, "RH/rh_pft1_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 23, "RH/rh_pft2_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 24, "RH/rh_pft3_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 25, "RH/rh_pft4_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 26, "RH/rh_pft5_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 27, "RH/rh_pft6_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 28, "RH/rh_pft7_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 29, "RH/rh_pft8_mean");

  pkgMgr_SetThemes_ChildName(pThme_p, 30, "SOC/soc_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 31, "SOC/soc_std_dev");
  pkgMgr_SetThemes_ChildName(pThme_p, 32, "SOC/soc_pft1_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 33, "SOC/soc_pft2_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 34, "SOC/soc_pft3_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 35, "SOC/soc_pft4_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 36, "SOC/soc_pft5_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 37, "SOC/soc_pft6_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 38, "SOC/soc_pft7_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 39, "SOC/soc_pft8_mean");
  // <JMG> 2016-02-18 changed all mult_av to mult_mean to match Lucas names
  pkgMgr_SetThemes_ChildName(pThme_p, 40, "EC/frozen_area");
  pkgMgr_SetThemes_ChildName(pThme_p, 41, "EC/emult_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 42, "EC/tmult_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 43, "EC/wmult_mean");

  // DEPRECATED. Now allow entry 44..51 inclusive to be blank
  pkgMgr_SetThemes_ChildName(pThme_p, 44, "QA/carbon_model_qual_pft1");
  pkgMgr_SetThemes_ChildName(pThme_p, 45, "QA/carbon_model_qual_pft2");
  pkgMgr_SetThemes_ChildName(pThme_p, 46, "QA/carbon_model_qual_pft3");
  pkgMgr_SetThemes_ChildName(pThme_p, 47, "QA/carbon_model_qual_pft4");
  pkgMgr_SetThemes_ChildName(pThme_p, 48, "QA/carbon_model_qual_pft5");
  pkgMgr_SetThemes_ChildName(pThme_p, 49, "QA/carbon_model_qual_pft6");
  pkgMgr_SetThemes_ChildName(pThme_p, 50, "QA/carbon_model_qual_pft7");
  pkgMgr_SetThemes_ChildName(pThme_p, 51, "QA/carbon_model_qual_pft8");

  // Lucas output(s) named: tcf_nee_qa_dayDDD_m09_pftX_cnt.uint8
  // PSB v1.5, p.4-43, Table 15, this is labelled as "pft_freq{pft:1..8}"
  pkgMgr_SetThemes_ChildName(pThme_p, 52, "QA/qa_count_pft1");
  pkgMgr_SetThemes_ChildName(pThme_p, 53, "QA/qa_count_pft2");
  pkgMgr_SetThemes_ChildName(pThme_p, 54, "QA/qa_count_pft3");
  pkgMgr_SetThemes_ChildName(pThme_p, 55, "QA/qa_count_pft4");
  pkgMgr_SetThemes_ChildName(pThme_p, 56, "QA/qa_count_pft5");
  pkgMgr_SetThemes_ChildName(pThme_p, 57, "QA/qa_count_pft6");
  pkgMgr_SetThemes_ChildName(pThme_p, 58, "QA/qa_count_pft7");
  pkgMgr_SetThemes_ChildName(pThme_p, 59, "QA/qa_count_pft8");
  pkgMgr_SetThemes_ChildName(pThme_p, 60, "QA/qa_count");

  // <JMG> 2014-09-03T14:19, changed order to plut bitflag last
  pkgMgr_SetThemes_ChildName(pThme_p, 61, "QA/carbon_model_bitflag");

  // we depend on GEO lat,lon being at the very END of this
  // collection.

  // Assign names for each child object in theme "L4C_OUTPUTS" as
  // required to implement the pending QA "rmse" variables already
  // computed but not yet written out as L4_C HDF5 granule datasets

  pkgMgr_SetThemes_ChildName(pThme_p, 62, "QA/nee_rmse_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 63, "QA/nee_rmse_pft1_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 64, "QA/nee_rmse_pft2_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 65, "QA/nee_rmse_pft3_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 66, "QA/nee_rmse_pft4_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 67, "QA/nee_rmse_pft5_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 68, "QA/nee_rmse_pft6_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 69, "QA/nee_rmse_pft7_mean");
  pkgMgr_SetThemes_ChildName(pThme_p, 70, "QA/nee_rmse_pft8_mean");

  pkgMgr_SetThemes_ChildName(pThme_p, 71, "GEO/latitude");
  pkgMgr_SetThemes_ChildName(pThme_p, 72, "GEO/longitude");

  l4c_EventCompletionMsg(fxName);

} // end:: l4c_PkgInit_L4C_OUTPUTS()

// ---------------------------------------------------------------
// function  : l4c_PkgInit_MetArrayAdt()
// purpose   : bind ALL the theme's children to the theme..
// inputs    : PkgTheme *pkgTheme_p,char *themeName_p
// revision  : 2015-07-29, added 1 for new 9km GEOS5 F/T or passive radio F/T
// revision  : 2014-08-07, replaced surface_flag_3km with ft_state_um_3km
// revision  : 2014-05-16T13:56:00, added "viirsNdvi_qc" analog to fpar_qc
// revision  : 2014-04-29, updated TOTAL=18 now to address fpar_clim addition
//             made on 4/16/2014 by LAJ
// history   : 2014-03-14, D5.0
// notes     : TBD: change all "fpar_qa" to "fpar_qc" once all dependencies
//             are noted, for compatibility with MODIS HDF4 usage 'fpar_qc'
// notes     : Here we bind up ALL the theme's children regardless
//             if they are used or not, for completeness, in case they
//             are used later.
void l4c_PkgInit_MetArrayAdt(PkgTheme *pkgTheme_p, char *themeName_p) {
  pkgMgr_SetThemeName(pkgTheme_p, themeName_p);

  // assign childs name, via its base0 index
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 0, "metArrayObj");

  pkgMgr_SetThemes_ChildName(pkgTheme_p, 1, "metArrayObj.viirsNdvi");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 2, "metArrayObj.viirsNdvi_qc");

  pkgMgr_SetThemes_ChildName(pkgTheme_p, 3, "metArrayObj.fpar");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 4, "metArrayObj.fpar_qa");

  // <JMG> 2015-07-30, now the 3km F/T may also come from GEOS5 TS regridded to
  // 3km <JMG> 2014-08-07, replaced surflag_3km with uint8 ft_state_um_3km
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 5, "metArrayObj.ft_state_um_3km");

  pkgMgr_SetThemes_ChildName(pkgTheme_p, 6, "metArrayObj.tmin");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 7, "metArrayObj.tmax");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 8, "metArrayObj.vpd");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 9, "metArrayObj.par");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 10, "metArrayObj.tsoil");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 11, "metArrayObj.smtop");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 12, "metArrayObj.smrz");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 13, "metArrayObj.smrz_mn");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 14, "metArrayObj.smrz_mx");
  // <JMG> added next 3 to consolidate
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 15, "metArrayObj.qv2m");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 16, "metArrayObj.t2m_avg");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 17, "metArrayObj.ps_avg");

  // <JMG> added to sync up with new fpar_clim (ARRAY) added 4/16/2014
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 18, "metArrayObj.fpar_clim");

  l4c_EventCompletionMsg((char *)__func__);

} // end:: l4c_PkgInit_MetArrayAdt()()

// ---------------------------------------------------------------
// function  : l4c_PkgInit_PRE_PROC_DEFLATE_1KM()
// purpose   : for theme, "PRE_PROC_DEFLATE_1KM" define all child objects
//             for which we'll later define all (25) specific dataset
//             properties (using "set" pkgObj_Set{Property}() calls)
// inputs    : PkgTheme *pkgTheme_p,char *themeName_p
// revised   : 2015-07-29, added 1 to support new 9KM GEOS5 F/T or passive
//             radiometer F/T orig at 36KM, regridded to 9KM.
// revised   : 2014-08-07,  replaced surflag_3km with uint8 ft_state_um_3km and
//              added ndvi_M01,ndvi_qc_M01,fpar_clim...
// notes     : Streamlined l4c_PreProc1km_func.c to fully utilize the
//             dataset properties in theme "PRE_PROC_DEFLATE_1KM" which
//             store all dataset properties, so we don't need to use the
//             original (2) cumbersome, error-prone string tables to
//             point into various themes...
// notes     : The actual property assignments using pkgObjSet{Prop}()
//             calls are performed by l4c_PreProc_Setup() and by
//
void l4c_PkgInit_PRE_PROC_DEFLATE_1KM(PkgTheme *pkgTheme_p, char *themeName_p) {
  // SesCtl *sesCtl_p= l4c_GetSesCtl();
  NsbObj *nsbObj_p = l4c_GetNsbOverseer();
  char *dsetName_p = NULL;
  char nvpBuf[256] = {0};

  // themeName_p here is "PRE_PROC_DEFLATE_1KM" unless changed...
  pkgMgr_SetThemeName(pkgTheme_p, themeName_p);

  pkgMgr_SetThemes_ChildName(pkgTheme_p, 0, "fpar_M01");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 1, "fpar_qc_M01");

  // 5/16/2014, dynamically assign name for "fpar_clim_dayDDD_perPP"
  // as this is not yet contained in sesCtl_p->fparClimDatasetname,
  // will be once the realtime definition of year-day (and period PP)
  // modeled is known and defined. Yes, this is complicated.
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 2, "fpar_clim_dayDDD_perPP");

  // 5/16/2014, <JMG> add NDVI 16-day, and its QC field..
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 3, "ndvi_M01");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 4, "ndvi_qc_M01");

  pkgMgr_SetThemes_ChildName(pkgTheme_p, 5, "FT_ROLLING_M03");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 6, "FT_ROLLING_AGE_M03");

  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOC_C0", nvpBuf);
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 7, dsetName_p);
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOC_C1", nvpBuf);
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 8, dsetName_p);
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOC_C2", nvpBuf);
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 9, dsetName_p);
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_NPP_SUM", nvpBuf);
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 10, dsetName_p);

  pkgMgr_SetThemes_ChildName(pkgTheme_p, 11, "QV2M_M09_AVG");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 12,
                             "RADIATION_SHORTWAVE_DOWNWARD_FLUX");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 13, "SM_ROOTZONE_WETNESS");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 14, "SM_SURFACE_WETNESS");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 15, "SOIL_TEMP_LAYER1");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 16, "SURFACE_PRESSURE");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 17, "T2M_M09_AVG");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 18, "T2M_M09_MAX");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 19, "T2M_M09_MIN");

  // <JMG> 2015-07-30, added to provide for new GEOS5-FP TS driven Freeze thaw
  // still at 3km, as a work around for failed SMAP active radar, providing
  // defunct radar F/T <JMG> 2014-0-8-07, DEPRECATED
  // pkgMgr_SetThemes_ChildName(pkgTheme_p,20, "SURFACE_FLAG_M03" ); replaced
  // with FT_STATE_UM_M03 which is now in the MET not SOC file
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 20, "FT_STATE_UM_M03");

  // <JMG> 2015-07-29, re-ordered all subsequent indices 21..28 to {22..29} to
  // accomodate insertion of new 9KM GEOS-5 FP TS driven Freeze Thaw regridded
  // in preprocessor to 9KM not 3KM.

  // Note: deprecated MERRA qualifier, and, these stored in SOC file not
  // Ancillary <JMG> 2015-05-27, TBD FIX, don't hard code these names, read from
  // RIP file!
  dsetName_p = nsbGetValueByKey(
      nsbObj_p, "L4C_INPUT_DATASET_SOILROOTZONE_M09_MIN", nvpBuf);
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 21, dsetName_p);

  dsetName_p = nsbGetValueByKey(
      nsbObj_p, "L4C_INPUT_DATASET_SOILROOTZONE_M09_MAX", nvpBuf);
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 22, dsetName_p);

  // This PkgObj stores a single INPUT re-usable 1KM M01 buffer sized as a
  // float32, but also usable for reading in uint8, uint16 etc types
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 23, "PREPROC_1KM_BUFFER");

  // <JMG> 2014-05-19, mirror of properties of HDF5 files, filepath, file_id,
  // etc
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 24, "SMAP_L4_C_FPAR_8DAY");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 25, "SMAP_L4_C_MET");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 26, "SMAP_L4_C_SOC");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 27, "SMAP_L4_C_FparClimatology");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 28, "SMAP_L4_C_Ancillary");

  l4c_EventCompletionMsg((char *)__func__);

} // end:: l4c_PkgInit_PRE_PROC_DEFLATE_1KM()

// ---------------------------------------------------------------
// function  : l4c_PkgInit_POST_PROC_INFLATE_9KM()
// purpose   : Declare the names of target variables for the
//             POST PROCESSING phase.
// inputs    : PkgTheme *pkgTheme_p,char *themeName_p
// notes     : n/a
void l4c_PkgInit_POST_PROC_INFLATE_9KM(PkgTheme *pkgTheme_p,
                                       char *themeName_p) {
  pkgMgr_SetThemeName(pkgTheme_p, themeName_p);

  pkgMgr_SetThemes_ChildName(pkgTheme_p, 0, "STATE.dom_1km");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 1, "STATE.avg_9km");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 2, "STATE.std_9km");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 4, "STATE.cnt_9km ");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 5, "STATE.rank_9km");

  pkgMgr_SetThemes_ChildName(pkgTheme_p, 6, "PostProc-Placeholder_06");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 7, "PostProc-Placeholder_07");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 8, "PostProc-Placeholder_08");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 9, "PostProc-Placeholder_09");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 10, "PostProc-Placeholder_10");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 11, "PostProc-Placeholder_11");

  l4c_EventCompletionMsg((char *)__func__);

} // end:: l4c_PkgInit_POST_PROC_INFLATE_9KM()

// ---------------------------------------------------------------
// function  : l4c_PkgInit_EcoStateArrayAdt()
// purpose   :
// inputs    : PkgTheme *pkgTheme_p,char *themeName_p
// notes     : n/a
void l4c_PkgInit_EcoStateArrayAdt(PkgTheme *pkgTheme_p, char *themeName_p) {
  pkgMgr_SetThemeName(pkgTheme_p, themeName_p);

  // these are M01land or M09land 1D-Sparse Vectors at this point...
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 0, "ecoStateArrayObj.nee");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 1, "ecoStateArrayObj.gpp");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 2, "ecoStateArrayObj.npp");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 3, "ecoStateArrayObj.reco");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 4, "ecoStateArrayObj.rh[NPOOL]");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 5, "ecoStateArrayObj.rh[0]");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 6, "ecoStateArrayObj.rh[1]");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 7, "ecoStateArrayObj.rh[2]");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 8, "ecoStateArrayObj.rhtot");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 9, "ecoStateArrayObj.C[NPOOL]");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 10, "ecoStateArrayObj.C0");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 11, "ecoStateArrayObj.C1");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 12, "ecoStateArrayObj.C2");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 13, "ecoStateArrayObj.Ctot");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 14, "ecoStateArrayObj.dCtot");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 15, "ecoStateArrayObj.Emult");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 16, "ecoStateArrayObj.Tmult");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 17, "ecoStateArrayObj.Wmult");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 18, "ecoStateArrayObj.npp_sum");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 19, "ecoStateArrayObj.kmult_sum");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 20, "ecoStateArrayObj.nee_sum");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 21, "ecoStateArrayObj.gpp_sum");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 22, "ecoStateArrayObj.reco_sum");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 23, "ecoStateArrayObj.rhtot_sum");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 24,
                             "ecoStateArrayObj.ecoStateArrayQaObj");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 25,
                             "ecoStateArrayObj.ecoStateArrayQaObj.gpp_rank");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 26,
                             "ecoStateArrayObj.ecoStateArrayQaObj.nee_rank");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 27,
                             "ecoStateArrayObj.ecoStateArrayQaObj.gpp_error");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 28,
                             "ecoStateArrayObj.ecoStateArrayQaObj.nee_error");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 29, "ecoStateArrayObj.spinup_iter");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 30, "ecoStateArrayObj.spinup_stop");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 31, "ecoStateArrayObj.FrozenArea_9km");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 32, "ecoStateArrayObj.ftRolling_3km");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 33,
                             "ecoStateArrayObj.ftRollingAge_3km");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 34,
                             "ecoStateArrayObj.ftRollingQA_3km"); // jmg added!
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 35,
                             "ecoStateArrayObj.litter_rate"); // <KAE> 2024-07-16 added!
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 36, "ecoStateArrayObj");

  l4c_EventCompletionMsg((char *)__func__);

} // end:: l4c_PkgInit_EcoStateArrayAdt()

// ---------------------------------------------------------------
// function  : l4c_PkgInit_SMAP_L4_C_Ancillary()
// purpose   : Bind all datasets in SMAP_L4_C_Ancillary into the
//           : theme of the same name "SMAP_L4_C_Ancillary" where each
//             HDF5 dataset is mapped to one child PkgObj instance.
// inputs    : PkgTheme *pkgTheme_p,char *themeName_p
// revision  : <JMG> 2014-08-07, deprecated MERRA climate min,max layers
// notes     : n/a
void l4c_PkgInit_SMAP_L4_C_Ancillary(PkgTheme *pkgTheme_p, char *themeName_p) {
  pkgMgr_SetThemeName(pkgTheme_p, themeName_p);

  // these are all HDF5 dataset names in the ancillary .h5
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 0, "L4C_M01_PixFilter_uint8");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 1, "L4C_M01land_PixFilter_uint8");

  pkgMgr_SetThemes_ChildName(pkgTheme_p, 2, "MCD12Q1_M01_lc_dom_uint8");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 3, "MCD12Q1_M01_lc_mask_uint8");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 4, "MCD12Q1_M01land_lc_dom_uint8");

  pkgMgr_SetThemes_ChildName(pkgTheme_p, 5, "MCD12Q1_M03_lc_mask_uint8");

  pkgMgr_SetThemes_ChildName(pkgTheme_p, 6, "MCD12Q1_M09_lc_mask_uint8");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 7, "MCD12Q1_M09_lc_nmargveg_uint8");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 8, "MCD12Q1_M09_lc_nnonveg_uint8");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 9, "MCD12Q1_M09_lc_nveg_uint8");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 10, "MCD12Q1_M09land_col_uint16");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 11, "MCD12Q1_M09land_row_uint16");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 12, "MCD12Q1_M09_lc_dom.uint8");

  // Remember that these (input) dataset names differ from the (output)
  // names the PSD requires them to have, e.g. GEO/latitude, /GEO/longitude
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 13,
                             "SMAP_L4_C_LAT_1624_x_3856_M09_flt32");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 14,
                             "SMAP_L4_C_LON_1624_x_3856_M09_flt32");

  pkgMgr_SetThemes_ChildName(pkgTheme_p, 15,
                             "SMAP_L4_C_LAT_14616_x_34704_M01_flt32");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 16,
                             "SMAP_L4_C_LON_14616_x_34704_M01_flt32");

  pkgMgr_SetThemes_ChildName(pkgTheme_p, 17,
                             "tile_c_M01_1km_ease2_template_int16");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 18,
                             "tile_h_M01_1km_ease2_template_uint8");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 19,
                             "tile_r_M01_1km_ease2_template_int16");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 20,
                             "tile_v_M01_1km_ease2_template_uint8");
  // <JMG> 2014-05-19, added one to represent self-referential, file level info
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 21, "SMAP_L4_C_Ancillary");

  l4c_EventCompletionMsg((char *)__func__);

} // end:: l4c_PkgInit_SMAP_L4_C_Ancillary()

// ---------------------------------------------------------------
// function  : l4c_PkgInit_SMAP_L4_C_SOC()
// purpose   : declare child names owned by theme "SMAP_L4_C_SOC"
// inputs    : PkgTheme *pkgTheme_p,char *themeName_p
// revised   : 2014-05-19T11:57, added pkgObj "SMAP_L4_C_SOC" for file level
// info notes     : The PkgObj (child) name here is the verbatim dataset name.
//             Only the actual SOC and NPP_SUM datasets are in the /SOC
//             group, whereas the others are in the root / group.
void l4c_PkgInit_SMAP_L4_C_SOC(PkgTheme *pkgTheme_p, char *themeName_p) {
  NsbObj *nsbObj_p = l4c_GetNsbOverseer();
  char fxName[256];
  char *dsetName_p = NULL;
  char nvpBuf[256];

  strcpy(fxName, __func__);
  pkgMgr_SetThemeName(pkgTheme_p, themeName_p);

  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_FT_ROLLING", nvpBuf);
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 0, dsetName_p);

  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_FT_ROLLING_AGE", nvpBuf);
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 1, dsetName_p);

  pkgMgr_SetThemes_ChildName(pkgTheme_p, 2,
                             "SOC/FT_QC_M03-NOT_CURRENTLY-USED!");

  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOC_C0", nvpBuf);
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 3, dsetName_p);

  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOC_C1", nvpBuf);
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 4, dsetName_p);

  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOC_C2", nvpBuf);
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 5, dsetName_p);

  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_NPP_SUM", nvpBuf);
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 6, dsetName_p);

  // <JMG> 2014-08-07T14:05
  dsetName_p = nsbGetValueByKey(
      nsbObj_p, "L4C_INPUT_DATASET_SOILROOTZONE_M09_MIN", nvpBuf);
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 7, dsetName_p);

  dsetName_p = nsbGetValueByKey(
      nsbObj_p, "L4C_INPUT_DATASET_SOILROOTZONE_M09_MAX", nvpBuf);
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 8, dsetName_p);

  // <JMG> 2014-05-19, added self-referential file level info
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 9, "SMAP_L4_C_SOC");

  l4c_EventCompletionMsg(fxName);

} // end:: l4c_PkgInit_SMAP_L4_C_SOC()

// ---------------------------------------------------------------
// function  : l4c_PkgInit_SMAP_L4_C_MET()
// purpose   : initialize children in theme "SMAP_L4_C_MET"
// inputs    : PkgTheme *pkgTheme_p,char *themeName_p
// revised   : 2015-07-29, added 1, total now 12, to support new
//             9KM GEOS5 F/T or Passive radiometer F/T
// revised   : 2014-08-07, replaced SURFACE_FLAG_M03 with FT_STATE_UM_M03
// revised   : 2014-05-19T11:57, added pkgObj "SMAP_L4_C_MET" for file level
// info notes     : n/a
void l4c_PkgInit_SMAP_L4_C_MET(PkgTheme *pkgTheme_p, char *themeName_p) {
  NsbObj *nsbObj_p = l4c_GetNsbOverseer();
  char fxName[256];
  char *dsetName_p = NULL;
  char nvpBuf[256];

  strcpy(fxName, __func__);

  pkgMgr_SetThemeName(pkgTheme_p, themeName_p);

  pkgMgr_SetThemes_ChildName(pkgTheme_p, 0, "QV2M_M09_AVG");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 1,
                             "RADIATION_SHORTWAVE_DOWNWARD_FLUX");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 2, "SM_ROOTZONE_WETNESS");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 3, "SM_SURFACE_WETNESS");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 4, "SOIL_TEMP_LAYER1");

  // replaced SURFACE_FLAG_M03 with FT_STATE_UM_M03
  // pkgMgr_SetThemes_ChildName(pkgTheme_p, 5,"SURFACE_FLAG_M03");
  // L4C_INPUT_DATASET_FT_STATE_UM_M03
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_FT_STATE_UM_M03", nvpBuf);
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 5, dsetName_p);

  pkgMgr_SetThemes_ChildName(pkgTheme_p, 6, "SURFACE_PRESSURE");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 7, "T2M_M09_AVG");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 8, "T2M_M09_MAX");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 9, "T2M_M09_MIN");
  // <JMG> 2014-05-19, added one for self-referential file level info
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 10, "SMAP_L4_C_MET");

  l4c_EventCompletionMsg(fxName);

} // end:: l4c_PkgInit_SMAP_L4_C_MET()

// ---------------------------------------------------------------
// function  : l4c_PkgInit_FPAR_8DAY()
// purpose   : initialize children in theme "SMAP_L4_C_FPAR_8DAY"
// inputs    : PkgTheme *pkgTheme_p,char *themeName_p
// revised   : 2014-05-19T11:57, added pkgObj "SMAP_L4_C_FPAR_8DAY" for file
// level info revised   : 2014-05-16T12:32:00, added (2) for ndvi,ndvi_qc notes
// : Entries for both FPAR and NDVI are declared regardless of
//             which mode gppMethodStr indicates for this day!
void l4c_PkgInit_SMAP_L4_C_FPAR_8DAY(PkgTheme *pkgTheme_p, char *themeName_p) {
  NsbObj *nsbObj_p = l4c_GetNsbOverseer();
  char fxName[256];
  char *dsetName_p = NULL;
  char nvpBuf[256];

  strcpy(fxName, __func__);

  pkgMgr_SetThemeName(pkgTheme_p, themeName_p);

  pkgMgr_SetThemes_ChildName(pkgTheme_p, 0, "fpar_M01");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 1, "fpar_qc_M01");
  // <JMG> 2014-08-20, deprecated!
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 2, "fpar_qc_extra_M01");

  // <JMG> 2014-05-16, added to fully support NDVI
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 3, "ndvi_M01");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 4, "ndvi_qc_M01");
  // <JMG> 2014-05-19, added for file level info
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 5, "SMAP_L4_C_FPAR_8DAY");

  l4c_EventCompletionMsg((char *)__func__);

} // end:: l4c_PkgInit_Fpar_8day()

// ---------------------------------------------------------------
// function  : l4c_PkgInit_SMAP_FparClimatology()
// purpose   : Initialize the theme "SMAP_L4_C_FparClimatology"
// inputs    : PkgTheme *pkgTheme_p,char *themeName_p
// revised   : 2014-05-19T11:57, added pkgObj "SMAP_L4_Climatology" for file
// level info revision  : 2014-04-22T11:52, prep for v5.0.1+ re-verify,
//             enabling FparClimatology pre-proc as part of
//             standard actions
// notes     : n/a
void l4c_PkgInit_SMAP_FparClimatology(PkgTheme *pkgTheme_p, char *themeName_p) {
  pkgMgr_SetThemeName(pkgTheme_p, themeName_p);

  // Warning: the actual datasetnames are of the form
  // "fpar_clim_M01_dayDDD_perPP" but we deliberately omit these
  // datetime qualifiers since these are late-bound, based on which
  // day-of-year and period is known only at run time.
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 0, "fpar_clim_M01_day");
  // <JMG> 2014-05-19, added for file level info
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 2, "SMAP_L4_C_FparClimatology");

  l4c_EventCompletionMsg((char *)__func__);

} // end:: l4c_PkgInit_SMAP_FparClimatology()

// ---------------------------------------------------------------
// function  : l4c_PkgInit_SMAP_LongtermClimatology()
// purpose   : Initialize the PkgMgr LongtermClimatology PkgObj child names
// inputs    : PkgTheme *pkgTheme_p,char *themeName_p
// notes     : was MerraClimatology but now made more generic since
//             we are migrating to a different (non-Merra) source such
//             as new GEOS5 FP or FP-IT long term climatology
void l4c_PkgInit_SMAP_LongtermClimatology(PkgTheme *pkgTheme_p,
                                          char *themeName_p) {
  pkgMgr_SetThemeName(pkgTheme_p, themeName_p);

  // Remember in operation to search-replace the 'DDD' with 001..366
  // based on day-of-year..
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 0, "gwetroot_mu_dayDDD");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 1, "gwetroot_sig2_dayDDD");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 2, "gwettop_mu_dayDDD");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 3, "gwettop_sig2_dayDDD");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 4, "swgdn_mu_dayDDD");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 5, "swgdn_sig2_dayDDD");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 6, "t2m_avg_mu_dayDDD");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 7, "t2m_avg_sig2_dayDDD");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 8, "t2m_max_mu_dayDDD");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 9, "t2m_max_sig2_dayDDD");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 10, "t2m_min_mu_dayDDD");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 11, "t2m_min_sig2_dayDDD");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 12, "tsoil1_mu_dayDDD");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 13, "tsoil1_sig2_dayDDD");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 14, "vpd_avg_mu_dayDDD");
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 15, "vpd_avg_sig2_dayDDD");

  // <JMG> 2014-05-19, addedfor file level info
  pkgMgr_SetThemes_ChildName(pkgTheme_p, 16, "SMAP_L4_C_LongtermClimatology");

  l4c_EventCompletionMsg((char *)__func__);

} // end:: l4c_PkgInit_SMAP_LongtermClimatology()

// ---------------------------------------------------------------
// function  : l4c_PkgInit_SMAP_Metagroups()
// purpose   : Initialize the PkgMgr Metagroups theme to facilitate
//             ISO metadata element assignment by HDF5 GROUP
// inputs    : PkgTheme *pkgTheme_p -- ptr to PkgMgr "theme" object
//             that owns a collection of PkgObj child objects
//           : char *themeName_p
// returns   : void
// notes     : Here we map each HDF5 ISO related metadata GROUP
//             to its own theme, using a "META_" preface before the
//             literal group name. Performing this mapping is
//             entirely separate from writing out these GROUPS to
//             the HDF5 output file, now typically done only by the
//             GMAO L4 post-processor. For local runs, however,
//             we can opt to include these.
/*
    pObjCounts[12] = 1 ;  // Theme: "Meta_CRID"
    pObjCounts[13] = 1 ;  // Theme: "Meta_Config"
    pObjCounts[14] = 1 ;  // Theme: "Meta_DatasetIdentification"
    pObjCounts[15] = 1 ;  // Theme: "Meta_Extent"
    pObjCounts[16] = 1 ;  // Theme: "Meta_GridsSpatialRepresentation"
    pObjCounts[17] = 1 ;  // Theme: "Meta_ProcessStep"
    pObjCounts[18] = 1 ;  // Theme: "Meta_SeriesIdentification"
*/
void l4c_PkgInit_Metagroups(PkgMgr *pkgMgr_p) {
  NsbObj *nsbObj_p = l4c_GetNsbOverseer();
  char fxName[256];
  char *dsetName_p = NULL;
  char nvpBuf[256];

  strcpy(fxName, __func__);

  PkgTheme *pThme_p = NULL; // convenience ptr

  // themes 12...19 are addressed here

  // GROUP: CRID ---------------------------------
  pThme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 11);
  pkgMgr_SetThemeName(pThme_p, "Metagroup_CRID");
  pkgMgr_SetThemes_ChildName(pThme_p, 0, "CRID");

  // GROUP: Config -------------------------------
  pThme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 12);
  pkgMgr_SetThemeName(pThme_p, "Metagroup_Config");
  pkgMgr_SetThemes_ChildName(pThme_p, 0, "L4_C_ATBD_VERSION");
  pkgMgr_SetThemes_ChildName(pThme_p, 1, "L4_C_ATBD_DATE");
  pkgMgr_SetThemes_ChildName(pThme_p, 2, "L4_C_PSD_EDITION");
  pkgMgr_SetThemes_ChildName(pThme_p, 3, "L4_C_PSD_DATE");
  // CompositeReleaseID value default is "V050010001"
  pkgMgr_SetThemes_ChildName(pThme_p, 4, "L4_C_CompositeReleaseID");
  pkgMgr_SetThemes_ChildName(pThme_p, 5, "L4_C_documentation");
  pkgMgr_SetThemes_ChildName(pThme_p, 6, "L4_C_identifier_product_DOI");
  pkgMgr_SetThemes_ChildName(pThme_p, 7, "L4_C_VersionID");
  pkgMgr_SetThemes_ChildName(pThme_p, 8, "L4_C_algorithmVersionID");
  pkgMgr_SetThemes_ChildName(pThme_p, 9, "L4_C_datasetVersionID");

  // GROUP: DatasetIdentification------------------
  pThme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 13);
  pkgMgr_SetThemeName(pThme_p, "Metagroup_DatasetIdentification");
  pkgMgr_SetThemes_ChildName(pThme_p, 0, "CompositeReleaseID");
  pkgMgr_SetThemes_ChildName(pThme_p, 1, "ECSVersion");
  pkgMgr_SetThemes_ChildName(pThme_p, 2, "QA_Abstract");
  pkgMgr_SetThemes_ChildName(pThme_p, 3, "QA_CreationDate");
  pkgMgr_SetThemes_ChildName(pThme_p, 4, "QA_Filename");
  pkgMgr_SetThemes_ChildName(pThme_p, 5, "SMAP_short_name");
  pkgMgr_SetThemes_ChildName(pThme_p, 6, "UUID");
  // here, this is the dataset version ID
  pkgMgr_SetThemes_ChildName(pThme_p, 7, "VersionID");
  pkgMgr_SetThemes_ChildName(pThme_p, 8, "abstract");
  pkgMgr_SetThemes_ChildName(pThme_p, 9, "characterSet");
  pkgMgr_SetThemes_ChildName(pThme_p, 10, "creationDate");
  pkgMgr_SetThemes_ChildName(pThme_p, 11, "credit");
  // here, this is the granulename
  pkgMgr_SetThemes_ChildName(pThme_p, 12, "filename"); // granule-name
  //
  pkgMgr_SetThemes_ChildName(pThme_p, 13, "language"); // "eng"
  pkgMgr_SetThemes_ChildName(pThme_p, 14, "long_name");
  pkgMgr_SetThemes_ChildName(pThme_p, 15, "originatorOrganizationName");
  pkgMgr_SetThemes_ChildName(pThme_p, 16, "long_name");
  pkgMgr_SetThemes_ChildName(pThme_p, 17, "purpose");
  pkgMgr_SetThemes_ChildName(pThme_p, 18, "short_name");
  // on-going" ("completed" after final reprocessing)
  pkgMgr_SetThemes_ChildName(pThme_p, 19, "status");
  // SPL4CMDL
  pkgMgr_SetThemes_ChildName(pThme_p, 20, "short_name");
  // "geoscientificInformation"
  pkgMgr_SetThemes_ChildName(pThme_p, 21, "topicCategoryCode");

  // GROUP: Extent --------------------------------
  pThme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 14);
  pkgMgr_SetThemeName(pThme_p, "Metagroup_Extent");
  // "Global land excluding inland water and permanent ice."
  pkgMgr_SetThemes_ChildName(pThme_p, 0, "description");
  pkgMgr_SetThemes_ChildName(pThme_p, 1, "eastBoundLongitude");
  pkgMgr_SetThemes_ChildName(pThme_p, 2, "northBoundLatitude");
  pkgMgr_SetThemes_ChildName(pThme_p, 3, "southBoundLatitude");
  pkgMgr_SetThemes_ChildName(pThme_p, 4, "westBoundLongitude");
  pkgMgr_SetThemes_ChildName(pThme_p, 5, "rangeBeginningDateTime");
  pkgMgr_SetThemes_ChildName(pThme_p, 6, "rangeEndingDateTime");

  // GROUP: GridSpatialRepresentation -------------
  pThme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 15);
  pkgMgr_SetThemeName(pThme_p, "Metagroup_GridSpatialRepresentation");
  // nested groups, Latitude and Longitude!!
  // latitude resolution "9km on average"
  pkgMgr_SetThemes_ChildName(pThme_p, 0, "latitude_resolution");
  // longitude: "9km on average"
  pkgMgr_SetThemes_ChildName(pThme_p, 1, "longitude_resolution");
  pkgMgr_SetThemes_ChildName(pThme_p, 2, "northBoundLatitude");
  pkgMgr_SetThemes_ChildName(pThme_p, 3, "southBoundLatitude");
  pkgMgr_SetThemes_ChildName(pThme_p, 4, "westBoundLongitude");
  pkgMgr_SetThemes_ChildName(pThme_p, 5, "rangeBeginningDateTime");
  pkgMgr_SetThemes_ChildName(pThme_p, 6, "rangeEndingDateTime");

  // GROUP: ProcessStep ---------------------------
  pThme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 16);
  pkgMgr_SetThemeName(pThme_p, "Metagroup_ProcessStep");
  pkgMgr_SetThemes_ChildName(pThme_p, 0, "description");
  // TBD, complete the rest of these

  // GROUP: SeriesIdentification ------------------
  pThme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 17);
  pkgMgr_SetThemeName(pThme_p, "Metagroup_SeriesIdentification");
  pkgMgr_SetThemes_ChildName(pThme_p, 0, "description");

  // GROUP: AcquisitionInformation ------------------
  pThme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 18);
  pkgMgr_SetThemeName(pThme_p, "Metagroup_AcquisitionInformation");
  pkgMgr_SetThemes_ChildName(pThme_p, 0, "description");

  // THEME: "L4C_OUTPUT_VARIABLE_PROPERTIES" ----------
  pThme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 19);
  pkgMgr_SetThemeName(pThme_p, "L4C_OUTPUT_VARIABLE_PROPERTIES");
  pkgMgr_SetThemes_ChildName(pThme_p, 0, "description");

  l4c_EventCompletionMsg((char *)__func__);

} // end:: l4c_PkgInit_Metagroups()
