// ----------------------------------------------------------------
// module       : l4c_DAILY_PostProcessor.c (was orig l4c_PostProc9km_func.c)
// purpose      :  tcfPostProc9km_func.c implements 9km aggregation,
//                 1km 1D-sparse vector data to 9km 1D-sparse vector data
// purpose       : Implements core processing functions for tcfPostProc9km.  These functions handle spatial looping
//                 and aggregation from 1km data to 9km and 9km x PFT averages.  Also implements rescaling
//                  logic for QA because QA 1km psuedo-errors must be averaged as floating point data during 
//                 spatial aggregation PRIOR to rescaling.
//  project      : SMAP L4C science development.
//  dependencies : HDF4, HDF5, FRM_API, uuid libraries for runtime architecture
//  revision     : v7.4.0 Updated to support metadata for VIIRS fPAR
//  revision     : v7.2.1-2017-05-24T17:05, also corrected RMSE QA var initialization to 
//                 missing values for the RMSE average as well as RMSE PFT wise QA vars.
//  revision     : v7.2.1-2017-05-24T14:44, the scaling by 100.0 for {EC/emult_mean,...}
//                 previously didn't skip over -9999.0 valued pixels, so this
//                 was changed (line 944) such that -9999 valued pixels are left intact
//                 and NOT scaled. Incorrectly scaling them results in bad -999900.0 valued pixels.
//  revision     : v7.1.1-2016-03-02T09:59, investigate potential "negative RMSE" issue
//                       in either PostProcessor, or mod_mult/mod_func. In progress
//  history      : 2016-02-17, changed module name to l4c_DAILY_PostProcessor.c from
//                 orig l4c_PostProc1km_func.c
//  history      : 2016-01-07T16:40,jmg, implemented RSME implementation on-going
//  history      : 2015-12-04T09:44:00,jmg, implemented RSME implementation on-going
//  history      : 2015-10-15T11:22:00, noted valgrind non-fatal warnings/errors on l4c_SetBits() calls
//                 for the "SOC/soc_mean" if bloc, (around line 1087, now 1110) on "use of 
//                 uninitialized values and "Conditional jump or move depends on
//                 uninitialised value(s)". I haven't yet been able to adequately
//                 interpret these warnings/errors(TBD), since all variables appear to
//                 have been initialized except the receiving variable.
//                 Note that carbon_model_bitflag is allocated at l4c_PostProc9km_func.c:1413:
//                   stepVar_p.carbon_model_bitflag = (uint16*)memCalloc(LDOM9KM, sizeof(uint16));
//                 using count, LDOM9KM 1664040ul
//  history      : 2014-11-05T18:37, added realtime guard on NaN|Inf for std_9km variables
//                 by using special (isfinite(x) and fpclassify() macros to trap NaN's 
//                 when they occur. Note that fenv.h also contains tests but these
//                 cause non-recoverable (core dump) when a NaN occurs.
//  history      : 2014-10-29T16:00:00, <LAJ> added CompletenessOmission,DomainConsistency sums and
//                 generalized state variable limit checking.  
//  revised      : 2014-10-28T16:48:00, defined CompletenessOmission,DomainConsistency
//                 using qa_count 9km array, which is only complete AFTER the forward step
//                 loop is completely done.
//  revised      : 2014-10-21T16:35 changed to rescale (3) EC 1D-sparse vector 
//                 Emult,Tmult,Wmult 1D vector arrays from {0.0<= a <=1.0} to {0.0< a < 100.0}
//                 prior to post processing
//  history      : 2014-10-16T15:18:00, added verification check of QA Score (bits 8,9,10,11)
//                 to double check range of score values inserted into these bits.
//  history      : 2014-07-23T120:00:00, modified module to do aggregation of 1km to 9km within forward step, laj
//  history      : 2014-04-22T17:13:00, begin v5.0.1+ era fixes,enhancements
//  history      : 2014-03-13T18:53:00, corrected cast of qcScore to uint8, from uint16
//  history      : 2014-03-12T11:37:00, final stage integration
//  history      : 2014-02-15T14:35:00-07:00,jmg
//  history      : Initial reformating -- laj 09/10/13, Lucas A. Jones; lucas@ntsg.umt.edu
//  author       : joe glassy, Lupine Logic Inc.
//  notes        : The orderf of all HDF5 output dataset (properties) is defined by the 
//                 order they appear in their PkgMgr theme "L4C_OUTPUT" for primary (.h5) 
//                 variables and PkgMgr theme "L4C_QA_OUTPUT" for QA (.qa) variables. These "theme"
//                 definitions appear in source module l4c_DAILY_PkgMgrInit_Themes.c
//                 See functions : l4c_PkgInit_L4C_OUTPUTS() 

//  notes        : RECAP OF OVERALL FLOW: {AGGREGATE, INFLATE, WRITE-HDF5} 
//                 a) submit given 1km 1D-sparse vec to POST-PROCESSOR, which aggregates
//                    the variable to a 9KM, 1D-sparse vector form.
//                 b) After all PFT-wise output variables, and singular output variables have
//                 undergone AGGREGATION, then INFLATE each to a 2D full grid form, 
//                 and finally, WRITE to HDF5 the inflated 2D variable(s) as 
//                 2D full grid raster.
//                 
//  notes        : Method: Initially, submit 1km 1D-land variables serially, 
//                 one at a time, until a list-drive (loop) method can be perfected
// BY PFT:
//    varFlt32_Obj_p->pft_9km[p].rank       {1..8} of these
//    varFlt32_Obj_p->pft_9km[p].cnt        {1..8} of these
//    varFlt32_Obj_p->pft_9km[p].avg        {1..8} of these
// SINGULARS :
//    varFlt32_Obj_p->rank_9km
//    varFlt32_Obj_p->cnt_9km
//    varFlt32_Obj_p->avg_9km ..singulars for nee_mean,gpp_mean,rh_mean,soc_mean
//    varFlt32_Obj_p->std_9km ..singulars for nee_std_dev,gpp_std_dev,rh_std_dev,soc_std_dev
//
//  PkgMgr functions we use here:
//  pkgMgr_GetThemes_ChildOwnerByName(pThme_p,"NEE/nee_mean") fetches its HDF5 "file_id"
//  pkgMgr_GetThemes_ChildIdByName(pThme_p,"NEE/nee_mean") fetches its HDF5 "dataset_id"
//  pkgMgr_GetThemes_ChildDatatypeByName(pThme_p,"NEE/nee_mean") fetches its HDF4 datatype
//

#include "l4c_DAILY.h"
#include "l4c_DAILY_bplut.h"
#include "l4c_DAILY_prot.h"


// Instantiate a counter object tracking invalid states
static PostInvalid pInv ;

// local devel diag structures to track frequency counts of
// the PFT dom values assigned to bits {4,5,6,7}
static PostProcFreqVerifyMetrics pfvmPftDom ;

// local devel diag structure to track frequency counts of
// the QaScore {0,1,2,3} assigned to bits {8,9,10,11}
static PostProcFreqVerifyMetrics pfvmQaScore;


// ---------------------------------------------------------------------
//  module       : tcfPostProc_func.c
//  purpose      : Core processing functions for tcfPostProc9km.  
//   These functions handle spatial looping
//   and aggregation from 1km data to 9km and 9km x PFT averages.  
//   Also implements rescaling logic for QA because QA 1km psuedo-errors
//   must be averaged as floating point data during spatial aggregation
//   PRIOR to rescaling.
//
//  project      : SMAP L4C science development.
//  dependencies : None
//  revision     : laj 02/12/14
//  original author : Lucas A. Jones; lucas@ntsg.umt.edu
//  notes        :

//   
//  PRIMARY OUTPUT VARIABLES ARE: 
//  "ecoStateArrayObj_p->nee.dom_1km"
//  "ecoStateArrayObj_p->gpp.dom_1km"
//  "ecoStateArrayObj_p->rhtot.dom_1km"
//  "ecoStateArrayObj_p->Ctot.dom_1km"
//  "ecoStateArrayObj_p->Emult.dom_1km
//  "ecoStateArrayObj_p->Tmult.dom_1km
//  "ecoStateArrayObj_p->Wmult.dom_1km
//  "ecoStateArrayObj_p->ecoStateArrayQaObj.nee_error
// ---------------------------------------------------------------------


// ---------------------------------------------------------------------
// function: l4c_GetPostProcFreqVerify_PftDom(void)
// purpose : accessor for PftDom bitfield freq count tracking
// returns : ptr to master instance of pfvmPftDom tracker
// notes   : This is one of many software instrumentation mechanisms
//           for tracking suspicious ruutime events.
PostProcFreqVerifyMetrics *l4c_GetPostProcFreqVerify_PftDom(void)
{
  return &pfvmPftDom;   
} // end:: l4c_GetPostProcFreqVerify_PftDom()


// ---------------------------------------------------------------------
// function: l4c_GetPostProcFreqVerify_QaScore(void)
// purpose: accessor for QaScore bitfield freq count tracking
// returns: ptr to master instance of pfvmQaScore tracker
// notes   : This is one of many software instrumentation mechanisms
//           for tracking suspicious ruutime events.
// author  : Joe Glassy
PostProcFreqVerifyMetrics *l4c_GetPostProcFreqVerify_QaScore(void)
{
  return &pfvmQaScore;  
} // end:: l4c_GetPostProcFreqVerify_QaScore()


// ---------------------------------------------------------------------
// function: l4c_GetInvalidEventCounts(void)
// returns : PostInvalid *, pointer to local static structure of event counts
//           that track occurrences of suspicious runtime events
// purpose : return pointer to local static package of counts of potentially 
//           invalid events
// notes   : This is one of many software instrumentation mechanisms
//           for tracking suspicious ruutime events. Should have been
//           named l4c_GetInvalidEventCounts(), now is renamed.
// author  : Joe Glassy
// OLD NAME PostInvalid *l4c_ReportInvalidNee(void)
PostInvalid *l4c_GetInvalidEventCounts(void)
{
    // Return pointer to entire structure containing many different
    // runtime (suspicious or interesting) event counts
    return &pInv;
    
} // end:: l4c_GetInvalidEventCounts()


// ---------------------------------------------------------------------
// function : int l4c_PostProc_PrepQaCountObsStats()
// purpose  : Prep qa_count arrays, for observe min,max,mean calc
// inputs   : 
// returns  : status, 0
// notes    :
// revised  : 2016-03-04T11:47:00-07:00
// author   : Joe Glassy
int l4c_PostProc_PrepQaCountObsStats(hid_t dset_id,char *dsetName_p,
                  uint8 *uint8_9kmAry_p,float32 *wrk9kmFltCount_p)
{
  SesCtl  *sesCtl_p = l4c_GetSesCtl();         
  int pix = 0 ;
  int nSparsePixelCount = LDOM9KM;
  
  if(sesCtl_p->enableWriteObsMinMax)
  {
     // <JMG> 2016-03-04, transfer uint8 STATE_p->pft_9km[p].cnt 
     // to float32 array for statistics work
     for(pix=0;pix< nSparsePixelCount; pix++) 
     {
       // Translate BYTE_FILL_VAL to FLOAT_FILL_VAL locally for 
       // consistent handling in stats summary
        if(uint8_9kmAry_p[pix] == BYTE_FILL_VALUE)
           wrk9kmFltCount_p[pix] = FLOAT_FILL_VALUE;
        else
           wrk9kmFltCount_p[pix] = uint8_9kmAry_p[pix];
     }
     
     // Execute qa_count summary stats on temporary float32 array
     l4c_PostProc9km_WriteObsMinMaxMean(dset_id,
        dsetName_p, nSparsePixelCount, FLOAT_FILL_VALUE,
        wrk9kmFltCount_p);
        
  } // end:: block if enabled obs output
    
 return 0 ;    
 
} // end:: l4c_PostProc_PrepQaCountObsStats()


// ---------------------------------------------------------------------
// function: l4c_PostProc9km_WriteObsMinMaxMean()
// inputs  : diagLabel (like: "Verify_nee_qa_nee_rmse_mean")
//           nCount -- no. of pixels in float32 array 
//           float32 *avg9km_p -- array of float32 to analyze
// revised : 2016-03-02T10:57:00, added new.
// returns : minimum valid diagnostic value from pixel array 
// purpose : Calculate observed min,max,mean value for a float32 dataset and
//           write this out as a dataset attribute
// author  : Joe Glassy
// notes   : LIMITATION the missingVal MUST be low like -9999.0 in order for
//           this logic to work.
int l4c_PostProc9km_WriteObsMinMaxMean(hid_t dset_id,char *diagLabel_p,const int nCount,const float32 missingVal,float32 *avg9km_p)
{
 SesCtl  *sesCtl_p = l4c_GetSesCtl();         
 float32 diagPix = 0.0 ;
 float32 diagSum = 0.0 ;
 float32 diagCount=0.0 ;
 float32 diagMinVal = missingVal; // init for sentinal use
 float32 diagMaxVal = missingVal; // init for sentinal use
 float32 diagMeanVal=0.0;
 int d = 0 ;
 int hitFirst =1 ;
 char fxName[256] = {0};
 hid_t ret =0 ;
 
    // register fx name
    strcpy(fxName,__func__);
    
    // iterate through all test array pixels to analyze, find min,max,mean
   for(d=0; d < nCount; d++)
   {
      diagPix = avg9km_p[d];
      if( diagPix > missingVal)
      {
          diagSum += diagPix;
          diagCount += 1.0 ;
          // accumulate min,max. Initialize ONCE ONLY via hitFirst state
          if(hitFirst == 1)
          {   // INITIALIZE on first valid pixel
              diagMinVal = diagPix;
              diagMaxVal = diagPix;
              hitFirst = 0 ;
          } else {
              if(diagPix < diagMinVal) diagMinVal = diagPix;
              if(diagPix > diagMaxVal) diagMaxVal = diagPix;
          }
      } // end:: block on encountering first valid pixel
   } // end:: d loop through all pixels
   
    if(diagCount > 0.0)
    {
       diagMeanVal = diagSum / diagCount;
    }
    xLogMsg("@A In(%s) OVERVIEW %s Count %.0f Min %.3f Max %.3f Mean %.3f~\n",
         fxName, diagLabel_p, diagCount, diagMinVal, diagMaxVal, diagMeanVal);

    // write out observed min,max, and mean for each dataset         
    if( sesCtl_p->enableWriteObsMinMax )
    {
     ret = l4c_Hdf5_CreateAttribute(dset_id, IsFloat32, ".", "observed_min" ,&diagMinVal);
     ret = l4c_Hdf5_CreateAttribute(dset_id, IsFloat32, ".", "observed_max" ,&diagMaxVal);
     ret = l4c_Hdf5_CreateAttribute(dset_id, IsFloat32, ".", "observed_mean",&diagMeanVal);
    }
         
 // return minimum valid value
 return diagMinVal;
} // end:: l4c_PostProc9km_WriteObsMinMaxMean()


// ---------------------------------------------------------------------
// function : l4c_PostProcess_PrimaryVars()
// inputs   : char *varName_p, as:"nee","gpp","rh","soc", "nee_rmse"
//            VarFloat32Adt *STATE_p -- as needed for nee,gpp, ...rmse
//            void *DEST_p -- ptr to shared,re-usable array
// returns  : int , status.
// revision : 2016-01-07T16:31, modified to support RMSE processing of
//            (1) nee_rmse_mean, and (8) PFT wise 
//            nee_rmse_pft{1..8}_mean float32 9km variables
// historyn : 2014-09-04T12:39, refactored to eliminate fragile
//            order-dependence of the original startIndex scheme.
//
// notes    : pattern for these (4) is: "nee","gpp","rh","soc"
//            defined as: char varNameList[4][8] = {"nee","gpp","rh","soc"} ;
//            Process NEE, GPP, RH, SOC variables.
//            --Each of these has a 'mean' and 'stddev' and (8) PFT wise entries
//            --"STATE" is really an output classifier and NOT an
//               HDF5 group, since there is no "STATE" group
// notes     : To implement RMSE float32's, we now pass in "nee_qa" variable, 
//             which ultimately reside in QA/nee_rmse_mean and its PFT wise variants
// ----------------------------------------------------------------
int l4c_PostProcess_PrimaryVars(const char *varName_p,VarFloat32Adt *STATE_p,void *DEST_p)
{
 SesCtl    *sesCtl_p = l4c_GetSesCtl();
 PkgMgr    *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
 PkgTheme  *pThme_p  = NULL;
 AncImagePkgAdt *ancImagePkgObj_p   = l4c_GetAncImagePkgAdt();  
 
 // <JMG> fixed out_file_id since sep. QA file deprecated
 hid_t out_file_id = sesCtl_p->l4cOutput_file_id;
 hid_t output_H5T_datatype = H5T_NATIVE_FLOAT ;
 hid_t dset_id = 0 ; // HDF5 dataset id 
 
 double elapSec    = 0.0;
 
 // <JMG> add diagnostic monitor on nee_rmse_mean to verify no negative values
 int     d=0; // diagnostic pixel index
 float32 diagPix = 0.0 ;
 float32 diagSum = 0.0 ;
 float32 diagCount=0.0 ;
 float32 diagMinVal=0.0;
 float32 diagMaxVal=0.0;
 float32 diagMeanVal=0.0;
 
 char fxName[256] ={0} ;
 char outputThemeName[32] = {0} ; // now static as "L4C_OUTPUTS"
 char hGroupLabel[8] = {0} ;  // {"NEE"|"GPP"|"RH"|"SOC|QA"} inferred from varName_p
 // <JMG> 2021-10-25T12:45, modified length to 256 to resolve buffer overrun
 char fullyQualifiedVarname[256] = {0} ; // e.g. "NEE/nee_mean", "GPP/gpp_std_dev" etc
 // BASE SUFFIX :"_mean" for "gpp","nee","rh","soc",    or, "_rmse" for "nee_rmse"
 char baseSuffixLabel[16] = {0} ; // _mean for GPP,NEE etc, or _rmse for QA use-case context
 char stddevSuffixLabel[16] = {0}; // "stddev"
 
 char dsetName[128] = {0} ;
 int  p=0;

 // <JMG> fixed: all in this sci var category share the same 9km extent
 // size_t nElem = sizeof(float32) * EGv2_GLOBAL_9KM_PIXELS;
 
 SplandRefAdt SPLAND  ;
 
 // strongly typed pointers 
 float32  *destGrid_float32_p =NULL;
 
    strcpy(fxName,(char *)__func__);

    // ----------------------------------------------------------
    // STEP 1: Attach pre-existing {row,col} LUTs to local SPLAND
    SPLAND.row = ancImagePkgObj_p->imgM09land_row_uint16_p;
    SPLAND.col = ancImagePkgObj_p->imgM09land_col_uint16_p;

    // Set context into requested THEME, e.g.L4C_OUTPUTS which is now
    // fixed since there is just a single theme 
    strcpy(outputThemeName, "L4C_OUTPUTS");
    pThme_p = pkgMgr_GetThemeByName(pkgMgr_p,outputThemeName);

    // Assign dataset name suffix based on context:{"gpp","nee","rh","soc"} or "_rmse"
    // set defaults for majority cases
    strcpy(hGroupLabel,varName_p);
    // defaults to: "gpp","nee","rh","soc", "nee_rmse" (for nee_rmse_mean etc)
    strcpy(dsetName,   varName_p);
    // Intentionally omit the leading underscore here...
    strcpy(baseSuffixLabel  ,"mean");
    strcpy(stddevSuffixLabel,"std_dev");
    
    // <JMG> 2014-11-14T12:37, pre-work added to soon enable RMSE 
    // outputs when ready. Next, allow for use-case overrides,
    // to handle RMSE as well as "nee","gpp" etc
    
    // Establish all base names, suffixes and the HDF5 group label for RMSE case
    if( strstr(varName_p,"rmse") != NULL)
    {
        strcpy(dsetName,"nee_rmse"); 
        strcpy(baseSuffixLabel,"mean"); // goal, to result in "nee_rmse_mean"
        strcpy(stddevSuffixLabel,""); //empty, trigger suppression of STDDEV handling
        // <JMG> 2014-11-14T12:22, change to "QA" or other more appropriate GROUP for RMSE
        strcpy(hGroupLabel,"QA");
        xLogMsg("@A In(%s) RMSE processed ~ \n",fxName);
    }
    // In all cases, transform: "nee" to "NEE", "gpp" to "GPP", "qa" to "QA" etc
    frm_strToUpper(hGroupLabel); 

    // Process NEE, GPP, RH, SOC variables (or QA/xxx_rmse) ------------
    // Each of these has a 'mean' and 'stddev' and (8) PFT wise entries
    // note that "STATE" is really an output classifier and NOT at all a
    // HDF5 group, since there is no "STATE" group

    // Bind generic re-usable array (DEST_p) to a strongly-typed (float32) entity
    destGrid_float32_p = (float32 *)DEST_p;
    // consider if a memset(destGrid_float32_p,'\0',nElem) would be more robust

    // STEP 1: process statistical summaries (mean,std-deviation) for 9-km 
    // pattern: iterate through list of var base names: "nee","gpp","rh","soc" 

    // ------------------------- PROCESS "MEAN" instance ---------------    
    // Construct: NEE/nee_mean, GPP/gpp_mean, or for QA:
    //  QA/nee_rmse, QA/nee_rmse_pft{1..8}_mean
    // 
    sprintf(fullyQualifiedVarname, "%s/%s_%s",hGroupLabel, dsetName, baseSuffixLabel);

    xLogMsg("@A In(%s) INFLATE-and-WRITE(%s)~\n",fxName,fullyQualifiedVarname);
    
    l4c_inflate_init_9km(destGrid_float32_p, DFNT_FLOAT32);
    l4c_inflate_9km(STATE_p->avg_9km, destGrid_float32_p, DFNT_FLOAT32);
    
    // STEP 2) WRITE: "NEE/XXXX_mean  via single "avg_9km" variable to hdf5 here
    // or, if RMSE, WRITE QA/nee_rmse_mean via single "avg_9km" variable.
    
    // fetch this dataset's ID here
    dset_id = pkgMgr_GetThemes_ChildIdByName(pThme_p, fullyQualifiedVarname);

    if(sesCtl_p->enableWriteObsMinMax)
    {
       l4c_PostProc9km_WriteObsMinMaxMean(dset_id,
          fullyQualifiedVarname, LDOM9KM, FLOAT_FILL_VALUE, STATE_p->avg_9km);
    }
    
    l4c_Hdf5_WriteDataset(fxName, out_file_id, fullyQualifiedVarname,
        dset_id, output_H5T_datatype, destGrid_float32_p);

    // ------------------------- PROCESS "STDDEV" instance -------------
    // Next, repeat for STANDARD DEVIATION "('nee'|'gpp'|'rh'|'soc')_std_dev" variable
    // noting the test here is used to screen out use cases where STDDEV 
    // is not applicable, such as the RMSE workflow
    if( strlen(stddevSuffixLabel) >= 1)
    {
        // construct: NEE/nee_std_dev, GPP/gpp_std_dev, from "nee","gpp" etc
        sprintf(fullyQualifiedVarname, "%s/%s_%s",hGroupLabel, dsetName, stddevSuffixLabel);
        
        // xLogMsg("@A In(%s) INFLATE-and-WRITE(%s)~\n",fxName,dsetName);
        l4c_inflate_init_9km(destGrid_float32_p, DFNT_FLOAT32);
        l4c_inflate_9km(STATE_p->std_9km, destGrid_float32_p, DFNT_FLOAT32);

          // STEP 3) WRITE next single xxx_std_dev, to hdf5
          dset_id = pkgMgr_GetThemes_ChildIdByName(pThme_p, fullyQualifiedVarname);

          if(sesCtl_p->enableWriteObsMinMax)
          {
            l4c_PostProc9km_WriteObsMinMaxMean(dset_id,
                fullyQualifiedVarname, LDOM9KM, FLOAT_FILL_VALUE, STATE_p->std_9km);
          }
          
          l4c_Hdf5_WriteDataset(fxName, out_file_id, fullyQualifiedVarname, // dataset's name
            dset_id, output_H5T_datatype, destGrid_float32_p);
            
    } // end:: block if this variable includes a {varname}_std_dev component

    
    // STEP 4: Iterate through NPFT=8 wise variables for "nee"|"gpp"|"rh"|"soc" OR _rmse

    // Produce PFT-wise mean summaries, for 9km x PFT:  Iterate through (8) PFTs for varName_p.
    // construct: names like: nee_pft{1..8}_mean, gpp_mean_pft{1..8}_mean, etc.
    // OR, for RMSE, does equivalent
    for(p=0; p < NPFT ; p++)
    {
        // construct: fully qualified dataset name like "NEE/nee_pftNN_mean" or "QA/nee_pftNN_rmse"
        sprintf(fullyQualifiedVarname, "%s/%s_pft%d_%s",hGroupLabel, dsetName, p+1,baseSuffixLabel);
        // xLogMsg("@A In(%s) INFLATE-and-WRITE(%s)~\n",fxName, dsetName);
            
        // INFLATE the current variable from 1D-sparse vector to 2D full grid
        l4c_inflate_init_9km(destGrid_float32_p, DFNT_FLOAT32);
        l4c_inflate_9km(STATE_p->pft_9km[p].avg, destGrid_float32_p, DFNT_FLOAT32);
            
        // WRITE current variable (at dset_id) out to the L4_C HDF5 granule 
        dset_id = pkgMgr_GetThemes_ChildIdByName(pThme_p, fullyQualifiedVarname);

        if(sesCtl_p->enableWriteObsMinMax)
        {
            l4c_PostProc9km_WriteObsMinMaxMean(dset_id,
                fullyQualifiedVarname, LDOM9KM, 
                FLOAT_FILL_VALUE, STATE_p->pft_9km[p].avg);
        }
        
        l4c_Hdf5_WriteDataset(fxName, out_file_id, fullyQualifiedVarname,
            dset_id, output_H5T_datatype, destGrid_float32_p);   
        
    } // end:: p loop through PFT

   elapSec = l4c_EventCompletionMsg(fxName);

 return 0;
} // end:: l4c_PostProcess_PrimaryVars()



// ---------------------------------------------------------------------
// function : l4c_PostProcess_EC_Vars()
// inputs   : char *varName_p, as:"emult_mean","tmult_mean","wmult_mean","frozen_area"
//            VarFloat32Adt *STATE_p as: 
//            void *DEST_p -- ptr to shared,re-usable array
// returns  : int , status.
// revision : 2014-09-04T10:55:00, refactored to eliminate fragile
//            order-dependence of the original startIndex scheme.
// notes    : pattern for these (4) is: "Emult","Tmult","Wmult"
//   define as: char varNameList[4][8] = {"Emult","Tmult","Wmult"} ;
// ----------------------------------------------------------------
int l4c_PostProcess_EC_Vars(const char *varName_p,VarFloat32Adt *STATE_p,void *DEST_p)
{
 SesCtl    *sesCtl_p = l4c_GetSesCtl();
 PkgMgr    *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
 PkgTheme  *pThme_p  = NULL;
 AncImagePkgAdt *ancImagePkgObj_p   = l4c_GetAncImagePkgAdt();  
 PostProcMetrics postMetrics ; // post-proc metrics object
 // <JMG> fixed out_file_id since sep. QA file deprecated
 hid_t out_file_id = sesCtl_p->l4cOutput_file_id;
 hid_t output_H5T_datatype = H5T_NATIVE_FLOAT ;
 hid_t dset_id = 0 ;
 
 size_t nQualfied = 0; // count of non fill value cells
 size_t ithPix = 0;
 float32 sciPix   = 0.0 ; // pixel DN for post-proc transform
 float32 obs_min  = 0.0,   // observed minimum
         obs_max  = 0.0,   // observed maximum
         obs_range= 0.0; // observed range
 
 char fxName[256] ={0} ;
 char outputThemeName[64] = {0} ;
 char hGroupLabel[8] = {0} ; // "EC"
 char fullyQualifiedVarname[128] = {0} ; // e.g. "EC/emult, EC/tmult, EC/tmult
 char dsetName[128] = {0} ;
 
 double elapSec    = 0.0;
 
 SplandRefAdt SPLAND  ;
 
 // strongly typed pointers specific to EC variables
 float32  *destGrid_float32_p =NULL;

    strcpy(fxName,(char *)__func__);

    // ----------------------------------------------------------
    // STEP 1: Attach pre-existing {row,col} LUTs to local SPLAND
    SPLAND.row = ancImagePkgObj_p->imgM09land_row_uint16_p;
    SPLAND.col = ancImagePkgObj_p->imgM09land_col_uint16_p;

    // Set context into requested THEME, e.g.L4C_OUTPUTS which is now
    // fixed since there is just a single theme 
    strcpy(outputThemeName, "L4C_OUTPUTS");
    pThme_p = pkgMgr_GetThemeByName(pkgMgr_p,outputThemeName);

    // pattern: iterate through list of var base names: "nee","gpp","rh","soc" 
    // step 1 : derive upper case group, assemble into a fully qualified datasetname
    
    // Step 1: fix the group as "EC" here
    strcpy(hGroupLabel,"EC");

    // Process EC variables------------------------
    // Bind generic re-usable array (DEST_p) to a strongly-typed (float32) entity
    destGrid_float32_p = (float32 *)DEST_p;
    // consider is a memset(destGrid_float32_p,'\0',nElem) would be more robustt

    // STEP 1: process statistical summaries (mean,std-deviation) for 9-km 
    // pattern: iterate through list of var base names: "nee","gpp","rh","soc" 
    // step 1 : derive upper case group, assemble into a fully qualified datasetname
    // construct: EC/{varname} where varname DOES include the _av suffix
    sprintf(fullyQualifiedVarname, "%s/%s",hGroupLabel,varName_p);
    strcpy(dsetName, fullyQualifiedVarname); // for backward compatibility

    // only derive (L4C_GRANULE_METRICS = enable) if enabled for
    // calculating and reporting actual observed min,max,mean, etc
    if(sesCtl_p->enableGranuleMetrics )
    {
        // remember this only affects postMetrics -- upscale from 0..1 to {0.0<pix<100.}
        postMetrics.optionScaleFactor = 100.0; 
        l4c_SummarizePixels(&postMetrics, dsetName,
            pkgMgr_GetThemes_ChildIdByName(pThme_p, dsetName),
            EGv2_GLOBAL_9KM_PIXELS, (float32)FLOAT_FILL_VALUE, STATE_p->avg_9km);
    }

    // xLogMsg("@A In(%s) INFLATE-and-WRITE(%s)~\n",fxName,dsetName);
    l4c_inflate_init_9km(destGrid_float32_p, DFNT_FLOAT32);
    l4c_inflate_9km(STATE_p->avg_9km, destGrid_float32_p, DFNT_FLOAT32);
    
    dset_id = pkgMgr_GetThemes_ChildIdByName(pThme_p, fullyQualifiedVarname);
    
    if(sesCtl_p->enableWriteObsMinMax)
    {
        l4c_PostProc9km_WriteObsMinMaxMean(dset_id,
            fullyQualifiedVarname, LDOM9KM,
            FLOAT_FILL_VALUE, STATE_p->avg_9km);
    }
    // STEP 2) WRITE: "EC/{ec-variable}  into single "avg_9km" variable to hdf5 here
    l4c_Hdf5_WriteDataset(fxName, out_file_id, dsetName,
        dset_id,  output_H5T_datatype, destGrid_float32_p);

    // STEP 3) Add several empirical "observed" metrics as dataset attributes
   elapSec = l4c_EventCompletionMsg(fxName);

 return 0;
} // end:: l4c_PostProcess_EC_Vars()

// ---------------------------------------------------------------------
// function : l4c_PostProcess_QA()
// inputs   : char *varName_p, 
//            VarFloat32Adt *STATE_p as: nee
//            void *DEST_p -- ptr to shared,re-usable array
// returns  : int , status.
// revision : 2014-09-04T10:55:00, refactored to eliminate fragile
//            order-dependence of the original startIndex scheme.
// notes    : The QA series variables include (2) PFT wise sets of (8) each
//            and (2) "singletons" as:
// Author   : Lucas Jones
/*  
    "QA/carbon_model_qual_pft1" 
    "QA/carbon_model_qual_pft2" 
    "QA/carbon_model_qual_pft3" 
    "QA/carbon_model_qual_pft4" 
    "QA/carbon_model_qual_pft5" 
    "QA/carbon_model_qual_pft6" 
    "QA/carbon_model_qual_pft7" 
    "QA/carbon_model_qual_pft8" 

    "QA/qa_count_pft1" 
    "QA/qa_count_pft2" 
    "QA/qa_count_pft3" 
    "QA/qa_count_pft4" 
    "QA/qa_count_pft5" 
    "QA/qa_count_pft6" 
    "QA/qa_count_pft7" 
    "QA/qa_count_pft8" 

    TWO SINGLETONS are: 
    "QA/qa_count");
    "QA/carbon_model_bitflag"
*/
// ---------------------------------------------------------------------
int l4c_PostProcess_QA(VarFloat32Adt *STATE_p,void *DEST_p)
{
 SesCtl    *sesCtl_p = l4c_GetSesCtl();
 PkgMgr    *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
 PkgTheme  *pThme_p  = NULL;
 AncImagePkgAdt *ancImagePkgObj_p   = l4c_GetAncImagePkgAdt();  
 
 // <JMG> fixed out_file_id since sep. QA file deprecated
 hid_t out_file_id = sesCtl_p->l4cOutput_file_id;
 hid_t output_H5T_datatype = H5T_NATIVE_UINT8 ;
 hid_t dset_id = 0 ; // <JMG> 2016-03-04T11:08 added
 
 char fxName[256] ={0} ;
 char varName_p[64] = {0} ;
 char outputThemeName[64] = {0} ;
 char hGroupLabel[8] = {0} ; // {"QA"}
 // QA/qa_count_pft{1..8}, QA/carbon_model_qual_pft{1..8}
 // and (2) singletons, QA/qa_count and QA/carbon_model_bitflag
 char fullyQualifiedVarname[128] = {0} ; 
 char dsetName[128] = {0} ;
 double elapSec    = 0.0;
 float32 *wrk9kmFltCount_p = NULL;
 int pix=0,p=0;

 // <JMG> all in this sci var category share the same 9km extent
 // careful: some are uint8 and some are uint16 here
 size_t nElem = 0;
 
 SplandRefAdt SPLAND  ;
 
 // strongly typed pointers specific to QA variables
 uint16   *destGrid_uint16_p  =NULL;
 uint8    *destGrid_uint8_p   =NULL;

    strcpy(fxName,(char *)__func__);

    // ----------------------------------------------------------
    // STEP 1: Attach pre-existing {row,col} LUTs to local SPLAND
    SPLAND.row = ancImagePkgObj_p->imgM09land_row_uint16_p;
    SPLAND.col = ancImagePkgObj_p->imgM09land_col_uint16_p;

    // <JMG> 2016-03-04T11:12, allocate temporary float32 LDOM9KM array
    wrk9kmFltCount_p = memCalloc(sizeof(float32),LDOM9KM);

    // Set context into requested THEME, e.g.L4C_OUTPUTS which is now
    // fixed since there is just a single theme 
    strcpy(outputThemeName, "L4C_OUTPUTS");
    pThme_p = pkgMgr_GetThemeByName(pkgMgr_p,outputThemeName);

    // for QA, fix this HDF5 GROUP asn a constant label
    strcpy(hGroupLabel,"QA");
    
    // Bind generic re-usable array (DEST_p) to a
    // strongly-typed (uint8) entity
    // Clear this re-usable 2D 9km container for re-bind
    nElem = sizeof(uint8) * EGv2_GLOBAL_9KM_PIXELS;
    memset(DEST_p, '\0', nElem);

    // Bind uint8 data-type for generic 2D grid array
    destGrid_uint8_p = (uint8 *)DEST_p;

    // PHASE 1: ITERATE through QA variables, starting with "carbon_model_qual"
#ifdef CARBON_MODEL_QUAL_ENABLED
    // set a basename: carbon_model_qual
    strcpy(varName_p,"carbon_model_qual");

    // <JMG> 2016-02-10T16:09, Officially deprecated output of all (8) 
    // carbon_model_qual_pft{1..8} datasets since RMSE replaces these.
    // {carbon_model_qual_pft1,carbon_model_qual_pft2, ... carbon_model_qual_pft8}
    for(p=0; p < NPFT; p++)
    {
        // construct a name like: QA/carbon_model_qual_pft {1..8}
        sprintf(fullyQualifiedVarname, "%s/%s_pft%d",hGroupLabel,varName_p,p+1);
        strcpy(dsetName, fullyQualifiedVarname); // for backward compatibility
        // first, transform from 1D-sparse vector to 2D full grid
        l4c_inflate_init_9km(destGrid_uint8_p, DFNT_UINT8);
        l4c_inflate_9km(STATE_p->pft_9km[p].rank,
            destGrid_uint8_p, DFNT_UINT8);
        // WRITE the NPFT wise QA variables out to HDF5
        dset_id = pkgMgr_GetThemes_ChildIdByName(pThme_p, dsetName);
        l4c_Hdf5_WriteDataset(fxName, out_file_id, dsetName,
            dset_id, output_H5T_datatype, destGrid_uint8_p); 
    } // end:: p loop through (8) PFT classes for QA/carbon_model_qual_pft{1..8}
#endif

    // PHASE 2: ITERATE through (8) QA vars: qa_count_pft{1..8}
    strcpy(varName_p,"qa_count");
    for(p =0; p < NPFT ; p++)
    {
        // construct a name like: QA/qa_count_pft {1..8}
        sprintf(fullyQualifiedVarname, "%s/%s_pft%d",hGroupLabel,varName_p,p+1);
        strcpy(dsetName, fullyQualifiedVarname); // for backward compatibility
        // xLogMsg("@A In(%s) INFLATE-and-WRITE(%s)~\n",fxName, dsetName);

        // INFLATE AND WRITE FOR qa_count_pft{1..8} SERIES
        // from 1D-sparse vector to 2D full grid
        l4c_inflate_init_9km(destGrid_uint8_p , DFNT_UINT8);

        l4c_inflate_9km(STATE_p->pft_9km[p].cnt,
                            destGrid_uint8_p , DFNT_UINT8);

        // WRITE qa_count_pft{1..8} to HDF5 output file 
        dset_id = pkgMgr_GetThemes_ChildIdByName(pThme_p, dsetName);

        // Prep qa_count, then call l4c_PostProc9km_WriteObsMinMaxMean() in here
        l4c_PostProc_PrepQaCountObsStats(dset_id,dsetName,
             STATE_p->pft_9km[p].cnt, wrk9kmFltCount_p);

        l4c_Hdf5_WriteDataset(fxName, out_file_id, dsetName,
            dset_id, output_H5T_datatype, destGrid_uint8_p); 

    } // end:: p loop through (8) PFT classes for QA/qa_count_pft{1..8}

    // PHASE 3: Post process (2) singletons: QA/qa_count and QA/carbon_model_bitflags
    strcpy(varName_p, "qa_count");
    // constuct: "QA/qa_count"
    sprintf(fullyQualifiedVarname, "%s/%s",hGroupLabel,varName_p);
    strcpy(dsetName, fullyQualifiedVarname); // for backward compatibility
    // xLogMsg("@A In(%s) INFLATE-and-WRITE(%s)~\n",fxName, dsetName);

    l4c_inflate_init_9km(destGrid_uint8_p, DFNT_UINT8);

    l4c_inflate_9km(STATE_p->cnt_9km, destGrid_uint8_p, DFNT_UINT8);
    
    // WRITE to HDF5 output file 
    dset_id = pkgMgr_GetThemes_ChildIdByName(pThme_p, dsetName);    
    
    // <JMG> 2016-03-04T11:58 Prep qa_count, then 
    //  call l4c_PostProc9km_WriteObsMinMaxMean() in here
    //  to write dataset attributes: observed_min,max,mean
    l4c_PostProc_PrepQaCountObsStats(dset_id,dsetName,
             STATE_p->cnt_9km, wrk9kmFltCount_p);
    
    l4c_Hdf5_WriteDataset(fxName, out_file_id, dsetName,
        dset_id, output_H5T_datatype, destGrid_uint8_p);

    // STARTING INFLATE AND WRITE FOR BITFLAG SERIES
    // QA/bitflag
    strcpy(varName_p, "carbon_model_bitflag");
    // constuct: "QA/carbon_model_bitflag"
    sprintf(fullyQualifiedVarname, "%s/%s",hGroupLabel,varName_p);
    strcpy(dsetName, fullyQualifiedVarname); // for backward compatibility

    // xLogMsg("@A In(%s) INFLATE-and-WRITE(%s)~\n",fxName, dsetName);

    // Clear container for re-bind to another datatype (uint16)
    nElem = sizeof(uint16) * EGv2_GLOBAL_9KM_PIXELS;
    memset(DEST_p, '\0', nElem);
    // Bind uint16 data-type for generic 2D grid array
    destGrid_uint16_p = (uint16 *)DEST_p;

    l4c_inflate_init_9km(destGrid_uint16_p, DFNT_UINT16);

    l4c_inflate_9km(STATE_p->carbon_model_bitflag, destGrid_uint16_p, DFNT_UINT16);

    // WRITE to HDF5 output file 
    l4c_Hdf5_WriteDataset(fxName, out_file_id, dsetName,
        pkgMgr_GetThemes_ChildIdByName(pThme_p, dsetName),
        H5T_NATIVE_USHORT, destGrid_uint16_p);  

   // Release memory for local float32 LDOM9KM work array
   memFree(wrk9kmFltCount_p);

   elapSec = l4c_EventCompletionMsg(fxName);

 return 0;
} // end:: l4c_PostProcess_QA()
 
// =====================================================================
// PRIMARY POST-PROCESSOR DRIVER LOGIC
// =====================================================================
 
// ----------------------------------------------------------------
// function: l4c_PostProc9km_driverConvert()   
// Purpose : This function is the high level driver for the  1km to
//           9km and 9km x PFT averages spatial aggregation steps,
//           and is responsible for WRITING all HDF5 output datasets 
//           to the current days Level 4 Carbon output granule file.
//           e.g. L4_C_mdl_20151112T165400_Vb1010_001.h5
//
// inputs  : EcoStateArrayAdt *ecoStateArrayObj_p
// Notes   : Loops through 1km file name list input at command line,
//           processesing each file individually
//           into 9km and 9km x PFT averages using 
//           convert_1km_to_9km_by_pft(). The resulting arrays of averages 
//           are then written out to the primary HDF5 output file within
//           convert_1km_to_9km_by_pft().
// 
// notes    : Output path to directly to store results 
//            (command line arg from main()), array of filenames, and
//            total number of files to process.
// Author   : Joe Glassy
// Returns  : time_t --snapshot of ending time, as a time_t value
// Outputs  : One 9km average file and eight 9km x PFT average files
//            written per each 1km input file.
// Revised  : 2015-11-12T10:00 (only documentation added)
//
time_t l4c_PostProc9km_driverConvert(EcoStateArrayAdt *ecoStateArrayObj_p)
{
  SesCtl *sesCtl_p = l4c_GetSesCtl();
  AncImagePkgAdt *ancImagePkgObj_p   = l4c_GetAncImagePkgAdt();
  PrePostImgPkgAdt *prePostAdt_p = l4c_GetPrePostImagePkgAdt();
  PkgMgr   *pkgMgr_p = sesCtl_p->l4cPkgMgr_p ; // set ptr to OVERSEER
  PkgTheme *pThme_p= NULL;
  PostProcMetrics postMetrics ;
  
  PftAdt           PFT ; // local PFT..
  
  // local container for row,col luts, referencing layers in Ancillary file
  SplandRefAdt SPLAND  ;
  
  // The stateOutput2D_p pointer is treated as a re-usable buffer, sized to the greatest
  // common denominator of variable datatype/extents, to allow it to
  // store any OUTPUT variable 2D FULL GRID (destination array)
  uint8  *stateOutput2D_p = NULL;
  
  size_t pix9km =0;  // for EC recaling (0..1) to (0..100)
  size_t nElem =0;
  size_t nSum =0;
    
  int t =0;
  int nUnique =0;
  
  time_t postStrtSec  = time(0);  
  time_t postEndiSec  = time(0);
  double postElapSec  = 0.0;
  double cumElapSec   = 0.0;

  // This is the pointer to destination OUTPUT array that represents the
  // 9km 2D instance of ->frozen_area
  float32 *frozenArea_M09_2DFullGrid_p = NULL; // special case, already AT 9km...
  
  char fxName[256] = {0};
  char timeStr[32] = {0};
  
    strcpy(fxName,(char *)__func__);

    xLogMsg("\n");
    xLogMsg("@A ============== L4_C POST-PROCESSOR ================== ~\n");
    xLogMsg("@A In(%s) CONVERTING 9km 1D-sparse land vectors to 9KM 2-D global grids at (%s) ~\n",
        fxName,frm_utilGetCtime(timeStr,postStrtSec)  );

    // RMSE development. Stub allocate (1)+(8) 9KM RMSE variables
    l4c_PostProcess_RMSE_Alloc_Memory();

    // Step 0: output bitfield assignment (PftDom,QaScore) tracker counts
    nUnique = 0;
    nSum    = 0 ;
    for(nUnique=0,t=0; t < 256; t++)
        if( pfvmPftDom.pfvm_Uint8freq[t] > 0)
        {
            xLogMsg("@D In(%s) PftDom index(%3d) absfreq(%zu)~\n",
                fxName, t, pfvmPftDom.pfvm_Uint8freq[t]);
            nUnique++ ; // tally n. of unique values with positive counts
            nSum += pfvmPftDom.pfvm_Uint8freq[t] ;
        }

    xLogMsg("@D POST-PROC BITFIELD PftDom tracked %d unique values. Sum(%zu)~\n",
        nUnique, nSum);     

    // Next, output freq-of-occurance of QaScores, validate they are in range
    nSum =0;
    nUnique = 0 ;
    for(t=0; t < 256; t++)
        if( pfvmQaScore.pfvm_Uint8freq[t] > 0)
        {
            xLogMsg("@D In(%s) QaScore index(%3d) absfreq(%zu)~\n",
                fxName, t, pfvmQaScore.pfvm_Uint8freq[t]);
            nUnique++; // tally n. unique values with positive counts
            nSum += pfvmQaScore.pfvm_Uint8freq[t] ;
        }
    xLogMsg("@D POST-PROC BITFIELD QaScore tracked %d unique values. Sum(%zu) ~\n",
      nUnique, nSum);

    // ----------------------------------------------------------
    // PHASE 1: Attach pre-existing {row,col} LUTs to local SPLAND
    SPLAND.row = ancImagePkgObj_p->imgM09land_row_uint16_p;
    SPLAND.col = ancImagePkgObj_p->imgM09land_col_uint16_p;
    
    // DIRECT ATTACH to pre-existing PFT 1km 1D sparse mask
    PFT.dom_1km = ancImagePkgObj_p->imgPftM01land_dom_p ;
    
    // nElem here is the total n. 'bytes' to handle even float32 etc
    nElem = sizeof(float32) * EGv2_GLOBAL_9KM_PIXELS;
    
    // Allocate up front one re-usable generic container,
    // to hold full 2-D 9KM grid output fields, one at a time, noting
    // that 'nElem' is 1624x3856 pixels by float32.
    stateOutput2D_p  = (uint8*)memCalloc(nElem, sizeof(uint8));

    pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "L4C_OUTPUTS" );

    // STEP 2: Aggregate EACH primary variable from 1km to 9km. Noting
    //  each of these spawns the FULL SET of companion output 9km arrays, 
    //  each of which separately need a INFLATE transformation then WRITE.
    //  Remember we are re-using the STATE and DEST compound structures 
    //  where the DEST contains a re-usable array.

    // PHASE 2: POST-PROCESS: iterate: "NEE","GPP","RH","SOC" of (4) science variables
    // refactored for more clarity
    // nElem = sizeof(float32) * EGv2_GLOBAL_9KM_PIXELS;
    l4c_PostProcess_PrimaryVars("nee", &ecoStateArrayObj_p->nee  , stateOutput2D_p);
    l4c_PostProcess_PrimaryVars("gpp", &ecoStateArrayObj_p->gpp  , stateOutput2D_p);
    l4c_PostProcess_PrimaryVars("rh" , &ecoStateArrayObj_p->rhtot, stateOutput2D_p);
    l4c_PostProcess_PrimaryVars("soc", &ecoStateArrayObj_p->Ctot , stateOutput2D_p);

    // -----------------------------------------------------------------
    // PHASE 3: POST-PROCESS: (4) "EC" variables and Frozen area
    // changed all mult_av to mult_mean names
    // Emult Tmult Wmult Frozen area, {"EC/emult_mean","EC/tmult_mean","EC/wmult_mean","EC/frozen"}
    
    // <JMG> 2014-10-21, prior to post-processing, rescale (3) EC 1D-sparse vector 
    // arrays from {0.0<= a <=1.0} to {0.0< a < 100.0}
    // <LAJ> 2017-05-24, this is where fill values are getting rescaled in the EC vars
    xLogMsg("@A In(%s) Rescaling %zu pixels ea for EC (Emult,Tmult,Wmult) from (0..1) to (0..100)~ \n",
            fxName, (size_t)EGv2_LDOM9KM );          
              
    for(pix9km=0; pix9km < EGv2_LDOM9KM; pix9km++)
    {
        // <JMG> 2017-05-24, add check to skip OVER any missing value pixels
        if(ecoStateArrayObj_p->Emult.avg_9km[pix9km] > FLOAT_FILL_VALUE)
        {
           ecoStateArrayObj_p->Emult.avg_9km[pix9km] *= 100.0 ;
           ecoStateArrayObj_p->Tmult.avg_9km[pix9km] *= 100.0 ;
           ecoStateArrayObj_p->Wmult.avg_9km[pix9km] *= 100.0 ;
        } // otherwise, leave existing -9999 value intact
        
    } // end::pix3km rescaling EC 9km vars
    
    xLogMsg("@A In(%s) Postprocessing EC terms (Emult,Tmult,Wmult) ~\n",fxName);
    // <JMG> 2016-02-18 changed all EC mult_av to mult_mean names
    l4c_PostProcess_EC_Vars("emult_mean"   , &ecoStateArrayObj_p->Emult, stateOutput2D_p);
    l4c_PostProcess_EC_Vars("tmult_mean"   , &ecoStateArrayObj_p->Tmult, stateOutput2D_p);
    l4c_PostProcess_EC_Vars("wmult_mean"   , &ecoStateArrayObj_p->Wmult, stateOutput2D_p);
    
    xLogMsg("@A In(%s) Postprocessing EC frozen_area ~\n",fxName);
    //  Process EC/frozen area, noting it does NOT require aggregation
    //   from 1km to 9km, because it is ALREADY defined at 9km, 
    //   however it requires inflating from 1D 9km to 2D 9km 
    frozenArea_M09_2DFullGrid_p = memCalloc(EGv2_GLOBAL_9KM_PIXELS,sizeof(float32));

    // Devel mode only: if RIP L4C_GRANULE_METRICS = enable, and
    //   calculate observed min,max,mean etc
    if(sesCtl_p->enableGranuleMetrics )
    {
        // upscale from 0..1 to {0.0<pix<100.}
        postMetrics.optionScaleFactor = 100.0; // upscale from 0..1 to {0.0<pix<100.}

        l4c_SummarizePixels(&postMetrics, "EC/frozen_area",
            pkgMgr_GetThemes_ChildIdByName(pThme_p, "EC/frozen_area"),
            EGv2_GLOBAL_9KM_PIXELS, (float32)FLOAT_FILL_VALUE, ecoStateArrayObj_p->FrozenArea_9km);
    }

    // Inflate 1D 'frozenArea' to full grid 2D form    
    l4c_inflate_init_9km(frozenArea_M09_2DFullGrid_p, DFNT_FLOAT32);
    l4c_inflate_9km(ecoStateArrayObj_p->FrozenArea_9km,
                    frozenArea_M09_2DFullGrid_p, DFNT_FLOAT32);
                    
    l4c_Hdf5_WriteDataset(fxName,  
        sesCtl_p->l4cOutput_file_id, "EC/frozen_area",
        pkgMgr_GetThemes_ChildIdByName(pThme_p,"EC/frozen_area"), // fetch datset_id 
        H5T_NATIVE_FLOAT, frozenArea_M09_2DFullGrid_p);
    // Release memory used for local, special-case frozen area 9km array
    memFree(frozenArea_M09_2DFullGrid_p) ;

    // --------------------------------------------------------
    // PHASE 4: QA processing block
    // <JMG> 2014-09-03T14:02, all QA variables have been consolidated
    // into one theme "L4C_OUTPUTS" and are stored now in one L4_C .h5 output granule
    l4c_PostProcess_QA(&ecoStateArrayObj_p->nee_qa, stateOutput2D_p);

    // --------------------------------------------------------
    // PHASE 5: RMSE QA (optional) processing block - Uses "nee_qa" variable
    //  <LAJ> 11-14-14  Do post-processing of RMSE fields due to their
    //  datatype being float32 and due to these having a xxx_pft members.
    //  For generating RMSE we pass "nee_qa" to l4c_PostProcess_PrimaryVars)
    //  where internally (triggered by presence of "nee_rmse" label, and for
    //  the RMSE ops, we suppress the (un-used) stdev field not present in RMSE
    if( sesCtl_p->enableRMSE )
    {
        // Assert: produce nee_rmse_mean, nee_rmse_pft{1..8}_mean ...
        xLogMsg("@A In(%s) Enabled generation of optional RMSE ~\n",fxName);
        l4c_PostProcess_PrimaryVars("nee_rmse", &ecoStateArrayObj_p->nee_qa, stateOutput2D_p);
    }

    // -----------------------------------------------------------------
    // PHASE 6: RELEASE Local memory
    xLogMsg("@A In(%s) Releasing remaining local L4_C post-processing arrays ~\n",fxName);
    // Free main PFT array
    memFree(PFT.dom_1km);  // release memory for large 1km 2D working grid
    // Free up the (parent) generic container object(s) 
    // used for post processing
    memFree(stateOutput2D_p);

    // -----------------------------------------------------------------
    // PHASE 7: POST-PROCESS: Use qa_count to populate ISO CompletenessOmission
    // and DomainConsistency for NEE and rest.
    // <JMG> 2014-11-11,  NOTE: ecoStateArrayObj_p->nee_qa is actually NEVER
    // passed in to l4c_PostProc_fltVar9km() whereas in l4c_ForwardStep
    // on line 440, &ecoStateArrayObj_p->nee IS passed in, and deep inside
    // l4c_PostProc_fltVar9km() , on line 475, is the only place that 
    // varFlt32_Obj_p->domainConsistencySum_1km++; is incremented! So,
    // It appears that &ecoStateArrayObj_p->nee_qa is NOT the array to 
    // pass in here, but rather it should be &ecoStateArrayObj_p->nee !!
    
    l4c_PostProcessGranuleCompleteness(&ecoStateArrayObj_p->nee);
    // -----------------------------------------------------------------
    
    l4c_Populate_ISO_MetadataElements();
    
    // WRITE ISO Metadata out, incl DomainConsistency,CompletenessOmimssion
    l4c_Hdf5_Add_ISO_FileMetadata(sesCtl_p->l4cOutput_file_id) ;
    // -----------------------------------------------------------------

    // -----------------------------------------------------------------
    // <JMG> 2016-01-04 RMSE -- release memroy
    l4c_PostProcess_RMSE_Release_Memory();

    // snapshot local subprocess time 
    postEndiSec = time(0);
    postElapSec = difftime(postEndiSec,postStrtSec);
    // snapshot global time
    sesCtl_p->sesEndiSec = time(0); // snapshot cumulative time thus far
    cumElapSec = sesCtl_p->sesEndiSec - sesCtl_p->sesStrtSec;
    
    xLogMsg("@A In(%s) POST-PROCESSOR FINISHED (%s) requiring %.2lf sec (local)~\n",
        fxName,frm_utilGetCtime(timeStr,postEndiSec),  postElapSec  );
    xLogMsg("@A In(%s) session-cumulative elapsed %.2lf sec (global)~\n",
        fxName,  cumElapSec );
    
   cumElapSec = l4c_EventCompletionMsg(fxName);
   
   // return ending time
   return postEndiSec;

} // end :: l4c_PostProc9km_driverConvert()

// ---------------------------------------------------------------------
// function : l4c_PostProc_fltVar9km(()
// purpose  : Summarizes 1km floating point arrays (81 elements) into 
//            9km and 9km x PFT and places the results in global 9km grid.
// revised  : 2014-08-21T18:51:00, adapted for NDVI use
// inputs   : uint32 s9km       -- base0 pixel index into s9km model domain
//           PftAdt *PFT_ARRAY  -- ptr to PFT classification image
//           const char *datasetName_p -- name of current HDF5 output dataset
//           VarFloat32Adt *varFlt32_Obj_p -- pointer to compound container ADT
//
// Notes    : l4c_PostProc_fltVar9km() is now repetitively called in a 
//            9km 1D-Sparse vector loop, once per unique s9km index value.
//
// Note on Inputs: Index into 9km global array, ancillary 1km PFT data 
//         for enclosing 9km grid cell, allocated state structure field 
//         containing 1km input data, scalars for intermediate results
//         (sums, etc.) and empty allocated 9km spatial arrays for 
//          output data.
// Author  : Lucas Jones
// Outputs : science code implementation: outputs one 9km average raw binary file 
//           and eight 9km x PFT average raw binary file(s). These are
//           are written per each 1km input file (i.e. each call to this function).
// Notes   : <JMG> 2015-10-15, valgrind reporting use of uninitialized (size 8)
//           variables in the if block (around line 1110) for SOC/soc_mean
           
// Notes   : <JMG> 2014-08-21, (CORRECTED OK NOW) prior, a SEGFAULT
//           tripped in l4c_SetBits() encountered when in NDVI mode, works fine in FPAR mode, 
//           where the search for cause is pending. This may be risky 
//           code where string literals are used for strcmp() ops on 
//           dataset names, as if any name is mis-spelled, the
//           resulting error will be very hard to find.
 
void l4c_PostProc_fltVar9km(uint32 s9km, PftAdt *PFT_ARRAY, const char *datasetName_p, VarFloat32Adt *varFlt32_Obj_p)
{
 // <JMG> 2014-07-23T17:29, Now define valid_min,valid_max and thresholds
  // via QcStateLimitAdt *qcStateLim_p = l4c_GetQcStateLimits();
 // Fetch valid min,max and thresholds assigned to QcStateLimitAdt qaLim structure
 // qcStateLim_p = l4c_GetQcStateLimits();
 char fxName[256] = {0} ;
 SesCtl *sesCtl_p = l4c_GetSesCtl();
 int       ithPft =0;
 int       rn =0 , cn =0;
 float32   fltCount =0.0 ;  
 float32   sumXsum = 0.0 ;
 float32   chkFlt =0.0 ;
 size_t    DoNoOp = 0 ;
 uint8     pft_index    = 0;
 uint8     byteIsSet = 1 ;  // placeholder allows use of AssignBits
 uint16    wordIsSet = 1 ;  // placeholder allows use of AssignBits
 uint32    s1km_nested  = 0;
 uint16    thisCarbonModelBitflag = 0;

    strcpy(fxName, __func__);
    
    // Assign current carbon model bitflag immediately. Dereferencing
    // this right way reveals if this array is intact or corrupted
    thisCarbonModelBitflag = varFlt32_Obj_p->carbon_model_bitflag[s9km] ;

    // Initialize sums, 9km x pft sums for each pft
    for (ithPft=0; ithPft < NPFT; ithPft++)
    {
        varFlt32_Obj_p->pft_9km[ithPft].sum = 0.0 ;     // sum
        varFlt32_Obj_p->pft_9km[ithPft].cnt[s9km] = 0;  // count, a uint8
    } // end :: ithPft < NPFT loop

    // Initialize 9km sums and counts
    varFlt32_Obj_p->sum_9km       = 0.0;    // sum, 9km
    varFlt32_Obj_p->sum2_9km      = 0.0;    // sum of squares, 9km
    varFlt32_Obj_p->cnt_9km[s9km] =   0;    // count, 9km
    varFlt32_Obj_p->limitExceedCnt_9km = 0; // count of limit exceedence for 9km grid cell

    // Begin nested 1km 2D grid loops  (9x9=81), [NESTED_1KM_IN_9KM == 9]
    // defined in tcf_const.h
    for (rn=0; rn < NESTED_1KM_IN_9KM; rn++)
    {
        for (cn=0; cn < NESTED_1KM_IN_9KM; cn++)
        {
            // Serial array index for 1km grid
            s1km_nested = M_2D_B0(rn, cn, NESTED_1KM_IN_9KM);

            // Unpack pft
            pft_index = PFT_ARRAY->nested_1km[s1km_nested];

            // Process only veg pfts with valid state 
            //  data (FLOAT_FILL_VALUE already reflects
            //  valid model simulation grid cells)
            if(varFlt32_Obj_p->nested_1km[s1km_nested] > FLOAT_FILL_VALUE)
            {

                // De-increment 1-based pft_index to 0-based array index
                ithPft = pft_index-1;

                // update count, Add specified pft to 9km x pft tallies
                varFlt32_Obj_p->pft_9km[ithPft].cnt[s9km]++;

                // update count, Add 1km grid cell to 9km tallies
                varFlt32_Obj_p->cnt_9km[s9km]++;

                // Tally global total number of grid cells where simulation attempted ("Domain Consistancy")
                varFlt32_Obj_p->domainConsistencySum_1km++;

                // Add to running sums. 
                // <JMG> 2014-11-05, changed coerce to float32 since these are not float64
                varFlt32_Obj_p->sum_9km  += (float32)varFlt32_Obj_p->nested_1km[s1km_nested];
                
                sumXsum = varFlt32_Obj_p->nested_1km[s1km_nested] * varFlt32_Obj_p->nested_1km[s1km_nested] ;
                varFlt32_Obj_p->sum2_9km += sumXsum ;
                
                // CAREFUL: one instance of a "nested_1km[]" is a uint8 and another 
                //  identically named variable (at another level) is a float32.
                //  This is a quite risky practice, if these ever get confused,
                // <LAJ> 11-14-14, The "nested_1km" field applies to a PFT structure, whereas the other
                //   "nested_1km" fields are in state variable structures.  I don't expect these two structures 
                //   to be easily confused.
                varFlt32_Obj_p->pft_9km[ithPft].sum += (float32)varFlt32_Obj_p->nested_1km[s1km_nested];

                // Check state variable limits, increment tallies
                if ((varFlt32_Obj_p->nested_1km[s1km_nested] < varFlt32_Obj_p->limitMin) || 
                    (varFlt32_Obj_p->nested_1km[s1km_nested] > varFlt32_Obj_p->limitMax))
                {                 
                    // Tally 1km grid cells exceeding expected bounds for overlying 9km grid 
                    varFlt32_Obj_p->limitExceedCnt_9km++;

                    // Tally global total number of grid cells where simulation attempted BUT value exceeds expected bounds
                    varFlt32_Obj_p->completenessOmissionSum_1km++;
                }
        
            } // end :: if varFlt32_Obj_p->nested_1km[s1km_nested] > FLOAT_FILL_VALUE (if veg pft check)

        } // end :: 1km row loop
    } // end :: 1km col loop

    // STATE VARIABLE 9-km SUMMARY----------------------------------------------------------
    // Get 9km averages and stdev. and 9km x pft averages based on 1km results

    // Loop through Pft and increment 9km x pft averages based on 1km counts
    for (ithPft=0; ithPft < NPFT; ithPft++)
    {
        // Guard against div by zero. If so, assign fill values
        fltCount = (float32)varFlt32_Obj_p->pft_9km[ithPft].cnt[s9km] ;
        if( fltCount > 0.0)
        {
            varFlt32_Obj_p->pft_9km[ithPft].avg[s9km] = (float32)varFlt32_Obj_p->pft_9km[ithPft].sum / fltCount;
        } else {
            varFlt32_Obj_p->pft_9km[ithPft].avg[s9km] = FLOAT_FILL_VALUE;
        } // end :: check 1km pft count > 0

    } // end :: ithPft, pft loop

    // Test diag: reset to known values first
    varFlt32_Obj_p->var_9km       = 0.0;
    varFlt32_Obj_p->std_9km[s9km] = 0.0;
    
    // Check for 1km count > 0, if not assign fill values
    // TBD in future, also check the upper end value too in case 
    //  something went wrong...
    fltCount = (float32)varFlt32_Obj_p->cnt_9km[s9km] ;

    // Initialize as fill, then overwrite if fltCount > 0.0...
    varFlt32_Obj_p->avg_9km[s9km] = FLOAT_FILL_VALUE;
    varFlt32_Obj_p->std_9km[s9km] = FLOAT_FILL_VALUE;

    if (fltCount > 0.0)
    {
        // Do the grand averages and standard devs. for the 9-km grid cell
        varFlt32_Obj_p->avg_9km[s9km] = (float32) varFlt32_Obj_p->sum_9km / fltCount; 

        // <JMG> 2014-11-05T17:18, many NaN and Inf detected in std_9km outputs
        // so work on making this more robust...
        // stddev   = sqrt((sum_sq-(sum * sum)/n)/(n-1.0));
      
        // Do this once only
        sumXsum = varFlt32_Obj_p->sum_9km * varFlt32_Obj_p->sum_9km ;
        varFlt32_Obj_p->var_9km  = (float32)varFlt32_Obj_p->sum2_9km - sumXsum  / fltCount;

        // <LAJ> 11-14-14, Assign zero to stdev and variance when only one obs available
        if (fltCount > 1.0)
        {
            varFlt32_Obj_p->var_9km *= (float32) 1.0 / (fltCount - 1.0);  

            // TBD: check-guard against NaN here using custom verification code..
            varFlt32_Obj_p->std_9km[s9km] = (float32)sqrt( (float64)varFlt32_Obj_p->var_9km );
        
        } else {
            // <JMG> 2016-04-21 re-verify here if for the RMSE case we should be
            // leaving this as FLOAT_FILL_VALUE, still taking care not to 
            // trigger a divide-by-zero error just below in sqrt() or other calcs
            varFlt32_Obj_p->var_9km = 0.0 ;  
            varFlt32_Obj_p->std_9km[s9km] = 0.0;
        } // end :: one obs check

        // Implement special C99 trap any NaN or Inf to detect hits right away
        chkFlt =  (float32)sqrt( (float64)varFlt32_Obj_p->var_9km );
      
        // count occurrences of NaN and Inf, to suppress...
        if(fpclassify(chkFlt) == FP_NAN)
        {
            pInv.invalidHitNaN++ ;
        }
      
        if(fpclassify(chkFlt) == FP_INFINITE)
        { 
            pInv.invalidHitInf++;
        }
      
        // Last ditch robustness test uses isfinite() macro, where in
        //  either case, if its a NaN OR it is a Inf, we suppress and 
        //  substitute a discrete sentinal value -9998.0 so we can 
        //  distinguish these cases from all other uses of -9999.0 
        if( isfinite(chkFlt))
        {
            varFlt32_Obj_p->std_9km[s9km] = chkFlt ;
        } else {
            // 2014-11-11 consider if suppression result should be 0.0 or -9999.0
            varFlt32_Obj_p->std_9km[s9km] = FLOAT_FILL_VALUE;
        }

        // Check for for exceedence of expected reasonable bounds 
        // <LAJ> 7/21/14 : Set QA min/max bounds here (must switch on or index on dataset name)
        if(varFlt32_Obj_p->limitExceedCnt_9km > 0)
        {

            // NEE limits and bitflag
            if(strcmp("NEE/nee_mean", datasetName_p) == 0)
            {
				// <JMG> 2021-10-19, substitute frm_utilAssignBits for frm_utilSetBits
				// invocation form: destQcWordPix =(uint8)frm_utilAssignBits(0,0,&destQcWordPix,sizeof(uint8),&modlandQc,sizeof(uint8)) ;
			  varFlt32_Obj_p->carbon_model_bitflag[s9km] = (uint16)frm_utilAssignBits(0,0,
							&varFlt32_Obj_p->carbon_model_bitflag[s9km], sizeof(uint16),
							&wordIsSet, sizeof(uint16));
				

              // <JMG> 2014-10-01, FATAL BUG 2014-10-01 in gdb test run
              // "Program received signal SIGSEGV, Segmentation fault."
              //  0x0000003b0648968e in memcpy () from /lib64/libc.so.6 on this l4c_SetBits() call
              //  AGAIN, but this time at s9km=42444, using FPAR, not NDVI
              //  so a transient bug appears to have returned. The call-trace where this bug
              //  appears  is virtually identical to the 2014-08-21 fatal bug earlier (see above)
              
              // <JMG> 2021-10-25, change to use AssignBits
              varFlt32_Obj_p->carbon_model_bitflag[s9km] = (uint16)l4c_AssignBits(0,0,
                            &varFlt32_Obj_p->carbon_model_bitflag[s9km], sizeof(uint16),
                            &wordIsSet,sizeof(uint16));
              
              // 2014-08-21, added counters to tally n. of "invalid" NEES,
              // so here we increment special counters of exceptions to report out later
              
              pInv.invalidNee++ ; // increment count of apparently invalid NEE's               
                
              if(varFlt32_Obj_p->avg_9km[s9km] < varFlt32_Obj_p->limitMin)
                    pInv.invalidNeeMin++ ;
                    
              if(varFlt32_Obj_p->avg_9km[s9km] > varFlt32_Obj_p->limitMax)
                    pInv.invalidNeeMax++ ;
                
            // GPP limits and bitflag
            } else if (strcmp("GPP/gpp_mean", datasetName_p) == 0) {

               // <JMG> 2014-08-21T19:15, ...check on potential SEGFAULT in l4c_SetBits()
               
               // <JMG> 2021-10-25, change to use AssignBits
               varFlt32_Obj_p->carbon_model_bitflag[s9km] = (uint16)l4c_AssignBits(1,1, 
                     &varFlt32_Obj_p->carbon_model_bitflag[s9km], sizeof(uint16),
                     &wordIsSet, sizeof(uint16));
        
               pInv.invalidGpp++ ; // increment count of apparently invalid GPPs
                
               if(varFlt32_Obj_p->avg_9km[s9km] < varFlt32_Obj_p->limitMin)
                    pInv.invalidGppMin++;
                    
               if(varFlt32_Obj_p->avg_9km[s9km] > varFlt32_Obj_p->limitMax)
                    pInv.invalidGppMax++;

            // RH limits and bitflag
            } else if (strcmp("RH/rh_mean", datasetName_p) == 0) {

            // <JMG> 2021-10-25, change to use AssignBits
            varFlt32_Obj_p->carbon_model_bitflag[s9km] = (uint16)l4c_AssignBits(2,2, 
                    &varFlt32_Obj_p->carbon_model_bitflag[s9km], sizeof(uint16),
                    &wordIsSet,sizeof(uint16) );

            // SOC limits and bitflag
            } else if (strcmp("SOC/soc_mean", datasetName_p) == 0) {
            
               // <JMG> 2015-10-15: valgrind reports (2) types of non-fatal errors here,
               // for each separate assignment occurance in this block. Line numbers
               // reported have been adjusted (orignally flagged 1087, is now 1134)
               // 1) Use of uninitialised value of size 8
               // ==24351== at 0x4A08DAC: memcpy (mc_replace_strmem.c:882)
               // ==24351== by 0x4074EF: l4c_SetBits (l4c_DAILY_sci.c:264)
               // ==24351== by 0x423BCF: l4c_PostProc_fltVar9km (l4c_PostProc9km_func.c:1087)
               // ==24351== by 0x420284: l4c_ModDriverGrid_forwardStep
               // ==24351== Conditional jump or move depends on uninitialised value(s)

            
               varFlt32_Obj_p->carbon_model_bitflag[s9km] = (uint16)l4c_AssignBits(3,3,
                    &varFlt32_Obj_p->carbon_model_bitflag[s9km], sizeof(uint16),
                    &wordIsSet, sizeof(uint16) );
                    
               // diag only: increment count of times we reach this branch
               pInv.invalidSocMean++ ;
            
            } else {
                // Default : don't do anything if EC datasets, etc.
                DoNoOp++; // count NO-OPs ...
                
            } // end :: chain, if(strcmp("NEE/nee_mean", datasetName_p))...

        }  // end :: if(varFlt32_Obj_p->limitExceedCnt_9km > 0)
      
    } // end :: check 1km count > 0


} // end :: l4c_PostProc_fltVar9km()


// ---------------------------------------------------------------------
// l4c_PostProc_qaRank9km(()
//
// Purpose : Summarizes 1km floating point QA pseudo-error arrays
//            (81 elements) into 9km and 9km x PFT, ranks into
//            integer classes and places results in global 9km grid.
// 
// inputs  : uint32 s9km, 
//           PftAdt *PFT_ARRAY, 
//           VarFloat32Adt *varFlt32_Obj_p --input var is the "nee_qa" instance of VarFloat32Adt
// revised : 2014-10-16T15:21, reviewed, added minor doc only
// revised : 2014-09-04T13:17, removed param dataset name
// notes   : This function is called for each pixel in 9km space.
// Notes   : Index into 9km global array, ancillary 1km PFT data for 
//           enclosing 9km grid cell, allocated state 
//           structure field containing 1km input data, scalars for 
//           intermediate results (sums, etc.) and empty allocated 9km 
//           spatial arrays for output data.
// Notes   : Reads BPLUT, which contains ranking information (how many classes, breaks, etc.).
// Author  : Lucas Jones
// Outputs : One 9km QA rank file and eight 9km x PFT rank files written per each 1km input file
//          (i.e. each call to this function). Also, writes out cell count for each PFT and grand averages
//          for the enclosing 9km grid cell.
// ---------------------------------------------------------------------
void l4c_PostProc_qaRank9km(uint32 s9km, PftAdt *PFT_ARRAY, VarFloat32Adt *varFlt32_Obj_p)
{
  SesCtl *sesCtl_p = l4c_GetSesCtl() ;
  PrePostImgPkgAdt *prePostAdt_p = l4c_GetPrePostImagePkgAdt();    
  // <JMG> 2014-10-16T16:12, added monitors tracking multi-byte bitfield assignments
  // Note: these freq [0..255] arrays were initalized in the forward step fx.
  PostProcFreqVerifyMetrics *pfvmPftDom_p = l4c_GetPostProcFreqVerify_PftDom();
  PostProcFreqVerifyMetrics *pfvmQaScore_p= l4c_GetPostProcFreqVerify_QaScore();
  // <JMG> 2014-07-23T17:29, added for storing valid_min,valid_max and thresholds
  QcStateLimitAdt *qcStateLim_p = l4c_GetQcStateLimits();
  BPLUT     *bp_p = NULL; // ptr to BPLUT record selected via PFT class
  
  int       ithPft =0, rn =0, cn=0;

  uint8     pft_index      = 0 ;
  uint8     pft_dom9km     = 0;
  uint8     pft_dom9km_cnt = 0;
  uint8     qaScore =0;
  
  uint32    s1km_nested = 0;
  float32   sumXsum     = 0.0 ;
  float32   useFparClimPerc = 0.0;
  float32   useNdviPerc = 0.0 ;
  float64   tmp = 0.0;

  // FETCH valid min,max and thresholds assigned to QcStateLimitAdt qaLim structure
  qcStateLim_p = l4c_GetQcStateLimits();

  // Initialize 9km x pft sums for each pft
    for (ithPft=0; ithPft < NPFT; ithPft++)
    {
      varFlt32_Obj_p->pft_9km[ithPft].sum2 = 0.0;     // sum of squares
      varFlt32_Obj_p->pft_9km[ithPft].cnt[s9km] = 0;  // count
    }

    // Initialize 9km fields
    varFlt32_Obj_p->sum2_9km       = 0.0; // sum of squares
    varFlt32_Obj_p->cnt_9km[s9km]  = 0;   // count
    varFlt32_Obj_p->rank_9km[s9km] = 0;   // qa rank

    // Initialize dominant pft variables
    pft_dom9km = 0;          // dominant pft within 9km grid cell
    pft_dom9km_cnt = 0;      // 1km grid cell count of dominant pft 

    // Begin nested 1km 2D grid loops. 
    // Note NESTED_1KM_IN_9KM is 9, so iterate through (81) pixels
    for (rn=0; rn < NESTED_1KM_IN_9KM; rn++)
    {
        for (cn=0; cn < NESTED_1KM_IN_9KM; cn++)
        {
            // Serial array index for 1km grid. This derives
            // the equivalent index into the 1km grid
            s1km_nested = M_2D_B0(rn, cn, NESTED_1KM_IN_9KM);
            // one-time-only, calc product
            sumXsum =  varFlt32_Obj_p->nested_1km[s1km_nested] * varFlt32_Obj_p->nested_1km[s1km_nested];

            // Unpack pft
            pft_index = PFT_ARRAY->nested_1km[s1km_nested];

            // FILTER: Process only vegetated PFTs with valid state data
            if (varFlt32_Obj_p->nested_1km[s1km_nested] > FLOAT_FILL_VALUE)
            {
                // De-increment 1-based pft_index to 0-based array index
                ithPft = pft_index-1;

                // Increment specified pft to 9km x pft tallies
                varFlt32_Obj_p->pft_9km[ithPft].cnt[s9km]++;

                // Increment 1km grid cell to 9km tallies
                varFlt32_Obj_p->cnt_9km[s9km]++;

                // Add to running sums, doing arithmatic in float32
                varFlt32_Obj_p->sum_9km  += (float32)varFlt32_Obj_p->nested_1km[s1km_nested];
                varFlt32_Obj_p->sum2_9km += sumXsum;

                varFlt32_Obj_p->pft_9km[ithPft].sum += (float32)varFlt32_Obj_p->nested_1km[s1km_nested];
                varFlt32_Obj_p->pft_9km[ithPft].sum2 += sumXsum;
            } // end :: veg pft check*/

        } // end :: 1km row loop
    } // end :: 1km col loop

  // QA 9-km SUMMARY-------------------------------------------------------------

    // Loop through Pft and classify 9km x pft averages into integer ranks
    for (ithPft=0; ithPft < NPFT; ithPft++)
    {
        // Retrieve this PFT's BPLUT record's parameters.
        //  CRITICAL NOTE: a base-1 index is specified here so
        //  be sure it IS base1 and not base0.
        pft_index = ithPft + 1;
        bp_p = l4c_GetBPLUT_ByIndex(pft_index, NPFT);
      
        // Check for 1km pft count > 0, if not assign fill values
        if (varFlt32_Obj_p->pft_9km[ithPft].cnt[s9km] > 0)
        {
      
            // Calculate Quadratic averages based on 1km counts
            tmp = (float64) varFlt32_Obj_p->pft_9km[ithPft].sum2 / varFlt32_Obj_p->pft_9km[ithPft].cnt[s9km];
            // Take SQRT sqrt() of sum/count...
            varFlt32_Obj_p->pft_9km[ithPft].avg[s9km] = (float32) sqrt(tmp);

            // Reclassify for 9km x pft averages

            // <LAJ> 01/05/2016
            // SKIP RECLASSIFICATION HERE, INSTEAD POPULATE FLOAT FIELD WITH
            // VARFLT32.AVG FIELD WHICH CONTAIN THE 'RMSE' CALCULATION
            
            // <JMG> 2016-03-02, re-verify for new RMSE enabled logic that tcfModQa_reclas() should still be called here
            varFlt32_Obj_p->pft_9km[ithPft].rank[s9km] = tcfModQa_reclas(varFlt32_Obj_p->pft_9km[ithPft].avg[s9km],
                   bp_p->Nee_QA_Error_min, bp_p->Nee_QA_Error_max,
                   bp_p->Nee_QA_Rank_min, bp_p->Nee_QA_Rank_max);

            // Check for dominant pft
            if (varFlt32_Obj_p->pft_9km[ithPft].cnt[s9km] > pft_dom9km_cnt) 
            {
                // Assign "pft_dom9km" when count warrants it
                pft_dom9km = pft_index;
                pft_dom9km_cnt = varFlt32_Obj_p->pft_9km[ithPft].cnt[s9km];
            }  // end :: Check for dominant pft

        } else {

          // <LAJ> 2017-05-24 Added the below statement to assign fill values to 'nee_rmse' for each PFT stored in 'varFlt32_Obj_p->pft_9km[ithPft].avg'
          varFlt32_Obj_p->pft_9km[ithPft].avg[s9km] = FLOAT_FILL_VALUE;      
          varFlt32_Obj_p->pft_9km[ithPft].rank[s9km] = BYTE_FILL_VALUE;
      
        } // end:: if block, Check 1km pft count > 0
      
    } // end :: Pft loop


    // Check for 1km count > 0, if not assign fill values
    if (varFlt32_Obj_p->cnt_9km[s9km] > 0)
    {

        //  Grand quadratic average for the 9-km grid cell
        tmp = (float64) varFlt32_Obj_p->sum2_9km / varFlt32_Obj_p->cnt_9km[s9km];
        varFlt32_Obj_p->avg_9km[s9km] = (float32) sqrt(tmp);

        // Reclassify for 9km averages
        // NOTE: Will need to know dominant cover type here if QA rank
        // min/max varies with pft in the BPLUT)

        // <LAJ> 01/05/2016
        // SKIP RECLASSIFICATION HERE, and do POPULATE FLOAT FIELD WITH VARFLT32.AVG FIELD
        // WHICH CONTAIN THE 'RMSE' CALCULATION
        
        // <JMG> 2016-03-02, re-verify with Lucas for RMSE, that this should still be called
        varFlt32_Obj_p->rank_9km[s9km] = tcfModQa_reclas(varFlt32_Obj_p->avg_9km[s9km],
              bp_p->Nee_QA_Error_min, bp_p->Nee_QA_Error_max,
              bp_p->Nee_QA_Rank_min, bp_p->Nee_QA_Rank_max);
              
            
        // Set QC bit fields
        // Set dominant pft to QC 4-bit wide bit field, ('PFT_dominant')
        varFlt32_Obj_p->carbon_model_bitflag[s9km] = (uint16)l4c_AssignBits(4, 7,
            &varFlt32_Obj_p->carbon_model_bitflag[s9km],
            sizeof(uint16), &pft_dom9km, sizeof(uint8));

        // Set overall 9km QA rank to QC 4 bit wide bit field ('QA_score')
        varFlt32_Obj_p->carbon_model_bitflag[s9km] = (uint16)l4c_AssignBits(8, 11,
            &varFlt32_Obj_p->carbon_model_bitflag[s9km],
            sizeof(uint16), &varFlt32_Obj_p->rank_9km[s9km], sizeof(uint8));

        // <JMG> 2014-10-16T15:23, added frequency distrib check on inputs 
        // to bits {4,5,6,7} and bits {8,9,10,11} to re-verify ranges of values
        // inserted into these bits.  Increments freq-of-occurrence counts
        pfvmPftDom_p->pfvm_Uint8freq[pft_dom9km]++ ;
        
        // the SOURCE of QaScore is presumed to be rank_9km[]
        qaScore = varFlt32_Obj_p->rank_9km[s9km];
        pfvmQaScore_p->pfvm_Uint8freq[qaScore]++;

        // Determine dominant methods used for GPP
        // Calculate percentage of 9km grid cell where each method (FparClim or Ndvi) was applied
        useFparClimPerc =  (float32) varFlt32_Obj_p->cnt_useFparClim_9km[s9km] / varFlt32_Obj_p->cnt_9km[s9km] * 100.0;
    
        useNdviPerc = (float32) varFlt32_Obj_p->cnt_useNdvi_9km[s9km] / varFlt32_Obj_p->cnt_9km[s9km] * 100.0;

        // Set ('GPP_Method') bit (12,12)
        if (useFparClimPerc > qcStateLim_p->useFparClimThres)
        {
            varFlt32_Obj_p->carbon_model_bitflag[s9km] = (uint16)l4c_SetBits(12,12,
                &varFlt32_Obj_p->carbon_model_bitflag[s9km], sizeof(uint16));
        }

        // Set ('fPAR_Source') bit (13,13); 0 == MODIS fPAR, 1 == VIIRS fPAR
        // <KAE> 2021-11-01 Changed this so as to reflect fPAR data source
        if ( strcmp(sesCtl_p->fparSourceStr, "VIIRS") == 0 )
        {
        varFlt32_Obj_p->carbon_model_bitflag[s9km] = (uint16)l4c_SetBits(13,13,
            &varFlt32_Obj_p->carbon_model_bitflag[s9km], sizeof(uint16));
        }

        // Set ('FT_Method') bit (14,14)
        if (varFlt32_Obj_p->cnt_useTsurfFT_9km[s9km] > varFlt32_Obj_p->cnt_useRadarFT_9km[s9km]) 
        {
            varFlt32_Obj_p->carbon_model_bitflag[s9km] = (uint16) l4c_SetBits(14,14,
                &varFlt32_Obj_p->carbon_model_bitflag[s9km], sizeof(uint16));
        }

    
    } else {

        // Otherwise assign fill values
        // <LAJ> 2017-05-24 Added the below statement to assign fill values to 'nee_rmse' stored in 'varFlt32_Obj_p->avg_9km'
        varFlt32_Obj_p->avg_9km[s9km] = FLOAT_FILL_VALUE;
        varFlt32_Obj_p->rank_9km[s9km] = BYTE_FILL_VALUE;

        // Set Missing value bitflag ('IsFill')
        varFlt32_Obj_p->carbon_model_bitflag[s9km] = (uint16)l4c_SetBits(15,15,
            &varFlt32_Obj_p->carbon_model_bitflag[s9km], sizeof(uint16));

    } // end :: Check 1km count > 0

} // end :: l4c_PostProc_qaRank9km()


// -------------------------------------------------------------------------------------------------------------
// function : l4c_PostProc9km_stepVarAlloc()
// purpose  : Allocate memory for (VarFloat32Adt) internal arrays inside
//            VarFloat32Adt structure, as nstances of model variable grid arrays. 
//            These currently contain fields for all various 
//            aggregration levels: 
//              1km, 9km x PFT, and 9km. All arrays contain global grids,
//              except for 1km which include only 81 1km gridcells nested 
//              within each 9km grid cell.
// inputs   : String specifier for variable type.  Float variables include
//            stdev. calculations, wheres QA variables 
//            include rank reclassification from original float fields.
// Outputs  : Allocated variable structure fields.
// notes    : Also used for QA-related arrays. 
// Author   : Lucas Jones
//
VarFloat32Adt l4c_PostProc9km_stepVarAlloc(const char *stateMode)
{
    size_t            ithPft = 0;
    VarFloat32Adt     stepVar_p; // declare static instance of a stepVar_p
    char vNameBuf[64] = {0} ;

    // Allocate space for 9-km gridcell (81 1-km elements)

    // Initialize grand sums for Domain Consistency and Completeness Omission
    stepVar_p.domainConsistencySum_1km    = 0;  
    stepVar_p.completenessOmissionSum_1km = 0;

    // Internal 1km States (passed to post processor)
    stepVar_p.nested_1km = (float32*)memCalloc(NESTED_1KM_IN_9KM_SQ, sizeof(float32));
    memObjSetName(stepVar_p.nested_1km,"stepVar_p.nested_1km");

    stepVar_p.avg_9km = (float32*)memCalloc(LDOM9KM, sizeof(float32));
    memObjSetName(stepVar_p.avg_9km,"stepVar_p.avg_9km");
    
    stepVar_p.cnt_9km = (uint8*)memCalloc(LDOM9KM, sizeof(uint8));
    memObjSetName(stepVar_p.cnt_9km,"stepVar_p.cnt_9km");

    // Universal output summary states for 9km x PFT
    for (ithPft=0; ithPft < NPFT; ithPft++)
    {
        sprintf(vNameBuf,"stepVar_p.pft_9km[pft_%zu].avg",ithPft);
        
        stepVar_p.pft_9km[ithPft].avg = (float32*)memCalloc(LDOM9KM, sizeof(float32));
        memObjSetName(stepVar_p.pft_9km[ithPft].avg,vNameBuf);
        
        sprintf(vNameBuf,"stepVar_p.pft_9km[pft_%zu].cnt",ithPft);
        
        stepVar_p.pft_9km[ithPft].cnt = (uint8*)memCalloc(LDOM9KM, sizeof(uint8));
        memObjSetName(stepVar_p.pft_9km[ithPft].cnt, vNameBuf);
    }  // end :: PFT loop

    // Additional rank summary data if QA variable
    if (strcmp(stateMode, "QA") == 0)
    {
        // <JMG> 2014-10-01, put carbon_model_bitflag alloc up first, to try to
        // diagnose why and where it gets corrupted, causes FATAL SEGVIOLATION
        // in the NEE l4c_SetBits(0,0...) call
        stepVar_p.carbon_model_bitflag = (uint16*)memCalloc(LDOM9KM, sizeof(uint16));
        memObjSetName(stepVar_p.carbon_model_bitflag,"stepVar_p.carbon_model_bitflag");
        
        stepVar_p.cnt_useFparClim_9km = (uint8*)memCalloc(LDOM9KM, sizeof(uint8));
        memObjSetName(stepVar_p.cnt_useFparClim_9km,"stepVar_p.cnt_useFparClim_9km");
        
        stepVar_p.cnt_useNdvi_9km = (uint8*)memCalloc(LDOM9KM, sizeof(uint8));
        memObjSetName(stepVar_p.cnt_useNdvi_9km,"stepVar_p.cnt_useNdvi_9km");
        
        stepVar_p.cnt_useRadarFT_9km = (uint8*)memCalloc(LDOM9KM, sizeof(uint8));
        memObjSetName(stepVar_p.cnt_useRadarFT_9km,"stepVar_p.cnt_useRadarFT_9km");
        
        stepVar_p.cnt_useTsurfFT_9km = (uint8*)memCalloc(LDOM9KM, sizeof(uint8));
        memObjSetName(stepVar_p.cnt_useTsurfFT_9km,"stepVar_p.cnt_useTsurfFT_9km");

        stepVar_p.rank_9km = (uint8*)memCalloc(LDOM9KM, sizeof(uint8));
        memObjSetName(stepVar_p.rank_9km,"stepVar_p.rank_9km");

        for (ithPft=0; ithPft<NPFT; ithPft++)
        {
            sprintf(vNameBuf,"stepVar_p.pft_9km[Pft_%zu].rank",ithPft);
            stepVar_p.pft_9km[ithPft].rank = (uint8*)memCalloc(LDOM9KM, sizeof(uint8));
            memObjSetName(stepVar_p.pft_9km[ithPft].rank,vNameBuf);
        }

    // Additional stdev. summary data if float state variable
    } else {  // not a "QA"..
        
      stepVar_p.std_9km = (float32*)memCalloc(LDOM9KM, sizeof(float32));
      memObjSetName(stepVar_p.std_9km,"stepVar_p.std_9km");
      
    } // end :: QA variable check

    // return fully allocated container stepVar structure
    return stepVar_p;
    
} // end :: l4c_PostProc9km_stepVarAlloc()


// ---------------------------------------------------------------------
// function : l4c_PostProc9km_stepVarFree()
// Purpose  : Frees memory allocation for individual model variable grid 
//            arrays allocated with tcfMem_stepVarAlloc(). 
// Notes    : Also used for QA-related arrays.
// Inputs   : String specifier for variable type and pointer to variable 
//            structure to free. 
// Author   : Lucas Jones
// Outputs  : None. 
void l4c_PostProc9km_stepVarFree(const char *stateMode, VarFloat32Adt *stepVar_p)
{

  size_t   ithPft;

  // Internal 1km States (passed to post processor)
  memFree(stepVar_p->nested_1km);

  // Universal output summary states for 9km
  memFree(stepVar_p->avg_9km);
  memFree(stepVar_p->cnt_9km);

  // Universal output summary states for 9km x PFT
  for (ithPft=0; ithPft<NPFT; ithPft++)
  {
      memFree(stepVar_p->pft_9km[ithPft].avg);
      memFree(stepVar_p->pft_9km[ithPft].cnt);
  } // end:: for ithPft..

  // Additional rank summary data if QA variable
  if (strcmp(stateMode, "QA") == 0)
  {
      memFree(stepVar_p->cnt_useFparClim_9km);
      memFree(stepVar_p->cnt_useNdvi_9km);
      memFree(stepVar_p->cnt_useRadarFT_9km);
      memFree(stepVar_p->cnt_useTsurfFT_9km);

      memFree(stepVar_p->rank_9km);
      memFree(stepVar_p->carbon_model_bitflag);

        for (ithPft=0; ithPft < NPFT; ithPft++)
            memFree(stepVar_p->pft_9km[ithPft].rank);

      // Additional stdev. summary data if float state variable
  } else {
      memFree(stepVar_p->std_9km);
  } // end:: strcmp(stateMode)...

} // end :: l4c_PostProc9km_stepVarFree()


// ---------------------------------------------------------------------
// function : l4c_tcf_qa_reclas()
// Purpose  : Generic utility function to reclassify float value  with specified 
//            float range into an integer value with a specified integer range. 
//            Reclassify a float value with a float range into an integer value 
//            with a specified integer range. 
// Notes    : CURRENTLY IS NOT USED..  Was required for assignment of 
//            output QA integer rank fields done as a post-process,
//            but retain the capability here in case 1km QA fields are desired.
// inputs   :  float32 error, 
//             float32 error_mn,
//             float32 error_mx,
//             uint32 qa_rank_mn,
//             uint32 qa_rank_mx
// Author   : Lucas Jones
// notes    : <JMG> 2016-01-04, function IS NOT currently called anywhere in l4c_DAILY.ex 
//            at this time, so it is a deprecation candidate.
// notes    : Float scalar, and pre-specified float extrema and integer rank extrema.
// Returns  : uint8, reclassified rank, expressed as {0..3} or {1..4}
// Outputs  : Reclassified float scalar.
// Notes    : uses FRM_API_ROUND(x) (((x)<0.50)?ceil(x-0.5):floor(x+0.50))

uint8 l4c_tcf_qa_reclas(float32 error, float32 error_mn,
    float32 error_mx, uint32 qa_rank_mn, uint32 qa_rank_mx)
{
 
  uint8      qa_rank =0;

  float32    tmp  = 0.0;
  float32    qa_rank_flt = 0.0;

  // Step 1: Rescale float (variable: "error") to range from 0 to 1
  qa_rank_flt = (error - error_mn) / (error_mx - error_mn);

  // Step 2: Rescale rank floats (float [0,1]) to state error units (float, [state units])
  tmp = (float32) qa_rank_flt * (qa_rank_mx - qa_rank_mn) + qa_rank_mn;

  // <JMG> standardized to use FRM_API_ROUND()
  qa_rank = (uint8)FRM_API_ROUND(tmp);

  return(qa_rank);

} // end :: l4c_tcf_qa_reclas() 


// ---------------------------------------------------------------------
// function  : int l4c_PostProcessGranuleCompleteness(ecoStateArrayObj_p);
// purpose   : Derive granule level summary statistics using qa_count
// inputs    : EcoStateArrayAdt *ecoStateArrayObj_p
// outputs   : n/a
// revised   : 2014-10-28T16:55:00
// author    : joe glassy
// notes     : Normally called one time only, by {TBD}.

/*
 (Lucas) proposed method for Assigning CompletenessOmission and DomainConsistency
  The metric QA/qa_count most closely aligns with the definition of “CompletenessOmission.” 
  I would add these up and divide by the number of land 9km grid cells multiplied by 81
  (to get the total number of land 1km grid cells in the denominator).
  For “DomainConsistency,” we could add up the “NEE bit” (bit 0 of QA/carbon_model_bitflag) 
  and divide by the number of land 9km grid cells.  
  Alternatively, I would advocate counting up the internal 1km sums used for
  checking the “NEE bit” 9km threshold criteria and dividing this grand sum
  by the number of land 9km grid cells multiplied by 81 (to get the total
  number of land 1km grid cells in the denominator), which would be more
  consistent with the use of QA/qa_count for “CompletenessOmission”, 
  in which case we are assured
  that  “DomainConsistency[%]”  =< “CompletenessOmission [%]”.
 
*/
int l4c_PostProcessGranuleCompleteness(VarFloat32Adt *varFlt32_Obj_p)
{
 SesCtl *sesCtl_p   = l4c_GetSesCtl();
 char fxName[256] = {0} ;
 
 strcpy(fxName,__func__);
 
 // Derive global "DomainConsistency" and "CompletenessOmission" for NEE
 xLogMsg("@A In(%s) varFlt domainConsistencySum_1km (%zu) completenessOmissionSum_1km (%zu)~\n",
    fxName,varFlt32_Obj_p->domainConsistencySum_1km, varFlt32_Obj_p->completenessOmissionSum_1km);
 
 pInv.qa_DomainConsistency    = (float32)varFlt32_Obj_p->domainConsistencySum_1km / (float32)EGv2_LDOM1KM * 100.0;
 pInv.qa_CompletenessOmission = (float32)varFlt32_Obj_p->completenessOmissionSum_1km / (float32)EGv2_LDOM1KM * 100.0;

 xLogMsg("@A In(%s) domainConsistencySum_1km %.4f completenessOmissionSum_1km %.4f~\n",
    fxName, pInv.qa_DomainConsistency, pInv.qa_CompletenessOmission );

 return 0 ;
    
} // end:: l4c_PostProcessGranuleCompleteness()


// ---------------------------------------------------------------------
// function  : int l4c_PostProcess_RMSE_Alloc_Memory()
// purpose   : Allocate memory for all diagnostic monitor 9km RMSE
//             float32 variables. These are NOT relevant to OPS processing,
//             and are used only within development tests. These include
//             (1) rmse "MEAN" and a set of (8) PFT wise RMSE by PFT variables
// inputs    : void TBD, will change as needed
// returns   : int, status 0 when successful
// outputs   : n/a
// revised   : 2016-01-04T14:46:00
// author    : joe glassy
// notes     : Called once, by {TBD}, each array is GRID_9KM_CELLCOUNT by sizeof(float32) bytes
int l4c_PostProcess_RMSE_Alloc_Memory(void)
{
 int status = 0 ;
 int ithPft = 0 ;
 char fxName[256] = {0};
  
 // fetch pointer to Pre and Post Processor image package container    
 PrePostImgPkgAdt *prePostAdt_p = l4c_GetPrePostImagePkgAdt();
    

 strcpy(fxName,(char *)__func__);
 xLogMsg("@A In(%s) Allocating RMSE variables (one mean, 8 PFT wise) ~\n",fxName);    

 // Allocate main RMSE MEAN 9km
 prePostAdt_p->rmse_mean_M09_p = (float32 *)memCalloc(EGv2_GLOBAL_9KM_PIXELS,sizeof(float32)) ;
 
 // Allocate each PFT wise RMSE 9km
 for(ithPft =0; ithPft < NPFT ; ithPft++)
 {
     
   prePostAdt_p->rmse_pftList_M09_p[ithPft] = (float32 *)memCalloc(EGv2_GLOBAL_9KM_PIXELS,sizeof(float32));
   
 } // end::ithPft loop iterating through (8) PFT indices
 
 return status ;
} // end:: l4c_PostProcess_RMSE_Alloc()


// ---------------------------------------------------------------------
// function  : int l4c_PostProcess_RMSE_Release_Memory(void)
// purpose   : Releases memory for all diagnostic, monitor-only
//             9km RMSE float32 variables, including
//             (1) rmse "MEAN" and a set of (8) PFT wise RMSE by PFT variables
// inputs    : void TBD, will change as needed
// returns   : int, status 0 when successful
// outputs   : n/a
// revised   : 2016-01-04T14:46:00
// author    : joe glassy
// notes     : Called once by {TBD}
int l4c_PostProcess_RMSE_Release_Memory(void)
{
 int status = 0 ;
 int ithPft = 0 ;
 char fxName[256] = {0};
 // fetch pointer to Pre and Post Processor image package container
 PrePostImgPkgAdt *prePostAdt_p = l4c_GetPrePostImagePkgAdt();

 strcpy(fxName,(char *)__func__);
 xLogMsg("@A In(%s) Releasing memory for RMSE variables (one mean, 8 PFT wise) ~\n",fxName);    

 // Release memory for each PFT wise RMSE 9km
 for(ithPft =0; ithPft < NPFT ; ithPft++)
 {
   memFree( prePostAdt_p->rmse_pftList_M09_p[ithPft] );
 }

  // Release memory for RMSE mean 9KM
  memFree(prePostAdt_p->rmse_mean_M09_p);

 return status ;
} // end:: l4c_PostProcess_RMSE_ReleaseMemory()



