// ---------------------------------------------------------------------
// module       : l4c_DAILY_PreProcessor.c (was orig l4c_PreProc1km_func.c
// purpose      : Implements L4C_DAILY pre-processor (DEFLATE) logic
//                on 1km,3km, and 9km 2D full grid inputs deflated to
//                1D sparse vectors
//                 1km 1D-sparse vector data to 9km 1D-sparse vector data
//  project      : SMAP L4C science development.
//  dependencies : HDF4, FRM_API etc
//  revised      : 2016-02-17, changed module name to l4c_DAILY_PreProcessor.c
//  from
//                 orig l4c_PreProc1km_func.c
//  revised      : 2015-07-30, changed to support new v7 FORK GEOS5 F/T at 3KM
//  or
//                 passive radiometer derived 3km F/T
//  revised      : 2015-05-27, changed to use RIP file dynamic lookups on
//  MIN,MAX for:
//                 nsbGetValueByKey(nsbObj_p,"L4C_INPUT_DATASET_SOILROOTZONE_M09_MIN",
//                 nvpBuf)
//
//  revised      : 2014-11-14 <JMG> changed incorrect fpar_qa_M01 changed to
//  fpar_qc_M01 at line 926 revised      : 2014-11-11T15:17, corrected datatypes
//  to be int16 for ft_calendar,soc_calendar
//                 in the SOC file, to properly represent {1..366} dayOfYear
//                 slots!
//  history      : 2014-11-06, <JMG>  when L4C_GPP_SOURCE = CLIM (forced
//                 use of only FparClimatology) added SKIP-LOGIC to avoid
//                 pre-processing fpar_M01,fpar_qc_M01 and/or NDVI
//                 since in this case these are not used or relevant.
//  history      : 2014-06-13T18:54, updated to use new frm_hdf5_ReadAttribute()
//                 call that omits the un-used obj_name param.
//  history      : 2014-05-16T13:00:00, NDVI integration started. When
//                 available, will be stored in viirsNdvi"
//  history      : 2014-05-09T10:34, added READ of (2) FLOAT32 arrays of 370,
//                 "soc_updated_datetime", and "ft_updated_datetime" to match
//                 the calendars in SMAP_L4_C_SOC.h5
//  history      : 2014-04-28T18:08:00, extend to pre-process FparClimatology
//                 dataset for current 8-day period. Note that the surface_flag
//                 3km MUST still be read IN as an input, so the deprecation
//                 applies only to WRITING it out!
//  history      : 2014-03-01T13:00, re-organized code in l4c_GridManager.c
//                 pulling out most of the pre-processor specific code to
//                 this module file.  Moved set_fillval_UUTA to the
//                 post-processor code since it is NOT used in pre proc DEFLATE.
//  author       : joe glassy, Lupine Logic Inc.

//  notes        : MASTER LIST of all variables that are PRE-PROCESSED here,
//                 noting that memory for each of these is allocated here
//                 with the variables carried forward intact to the
//                 main model.
/*
  ecoStateArrayObj_p->ftRolling_3km
  ecoStateArrayObj_p->ftRollingAge_3km
  ecoStateArrayObj_p->C[0].dom_1km
  ecoStateArrayObj_p->C[1].dom_1km
  ecoStateArrayObj_p->C[2].dom_1km
  ecoStateArrayObj_p->npp_sum.dom_1km

  metArrayObj_p->vpd
  metArrayObj_p->fpar
  metArrayObj_p->fpar_clim
  metArrayObj_p->fpar_qa
  metArrayObj_p->viirsNdvi
  metArrayObj_p->viirsNdvi_qc
  metArrayObj_p->qv2m
  metArrayObj_p->par
  metArrayObj_p->smrz
  metArrayObj_p->smtop
  metArrayObj_p->tsoil
  metArrayObj_p->ps_avg
  metArrayObj_p->t2m_avg
  metArrayObj_p->tmax
  metArrayObj_p->tmin
  metArrayObj_p->ft_state_um_3km
  metArrayObj_p->smrz_mn
  metArrayObj_p->smrz_mx

 count in PRE_PROC_DEFLATE_1KM theme is 24 TBD
  0 DEFLATE-CANDIDATE (fpar_M01)~
  1 DEFLATE-CANDIDATE (fpar_qc_M01)~
  2 DEFLATE-CANDIDATE (fpar_clim_M01_day001_per01)~
  3 DEFLATE-CANDIDATE (ndvi_M01)~
  4 DEFLATE-CANDIDATE (ndvi_qc_M01)~
  5 DEFLATE-CANDIDATE (FT_ROLLING_M03)~
  6 DEFLATE-CANDIDATE (FT_ROLLING_AGE_M03)~
  7 DEFLATE-CANDIDATE (SOC/SOC_C0_M01)~
  8 DEFLATE-CANDIDATE (SOC/SOC_C1_M01)~
  9 DEFLATE-CANDIDATE (SOC/SOC_C2_M01)~
 10 DEFLATE-CANDIDATE (SOC/NPP_SUM_M01)~
 11 DEFLATE-CANDIDATE (QV2M_M09_AVG)~
 12 DEFLATE-CANDIDATE (RADIATION_SHORTWAVE_DOWNWARD_FLUX)~
 13 DEFLATE-CANDIDATE (SM_ROOTZONE_WETNESS)~
 14 DEFLATE-CANDIDATE (SM_SURFACE_WETNESS)~
 15 DEFLATE-CANDIDATE (SOIL_TEMP_LAYER1)~
 16 DEFLATE-CANDIDATE (SURFACE_PRESSURE)~
 17 DEFLATE-CANDIDATE (T2M_M09_AVG)~
 18 DEFLATE-CANDIDATE (T2M_M09_MAX)~
 19 DEFLATE-CANDIDATE (T2M_M09_MIN)~
 20 DEFLATE-CANDIDATE (FT_STATE_UM_M03)~
 21 DEFLATE-CANDIDATE Nature run smrz min  layer
 22 DEFLATE-CANDIDATE Nature run smrz max layer


 */

// ---------------------------------------------------------------------

#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"

// --------------------------------------------------------------------
// function : l4c_GetThemePropertySummary()
// purpose  : For a given HDF5 input dataset ('childName_p') whose
//            key properties are stored in PkgMgr theme 'themeLabel_p',
//            return a set of the properties listed below.
// inputs   : const char *hdf5_datasetName_p -- fully qualified name of HDF5
// dataset itself
//            int *childDatatype,     -- HDF4 /TIV style datatype identifier
//            const, DFNT_FLOAT32 etc int *childResolution    -- assigned grid
//            resolution {1,3,9} for 1km,3km or 9km size_t *child_nElements --
//            n. of pixels, usually as 1D sparse vector, and NOT in 2D grid,
//            size_t *objSiz          -- n. of bytes per pixel
// notes    : we store the file_id for each dataset in "Owner"
// pkgMgr_GetObjOwner(pObj_p) revised  : 2014-05-19T15:21:00, all dataset
// properties now stored in
//            theme "PRE_PROC_DEFLATE_1KM" now, simplying property lookups
// returns  : pkgObj's ID

int l4c_GetThemePropertySummary(const char *hdf5_datasetName_p,
                                int *childDatatype, int *childResolution,
                                size_t *child_nElements, size_t *objSiz) {
  SesCtl *sesCtl_p = l4c_GetSesCtl();
  PkgMgr *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
  PkgTheme *pThme_p = NULL;
  PkgObj *pObj_p = NULL;
  int dType = 0;
  char fxName[256] = {0};

  // all datasets have their properties stored in an entry in one theme
  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "PRE_PROC_DEFLATE_1KM");

  // in theme, lookup entry by dataset's name
  pObj_p = pkgMgr_GetObjByName(pThme_p, hdf5_datasetName_p);

  // fetch n. elements assigned to this (HDF5 dataset's) PkgObj
  *child_nElements = (size_t)pkgMgr_GetObjNelem(pObj_p);

  // fetch grid resolution value
  *childResolution = (int)pkgMgr_GetObjResolution(pObj_p);

  // fetch pixel datatype identifier (sync'd HDF4 with TIV now)
  dType = pkgMgr_GetObjDatatype(pObj_p);
  *childDatatype = dType;

  // Derive size in bytes of one pixel
  *objSiz = DFKNTsize(*childDatatype);

  // build a custom fxName
  sprintf(fxName, "<%s Dataset(%s) Type(%d)>", (char *)__func__,
          hdf5_datasetName_p, dType);

  l4c_EventCompletionMsg(fxName);

  // return ID as well
  return pObj_p->pkObjId;

} // end:: l4c_GetThemePropertySummary()

// ---------------------------------------------------------------
// function : l4c_PrepDeflate_AssignDatasetProperties()
// inputs   : PkgMgr *pkgMgr_p -- ptr to master pkgMgr overseer
//          : NbsObj *nsbObj_p -- ptr to master NSB overseer, to fetch
//            externally defined dataset names from the session RIP file
// revised  : 2015-07-29, change to support new v7 FORK GEOS5 9KM F/T also
// returns  : int, status
// dependencies : tihs routine is dependent on the RIP file containing
//            ALL L4_C_INPUT_DATASET_{var} definitions.
// purpose  : This function assigns ALL relevant dataset properties into the
//            different PkgMgr themes, to manage HDF5 dataset properties.
//            The common theme "PRE_PROC_DEFLATE_1KM" includes JUST the
//            SUBSET of INPUT HDF5 full 2D 1km datasets that need to
//            be DEFLATED from 1km to 9km, noting that the
//            FILE-LEVEL PkgMgr themes also store each INPUT DATASET's
//            relevant properties. Note that it is entirely permissible
//            for a given HDF5 dataset to have its properties defined in
//            more than one theme, since the whole point of PkgMgr themes
//            is to help handle special cases.
// notes    : CALLED BY: l4c_GridManager() ahead of call to pre-processor
// notes    : Note we need to store the HDF5 file_id (handle) in each
//            dataset's pkgObj "ID" member field since we'll need it later

int l4c_PrepDeflate_AssignDatasetProperties(PkgMgr *pkgMgr_p,
                                            NsbObj *nsbObj_p) {
  SesCtl *sesCtl_p = l4c_GetSesCtl();
  PkgTheme *pThme_p = NULL;
  PkgObj *pObj_p = NULL;
  char nvpBuf[FILENAME_MAX] = {0}; // reusable buffer for NSB "Get" retrievals
  char *dsetName_p = NULL;         // future: uses nsbObj_p to supply all these!
  char fxName[256] = {0};

  strcpy(fxName, (char *)__func__);

  xLogMsg("@A In(%s) Assigning dataset properties~\n", fxName);

  // -----------------------------------------------------------------
  // THEME: PRE_PROC_DEFLATE_1KM single them, define dataset properties
  // for ALL the datasets that require pre-processing (Deflate from 1km to 9km)
  // -----------------------------------------------------------------
  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "PRE_PROC_DEFLATE_1KM");

  // -----------------------------------------------------------------
  // FPAR climatology fpar_clim_dayDDD_perPP daily dataset. Assignment of
  //  this (name) is handled as a special case since it is dynamic (calendar
  //  driven) based on current model 'day-of-year' being run now
  dsetName_p = sesCtl_p->fparClimDatasetname;
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputFparClimatologyPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->fparClima_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  // CAUTION: this count is the n. pix of FUTURE 1D vec not current 2D vec
  pkgMgr_SetObjNelem(pObj_p, EGv2_LDOM1KM);
  pkgMgr_SetObjFillvalue(pObj_p, 254);

  // -----------------------------------------------------------------
  // ndvi_M01 NDVI 16-day and its QC, datasets
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_NDVI", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputFparPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->fpar_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  // CAUTION: this count is the n. pix of FUTURE 1D vec not current 2D vec
  pkgMgr_SetObjNelem(pObj_p, EGv2_LDOM1KM);
  pkgMgr_SetObjFillvalue(pObj_p, FPAR_FILL_VALUE);

  // ndvi_qc_M01 16-day QC
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_NDVI_QC", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputFparPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->fpar_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  // CAUTION: this count is the n. pix of FUTURE 1D vec not current 2D vec
  pkgMgr_SetObjNelem(pObj_p, EGv2_LDOM1KM);
  pkgMgr_SetObjFillvalue(pObj_p, FPAR_FILL_VALUE);

  // -----------------------------------------------------------------
  // fpar_M01 FPAR 8DAY dataset
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_FPAR", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputFparPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->fpar_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  // CAUTION: this count is the n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjNelem(pObj_p, EGv2_LDOM1KM);
  pkgMgr_SetObjFillvalue(pObj_p, FPAR_FILL_VALUE);

  // -----------------------------------------------------------------
  // fpar_qc_M01 FPAR QC_M01 8-day composite dataset
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_FPAR_QC", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputFparPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->fpar_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  // CAUTION: Nelem here is n. pix of future 1D vec, and NOT of current 2D vec
  pkgMgr_SetObjNelem(pObj_p, EGv2_LDOM1KM);
  pkgMgr_SetObjFillvalue(pObj_p, FPAR_FILL_VALUE);

  // -----------------------------------------------------------------
  // FT_ROLLING_M03 3km dataset
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_FT_ROLLING", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputSOCPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->soc_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8);
  pkgMgr_SetObjResolution(pObj_p, 3);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM3KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, 255);

  // -----------------------------------------------------------------
  // FT_ROLLING_AGE_M03 3km dataset
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_FT_ROLLING_AGE", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputSOCPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->soc_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8);
  pkgMgr_SetObjResolution(pObj_p, 3);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM3KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, 255);

  // -----------------------------------------------------------------
  // SOC_C0_M01 1km dataset
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOC_C0", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputSOCPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->soc_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM1KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // -----------------------------------------------------------------
  // SOC_C1_M01 1km dataset
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOC_C1", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputSOCPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->soc_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM1KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // -----------------------------------------------------------------
  // SOC_C2_M01 1km dataset
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOC_C2", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputSOCPath);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->soc_file_id);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM1KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // -----------------------------------------------------------------
  // NPP_SUM_M01 1km dataset
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_NPP_SUM", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputSOCPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->soc_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM1KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // <JMG> 2015-05-27, verified, names of smrz min,max ARE now dynamically read
  // from RIP file L4C_INPUT_DATASET_SOILROOTZONE_M09_MIN, MAX

  // -----------------------------------------------------------------
  // SMAP_NatureRun minimum
  dsetName_p = nsbGetValueByKey(
      nsbObj_p, "L4C_INPUT_DATASET_SOILROOTZONE_M09_MIN", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputSOCPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->soc_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  // <JMG> 2014-09-27, found, corrected bug where smrz_mn,mx WAS 1km but
  // should have been 9 since it is M09
  pkgMgr_SetObjResolution(pObj_p, 9); // changed from 1 to 9km
  pkgMgr_SetObjRank(pObj_p, 2);
  // n. pix applies to future 1D vector form, not current 2D grid form
  pkgMgr_SetObjNelem(pObj_p, EGv2_LDOM9KM);
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // -----------------------------------------------------------------
  // SMAP_NatureRun maximum
  dsetName_p = nsbGetValueByKey(
      nsbObj_p, "L4C_INPUT_DATASET_SOILROOTZONE_M09_MAX", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputSOCPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->soc_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  // <JMG> 2014-09-27, found, corrected bug where smrz_mn,mx was 1KM
  // but should have been 9km
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // -----------------------------------------------------------------
  // MET QV2M_M09_AVG dataset
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_QV2M_M09_AVG", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, "QV2M_M09_AVG");
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // -----------------------------------------------------------------
  // MET RADIATION_SHORTWAVE_DOWNWARD_FLUX dataset
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_RAD_SW_DOWN", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // -----------------------------------------------------------------
  // "SM_ROOTZONE_WETNESS" dataset
  dsetName_p = nsbGetValueByKey(
      nsbObj_p, "L4C_INPUT_DATASET_SM_ROOTZONE_WETNESS", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // -----------------------------------------------------------------
  // "SM_SURFACE_WETNESS" dataset, L4C_INPUT_DATASET_SM_SURFACE_WETNESS
  dsetName_p = nsbGetValueByKey(nsbObj_p,
                                "L4C_INPUT_DATASET_SM_SURFACE_WETNESS", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // -----------------------------------------------------------------
  // "SOIL_TEMP_LAYER1" dataset
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOIL_TEMP_LAYER1", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // -----------------------------------------------------------------
  // "FT_STATE_UM_M03" uint8 dataset (stores 4 bits extracted from uint16
  // surface_flag)

  // "FT_STATE_UM_M09" uint8 dataset, <JMG> 2015-07-30,added to support new
  // GEOS5 3KM F/T (stores 4 bits extracted from uint16 surface_flag)

  // <JMG> 4/24/2014, still must allow for read of surface_flag BUT DO SUPPRESS
  // output dsetName_p =
  // nsbGetValueByKey(nsbObj_p,"L4C_INPUT_DATASET_SURFACE_FLAG_M03", nvpBuf);
  // <JMG> 2014-08-07, replace L4C_INPUT_DATASET_SURFACE_FLAG_M03 with
  // L4C_INPUT_DATASET_FT_STATE_UM_M03
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_FT_STATE_UM_M03", nvpBuf);

  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  // now: ft_state_um_3km, so changed datatype to uint8 from uint16
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8);
  pkgMgr_SetObjResolution(pObj_p, 3);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM3KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, BYTE_FILL_VALUE);

  // -----------------------------------------------------------------
  // "SURFACE_PRESSURE" dataset
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SURFACE_PRESSURE", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // -----------------------------------------------------------------
  // "T2M_M09_AVG" dataset
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_T2M_M09_AVG", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // -----------------------------------------------------------------
  // "T2M_M09_MAX" dataset
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_T2M_M09_MAX", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // -----------------------------------------------------------------
  // "T2M_M09_MIN" dataset
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_T2M_M09_MIN", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // -----------------------------------------------------------------
  // IMPLEMENT DATASET PROPERTY ASSIGNMENTS FOR ALL File-Level theme(s)
  // -----------------------------------------------------------------

  // -----------------------------------------------------------------
  // SMAP_L4_C_Ancillary  THEME
  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "SMAP_L4_C_Ancillary");

  // TBD: fill in the RIP keywords for each dataset again below...
  // dsetName_p = nsbGetValueByKey(nsbObj_p,"L4C_INPUT_DATASET_XXXXXX", nvpBuf);

  // Custom 1D-sparse vector form of global L4_C pixel-filter, used only
  // special "point" (or continent etc) subset tasks, not normally used in
  // production, so ALL pixel mask slots are set to '1' to enable use of.
  pObj_p = pkgMgr_GetObjByName(pThme_p, "L4C_M01land_PixFilter_uint8");

  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputAncPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->anc_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(pObj_p, EGv2_LDOM1KM);
  pkgMgr_SetObjFillvalue(pObj_p, BYTE_FILL_VALUE);

  // Custom full 2D grid form of global L4_C pixel-filter, used only
  // special "point" (or continent etc) subset tasks, not normally used in
  // production, so ALL pixel mask slots are set to '1' to enable use of.
  pObj_p = pkgMgr_GetObjByName(pThme_p, "L4C_M01_PixFilter_uint8");
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputAncPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->anc_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(pObj_p, EGv2_LDOM1KM); // Marks with 1D-sparse vec count
  pkgMgr_SetObjFillvalue(pObj_p, BYTE_FILL_VALUE);

  // -------------------------------------------------------------------------
  // FPAR CLIMATOLOGY THEME "SMAP_L4_C_FparClimatology"
  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "SMAP_L4_C_FparClimatology");

  pObj_p = pkgMgr_GetObjByIndex(pThme_p, 0);
  // SPECIAL EXCEPTION: Assign late-bound name for Fpar Climatology dataset
  // that varies by composite period
  pkgMgr_SetObjName(pObj_p, sesCtl_p->fparClimDatasetname);

  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputFparClimatologyPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->fparClima_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  // CAUTION: this count is the n. pix of FUTURE 1D vec not current 2D vec
  pkgMgr_SetObjNelem(pObj_p, EGv2_LDOM1KM);
  pkgMgr_SetObjFillvalue(pObj_p, FPAR_FILL_VALUE);

  // -------------------------------------------------------------------------
  // FPAR 8DAY THEME "SMAP_L4_C_FPAR_8DAY"
  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "SMAP_L4_C_FPAR_8DAY");
  pObj_p = pkgMgr_GetObjByName(pThme_p, "fpar_M01");
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputFparPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->fpar_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  // CAUTION: this count is the n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjNelem(pObj_p, EGv2_LDOM1KM);
  pkgMgr_SetObjFillvalue(pObj_p, FPAR_FILL_VALUE);

  pObj_p = pkgMgr_GetObjByName(pThme_p, "fpar_qc_M01");
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputFparPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->fpar_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  // CAUTION: Nelem here is n. pix of future 1D vec, and NOT of current 2D vec
  pkgMgr_SetObjNelem(pObj_p, EGv2_LDOM1KM);
  pkgMgr_SetObjFillvalue(pObj_p, FPAR_FILL_VALUE);

  // -------------------------------------------------------------------------
  // <JMG> 2014-08-07, be sure to add and support NDVI here!
  // FPAR 8DAY THEME "NDVI_16DAY" as available

  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "SMAP_L4_C_FPAR_8DAY");
  pObj_p = pkgMgr_GetObjByName(pThme_p, "ndvi_M01");
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputFparPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->fpar_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  // CAUTION: this count is the n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjNelem(pObj_p, EGv2_LDOM1KM);
  pkgMgr_SetObjFillvalue(pObj_p, BYTE_FILL_VALUE);

  pObj_p = pkgMgr_GetObjByName(pThme_p, "ndvi_qc_M01");
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputFparPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->fpar_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  // CAUTION: Nelem here is n. pix of future 1D vec, and NOT of current 2D vec
  pkgMgr_SetObjNelem(pObj_p, EGv2_LDOM1KM);
  pkgMgr_SetObjFillvalue(pObj_p, BYTE_FILL_VALUE);

  // -------------------------------------------------------------------------
  // SOC THEME
  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "SMAP_L4_C_SOC");

  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_FT_ROLLING", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputSOCPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->soc_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8);
  pkgMgr_SetObjResolution(pObj_p, 3);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM3KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, BYTE_FILL_VALUE);

  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_FT_ROLLING_AGE", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputSOCPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->soc_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8);
  pkgMgr_SetObjResolution(pObj_p, 3);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM3KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, BYTE_FILL_VALUE);

  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOC_C0", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputSOCPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->soc_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM1KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOC_C1", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputSOCPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->soc_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM1KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOC_C2", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputSOCPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->soc_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM1KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_NPP_SUM", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputSOCPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->soc_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 1);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM1KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // <JMG> 2014-08-07 added...
  dsetName_p = nsbGetValueByKey(
      nsbObj_p, "L4C_INPUT_DATASET_SOILROOTZONE_M09_MIN", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputSOCPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->soc_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  // CAUTION: this count is the n. pix of future 1D vec NOT the current 2D vec!!
  pkgMgr_SetObjNelem(pObj_p, EGv2_LDOM9KM);
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  dsetName_p = nsbGetValueByKey(
      nsbObj_p, "L4C_INPUT_DATASET_SOILROOTZONE_M09_MAX", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputSOCPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->soc_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  // CAUTION: this count is the n. pix of future 1D vec NOT the current 2D vec!!
  pkgMgr_SetObjNelem(pObj_p, EGv2_LDOM9KM);
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // -------------------------------------------------------------------------
  // MET THEME
  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "SMAP_L4_C_MET");

  pObj_p = pkgMgr_GetObjByName(pThme_p, "QV2M_M09_AVG");
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  pObj_p = pkgMgr_GetObjByName(pThme_p, "RADIATION_SHORTWAVE_DOWNWARD_FLUX");
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  pObj_p = pkgMgr_GetObjByName(pThme_p, "SM_ROOTZONE_WETNESS");
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  pObj_p = pkgMgr_GetObjByName(pThme_p, "SM_SURFACE_WETNESS");
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  pObj_p = pkgMgr_GetObjByName(pThme_p, "SOIL_TEMP_LAYER1");
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // <JMG> 2015-07-30, add support for 3KM GEOS5 FP TS derived F/T
  // <JMG> 2014-08-07T12:26, NEW, replaced uint16 SUIRFACE_FLAG_M03 with uint8
  // FT_STATE_UM_M03 here OLD: <JMG> 4/24/2014, still must allow for read of
  // surface_flag BUT DO SUPPRESS output

  // L4C_INPUT_DATASET_FT_STATE_UM_M03        = FT_STATE_UM_M03
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_FT_STATE_UM_M03", nvpBuf);
  pObj_p = pkgMgr_GetObjByName(pThme_p, dsetName_p);
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8);
  pkgMgr_SetObjResolution(pObj_p, 3);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM3KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, 254);

  pObj_p = pkgMgr_GetObjByName(pThme_p, "SURFACE_PRESSURE");
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  pObj_p = pkgMgr_GetObjByName(pThme_p, "T2M_M09_AVG");
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  pObj_p = pkgMgr_GetObjByName(pThme_p, "T2M_M09_MAX");
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  pObj_p = pkgMgr_GetObjByName(pThme_p, "T2M_M09_MIN");
  pkgMgr_SetObjGenericPath(pObj_p, sesCtl_p->inputMetPath);
  pkgMgr_SetObjOwner(pObj_p, sesCtl_p->met_file_id);
  pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_FLOAT32);
  pkgMgr_SetObjResolution(pObj_p, 9);
  pkgMgr_SetObjRank(pObj_p, 2);
  pkgMgr_SetObjNelem(
      pObj_p, EGv2_LDOM9KM); // n. pix of future 1D vec not current 2D vec
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  l4c_EventCompletionMsg(fxName);

  return 0;
} // end:: l4c_PrepDeflate_AssignDatasetProperties()

// ---------------------------------------------------------------
// function     : l4c_PreProc_Calculate_VPD(metArrayObj_p)
// purpose      : allocate vpd array and then calculate pixel wise VPD
//                to store in 1D-sparse vector space metArrayObj_p->vpd array
// inputs       : MetArrayAdt *metArrayObj_p --ptr to container of
//                all met variables including VPD
// returns      : size_t, count of pixels processed
// notes        :l4c_calc_vpd () is located in l4c_sci.c
size_t l4c_PreProc_Calculate_VPD(MetArrayAdt *metArrayObj_p) {
  size_t pix = 0;
  time_t secNow = time(0);

  xLogMsg("@A In(%s) allocating and calculating VPD at (%s)~\n", __func__,
          ctime(&secNow));

  // ALLOCATE VPD 1D-sparse vector form, prior to calc
  metArrayObj_p->vpd = memCalloc((size_t)EGv2_LDOM9KM, sizeof(float32));
  memObjSetName(metArrayObj_p->vpd, "metArrayObj_p->vpd");

  // calculate VPD as a special case, since it was not inherited..
  // using 1D-sparse vector variant of t2m_avg, qv2m_p, and ps_avg_p
  for (pix = 0; pix < EGv2_LDOM9KM; pix++) {
    // float32 l4c_calc_vpd(const float32 tavg,const float32 qv2m,const float32
    // ps)
    metArrayObj_p->vpd[pix] =
        l4c_calc_vpd(metArrayObj_p->t2m_avg[pix], metArrayObj_p->qv2m[pix],
                     metArrayObj_p->ps_avg[pix]);
  } // end:: for pix...

  l4c_EventCompletionMsg(__func__);

  return pix;

} // end:: l4c_PreProc_Calculate_VPD()

// --------------------------------------------------------------------
// function : l4c_PreProc_Read_2D_Image()
// purpose  : Reads 2D full grid image 'datasetName' into a
//          : reusable 2D input buffer, and performs
//          : one-time-only configuration prior to running
//          : the pre-processor DEFLATE. In SOC, reads the
//            (2) CORRECTED DATATYPE int16 370 slot calendars, "ft_calendar" and
//            "soc_calendar"
// inputs   : const hid_t curr_file_id -- the open HDF5 file_id owning dataset
// named 'datasetName_p'
//            char *datasetName_p -- the name of the dataset to pre-process
//            char *callerLbl_p -- a callers label
// returns  :void
// history  : 2014-05-19T15:24, simplified by removing string tables for
//            theme name, since we store all dataset properties now in
//            a single theme, "PRE_PROC_DEFLATE_1KM" now.
// notes    : To be fully qualified, a datasetName_p should have any "GROUP"
//            it belong to prepended to it, e.g. /SOC/SOC_myVar
void l4c_PreProc_Read_2D_Image(const hid_t curr_file_id, char *datasetName_p,
                               char *callerLbl_p)

{
  SesCtl *sesCtl_p = l4c_GetSesCtl();
  AncImagePkgAdt *ancImagePkgObj_p = l4c_GetAncImagePkgAdt();

  PkgMgr *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
  PkgTheme *pThme_p =
      NULL; // ptr to the THEME that owns this variable, among others
  PkgObj *pObj_p = NULL; // ptr to this variables own PkgObj object

  char *inputPath_p = NULL; // ptr to HDF5 file's path we'll read dataset from

  void *input_M01_2D_p =
      NULL; // ptr to common,re-useable 1KM 2D full grid buffer

  int32 hdf4Datatype = 0;
  int hRank = 0;
  int rastResol = 0;
  time_t secNow = 0;

  size_t nPix1km = EGv2_GLOBAL_1KM_PIXELS;
  size_t n_1DSparsePixels = 0;
  size_t siz = 0;

  hid_t hDtype;
  hid_t dcpl;
  hid_t dspace;
  hid_t dataset_id;
  hsize_t dims[4];
  herr_t ret = 0;
  char fxName[256] = {0}; // shields __func__
  // read L4_C_MET_FT_SOURCE string as reported by L4CMET for verification
  char ftSourceMethodStr_RptByMet[32] = {0};

  secNow = time(0);

  strcpy(fxName, (char *)__func__);

  // first fetch pointer to shared big global 1km full-grid input buffer
  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "PRE_PROC_DEFLATE_1KM");

  // fetch common large 1KM buffer sized to fit any datatype
  input_M01_2D_p =
      pkgMgr_GetThemes_ChildDataByName(pThme_p, "PREPROC_1KM_BUFFER");

  // RESET the re-usable INPUT 1KM 2D FULL GRID buffer for this raster
  // note we re-use this buffer for ANY INPUT datatype that fits in 4 bytes,
  // including uint8, uint16, and float32.
  memset(input_M01_2D_p, '\0', (nPix1km * sizeof(float32)));

  // Given this theme, now fetch the dataset's name as present in
  // its parent HDF5 file
  pObj_p = pkgMgr_GetObjByName(pThme_p, datasetName_p);
  // <JMG> 4/28/2014, added more tests on pObj_p in case it fails
  if (pObj_p == NULL) {
    xLogMsg("@F In(%s) Dataset(%s)'s entry could NOT be found!\n",
            datasetName_p);
  }

  // Retrieve this input dataset's file path
  inputPath_p = pkgMgr_GetObjGenericPath(pObj_p);

  // CHECK -- pre-qualify existence of HDF5 file at this path!!
  if (!frm_ioDoesFileExist(inputPath_p)) {
    xLogMsg(
        "@F (%s) ~ SERIOUS ERROR cannot find input file(%s) for DEFLATE OP!\n",
        fxName, inputPath_p);
  }

  // <JMG> 2014-09-28, VERIFY if this is still relevant and correct
  // Be sure to store the dataset_id in objects ID
  // pkgMgr_SetObjId(pObj_p,dataset_id);

  // xLogMsg("@A DIAG In(%s) HDF5-DatasetInfo: (%s) file_id(%d)~\n",
  //  fxName, datasetName_p, curr_file_id);

  // FOR READ I/O switch on native datatype. This is cumbersome but
  // for now, necessary.

  // Note that HDF5's H5T_xxx types do NOT apparently behave as
  //  integer resolvable macros or enums, so this complicates the
  //  generalization of type-dependent handling considerably...
  //  Eventually this will likely be refactored to use a switch(h4DatatypeId)...

  //  TBD: replace with calls to
  //  nsbGetValueByKey(nsb_p,"L4C_INPUT_DATASET_FPAR_M01",nvpBuf) the fpar_clim
  //  entry here is just a wildcard (see use of strstr()

  if (strcmp(datasetName_p, "fpar_M01") == 0 ||
      strcmp(datasetName_p, "fpar_qc_M01") ==
          0 || // <JMG> 2014-11-14 incorrect fpar_qa_M01 changed to fpar_qc_M01
      strstr(datasetName_p, "fpar_clim_M01_day") !=
          NULL || // <JMG> added 4/22/2014
      strcmp(datasetName_p, "FT_ROLLING_M03") == 0 ||
      strcmp(datasetName_p, "FT_ROLLING_AGE_M03") == 0) {

    ret = frm_hdf5_ReadData(curr_file_id, H5T_NATIVE_UINT8, datasetName_p,
                            input_M01_2D_p, callerLbl_p);

  } else if (strcmp(datasetName_p, "ndvi_M01") == 0) {

    ret = frm_hdf5_ReadData(curr_file_id, H5T_NATIVE_UINT8, datasetName_p,
                            input_M01_2D_p, callerLbl_p);

  } else if (strcmp(datasetName_p, "ndvi_qc_M01") == 0) {

    ret = frm_hdf5_ReadData(curr_file_id, H5T_NATIVE_UINT8, datasetName_p,
                            input_M01_2D_p, callerLbl_p);

  } else if (strcmp(datasetName_p, "FT_STATE_UM_M03") == 0) {
    // <JMG> 2014-08-07T12:28, replace uint16 SURFACE_FLAG_M03 with uint8
    // FT_STATE_UM_M03
    ret = frm_hdf5_ReadData(curr_file_id, H5T_NATIVE_UINT8, datasetName_p,
                            input_M01_2D_p, callerLbl_p);

    // <JMG> 2015-09-06 also read and report out the MET file
    //  attribute "L4_C_MET_FT_SOURCE"
    frm_hdf5_ReadAttribute(curr_file_id, &hDtype, "L4_C_MET_FT_SOURCE",
                           (char *)ftSourceMethodStr_RptByMet);

    xLogMsg("@A In(%s) L4CMET reports FT_SOURCE Method (%s) is used ~\n",
            fxName, ftSourceMethodStr_RptByMet);

  } else {
    // <JMG> 2014-09-04, all other named float32 images are read here
    ret = frm_hdf5_ReadData(curr_file_id, H5T_NATIVE_FLOAT, datasetName_p,
                            input_M01_2D_p, callerLbl_p);
  }

  // Assert, in: AncImagePkgAdt:
  // set 1 in slot (1..366) as each day-of-year is processed
  //  SOC_CalendarCheck[L4C_N_CALENDAR_SLOTS] ;
  // set 1 in slot (1..366) as each day-of-year is processed
  //  FT_CalendarCheck[L4C_N_CALENDAR_SLOTS]  ;
  // <JMG> 2014-05-09T10:33:00, added to exactly
  //  match SMAP_L4_C_SOC.h5 calendar check-offs

  // DO THIS ONCE ONLY, triggered here when hitting the FT_ROLLING_M03,
  // fetch the "soc_calendar" (uint8) attribute array from the soc_file_id
  if (strcmp(datasetName_p, "FT_ROLLING_M03") == 0) {
    // read the soc_calendar[] uint8 array.
    // --assume buffer (FT_CalendarCheck[]) is large enough
    frm_hdf5_ReadAttribute(curr_file_id, &hDtype, "soc_calendar",
                           (uint8 *)ancImagePkgObj_p->SOC_CalendarCheck);

    // fetch FT calendar for period up to today
    frm_hdf5_ReadAttribute(curr_file_id, &hDtype, "ft_calendar",
                           (uint8 *)ancImagePkgObj_p->FT_CalendarCheck);

#ifdef USING_NEW_SOC
    // also read in 2 new int16 CUMULATIVE calendar dayofyear slot arrays for
    // cross-check
    frm_hdf5_ReadAttribute(curr_file_id, &hDtype, "soc_calendar_cumulative",
                           (int16 *)ancImagePkgObj_p->SOC_CalendarCumulative);
    // fetch FT calendar for period up to today
    frm_hdf5_ReadAttribute(curr_file_id, &hDtype, "ft_calendar_cumulative",
                           (int16 *)ancImagePkgObj_p->FT_CalendarCumulative);
#endif
  }

  // <JMG> curr_file_id intentionally left open, closed now at end of session

  l4c_EventCompletionMsg(fxName);

} // end:: int l4c_PreProc_Read_2D_Image()

// --------------------------------------------------------------------
// function : l4c_PreProc_Setup(()
// purpose  : Allocates shared 1km 2D buffer, sets up various metadata
//            and dataset properties prior to running the pre-process
// inputs   : PkgMg *pkgMgr_p, MetArrayAdt *metArrayObj_p
// outputs  :return populated list of target HDF5 datasets to DEFLATE
// revision : 2014-04-22T11:21:00, v5.0.1 add FparClimatology to collection
//            of datasets to run through (inflate) preprocessor
// revision : 2014-03-14, D5.0 delivery
// returns  : count of datasets provisioned.
// notes    : CALLED BY
// notes    : added (2) SMRZ layers omitted earlier
//             MERRA_climate_M09_gwetroot_min_flt32" and
//             MERRA_climate_M09_gwetroot_min_flt32")
int l4c_PreProc_Setup(PkgMgr *pkgMgr_p, const int nPreProcVariables) {
  // SesCtl *sesCtl_p= l4c_GetSesCtl();
  NsbObj *nsbObj_p = NULL;
  PkgTheme *pThme_p; // ptr to (parent), the Theme itself
  PkgObj *pObj_p;    // ptr to child elements of this Themes collection
  void *preProc_M01_image_p;
  size_t dimsAry[4] = {0};
  size_t nPix1km = EGv2_GLOBAL_1KM_PIXELS;
  char fxName[256] = {0}; // shields __func__

  strcpy(fxName, (char *)__func__);

  // ----------------------------------------------------------
  // STEP 1: allocate, read the (2) {row,col}
  // "MCD12Q1_M09land_col_uint16", "MCD12Q1_M09land_row_uint16"
  l4c_ReadAnc_M09land_row_col_9km();

  // ----------------------------------------------------------
  // STEP 2: Create a single re-usable 1k 2D FULL GRID raster
  // to successively fetch the larger 1km images into
  dimsAry[0] = EGv2_GLOBAL_1KM_NS_PIXELS; // n. rows
  dimsAry[1] = EGv2_GLOBAL_1KM_EW_PIXELS; // n. columns

  // First, fetch the Theme: re-usable 1km 2D full grid array
  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "PRE_PROC_DEFLATE_1KM");

  // Under Theme "PRE_PROC_DEFLATE_1KM" fetch ptr to "PREPROC_1KM_BUFFER"
  pObj_p = pkgMgr_GetObjByName(pThme_p, "PREPROC_1KM_BUFFER");

  // Create a single, large, anonymous, journaled M01 (1km 2D) memory block
  preProc_M01_image_p = (float32 *)memCalloc(nPix1km, sizeof(float32));
  // Provide re-usable buffer with a distinct name
  memObjSetName(preProc_M01_image_p, "PREPROC_1KM_BUFFER");

  // Bind the common 1K M01 re-usable array to this Theme -->
  // "PRE_PROC_DEFLATE_1KM"
  pkgMgr_BindObjByName(pThme_p, "PREPROC_1KM_BUFFER", preProc_M01_image_p);

  // Use this PkgObj's methods to to assign key properties
  // NOTE: 2/27/2014, TIV datatypes now sync'd with HDF5 type identifiers
  // so, IsFloat32 == DFNT_FLOAT32 ,etc.
  pkgMgr_SetObjDatatype(pObj_p, IsFloat32);
  pkgMgr_SetObjNelem(pObj_p, EGv2_GLOBAL_1KM_PIXELS);
  pkgMgr_SetObjRank(pObj_p, 2); // must set rank >0 prior to dims!!
  pkgMgr_SetObjDims(pObj_p, dimsAry);
  pkgMgr_SetObjFillvalue(pObj_p, -9999.0);

  // fetch pointer to NSB name-value-pair mgr (overseer) to provide
  // access to external RIP file name value pair pre-parsed definitions
  nsbObj_p = l4c_GetNsbOverseer();

  // Assign all dataset PROPERTIES into their PkgObj
  l4c_PrepDeflate_AssignDatasetProperties(pkgMgr_p, nsbObj_p);

  // return count of HDF5 datasets to be pre-processed (DEFLATE)
  return nPreProcVariables;

} // end::l4c_PreProc_Setup()

// --------------------------------------------------------------------
// function : l4c_PreProcAllocate_1DSparseVector()
// purpose  : Allocate a reusable, 1D sparse vector memory for the current
// variable
//            processed, storing the OUTPUT (result) of DEFLATE operation.
// inputs   : const char *h5datasetName_p -- name of dataset to init
// returns  : void *, ptr to 1D sparse vector, OUTPUT OF DEFLATE, allocated
//            for current theme's variable, e.g. {fpar_M01,fpar_qa_M01,
//            SOC_C0,C1,C2,npp_sum etc}
// notes    : CALLED BY: l4c_PreProc_ByLayerDriver()
// notes    : We re-use a single 1KM output buffer sized to fit the
//            greatest-common-denominator properties of all variables used.
//
void *l4c_PreProcAllocate_1DSparseVector(const char *h5datasetName_p) {
  SesCtl *sesCtl_p = l4c_GetSesCtl();
  PkgMgr *pkgMgr_p =
      sesCtl_p->l4cPkgMgr_p; // fetch pointer to session 'overseer'
  PkgTheme *pThme_p = NULL;
  PkgObj *pObj_p = NULL;

  size_t siz = 0;
  size_t n_1DSparsePixels = 0;
  int datatype = -1;
  char fxName[256] = {0};
  char dsetName[256] = {0};
  char themeLabel[128] = {0};
  char addrLbl[64] = {0};
  void *newOutput_1DSparseVec_p = NULL;

  strcpy(fxName, __func__);
  strcpy(themeLabel, "PRE_PROC_DEFLATE_1KM");

  // ---------------------------------------------------------------
  // STEP 1: Retrieve the 'theme' storing all variables to DEFLATE
  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, themeLabel);
  // ...retrieve ptr to (child) PkgObj object, assoc with this 'datasetName'
  pObj_p = pkgMgr_GetObjByName(pThme_p, h5datasetName_p);

  // Fetch its pre-defined datatype, and derive the per-pixel size (byte)
  // for this datatype
  datatype = pkgMgr_GetObjDatatype(pObj_p);
  siz = DFKNTsize(datatype);

  // TBD: GetObjNelem returns int, eventually will change this to size_t
  n_1DSparsePixels = (size_t)pkgMgr_GetObjNelem(pObj_p);

  if (datatype <= 0 || siz > sizeof(float64)) {
    xLogMsg(
        "@F FATAL In(%s) key(%s) Invalid scalarSize (%zu) for dataset %s~\n",
        (char *)__func__, pObj_p->pkObjKey, siz, h5datasetName_p);
  }

  // ---------------------------------------------------------------
  // STEP 2: allocate the specific object's DEST destination 1D-sparse buffer
  newOutput_1DSparseVec_p = memCalloc(n_1DSparsePixels, siz);

  // For easier verification, provide buffer a sensible name!
  strcpy(dsetName, h5datasetName_p);
  memObjSetName(newOutput_1DSparseVec_p, dsetName);
  // report match-up between the dsetName here and the DNA for later
  // verification
  sprintf(addrLbl, "%p", newOutput_1DSparseVec_p);

  // xLogMsg("@A In(%s) Alloc 1D 1km vec (%zu pix) for (%s) h5name(%s) DNA(%s)
  // type(%d) theme(%s)~\n",
  //  fxName, n_1DSparsePixels, dsetName, h5datasetName_p, addrLbl,datatype,
  //  themeLabel);

  // xLogMsg("@A In(%s) Alloc 1D 1km vec (%zu pix) for (%s) h5name(%s) type(%d)
  // ~ \n",
  //  fxName, n_1DSparsePixels, dsetName, h5datasetName_p, datatype);

  // NOTE: we may still need to store each datasets PRIMARY theme in a member
  //  of the "PRE_PROC_DEFLATE_1KM" theme's dataset level pkgObj in order to
  //  properly do this binding...however note that in the
  //  l4c_PreProc_Bind_1D_DeflatedVars() we ALSO do a direct bind of each
  //  dataset's 1D sparse vector to the internal array pointer it ultimately
  //  needs to attach to... so the binding done HERE is really just a
  //  "nice-to-have" not "must-have" binding since we never actually USE it.
  //  Furthermore, if we DO need to fetch a binding from a theme we could always
  //  retrieve the valid bindings done here in PRE_PROC_DEFLATE_1KM theme
  pkgMgr_BindObjByName(pThme_p, h5datasetName_p, newOutput_1DSparseVec_p);

  l4c_EventCompletionMsg(__func__);

  // ---------------------------------------------------------------
  // STEP 3: return pointer to newly allocated, 1D sparse vector for current
  // variable
  return newOutput_1DSparseVec_p;

} // end:: l4c_PreProcAllocate_1DSparseVector()

// ---------------------------------------------------------------------
// function: l4c_SummarizeFreezeThaw(sesCtl_p,deflated_1DSparseVec_p)
// purpose : Scan FT_STATE_UM_M09 as a 1D (3KM) sparse vector, and calc
//           frequencies of bit0 (assuming tsurf used if derived from L3_SM_A,
//           or, assuming bit0 populated by GEOS5 FP "TS" proxy approach
//           where 1=Frozen,0=Thawed.
// revised : 2015-09-07T15:18, corrected array indices (were backwards)
//           on report stanza.
// notes   : We retain L3_SM_A logic where 1=Frozen, 0=Thawed
//
// ASSIGNMENT LOGIC for L3_SM_A usage for FT_STATE_UM_M03 is:
//             if(ft_tsurf)
//                 ftStateUm_M03 = (uint8)frm_utilAssignBits(0,0,
//                 &ftStateUm_M03,sizeof(uint8));
//             if(ft_radar)
//                 ftStateUm_M03 = (uint8)frm_utilAssignBits(1,1,
//                 &ftStateUm_M03,sizeof(uint8));
//             if(ft_radar_qc)
//                 ftStateUm_M03 = (uint8)frm_utilAssignBits(2,2,
//                 &ftStateUm_M03,sizeof(uint8));
//             if(ft_in_swath)
//                 ftStateUm_M03 = (uint8)frm_utilAssignBits(3,3,
//                 &ftStateUm_M03,sizeof(uint8));
//
// ASSIGNMENT LOGIC when GEOS5 FP "TS" is used:
//             Bit0 set 1=TRUE when frozen, 0=false=thawed from GEOS5 FP "TS" as
//             deg C.
//                 ftStateUm_M03 = (uint8)frm_utilAssignBits(0,0,
//                 &ftStateUm_M03,sizeof(uint8));
//             Bit1 set 1=TRUE when InTerrestrial, 0=FALSE when outside land
//             mask (in ocean,rock etc)
//                 ftStateUm_M03 = (uint8)frm_utilAssignBits(1,1,
//                 &ftStateUm_M03,sizeof(uint8));
//             Bit2 set 1=TRUE ALWAYS if GEOS5 FP "TS" used, to match radar QC
//             being GOOD all the time
//                 ftStateUm_M03 = (uint8)frm_utilAssignBits(2,2,
//                 &ftStateUm_M03,sizeof(uint8));
//             Bit3 (ft_in_swath) set 1=TRUE always
//                 ftStateUm_M03 = (uint8)frm_utilAssignBits(3,3,
//                 &ftStateUm_M03,sizeof(uint8));
//
int l4c_SummarizeFreezeThaw(SesCtl *sesCtl_p, uint8 *deflated_1DSparseVec_p) {
  size_t pix3km = 0;
  // 56359296 pixels in 2D full (4872x11568) raster, or LDOM3KM=14976360 in 1D
  // sparse
  size_t nElem3km_1Dsparse = LDOM3KM;
  uint8 ftState3kmUint8 = 0; // whole uint8 pixel for FT_STATE_UM_M03
  uint8 ftBit0State = 0;     // returned boolean {1|0} 1=frozen, 0=thawed
  uint8 ftBit1State = 0;     // bit1 is InTerrestrial or not. Substitutes for
                             // ft_radar (0|1) when L3_SM_A used
  uint8 ftBit2State = 0;     // bit2 is ALWAYS FALSE if GEOS5 FP "TS" used,
                             // otherwise may be true or false if L3_SM_A used
  uint8 ftBit3State =
      0; // bit3 InSwath always TRUE if GEOS FP "TS" used otherwise is (0|1)

  size_t ftStateFreq[2] = {0}; // bit0 freqs
  size_t ftBit1Freq[2] = {0};  // bit1 freqs, InTerrestrial or not
  size_t ftQcFreq[2] = {0};    // bit2 freqs, proxy for QC
  size_t ftInSwathFreq[2] = {
      0}; // bit3 freqs, InSwath or not. Should always be 1

  char fxName[256] = {0};

  strcpy(fxName, __func__);
  xLogMsg("@A In(%s) FT_STATE_UM_M03 freeze thaw summary of %zu 1D-sparse "
          "pixels ~\n",
          fxName, nElem3km_1Dsparse);
  xLogMsg("@A In(%s) RIP defined FT policy %d %s ~ \n", fxName,
          sesCtl_p->ftState, sesCtl_p->ftMethodString);

  for (pix3km = 0; pix3km < nElem3km_1Dsparse; pix3km++) {
    // Retrieve whole uint8 3KM sparse pixel
    ftState3kmUint8 = deflated_1DSparseVec_p[pix3km];

    // Alternatives:
    // uint32 frm_utilRetrieveBits(const int16 startPos,const int16 endPos, void
    // *data,size_t dataSize) l4c_GetBits(0,0,&ftState3kmUint8,sizeof(uint8));

    // 1) Extract the bit0 value. Others are NOT yet scanned
    ftBit0State = l4c_GetBits(0, 0, &ftState3kmUint8, sizeof(uint8));
    // 2) Extract the bit1, as 1=TRUE when in land domain, false otherwise
    ftBit1State = l4c_GetBits(1, 1, &ftState3kmUint8, sizeof(uint8));
    // 3) Extract the bit2, as always true, matching a radar_qc of always good
    ftBit2State = l4c_GetBits(2, 2, &ftState3kmUint8, sizeof(uint8));
    // 4) extract bit3 (InSwath), which should ALWAYS be true if GEOS5 FP "TS"
    // used, otherwise may be true or false if L3_SM_A used.
    ftBit3State = l4c_GetBits(3, 3, &ftState3kmUint8, sizeof(uint8));
    // NEXT, collect frequencies for each bit 0,1,2,3 used in FT_STATE_UM_M03's
    // 8-bit byte noting that bits 4,5,6,7 are NOT yet used. Tests here
    // shouldn't be needed as we pulling off only 1 bit valued(0|1) but just to
    // be sure
    if (ftBit0State <= 1)
      ftStateFreq[ftBit0State]++;
    if (ftBit1State <= 1)
      ftBit1Freq[ftBit1State]++;
    if (ftBit2State <= 1)
      ftQcFreq[ftBit2State]++;
    // remember that InSwath=0 NOT 1 as per Barry Weiss convention
    if (ftBit3State <= 1)
      ftInSwathFreq[ftBit3State]++;
  } // end:: pix3km 1D sparse loop

  // NEXT, report on frequency results, by bit {0,1,2,3}
  xLogMsg(
      "@A In(%s) bit0 Freeze-Thaw               frozen(%9zu) thawed(%9zu) ~\n",
      fxName, ftStateFreq[1], ftStateFreq[0]);
  xLogMsg(
      "@A In(%s) bit1 ft_radar or InTerrestrial true  (%9zu) false (%9zu) ~\n",
      fxName, ftBit1Freq[1], ftBit1Freq[0]);
  xLogMsg(
      "@A In(%s) bit2 radar_qc or (always true) good  (%9zu) bad   (%9zu) ~\n",
      fxName, ftQcFreq[1], ftQcFreq[0]);
  xLogMsg(
      "@A In(%s) bit3 InSwath (0=In 1=Out-of-swath) In(%9zu) Out   (%9zu) ~\n",
      fxName, ftInSwathFreq[0], ftInSwathFreq[1]);
  xLogMsg("@A In(%s) completed ~\n", fxName);

  return 0;
} // end:: l4c_SummarizeFreezeThaw()

// --------------------------------------------------------------------
// function : l4c_PreProc_Bind_1D_DeflatedVars()
// purpose  : Takes final populated 1D sparse vector data for each input
//            ECO or MET variable, and attaches the DEFLATED
//            data to the "core" variable where it will be used.
// inputs   : const char *childName_p,
// returns  : int, cumulative count of variables bound
// notes    : CALLED BY: l4c_PreProc_ByLayerDriver()
// notes    : The dataset name hard-coded here MUST exactly match
//            the dataset name initialized in l4c_DAILY_PkgMgrInit_Themes.c
//            A more robust verification should be pursued.
//
// notes    : Currently the dataset names are hard coded-- the preferred
//            way to migrate to is to use statements like:
//            if(strcmp(datasetName_p,nsbGetValueByKey(nsb_p,"L4C_INPUT_DATASET_FPAR_M01",nvpBuf))==0)
//            { do logic here }.
//            be checked against their RIP nsb retrieved equivalents for
//            additional robustness. And, this routine is NOT currently
//            thread-safe due to presence of the 'static int'
//
int l4c_PreProc_Bind_1D_DeflatedVars(char *datasetName_p,
                                     void *deflated_1DSparseVec_p) {
  SesCtl *sesCtl_p = l4c_GetSesCtl();
  // fetch pointers to these master instances that live in  l4c_DAILY_Action.c
  EcoStateArrayAdt *ecoStateArrayObj_p = l4c_GetEcoStateArrayAdt();
  MetArrayAdt *metArrayObj_p = l4c_GetMetArrayAdt();
  NsbObj *nsbObj_p = l4c_GetNsbOverseer();
  char nvpBuf[256] = {0};
  char *dsetName_p = NULL;
  static int successCount = 0; // track cumulative count of all variables attached

  if (strcmp(datasetName_p, "fpar_M01") == 0) {
    // ATTACH completed 1D sparse data to the 'fpar'
    metArrayObj_p->fpar = deflated_1DSparseVec_p;
    successCount++;
  } else if (strcmp(datasetName_p, "fpar_qc_M01") == 0) {

    // <JMG> 2014-02-28T13:38,CRITICAL fpar_qc vs. fpar_qa confusion:
    //  Historically the HDF5 file uses fpar_qc_xx
    //  while the latest core code uses "fpar_qa" so sync these
    //  to conform to MODIS usage of fpar_qc (and change tcf_struct.h)
    //  as soon as possible!
    //  <LAJ> 2014-11-13, Has the above change occured?  The fpar_qc data have
    // bad values onward from at least this location.
    metArrayObj_p->fpar_qa = deflated_1DSparseVec_p;
    successCount++;
  } else if (strstr(datasetName_p, "fpar_clim_M01_day") != NULL) {
    // <JMG> added 4/22/2014, enable FparClimatology pre-proc
    metArrayObj_p->fpar_clim = deflated_1DSparseVec_p;
    // explictly label the time-variant FparClimatology name!
    memObjSetName(metArrayObj_p->fpar_clim, sesCtl_p->fparClimDatasetname);
    successCount++;

  } else if (strcmp(datasetName_p, "ndvi_M01") == 0) {
    metArrayObj_p->viirsNdvi = deflated_1DSparseVec_p;
    successCount++;
  } else if (strcmp(datasetName_p, "ndvi_qc_M01") == 0) {
    metArrayObj_p->viirsNdvi_qc = deflated_1DSparseVec_p;
    successCount++;
  } else if (strcmp(datasetName_p, "FT_ROLLING_M03") == 0) {
    // bind up Freeze thaw related states. Note that
    // ecoStateArrayObj_p->FrozenArea_9km is not created until inside
    // main model.
    ecoStateArrayObj_p->ftRolling_3km = deflated_1DSparseVec_p;
    successCount++;
  } else if (strcmp(datasetName_p, "FT_ROLLING_AGE_M03") == 0) {
    ecoStateArrayObj_p->ftRollingAge_3km = deflated_1DSparseVec_p;
    successCount++;
  } else if (strcmp(datasetName_p,
                    nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOC_C0",
                                     nvpBuf)) == 0) {
    ecoStateArrayObj_p->C[0].dom_1km = deflated_1DSparseVec_p;
    successCount++;
  } else if (strcmp(datasetName_p,
                    nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOC_C1",
                                     nvpBuf)) == 0) {
    ecoStateArrayObj_p->C[1].dom_1km = deflated_1DSparseVec_p;
    successCount++;
  } else if (strcmp(datasetName_p,
                    nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOC_C2",
                                     nvpBuf)) == 0) {
    ecoStateArrayObj_p->C[2].dom_1km = deflated_1DSparseVec_p;
    successCount++;
  } else if (strcmp(datasetName_p,
                    nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_NPP_SUM",
                                     nvpBuf)) == 0) {
    ecoStateArrayObj_p->npp_sum.dom_1km = deflated_1DSparseVec_p;
    successCount++;
  } else if (strcmp(datasetName_p, "QV2M_M09_AVG") == 0) {
    metArrayObj_p->qv2m = deflated_1DSparseVec_p;
    successCount++;
  } else if (strcmp(datasetName_p, "RADIATION_SHORTWAVE_DOWNWARD_FLUX") == 0) {
    metArrayObj_p->par = deflated_1DSparseVec_p;
    successCount++;
  } else if (strcmp(datasetName_p, "SM_ROOTZONE_WETNESS") == 0) {
    metArrayObj_p->smrz = deflated_1DSparseVec_p;
    successCount++;
  } else if (strcmp(datasetName_p, "SM_SURFACE_WETNESS") == 0) {
    metArrayObj_p->smtop = deflated_1DSparseVec_p;
    successCount++;
  } else if (strcmp(datasetName_p, "SOIL_TEMP_LAYER1") == 0) {
    metArrayObj_p->tsoil = deflated_1DSparseVec_p;
    successCount++;
  } else if (strcmp(datasetName_p, "SURFACE_PRESSURE") == 0) {
    metArrayObj_p->ps_avg = deflated_1DSparseVec_p;
    successCount++;
  } else if (strcmp(datasetName_p, "T2M_M09_AVG") == 0) {
    metArrayObj_p->t2m_avg = deflated_1DSparseVec_p;
    successCount++;
  } else if (strcmp(datasetName_p, "T2M_M09_MAX") == 0) {
    metArrayObj_p->tmax = deflated_1DSparseVec_p;
    successCount++;
  } else if (strcmp(datasetName_p, "T2M_M09_MIN") == 0) {
    metArrayObj_p->tmin = deflated_1DSparseVec_p;
    successCount++;
  } else if (strcmp(datasetName_p, "FT_STATE_UM_M03") == 0) {
    // <JMG> 2015-07-29, support new 3KM GEOS5 or passive radiometer F/T
    // <JMG> 2014-08-07T12:13, replaced uint16 SURFACE_FLAG with uint8
    // FT_STATE_UM_M03 deprecated old metArrayObj_p->surflag_3km
    metArrayObj_p->ft_state_um_3km = deflated_1DSparseVec_p;
    successCount++;
    // <JMG> 2015-09-06 added quick frequency count verification of bit0 in
    // FT_STATE_UM_M03 to verify when GEOS5 FP "TS" derived F/T was performed
    // NOTE: here we interpret this generic 1KM 1D buffer as a 3KM uint8 inside
    l4c_SummarizeFreezeThaw(sesCtl_p, deflated_1DSparseVec_p);

  } else if (strcmp(datasetName_p,
                    nsbGetValueByKey(nsbObj_p,
                                     "L4C_INPUT_DATASET_SOILROOTZONE_M09_MIN",
                                     nvpBuf)) == 0) {
    // <JMG> 2015-05-27 TBD, change to use RIP dynamic lookup of these names
    // dsetName_p =
    // nsbGetValueByKey(nsbObj_p,"L4C_INPUT_DATASET_SOILROOTZONE_M09_MIN",
    // nvpBuf);
    metArrayObj_p->smrz_mn = deflated_1DSparseVec_p;
    // <JMG> 2014-09-27, added name assignment here for these 2
    memObjSetName(metArrayObj_p->smrz_mn, "metArrayObj_p->smrz_mn");
    successCount++;
  } else if (strcmp(datasetName_p,
                    nsbGetValueByKey(nsbObj_p,
                                     "L4C_INPUT_DATASET_SOILROOTZONE_M09_MAX",
                                     nvpBuf)) == 0) {
    metArrayObj_p->smrz_mx = deflated_1DSparseVec_p;
    memObjSetName(metArrayObj_p->smrz_mx, "metArrayObj_p->smrz_mx");
    successCount++;
  } // end:: multi-way branching ...

  l4c_EventCompletionMsg(__func__);

  return successCount;

} // end:: l4c_PreProc_Bind_1D_DeflatedVars()

// --------------------------------------------------------------------
// function : l4c_PreProc_ByLayerDriver()
// purpose  : Driver to iterate through each of (18) 2D layers, DEFLATING each
// inputs   : const int nTotalToDeflate --count of 2D image layers to
//            transform to 1D sparse images
// revision : 2014-05-19T15:15, deprecated dataset name and theme name
//            string tables, now retrieves dataset names directly from
//            pkgMgr pObj entries initialized directly via NSB from the
//            external session's runtime-input-parameter (RIP) file.
//
// revision : Add the (2) SMRZ layers stored in SMAP_L4_C_Ancillary.h5 to
// DEFLATE lists
//            since these were inadvertantlhy stored as 2D grids NOT M09land as
//            I thought
// returns  : int,
// notes    : we store the file_id for each dataset in "Owner"
// pkgMgr_GetObjOwner(pObj_p) notes    : CRITICAL count "nTotalToDeflate" is
// passed in from l4c_GridManager and this
//            MUST match the actual no. of (currently) hard coded operations!!
//
int l4c_PreProc_ByLayerDriver(const int nTotalToDeflate) {
  SesCtl *sesCtl_p = l4c_GetSesCtl();
  PkgMgr *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
  PkgTheme *pThme_p = NULL;
  PkgObj *pObj_p = NULL;

  void *preProc_M01_image_p =
      NULL; // ptr to COMMON workhorse 1km SOURCE (input) M01 buffer
  void *sparse1d_vec_p =
      NULL; // ptr to 1D-sparse vector to populate with DEFLATEd vector data

  // <KAE> 2024-01-06 Manually allocating a sparse 1D vector so I can read
  //  and work with datasets that are not registered in the package manager
  void *sparse1d_alloc_p = NULL;

  uint16 *row_1D_lut_p =
      NULL; // "MCD12Q1_M09land_col_uint16" in SMAP_L4_C_Ancillary
  uint16 *col_1D_lut_p =
      NULL; // "MCD12Q1_M09land_row_uint16" in SMAP_L4_C_Ancillary

  int ithImage = 0;
  int rasterResolution = 0;
  int32 hdf4_Datatype = 0;

  int nPreProcVariables = nTotalToDeflate;
  int verifyCount = 0;
  hid_t curr_file_id = 0;
  size_t n1D_pix = 0;
  size_t pixSiz = 0;
  char fxName[256] = {0};
  char *dsetName_p = NULL; // ptr to each dataset's name in turn

  // <KAE> 2024-01-05 For reading (already) deflated binary files
  char file_path[FILENAME_MAX];
  // <KAE> 2024-01-05 Need to get this in order to load data not registered
  //  with the pkgMgr; I'm told that this will fetch pointers to the master
  //  instance that lives in l4c_DAILY_Action.c
  EcoStateArrayAdt *ecoStateArrayObj_p = l4c_GetEcoStateArrayAdt();

  strcpy(fxName, (char *)__func__);

  // <KAE> 2024-01-03 If the litterfall allocation scheme is to follow a schedule...
  xLogMsg("@A In(%s) Assessing the litterfall scheme ~\n", fxName);
  // <KAE> 2024-07-17 Allocating memory with memObjCalloc() instead
  sparse1d_alloc_p = memObjCalloc( // NOTE: Freed later, in l4c_DailyDriver()
    "litter_rate", sizeof(float32), LDOM1KM, IsFloat32, "l4c_PreProc_ByLayerDriver", __FILE__);
  if (strstr(sesCtl_p->litterfallScheme, "SCHEDULED") != NULL)
  {
    snprintf(file_path, FILENAME_MAX, sesCtl_p->litterfallFileTemplate, sesCtl_p->runCompositePeriod);
    if (!frm_ioDoesFileExist(file_path))
    {
      xLogMsg("@F In(%s) Mandatory LITTERFALL FILE TEMPLATE (%s) does not yet exist where stated! ~\n",
        fxName, file_path);
    }
    // We can use sparse1d_vec_p to receive binary array data because
    //  the litterfall arrays are already deflated
    frm_ioReadbinary(file_path, IsFloat32, LDOM1KM, sparse1d_alloc_p);
    ecoStateArrayObj_p->litter_rate = sparse1d_alloc_p;
    xLogMsg("@A In(%s) COMPLETED loading the litterfall allocation rate ~\n", fxName);
  }

  // DIAG: list out all names in PRE_PROC_DEFLATE_1KM theme
  // Fetch HDF5 dataset name from PRE_PROC_DEFLATE_1KM theme...
  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "PRE_PROC_DEFLATE_1KM");

  // Note: now possible to dynamically assign nTotalToDeflate by
  // fetching the count of child PkgObj in theme "PRE_PROC_DEFLATE_1KM"...
  verifyCount = pkgMgr_GetThemeObjectCount(pThme_p);

  // DIAG: echo ALL layers we intend to deflate
  for (ithImage = 0; ithImage < nTotalToDeflate; ithImage++) {
    pObj_p = pkgMgr_GetObjByIndex(pThme_p, ithImage);
    dsetName_p = pkgMgr_GetObjName(pObj_p);
    // xLogMsg("@A In(%s) %3d DEFLATE-CANDIDATE (%s)~
    // \n",fxName,ithImage,dsetName_p);
  } // end:: ithImage loop, passively report all layers to work
  // xLogMsg("@A~ \n");

  // Pre-Qualify the entire list of target images to deflate, making
  // sure each exists where advertised
  for (ithImage = 0; ithImage < nTotalToDeflate; ithImage++) {
    pObj_p = pkgMgr_GetObjByIndex(pThme_p, ithImage);
    dsetName_p = pkgMgr_GetObjName(pObj_p);

    curr_file_id = pkgMgr_GetObjOwner(pObj_p);

    // SPECIAL case..override for SOC layers and NPP_SUM, retrieving
    // the HDF5 id from pre-cached soc_group_id
    if (strstr("SOC/", dsetName_p) != NULL)
      curr_file_id = sesCtl_p->soc_group_id;

    // based on gppMethodStr as {FPAR|NDVI} may need to skip the
    // check on whichever dataaset {fpar_M01|ndvi_M01} is not required

    // CASE 1: if CLIM override chosen and if current dataset name is FPAR 8day
    // or NDVI then skip over this, go to next entry
    if (strstr(sesCtl_p->gppMethodStr, "CLIM") != NULL &&
        (strcmp(dsetName_p, "fpar_M01") == 0 ||
         strcmp(dsetName_p, "fpar_qc_M01") == 0 ||
         strstr(dsetName_p, "ndvi_") != NULL)) {
      // if L4C_GPP_SOURCE = CLIM, skip pre-proc of
      // fpar_M01,fpar_qc_M01,ndvi_M01 etc
      continue;
    } else if (strcmp(sesCtl_p->gppMethodStr, "FPAR") == 0 &&
               (strcmp(dsetName_p, "ndvi_M01") == 0 ||
                strcmp(dsetName_p, "ndvi_qc_M01") == 0)) {
      continue;
    } else if (strcmp(sesCtl_p->gppMethodStr, "NDVI") == 0 &&
               (strcmp(dsetName_p, "fpar_M01") == 0 ||
                strcmp(dsetName_p, "fpar_qc_M01") == 0)) {
      continue;
    }

    if (frm_hdf5_DatasetExists(curr_file_id, dsetName_p) != 1) {
      xLogMsg("@W In(%s) %-3.3d WARNING! Verify! dataset (%s) may NOT exist in "
              "file_id %d~\n",
              fxName, ithImage, dsetName_p, curr_file_id);
    } else {
      xLogMsg("@A In(%s) %-3.3d dataset (%36s) confirmed in file_id %d~\n",
              fxName, ithImage, dsetName_p, curr_file_id);
    }

  } // end:: ithImage pre-qualify entire list of images to deflate

  // Retrieve ptr to previously created anonymous, journaled re-usable
  // INPUT BUFFER for 2D 1km (e.g M01) data, retrieved via its PkgMgr Theme.
  preProc_M01_image_p = pkgMgr_GetMgr_ThemesChildDataByName(
      pkgMgr_p, "PRE_PROC_DEFLATE_1KM", "PREPROC_1KM_BUFFER");

  // Get: "MCD12Q1_M09land_row_uint16" in theme: SMAP_L4_C_Ancillary
  row_1D_lut_p = pkgMgr_GetMgr_ThemesChildDataByName(
      pkgMgr_p, "SMAP_L4_C_Ancillary", "MCD12Q1_M09land_row_uint16");

  // Get: "MCD12Q1_M09land_col_uint16" in SMAP_L4_C_Ancillary
  col_1D_lut_p = pkgMgr_GetMgr_ThemesChildDataByName(
      pkgMgr_p, "SMAP_L4_C_Ancillary", "MCD12Q1_M09land_col_uint16");

  // Fetch HDF5 dataset name from PRE_PROC_DEFLATE_1KM theme...
  pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "PRE_PROC_DEFLATE_1KM");

  // MAIN PRE-PROCESSING LOOP ----------------------------------------
  // Finally, iterate through all (21) INPUT datasets to DEFLATE
  // MAIN PRE-PROCESSING LOOP ----------------------------------------

  for (ithImage = 0; ithImage < nPreProcVariables; ithImage++) {
    // fetch ptr to THIS dataset's PkgObj container
    pObj_p = pkgMgr_GetObjByIndex(pThme_p, ithImage);

    // fetch HDF5 file handle from l4c_PrepDeflate_AssignDatasetProperties()
    curr_file_id = pkgMgr_GetObjOwner(pObj_p);

    // ASSERT: dsetName_p matches pkgMgr_GetObjName() here
    dsetName_p = pkgMgr_GetThemes_ChildName(pThme_p, ithImage);

    // <JMG> 2014-08-21, based on gppMethodStr as {FPAR|NDVI} need to skip the
    // check on whichever dataaset {fpar_M01|ndvi_M01} is not required

    // <JMG> 2014-11-06, when L4C_GPP_SOURCE = CLIM (forced use of only
    // FparClimatology) added SKIP-LOGIC to avoid pre-processing
    // fpar_M01,fpar_qc_M01 and/or NDVI since in this case these are not used or
    // relevant.
    if (strcmp(sesCtl_p->gppMethodStr, "FPAR") == 0 &&
        (strcmp(dsetName_p, "ndvi_M01") == 0 ||
         strcmp(dsetName_p, "ndvi_qc_M01") == 0)) {
      continue;
    } else if (strcmp(sesCtl_p->gppMethodStr, "NDVI") == 0 &&
               (strcmp(dsetName_p, "fpar_M01") == 0 ||
                strcmp(dsetName_p, "fpar_qc_M01") == 0)) {
      continue;
    } else if (strstr(sesCtl_p->gppMethodStr, "CLIM") !=
                   NULL && // allows CLIM or FPARCLIM etc
               (strcmp(dsetName_p, "fpar_M01") == 0 ||
                strcmp(dsetName_p, "fpar_qc_M01") == 0 ||
                strstr(dsetName_p, "ndvi_") != NULL)) {
      // if L4C_GPP_SOURCE = CLIM, skip pre-proc of
      // fpar_M01,fpar_qc_M01,ndvi_M01 etc
      continue;
    }

    // Removed diagnostic
    // xLogMsg("@A In(%s) DEFLATE RASTER %d [%s]\n",fxName, ithImage,
    // dsetName_p);

    // a NEW 1D sparse vector is allocated via memCalloc() for each (dsetName_p)
    sparse1d_vec_p = l4c_PreProcAllocate_1DSparseVector(dsetName_p);

    // Note: each input dataset's properties (datatype,Res,nPix,pixSiz etc) are
    // assigned earlier, within l4c_gridPre_Deflate.c in function
    // l4c_PrepDeflate_HDF5_Datasets()
    l4c_GetThemePropertySummary(dsetName_p, &hdf4_Datatype, &rasterResolution,
                                &n1D_pix, &pixSiz);

    // Avoid wasting time if somehow the resolution hasn't been set correctly
    assert(rasterResolution >= 1 && rasterResolution <= 9);

    // l4c_PreProc_Read_2D_Image performs the actual READ of the
    //   current INPUT 2D FULL GRID variable (of ca 20) from its
    //   HDF5 file, using the shared, re-usable 1km 2D full grid
    //   array as the INPUT to read HDF5 dataset, to eventually load
    //   the deflated result into the per-variable collection of
    //   sparse1d_vec_p arrays...
    l4c_PreProc_Read_2D_Image(curr_file_id, dsetName_p, fxName);

    // SWITCH based on KM resolution of this 'ithImage' processed
    switch (rasterResolution) {
    case 1:
      l4c_deflate_1km(row_1D_lut_p, col_1D_lut_p, preProc_M01_image_p,
                      sparse1d_vec_p, hdf4_Datatype);
      break;
    case 3:
      l4c_deflate_3km(row_1D_lut_p, col_1D_lut_p, preProc_M01_image_p,
                      sparse1d_vec_p, hdf4_Datatype);
      break;
    case 9:
      l4c_deflate_9km(row_1D_lut_p, col_1D_lut_p, preProc_M01_image_p,
                      sparse1d_vec_p, hdf4_Datatype);
      break;
    default:
      xLogMsg("@F In(%s) ds(%s) Invalid res(%d) for l4c_deflate_1km~\n",
              __func__, dsetName_p, rasterResolution);
      break;
    } // end:: switch on {1,3,9} KM resolution

    // Now the 'ith' DEFLATED sparse1d_vec_p is POPULATED, we bind
    //  it back (attach) to the correct ecoStateArrayAdt variable or
    //  metArrayAdt variable.  Here we match on the dataset name and
    //  bind the var directly to its equiv MetArrayAdt or EcoStateArrayAdt
    //  member
    // NOTE: this is hard coded for now, will be made table-drive later if
    // if possible.
    verifyCount = l4c_PreProc_Bind_1D_DeflatedVars(dsetName_p, sparse1d_vec_p);

    // verifyCount keeps a running total of n. of images pre-processed...
    xLogMsg(
        "@A DEFLATE-STATUS In(%s) [%-36s] %2d 1D-sparse variables completed~\n",
        fxName, dsetName_p, verifyCount);

  } // end:: ithImage loop

  l4c_EventCompletionMsg(fxName);

  // return count of rasters processed
  return ithImage;

} // end:: l4c_PreProc_ByLayerDriver()
