// ---------------------------------------------------------------
// module       : l4c_DAILY_WriteMetadata.c
// purpose      : Implement writes of formal metadata to L4_C HDF5 file(s)
//                to consolidate all metadata writes to this 1 file.
// project      : SMAP L4C science development.
// dependencies : HDF4 4.2.x, HDF5 v1.8.x or later libraries, L4C FRAMEWORK APIs
// revised      : 2021-11-01 Changed carbon_model_bitflag legend (l4cQaBitflagLegendStr)
//                  and added L4C_FPAR_SOURCE as an attribute on the output HDF5 granule,
//                  both to support VIIRS fPAR option
// revised      : 2017-07-24-- re-verify that "missing_value" is set correctly as
//                currently may inherit default 0.0 when it may need to be -9999.0
// revised      : 2016-04-13T11:23 <JMG> A reported missing comma after "2s" inserted into the
//                "flag_masks" string, responding to a bug report by Scott Dunbar, Ed Armstrong and 
//                Christin Alarcon (issued Feb 9th 2016 that I only now received word of).
//                Note since these are NOT used or assigned in L4_C we should look
//                into removing them entirely unless (as placeholders only) they help compatibility.

// revised      : 2016-03-03T16:29, consolidated by moving metadata calls located
//                prior tin l4c_DAILY_hdf5.c into this module file, to keep
//                all (or more of) the metadata handling in this one module.
// revised      : 2015-06-10T16:48, changed L4C_SOFTWARE_VERSION to L4C_SW_VERSION
// revised      : 2014-11-10T11:24, in 'units' changed curly brace to ')' symbol
//                to fix Ops error when Joe A's perl script post-processes these attributes.
// history      : 2014-06-23T18:10:00, added group-wise handlers.
// history      : 2014-06-23T17:33:00, adding creation of formal metadata
// author       : Joe Glassy, K. Arthur Endsley
// notes        : All HDF5 groups referenced by these write operations 
//                must pre-exist! note this module is in development and
//                has not yet been added to L4_C makefile. TBD..

#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"

char l4cQaBitflagLegendStr[] =
 "\n----------------------------------------------------------------------------------------\n"
"QA    Bitflag      Bit Position(s)  N.    Value        Description\n"
"flag  Name         (Start,End)      Bits  Range\n"
"------------------------------------------------------------------------------------------\n"
"01    NEE bit       00 - 00         (1)   {0|1}        0=NEE within valid range,1=out of range\n"
"02    GPP bit       01 - 01         (1)   {0|1}        0=GPP within valid range,1=out of range\n"
"03    Rh  bit       02 - 02         (1)   {0|1}        0=Rh  within valid range,1=out of range\n"
"04    SOC bit       03 - 03         (1)   {0|1}        0=SOC within valid range,1=out of range\n"
"05    PFT dominant  04 - 07         (4)   {1..8}       Dominant PFT class, decimal integer {1..8} inclusive\n"
"06    QA score      08 - 11         (4)   {0,1,2,3}    0=best, 3=worst; overall Quality score\n"
"07    GPP method    12 - 12         (1)   {0|1}        0=used fPAR, 1=used fPAR climatology\n"
"08    fPAR source   13 - 13         (1)   {0|1}        0=used MODIS fPAR, 1=used VIIRS fPAR\n"
"09    F/T method    14 - 14         (1)   {0|1}        0=used RADAR F/T, 1=used Tsurf F/T\n"
"10    IsFill        15 - 15         (1)   {0|1}        0=is NOT a Fillvalue, 1=Is a Fillvalue\n" ;


// ----------------------------------------------------------------------
// function : l4c_MetadataWriteDataset(hid_t file_id)
// purpose  : purpose: Write selective dataset metadata, including 
//            units, shortname. longname, valid_min,valid_max,
//            flagMeanings, QA bitflag legend,
//            and GEO coordinate variable pointer
// inputs   : hid_t file_id -- HDF5 file_id identifying output granule
// returns  : int,status
// revised  : 2016-03-03T1659, added to consolidate writes
// notes    : 
int l4c_MetadataWriteDataset(hid_t file_id)
{
 SesCtl   *sesCtl_p = l4c_GetSesCtl();    
 PkgMgr   *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
 PkgTheme *pThme_p = NULL;
 PkgObj   *pObj_p  = NULL;    
 // Hdf5GridPropAdt *hGridPropObj_p = l4c_GetHdf5GridProp();
 // NsbObj   *nsb_p   = l4c_GetNsbOverseer();
  
 int  idx = 0 ;
 int  startIndex = 0;
 int  endIndex   = 0;
 int  nEntries   = 0 ;
 int  nChildren = 0 ;
 int  dtype  = 0 ;      // to verify dataset's datatype for fillvalue and missing value assignments 
 char fxName[256] = {0} ;
 char coordLatLonRef[64] = {0};
 char *dsetName_p = NULL;
 char *themeName_p = NULL;
 char dsetAttribValue[FILENAME_MAX] = {0} ; 
 char flagMasks[32]      = {0} ; // added to match flagMasks
 char flagMeanings[256]  = {0} ; // added to match flag_meanings
 
 int32   intValue = 0;  // for type-specific fillvalue or missing value assignments
 float64 dblValue = 0.0;
 
 hid_t dset_id = -1 ;
 hid_t ret = -1; 
 
 
    strcpy(fxName,__func__);
 
    pThme_p     = pkgMgr_GetThemeByName(pkgMgr_p,"L4C_OUTPUTS");
    // retrieve name of theme, to confirm it
    themeName_p = pkgMgr_GetThemeName(pThme_p);
    // Establish a loop count, the number of PkgObj (child) one per dataset
    nChildren = pkgMgr_GetThemeObjectCount(pThme_p);

    xLogMsg("@A In(%s) CREATING %d L4_C (file id %d) DATASET Attributes from theme %s ~\n",
         fxName, nChildren, (int)file_id, themeName_p);        
        
    // TASK: next, implement WRITES of all dataset attributes.
    // Mild Caution: this pkgMgr sequenced approach (can) be error prone if
    //  the ORDER of these named pkgMgr entries in the OUTPUT PkgMgr instance 
    //  is changed from the order assumed below,
    //  so, be advised that once created, leave their ordering alone!
    
    // <JMG> 2020-07-13: a particular 'start' and 'end' index are determined here in order to 
    // work with a selective interior set of (L4C_OUTPUTS) objects ONLY
    // A better way might be to assemble an array of names in the required
    // order, and perform search-by-name explictly this way.
    
    // RECAP: the "NEE/nee_mean" entry MUST be the first (0-th) in the L4C_OUTPUTS pkgMgr theme
    startIndex = pkgMgr_GetThemes_ChildIndexByName(pThme_p,"NEE/nee_mean") ;
    // RECAP: the "GEO/longitude" entry MUST be the last (N-th-1) in the L4C_OUTPUTS pkgMgr theme
    endIndex   = pkgMgr_GetThemes_ChildIndexByName(pThme_p,"GEO/longitude");
    // establish the n. of entries...
    nEntries   = (endIndex - startIndex)+1;
    
    xLogMsg("@A In(%s) %s theme INDEX-ENDPOINTS First(%d)Last(%d) ~\n",
         fxName, themeName_p, startIndex,endIndex);
    
    // reminder: use a blank delimiter here to separate lat,lon --> and NOT a comma!
    strcpy(coordLatLonRef,"GEO/latitude GEO/longitude");
    
    // ITERATE through all OUTPUT datasets, writing attributes (units,shortname, _FillValue)...
    // This loop style (idx) process depends on the ORDER of entries in the pkgMgr theme,
    // which introduces risks if the ORDER is changed!
    
    // ASSERT: the expression idx <= endIndex is correct here as is. It should not be "<" 
    for(idx= startIndex ; idx <= endIndex ; idx++)
    {
        // Using base0 dataset index, retrieve this dataset's child obj
        // in the PkgMgr hierarchy 
        pObj_p = pkgMgr_GetObjByIndex(pThme_p,idx);

        // Retrieve the dataset_id prior assigned earlier, via HDF5 library call H5Dcreate()
        dset_id = pkgMgr_GetObjId(pObj_p);
    
        // <JMG> 2017-07-24, fetch of datasets datatype, 
        // for use in typing _Fillvalue or missing value
        dtype   = pkgMgr_GetObjDatatype(pObj_p);

        // Retrieve next KEY, e.g. same as dataset name, like "NEE/nee_mean"
        dsetName_p = pkgMgr_GetObjName(pObj_p);
        
        // SKIP over any DEPRECATED datasets
        if(strstr(dsetName_p,"QA/carbon_model_qual_pft")!= NULL)
                continue;
        
        // WRITE coordinates attribute to current dataset, suppressing for lat,lon itself
        if(strstr(dsetName_p,"GEO") == NULL)
        {
            ret = l4c_Hdf5_CreateAttribute(dset_id,IsString,".","coordinates",coordLatLonRef);
        }

        // fetch and WRITE units
        strcpy(dsetAttribValue, pkgMgr_GetObjUnits(pObj_p) );
        ret = l4c_Hdf5_CreateAttribute(dset_id,IsString,".","units",dsetAttribValue);
        
        // fetch and WRITE shortname
        strcpy(dsetAttribValue,pkgMgr_GetObjShortname(pObj_p));
        ret = l4c_Hdf5_CreateAttribute(dset_id,IsString,".","short_name",dsetAttribValue);
        
        // fetch and WRITE longname
        strcpy(dsetAttribValue,pkgMgr_GetObjLongname(pObj_p));
        ret = l4c_Hdf5_CreateAttribute(dset_id,IsString,".","long_name" ,dsetAttribValue);

        // xLogMsg("@D In(%s) %3.3d WRITING-DATASET (%s) ID %ld (called CreateAttirbute, ABOUT TO call AssignFillValue ~\n",fxName,idx,dsetName_p,dset_id);
        
        
        // ASSIGN FILLVALUE based on dataset name and its datatype 
        // <JMG> 2020-08-06, added write of missing_value inside of "l4c_Hdf5_AssignFillValue()"
        
        // <JMG> 2016-02-10, TBD --> change _FillValue's output datatype and value
        // to exactly match EACH DATASET's defined datatype
        // ERROR -- <JMG> 2020-07-14T10:22 error on GetObjByName() is occurring inside AssignFillValue
        // CHANGED call-signature to pass in pObj_p since its already available
        // check if "dsetName_p is corrupted?  debugging in progress....
        l4c_Hdf5_AssignFillValue(dset_id,pObj_p);
     
        // Define 'missing_value'. TBD, match missing_value's datatype to each dataset's actual datatype
        // dblValue = FLOAT_FILL_VALUE ;   ...start with baseline default
        // <JMG> 2017-07-24, new missing_value approach keys on datasets actual datatype to refine the
        // assignment of missing_value

        // WRITE valid_min,valid_max,fillvalue,missing_value next. Ideally, match datatype of
        // _FillValue.
        dblValue = pkgMgr_GetObjValidMin(pObj_p) ;
        ret = l4c_Hdf5_CreateAttribute(dset_id,IsFloat64,".","valid_min" ,&dblValue);
        ret = l4c_Hdf5_CreateAttribute(dset_id,IsFloat64,".","vmin" ,&dblValue);
        
        dblValue = pkgMgr_GetObjValidMax(pObj_p) ;
        ret = l4c_Hdf5_CreateAttribute(dset_id,IsFloat64,".","valid_max" ,&dblValue);
        ret = l4c_Hdf5_CreateAttribute(dset_id,IsFloat64,".","vmax" ,&dblValue);
        
        // for CF compliance, include a generic (unit) scale_factor and add_offset
        // excluding QA and coord vars
        if(strstr(dsetName_p,"GEO")==NULL || strstr(dsetName_p,"QA")==NULL)
        {
            dblValue = 1.0;
            ret = l4c_Hdf5_CreateAttribute(dset_id, IsFloat64, ".", "scale_factor" ,&dblValue);
            dblValue =0.0;
            ret = l4c_Hdf5_CreateAttribute(dset_id, IsFloat64, ".", "add_offset" ,&dblValue);
        }
 
        // <JMG> 2016-03-03T15:41, added carbon model bit flag 'flag_meanings' and
        // flag_masks back in 
               
        // Add QA bitflag legend, to just the carbon_model_bitflag 
        //  to attempt to mimic the incredibly obtuse way and anti-intuitive 
        //  way that L1, L2,L3 products handle these. Consider whether enum
        //  style (like in H5py) could be implemented using C library here...
        if(strstr(dsetName_p,"carbon_model_bitflag")!=NULL)
        {
            ret = l4c_Hdf5_CreateAttribute(dset_id,IsString,".",
                "legend_carbon_model_bitflag" ,l4cQaBitflagLegendStr);
                
            // <JMG> 2016-04-13T11:23 missing comma after "2s" inserted, responding to
            // a bug report by Scott Dunbar, Ed Armstrong Christin Alarcon (Feb 9th 2016) that we
            // just received. Note since these are NOT used or assigned in L4_C we should look
            // into removing them entirely unless (as placeholders only) they help compatibility.
            strcpy(flagMasks,"1s, 2s, 4s, 8s, 16s");
            
            ret = l4c_Hdf5_CreateAttribute(dset_id,IsString,".",
                "flag_masks" ,flagMasks);

            // <KAE> 2017-07-15 Updated from "NOT_USED ..." with best-guess at meeting NSIDC reqs.
            strcpy(flagMeanings,"NEE GPP RH SOC PFT PFT PFT PFT QA QA QA QA fPAR NDVI FT IsFill");
                
            // <JMG> 2016-03-03T16:46:00, placeholder 'flag_meanings' added but I have
            // not yet been able to trace this add-request to a specific requirement.
            // <JMG> 2016-04-13T11:22:00, flag_masks and flag_meanings present here are
            //       NOT CURRENTLY ASSIGNED ANY VALUES
            ret = l4c_Hdf5_CreateAttribute(dset_id,IsString,".",
                "flag_meanings" ,flagMeanings);
        }
        
    } // end:: loop startIdx to endIdx

    // <JMG> added 2020-07-13
    xLogMsg("@A In(%s) Completed  %d L4_C (file id %ld) DATASET Attributes from theme %s ~\n",
         fxName, nChildren, (int)file_id, themeName_p);

 return 0 ;
    
} // end:: l4c_MetadataWriteDataset()


// ----------------------------------------------------------------------
// function : l4c_Populate_ISO_MetadataElements(void)
// purpose  : Initialize master L4C_ISO_MetaElements(), assigning fallback
//            defaults, to later update based on granule level performance.
// inputs   : void
// returns  : int,status
// revised  : 2014-08-13T13:36, changed L4C_PRODUCTION_UTC_DATETIME
//            back to L4C_PRODUCTION_UTC_DATETIME to preserve Joe A's
//            Perl parser logic on post-production ISO metadata injection.
// history  : 2014-05-12T13:37:00,jmg, corrected duplicate UUID tag gen.
// notes    : DomainConsistency: pct of all granule pixels for named variable,
//            falling within acceptable limits
//            CompletenessOmission: pct of all granule pixels for named variable
//            with unacceptable data relative to the total Terrestrial pixels modeled.
void l4c_Populate_ISO_MetadataElements(void)
{
 // fetch pointer to master instance of ISO metadata elements   
 L4C_ISO_MetaElements *isoMeta_p = l4c_GetISO_MetaElements();
 // <JMG> changed name 2015-10-15 to l4c_GetInvalidEventCounts() to reflect
 // its expanding purpose
 PostInvalid *pInv_p   = l4c_GetInvalidEventCounts() ;
 
 char fxName[256] = {0};
 
 strcpy(fxName,__func__);
 xLogMsg("@A In(%s) Transfer ISO DomainConsistency %.4f and CompletenessOmission %.4f from pInv_p to isoMeta_p\n",
        fxName,pInv_p->qa_DomainConsistency,pInv_p->qa_CompletenessOmission);
 
 // DomainConsistency --pct of all granule pixels within acceptable limits
 isoMeta_p->nee_DomainConsistency    = pInv_p->qa_DomainConsistency;
 isoMeta_p->gpp_DomainConsistency    = pInv_p->qa_DomainConsistency; 
 isoMeta_p->rh_DomainConsistency     = pInv_p->qa_DomainConsistency;   
 isoMeta_p->soc_DomainConsistency    = pInv_p->qa_DomainConsistency;
 isoMeta_p->ec_DomainConsistency     = pInv_p->qa_DomainConsistency;

 // CompletenessOmission
 isoMeta_p->nee_CompletenessOmission = pInv_p->qa_CompletenessOmission ;
 isoMeta_p->gpp_CompletenessOmission = pInv_p->qa_CompletenessOmission ;
 isoMeta_p->rh_CompletenessOmission  = pInv_p->qa_CompletenessOmission ;
 isoMeta_p->soc_CompletenessOmission = pInv_p->qa_CompletenessOmission ;   
 isoMeta_p->ec_CompletenessOmission  = pInv_p->qa_CompletenessOmission ;


} // end:: l4c_Populate_ISO_MetadataElements()


// ----------------------------------------------------------------------
// function : l4c_Hdf5_AddFileMetadata(()
// purpose  : Add core file level metadata
// inputs   : void
// returns  : int,status
// revised  : 2014-08-13T13:36, changed L4C_PRODUCTION_UTC_DATETIME
//            back to L4C_PRODUCTION_UTC_DATETIME to preserve Joe A's
//            Perl parser logic on post-production ISO metadata injection.
// revised  : 2014-05-21, changed RELEASE_TAG to L4_C_SUBSYSTEM_RELEASE
// history  : 2014-05-12T13:37:00,jmg, corrected duplicate UUID tag gen.
// notes    : TBD, plan is to replace these calls with newer
//            frm_hdf5_WriteAttributeByType|CustomString() functions

void l4c_Hdf5_AddFileMetadata(void)
{
SesCtl         *sesCtl_p = l4c_GetSesCtl();
// AncImagePkgAdt *ancImagePkgMasterObj_p = l4c_GetAncImagePkgAdt();
NsbObj *nsb_p = l4c_GetNsbOverseer(); // <JMG> added 09/11/2014...
PkgMgr *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
PkgTheme *pThme_p= NULL;

L4C_ISO_MetaElements *isoMeta_p = l4c_GetISO_MetaElements();

// PkgObj   *pObj_p = NULL;
hid_t file_id    = sesCtl_p->l4cOutput_file_id;

// char taskUUID[64] = {0}; // buffer storing latest UUID generated
char granuleUUID[64] = {0} ;
char taskCompositePeriodStr[16] = {0} ;
char taskYearStr[16] = {0} ;
char timeStr[64] = {0};
char modelDayOfYear[8] = {0};
char fxName[256] ;
char valueBuffer[512] = {0} ;   
char nvpBuf[FILENAME_MAX] = {0};

char localDatetimeStr[64] ={0} ;
char utcDatetimeStr[64] = {0};
char govDocBuffer[64] = {0}; // for ATBD and PSD version,date etc

int status = -1 ;

    strcpy(fxName,(char *)__func__);
    
    xLogMsg("@A In(%s) l4c_Hdf5_AddFileMetadata \n",fxName);
 
    pThme_p  = pkgMgr_GetThemeByName(pkgMgr_p,"L4C_OUTPUTS");

    // PkgMgr references setup here but not used quite yet


    // ======================================================================
    // write formal "group-of-groups" metadata to primary (file_id) ...
    // ======================================================================
    status = l4c_Hdf5_AddFormalMetadata(file_id);

    // ======================================================================
    // write local baseline metadata to primary (file_id) ...
    // ======================================================================

    //  frm_hdf5_WriteAttributeCustomString(file_id,"/", "L4C_SUBSYS_RELEASE",sesCtl_p->SubsysReleaseTag);

    xLogMsg("@A In(%s) OUTPUT: L4_C Subsystem Release     (%s)~\n", fxName,sesCtl_p->SubsysReleaseTag);
    xLogMsg("@A In(%s) OUTPUT: L4_C Task label            (%s)~\n"  , fxName,sesCtl_p->l4c_Task_Label);

    // new syntax, uses frm_hdf5_WriteAttributeAsType() :
    strcpy(valueBuffer,sesCtl_p->swVersion);

    // status = frm_hdf5_WriteAttributeAsType(file_id,"/","L4_C_SW_VERSION",
    //      IsString, 1, valueBuffer);
    // Note: this is the same as the executable ID tag...
    l4c_Hdf5_CreateAttributeString( file_id,"L4_C_SW_VERSION"   ,sesCtl_p->swVersion);
    // <JMG> added 2014/09/11 macro L4C_VERS_TAG written here as
    // L4_C_SW_DATETIME... TBD: standardize across ALL l4c_MET,l4c_MODIS
    l4c_Hdf5_CreateAttributeString( file_id,"L4_C_SW_DATETIME"   ,sesCtl_p->swRelease);
    
    // <JMG> 2014-10-06, inherit from RIP, for consistency across produce + pre-processor outputs
    l4c_Hdf5_CreateAttributeString( file_id,"L4_C_DATASET_VERSION" ,sesCtl_p->datasetVersion);

    // <JMG> 2014-10-06, inherit from RIP, or hard-wire since DOI for 
    // consistency across produce + pre-processor outputs
    // note: for L4CMET, writes L4_C_MET_DOI and for L4CMOD writes L4_C_MOD_DOI, both as "TBD" for ow
    strcpy(valueBuffer, "10.5067/22TFAUSNLO9R" );
    l4c_Hdf5_CreateAttributeString( file_id,"L4_C_MDL_DOI" ,valueBuffer);

    // RELEASE tag here is the UM specific system-release identifier that includes the
    //  system version+datetime identifying not just the software but also the combined
    //  suite of static ancillary data (BPLUT,grid lookup tables, etc) that in total make
    //  up the "L4_C system"
    l4c_Hdf5_CreateAttributeString( file_id,"L4_C_SUBSYSTEM_RELEASE",sesCtl_p->SubsysReleaseTag);
    l4c_Hdf5_CreateAttributeString( file_id,"L4_C_TASK_LABEL"       ,sesCtl_p->l4c_Task_Label);
    
    // <JMG> 2015-11-11T10:12 suppress writing CRID here as this is incorrect anyway
    // as inherited from RIP NVP that is never actually assigned correctly or maintained by OPS
    // l4c_Hdf5_CreateAttributeString( file_id,"L4_C_CRID"             ,sesCtl_p->CRID);

    // <JMG> 2014-09-12T11:39, added, was omitted earlier
    // choices: "local"|"production"
    l4c_Hdf5_CreateAttributeString( file_id,"L4_C_PRODUCTION_LEVEL" ,sesCtl_p->productionLevel);

    // PRODUCTION_DATETIME orig format returned was is: "2014-09-11T22:59:21Z"
    // but GMAO wants it in format "2014-09-11T22:59:00.000Z"
    // Joe A, needs the form to look like: 2014-08-26T06:55:28.000Z
    l4c_GetISO8601_datetime(sesCtl_p->sesStrtSec,localDatetimeStr,utcDatetimeStr);
    l4c_Hdf5_CreateAttributeString( file_id,"L4_C_PRODUCTION_DATETIME", utcDatetimeStr);
    
    l4c_Hdf5_CreateAttributeString( file_id,"L4_C_HOST_PLATFORM",sesCtl_p->thisHost);

    // <KAE> 2021-11-01 Write "MODIS" or "VIIRS" to the output
    l4c_Hdf5_CreateAttributeString( file_id, "L4C_FPAR_SOURCE", sesCtl_p->fparSourceStr);
    
    // Write out fparState and ft settings (assumed available) for better provenance 
    l4c_writeNvpAsAttribute(file_id,"L4C_GPP_SOURCE","L4C_GPP_SOURCE");
    l4c_writeNvpAsAttribute(file_id,"L4C_FT_SOURCE","L4C_FT_SOURCE");
    
    // Do not use sesCtl_p->doy_Str[] as it appears to be contaminated in tcfAction etc.
    // runDayOfYear;  modelDayOfYear[8] = 
    sprintf(modelDayOfYear,"%d",sesCtl_p->runDayOfYear);
    l4c_Hdf5_CreateAttributeString( file_id,"L4_C_DAY_OF_YEAR",modelDayOfYear);
    // decide which of these to standardize on
    l4c_Hdf5_CreateAttributeString( file_id,"YEARDAY",modelDayOfYear);

    sprintf(taskCompositePeriodStr,"%d",sesCtl_p->runCompositePeriod);
    l4c_Hdf5_CreateAttributeString( file_id,"L4_C_COMPOSITE_PERIOD",taskCompositePeriodStr);
 
    sprintf(taskYearStr,"%d",sesCtl_p->runModelYear);
    l4c_Hdf5_CreateAttributeString( file_id,"L4_C_MODEL_YEAR",taskYearStr);
    l4c_Hdf5_CreateAttributeString( file_id,"L4_C_BPLUT_USED",sesCtl_p->inputBplutPath);
 
    l4c_Hdf5_CreateAttributeString( file_id,"L4_C_OUTPUT_GRANULE",sesCtl_p->output_L4C_DAILY_Path);
    // generate unique GRANULE_UUID for primary file
    frm_utilGenerateUUID(granuleUUID,1);
    l4c_Hdf5_CreateAttributeString( file_id,"L4_C_OUTPUT_GRANULE_UUID",granuleUUID);

    // <JMG> 2014-11-10T11:18:00, corrected INPUTPOINTER_FPAR to read SMAP_L4_C_FparClimatology.h5 file
    // when a forced option to use the FPAR Climatology is used. To correct post-processing Ops error
    // Joe A. documented. 
    if(strstr(sesCtl_p->gppMethodStr,"CLIM")!= NULL)
        l4c_Hdf5_CreateAttributeString( file_id,"INPUTPOINTER_FPAR" ,sesCtl_p->inputFparClimatologyPath);
    else
        l4c_Hdf5_CreateAttributeString( file_id,"INPUTPOINTER_FPAR" ,sesCtl_p->inputFparPath);
    
    
    l4c_Hdf5_CreateAttributeString( file_id,"INPUTPOINTER_MET"  ,sesCtl_p->inputMetPath);
    l4c_Hdf5_CreateAttributeString( file_id,"INPUTPOINTER_SOC"  ,sesCtl_p->inputSOCPath);
    
    // <JMG> 2014-10-08T16:23, added, always written regardless of gppMethodStr setting.
    //  until further notice.
    l4c_Hdf5_CreateAttributeString( file_id,"INPUTPOINTER_FPARCLIM",sesCtl_p->inputFparClimatologyPath);
    
    l4c_Hdf5_CreateAttributeString( file_id,"INPUTPOINTER_ANC"  ,sesCtl_p->inputAncPath);
    l4c_Hdf5_CreateAttributeString( file_id,"INPUTPOINTER_BPLUT",sesCtl_p->inputBplutPath);

    l4c_Hdf5_CreateAttributeString( file_id,"L4_C_RUNTIME_INPUTS_USED"  ,sesCtl_p->inputRipPath);
    
    // <JMG> 2014-09-11, added governing ATBD and PSD parameters from RIP pre-loaded in CollectNVPs
    // a) governing ATBD and PSD attribute
    // b) governing ATBD and PSD attributes
    if( nsbDoesKeyExist(nsb_p,"L4C_ATBD_VERSION"))
    {
        strcpy(sesCtl_p->govATBD_Vers, nsbGetValueByKey(nsb_p,"L4C_ATBD_VERSION",nvpBuf));
        l4c_Hdf5_CreateAttributeString( file_id,"L4_C_ATBD_VERSION"  ,sesCtl_p->govATBD_Vers);
    }
    if( nsbDoesKeyExist(nsb_p,"L4C_ATBD_DATE"))
    {
        strcpy(sesCtl_p->govATBD_Date, nsbGetValueByKey(nsb_p,"L4C_ATBD_DATE",nvpBuf));
        l4c_Hdf5_CreateAttributeString( file_id,"L4_C_ATBD_DATE"  ,sesCtl_p->govATBD_Date);
    }
    if( nsbDoesKeyExist(nsb_p,"L4C_PSD_EDITION"))
    {
        strcpy(sesCtl_p->govPSD_Edition, nsbGetValueByKey(nsb_p,"L4C_PSD_EDITION",nvpBuf));     
        l4c_Hdf5_CreateAttributeString( file_id,"L4_C_PSD_EDITION"  ,sesCtl_p->govPSD_Edition);     
    }
    if( nsbDoesKeyExist(nsb_p,"L4C_PSD_DATE"))
    {
        strcpy(sesCtl_p->govPSD_Date, nsbGetValueByKey(nsb_p,"L4C_PSD_DATE",nvpBuf));
        l4c_Hdf5_CreateAttributeString( file_id,"L4_C_PSD_DATE"  ,sesCtl_p->govPSD_Date);
    }
    
 
 l4c_EventCompletionMsg( fxName);
 // done
    
} // end:: l4c_Hdf5_AddFileMetadata()


// ---------------------------------------------------------------------
// function  : l4c_Hdf5_Add_ISO_FileMetadata(hid_t file_id)
// purpose   : Add DomainConsistency, CompletenessOmission and any others
// revised   : 2014-10-30T15:12:00-[07:00
// author    : Joe Glassy
// notes     : Split out from prior l4c_HDF5_AddFileMetadata() since
//             these values are NOT yet ready until post-processor is done!
// notes     : <JMG> 2014-10-06T16:17, added ISO element missing that Joe A reported
int l4c_Hdf5_Add_ISO_FileMetadata(const hid_t file_id)
{
 char fxName[256];
 int status =0;
 L4C_ISO_MetaElements *isoMeta_p = l4c_GetISO_MetaElements();   
 // for: DomainConsistency and CompletenessOmission
 char dcDescriptionStr[128] = {0};
 char dcUnitStr[128] = {0};

    
    strcpy(fxName,__func__);
    xLogMsg("@A In(%s) writing COMPLETENESS_OMISSION and DOMAIN_CONSISTENCY ~\n",fxName);

 strcpy(dcDescriptionStr,"CompletenessOmission description is TBD");
 // <JMG> 2014-11-10, changed mistaken "}" to correct ")" symbol. The curly brace
 // caused an error in Joe A's perl Ops script.
 strcpy(dcUnitStr,"percent (range:1..100)");
 
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","NEE_COMPLETENESS_OMISSION_DESCRIPTION",
            IsString, 1, dcDescriptionStr);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","GPP_COMPLETENESS_OMISSION_DESCRIPTION",
            IsString, 1, dcDescriptionStr);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","RH_COMPLETENESS_OMISSION_DESCRIPTION",
            IsString, 1, dcDescriptionStr);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","SOC_COMPLETENESS_OMISSION_DESCRIPTION",
            IsString, 1, dcDescriptionStr);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","EC_COMPLETENESS_OMISSION_DESCRIPTION",
            IsString, 1, dcDescriptionStr);

 status = frm_hdf5_WriteAttributeAsType(file_id,"/","NEE_COMPLETENESS_OMISSION_UNITS",
            IsString, 1, dcUnitStr);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","GPP_COMPLETENESS_OMISSION_UNITS",
            IsString, 1, dcUnitStr);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","RH_COMPLETENESS_OMISSION_UNITS",
            IsString, 1, dcUnitStr);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","SOC_COMPLETENESS_OMISSION_UNITS",
            IsString, 1, dcUnitStr);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","EC_COMPLETENESS_OMISSION_UNITS",
            IsString, 1, dcUnitStr);

 status = frm_hdf5_WriteAttributeAsType(file_id,"/","NEE_COMPLETENESS_OMISSION_VALUE",
            IsFloat32, 1, &isoMeta_p->nee_CompletenessOmission);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","GPP_COMPLETENESS_OMISSION_VALUE",
            IsFloat32, 1, &isoMeta_p->gpp_CompletenessOmission);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","RH_COMPLETENESS_OMISSION_VALUE",
            IsFloat32, 1, &isoMeta_p->rh_CompletenessOmission);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","SOC_COMPLETENESS_OMISSION_VALUE",
            IsFloat32, 1, &isoMeta_p->soc_CompletenessOmission);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","EC_COMPLETENESS_OMISSION_VALUE",
            IsFloat32, 1, &isoMeta_p->ec_CompletenessOmission);


 // --------------------------------------------------------------------
 // DOMAIN CONSISTENCY
 strcpy(dcDescriptionStr,"DomainConsistency description is TBD");
 strcpy(dcUnitStr,"percent (range:1..100)");

 status = frm_hdf5_WriteAttributeAsType(file_id,"/","NEE_DOMAIN_CONSISTENCY_DESCRIPTION",
            IsString, 1, dcDescriptionStr);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","GPP_DOMAIN_CONSISTENCY_DESCRIPTION",
            IsString, 1, dcDescriptionStr);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","RH_DOMAIN_CONSISTENCY_DESCRIPTION",
            IsString, 1, dcDescriptionStr);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","SOC_DOMAIN_CONSISTENCY_DESCRIPTION",
            IsString, 1, dcDescriptionStr);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","EC_DOMAIN_CONSISTENCY_DESCRIPTION",
            IsString, 1, dcDescriptionStr);

 status = frm_hdf5_WriteAttributeAsType(file_id,"/","NEE_DOMAIN_CONSISTENCY_UNITS",
            IsString, 1, dcUnitStr);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","GPP_DOMAIN_CONSISTENCY_UNITS",
            IsString, 1, dcUnitStr);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","RH_DOMAIN_CONSISTENCY_UNITS",
            IsString, 1, dcUnitStr);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","SOC_DOMAIN_CONSISTENCY_UNITS",
            IsString, 1, dcUnitStr);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","EC_DOMAIN_CONSISTENCY_UNITS",
            IsString, 1, dcUnitStr);
 
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","NEE_DOMAIN_CONSISTENCY_VALUE",
            IsFloat32, 1, &isoMeta_p->nee_DomainConsistency);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","GPP_DOMAIN_CONSISTENCY_VALUE",
            IsFloat32, 1, &isoMeta_p->gpp_DomainConsistency);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","RH_DOMAIN_CONSISTENCY_VALUE",
            IsFloat32, 1, &isoMeta_p->rh_DomainConsistency);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","SOC_DOMAIN_CONSISTENCY_VALUE",
            IsFloat32, 1, &isoMeta_p->soc_DomainConsistency);
 status = frm_hdf5_WriteAttributeAsType(file_id,"/","EC_DOMAIN_CONSISTENCY_VALUE",
            IsFloat32, 1, &isoMeta_p->ec_DomainConsistency);
            
 return 0;  
 
} // end:: l4c_Hdf5_()



// ----------------------------------------------------------------------
// function : l4c_Hdf5_AddFormalMetadata(()
// purpose  : Serves as a driver to call a series of group-specific
//            metadata writing handlers, writing attributes into the
//            set of pre-defined groups:
//            {CRID,Config,DatasetIdentification, Extent, 
//             GridSpatialRepresentation,ProcessStep,SeriesIdentification}
// inputs   : const hid_t file_id  --target ID to write info to
// returns  : int,status
// revised  : 2014-10-06T16:08:00, corrected name '_hdf5_' to "_Hdf5_" to
//            match other l4c_Hdf5_xxx functions and added 
//            
// history  : 2014-06-24T18:20:00, orig
// dependency: must call l4c_CreateAllGroups() BEFORE calling this
//             function since all HDF5 groups must already be in place.
// notes    : n/a
int l4c_Hdf5_AddFormalMetadata(const hid_t file_id)
{
    SesCtl *sesCtl_p = l4c_GetSesCtl();
    int status = 0 ;    
    char attribBuf[512] = {0} ;
    char nvpBuf[512]    = {0} ;
    // fetch ptr to the RIP name-value-pair stringblock database
    NsbObj *nsbObj_p          = sesCtl_p->sesNsbDb_p  ;
    // fetch ptr to overseer Package Manager instance
    PkgMgr *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
    
    // <JMG> 06/17/2014 work-in-progress
    // NOTE: list of L4_SM CF style subset of high-level attributes Rolf uses:
    
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/", "Contact", "http://www.ntsg.umt.edu");
        
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/", "Comment", "HDF-5 file format used");
        
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/", "Conventions", "CF 1.x");
        
    // THIS granule name
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/", "Filename", sesCtl_p->output_L4C_DAILY_Path);
        
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/", "History", "file written by l4c_DAILY.ex");
        
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/", "Institution", "University of Montana NTSG");
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/", "References", "See NASA SMAP Level 4 Carbon Product Specification Document");
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/", "Source", "TBD");
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/", "Title", "NASA SMAP Level 4 Carbon (L4_C) daily global 9km data product granule");
    
    // STUB in placeholders for the (2) ISO 19139 XML files that will be
    // added later by a post-processor.
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/Metadata", "iso_19139_dataset_xml", "XML 19139 added added via post-processor");
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/Metadata", "iso_19139_series_xml", "XML 19139 series added via post-processor");
    
    // Call a succession of subsidery handlers, one per HDF5 group
    // call interface: file_id plus groupFullLabel
    l4c_MetadataWriteGroup_CRID(file_id) ;  
    
    l4c_MetadataWriteGroup_SeriesIdentification(file_id) ;

    l4c_MetadataWriteGroup_Config(file_id);

    l4c_MetadataWriteGroup_DatasetIdentification(file_id) ;

    l4c_MetadataWriteGroup_Extent(file_id) ;

    l4c_MetadataWriteGroup_GridSpatialRepresentation(file_id) ;

    l4c_MetadataWriteGroup_ProcessStep(file_id) ;

    // <JMG> 2014-09-03T14:18, TBD yet to add:
    //  l4c_MetadataWriteGroup_AcquisitionInformation() 
    //  and ALL OTHER missing data-source entries

    l4c_EventCompletionMsg( (char *)__func__);  
    // return status
    return 0 ;
    
} // end:: l4c_Hdf5_AddFormalMetadata()


// ----------------------------------------------------------------------
// function : l4c_Hdf5_MetadataWriteGroup_CRID()
// purpose  : Add HDF5 "group-of-groups" metadata to the CRID group
//            to mirror L4_SM scheme.
// inputs   : const hid_t file_id  --target ID to write info to
// returns  : int,status
// revised  : 2014-06-24T12:12:00, orig
// history  : n/a
// notes    : n/a
void l4c_MetadataWriteGroup_CRID(const hid_t file_id)
{
    SesCtl *sesCtl_p = l4c_GetSesCtl();
    
    // fetch ptr to the RIP name-value-pair stringblock database
    NsbObj *nsbObj_p          = sesCtl_p->sesNsbDb_p  ;
    // fetch ptr to overseer Package Manager instance
    PkgMgr *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
    PkgTheme *pThme_p= NULL;
    // PkgObj   *pObj_p = NULL;
    // hid_t qa_file_id = sesCtl_p->l4cQa_file_id;
    char attribBuf[512] = {0} ;
    char nvpBuf[512]    = {0} ;
    int status = 0 ;

    strcpy(attribBuf, "TBD");
    status = frm_hdf5_WriteAttributeAsType(file_id, 
        "Metadata/Config","CRID",IsString,1,attribBuf);
    
    l4c_EventCompletionMsg( (char *)__func__);  
    
} // end:: l4c_MetadataWriteGroup_CRID()


// ----------------------------------------------------------------------
// function : l4c_Hdf5_MetadataWriteGroup_SeriesIdentification()
// purpose  : Add HDF5 "group-of-groups" metadata to the ProcessStep group
//            to mirror L4_SM scheme.
// inputs   : const hid_t file_id  --target ID to write info to
// returns  : int,status
// revised  : 2014-06-24T12:12:00, orig
// history  : n/a
// notes    : n/a
void l4c_MetadataWriteGroup_SeriesIdentification(const hid_t file_id)
{
    SesCtl *sesCtl_p = l4c_GetSesCtl();
    
    // fetch ptr to the RIP name-value-pair stringblock database
    NsbObj *nsbObj_p          = sesCtl_p->sesNsbDb_p  ;
    // fetch ptr to overseer Package Manager instance
    PkgMgr *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
    PkgTheme *pThme_p= NULL;
    PkgObj   *pObj_p = NULL;
    hid_t group_id   = 0;
    // hid_t qa_file_id = sesCtl_p->l4cQa_file_id;
    char attribBuf[512] = {0} ;
    char nvpBuf[512]    = {0} ;
    int status = 0 ;    
    
    l4c_EventCompletionMsg( (char *)__func__);
        
} // end:: l4c_MetadataWriteGroup_SeriesIdentification()


// ----------------------------------------------------------------------
// function : l4c_Hdf5_MetadataWriteGroup_Config()
// purpose  : Add HDF5 "group-of-groups" metadata to the "Config" group
//            to mirror L4_SM scheme.
// inputs   : const hid_t file_id  --target ID to write info to
// returns  : int,status
// revised  : 2014-06-24T17:35:00, orig
// history  : n/a
// notes    : The RIP keywords preface is "L4C_xxxx" which is different from
//             the HDF5 attribute qualifier, which is tpically "L4_C_xxxx"
void l4c_MetadataWriteGroup_Config(const hid_t file_id)
{
    SesCtl *sesCtl_p = l4c_GetSesCtl();
    
    // fetch ptr to the RIP name-value-pair stringblock database
    NsbObj *nsbObj_p          = sesCtl_p->sesNsbDb_p  ;
    
    // fetch ptr to overseer Package Manager instance
    PkgMgr *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
    PkgTheme *pThme_p= NULL;
    hid_t group_id   = 0;
    
    // hid_t qa_file_id = sesCtl_p->l4cQa_file_id;
    char nvpBuf[512]     = {0} ;
    char groupLabel[128] = {0} ;
    
    int status = 0 ;
    
    // Define common group qualified location for these Config attributes
    strcpy(groupLabel,"Metadata/Config");   
    
    // Example syntax, using WriteAttributeAsType() to a nested group
    // <JMG> 6/23/2014, TBD, change to fetch ATBD date,version from RIP!
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        groupLabel, "L4_C_ATBD_Date", 
        nsbGetValueByKey(nsbObj_p,"L4C_ATBD_DATE",nvpBuf));
        
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        groupLabel,"L4_C_ATBD_Version",
        nsbGetValueByKey(nsbObj_p,"L4C_ATBD_VERSION",nvpBuf));

    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        groupLabel,"L4_C_PSD_Edition",
        nsbGetValueByKey(nsbObj_p,"L4C_PSD_EDITION",nvpBuf));

    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        groupLabel,"L4_C_PSD_PublicationDate",
        nsbGetValueByKey(nsbObj_p,"L4C_PSD_PUBDATE",nvpBuf));

    // Note that in L4_C we combine the SW version and date in one field
        // <JMG> 2015-06-10 changed L4C_SOFTWARE_VERSION to L4C_SW_VERSION
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        groupLabel, "L4_C_SW_VersionID",
        nsbGetValueByKey(nsbObj_p,"L4C_SW_VERSION",nvpBuf));
    
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        groupLabel,"L4_C_identifier_DOI","TBD-for-DOI");
    
    l4c_EventCompletionMsg( (char *)__func__);  
    
} // end:: l4c_MetadataWriteGroup_Config()


// ----------------------------------------------------------------------
// function : l4c_Hdf5_MetadataWriteGroup_DatasetIdentification()
// purpose  : Add HDF5 "group-of-groups" metadata to the 
//            DatasetIdentification group to mirror the L4_SM scheme.
// inputs   : const hid_t file_id  --target ID to write info to
// returns  : int,status
// revised  : 2014-06-24T12:12:00, orig
// history  : n/a
// notes    : n/a
void l4c_MetadataWriteGroup_DatasetIdentification(const hid_t file_id)
{
    SesCtl *sesCtl_p = l4c_GetSesCtl();
    
    // fetch ptr to the RIP name-value-pair stringblock database
    NsbObj *nsbObj_p          = sesCtl_p->sesNsbDb_p  ;
    // fetch ptr to overseer Package Manager instance
    PkgMgr *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
    PkgTheme *pThme_p= NULL;
    // PkgObj   *pObj_p = NULL;
    hid_t group_id   = 0;
    // hid_t qa_file_id = sesCtl_p->l4cQa_file_id;
    char attribBuf[512] = {0} ;
    char nvpBuf[512]    = {0} ;
    int status = 0 ;    
    
    l4c_EventCompletionMsg( (char *)__func__);  
    
} // end:: l4c_MetadataWriteGroup_DatasetIdentification()


// ----------------------------------------------------------------------
// function : l4c_Hdf5_MetadataWriteGroup_Extent()
// purpose  : Add HDF5 "group-of-groups" metadata to the Extent group
//            to mirror the L4_SM scheme.
// inputs   : const hid_t file_id  --target ID to write info to
// returns  : int,status
// revised  : 2014-06-24T12:12:00, orig
// history  : n/a
// notes    : n/a
void l4c_MetadataWriteGroup_Extent(const hid_t file_id)
{
    SesCtl *sesCtl_p = l4c_GetSesCtl();
    
    // fetch ptr to the RIP name-value-pair stringblock database
    NsbObj *nsbObj_p          = sesCtl_p->sesNsbDb_p  ;
    // fetch ptr to overseer Package Manager instance
    PkgMgr *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
    // PkgTheme *pThme_p= NULL;
    // PkgObj   *pObj_p = NULL;
    hid_t group_id   = 0;
    // hid_t qa_file_id = sesCtl_p->l4cQa_file_id;
    char attribBuf[512] = {0} ;
    char nvpBuf[512]    = {0} ;
    int status = 0 ;    
    
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "Metadata/Extent", "description", "L4_C land domain excluding inland water and permanent ice");
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "Metadata/Extent", "eastBoundLongitude", "179.999");
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "Metadata/Extent", "northBoundLatitude", "90.0");
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "Metadata/Extent", "southBoundLatitude", "-90.0" );
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "Metadata/Extent", "westBoundLongitude", "-179.999");

    l4c_EventCompletionMsg( (char *)__func__);
        
} // end:: l4c_MetadataWriteGroup_Extent()


// ----------------------------------------------------------------------
// function : l4c_Hdf5_MetadataWriteGroup_GridSpatialRepresentation()
// purpose  : Add HDF5 "group-of-groups" metadata to the
//            GridSpatialRepresentation group to mirror the L4_SM scheme.
// inputs   : const hid_t file_id  --target ID to write info to
// returns  : int,status
// revised  : 2014-06-24T12:12:00, orig
// history  : n/a
// notes    : n/a
void l4c_MetadataWriteGroup_GridSpatialRepresentation(const hid_t file_id)
{
    SesCtl *sesCtl_p = l4c_GetSesCtl();
    
    // fetch ptr to the RIP name-value-pair stringblock database
    NsbObj *nsbObj_p = sesCtl_p->sesNsbDb_p  ;
    
    // fetch ptr to overseer Package Manager instance
    PkgMgr *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
    
    // PkgTheme *pThme_p= NULL;
    // PkgObj   *pObj_p = NULL;
    // hid_t qa_file_id = sesCtl_p->l4cQa_file_id;
    char attribBuf[512] = {0} ;
    char nvpBuf[512]    = {0} ;
    int status = 0 ;    

    l4c_EventCompletionMsg( (char *)__func__);

} // end:: l4c_MetadataWriteGroup_GridSpatialRepresentation()


// ----------------------------------------------------------------------
// function : l4c_Hdf5_MetadataWriteGroup_ProcessStep()
// purpose  : Add HDF5 "group-of-groups" metadata to the ProcessStep group
//            to mirror L4_SM scheme.
// inputs   : const hid_t file_id  --target ID to write info to
// returns  : int,status
// revised  : 2014-06-24T12:12:00, orig
// history  : n/a
// notes    : n/a
void l4c_MetadataWriteGroup_ProcessStep(const hid_t file_id)
{
    SesCtl *sesCtl_p = l4c_GetSesCtl();
    
    // fetch ptr to the RIP name-value-pair stringblock database
    NsbObj *nsbObj_p          = sesCtl_p->sesNsbDb_p  ;
    // fetch ptr to overseer Package Manager instance
    PkgMgr    *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
    PkgTheme  *pThme_p= NULL;
    PkgObj    *pObj_p = NULL;

    hid_t group_id   = 0;

    char attribBuf[512] = {0} ;
    char nvpBuf[512]    = {0} ;
    int status = 0 ;    
    
    l4c_EventCompletionMsg( (char *)__func__);
        
} // end:: l4c_MetadataWriteGroup_ProcessStep()
