// ---------------------------------------------------------------------
// module   : l4c_DAILY_bplut.c
// revised  : 2016-03-02, updated diag report to reflect BPLUT used by L4C OPS v7.1 (beta delivery)
// revised  : 2014-10-16T12:37, reviewed semantics of l4c_GetBPLUT_ByIndex()
//            regarding use of base0 vs. base1 PFT indices passed in.
// history  : 2014-02-22T12:26, split out scan and allocation of BPLUT
//             to try to better facilitate multiple uses (or READS) of BPLUT within
//             a complex application, without the risk of redundant allocations
// history  : 2014-02-15T11:35-07:00, modified to use new ADT namespaces
// history  : rev 2012-08-27T15:36:00-07:00
// author   : joe glassy
// Purpose  : this module reads a BPLUT (based on UMD veg index) organized
//           for row-oriented retrievals
//  Version : The version of the BPLUT: SMAP_BPLUT_2014-02-11.csv
//            BPLUT internal UUID must match "6640b8c0-9a01-11e3-a29d-000c29ddf1f0"
//
//  Warning: This software will NOT work with the original column oriented 
//    SMAP_BPLUT_20111214.txt or   SMAP_BPLUT_20120808.txt tables!
//        
//  Notes:        
//  --The logic for BPLUT processing now splits out the SCAN, the ALLOCATION
//  from the actual READ.<JMG>
//  --I have prepared a SMAP_BPLUT_2011-12-14.csv "row-oriented" CSV file
//    designed to work with this code.
//  --For stand alone Unit Test, compile using: 
//      gcc -g -Wall -std=c99 -D_GNU_SOURCE -DBPLUT_SELF_CONTAINED 
//      -DBPLUT_UNIT_TEST l4c_bplut.c  -o bplut.exe
//  --To execute, invoke using: ./bplut.exe ./SMAP_BPLUT_2011-12-14.csv                               
//
 

#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"

// (8) keys intended to match the Plant Function Type (PFT) classification
char *SMAP_BPLUT_Keys[] ={"ENF","EBF","DNF","DBF","SHR","GRS","CCP","BCP" };


// ---------------------------------------------------------------------
// Define private MASTER instance of BPLUT table
// Note: interface for external callers is primarily via l4c_GetBPLUT_Record(landcoverKey)
// and rarely, when absolutely needed, via l4c_GetBPLUT_Table(void).
//
static BPLUT *bpLUT_Master_p = NULL ; // pointer to array of BPLUT structures to populate

// ---------------------------------------------------------------------
// function  : l4c_BPLUT_Allocate(int nValidRecords)
// inputs    : FILE *bplutf -- ptr to open BPLUT file
// purpose   : Allocate memory for an array of BPLUT records
// returns   : BPLUT *bpLUT_p -- ptr to allocated array of BPLUT structures
// notes     : memObjCalloc() allocates the bpLUT_Master_p which is a static
//              so we intentionally do NOT return a pointer to it -- it is 
//              already directly visible to any function in this module file.
//  Version of BPLUT used here is: SMAP_BPLUT_2014-02-11.csv
//
int l4c_BPLUT_Allocate(const int nValidRecs,const char *bpLUT_Label_p)
{
 time_t actionSec = time(0);
 static int bpLutAllocState = 0 ; // after 1st allocation, set true, to avoid multiples..
 
    if( bpLutAllocState == 0 && bpLUT_Master_p == NULL)
    {
        // REMEMBER: this allocates the module level static bpLUT_p and NOT
        // a local pointer!!!
        bpLUT_Master_p = (BPLUT *)memObjCalloc(bpLUT_Label_p, sizeof(BPLUT),
                        nValidRecs, IsCustomUserType,"","");
                                
        xLogMsg("@A BPLUT Named(%s) allocated with %d records on %s ~\n",
            bpLUT_Label_p, nValidRecs, ctime(&actionSec));
    } else {
        // if an attempt is made to re-allocate this, block it and emit warning
        xLogMsg("@E ~l4c_BPLUT_Allocate: ERROR! BPLUT (%s) allocated previously to %d recs\n",
            bpLUT_Label_p, bpLutAllocState);
    }

 // Adjust allocation state after 1st attempt, allows showing this count on second or 
 // other alloc attempts...
 bpLutAllocState = nValidRecs ;
    
 // return count of recs allocated. Intentionally DO NOT return ptr to
 // the module level static, bpLUT_Master_p since all code in this file
 // ALREADY has direct access and visibility to this!!
 return nValidRecs ;    
} // end:: l4c_BPLUT_Allocate()


// ---------------------------------------------------------------------
// function  : l4c_BPLUT_Scan(inf)
// inputs    : FILE *bplutf -- ptr to open BPLUT file
// purpose   : Passive SCAN the SMAP L4_C BPLUT from CSV file, returning
//             a count of qualifying BPLUT records
// returns   : int, count of BPLUT records
// notes     : Version of BPLUT used here is: SMAP_BPLUT_2014-02-11.csv
//
int l4c_BPLUT_Scan(FILE *bplutf)
{
 char lineBuffer[MAX_LINE_BUFFER_LEN] = {0} ;  // string buffer to hold each line of text in
 char fxName[256];
 int ithColumn  =0;
 int  nFields = 0 ;         // line structure validation counter for n. of fields
 int validRecdCount = 0;
 
 // force starting at the top, just in case. who knows it was last left..
 strcpy(fxName,(char *)__func__);
 rewind(bplutf);
 
 // iterate through all lines, counting qualified BPLUT records
 while((fgets(lineBuffer,MAX_LINE_BUFFER_LEN-1,bplutf))!= NULL)
 {
    // first, skip over any comment lines
    if(strchr(lineBuffer,'#') != NULL && lineBuffer[0] == '#')
                                                       continue;
    // Count up the commas to superfiscially validate each line
    // NOTES: there will be ONE LESS comma then valid n. of fields 
    for(nFields=0,ithColumn=0; ithColumn < (int)strlen(lineBuffer); ithColumn++)
                     if(lineBuffer[ithColumn] == ',') nFields++ ;

    // When approx no. of fields appear to be available,
    //    increment count of valid records
    if(nFields >= 40)
              validRecdCount++;
              
    // Explicitly reset line for next read...
    memset(lineBuffer,'\0',MAX_LINE_BUFFER_LEN);
    
 } // end:: fgets() loop
 
 // Rewind the file pointer back to beginning for 2nd pass
 rewind(bplutf);

 // Allocate array of validLineCount structures, as a "named" memObj ...
 // <JMG> 2016-03-02, added display of nFields, adding 1 to count to
 // reflect the base0 loop logic above
 xLogMsg("@A In(%s): encountered %d qualifying records and %d fields~\n",
        fxName,validRecdCount, nFields+1);
 
 // Return the count of qualifying records
 return validRecdCount ;
 
} // end:: l4c_BPLUT_Scan()


// ---------------------------------------------------------------------
// function  : l4c_BPLUT_Read
// inputs    : SesCtl *sesCtl_p --provides sesCtl_p->inputBplutPath to local bpLutPath
//           : int *nBplutRecords  -- ptr to count of BPLUT records successfully read (returned)
// purpose   : Read the SMAP L4_C BPLUT from CSV file, create and return
//             array of populated BPLUT structures
// returns   : BPLUT *, pointer to array of structures, AND assigns 
//             count of records to nBpLutRecords. 
// revised   : 2017-06-05T17:33, added fetch of sesCtl to facilitate (FUTURE) hard-verification
// of BPLUT's UUID and or MD5 from RIP NVPs L4C_BPLUT_UUID and L4C_BPLUT_MD5 respectively.
// via member functions sesCtl_p->l4c_BPLUT_UUID and sesCtl_p->BPLUT_MD5 
// notes     : Version of BPLUT used here is: SMAP_BPLUT_2017-05-27.csv
// notes     : sesCtl_p provided for hard check but IS NOT YET USED as of 2017-06-05.
BPLUT *l4c_BPLUT_Read(SesCtl *sesCtl_p,int *nBplutRecords)
{
 BPLUT *bp_p  = NULL ; // convenience pointer to a SINGLE RECORD of BPLUT
 FILE *bplutf = NULL;   // pointer to local BPLUT CSV file handle
 
 char fxName[256] ={0} ;
 char *ch_p            = NULL ; // pointer to successive tokens on a line
 char *brkb_p  ; // pointer required for strtok_r thread safety
 char lineBuffer[MAX_LINE_BUFFER_LEN] = {0} ;  // string buffer to hold each line of text in
 char bpLutPath[FILENAME_MAX] = {0} ; // Added 2017-06-05...
 char tokenSepChars[8] = ", \t" ;// supported characters for token separators
 // int  i =0;
 int  ithRecord =0;         //  BPLUT record counter
 int  ithColumn =0;         // column (e.g. field) counter for parser
 int  nComments =0; 
 int  validRecdCount =0 ;   // count of valid BPLUT records 
 // int  lutColumnIndex =0 ;   // no longer used, col label
 
  strcpy(fxName,__FILE__);
  strcat(fxName,".");
  strcat(fxName,__func__);

    strcpy(bpLutPath, sesCtl_p->inputBplutPath);

    // OPEN BPLUT file, to parse...
    bplutf = frm_Openfile(bpLutPath,"rt",fxName);

    xLogMsg("@A In(%s) External BPLUT text file (%s) opened successfully~\n",
        fxName, bpLutPath);

    // Pass 1: establish count of valid BPLUT rows to size array of BPLUT structures   
    // SCAN and store count of qualifying records 
    validRecdCount = l4c_BPLUT_Scan(bplutf);
 
    // ALLOCATE BPLUT only once!! returning result in a LOCAL pointer
    l4c_BPLUT_Allocate(validRecdCount,"L4_C_BPLUT") ;
 
 
 // Pass 2: read data into correctly sized array of structures 
 while((fgets(lineBuffer, MAX_LINE_BUFFER_LEN-1, bplutf))!= NULL)
 {
    lineBuffer[strlen(lineBuffer)-1] = '\0' ; // clip newline
    // skip over any comment lines
    if(strchr(lineBuffer,'#') != NULL && lineBuffer[0] == '#')
    {
            nComments++ ;
            continue;
    }
    
    // Set pointer DIRECTLY to the master BPLUT and get 
    //   next each successive (ithRecord) record in BPLUT table
    bp_p = &bpLUT_Master_p[ithRecord] ;

    ithColumn =0 ;
    // iterate on current line through all delimited tokens (columns) 
    for(ch_p = strtok_r(lineBuffer,tokenSepChars,&brkb_p);
        ch_p != NULL;
        ch_p = strtok_r(NULL,tokenSepChars,&brkb_p))
    {
       // just in case CSV created in Windows, trim off CR/LF
       if(ch_p[strlen(ch_p)-1] == 13) 
                ch_p[strlen(ch_p)-1] = '\0' ;
                
       // branch on base0 column index, to load structure member
       switch(ithColumn)
       {
       // ADDED 'lutIntegerIndex' to allow integer index lookups
       case  0: bp_p->lutIntegerIndex = atoi(ch_p);  break; 
       // ADDED 'lutKey' string, as char lutKey[32] for string driven lookups 
       case  1: strcpy(bp_p->lutKey,ch_p)                     ; break;       
       case  2: strcpy(bp_p->useNDVItoFPAR, ch_p)             ; break;
       case  3: bp_p->NDVItoFPAR_scale = atof(ch_p)           ; break;
       case  4: bp_p->NDVItoFPAR_offset= atof(ch_p)           ; break;
       case  5: bp_p->LUEmax     = atof(ch_p)                 ; break;
       case  6: bp_p->Tmin_min   = atof(ch_p)                 ; break;
       case  7: bp_p->Tmin_max   = atof(ch_p)                 ; break;
       case  8: bp_p->VPD_min    = atof(ch_p)                 ; break;
       case  9: bp_p->VPD_max    = atof(ch_p)                 ; break;
       case 10: bp_p->SMrz_min   = atof(ch_p)                 ; break;
       case 11: bp_p->SMrz_max   = atof(ch_p)                 ; break;
       case 12: bp_p->FT_min   = atof(ch_p)                   ; break;
       case 13: bp_p->FT_max   = atof(ch_p)                   ; break;
       case 14: bp_p->SMtop_min  = atof(ch_p)                 ; break;
       case 15: bp_p->SMtop_max  = atof(ch_p)                 ; break;
       case 16: bp_p->Tsoil_beta0  = atof(ch_p)               ; break;
       case 17: bp_p->Tsoil_beta1  = atof(ch_p)               ; break;
       case 18: bp_p->Tsoil_beta2  = atof(ch_p)               ; break;
       case 19: bp_p->fraut      = atof(ch_p)                 ; break;
       case 20: bp_p->fmet       = atof(ch_p)                 ; break;
       case 21: bp_p->fstr       = atof(ch_p)                 ; break;
       case 22: bp_p->kopt       = atof(ch_p)                 ; break;
       case 23: bp_p->kstr       = atof(ch_p)                 ; break;
       case 24: bp_p->kslw       = atof(ch_p)                 ; break;
       case 25: bp_p->Nee_QA_Rank_min = (uint32) atoi(ch_p)   ; break;
       case 26: bp_p->Nee_QA_Rank_max = (uint32) atoi(ch_p)   ; break;
       case 27: bp_p->Nee_QA_Error_min = atof(ch_p)           ; break;
       case 28: bp_p->Nee_QA_Error_max = atof(ch_p)           ; break;
       case 29: bp_p->Fpar_QA_Rank_min = (uint32) atoi(ch_p)  ; break;
       case 30: bp_p->Fpar_QA_Rank_max = (uint32) atoi(ch_p)  ; break;
       case 31: bp_p->Fpar_QA_Error_min = atof(ch_p)          ; break;
       case 32: bp_p->Fpar_QA_Error_max = atof(ch_p)          ; break;
       case 33: bp_p->FtMethod_QA_mult = atof(ch_p)           ; break;
       case 34: bp_p->FtAge_QA_Rank_min = (uint32) atoi(ch_p) ; break;
       case 35: bp_p->FtAge_QA_Rank_max = (uint32) atoi(ch_p) ; break;
       case 36: bp_p->FtAge_QA_Error_min = atof(ch_p)         ; break;
       case 37: bp_p->FtAge_QA_Error_max = atof(ch_p)         ; break;
       case 38: bp_p->Par_QA_Error = atof(ch_p)               ; break;
       case 39: bp_p->Tmin_QA_Error = atof(ch_p)              ; break;
       case 40: bp_p->Vpd_QA_Error = atof(ch_p)               ; break;       
       case 41: bp_p->Smrz_QA_Error = atof(ch_p)              ; break;
       case 42: bp_p->Tsoil_QA_Error = atof(ch_p)             ; break;
       case 43: bp_p->Smtop_QA_Error = atof(ch_p)             ; break;
       default: xLogMsg("@F In(%s): ~Unexpected column(%d) at record %d\n",
                    fxName, ithColumn,ithRecord);
                break;
       } //end:: switch on COLUMN of BPLUT
       ithColumn++ ; // increment count of valid columns processed 
       if(ithColumn == L4C_BPLUT_N_COLUMNS) 
           ithRecord++ ; // increment count of BPLUT records populated

    } // end:: strtok_r() loop for token parsing

 } // end::fgets loop
 
 xLogMsg("@A In(%s) BPLUT read %3d columns of (%d) records of (%d) expected~ \n",
    fxName, ithColumn, ithRecord, BPLUT_N_EXPECTED_RECORDS);

  // <JMG> 2015-05-28T16:23:00, reclassified this to a fatal non-recoverable
  // error, since so much depends on the BPLUT integrity.
  if(ithRecord < BPLUT_N_EXPECTED_RECORDS)
  {
      xLogMsg("@F In(%s) FATAL -- BPLUT apparent read failure at LUT column(%d)!~ \n",fxName,ithColumn);
      fclose(bplutf);
      l4c_CloseSession();
      exit(-1);
  }
 
 // close the BPLUT CSV file
 fclose(bplutf);
 
 // assign the finished number of valid records to pass back to caller
 *nBplutRecords = ithRecord ;

 l4c_EventCompletionMsg(fxName);

 // return pointer to populated (master) table as a convenience
 return bpLUT_Master_p ;

} // end::l4c_BPLUT_Read()


// ---------------------------------------------------------------------
// function :
// Purpose: l4c_GetBPLUT_Private() returns a pointer to the master
//   private instance of the array of structures. External callers should
//   never routinely access the BPLUT this way, and thus will RARELY if
//   ever need this function. The recommended practice is to instead use 
//   the l4c_GetBPLUT_Record("ENF") style of access for 1 record at a time. 
BPLUT *l4c_GetBPLUT_Private(void)
{
    
 return bpLUT_Master_p;
 
} // end::l4c_GetBPLUT_Private()

         
// ---------------------------------------------------------------------
// function : l4c_GetBPLUT_Record(BPLUT_KeyString_p)
// Returns  : pointer to located BPLUT record if found, otherwise returns NULL
// Purpose  : provide normal caller with interface to BPLUT rec retrieval
//            via the landcover abbreviated "string" key (or PFT) like "EBF", etc
// 
BPLUT *l4c_GetBPLUT_Record(const char *landcoverLutKey_p,const int nValidBPLUT_Records)
{
 BPLUT *bp_p  = NULL;    // pointer to one BPLUT record
 int    ithRecd  = 0   ;    // record counter
 // static char lastKeyFound[32] = {0} ; ...temp disabled the 1-deep cache'd key
 
     for(ithRecd=0; ithRecd < nValidBPLUT_Records; ithRecd++)
     {
        // set pointer to private MASTER instance of BPLUT array of structures
        bp_p = &bpLUT_Master_p[ithRecd] ;
        
        // check for match on key, take early out when found
        if(strcmp(landcoverLutKey_p,bp_p->lutKey)== STRING_MATCHED)
            break;
     } // end::ithRecd loop
  
 return bp_p ;           
} // end:: l4c_GetBPLUT_Record()


// ---------------------------------------------------------------------
// function :l4c_GetBPLUT_ByIndex(LCC_IntegerIndexKey)
// Returns  : pointer to located BPLUT record if found, otherwise returns NULL
// Inputs   : const int landcoverIntegerKey -- base1 PFT {1..8} index
//            const int nValidBPLUT_Records -- total n. of defined BPLUT records (8)
// Purpose  : Provide normal caller with interface to BPLUT rec retrieval
//            via a landcover or PFT index supplied as base1 {1..8}, not base0 {0..7}
// Revised  : 2014-10-16T12:35, documentation reviewed only.
// notes    : When passing in 'landcoverIntegerKey, it is absolutely
//            CRITICAL to be aware of whether a base0 or base1 index is
//            appropriate here!
BPLUT *l4c_GetBPLUT_ByIndex(const int landcoverIntegerKey,const int nValidBPLUT_Records)
{
 BPLUT *bp_p  = NULL;    // pointer to one BPLUT record
 int    ithRecd  = 0   ;    // record counter
 
     for(ithRecd=0; ithRecd < nValidBPLUT_Records; ithRecd++)
     {
        // set pointer to private MASTER instance of BPLUT array of structures
        bp_p = &bpLUT_Master_p[ithRecd] ;
        // Stop early on a Match of the base1 INTEGER index key {1..8} not {0..7}
        if(landcoverIntegerKey == bp_p->lutIntegerIndex)
                    break;
     } // end::ithRecd loop

 return bp_p ; 
 
} // end:: l4c_GetBPLUT_ByIndex()


#ifdef OLD_REPORT_ENABLED
// ---------------------------------------------------------------------
// function :l4c_BPLUT_OldReport()
// inputs   : FILE *outf -- ptr to open file handle
//            int nPftClasses ..set to NPFT.. 8
// Returns  : pointer to located BPLUT record if found, otherwise returns NULL
// Purpose  : Provides a method to output contents of a BPLUT for verification
// notes    : 
// #define   NPFT 8    Number of vegetated landcover types
int l4c_BPLUT_OldReport(FILE *outf,const int nPftClasses)
{

 BPLUT *bp_p = NULL;
 int r=0;
 char fxName[256] = {0} ;
 
   strcpy(fxName,__func__); 
   fprintf(outf,"@A l4c_UnitTest_BPLUT diagnostic report~\n");
   
   for(r=0; r < nPftClasses; r++)
   {
     // l4c_GetBPLUT_Record has been re-verified using string keys "EBF" etc
     // bp_p = l4c_GetBPLUT_Record(SMAP_BPLUT_Keys[r], nPftClasses);
     // test fetch on both key as well as index
     bp_p = l4c_GetBPLUT_ByIndex(r, nPftClasses);
     
     if( bp_p != NULL)
     {
        fprintf(stdout,"%3.3d Key_%s,Index%2.2d,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf,%lf\n",
        r,
        bp_p->lutKey ,        
        bp_p->lutIntegerIndex,
        bp_p->LUEmax,          
        bp_p->Tmin_min,        
        bp_p->Tmin_max,        
        bp_p->VPD_min,         
        bp_p->VPD_max,         
        bp_p->SMrz_min,        
        bp_p->SMrz_max,        
        bp_p->SMtop_min,       
        bp_p->SMtop_max,       
        bp_p->Tsoil_beta0,     
        bp_p->Tsoil_beta1,     
        bp_p->Tsoil_beta2,     
        bp_p->fraut ,          
        bp_p->fmet ,           
        bp_p->fstr ,           
        bp_p->kopt ,           
        bp_p->kstr ,           
        bp_p->kslw );
     }
   } // end::r loop

 return r;  
} // end::l4c_BPLUT_OldReport()
#endif

// ---------------------------------------------------------------------
// function :l4c_BPLUT_Report()
// inputs   : FILE *outf -- ptr to open file handle
//            int nPftClasses ..set to NPFT.. 8
// Returns  : pointer to located BPLUT record if found, otherwise returns NULL
// Revised  : 2016-03-02, to reflect BPLUT used by L4C OPS v7.1 (beta delivery)
// Purpose  : Provides a method to output contents of a BPLUT for verification
// notes    : 
// #define   NPFT 8    Number of vegetated landcover types
int l4c_BPLUT_Report(FILE *outf,const int nPftClasses)
{
 BPLUT *bp_p = NULL;
 int r=0;
 char fxName[256] = {0} ;
 
   strcpy(fxName,__func__); 
   
   fprintf(outf,"@A In(%s) l4c_UnitTest_BPLUT diagnostic report~\n",fxName);
   
   for(r=0; r < nPftClasses; r++)
   {
     // l4c_GetBPLUT_Record has been re-verified using string keys "EBF" etc
     // bp_p = l4c_GetBPLUT_Record(SMAP_BPLUT_Keys[r], nPftClasses);
     // test fetch on both key as well as index
     bp_p = l4c_GetBPLUT_ByIndex(r, nPftClasses);
     
     if(bp_p != NULL)
     {
       xLogMsg("@D REC %2d   %d  lutIntegerIndex    ~\n",r, bp_p->lutIntegerIndex   );
       xLogMsg("@D REC %2d   %s  lutKey             ~\n",r, bp_p->lutKey            );
       xLogMsg("@D REC %2d   %s  useNDVItoFPAR      ~\n",r, bp_p->useNDVItoFPAR     );
       xLogMsg("@D REC %2d   %f  NDVItoFPAR_scale   ~\n",r, bp_p->NDVItoFPAR_scale  ); 
       xLogMsg("@D REC %2d   %f  NDVItoFPAR_offset  ~\n",r, bp_p->NDVItoFPAR_offset );
       xLogMsg("@D REC %2d   %f  LUEmax             ~\n",r, bp_p->LUEmax            );  
       xLogMsg("@D REC %2d   %f  Tmin_min           ~\n",r, bp_p->Tmin_min          );  
       xLogMsg("@D REC %2d   %f  Tmin_max           ~\n",r, bp_p->Tmin_max          );  
       xLogMsg("@D REC %2d   %f  VPD_min            ~\n",r, bp_p->VPD_min           );  
       xLogMsg("@D REC %2d   %f  VPD_max            ~\n",r, bp_p->VPD_max           );  
       xLogMsg("@D REC %2d   %f  SMrz_min           ~\n",r, bp_p->SMrz_min          );  
       xLogMsg("@D REC %2d   %f  SMrz_max           ~\n",r, bp_p->SMrz_max          );  
       xLogMsg("@D REC %2d   %f  FT_min             ~\n",r, bp_p->FT_min            );  
       xLogMsg("@D REC %2d   %f  FT_max             ~\n",r, bp_p->FT_max            );  
       xLogMsg("@D REC %2d   %f  SMtop_min          ~\n",r, bp_p->SMtop_min         );  
       xLogMsg("@D REC %2d   %f  SMtop_max          ~\n",r, bp_p->SMtop_max         );  
       xLogMsg("@D REC %2d   %f  Tsoil_beta0        ~\n",r, bp_p->Tsoil_beta0       );  
       xLogMsg("@D REC %2d   %f  Tsoil_beta1        ~\n",r, bp_p->Tsoil_beta1       );  
       xLogMsg("@D REC %2d   %f  Tsoil_beta2        ~\n",r, bp_p->Tsoil_beta2       );  
       xLogMsg("@D REC %2d   %f  fraut              ~\n",r, bp_p->fraut             );  
       xLogMsg("@D REC %2d   %f  fmet               ~\n",r, bp_p->fmet              );  
       xLogMsg("@D REC %2d   %f  fstr               ~\n",r, bp_p->fstr              );  
       xLogMsg("@D REC %2d   %f  kopt               ~\n",r, bp_p->kopt              );  
       xLogMsg("@D REC %2d   %f  kstr               ~\n",r, bp_p->kstr              );  
       xLogMsg("@D REC %2d   %f  kslw               ~\n",r, bp_p->kslw              );  
       xLogMsg("@D REC %2d   %u  Nee_QA_Rank_min    ~\n",r, bp_p->Nee_QA_Rank_min   );  
       xLogMsg("@D REC %2d   %u  Nee_QA_Rank_max    ~\n",r, bp_p->Nee_QA_Rank_max   );  
       xLogMsg("@D REC %2d   %f  Nee_QA_Error_min   ~\n",r, bp_p->Nee_QA_Error_min  );  
       xLogMsg("@D REC %2d   %f  Nee_QA_Error_max   ~\n",r, bp_p->Nee_QA_Error_max  );  
       xLogMsg("@D REC %2d   %u  Fpar_QA_Rank_min   ~\n",r, bp_p->Fpar_QA_Rank_min  );  
       xLogMsg("@D REC %2d   %u  Fpar_QA_Rank_max   ~\n",r, bp_p->Fpar_QA_Rank_max  );  
       xLogMsg("@D REC %2d   %f  Fpar_QA_Error_min  ~\n",r, bp_p->Fpar_QA_Error_min );  
       xLogMsg("@D REC %2d   %f  Fpar_QA_Error_max  ~\n",r, bp_p->Fpar_QA_Error_max );  
       xLogMsg("@D REC %2d   %f  FtMethod_QA_mult   ~\n",r, bp_p->FtMethod_QA_mult  );  
       xLogMsg("@D REC %2d   %u  FtAge_QA_Rank_min  ~\n",r, bp_p->FtAge_QA_Rank_min );  
       xLogMsg("@D REC %2d   %u  FtAge_QA_Rank_max  ~\n",r, bp_p->FtAge_QA_Rank_max );  
       xLogMsg("@D REC %2d   %f  FtAge_QA_Error_min ~\n",r, bp_p->FtAge_QA_Error_min);  
       xLogMsg("@D REC %2d   %f  FtAge_QA_Error_max ~\n",r, bp_p->FtAge_QA_Error_max);  
       xLogMsg("@D REC %2d   %f  Par_QA_Error       ~\n",r, bp_p->Par_QA_Error      );  
       xLogMsg("@D REC %2d   %f  Tmin_QA_Error      ~\n",r, bp_p->Tmin_QA_Error     );  
       xLogMsg("@D REC %2d   %f  Vpd_QA_Error       ~\n",r, bp_p->Vpd_QA_Error      );
       xLogMsg("@D REC %2d   %f  Smrz_QA_Error      ~\n",r, bp_p->Smrz_QA_Error     );  
       xLogMsg("@D REC %2d   %f  Tsoil_QA_Error     ~\n",r, bp_p->Tsoil_QA_Error    );  
       xLogMsg("@D REC %2d   %f  Smtop_QA_Error     ~\n",r, bp_p->Smtop_QA_Error    );
     } // end:: end if block    

 } // end:: end loop
    
 return r ;
 
} // end:: l4c_BPLUT_Report()


#ifdef BPLUT_UNIT_TEST
#define N_TESTKEYS 12
char *testKeys[] ={"ENF","EBF","DNF","DBF","MF",
  "CShrub","OShrub","WSavannas","Savannas","Grass","Crop","Crop"} ;

void l4c_UnitTest_BPLUT(const int nRec)
{
 BPLUT *bp_p = NULL;
 int r=0;
 
   fprintf(stdout,"\nl4c_UnitTest_BPLUT\n");
   for(r=0; r < N_TESTKEYS; r++)
   {
     bp_p = l4c_GetBPLUT_Record(testKeys[r],nRec);
     if( bp_p != NULL)
     {
          fprintf(stdout,"Key(%12s) FPARcode %4d VPD_max %lf CUE %lf\n",
       bp_p->lutKey, bp_p->FPAR_code, bp_p->VPD_max,bp_p->CUE);
     }
   } // end::r loop
}// end:: l4c_UnitTest_BPLUT()
#endif 
