// ---------------------------------------------------------------
// module       : l4c_DAILY_hdf5.c
// purpose      : Implement general HDF5 services for L4_C.  
// project      : SMAP L4C science development.
// dependencies : HDF4 4.2.x, HDF5 v1.8.x or later libraries, L4C FRAMEWORK APIs
// revised      : 2021-11-01 Changed Metadata/Source/MODIS_FPAR to Metadata/Source/FPAR
// revised      : 2020-08-06T11:23, added write of missing_value , paired with _Fillvalue write.
// revised      : 2016-03-03T15:43, added QA Bitfield legend back in , as this
//                was mistakenly deprecated when carbon_model_qual was deprecated.
// revised      : 2016-02-11T14:01, updated to call l4c_Hdf5_AssignFillValue()
//                to improve consistency of fillvalues matched to dataset type
// history      : 2020-08-06, added write of missing_value, paired with _Fillvalue write. 
// history      : 2016-02-11T10:23:12 deprecated creation and output of all (8)
// history      : 2016-02-10T16:12 deprecated creation and output of all (8)
//                prior QA "carbon_model_qual_pft{1..8} variables, replaced by RMSE variables.
// history      : 2014-09-10T14:30, implement ISO metadata, file-level attributes
//                (and dataset attributes like 'coordinates' as per Joe A request.
// history      : 2014-09-03T12:34, removed all separate QA HDF5 file
//                logic. Removed  l4c_Hdf5_CreateQa() function and
//                added int l4c_H5DatatypeToTIVDatatype(const hid_t h5tDatatype)

// history      : 2014-08-13T13:36, changed L4C_PRODUCTION_UTC_DATETIME
//                back to L4C_PRODUCTION_UTC_DATETIME to preserve Joe A's
//                Perl parser logic on post-production ISO metadata injection.
// history      : 2014-08-07T16:51, adapting for 3km FT_STATE_UM_M03 and
//                9km SMAP NatureRun smrz  {min|max} saturated scaling images
// history      : 2014-06-23T17:33, adding creation of formal metadata
//                /Metadata, "group-of-groups" and child attributes under each.
// history      : 2014-05-21, changed RELEASE_TAG to SUBSYS_RELEASE_TAG to
//                better represent the tight coupling between all (3)
//                L4_C suite components (l4c_MET,l4c_MODIS,l4c_DAILY) and
//                the ancillary file(s) these depend on.
// history      : 2014-05-12T13:37:00,jmg, corrected duplicate UUID tag gen.
// history      : 2014-02-16T16:06:00, re-integrate NSB NVP mgr, a FRM_API client API
// history      : 2014-02-13T09:41:00, patched in frm_utilGetCtime().
// history      : 2014-01-27T12:17:00-07:00
// history      : 2014-12-09T12:17:00-07:00
// history      : 2014-12-11T12:17:00-07:00
// history      : 2013-11-25T17:38:00-07:00 
//              : 2013-11-18T09:36:00
// author       : jglassy
// notes        : Namespace signature: :"l4c_Hdf5_XXXXX"
// ---------------------------------------------------------------

#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"


static Hdf5GridPropAdt  hGridPropObj ;

/* REMINDER: Qa bitflags "legend" definitions: */
 
// ---------------------------------------------------------------

#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"

// --------------------------------------------------------------------
// function : l4c_GetHdf5GridProp(void)
// Purpose  : Return ptr to private master Hdf5GridPropAdt containing
//            HDF5 specific id mechanism -- e.g. the dpsace and dcpl's
//            for all supported grid configurations.
// Inputs   : void
// returns  : Hdf5GridPropAdt *h_p -ptr to Grid properties object
Hdf5GridPropAdt *l4c_GetHdf5GridProp(void)
{
    return &hGridPropObj ;
    
}  // end:: l4c_GetHdf5GridProp()


 
// --------------------------------------------------------------------
// function :l4c_DatatypeStringTo_Hdf4TypeId(
// Purpose  : Convert a HDF5 style datatype string identifier ("int32")to a
//            HDF4 style integer datatype descriptor like DFNT_INT32
// Inputs   : const char *dtype_str
// returns  : int , HDF4|TIV datatype identifier
// revision : 2014-09-02T15:55:00
int l4c_DatatypeStringTo_Hdf4TypeId(const char *dtype_str)
{
  int32   dataType =-1;
  char fxName[256] = {0} ;

    strcpy(fxName,__func__);
    
    if (strcmp("int8",dtype_str)==0)    dataType = DFNT_INT8;
    else if (strcmp("uint8"  , dtype_str)==0)    dataType = DFNT_UINT8   ;
    else if (strcmp("int16"  , dtype_str)==0)    dataType = DFNT_INT16   ;
    else if (strcmp("uint16" , dtype_str)==0)    dataType = DFNT_UINT16  ;    
    else if (strcmp("int32"  , dtype_str)==0)    dataType = DFNT_INT32   ;
    else if (strcmp("uint32" , dtype_str)==0)    dataType = DFNT_UINT32  ;    
    else if (strcmp("float32", dtype_str)==0)    dataType = DFNT_FLOAT32 ;
    else if (strcmp("flt32"  , dtype_str)==0)    dataType = DFNT_FLOAT32 ;  
    else if (strcmp("float64", dtype_str)==0)    dataType = DFNT_FLOAT64 ;  
    else if (strcmp("flt64"  , dtype_str)==0)    dataType = DFNT_FLOAT64 ;  
  else {    
    dataType = DFNT_NONE;
    xLogMsg("@F In(%s) ERROR: undefined data type: %s\n",fxName,dtype_str);
  }
  
  return(dataType);
} // end:: l4c_DatatypeStringTo_Hdf4TypeId()


// --------------------------------------------------------------------
// function :l4c_TIVDatatypeToDatatypeString()
// Purpose  : Convert a TIV type to a HDF4 style datatype string
//            identifier ("int32")
// Inputs   : const int TIV_datatypeId
char *l4c_TIVDatatypeToDatatypeString(const int TIV_datatypeId,char *typBuf)
{

    switch(TIV_datatypeId)
    {
    case IsInt8    : strcpy(typBuf,"int8"   ); break;
    case IsUint8   : strcpy(typBuf,"uint8"  ); break;
    case IsInt16   : strcpy(typBuf,"int16"  ); break;
    case IsUint16  : strcpy(typBuf,"uint16" ); break;
    case IsInt32   : strcpy(typBuf,"int32"  ); break;
    case IsUint32  : strcpy(typBuf,"uint32" ); break;
    case IsInt64   : strcpy(typBuf,"int64"  ); break;
    case IsUint64  : strcpy(typBuf,"uint64" ); break;
    case IsFloat32 : strcpy(typBuf,"float32"); break;
    case IsFloat64 : strcpy(typBuf,"float64"); break;
    default :
    xLogMsg("@E ERROR: l4c_TIVDatatypeToDatatypeString unsupported TIV type(%d)\n", 
                        TIV_datatypeId);
    
    } // end:: TIV_datatypeId
    
    return typBuf ;
} // end::  l4c_TIVDatatypeToDatatypeString()


// --------------------------------------------------------------------
// function : l4c_H5TDatatypeToTIVDatatype()
// Purpose  : Convert an HDF5 (native) datatype to a HDF4 integer form
//            identifier ("int32")
// Inputs   : const int TIV_datatypeId
// Notes    : only very common numeric types are currently supported.
// revision : 2014-09-03T16:14:00, added.
int l4c_H5TDatatypeToTIVDatatype(const hid_t h5tDatatype)
{
    int tivDatatype = -1;
    char fxName[256];
    
    strcpy(fxName,__func__);

    if( h5tDatatype == H5T_NATIVE_CHAR)
        tivDatatype = IsInt8;
    else if(h5tDatatype == H5T_NATIVE_UINT8 )
        tivDatatype = IsUint8;
    else if(h5tDatatype == H5T_NATIVE_SHORT )
        tivDatatype = IsInt16 ;
    else if(h5tDatatype == H5T_NATIVE_USHORT )
        tivDatatype = IsUint16 ;
    else if(h5tDatatype == H5T_NATIVE_LONG )
        tivDatatype = IsInt32 ;
    else if(h5tDatatype == H5T_NATIVE_ULONG )
        tivDatatype = IsUint32 ;
    else if(h5tDatatype == H5T_NATIVE_LLONG )
        tivDatatype = IsInt64 ;
    else if(h5tDatatype == H5T_NATIVE_ULLONG )
        tivDatatype = IsUint64 ;
    else if(h5tDatatype == H5T_NATIVE_FLOAT )
        tivDatatype = IsFloat32 ;
    else if(h5tDatatype == H5T_NATIVE_DOUBLE )
        tivDatatype = IsFloat64;
    else
        xLogMsg("@E In(%s) ERROR unsupported H5T datatype(%d) ~\n",fxName, h5tDatatype);

    return tivDatatype ;
} // end::  l4c_H5TDatatypeToTIVDatatype()


// --------------------------------------------------------------------
// function : l4c_TIVDatatypeToH5TDatatype()
// Purpose  : Convert a TIV (and HDF4) datatype identifier like DFNT_FLOAT32
//            to a HDF5 (native) datatype indentier like H5T_NATIVE_FLOAT32
// Inputs   : const int tivDatatype, like IsFloat32 or DFNT_FLOAT32
// Notes    : only very common numeric types are currently supported.
// revision : 2014-09-03T16:14:00, added.
hid_t l4c_TIVDatatypeToH5TDatatype(const int tivDatatype)
{
 hid_t resultH5T_datatype =0;
 char fxName[256];

    strcpy(fxName,__func__);
 
    switch(tivDatatype)
    {
    case IsInt8    : resultH5T_datatype = H5T_NATIVE_INT8    ; break;
    case IsUint8   : resultH5T_datatype = H5T_NATIVE_UINT8   ; break;
    case IsInt16   : resultH5T_datatype = H5T_NATIVE_SHORT   ; break;
    case IsUint16  : resultH5T_datatype = H5T_NATIVE_USHORT  ; break;
    case IsInt32   : resultH5T_datatype = H5T_NATIVE_LONG    ; break;
    case IsUint32  : resultH5T_datatype = H5T_NATIVE_ULONG   ; break;
    case IsInt64   : resultH5T_datatype = H5T_NATIVE_LLONG   ; break;
    case IsUint64  : resultH5T_datatype = H5T_NATIVE_ULLONG  ; break;
    case IsFloat32 : resultH5T_datatype = H5T_NATIVE_FLOAT   ; break;
    case IsFloat64 : resultH5T_datatype = H5T_NATIVE_DOUBLE  ; break;
    default :
    xLogMsg("@E In(%s) ERROR: unsupported TIV type(%d)~\n", 
        fxName, tivDatatype);
    } // end:: switch on tivDatatype

    return resultH5T_datatype ;
    
} // end::  l4c_TIVDatatypeToH5TDatatype()


// ---------------------------------------------------------------------
// function: l4c_Hdf5_AssignFillValue()
// revised : 2020-08-06T11:22, added write of missing_value
// revised : 2016-02-11T10:39:00 (created)
// author  : Joe Glassy
// inputs  : hid_t dset_id -- unique HDF5 dataset ID assigned prior
//           PkgObj *pObj_p -- ptr to current dataset's PkgObj avail upstream
//           DEPRECATED: char *datasetName_p -- ptr to datasets fully qualified name e.g. "NEE/nee_mean"
// returns : int, tivDatatype of processed dataset
// notes   : Added write of missing_value attribute, since we use _Fillvalue's value for this.
// notes   : For future hid_t compatibility added a case for INT64_FILL_VALUE, and
//           simplified call signature by passing in prior-available pObj_p
//           instead of performing redundant lookup using GetObjByName()
//           as prior.
// notes   : A series of calls are made to H5Pset_fill_value() based
//           on the particular HDF5 Access Property List (plist) applying
//           to each given datasetName. The goal is correctly match the
//           datatype of the written _FillValue (and internal fillvalue
//           written) specific to each dataset, upon creation.
int l4c_Hdf5_AssignFillValue(hid_t dset_id,PkgObj *pObj_p)
{
 SesCtl          *sesCtl_p = l4c_GetSesCtl();
 Hdf5GridPropAdt *hGridPropObj_p = l4c_GetHdf5GridProp();
 PkgMgr *pkgMgr_p  = sesCtl_p->l4cPkgMgr_p;
 // PkgTheme *pThme_p = NULL;
 // PkgObj   *pObj_p  = NULL;
 char fxName[256] = {0} ;
 char *dsetName_p = NULL ;
   
 hid_t   hRet = -1;
 hid_t   ret = -1 ;
 
 hid_t   plist1km = hGridPropObj_p->dcpl_1km  ; // property list for 1km variable
 hid_t   plist3km = hGridPropObj_p->dcpl_3km  ; // property list for 3km variable

 hid_t   plist9km = hGridPropObj_p->dcpl_9km; // property list for majority, 9km variables
 hid_t   plist9km_uint8  = hGridPropObj_p->dcpl_9km_uint8;
 hid_t   plist9km_int16  = hGridPropObj_p->dcpl_9km_int16; 
 hid_t   plist9km_uint16 = hGridPropObj_p->dcpl_9km_uint16; 
 hid_t   plist9km_int32  = hGridPropObj_p->dcpl_9km_int32; 
 hid_t   plist9km_int64  = hGridPropObj_p->dcpl_9km_int64; // added 2020-07-14
 hid_t   plist9km_uint32 = hGridPropObj_p->dcpl_9km_uint32; 

 uint8   byteFill   = BYTE_FILL_VALUE   ; // in l4c_DAILY.h, usually, 254
 int16   i16Fill    = SHORT_FILL_VALUE  ; // placeholder, until updated
 uint16  u16Fill    = USHORT_FILL_VALUE ; // 65534 from l4c_DAILY.h
 int64   int64Fill  = INT64_FILL_VALUE  ; // added 2020-07-14,65534, in l4c_DAILY.h, until updated
 int32   int32Fill  = INT32_FILL_VALUE  ; // Placeholder, 65534, in l4c_DAILY.h, until updated
 uint32  uint32Fill = UINT32_FILL_VALUE ; // Placeholder, 65534, in l4c_DAILY.h, until updated
 float32 fltFill    = FLOAT_FILL_VALUE  ;       // usually, -9999.0  
 float64 dblFill    = (double)FLOAT_FILL_VALUE; // same as for float32
 
 int tivDatatype = 0 ;
 int tivSizOfOne = 0 ;
 char fillvalueVarName[16] = "_FillValue" ;
 // <JMG> 2020-08-06T11:17 added to provide for missing_value write
 char missingvalueVarName[16] = "missing_value" ;
 char tivLabel[32]  = {0} ;
 
 strcpy(fxName,__func__);
 
 /* DEPRECATED 2020-07-14
 // first, retrieve the THEME for this collection (e.g. "L4C_OUTPUTS") of objects
 pThme_p  = pkgMgr_GetThemeByName(pkgMgr_p,"L4C_OUTPUTS");
 // next, retrieve the child object for THIS dataset within this THEME
 // <JMG> 2020-07-014T10:11 Apparent ERROR Here, FAILED TO find name(<correctName>) so
 // currently debugging why...
 pObj_p = pkgMgr_GetObjByName(pThme_p, datasetName_p);
 */ 
 
 dsetName_p = pkgMgr_GetObjName(pObj_p);
  
 // next, retrieve the TIV datatype (e.g .ike DFNT_FLOAT32) for this dataset
 tivDatatype = pkgMgr_GetObjDatatype(pObj_p);

 // <JMG> 2020-08-06T11:18, now write BOTH "_Fillvalue" and also "missing_value"
 switch(tivDatatype)
 {
     case DFNT_INT8   :
     case DFNT_UINT8  : hRet = H5Pset_fill_value(plist9km_uint8, H5T_NATIVE_UINT8,&byteFill);
                        ret = l4c_Hdf5_CreateAttribute(dset_id, IsUint8, ".", fillvalueVarName ,&byteFill);
                        ret = l4c_Hdf5_CreateAttribute(dset_id, IsUint8, ".", missingvalueVarName ,&byteFill);
                          break;
     case DFNT_INT16  : hRet = H5Pset_fill_value(plist9km_int16,  H5T_NATIVE_INT16,&i16Fill); 
                        ret = l4c_Hdf5_CreateAttribute(dset_id, IsInt16, ".", fillvalueVarName,    &i16Fill);
                        ret = l4c_Hdf5_CreateAttribute(dset_id, IsInt16, ".", missingvalueVarName ,&i16Fill);                        
                          break;
     case DFNT_UINT16 : hRet = H5Pset_fill_value(plist9km_uint16, H5T_NATIVE_USHORT, &u16Fill); 
                        ret = l4c_Hdf5_CreateAttribute(dset_id, IsUint16, ".", fillvalueVarName,&u16Fill);
                        ret = l4c_Hdf5_CreateAttribute(dset_id, IsUint16, ".", missingvalueVarName,&u16Fill);
                          break;
     case DFNT_INT32  : hRet = H5Pset_fill_value(plist9km_int32 , H5T_NATIVE_LONG,&int32Fill); 
                        ret = l4c_Hdf5_CreateAttribute(dset_id, IsInt32, ".", fillvalueVarName   ,&int32Fill);
                        ret = l4c_Hdf5_CreateAttribute(dset_id, IsInt32, ".", missingvalueVarName,&int32Fill);     
                          break;
     case DFNT_UINT32 : hRet = H5Pset_fill_value(plist9km_uint32, H5T_NATIVE_ULONG,&uint32Fill); 
                        ret = l4c_Hdf5_CreateAttribute(dset_id, IsUint32, ".",fillvalueVarName ,&uint32Fill);
                        ret = l4c_Hdf5_CreateAttribute(dset_id, IsUint32, ".",missingvalueVarName ,&uint32Fill);
                          break;
     // <JMG> 2020-07-14T10:43 added just in case a int64_t is now used
     case DFNT_INT64  : hRet = H5Pset_fill_value(plist9km_int64, H5T_NATIVE_LLONG,&int64Fill); 
                        ret = l4c_Hdf5_CreateAttribute(dset_id, IsInt64, ".", fillvalueVarName   ,&int64Fill);
                        ret = l4c_Hdf5_CreateAttribute(dset_id, IsInt64, ".", missingvalueVarName,&int64Fill);     
                          break;
     case DFNT_FLOAT32: hRet = H5Pset_fill_value(plist9km, H5T_NATIVE_FLOAT, &fltFill);
                        ret = l4c_Hdf5_CreateAttribute(dset_id, IsFloat32, ".", fillvalueVarName,&fltFill);
                        ret = l4c_Hdf5_CreateAttribute(dset_id, IsFloat32, ".", missingvalueVarName,&fltFill);
                          break;     
     case DFNT_FLOAT64: hRet = H5Pset_fill_value(plist9km, H5T_NATIVE_DOUBLE, &dblFill);
                        ret = l4c_Hdf5_CreateAttribute(dset_id, IsFloat64, ".", fillvalueVarName,&dblFill);
                        ret = l4c_Hdf5_CreateAttribute(dset_id, IsFloat64, ".", missingvalueVarName,&dblFill);                        
                          break;
 } // end:: switch() 
 
 // fetch readable NAME of this tivDatatype   
 strcpy(tivLabel, tiv_NameofTivType(tivDatatype,&tivSizOfOne));
 
 // <JMG> 2020-07-14 re-enabled for potential error diagnostics
 // xLogMsg("@D In(%s) missing_value written for dataset %s (type %d %s) at ID(%ld) hRet code %d ~\n",fxName,dsetName_p, tivDatatype,tivLabel, dset_id, (int)hRet);
 // xLogMsg("@D In(%s) FillValue written for dataset %s (type %d %s) at ID(%ld) hRet code %d ~\n",fxName,dsetName_p, tivDatatype,tivLabel, dset_id, (int)hRet);
    
 return tivDatatype ;
 
} // end:: l4c_Hdf5_AssignFillValue()

// ----------------------------------------------------------------------
// function : l4c_Hdf5_CreateDatasets()
// purpose  : Create initial L4_C OUTPUT file DATASETs
// inputs   : void
// returns  : int,status
// revision : 2016-02-10T16:12 deprecated creation and output of all (8)
//            prior QA "carbon_model_qual_pft{1..8} variables, since RMSE
//            replaces these.
// revision : 2014-08-14T17:58
// notes    : This legacy method manually enumerates each and every
//            output dataset, which is error prone and cumbersome.
//            A future proposed listwise method takes info from RIP file and
//            iterates these assignments in a loop, but it is still in
//            testing.
// notes    : called by: l4c_Hdf5_CreateOutput()
// notes    : the main L4_C and QA OUTPUT HDF5 files must have 
//            already been created and are open when this is called.
// notes    : Bug w016-01-07 is now found and corrected, where earlier
//            corruption of sesCtl->runDayOfYear occurred.
int l4c_Hdf5_CreateDatasets(hid_t file_id)
{
 SesCtl   *sesCtl_p = l4c_GetSesCtl();
 Hdf5GridPropAdt *hGridPropObj_p = l4c_GetHdf5GridProp();
 NsbObj   *nsb_p   = l4c_GetNsbOverseer();
 PkgMgr   *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
 PkgTheme *pThme_p = NULL;
 PkgObj   *pObj_p  = NULL;
  
 char **dsetOutputNames_p = NULL;
 char *dsetName_p = NULL;
 char *themeName_p= NULL;
 char fxName[256] = {0} ;
 char dsetAttribValue[FILENAME_MAX] = {0} ;
 char flagMasks[32]      = {0} ; // added to match flagMasks
 char flagMeanings[256]  = {0} ; // added to match flag_meanings
 char nvpResultName[512] = {0} ; // actual fully qualified name of a HDF5 dataset
 char tivTypeAsStr[64]   = {0} ;
 
 // <JMG> 2020-07-13 extra diagnostics for hid_t checking
 hid_t childsId = -1 ; 
 char buffer[256] ={0} ;
 char childName[256] = {0} ;
 
 int idx = 0 ;
 int nChildren =0 ;
 
 hid_t dType = -1 ;
 hid_t dcpl = -1;
 hid_t dspace = -1;
 hid_t dset_id = -1 ;
 hid_t chk_id = -1 ;
 hid_t ret = -1;
 hid_t chk_dset_id = 0;

 int startIndex =0;
 int endIndex   =0;
 int nEntries = 0;
 int tivDatatype = 0 ; // <JMG> 2016-02-11t12:26 added ...
 
 int ithIndex  = 0 ;

 char coordLatLonRef[64] = {0};
 float64 dblValue =0.0;
 
    strcpy(fxName,__func__);

    // fetch pointer to the theme "L4C_OUTPUTS" ...
    pThme_p     = pkgMgr_GetThemeByName(pkgMgr_p,"L4C_OUTPUTS");
    // retrieve name of theme, to confirm it
    themeName_p = pkgMgr_GetThemeName(pThme_p);

    // Establish a loop count, the number of PkgObj (child) one per dataset
    nChildren = pkgMgr_GetThemeObjectCount(pThme_p);

    xLogMsg("@A In(%s) CREATING %d L4_C OUTPUT GRANULE DATASET definitions from theme %s~\n",
         fxName, nChildren, themeName_p);

    // for now  we'll hard code in the dspace and dcpl

    // all L4_SM and GEOS-5 "float32" variables are handled in this block
    
    // Note: most externally defined properties of a dataset can now be retrieved
    // directly, using a RIP keyword for the datasetname...or, these can be
    // retrieved from the object's PkgMgr theme
    
    // EXAMPLE of one (future) idiom to use RIP keyword to resolve an 
    // actual fully qualified HDF5 dataset name..
    // nsbGetPropertyValueByKey(nsb_p, "L4C_OUTPUT_VARIABLE_NEE_MEAN", IsPropertyName,nvpResultName);
    // nsbGetPropertyValueByKey(nsb_p, "L4C_OUTPUT_VARIABLE_NEE_MEAN", IsPropertyDatatype,tivTypeAsStr);
    
    // DATATYPE Convert string form "float32" in RIP file, to integer code like IsFloat32 to store
    //      tivDatatype = l4c_DatatypeStringTo_Hdf4TypeId(tivTypeAsStr);
    // nsbGetPropertyValueByKey(nsb_p, "L4C_OUTPUT_VARIABLE_NEE_MEAN", IsPropertyName,nvpResultName);
    //   dset_id =H5Dcreate(file_id, nvpResultName, tivDatatype,
    //      hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    
    xLogMsg("@A In(%s) CREATING NEE, GPP, RH, SOC and EC output datasets ~\n",fxName);
    
    dset_id =H5Dcreate(file_id,"NEE/nee_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    // now, we store the child's dataset ID in its primary ID field
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "NEE/nee_mean", dset_id );
    
    // <JMG> 2020-07-13 immediately verify entry was stored corredtly
    chk_id = pkgMgr_GetThemes_ChildIdByName(pThme_p,"NEE/nee_mean");
    xLogMsg("@A In(%s) NEE/nee_mean VERIFYING NEW dataset ID (%ld)~ \n",fxName, chk_id);


    dset_id  =H5Dcreate(file_id,"NEE/nee_std_dev",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "NEE/nee_std_dev", dset_id  );

    dset_id = H5Dcreate(file_id,"NEE/nee_pft1_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "NEE/nee_pft1_mean", dset_id );
    
    dset_id = H5Dcreate(file_id,"NEE/nee_pft2_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "NEE/nee_pft2_mean", dset_id );
    
    dset_id  =H5Dcreate(file_id,"NEE/nee_pft3_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "NEE/nee_pft3_mean", dset_id );
    
    dset_id  =H5Dcreate(file_id,"NEE/nee_pft4_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "NEE/nee_pft4_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"NEE/nee_pft5_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "NEE/nee_pft5_mean", dset_id );
    
    dset_id  =H5Dcreate(file_id,"NEE/nee_pft6_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "NEE/nee_pft6_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"NEE/nee_pft7_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "NEE/nee_pft7_mean", dset_id );
    
    dset_id  =H5Dcreate(file_id,"NEE/nee_pft8_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "NEE/nee_pft8_mean", dset_id ); 
    
    dset_id  =H5Dcreate(file_id,"GPP/gpp_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "GPP/gpp_mean", dset_id );
    
    dset_id  =H5Dcreate(file_id,"GPP/gpp_std_dev",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "GPP/gpp_std_dev", dset_id );
    
    dset_id  =H5Dcreate(file_id,"GPP/gpp_pft1_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "GPP/gpp_pft1_mean", dset_id );
    
    dset_id  =H5Dcreate(file_id,"GPP/gpp_pft2_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "GPP/gpp_pft2_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"GPP/gpp_pft3_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "GPP/gpp_pft3_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"GPP/gpp_pft4_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "GPP/gpp_pft4_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"GPP/gpp_pft5_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "GPP/gpp_pft5_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"GPP/gpp_pft6_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "GPP/gpp_pft6_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"GPP/gpp_pft7_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "GPP/gpp_pft7_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"GPP/gpp_pft8_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "GPP/gpp_pft8_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"RH/rh_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "RH/rh_mean", dset_id );
    
    dset_id  =H5Dcreate(file_id,"RH/rh_std_dev",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "RH/rh_std_dev", dset_id );
    
    dset_id  =H5Dcreate(file_id,"RH/rh_pft1_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "RH/rh_pft1_mean", dset_id );
    
    dset_id  =H5Dcreate(file_id,"RH/rh_pft2_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "RH/rh_pft2_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"RH/rh_pft3_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "RH/rh_pft3_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"RH/rh_pft4_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "RH/rh_pft4_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"RH/rh_pft5_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "RH/rh_pft5_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"RH/rh_pft6_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "RH/rh_pft6_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"RH/rh_pft7_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "RH/rh_pft7_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"RH/rh_pft8_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "RH/rh_pft8_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"SOC/soc_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "SOC/soc_mean", dset_id );
    
    dset_id  =H5Dcreate(file_id,"SOC/soc_std_dev",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "SOC/soc_std_dev", dset_id );
    
    dset_id  =H5Dcreate(file_id,"SOC/soc_pft1_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "SOC/soc_pft1_mean", dset_id );
    
    dset_id  =H5Dcreate(file_id,"SOC/soc_pft2_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "SOC/soc_pft2_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"SOC/soc_pft3_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "SOC/soc_pft3_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"SOC/soc_pft4_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "SOC/soc_pft4_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"SOC/soc_pft5_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "SOC/soc_pft5_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"SOC/soc_pft6_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "SOC/soc_pft6_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"SOC/soc_pft7_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "SOC/soc_pft7_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"SOC/soc_pft8_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "SOC/soc_pft8_mean", dset_id );

    dset_id =H5Dcreate(file_id,"EC/frozen_area",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "EC/frozen_area", dset_id );
    
    // <JMG> 2016-02-18, changed all mult_av to mult_mean to match Lucas notation
    dset_id  =H5Dcreate(file_id,"EC/emult_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "EC/emult_mean", dset_id );
    
    dset_id  =H5Dcreate(file_id,"EC/tmult_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "EC/tmult_mean", dset_id );
    
    dset_id  =H5Dcreate(file_id,"EC/wmult_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "EC/wmult_mean", dset_id );
    
    // ----------------------------------------------------------------------------------
    // QA - CREATE ALL DATASET definitions
    xLogMsg("@A In(%s) CREATING legacy QA output datasets ~\n",fxName);
    
    // <JMG> 2016-02-10T10: 5, QA carbon_model_qual datasets NOW DEPRECATED

    // Next, QA qa_count datasets created here

    dset_id  = H5Dcreate(file_id,"QA/qa_count_pft1", H5T_NATIVE_UINT8,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km_uint8 , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "QA/qa_count_pft1", dset_id );

    dset_id  = H5Dcreate(file_id,"QA/qa_count_pft2", H5T_NATIVE_UINT8,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km_uint8 , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "QA/qa_count_pft2", dset_id );

    dset_id  = H5Dcreate(file_id,"QA/qa_count_pft3", H5T_NATIVE_UINT8,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km_uint8 , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "QA/qa_count_pft3", dset_id );

    dset_id  = H5Dcreate(file_id,"QA/qa_count_pft4", H5T_NATIVE_UINT8,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km_uint8 , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "QA/qa_count_pft4", dset_id );

    dset_id  = H5Dcreate(file_id,"QA/qa_count_pft5", H5T_NATIVE_UINT8,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km_uint8 , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "QA/qa_count_pft5", dset_id );
    
    dset_id  = H5Dcreate(file_id,"QA/qa_count_pft6", H5T_NATIVE_UINT8,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km_uint8 , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "QA/qa_count_pft6", dset_id );
    
    dset_id  = H5Dcreate(file_id,"QA/qa_count_pft7", H5T_NATIVE_UINT8,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km_uint8 , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "QA/qa_count_pft7", dset_id );
    
    dset_id  = H5Dcreate(file_id,"QA/qa_count_pft8", H5T_NATIVE_UINT8,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km_uint8 , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "QA/qa_count_pft8", dset_id );
    
    dset_id  = H5Dcreate(file_id,"QA/qa_count", H5T_NATIVE_UINT8,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km_uint8 , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "QA/qa_count", dset_id );

    // -----------------------------------------------------------------
    dset_id  = H5Dcreate(file_id,"QA/carbon_model_bitflag",H5T_NATIVE_USHORT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km_uint16 , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "QA/carbon_model_bitflag", dset_id );

    // -----------------------------------------------------------------
    xLogMsg("@A In(%s) CREATING (one) RMSE mean and (8) RMSE by PFT datasets ~\n",fxName);

    dset_id =H5Dcreate(file_id,"QA/nee_rmse_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    // now, we store the child's dataset ID in its primary ID field
    pkgMgr_SetThemes_ChildIdByName(pThme_p,"QA/nee_rmse_mean", dset_id );
    
    dset_id = H5Dcreate(file_id,"QA/nee_rmse_pft1_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "QA/nee_rmse_pft1_mean", dset_id );
    
    dset_id = H5Dcreate(file_id,"QA/nee_rmse_pft2_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p,"QA/nee_rmse_pft2_mean", dset_id );
    
    dset_id  =H5Dcreate(file_id,"QA/nee_rmse_pft3_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p,"QA/nee_rmse_pft3_mean", dset_id );
    
    dset_id  =H5Dcreate(file_id,"QA/nee_rmse_pft4_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p,"QA/nee_rmse_pft4_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"QA/nee_rmse_pft5_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p,"QA/nee_rmse_pft5_mean", dset_id );
    
    dset_id  =H5Dcreate(file_id,"QA/nee_rmse_pft6_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p,"QA/nee_rmse_pft6_mean", dset_id );

    dset_id  =H5Dcreate(file_id,"QA/nee_rmse_pft7_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "QA/nee_rmse_pft7_mean", dset_id );
    
    dset_id  =H5Dcreate(file_id,"QA/nee_rmse_pft8_mean",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "QA/nee_rmse_pft8_mean",dset_id );   

    // -----------------------------------------------------------------
    // create GEO group's lat and lon coord variables
    xLogMsg("@A In(%s) CREATING GEO latitude and longitude coordinate variable datasets ~\n",fxName);
    
    dset_id  =H5Dcreate(file_id,"GEO/latitude",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "GEO/latitude", dset_id );
    
    // VERIFY: immediately set, then verify, e.g. fetch HDF5 id, to echo
    chk_id = pkgMgr_GetThemes_ChildIdByName(pThme_p,"GEO/latitude");
    xLogMsg("@A In(%s) GEO/latitude NEW dataset id (%ld)~ \n",fxName, chk_id);
        
    dset_id  =H5Dcreate(file_id,"GEO/longitude",H5T_NATIVE_FLOAT,
         hGridPropObj_p->dspace_9km, H5P_DEFAULT, hGridPropObj_p->dcpl_9km , H5P_DEFAULT);
    pkgMgr_SetThemes_ChildIdByName(pThme_p, "GEO/longitude", dset_id );

    chk_id = pkgMgr_GetThemes_ChildIdByName(pThme_p,"GEO/longitude");
    xLogMsg("@A In(%s) GEO/longitude NEW dataset id (%ld)~ \n",fxName, chk_id);

    // ----------------------------------------------------------------
    // Pre-assign the rest of the dataset vital attribute properties
    // using l4c_DAILY_DatasetProperties.c --> l4c_AssignOutputDatasetProperties() function
    l4c_AssignOutputDatasetProperties() ;

    /*
    // <JMG> new diags to confirm all dataset ID's. Note these are correct here.
    for(ithIndex=0; ithIndex < nChildren; ithIndex++)
    {
        strcpy(childName,pkgMgr_GetNameByIndex(pThme_p,ithIndex,buffer));
        chk_id = pkgMgr_GetThemes_ChildIdByName(pThme_p,childName);
        xLogMsg("@A In(%s) CHILD-INVENTORY idx %d dsetId %ld (%s) ~\n",
            fxName,ithIndex,chk_id,childName);
    }
    */


    // WRITE core dataset attributes to ALL datasets requiring them, residing
    //   in the primary output HDF5 granule, identified by file_id
    // <JMG> 2016-03-03T17:02, added function l4c_MetadataWriteDataset() 
    //   inside l4c_DAILY_Writemetadata() to consolidate more metadata writing in one module...
    l4c_MetadataWriteDataset(file_id) ;
    
    l4c_EventCompletionMsg( (char *)__func__);

    return 0 ;
    
} // end:: l4c_Hdf5_CreateDatasets()


// ---------------------------------------------------------------------
// function : l4c_Hdf5_CloseSession()
// inputs   : Hdf5GridPropAdt *hGridPropObj_p,
//            const hid_t l4cMainOutput_file_id
// returns  : int,status
// revision : 2014-09-03T12:45:00, deprecated qa_file
// purpose  : formally close out all remaining HDF5 mechanisms that are 
//            open prior to session close.
//
int l4c_Hdf5_CloseSession(Hdf5GridPropAdt *hGridPropObj_p,const hid_t l4cMainOutput_file_id)
{
 herr_t status = -1;

    // DATASPACE close out
    status = H5Sclose(hGridPropObj_p->dspace_1km ) ; 
    status = H5Sclose(hGridPropObj_p->dspace_3km ) ; 
    status = H5Sclose(hGridPropObj_p->dspace_9km ) ; 

    // DCPL close out
    status = H5Pclose(hGridPropObj_p->dcpl_1km ) ; 
    status = H5Pclose(hGridPropObj_p->dcpl_3km ) ; 
    status = H5Pclose(hGridPropObj_p->dcpl_9km ) ; 
    
    status = H5Pclose(hGridPropObj_p->dcpl_9km_uint8 ) ; 
    status = H5Pclose(hGridPropObj_p->dcpl_9km_uint16 ) ; 

    // close out L4_C main and QA output file(s)
    H5Fclose( l4cMainOutput_file_id );

 return status;
 
} // end:: l4c_Hdf5_CloseSession()
    

// -----------------------------------------------------------------------------
// function : l4c_Hdf5_CreateGroup(owner_file_id)
// inputs   :const hid_t owner_file_id,char *groupLabel_p
// returns  : new GROUP id >0 if ok, or returns -1 if failed
// purpose  : create a group named 'groupLabel_p' e.g. "/SOC"
// notes    : A leading slash qualifier is permissible but not
//            required if the group is relative to the root "/" group.
//  hid_t H5Gcreate( hid_t loc_id, const char *name, hid_t lcpl_id, hid_t gcpl_id, hid_t gapl_id )
hid_t l4c_Hdf5_CreateGroup(const hid_t owner_file_id,char *groupLabel_p)
{
 hid_t new_group_id = -1;
    
    new_group_id = H5Gcreate(owner_file_id,groupLabel_p,
                H5P_DEFAULT,H5P_DEFAULT,H5P_DEFAULT); // lcpl,gcpl,gapl...
    
    xLogMsg("@A [%s] CREATE-GROUP <%s> at HDF5 object ID (%d)\n",
     (char *)__func__,  groupLabel_p,  new_group_id);
    
    return new_group_id ;
    
} // end:: l4c_Hdf5_CreateGroup()


// -----------------------------------------------------------------------------
// function : l4c_Hdf5_CreateAllOutputGroups(owner_file_id)
// inputs   : const hid_t owner_file_id,
//            const hid_t qa_file_id,
//            const int doCreateMetataGroups, if 1=do it, if 0 suppress.
//            hid_t *groupIdAry_p -- array of (open) HDF5 GROUP ID's after creation
// returns  : status
// outputs  : hid_t *groupIdAry_p -- array of new group ID's created, in
//            case these are needed later.
// purpose  : Create all HDF5 groups in BOTH the main SMAP_L4_C_MDL_xxxx .h5 and .qa file
// revision : 2014-09-03T12:40, added QA group to consolidate all QA into
//            the single daily "_mdl_" .h5 file.
// revision : 2014-06-23T17:35, added creation of formal metadata GROUPS
// notes    : Called by: l4c_Hdf5_CreateOutput(0
// notes    : Question to resolve: how much of /Metadata (attribs) need
//            to be echoed out to the .qa file?
int l4c_Hdf5_CreateAllOutputGroups(const hid_t owner_file_id,const int doCreateMetataGroups, hid_t *groupIdAry_p)
{
 // count of groups created. will be <= 8 if just the min. std are done
 int nGroupsCreated = 0 ;  
 char fxName[256] = {0} ;
 
 strcpy(fxName,__func__);
 
 // Create all groups in the single main SMAP_L4_C_MDL_xxxx .h5 file
 xLogMsg("@A In(%s) creating all HDF5 output granule GROUPS ~ \n",fxName);
 // currently these are hard wired since they vary so infrequently, but
 // ideally we want to make these externally defined as RIP driven entries
 groupIdAry_p[0] = frm_hdf5_CreateGroup( owner_file_id, "/NEE");
 groupIdAry_p[1] = frm_hdf5_CreateGroup( owner_file_id, "/GPP");
 groupIdAry_p[2] = frm_hdf5_CreateGroup( owner_file_id, "/RH");
 groupIdAry_p[3] = frm_hdf5_CreateGroup( owner_file_id, "/SOC");
 groupIdAry_p[4] = frm_hdf5_CreateGroup( owner_file_id, "/EC");
 groupIdAry_p[5] = frm_hdf5_CreateGroup( owner_file_id, "/QA");
 groupIdAry_p[6] = frm_hdf5_CreateGroup( owner_file_id, "/GEO");
 nGroupsCreated = 7 ; // base condition

 // NOTE: at NTSG is normally suppressed, since GMAO performs this as a postprocess
 if(doCreateMetataGroups)
 {
    // be sure now to increment nGroupsCreated once for each of these..
    
    // hid_t frm_hdf5_CreateGroup(hid_t loc_id,const char *newGroupName_p)
    groupIdAry_p[ 8]  = frm_hdf5_CreateGroup(owner_file_id,"/Metadata");
    groupIdAry_p[ 9]  = frm_hdf5_CreateGroup(owner_file_id,"/Metadata/AcquisitionInformation");
    groupIdAry_p[10]  = frm_hdf5_CreateGroup(owner_file_id,"/Metadata/AcquisitionInformation/Platform");
    groupIdAry_p[11]  = frm_hdf5_CreateGroup(owner_file_id,"/Metadata/AcquisitionInformation/PlatformDocument");
 
    groupIdAry_p[12]  = frm_hdf5_CreateGroup(owner_file_id,"/Metadata/CRID");
    groupIdAry_p[13]  = frm_hdf5_CreateGroup(owner_file_id,"/Metadata/CRID/MDL");
    groupIdAry_p[14]  = frm_hdf5_CreateGroup(owner_file_id,"/Metadata/CRID/Root");
 
    groupIdAry_p[15]  = frm_hdf5_CreateGroup(owner_file_id,"/Metadata/Config");
 
    groupIdAry_p[16]  = frm_hdf5_CreateGroup(owner_file_id,"/Metadata/DataQuality");
    groupIdAry_p[17]  = frm_hdf5_CreateGroup(owner_file_id,"/Metadata/DataQuality/L4C");
    groupIdAry_p[18]  = frm_hdf5_CreateGroup(owner_file_id,"/Metadata/DataQuality/L4C/CompletenessOmission");
    groupIdAry_p[19]  = frm_hdf5_CreateGroup(owner_file_id,"/Metadata/DataQuality/L4C/DomainConsistency");

    groupIdAry_p[20]  = frm_hdf5_CreateGroup(owner_file_id,"/Metadata/DatasetIdentification");
 
    groupIdAry_p[21]  = frm_hdf5_CreateGroup(owner_file_id,"/Metadata/Extent");
    // Attribs: {description,rangeBeginningDateTime,rangeEndingDateTime,
    //           eastBoundLongitude,northBoundLatitude, southBoundLatitude, westBoundLongitude}
 
    groupIdAry_p[22] = frm_hdf5_CreateGroup(owner_file_id,"/Metadata/GridSpatialRepresentation");
    // attribs: dimensionSize, resolution
    groupIdAry_p[23] = frm_hdf5_CreateGroup(owner_file_id,"/Metadata/GridSpatialRepresentation/Latitude");
    // attribs: dimensionSize, resolution
    groupIdAry_p[24] = frm_hdf5_CreateGroup(owner_file_id,"/Metadata/GridSpatialRepresentation/Longitude");
 
    groupIdAry_p[25] = frm_hdf5_CreateGroup(owner_file_id,"/Metadata/ProcessStep");
    // includes: SWVersionID,algorithmDescription,epochJulianDate,identifier,processor, softwareDate,
    // softwareTitle,stepDateTime,timeVariableEpoch , and, 
    // Echo most (all?) RIP file NVP key=value out as an attribute to ProcessStep
 
    groupIdAry_p[26] = frm_hdf5_CreateGroup(owner_file_id,"/Metadata/SeriesIdentification");

    // "Source" is like INPUTPOINTERS in ECS, list each major input 
    groupIdAry_p[27] = frm_hdf5_CreateGroup(owner_file_id,"/Source");
    
    // <JMG> 2014-09-03, re-order as per Joe A's recent additions {1..11}
    // attribs: list the var names in L4_C_MET.h5 file
    groupIdAry_p[28] = frm_hdf5_CreateGroup(owner_file_id,"/Source/L4_SM");
    groupIdAry_p[29] = frm_hdf5_CreateGroup(owner_file_id,"/Source/FPAR");
    groupIdAry_p[30] = frm_hdf5_CreateGroup(owner_file_id,"/Source/GEOS-5");
    groupIdAry_p[31] = frm_hdf5_CreateGroup(owner_file_id,"/Source/VIIRS_NDVI");
    groupIdAry_p[32] = frm_hdf5_CreateGroup(owner_file_id,"/Source/L3_SM_A");
    groupIdAry_p[33] = frm_hdf5_CreateGroup(owner_file_id,"/Source/L4_C_MET");
    groupIdAry_p[34] = frm_hdf5_CreateGroup(owner_file_id,"/Source/L4_C_MODIS");
    groupIdAry_p[35] = frm_hdf5_CreateGroup(owner_file_id,"/Source/L4_C_SOC");
    groupIdAry_p[36] = frm_hdf5_CreateGroup(owner_file_id,"/Source/SMAP_L4_C_FparClimatology");
    groupIdAry_p[37] = frm_hdf5_CreateGroup(owner_file_id,"/Source/SMAP_L4_C_Ancillary");
    groupIdAry_p[38] = frm_hdf5_CreateGroup(owner_file_id,"/Source/BPLUT");
    
    // <JMG> 2014-09-03t12:42, added group to respond to Joe Ardizzones request
    groupIdAry_p[39] = frm_hdf5_CreateGroup(owner_file_id,"/AcquisitionInformation");
    nGroupsCreated = 40 ; // full set of groups are usually just for development mode
    
 } // end:: block to create all Metadata groups
 
  // NOTE: L4_C does NOT currently define a "time" dataset like L4_SM does...
  //   in L4_SM, "time" is a FLOAT64 storing a seconds count =time(), and (4) attributes
  //   {actual_range,delta_t,long_name,units}

 l4c_EventCompletionMsg( (char *)__func__);

 // return count of HDF5 GROUPS created in output granule
 return nGroupsCreated ;
    
} // end:: l4c_CreateAllOutputGroups()



// -----------------------------------------------------------------------------
// function: l4c_Hdf5_CreateOutput()
// Purpose : create a 9km L4_C OUTPUT file, once only per session. This
//           function also calls a number of subordinate functions that
//           create the empty datasets, and create all attributes (as known
//           at creation time) at both the file level and dataset levels.
// Inputs  : void
// returns : hid_t -- HDF5 file id , unique
// revision: 2020-07-13, hid_t format spec adjustments
// revision: 2016-02-11T14:01, updated to call l4c_Hdf5_AssignFillValue()
//           to improve consistency of fillvalues matched to dataset type
// revision: 2014-09-04T09:33, updated to consolidate all output to a 
//           a single "_mdl_" .h5 file (deprecated .qa file)
// notes   : 
// NOTES:  The NASA SMAP Product Spec Document shows the filename format:
//          SMAP_L4_C_MDL_20140216T221400_R01100_001.h5
//
hid_t l4c_Hdf5_CreateOutput(void)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();
 Hdf5GridPropAdt *hGridProp_p = l4c_GetHdf5GridProp();
 char fxName[256]= {0} ;
 
 int32 rank = 2; // shared rank, all are 2D
 
 float32 fltFill    = FLOAT_FILL_VALUE  ; // usually, -9999.0
 uint8   byteFill   = BYTE_FILL_VALUE   ; // usually, 254
 uint16  ushortFill = USHORT_FILL_VALUE ; // usually, 65534
 
 hid_t   file_id    = -1 ; // file_id of PRIMARY L4_C OUTPUT file
 hid_t   qa_file_id = -1 ; // file_id of separate QA file
 
 hid_t   plist9km = -1; // property list for majority, 9km variables
 hid_t   plist3km = -1; // property list for 3km variable
 hid_t   plist1km = -1; // property list for 1km variable
  
 hid_t   plist9km_uint8 ; // prop list for 9km QA UINT8 fields...
 hid_t   plist9km_uint16; // prop list for 9km QA UINT16 fields...

 hid_t   plist9km_int16 ; // <JMG> 2016-02-11 added
 hid_t   plist9km_int32 ; // <JMG> 2016-02-11 added
 hid_t   plist9km_uint32; // <JMG> 2016-02-11 added
  
 hsize_t dims1km[3] = {0}   ; // dims for 1km 
 hsize_t dims3km[3] = {0}   ; // dims for 3km 
 hsize_t dims9km[3] = {0}   ; // dims for 9km
 
 hsize_t chunk_size[2] ;  // shared chunks
 herr_t ret = -1 ;
 
 hid_t  dspace9km = -1;   // for 9km vars
 hid_t  dspace3km = -1;   // for 3km vars
 hid_t  dspace1km = -1;   // for 1km vars
 hid_t groupIdAry_p[32] = {0} ;// array of GROUP id's as created
 
 int32 compressLevel = sesCtl_p->compressLev ; // gzip, typically at level 3
 
 // <JMG> 6/26/2014, added as per Joe Aridizzone request, since they post-process
 int createMetaGroupSwitch = 0;
 
    strcpy(fxName,__func__);
    // -------------------------------------------------------------------------------------
    // TASK 1: Create empty shell of primary and matched QA files
    xLogMsg("@A In(%s) CREATING-L4_C-OUTPUT file and matching QA file~\n",fxName);

    // Create the FILE -- empty shell of the new primary L4_C OUTPUT HDF5 file
    if((file_id = H5Fcreate(sesCtl_p->output_L4C_DAILY_Path, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT))<0)
        xLogMsg("@F In(%s): FAILED creating primary L4_C OUTPUT HDF5 [%s]\n",
        fxName,sesCtl_p->output_L4C_DAILY_Path);

    // Immediately store the main L4_C output HDF5 file's id    
    sesCtl_p->l4cOutput_file_id = file_id;

    // -------------------------------------------------------------------------------------
    // TASK 1b: 9KM Machinery: set dataspace dimensions for the majority of
    //          the 9km output variables
    dims9km[0] = EGv2_GLOBAL_9KM_NS_PIXELS; // n. 9km rows
    dims9km[1] = EGv2_GLOBAL_9KM_EW_PIXELS; // n. 9km columns
    
    // Create the first of two common 2D dataspaces, this one for 9km
    dspace9km  = H5Screate_simple(rank, dims9km, dims9km);
    hGridProp_p->dspace_9km = dspace9km;
    
    // Create and store in hGridProp the other DCPL's, for completeness only
    // in case datasets of these types are added in the future
    plist9km_int16 = H5Pcreate(H5P_DATASET_CREATE);
    hGridProp_p->dcpl_9km_int16 = plist9km_int16 ;
    plist9km_int32 = H5Pcreate(H5P_DATASET_CREATE);
    hGridProp_p->dcpl_9km_int16 = plist9km_int32 ;
    plist9km_uint32 = H5Pcreate(H5P_DATASET_CREATE);
    hGridProp_p->dcpl_9km_uint32 = plist9km_uint32 ; 

    // Create PRIMARY 9KM dcpl, a Dataset-Creation-Property-List instance
    // recall that this dcpl applies to all 9KM datasets in the output granule
    plist9km = H5Pcreate(H5P_DATASET_CREATE);
    hGridProp_p->dcpl_9km    = plist9km; // store this dcpl for future reference

    // Assign a best-guess chunking geometry, specific to a 9km.
    // Rank=2.  Ideally, fine-tune this later...
    chunk_size[0] = (dims9km[0] / 10);
    chunk_size[1] = (dims9km[1] / 10);
    H5Pset_chunk(plist9km, rank, chunk_size);
    // <JMG> 2016-02-12T11:00 added back.
    H5Pset_fill_value(plist9km, H5T_NATIVE_FLOAT, &fltFill);
    
    // Now, use the dcpl (plist) also request GZIP LZ77 compression at level 6 or 7
    H5Pset_deflate(plist9km, compressLevel);

    // ----------------------------------------------------------------
    // TASK 2a: CREATE QA file's separate dcpl for UINT8 datasets
    // Create a separate dcpl for QA variables whose datatype is UINT8 and
    // note the need to use the byteFill here, not fltFil!
    plist9km_uint8 = H5Pcreate(H5P_DATASET_CREATE);
    hGridProp_p->dcpl_9km_uint8 = plist9km_uint8;
    
    H5Pset_chunk (plist9km_uint8, rank, chunk_size);
    H5Pset_deflate(plist9km_uint8, compressLevel);
    // <JMG> 2016-02-17T14:13 re-enable this to address qa_count Fillvalue issue
    H5Pset_fill_value(plist9km_uint8, H5T_NATIVE_UINT8, &byteFill);
    
    // TASK 2b: CREATE QA file's separate dcpl for UINT16datasets
    // Create separate dcpl for QA variables of datatype uint16, and
    // note the need for use of ushortFill, not fltFill here
    plist9km_uint16 = H5Pcreate(H5P_DATASET_CREATE);
    hGridProp_p->dcpl_9km_uint16 = plist9km_uint16;
    H5Pset_chunk (plist9km_uint16, rank, chunk_size);
    H5Pset_deflate(plist9km_uint16, compressLevel);
    H5Pset_fill_value(plist9km_uint16, H5T_NATIVE_UINT16, &ushortFill);

    // ----------------------------------------------------------------
    // TASK 3: Create a 3km 2D "M03" specific dataspace, for use 
    //         with Freeze Thaw later
    dims3km[0] = EGv2_GLOBAL_3KM_NS_PIXELS;
    dims3km[1] = EGv2_GLOBAL_3KM_EW_PIXELS;

    dspace3km = H5Screate_simple(rank, dims3km, dims3km);
    hGridProp_p->dspace_3km = dspace3km;
    
    // set the chunk size, specific to a 9km 
    chunk_size[0] = (dims3km[0] / 10);
    chunk_size[1] = (dims3km[1] / 10);
    
    // Create a 3km dcpl, a Dataset-Creation-Property-List instance
    // to be used for 3KM Freeze/Thaw datasets
    plist3km = H5Pcreate(H5P_DATASET_CREATE);
    hGridProp_p->dcpl_3km = plist3km ;
    
    H5Pset_chunk (plist3km, rank, chunk_size);
    // Now, use the 3km dcpl (plist) also uses same GZIP LZ77 compression
    H5Pset_deflate(plist3km, compressLevel);
    // Set the 3km default fill value, 254 (for FT_ROLLING_M03 etc)
    H5Pset_fill_value(plist3km, H5T_NATIVE_UINT8, &byteFill);

    // ----------------------------------------------------------------
    // TASK 4: Create a 1km 2D "M01" specific dataspace, for primary vars
    // in the daily intermediate SOC 1km HDF5 file
    dims1km[0] = EGv2_GLOBAL_1KM_NS_PIXELS;
    dims1km[1] = EGv2_GLOBAL_1KM_EW_PIXELS;

    dspace1km = H5Screate_simple(rank, dims1km, dims1km);
    hGridProp_p->dspace_1km = dspace1km ;
    
    // set the chunk size, specific to a 1km 
    chunk_size[0] = (dims1km[0] / 10);
    chunk_size[1] = (dims1km[1] / 10);
    
    // Create a 1km dcpl, a Dataset-Creation-Property-List instance
    // to be used only for SOC 1km datasets..
    plist1km = H5Pcreate(H5P_DATASET_CREATE);
    hGridProp_p->dcpl_1km = plist1km ;
    // assign 1km CHUNK and DEFLATE parameters  
    H5Pset_chunk (plist1km, rank, chunk_size);
    // Next use 1km dcpl (plist) to enable GZIP LZ77 compression
    H5Pset_deflate(plist1km, compressLevel);

   // ----------------------- 9 KM -------------------------------------
   // 6) Next, store appropriate constructs made for the 9km use-cases
   
   // IsEGv2_Global_1km,IsEGv2_Global_3km,IsEGv2_Global_9km
   hGridProp_p->easeGridId  = IsEGv2_Global_9km;
   
    // <JMG> 6/26/2014, now assign switch to indicate if running locally
    //  at NTSG vs. running at GMAO (set to 0 here, since the GSFC
    //  GMAO production group prefers to create this group as a post-process step
    createMetaGroupSwitch = 0;

    l4c_Hdf5_CreateAllOutputGroups(file_id, createMetaGroupSwitch,groupIdAry_p);

    // WRITE initial, minimal LOCAL (PSA) metadata out to OUTPUT _mdl_ granule
    l4c_Hdf5_AddFileMetadata();
  
    // Next, CREATE ALL DATASET definitions, in both primary and QA file
    l4c_Hdf5_CreateDatasets(file_id);

    // WRITE lat,lon coordinate variables. Must do this AFTER the
    // call to CreateDatasets() though. Check to 
    // include the appropriate CF dataset level attributes as well.
    l4c_Hdf5_WriteLatLon_CoordVars(file_id);

    // -----------------------------------------------------------------
    // TASK 5: report back and package up id's for return
    xLogMsg("@A In(%s) CREATED NEW HDF5 [%s] file_id %d ~\n",fxName, 
        sesCtl_p->output_L4C_DAILY_Path, file_id);
    xLogMsg("@A In(%s) Compression level %d ~\n",fxName, compressLevel);
    
    xLogMsg("@A In(%s) 9km byte    fillvalue %u ~\n"   ,fxName, (unsigned int)byteFill);
    xLogMsg("@A In(%s) 9km ushort  fillvalue %u ~\n"   ,fxName, (unsigned int)ushortFill);
    xLogMsg("@A In(%s) 9km float32 fillvalue %.2lf ~\n",fxName, (double)fltFill);
    // <JMG> 2020-07-13T17:56, adjusted format spec to %ld from %d for proper hid_t display
    xLogMsg("@A In(%s) 1km dataspace properties dspace %ld dcpl_1km %ld ~\n",fxName, dspace1km, plist1km);
    xLogMsg("@A In(%s) 3km dataspace properties dspace %ld dcpl_3km %ld ~\n",fxName, dspace3km, plist3km);
    xLogMsg("@A In(%s) 9km dataspace properties dspace %ld dcpl_9km %ld ~\n",fxName, dspace9km, plist9km);

    xLogMsg("@A CREATED special case dcpls for uint8 and uint16~\n");
    xLogMsg("@A (In%s) 9km dcpl_9km_uint8  %d ~\n",fxName, plist9km_uint8);
    xLogMsg("@A In(%s) 9km dcpl_9km_uint16 %d ~\n",fxName, plist9km_uint16);
    
    // Next, explicitly FLUSH the output HDF5 files here, to keep things stable
    // and predictable. Options for H5F_scope_t arg are: {H5F_SCOPE_GLOBAL|H5F_SCOPE_LOCAL}
    ret = H5Fflush(file_id, H5F_SCOPE_LOCAL); 

    xLogMsg("@A In(%s) checkpoint flush L4CMDL OUTPUT granule at file_id (%ld) ~\n",
            fxName, file_id);
    
    l4c_EventCompletionMsg( (char *)__func__);

  // return unique file_id
 return file_id;    
 
} // end:: l4c_Hdf5_CreateOutput()


// DEPRECATED l4c_Hdf5_CreateQa()   2014-09-03 <JMG>


// ------------------------------------------------------------------
// function     : l4c_Hdf5_WriteData()
// purpose      : write an (entire) simple HDF5 dataset identified by name
//                from dataBuffer_p into the dataset in HDF5
// inputs       : const char *callerLbl
//                const hid_t thisFile_id --file's id
//                const dsTypeId  id like H5T_NATIVE_FLOAT32
//                char *datasetname_p group prefaced, qualified HDF5 
//                   dataset name.
//                void *dataBuffer_p --ptr to preallocated buffer sized
//                sufficiently to hold entire array
// revised      : 2014-02-16T11:52, adapted from FRM API's
// history      : n/a
// dependencies : HDF5 lib
// notes        : the H5LT interface doesn't support chunking, compression 
//                or explicit dcpl's, so use this with that understanding.
// notes        : It is critical that the dataBuffer_p passed in is sized for the right datatype and 
//                and no. of elements
// notes        :  H5T_NATIVE_UINT8, H5T_NATIVE_INT16,H5T_NATIVE_UINT16,H5T_NATIVE_INT32,
//                 H5T_NATIVE_UINT32, H5T_NATIVE_FLOAT32, H5T_NATIVE_FLOAT64, etc.
int l4c_Hdf5_WriteData(const char *callerLbl_p,const hid_t thisFile_id,const int rank,const hsize_t *dims,
        const hid_t dsTypeId,char *datasetName_p,void *dataBuffer_p)
{
 int status = -1 ;
 herr_t ret = 0;
 char fxName[256] = {0} ;
 
    // WRITE a simple HDF5 dataset...sans compression or chunking
    // noting that the dataset named here MUST NOT already exist. If it
    // does this routine fails.
    strcpy(fxName,__func__);
    
    if((ret = H5LTmake_dataset(thisFile_id, datasetName_p, rank,dims,
                            dsTypeId,(uint8 *)dataBuffer_p )) < 0)
    {
        xLogMsg("@F In(%s) READ FAILED dataset(%s) ret %d from file_id!\n",
                fxName, datasetName_p, ret, thisFile_id);
    } else {
        xLogMsg("@A In(%s) (%s) READ success on dataset (%s) from file_id %d\n",
                fxName, callerLbl_p, datasetName_p, thisFile_id);
    } // end:: block if data was read ok
 
    l4c_EventCompletionMsg( (char *)__func__ ); 
 
 return status ;
 
}  // end:: l4c_hdf5_WriteData()


// ------------------------------------------------------------------
// function     : l4c_Hdf5_WriteDataset()
// purpose      : write to a new or existing dataset
// inputs       : const char *callerLbl_p,
//                const hid_t file_id,
//                const char *dsName_p,
//                const hid_t dataset_id, 
//                const hid_t dsType_id  -- H5T datatype like H5T_NATIVE_FLOAT
//                void *dsData_p
// revised      : 2014-02-16T11:52, adapted from FRM API's
// history      : n/a
// dependencies : HDF5 lib
// notes        : from l4c_MODIS
//                  the H5LT interface doesn't support chunking, compression 
//                or explicit dcpl's, so use this with that understanding.
// notes        : It is critical that the dataBuffer_p passed in is sized for the right datatype and 
//                and no. of elements
// notes        : c99 allows: H5T_NATIVE_UINT8, H5T_NATIVE_INT16,H5T_NATIVE_UINT16,H5T_NATIVE_INT32,
//                 H5T_NATIVE_UINT32, H5T_NATIVE_FLOAT32, H5T_NATIVE_FLOAT64, or, H5T_NATIVE_FLOAT
//
// herr_t H5Dwrite( hid_t dataset_id, hid_t mem_type_id, hid_t mem_space_id, 
//               hid_t file_space_id, hid_t xfer_plist_id, const void * buf )
//
int l4c_Hdf5_WriteDataset(const char *callerLbl_p,const hid_t file_id,const char *dsName_p,
        const hid_t dataset_id, const hid_t dsType_id,void *dsData_p)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();
 char fxName[256] = {0} ;   
 herr_t ret = -1;
 herr_t retClose = -1 ;

    float chkSum  =0.0;
    float chkMean =0.0;
    float chkN = 0.0;
    float sciPix =0.0;
    float *flt_p = (float *)dsData_p;
    int noOp = 0 ;
    int doVerify = 0 ;
    size_t pixIndex =0;
    size_t nthQualified = 0 ;
 
        strcpy(fxName,(char *)__func__);   
 
        // set the condition of doVerify based on 'enableGranuleMetrics'
        doVerify = sesCtl_p->enableGranuleMetrics ;
 
        // Removed diagnostic, no longer needed
        // xLogMsg("@A (%s) caller: %s creating (%s) type(%d) with dset (%d) owner(%d)~\n",
        //        fxName, callerLbl_p, dsName_p, (int)dsType_id, dataset_id, file_id );
        
        // TBD: confirm: this has worked without an explicit call to H5Dopen() first,
        //  but, verify which exact use-cases require a prior H5Dopen() --probably a read does.
        //  The default semantics make it tricky where there is more than ONE write of a
        //  dataset during a session, such as for the SOC layers that are first read,
        //  then updated, within a session.
        if((ret=H5Dopen(file_id,dsName_p,H5P_DEFAULT)) < 0)
        {
            xLogMsg("@E In(%s) ERROR from (%s) EXPLICIT-OPEN of (%s) dset (%d) owner(%d) ret %d ~\n",
                fxName, callerLbl_p, dsName_p, dataset_id, file_id, ret );
        } else {
            noOp = 0;
            // xLogMsg("@A In(%s) caller(%s) EXPLICIT-OPEN of (%s) dset (%d) owner(%d) ret %d ~\n",
            //  fxName, callerLbl_p, dsName_p, dataset_id, file_id, ret );
        }
        
        if((ret=H5Dwrite(dataset_id, dsType_id, H5S_ALL, H5S_ALL, H5P_DEFAULT, dsData_p))< 0)
        {
            xLogMsg("@E (%s) ERROR on H5Dwrite: caller: %s creating (%s) type(%d) with dset (%d)~\n",
                fxName, callerLbl_p, dsName_p, (int)dsType_id, dataset_id );
        } else {
            
            // if requested, upon successfull completion of the dataset WRITE, also run a
            // quick statistical check and output it if in DIAG mode
            if( doVerify &&  dsType_id == H5T_NATIVE_FLOAT )
            {
                // scan the first LDOM9KM pixels, just as verification
                for(pixIndex =0; pixIndex < LDOM9KM; pixIndex++)
                {
                    sciPix = flt_p[pixIndex] ;
                    // should improve, verify this threshold of (100) better
                    if(sciPix >= -100.0) 
                    {
                        nthQualified++ ;
                        chkSum += sciPix;
                    }
                }
                
                chkN = (float)nthQualified ;
                if(nthQualified  > 0)
                    chkMean = chkSum / chkN;
                else
                    chkMean = -9999.0 ;
                    
                xLogMsg("@D DIAG (%s) dataset(%s) sum(%.2f) mean(%.2f) N(%.0f)~\n",
                 fxName, dsName_p, chkSum, chkMean, chkN);
            } // end:: doVerify block..
        } // end:: block if write went OK
        
        // explicitly CLOSE the dataset after the write is complete
        if((retClose = H5Dclose(dataset_id)) < 0)
        {
            xLogMsg("@E In(%s) ERROR CLOSING (%s) dataset_id(%d)~\n",
                (char *)__func__,dsName_p, dataset_id);
        }
        
        l4c_EventCompletionMsg( (char *)__func__ ); 
        
  return ret ;
  
 } // end:: l4c_Hdf5_WriteDataset()


// ----------------------------------------------------------------------
// function : l4c_Hdf5_CreateAttribute()
// purpose  : add an a single attribute to HDF5 object 'owner_id' named
//            objName_p, with attribute named 'attrib_name' having a value
//            of attribData_p whose datatype is tivDatatypeIdentifier (like DFNT_UINT8 etc)
// inputs   : const hid_t owner_id,
//            int tivDatatypeIdentifier: IsString,IsInt16,IsFloat32 etc NOT H5T_NATIVE_INT16'setc
//            const char *ownersName_p,  --string, naming the object who owns attribute
//            const char *attribName_p,  --string, naming attribute itself
//            void *attribData_p --ptr to raw attribute data to write
//
// returns  : int,status
// notes    : the test 'didExistPrior' now attempts to verify if the 
//            attribute first existed. If so, it uses it, otherwise a new
//            attribute instance is created.
// notes    : Reminder: hid_t H5Acreate( hid_t loc_id, const char *attr_name, 
//               hid_t type_id, hid_t space_id, hid_t acpl_id, hid_t aapl_id )
//           --acpl_id is attribute creation property list,
//           --aapl_id is attribute's access property list
//
// notes    : Reminder: type specific variants in the HDF5 library "Lite" interface are:
//  herr_t H5LTset_attribute_string( hid_t loc_id, const char *obj_name, const char *attr_name, const char *attr_data )
//  herr_t H5LTset_attribute_char ( hid_t loc_id, const char *obj_name, const char *attr_name, char *buffer, hsize_t size)
//  herr_t H5LTset_attribute_short ( hid_t loc_id, const char *obj_name, const char *attr_name, short *buffer, hsize_t size)
//  herr_t H5LTset_attribute_float( hid_t loc_id, const char *obj_name, const char *attr_name, float *buffer, hsize_t size )
//  herr_t H5LTset_attribute_double ( hid_t loc_id, const char *obj_name, const char *attr_name, const double *buffer, size_t size )
//
hid_t l4c_Hdf5_CreateAttribute(const hid_t owner_id,int tivDatatypeIdentifier,
        const char *objName_p,const char *attribName_p, void *attribData_p)
{
 hid_t attrib_id =0;
 hid_t attrDspace = 0 ;
 herr_t ret ;
 hid_t aSiz = 1 ; // size of 1D array for scalars, 
 int  didExistPrior = 0 ; // boolean, is 1 if attribute already exists, or 0 if it does NOT exist prior
 int  tivSizOfOne = 0 ;
 char fxName[256] = {0};
 char tivLabel[32] = {0};

    strcpy(fxName,__func__);
 
    // <JMG> add appropriate logic to FIRST test if attribute already exists
    if((attrib_id =H5Aopen(owner_id, attribName_p, H5P_DEFAULT)) < 0)
        didExistPrior = 0 ; // attribute does NOT already exist, so DO create it here
    else
        didExistPrior = 1 ; // attribute ALREADY exists at this location and name

    // fetch readable NAME of this tivDatatype   
    strcpy(tivLabel, tiv_NameofTivType(tivDatatypeIdentifier,&tivSizOfOne));

    if( didExistPrior == 0)
    {

        switch(tivDatatypeIdentifier)
        {
        case IsString : attrib_id = H5Acreate(owner_id,attribName_p,H5T_NATIVE_CHAR, attrDspace,
                                    H5P_DEFAULT,H5P_DEFAULT);
                        ret = H5LTset_attribute_string(owner_id,objName_p, attribName_p,attribData_p);
                                    break;
        case IsUint8  : attrib_id = H5Acreate(owner_id,attribName_p,H5T_NATIVE_UINT8, attrDspace,
                                    H5P_DEFAULT, H5P_DEFAULT);    
                        // <JMG> 2016-02-17t14:20, found +fixed bug, where _short() used when _char should have been used here.
                        ret = H5LTset_attribute_uchar(owner_id,objName_p,attribName_p,attribData_p,aSiz);
                                    break;
        case IsInt16  : attrib_id = H5Acreate(owner_id,attribName_p,H5T_NATIVE_SHORT, attrDspace,
                                    H5P_DEFAULT,H5P_DEFAULT);
                        ret = H5LTset_attribute_short(owner_id,objName_p,attribName_p,attribData_p,aSiz);
                                    break;
        case IsUint16 : attrib_id = H5Acreate(owner_id,attribName_p,H5T_NATIVE_USHORT, attrDspace,
                                    H5P_DEFAULT,H5P_DEFAULT);
                        ret = H5LTset_attribute_ushort(owner_id,objName_p,attribName_p,attribData_p,aSiz);
                                    break; 
        case IsInt32  : attrib_id = H5Acreate(owner_id,attribName_p,H5T_NATIVE_INT, attrDspace,
                                    H5P_DEFAULT,H5P_DEFAULT);
                        ret = H5LTset_attribute_long(owner_id,objName_p,attribName_p,attribData_p,aSiz);
                                    break;
        case IsUint32  : attrib_id = H5Acreate(owner_id,attribName_p,H5T_NATIVE_ULONG, attrDspace,
                                    H5P_DEFAULT,H5P_DEFAULT);
                        ret = H5LTset_attribute_long(owner_id,objName_p,attribName_p,attribData_p,aSiz);
                                    break;
        case IsFloat32: attrib_id = H5Acreate(owner_id,attribName_p,H5T_NATIVE_FLOAT, attrDspace,
                                    H5P_DEFAULT,H5P_DEFAULT);
                        ret = H5LTset_attribute_float(owner_id,objName_p, attribName_p,attribData_p,aSiz);
                                    break;
        case IsFloat64: attrib_id = H5Acreate(owner_id,attribName_p,H5T_NATIVE_DOUBLE, attrDspace,
                                    H5P_DEFAULT,H5P_DEFAULT);
                        ret = H5LTset_attribute_double(owner_id,objName_p, attribName_p,attribData_p,aSiz);
                                    break;
        default :
                    xLogMsg("@a l4c_Hdf5_CreateAttibute: attrib-name(%s) TIV datatype(%d [%s]) is NOT yet supported~\n",
                            attribName_p, tivDatatypeIdentifier, tivLabel);
        } // end:: hDtype

    } // end:: if didExistPrior test block
    
    ret = H5Aclose(attrib_id);
    
    l4c_EventCompletionMsg( (char *)__func__ ); 

 // return attrib_id BUT note after closing it will be stale...
 return attrib_id ; 
    
} // end:: l4c_Hdf5_CreateAttribute()



// ----------------------------------------------------------------------
// function : l4c_Hdf5_CreateAttributeString()
// purpose  : add (write) a string attribute (attribStringData_p)
//            for the HDF5 attribute named 'attribName_p' to the object
//            identified as 'owner_id'
// inputs   : hid_t owner_id,
//            const char *attribName_p,
//            char *attribStringData_p
void l4c_Hdf5_CreateAttributeString(hid_t owner_id,
            const char *attribName_p,char *attribStringData_p)
{
 hid_t attr_dspace = -1;
 hid_t attr_type   = -1;
 hid_t attr_id     = -1;
 herr_t ret ; 
 char fxName[256] = {0} ;
 
 
        strcpy(fxName,__func__);
        
       // Create attributes dataspace property
        attr_dspace = H5Screate(H5S_SCALAR);

        // Create attributes type as a string
        attr_type   = H5Tcopy(H5T_C_S1);

        // Set the SIZE of the attribute as the length of the single string
        H5Tset_size(attr_type, strlen(attribStringData_p)) ;

        // Create the container for this HDF5 string attribute
        attr_id = H5Acreate(owner_id, attribName_p, attr_type,
                attr_dspace, H5P_DEFAULT, H5P_DEFAULT);

        // Write the string attribute's value, held in string
        ret = H5Awrite(attr_id, attr_type, attribStringData_p);
        
        // Removed diagnostic
        // xLogMsg("@A In(%s) file_id %d Label<%s> = Value<%s>~\n",
        //    fxName, owner_id, attribName_p, attribStringData_p);

        H5Sclose(attr_dspace);
        H5Aclose(attr_id);
            
    l4c_EventCompletionMsg( (char *)__func__ ); 
    
} // end:: l4c_Hdf5_CreateAttributeString()


// ----------------------------------------------------------------------
// function : l4c_Hdf5_WriteLatLon_CoordVars()
// purpose  : Writes the /GEO latitude, longitude coordinate vars to
//            the single L4_C "_mdl_" .h5 output files.
// inputs   : void
// returns  : int,status
// revision : 2014-09-03T12:48, deprecated QA file entirely, merged it
//            into "_mdl_" .h5 file.
// note     : This is called by l4c_Hdf5_CreateOutput()
// notes    : This fx MUST be called after the l4c_Hdf5_CreateDatasets()
//            is performed since it relies on OUTPUT dataset ID's only available
//            after the OUTPUT HDF5 dataset objects are created.
// reminder : The (lat,lon) coord var dataset ID's relevant here are for the
//            OUTPUT HDF5 file, not the dataset ID from the INPUT SMAP_L4_C_Ancillary.h5 file

int l4c_Hdf5_WriteLatLon_CoordVars(const hid_t output_file_id)
{ 
 SesCtl *sesCtl_p = l4c_GetSesCtl();
 PkgMgr *pkgMgr_p = l4c_GetPkgMgrOverseer();
 NsbObj *nsb_p    = l4c_GetNsbOverseer();
 
 AncImagePkgAdt *ancImagePkgMasterObj_p = l4c_GetAncImagePkgAdt();
 PkgTheme *pThme_p= NULL;
 PkgObj   *pObj_p = NULL;

 hid_t anc_file_id =0 ; 
 hid_t dset_id = -1 ;
 hsize_t dims[2] = {0} ;
 herr_t ret =0;

 // Hdf5GridPropAdt *hGridPropObj_p = l4c_GetHdf5GridProp();
 char fxName[256]={0} ;
 char nvpBuf[FILENAME_MAX] = {0};
 char *latVarname_p = NULL;
 char *lonVarname_p = NULL;
 
    dims[0] = 1624;
    dims[1] = 3856 ;

    strcpy(fxName,(char *)__func__);

    // <JMG> anc_file_id and all other key INPUT HDF5 files opened now in l4c_SesInit()
    // fetch the HDF file id for the ancillary file
    anc_file_id = sesCtl_p->anc_file_id;

    // Allocate both coordinate variables
    ancImagePkgMasterObj_p->imgM09_latitude_img_p  = memCalloc(EGv2_GLOBAL_9KM_PIXELS, sizeof(float32));
    ancImagePkgMasterObj_p->imgM09_longitude_img_p = memCalloc(EGv2_GLOBAL_9KM_PIXELS, sizeof(float32));

    // the ancillary theme contains the lat,lon coord var names as present
    // in the ancillary file. Note how these differ from the OUTPUT names they
    // have (from the PSD) GEO/latitude,/GEO/longitude.
    pThme_p = pkgMgr_GetThemeByName(pkgMgr_p,"SMAP_L4_C_Ancillary");    

    // ------- READ LATITUDE AND LONGITUDE coordinate variables --------
    // READ using names in SMAP_L4_C_Ancillary.h5 which are different than those
    // to be used in the OUTPUT file
    // REMINDER: 2016-01-07, the dataset ID's for the INPUT (lat,lon) from SMAP Ancillary
    //   are distinct (obviously) from those assoc. with 
    //   OUTPUT (lat,lon) that are written in L4_C OUTPUT file

    // variable ancillary images from RIP file, reside in the SMAP_L4_C_Ancillary.h5 file
    latVarname_p = nsbGetValueByKey(nsb_p,"L4C_ANC_DATASET_LATITUDE_9KM_COORD",nvpBuf);
    xLogMsg("@A In(%s) preparing to read INPUT latitude coord var (%s)<%s>~ \n",
        fxName,latVarname_p,nvpBuf);
    // READ LATITUDE coordinate variable from (input) SMAP_L4_C_Ancillary.h5
    ret = frm_hdf5_ReadData(anc_file_id, H5T_NATIVE_FLOAT, latVarname_p,
        ancImagePkgMasterObj_p->imgM09_latitude_img_p ,fxName);

    lonVarname_p = nsbGetValueByKey(nsb_p,"L4C_ANC_DATASET_LONGITUDE_9KM_COORD",nvpBuf);
    xLogMsg("@A In(%s) preparing to read INPUT longitude coord var (%s)<%s>~ \n",
        fxName,lonVarname_p,nvpBuf);
        
    // READ LONGITUDE coordinate variable from (input) SMAP_L4_C_Ancillary.h5
    ret = frm_hdf5_ReadData(anc_file_id, H5T_NATIVE_FLOAT, lonVarname_p,
        ancImagePkgMasterObj_p->imgM09_longitude_img_p, fxName);

    // PHASE 2: lat,lon read, so NOTE CAREFULY we fetch the prior 
    // assigned dataset_id's for lat,lon from the L4C_OUTPUT theme
    // in order to WRITE them out to the L4_C output granule file
    pThme_p = pkgMgr_GetThemeByName(pkgMgr_p,"L4C_OUTPUTS");    
    
    // LATITUDE --------------------------------------------------------
    // <JMG> 2016-01-04 convert to use L4C_OUTPUT_VARIABLE_LATITUDE 
    // -- use IsPropertyName to retrieve JUST the 'name' from the delimited property string
    if(nsbDoesKeyExist(nsb_p,"L4C_OUTPUT_VARIABLE_LATITUDE"))
    {
        latVarname_p = nsbGetPropertyValueByKey(nsb_p,
           "L4C_OUTPUT_VARIABLE_LATITUDE", IsPropertyName,nvpBuf);
    } else {
        xLogMsg("@F In(%s) L4C_OUTPUT_VARIABLE_LATITUDE must exist! for dset_id (%d) ~\n",
            fxName,dset_id);
    }
    
    // orig L4C_OUTPUT_DATASET_GEO_9KM_LON is NOW called  L4C_OUTPUT_VARIABLE_LONGITUDE
    dset_id = pkgMgr_GetThemes_ChildIdByName(pThme_p,latVarname_p);
    xLogMsg("@A In(%s) LAT dataset (%s) id is %d ~\n",fxName,latVarname_p,dset_id);
    
    if( dset_id < 0)
    {
        xLogMsg("@F In(%s) LAT dataset (%s) id %d is INVALID ~\n",
                fxName,latVarname_p,dset_id);
    }
    l4c_Hdf5_WriteDataset( fxName, output_file_id, 
        latVarname_p,
        dset_id,
        H5T_NATIVE_FLOAT, 
        ancImagePkgMasterObj_p->imgM09_latitude_img_p);

    // LONGITUDE -------------------------------------------------------
    //  uses RIP nvp e.g. "L4C_OUTPUT_VARIABLE_LONGITUDE" to retrieve NAME of
    //  LATITUDE or LONGITUDE variable to output to new HDF5 granule
    
    // uses IsPropertyName to retrieve JUST the 'name' from the delimited property string
    if(nsbDoesKeyExist(nsb_p,"L4C_OUTPUT_VARIABLE_LONGITUDE"))
    {
        lonVarname_p = nsbGetPropertyValueByKey(nsb_p,
            "L4C_OUTPUT_VARIABLE_LONGITUDE", IsPropertyName,nvpBuf);
    } else {
        xLogMsg("@F In(%s) L4C_OUTPUT_VARIABLE_LONGITUDE must exist! for dset_id (%d) ~\n",
            fxName,dset_id);
    }

    // using the LONGITUDE's name, fetch its unique HDF5 dataset id
    dset_id = pkgMgr_GetThemes_ChildIdByName(pThme_p, lonVarname_p);
    
    xLogMsg("@A In(%s) LON dataset (%s) id is %d ~\n",fxName,lonVarname_p,dset_id); 
    if( dset_id < 0)
    {
        xLogMsg("@F In(%s) LON dataset (%s) id %d is INVALID~\n",
                fxName,lonVarname_p,dset_id);
    }
    // WRITE the LONGITUDE dataset out to the new L4_C product granule
    l4c_Hdf5_WriteDataset(fxName , output_file_id, 
        lonVarname_p,
        dset_id,
        H5T_NATIVE_FLOAT, 
        ancImagePkgMasterObj_p->imgM09_longitude_img_p);

    // NEXT UP: be sure to write the matching CF attribute coordinates="latitude longitude"
    // at each DATASET level, to point to these GEO/latitude /GEO/longitude

    // free up the local lat,lon buffers
    memFree( ancImagePkgMasterObj_p->imgM09_latitude_img_p );
    memFree( ancImagePkgMasterObj_p->imgM09_longitude_img_p);
    
    l4c_EventCompletionMsg( fxName);
    
    return 0;
} // end:: l4c_Hdf5_WriteLatLon_CoordVars()


// --------------------------------------------------------------------
// Function : writeNvpAsAttribute
// Purpose  : using a supported NVP keyword, create a string attribute
// Inputs   : const hid_t file,const char *keyword,char *attrForceName)
// Author   : joe glassy
// Revised  : 2014-10-21T21:23:00,ruby(ASUS,home,jmg)
// Returns  : hid_t attr_id if successful, or -1 if not
// Notes    : The HDF5 attribute written using the keyword as the
//            attributes name and the NVP's 'value' as the attribute value
//      
hid_t l4c_writeNvpAsAttribute(hid_t file_id,const char *keyword,const char *attrForceName)
{
 NsbObj *nsbObj_p = l4c_GetNsbOverseer(); 
    
 char attrValu[2048]= {0} ; 
 char nvpBuf[2048]= {0} ;
 int keyResult = -1 ;
 int status = 0 ;
 hid_t attr_id = 0 ;
  
    // search table of supported keywords
    
    if( nsbDoesKeyExist(nsbObj_p,keyword))
    {
        // fetch value of this supported keyword
        strcpy(attrValu, nsbGetValueByKey(nsbObj_p,keyword,nvpBuf));
        
        // Force use of 'attrForceName' instead of 'keyword' as the
        // HDF5 attribute name or label.
        if(attrForceName != NULL && strlen(attrForceName)>= 1)  
        {
            // attr_id = l4c_createHdf5Attribute(file_id,attrForceName, attrValu);
            status = frm_hdf5_WriteAttributeAsType(file_id,"/", attrForceName,IsString,1,attrValu);
            
        } else {
            // use the 'keyword' verbatim as the HDF5 attribute name
            // attr_id = l4c_createHdf5Attribute(file_id,keyword, attrValu);
            status = frm_hdf5_WriteAttributeAsType(file_id,"/", keyword,IsString,1,attrValu);
        }
    } else { // end:: block if keyword IS a supported one
        // notify caller that this keyword was invalid or not supported
        xLogMsg("@E l4c_writeNvpAsAttribute: NO key(%s) found - suppressed attrib write!~\n",
        keyword);
    }
    
 // TBD: must add code now to explicitly retrieve this most
 // recently created unique attr_id 
 // The attr_id returned here is not real, only a placeholder
 return 0 ; 
 
} // end::l4c_writeNvpAsAttribute()
