// ---------------------------------------------------------------
// module       : l4c_DAILY_rip.c
// purpose      : Implement RIP runtime-input-file related logic for
//                the NASA SMAP l4c_DAILY.ex
// project      : SMAP L4C science development.
// dependencies : HDF4 4.2.x, HDF5 v1.8.x or later libraries, L4C FRAMEWORK APIs
// revised      : 2015-07-29, <JMG> add support for GEOS5 FP TS derived F/T ,
// gridded
//                to same 3km as  before.
// revised      : 2015-06-10T16:47, <JMG> changed L4C_SOFTWARE_VERSION to
// L4C_SW_VERSION revised      : 2014-11-14, added optional L4C_ENABLE_RMSE =
// YES|NO keyword history      : 2014-11-03, added for local UM:
// L4C_MODEL_STOPDAY,L4C_MODEL_DAYSTEPINTERVAL
//                for future (local UM) multi-day mode. Not for Ops production
//                mode!
// history      : 2014-09-28T22:38, added enableSOCUpdate option
// history      : 2014-07-25TT16:06
// history      : 2014-07-25TT16:06, split out from l4c_DAILY_ses.c
// author       : jglassy
// notes        : n/a
// ---------------------------------------------------------------

#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"

// FparClimatology is time-ordered on a period basis NOT on a daily
// basis, so this is why the search key for this LUT is now normally
// the 8-day composite period index and not the day-of-year index.
// There are (46) 8-day periods in one year
// FIELDS: YEARDAY, IDX_8DAYPERIOD, IDX_16DAYPERIOD,FparClimDatasetName
static char *fparClimatologyNameLut[46] = {
    "001,01,01,fpar_clim_M01_day001_per01", // for: day 001..008
    "009,02,01,fpar_clim_M01_day009_per02", // for: day 009..016
    "017,03,02,fpar_clim_M01_day017_per03", // for: day 017..024
    "025,04,02,fpar_clim_M01_day025_per04", // for: day 020..032
    "033,05,03,fpar_clim_M01_day033_per05",
    "041,06,03,fpar_clim_M01_day041_per06",
    "049,07,04,fpar_clim_M01_day049_per07",
    "057,08,04,fpar_clim_M01_day057_per08",
    "065,09,05,fpar_clim_M01_day065_per09",
    "073,10,05,fpar_clim_M01_day073_per10",
    "081,11,06,fpar_clim_M01_day081_per11",
    "089,12,06,fpar_clim_M01_day089_per12",
    "097,13,07,fpar_clim_M01_day097_per13",
    "105,14,07,fpar_clim_M01_day105_per14",
    "113,15,08,fpar_clim_M01_day113_per15",
    "121,16,08,fpar_clim_M01_day121_per16",
    "129,17,09,fpar_clim_M01_day129_per17",
    "137,18,09,fpar_clim_M01_day137_per18",
    "145,19,10,fpar_clim_M01_day145_per19",
    "153,20,10,fpar_clim_M01_day153_per20",
    "161,21,11,fpar_clim_M01_day161_per21",
    "169,22,11,fpar_clim_M01_day169_per22",
    "177,23,12,fpar_clim_M01_day177_per23",
    "185,24,12,fpar_clim_M01_day185_per24",
    "193,25,13,fpar_clim_M01_day193_per25",
    "201,26,13,fpar_clim_M01_day201_per26",
    "209,27,14,fpar_clim_M01_day209_per27",
    "217,28,14,fpar_clim_M01_day217_per28",
    "225,29,15,fpar_clim_M01_day225_per29",
    "233,30,15,fpar_clim_M01_day233_per30",
    "241,31,16,fpar_clim_M01_day241_per31",
    "249,32,16,fpar_clim_M01_day249_per32",
    "257,33,17,fpar_clim_M01_day257_per33",
    "265,34,17,fpar_clim_M01_day265_per34",
    "273,35,18,fpar_clim_M01_day273_per35",
    "281,36,19,fpar_clim_M01_day281_per36",
    "289,37,20,fpar_clim_M01_day289_per37",
    "297,38,21,fpar_clim_M01_day297_per38",
    "305,39,22,fpar_clim_M01_day305_per39",
    "313,40,23,fpar_clim_M01_day313_per40",
    "321,41,24,fpar_clim_M01_day321_per41",
    "329,42,25,fpar_clim_M01_day329_per42",
    "337,43,26,fpar_clim_M01_day337_per43",
    "345,44,27,fpar_clim_M01_day345_per44",
    "353,45,28,fpar_clim_M01_day353_per45",
    "361,46,29,fpar_clim_M01_day361_per46"};

// ---------------------------------------------------------------
// function   : l4c_SesVerifyNVPs()
// Purpose    : Verifies the set of non-core name value pairs (NVPs)
//              supplied in the runtime input parameter (RIP) file via
//              keys.
// Inputs     : NsbObj *nsb_p -- ptr sessions NSB stringblock
//              name-value-pair in-memory database object, storing
//              ALL runtime input file parameters.
// Invocation : ./l4c_DAILY.ex <./l4c_DAILY.rip>
// Returns    : time_t --starting time in seconds.
// Notes      : Based on predefined RIP NVP keys, fetch values..
// Author     : joe glassy
void l4c_SesVerifyNVPs(NsbObj *nsb_p) {
  char valueString[FILENAME_MAX] = {0};
  char keyword[FILENAME_MAX] = {0};
  char fxName[256] = {0};

  int nKeys = nsbGetStringblockRecordCount(nsb_p);
  int k = 0;

  strcpy(fxName, __func__);

  xLogMsg("@A In(%s)l4c_SesFetchNVPs: scanning %5d keys", fxName, nKeys);

  // Iterate through ALL available RIP name-value-pair parameters
  for (k = 0; k < nKeys; k++) {
    // first, by index get the keyword it was stored under
    nsbGetKeywordByIndex(nsb_p, k, keyword);
    // next, fetch the value, interpreted as a universal string
    nsbGetValueByIndex(nsb_p, k, valueString);
    // for validation we've correctly got all the input parameters,
    // route a complete report of these to the high level log file
    xLogMsg("@A In(%s) RIP validation: entry %5d key[%s] value[%s]\n", fxName,
            k, keyword, valueString);

  } // end:: k loop through keys to test

} // end::l4c_SesVerifyNVPs()

// ---------------------------------------------------------------
// function : l4c_SesQualifyPathsWithRoot()
// purpose  : Check all input, output paths for a "/", and if
//            missing, prepend the appropriate ROOTDIR in front
// notes    : '/' is ASCII char 47. '\' is ASCII 92. "=34.
char *l4c_SesQualifyPathWithRoot(const char *rootDirPath_p, char *origPath_p) {
  // SesCtl *sesCtl_p = l4c_GetSesCtl();
  char workPath[FILENAME_MAX] = {0};

  if (strchr(origPath_p, '/') == NULL) {
    sprintf(workPath, "%s/%s", rootDirPath_p, origPath_p);
    strcpy(origPath_p, workPath);
    xLogMsg("@A l4c_SesQualifyPathWithRoot: qualified path(%s)~\n", origPath_p);
  } else {
    xLogMsg("@A l4c_SesQualifyPathWithRoot: path (%s) valid as is ~\n",
            origPath_p);
  }

  return origPath_p;
} // end:: l4c_SesQualifyPathWithRoot()

// ---------------------------------------------------------------------
// function :l4c_SesCollectCoreNvps()
// Purpose  : collect "CORE" standardized NVPs from RIP file, loading
//             into destination vars
// Inputs   : SesCtl *sesCtl_p -- ptr to session controls
//          : NsbObj *nsb_p    -- ptr to NSB overseer
// revised  : 2017-06-05T16:57, changed return type to NsbObj_p *, and
//            added parsing of (still optional) L4C_BPLUT_UUID
//            and L4C_BPLUT_MD5 name value pairs. If present, these are
//            stored in sesCtl_p members. If not present, no runtime check
//            is enabled that depends on them. We may make this more strict
//            in the future.
// returns  : NsbMgr *, ptr to name-value-pair string-block NSB overseer
// notes    : TBD, wrap in a mutex lock to make thread safe
NsbObj *l4c_SesCollectCoreNvps(SesCtl *sesCtl_p) {
  // inherit NSB overseer
  NsbObj *nsbMgr_p = l4c_GetNsbOverseer();
  // inherit pkgMgr_p
  // PkgMgr   *pkgMgr_p = l4c_GetPkgMgrOverseer();

  // ...and pThme_p used for FparClimatology special case handling
  // PkgTheme *pThme_p  = NULL;
  // PkgObj   *pObj_p   = NULL;
  char nvpBuf[FILENAME_MAX] = {0};

  char fxName[256] = {0};
  float64 numericBuf;
  // float64 resultFlt64 =0.0 ;

  // register this entry
  strcpy(fxName, (char *)__func__);
  xLogMsg("@A In(%s) organizing all runtime input parameters (RIP)~ \n",
          fxName);

  // fetch: Science Version ID, CRID, release tag, software sw version, "action"
  // operator etc
  strcpy(sesCtl_p->ScienceVersID,
         nsbGetValueByKey(nsbMgr_p, "L4C_SCIENCE_VERSION_ID", nvpBuf));

  strcpy(sesCtl_p->CRID, nsbGetValueByKey(nsbMgr_p, "L4C_CRID", nvpBuf));

  // Note: a "RELEASE_TAG" here refers to the combination of a
  // tightly-coupled collection of required ANCILLARY DATA
  // (BPLUT,SMAP_L4_C_Ancillary.h5,etc) making up a release instance
  strcpy(sesCtl_p->SubsysReleaseTag, "");
  strcpy(sesCtl_p->SubsysReleaseTag,
         nsbGetValueByKey(nsbMgr_p, "L4C_SUBSYSTEM_RELEASE", nvpBuf));

  // <JMG> 2015-06-10, standardized L4C_SOFTWARE_VERSION to L4C_SW_VERSION
  strcpy(sesCtl_p->swVersion,
         nsbGetValueByKey(nsbMgr_p, "L4C_SW_VERSION", nvpBuf));

  // <JMG> 2014-10-06, moved into "CoreNvps" group, DATASET vers from RIP,
  // as distinct from software version producing it.
  strcpy(sesCtl_p->datasetVersion,
         nsbGetValueByKey(nsbMgr_p, "L4C_DATASET_VERSION", nvpBuf));

  strcpy(sesCtl_p->appAction,
         nsbGetValueByKey(nsbMgr_p, "L4C_RUN_MODE", nvpBuf));

  strcpy(sesCtl_p->l4c_Task_Label,
         nsbGetValueByKey(nsbMgr_p, "L4C_TASK_LABEL", nvpBuf));
  strcpy(sesCtl_p->l4c_Task_UUID,
         nsbGetValueByKey(nsbMgr_p, "L4C_TASK_UUID", nvpBuf));
  strcpy(sesCtl_p->l4c_Platform,
         nsbGetValueByKey(nsbMgr_p, "L4C_PLATFORM", nvpBuf));

  // DOY, day of year...
  sesCtl_p->runDayOfYear =
      (int)nsbGetNumericByKey(nsbMgr_p, "L4C_MODEL_YEARDAY", &numericBuf);
  sesCtl_p->runToStopDay =
      sesCtl_p->runDayOfYear;       // same as runDayOfYear for GMAO Ops
  sesCtl_p->runDayStepInterval = 1; // always set to 1 for GMAP Ops

  // <JMG> 2014-11-03, added L4C_MODEL_STOPDAY for future (local UM) multi-day
  // mode
  if (nsbDoesKeyExist(nsbMgr_p, "L4C_MODEL_STOPDAY")) {
    sesCtl_p->runToStopDay =
        (int)nsbGetNumericByKey(nsbMgr_p, "L4C_MODEL_STOPDAY", &numericBuf);
  }
  // <JMG> 2014-11-03, added L4C_MODEL_DAYSTEPINTERVAL for future (local UM)
  // multi-day mode
  if (nsbDoesKeyExist(nsbMgr_p, "L4C_MODEL_DAYSTEPINTERVAL")) {
    sesCtl_p->runDayStepInterval = (int)nsbGetNumericByKey(
        nsbMgr_p, "L4C_MODEL_DAYSTEPINTERVAL", &numericBuf);
  }

  // DOY - Prepare and store the Day of Year string doyStr (now doy_Str so I can
  // tell them apart)
  strcpy(sesCtl_p->doy_Str, "");
  sprintf(sesCtl_p->doy_Str, "%-3.3d", sesCtl_p->runDayOfYear);

  // Get and store MODEL YEAR
  sesCtl_p->runModelYear =
      (int)nsbGetNumericByKey(nsbMgr_p, "L4C_MODEL_YEAR", &numericBuf);
  strcpy(sesCtl_p->year_Str, "");
  sprintf(sesCtl_p->year_Str, "%4d", sesCtl_p->runModelYear);
  // Get FPAR composite period index
  sesCtl_p->runCompositePeriod =
      (int)nsbGetNumericByKey(nsbMgr_p, "L4C_MODEL_PERIOD", &numericBuf);

  // <JMG> 2017-06-05T17:01, add check to parse L4C_BPLUT_UUID and L4C_BPLUT_MD5
  // if present Initialize BPLUT UUID and MD5 strings as empty
  strcpy(sesCtl_p->l4c_BPLUT_UUID, "");
  strcpy(sesCtl_p->l4c_BPLUT_MD5, "");
  if (nsbDoesKeyExist(nsbMgr_p, "L4C_BPLUT_UUID")) {
    strcpy(sesCtl_p->l4c_BPLUT_UUID,
           nsbGetValueByKey(nsbMgr_p, "L4C_BPLUT_UUID", nvpBuf));
  }
  if (nsbDoesKeyExist(nsbMgr_p, "L4C_BPLUT_MD5")) {
    strcpy(sesCtl_p->l4c_BPLUT_MD5,
           nsbGetValueByKey(nsbMgr_p, "L4C_BPLUT_MD5", nvpBuf));
  }

  xLogMsg("@A In(%s) YEARDAY (%d) string(%s) YEAR (%4d) string(%s)~ \n", fxName,
          sesCtl_p->runDayOfYear, sesCtl_p->doy_Str, sesCtl_p->runModelYear,
          sesCtl_p->year_Str);
  // <JMG> 2014-11-03, added for future multi-day UM use only
  xLogMsg("@A In(%s) STOPDAY (%d) DAY STEPINTERVAL %d~ \n", fxName,
          sesCtl_p->runToStopDay, sesCtl_p->runDayStepInterval);

  // Return pointer to allocated but only partially completed nsbMgr_p
  //  parent object, noting the downstream consumers can ALSO access this
  //  via a call to l4c_GetNsbOverseer() if they don't want to catch
  //  this return
  return nsbMgr_p;

} // end:: l4c_SesCollectCoreNvps()

// ---------------------------------------------------------------
// function :   l4c_SesCollectNVPs()
// purpose  : Session wise Name Value Pair (NVP) parameters in a sessions
//            RIP file are catagorized into (2) groups: "core" and "non-core".
//            This function parses the "non-core" NVPs.  Here we dereference
//            and collect 'controls' NVPs from the runtime input parameter (RIP)
//            file for this session. Runtime inputs are NOW stored in (2)
//            structures; SesCtl contains the (non-core) controls oriented
//            settings, and RunParamAdt -->runParamObj_p contains all the rest
//            of the (non-core) RIP parameters, many of which originally resided
//            inside of Lucas' now deprecated GLOB_PARAMS structure.
// inputs   :  SesCtl *sesCtl_p --ptr to session controls and friends
// revised  : 2017-06-05T17:20, changed return type to NsbObj *p.
// revised  : 2014-07-23T17:29, added QcStateLimitAdt processing with Lucas.
//            and note the set of L4C_INPUT_DATASET_xxxx names
//            are now mandatory and must be parsed here along with the others.
// revised  : 2014-02-19T16:51:00, added RunParamAdt..
// notes    : Fetch only the app-specific runtime inputs here, parsing and
// storing
//            these in the Name-value-pair "stringblock" in-memory database
//            ("nsb"), where values are later retrieved, via RIP keywords, using
//            the accessor functions:
//               valuString = nsbGetValueByKey(nsbObj_p,"L4C_CRID",nvpBuf)
//               valuNumeric=
//               nsbGetNumericByKey(nsbObj_p,"L4C_SUBTASK_ID",&resultFlt64)
//
NsbObj *l4c_SesCollectNVPs(void) {
  SesCtl *sesCtl_p =
      l4c_GetSesCtl(); // sesCtl_p = l4c_GetSesCtl() in any other module file
  NsbObj *nsbObj_p = l4c_GetNsbOverseer();

  // <JMG> 2014-07-23T17:29, added for storing valid_min,valid_max and
  // thresholds
  QcStateLimitAdt *qcStateLim_p = l4c_GetQcStateLimits();

  RunParamAdt *runtimeParamObj_p = &sesCtl_p->runParamObj;

  // Not yet used:
  //  PkgMgr *pkgMgr_p = sesCtl_p->l4cPkgMgr_p ;
  //  PkgTheme *pThme_p = NULL;
  //  PkgObj   *pObj_p  = NULL;

  char nvpBuf[FILENAME_MAX] = {0};
  char methodSrcBuf[32] = {0}; // used for GPP or FT source parsing
  char *dsetName_p = NULL;     // ptr to RIP keyword for a NVP to rerieve...
  char fxName[256] = {0};
  float64 numericBuf = 0.0; // float64 as biggest, for re-cycled numeric buffer

  // register this entry
  strcpy(fxName, (char *)__func__);

  // Set Default controls if any
  // Set Default FPAR gain,offset were in glob_grid_params, but now
  // in the RIP file as keys "fparGain","fparOffset"
  runtimeParamObj_p->FPAR_scale = 0.01; // used to be in BPLUT
  runtimeParamObj_p->FPAR_offset = 0.0; // used to be in BPLUT

  sesCtl_p->runPixStepInterval = 1;

  // <JMG> 2014-06-24, DATASET vers from RIP, is distinct from software version
  // producing it.
  strcpy(sesCtl_p->datasetVersion,
         nsbGetValueByKey(nsbObj_p, "L4C_DATASET_VERSION", nvpBuf));

  // <JMG> 2014-09-11, added as more explicit,
  // macro L4C_VERS_TAG, like: "v5.0.1-2014-09-10T16:44:00" attrib written out
  // as L4_C_SW_RELEASE
  strcpy(sesCtl_p->swRelease, l4c_GetExecutableVersionId());

  // <JMG> 2014-11-14, added optional QA RMSE generation. If this keyword
  // is missing, or is set to "NO" or "FALSE" the RMSE suite is NOT yet
  // generated Note that it currently now defaults to 1 (ON).
  if (nsbDoesKeyExist(nsbObj_p, "L4C_ENABLE_RMSE")) {
    nsbGetValueByKey(nsbObj_p, "L4C_ENABLE_RMSE", nvpBuf);
    sesCtl_p->enableRMSE = frm_utilTestBooleanStr(frm_strToUpper(nvpBuf));
  }

  // <JMG> added 2014-09-26, default is to always enable daily SOC update
  sesCtl_p->enableSOCUpdate = 1;
  if (nsbDoesKeyExist(nsbObj_p, "L4C_ENABLE_SOC_UPDATE")) {
    nsbGetValueByKey(nsbObj_p, "L4C_ENABLE_SOC_UPDATE", nvpBuf);
    sesCtl_p->enableSOCUpdate = frm_utilTestBooleanStr(frm_strToUpper(nvpBuf));
  } // end:: if block if ENABLE_SOC_UPDATE key was specified

  // if enabled, causes each registere function to display a
  // affirmative message when it completes. Used only in testing.
  sesCtl_p->enableCompleteMsg = 0;

  // test for YES,ON,TRUE,1,ENABLED, etc
  if (nsbDoesKeyExist(nsbObj_p, "L4C_ENABLE_COMPLETION_MSG")) {
    nsbGetValueByKey(nsbObj_p, "L4C_ENABLE_COMPLETION_MSG", nvpBuf);
    sesCtl_p->enableCompleteMsg =
        frm_utilTestBooleanStr(frm_strToUpper(nvpBuf));
  }

  // check if granule metrics should be dervied
  sesCtl_p->enableGranuleMetrics = 0;

  if (nsbDoesKeyExist(nsbObj_p, "L4C_GRANULE_METRICS")) {
    nsbGetValueByKey(nsbObj_p, "L4C_GRANULE_METRICS", nvpBuf);
    sesCtl_p->enableGranuleMetrics =
        frm_utilTestBooleanStr(frm_strToUpper(nvpBuf));
  }

  // <JMG>    2014-09-12T11:36, added, was omitted earlier
  // Fetch L4C_PRODUCTION_LEVEL  = local | production (when at GMAO)
  strcpy(sesCtl_p->productionLevel, "production"); // default
  if (nsbDoesKeyExist(nsbObj_p, "L4C_PRODUCTION_LEVEL"))
    strcpy(sesCtl_p->productionLevel,
           nsbGetValueByKey(nsbObj_p, "L4C_PRODUCTION_LEVEL", nvpBuf));

  // Retrieve SOC and FT calendar mode switch from RIP file
  // PROCESSING CALENDAR PROTOCOLS USED:
  //  opMode =0 forces a ZERO OUT of all (366) day slots in calendar.
  //  opMode =1 adjusts all slots to 1 from beginning of year to curr_doy,
  //  inclusive. opMode =2 adjusts JUST the curr_doy slot to 1.
  sesCtl_p->socOpMode =
      (int)nsbGetNumericByKey(nsbObj_p, "L4C_CALENDAR_MODE", &numericBuf);
  // set Freeze Thaw (FT) calendar mode to match SOC unless otherwise changed
  // later.
  sesCtl_p->ftOpMode = sesCtl_p->socOpMode;

  // mandatory: L4C_GPP_SOURCE, Fetch and store GPP algorithm source
  // from choices: {FPAR|NDVI|CLIM}.
  strcpy(methodSrcBuf, nsbGetValueByKey(nsbObj_p, "L4C_GPP_SOURCE", nvpBuf));
  // store FPAR|NDVI|CLIM string to match l4c_MODIS style convention on this
  // <JMG> 2014-10-21, added provision to FORCE use of FparClimatology
  //  where if RIP's gppMethodStr is "CLIM" or "FPARCLIM"
  //  this allows forced use of FparClimatology
  strcpy(sesCtl_p->gppMethodStr, methodSrcBuf);

  // Note that currently there is NO scenario trapped that results in
  // fparState being assigned to msngFparFile. This COULD be implemented
  // at the OPS level by writing a L4C_GPP_SOURCE = "MISSING"

  if (strstr(sesCtl_p->gppMethodStr, "CLIM") !=
      NULL) // allow forced use of FparClimatology
    sesCtl_p->fparState = DoUseFparClim;
  else if (strcmp(sesCtl_p->gppMethodStr, "FPAR") == 0)
    sesCtl_p->fparState = DoUseFpar;
  else if (strstr(sesCtl_p->gppMethodStr, "NDVI") != NULL) // use NDVI
    sesCtl_p->fparState = DoUseFparClim;
  else // just in case an unexpected option used, then default to FPAR CLIM use
    sesCtl_p->fparState = DoUseFparClim;

  // Note: if the L4C_FT_SOURCE directive is present, it takes precedence
  //  over the algorithmic logic for selecting L3_SM_A:surface_flag's
  // 'tsurf' at bit 7, or L3_SM_A_RADAR_FT
  //  RIP string Choices   : {USE_TSURF_FT|USE_RADAR_FT|USE_NO_FT}
  //  Enum matching choices:
  // DoUseFT_ActiveRadar, DoUseFT_ActiveTsurf,
  // DoUseFT_PassiveRadiom,DoUseFT_PassiveTsurf, DoUseFT_GEOS5, DoUse_NO_FT
  strcpy(methodSrcBuf, "");
  // option must be in all upper case here. Enforce this later via
  // frm_strToUpper() or equiv
  strcpy(methodSrcBuf, nsbGetValueByKey(nsbObj_p, "L4C_FT_SOURCE", nvpBuf));

  // <JMG> 2015-07-29, added support for new 3km F/T from GEOS5 or Passive
  // radiometer L3_SM_P DoUseFT_ActiveRadar, DoUseFT_ActiveTsurf,
  // DoUseFT_PassiveRadiom,DoUseFT_PassiveTsurf, DoUseFT_GEOS5
  if (strcmp(methodSrcBuf, "USE_RADAR_FT") == 0) // legacy, leave in
    sesCtl_p->ftState = DoUseFT_ActiveRadar;
  else if (strcmp(methodSrcBuf, "USE_TSURF_FT") == 0) // legacy, leave in
    sesCtl_p->ftState = DoUseFT_ActiveTsurf;
  else if (strcmp(methodSrcBuf, "USE_GEOS5_FT") == 0)
    sesCtl_p->ftState = DoUseFT_GEOS5;
  else if (strcmp(methodSrcBuf, "USE_PASSIVE_FT") == 0)
    sesCtl_p->ftState = DoUseFT_PassiveRadiom;
  else if (strcmp(methodSrcBuf, "USE_PASSIVE_TSURF_FT") == 0)
    sesCtl_p->ftState = DoUseFT_PassiveTsurf;
  else
    sesCtl_p->ftState = DoUse_NO_FT;
  // <JMG> 2015-09-06T18:55, also store in string form
  strcpy(sesCtl_p->ftMethodString, methodSrcBuf);

  // <JMG> 3/4/2014, verified subtaskID is a base0 counter that starts at 0 not
  // 1!!
  sesCtl_p->subtaskId =
      (int)nsbGetNumericByKey(nsbObj_p, "L4C_SUBTASK_ID", &numericBuf);
  sesCtl_p->nSubtasks =
      (int)nsbGetNumericByKey(nsbObj_p, "L4C_N_SUBTASKS", &numericBuf);
  // <JMG> 5/16/2014, facilitates testing via systematic skip-sampling in 9KM
  // space
  if (nsbDoesKeyExist(nsbObj_p, "L4C_PIXEL_STEPINTERVAL"))
    sesCtl_p->runPixStepInterval = (size_t)nsbGetNumericByKey(
        nsbObj_p, "L4C_PIXEL_STEPINTERVAL", &numericBuf);

  // Determine number of spatial array elements to process
  sesCtl_p->nElem9kmPerTask = EGv2_LDOM9KM / sesCtl_p->nSubtasks;
  sesCtl_p->nElem3kmPerTask = EGv2_LDOM3KM / sesCtl_p->nSubtasks;
  sesCtl_p->nElem1kmPerTask = EGv2_LDOM1KM / sesCtl_p->nSubtasks;

  // Offset of spatial array elements to process
  sesCtl_p->elem9kmOffset = sesCtl_p->nElem9kmPerTask * sesCtl_p->subtaskId;
  sesCtl_p->elem3kmOffset = sesCtl_p->nElem3kmPerTask * sesCtl_p->subtaskId;
  sesCtl_p->elem1kmOffset = sesCtl_p->nElem1kmPerTask * sesCtl_p->subtaskId;

  // Next,  just once, construct the taskIdStr (was in CONT...)
  // Prepare task id string. was in CONT.taskId,CONT.nTasks, migrated to SesCtl
  // sesCtl_p sprintf(sesCtl_p->taskIdStr,"%02dof%02d", CONT.taskId+1,
  // CONT.nTasks);
  sprintf(sesCtl_p->subtaskIdStr, "%-2.2d of %-2.2d", sesCtl_p->subtaskId,
          sesCtl_p->nSubtasks);

  sesCtl_p->compressLev =
      (int)nsbGetNumericByKey(nsbObj_p, "L4C_COMPRESSION_LEVEL", &numericBuf);

  strcpy(sesCtl_p->inputRootDir,
         nsbGetValueByKey(nsbObj_p, "L4C_INPUT_ROOTDIR", nvpBuf));
  strcpy(sesCtl_p->inputAncPath,
         nsbGetValueByKey(nsbObj_p, "L4C_INPUT_ANCILLARY_PATH", nvpBuf));
  strcpy(sesCtl_p->inputSOCPath,
         nsbGetValueByKey(nsbObj_p, "L4C_INPUT_SOC_PATH", nvpBuf));
  strcpy(sesCtl_p->inputMetPath,
         nsbGetValueByKey(nsbObj_p, "L4C_INPUT_MET_PATH", nvpBuf));

  // normally, SMAP_L4_C_FparClimatology.h5 lives in the ./ANC directory
  strcpy(sesCtl_p->inputFparClimatologyPath,
         nsbGetValueByKey(nsbObj_p, "L4C_INPUT_FPAR_CLIMATOLOGY_PATH", nvpBuf));

  // <JMG> 20o14-11-10, When forcing FparClimatology via L4C_GPP_SOURCE = CLIM,
  // FPAR 8DAY (may not) always be specified
  //  particularly if it was missing at Ops level due to latency.
  //  This same issue doesn't exist for when NDVI is used, since the NDVI (or
  //  FPAR) data itself is already inside the HDF5 file named by inputFparPath
  if (nsbDoesKeyExist(nsbObj_p, "L4C_INPUT_FPAR_8DAY_PATH")) {
    strcpy(sesCtl_p->inputFparPath,
           nsbGetValueByKey(nsbObj_p, "L4C_INPUT_FPAR_8DAY_PATH", nvpBuf));
  } else {
    strcpy(sesCtl_p->inputFparPath, "FPAR-8DAY-WAS-NOT-SPECIFIED");
  }

  // L4C_ANC_DATASET_LATITUDE_9KM_COORD  = SMAP_L4_C_LAT_1624_x_3856_M09_flt32
  strcpy(
      sesCtl_p->latCoordVarname,
      nsbGetValueByKey(nsbObj_p, "L4C_ANC_DATASET_LATITUDE_9KM_COORD", nvpBuf));
  // L4C_ANC_DATASET_LONGITUDE_9KM_COORD = SMAP_L4_C_LON_1624_x_3856_M09_flt32
  strcpy(sesCtl_p->lonCoordVarname,
         nsbGetValueByKey(nsbObj_p, "L4C_ANC_DATASET_LONGITUDE_9KM_COORD",
                          nvpBuf));

  // <JMG> fetch root of log file which MAY differ from OUTPUT root dir
  strcpy(sesCtl_p->logRootDir,
         nsbGetValueByKey(nsbObj_p, "L4C_OUTPUT_DAILY_LOG_ROOTDIR", nvpBuf));

  strcpy(sesCtl_p->inputBplutPath,
         nsbGetValueByKey(nsbObj_p, "L4C_INPUT_BPLUT_PATH", nvpBuf));

  // <KAE> 2024-01-03 For litterfall allocation scheme
  if (nsbDoesKeyExist(nsbObj_p, "L4C_LITTERFALL_SCHEME")) {
    strcpy(sesCtl_p->litterfallScheme,
           nsbGetValueByKey(nsbObj_p, "L4C_LITTERFALL_SCHEME", nvpBuf));
    strcpy(sesCtl_p->litterfallFileTemplate,
           nsbGetValueByKey(nsbObj_p, "L4C_LITTERFALL_FILE_TPL", nvpBuf));
  }

  // For now these are dummy calls but appear here in order to remind
  // and enforce that the L4C_INPUT_DATASET_xxxx names are
  // ALL NOW MANDATORY in the RIP file. These name the
  //  collection of datasets in the HDF5 files (MET,FPAR_8DAY, etc)
  // Note: these calls here intentionally perform a passive exercise of
  //  the fetch for various expected name-value-pairs, so that if any
  //  are missing, it triggers an informative error message to be sent
  //  to the log file.
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_FPAR", nvpBuf);
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_FPAR_QC", nvpBuf);
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_NDVI", nvpBuf);
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_NDVI_QC", nvpBuf);
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_FT_ROLLING", nvpBuf);
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_FT_ROLLING_AGE", nvpBuf);
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOC_C0", nvpBuf);
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOC_C1", nvpBuf);
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOC_C2", nvpBuf);
  dsetName_p = nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_NPP_SUM", nvpBuf);
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_QV2M_M09_AVG", nvpBuf);
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_RAD_SW_DOWN", nvpBuf);
  dsetName_p = nsbGetValueByKey(
      nsbObj_p, "L4C_INPUT_DATASET_SM_ROOTZONE_WETNESS", nvpBuf);
  dsetName_p = nsbGetValueByKey(nsbObj_p,
                                "L4C_INPUT_DATASET_SM_SURFACE_WETNESS", nvpBuf);
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SOIL_TEMP_LAYER1", nvpBuf);
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_SURFACE_PRESSURE", nvpBuf);
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_T2M_M09_AVG", nvpBuf);
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_T2M_M09_MAX", nvpBuf);
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_T2M_M09_MIN", nvpBuf);

  dsetName_p = nsbGetValueByKey(
      nsbObj_p, "L4C_INPUT_DATASET_SOILROOTZONE_M09_MIN", nvpBuf);
  dsetName_p = nsbGetValueByKey(
      nsbObj_p, "L4C_INPUT_DATASET_SOILROOTZONE_M09_MAX", nvpBuf);

  // Old (surface_flag) replaced by ft_state_M03
  // dsetName_p =
  // nsbGetValueByKey(nsbObj_p,"L4C_INPUT_DATASET_SURFACE_FLAG_M03", nvpBuf);
  dsetName_p =
      nsbGetValueByKey(nsbObj_p, "L4C_INPUT_DATASET_FT_STATE_UM_M03", nvpBuf);

  // dsetName_p = nsbGetValueByKey(nsbObj_p,"L4C_INPUT_DATASET_VPD_AVG"
  // ,nvpBuf);

  // fetch OUTPUT root
  strcpy(sesCtl_p->outputRootDir,
         nsbGetValueByKey(nsbObj_p, "L4C_OUTPUT_ROOTDIR", nvpBuf));

  // fetch primary L4_C HDF5 file's path.
  // <JMG> 2014-09-03T14:34, Note: the separate QA HDF5 file has now been
  // deprecated
  strcpy(sesCtl_p->output_L4C_DAILY_Path,
         nsbGetValueByKey(nsbObj_p, "L4C_OUTPUT_DAILY_PATH", nvpBuf));

  // FETCH valid min,max and thresholds assigned to QcStateLimitAdt qaLim
  // structure
  qcStateLim_p = l4c_GetQcStateLimits();

  // <JMG> 2014-08-21T19:29, bug in
  qcStateLim_p->neeLimitMin = (float32)nsbGetNumericByKey(
      nsbObj_p, "L4C_INPUT_NEE_VALID_MIN", &numericBuf);
  qcStateLim_p->neeLimitMax = (float32)nsbGetNumericByKey(
      nsbObj_p, "L4C_INPUT_NEE_VALID_MAX", &numericBuf);

  qcStateLim_p->gppLimitMin = (float32)nsbGetNumericByKey(
      nsbObj_p, "L4C_INPUT_GPP_VALID_MIN", &numericBuf);
  qcStateLim_p->gppLimitMax = (float32)nsbGetNumericByKey(
      nsbObj_p, "L4C_INPUT_GPP_VALID_MAX", &numericBuf);
  qcStateLim_p->rhLimitMin = (float32)nsbGetNumericByKey(
      nsbObj_p, "L4C_INPUT_RH_VALID_MIN", &numericBuf);
  qcStateLim_p->rhLimitMax = (float32)nsbGetNumericByKey(
      nsbObj_p, "L4C_INPUT_RH_VALID_MAX", &numericBuf);
  qcStateLim_p->socLimitMin = (float32)nsbGetNumericByKey(
      nsbObj_p, "L4C_INPUT_SOC_VALID_MIN", &numericBuf);
  qcStateLim_p->socLimitMax = (float32)nsbGetNumericByKey(
      nsbObj_p, "L4C_INPUT_SOC_VALID_MAX", &numericBuf);

  qcStateLim_p->useFparClimThres = (float32)nsbGetNumericByKey(
      nsbObj_p, "L4C_INPUT_GPP_METHOD_THRESHOLD", &numericBuf);
  qcStateLim_p->useNdviThres = (float32)nsbGetNumericByKey(
      nsbObj_p, "L4C_INPUT_NDVI_METHOD_THRESHOLD", &numericBuf);

  // Do path qualifications
  l4c_SesQualifyPathWithRoot(sesCtl_p->inputRootDir, sesCtl_p->inputAncPath);
  l4c_SesQualifyPathWithRoot(sesCtl_p->inputRootDir, sesCtl_p->inputSOCPath);
  l4c_SesQualifyPathWithRoot(sesCtl_p->inputRootDir, sesCtl_p->inputMetPath);
  l4c_SesQualifyPathWithRoot(sesCtl_p->inputRootDir, sesCtl_p->inputFparPath);
  l4c_SesQualifyPathWithRoot(sesCtl_p->inputRootDir,
                             sesCtl_p->inputFparClimatologyPath);
  l4c_SesQualifyPathWithRoot(sesCtl_p->inputRootDir,
                             sesCtl_p->inputMetClimatology);
  l4c_SesQualifyPathWithRoot(sesCtl_p->inputRootDir, sesCtl_p->inputBplutPath);

  // Validate, quality output and QA output paths
  l4c_SesQualifyPathWithRoot(sesCtl_p->outputRootDir,
                             sesCtl_p->output_L4C_DAILY_Path);

  runtimeParamObj_p->fpar_qa_year =
      atoi(nsbGetValueByKey(nsbObj_p, "fpar_qa_year", nvpBuf));
  runtimeParamObj_p->nsite = atoi(nsbGetValueByKey(nsbObj_p, "Nsites", nvpBuf));

  // SUBSET driver needs all these...
  runtimeParamObj_p->spinup_max_iter =
      (int)nsbGetNumericByKey(nsbObj_p, "spinup_max_iter", &numericBuf);

  runtimeParamObj_p->spinup_stop_tol =
      (float32)nsbGetNumericByKey(nsbObj_p, "spinup_stop_tol", &numericBuf);

  runtimeParamObj_p->FPAR_scale =
      (float32)nsbGetNumericByKey(nsbObj_p, "fparGain", &numericBuf);
  runtimeParamObj_p->FPAR_offset =
      (float32)nsbGetNumericByKey(nsbObj_p, "fparOffset", &numericBuf);

  // <Jl4c_DAILY.makMG> 2014-08-22T01:55:00, added
  runtimeParamObj_p->NDVI_scale =
      (float32)nsbGetNumericByKey(nsbObj_p, "ndviGain", &numericBuf);
  runtimeParamObj_p->NDVI_offset =
      (float32)nsbGetNumericByKey(nsbObj_p, "ndviOffset", &numericBuf);

  memset(sesCtl_p->fparClimDatasetname, '\0',
         sizeof(sesCtl_p->fparClimDatasetname));
  // <JMG> 4/22/2014, minimally validate runDayOfYear and runCompositePeriod
  assert(sesCtl_p->runDayOfYear >= 1);
  assert(sesCtl_p->runCompositePeriod >= 1);

  // Define FparClimatology dataset name from PERIOD and YEARDAY
#ifdef DEPRECATED_OLD_METHOD
  sprintf(sesCtl_p->fparClimDatasetname, "fpar_clim_M01_day%-3.3d_per%-2.2d",
          sesCtl_p->runDayOfYear, sesCtl_p->runCompositePeriod);
#endif

  // <JMG> 2014-10-22T13:04 refined to also use day of year index
  // SEARCH FparClimatolgy PERIOD LUT for match
  // on this period and day of year
  l4c_SetFparClimatologyDatasetName(
      sesCtl_p->runDayOfYear, sesCtl_p->runCompositePeriod,
      "USES-BUILT-IN-LUT-NOW", sesCtl_p->fparClimDatasetname);

  xLogMsg("@A In(%s): FparClimatology dataset name used [%s] ~\n", fxName,
          sesCtl_p->fparClimDatasetname);

  // DIAG Only -- generate report of entire NSB database.
  //  Re-verified as ALL CORRECT on 2014-09-012T10;10:00
  //  nsbExportDbToText(sesCtl_p->sesNsbDb_p,"L4C_RIP_NVP_VERIFY.txt",' ');

  l4c_EventCompletionMsg(fxName);

  // <JMG> 2017-06-05, return ptr to completed NSB parent object. Callers can
  //  optionally also fetch this ptr via a call to l4c_GetNsbOverseer() if
  //  they don't want to catch this return
  return nsbObj_p;

} // end:: l4c_SesCollectNVPs()

// --------------------------------------------------------------------
// purpose  : Fetch the HDF5 dataset name in SMAP_L4_C_FparClimatology.h5
//            matching this YEARDAY and 8-day PERIOD
// function : l4c_SetFparClimatologyDatasetName()
// inputs   : const int dayOfyear,  base1: 1..366 model day-of-year
//            const int compositePeriodIdx -- base1: 8day composite index 1..46
//            char *lutPath_p, -- LUT path
//            char *dsetName_p -- dataset name (string) returned for dataset
//            in SMAP_L4_C_FparClimatology.h5 holding requested days FPAR data.
// returns  : int, status 0 if ok -1 if failed
// revised  : 2014-10-22T12:29:00-07:00, reviewed for use-case of Jan 15-20 for
// ORT/ORR outputs  : char *dsetName_p -- FparClimatology dataset name requested
// notes    : may change to return char *dsetName_p
//
// notes    : Example of first few records of LUT:
// "001,01,01,fpar_clim_M01_day001_per01", // for: day 001..008
// "009,02,01,fpar_clim_M01_day009_per02", // for: day 009..016
// "017,03,02,fpar_clim_M01_day017_per03", // for: day 017..024
// "025,04,02,fpar_clim_M01_day025_per04", // for: day 020..032
// "033,05,03,fpar_clim_M01_day033_per05",
//
int l4c_SetFparClimatologyDatasetName(const int dayOfYear,
                                      const int compositePeriodIdx,
                                      char *lutPath_p, char *dsetName_p) {
  // SesCtl *sesCtl_p = l4c_GetSesCtl();
  char fxName[256] = {0};
  char **doyPeriodLutStrTbl_p = NULL;
  char lutRecdStr[64] = {0};
  char dsetBuffer[64] = {0};
  char prevLutName_p[64] = {0}; // allows for lag to fetch one past current key
  char *periodStr_p = NULL;
  char *dayStr_p = NULL;
  int ithLine = 0;
  int nLines = 0; // count of lines read as returned ...
  int chosenIdx = 0;
  int lutRecPeriodIndex = 0;
  int lutDayOfYearIndex = 0;

  strcpy(fxName, __func__);

  // LUT data now supplied via a static hard wired
  // string table instead of being read from a file as below.

  // doyPeriodLutStrTbl_p = frm_ioReadTextIntoTable(lutPath_p,
  //  DoIgnoreComments, &maxWidthCharFound, &nLines);

  // now uses static string table method
  nLines = 46;
  // set stringtable pointer to point to built in static table
  doyPeriodLutStrTbl_p = fparClimatologyNameLut;

  // scan for match on sesCtl_p->runCompositePeriod

  for (ithLine = 0; ithLine < nLines; ithLine++) {
    // extract the period index at end of record
    strcpy(lutRecdStr, doyPeriodLutStrTbl_p[ithLine]);

    periodStr_p = strstr(lutRecdStr, "_per") + strlen("_per");
    dayStr_p = strstr(lutRecdStr, "_day") + strlen("_day");

    // store LUT's 8-day-period as an integer index
    lutRecPeriodIndex = atoi(periodStr_p);
    // store LUT's day-of-year as an interger index
    lutDayOfYearIndex = atoi(dayStr_p);

    // DIAG only
    // xLogMsg("@D In(%s) Target DOY(%d) Period(%d) LUT DOY(%d)PER(%d)~\n",
    //    fxName, dayOfYear, compositePeriodIdx, lutDayOfYearIndex,
    //    lutRecPeriodIndex) ;

    // if invalid, skip ahead
    if (periodStr_p == NULL || lutRecPeriodIndex > 46) {
      xLogMsg("@E In(%s) error reading LUT(%s) at line %d ~\n", fxName,
              lutPath_p, ithLine);
      continue;
    }

    // On match, fetch whole 4th token as the
    // FparClimatology dataset name for this YEARDAY PERIOD combo
    if (compositePeriodIdx == lutRecPeriodIndex) {
      // fetch 4th token, identified as 3rd token since this is base0
      frm_strGetToken(dsetBuffer, lutRecdStr, 3, ",");
      if (strlen(dsetBuffer) > 0) {
        // <JMG> 2014-10-22T12:38, bug: the actual day-of-year
        // is not yet always correct...
        strcpy(dsetName_p, dsetBuffer);
        break;
      } // end if valid result
    }   // end:: block on correct match

    // store lag by-one entry, as fallback
    strcpy(prevLutName_p, dsetBuffer);
    // increment count of LUT records scanned
    chosenIdx++;
  } // end:: ithLine

  // fallback: if chosenIdx >= nLines, warn use and
  // pick the 1st or last rec arbitrarily so we don't crash!
  if (chosenIdx >= nLines) {
    strcpy(dsetName_p, prevLutName_p);
    xLogMsg("@E In(%s) Potential LUT match-failure! Fallback: uses "
            "FparClim(%s) dataset~\n",
            fxName, dsetName_p);
  }

  // also returns base0 index of line for matching LUT record
  return ithLine;
} // end:: l4c_SetFparClimatologyDatasetName()
