// --------------------------------------------------------------------
// module  : l4c_DAILY_sci.c
// purpose : Implement misc science functions not found elsewhere
// revision: 2014-11-03, moved l4c_AssignFillvalue to l4c_GridPost_Inflate.c
//           where it belongs with functions using it.
// revision: 2014-04-10T14:01, added BIT32[] lookup array, and added
//           l4c_testByteorder(), l4c_GetBits(), l4c_SetBits(), all from 
//           NASA MODIS L4 {FPAR,LAI,PSN,NPP} land instrument team code suite.
// revision: 2014-02-12T20:40:00
// author  : joe glassy
// notes   : The bit handling code implemented here is not guaranteed
//           (and has not been tested) to be thread-safe.
// --------------------------------------------------------------------

#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"


int L4C_BIT32[] = 
{
  0x1,        0x3,        0x7,        0xF,
  0x1F,       0x3F,       0x7F,       0xFF,
  0x1FF,      0x3FF,      0x7FF,      0xFFF,
  0x1FFF,     0x3FFF,     0x7FFF,     0xFFFF,
  0x1FFFF,    0x3FFFF,    0x7FFFF,    0xFFFFF,
  0x1FFFFF,   0x3FFFFF,   0x7FFFFF,   0xFFFFFF,
  0x1FFFFFF,  0x3FFFFFF,  0x7FFFFFF,  0xFFFFFFF,
  0x1FFFFFFF, 0x3FFFFFFF, 0x7FFFFFFF, 0xFFFFFFFF
} ;


// ---------------------------------------------------------------------
// function : l4c_TestBooleanStr()
// inputs   : const char *booleanOptionSwitch_p
// returns  : int, boolean, 1=if true 0=false
// revision : 2014-09-12T15:37:00
// author   : joe glassy, Lupine Logic Inc.
// purpose  : Optionally rescales output array, and, derives and returns
int l4c_TestBooleanStr(const char *optionSwitch_p)
{
 int resultBoolean = 0 ;
 char evalBuffer[32] = {0} ;
 
  strcpy(evalBuffer,optionSwitch_p);
  // normalize...
  frm_strToUpper(evalBuffer);
  // test
  
  // allow for a variety of responses to equal true...
  if(strcmp(evalBuffer,"YES")  ==0 || 
     strcmp(evalBuffer,"ON")   ==0 || 
     strcmp(evalBuffer,"TRUE") ==0 || 
     strcmp(evalBuffer,"1")    ==0 ||   
     strstr(evalBuffer,"ENABLE")!=NULL)
  {
        // set TRUE if any of these match
        resultBoolean = 1 ;
        
  } else if (strcmp(evalBuffer,"NO")  ==0 || 
     strcmp(evalBuffer,"OFF")   ==0 || 
     strcmp(evalBuffer,"FALSE") ==0 || 
     strcmp(evalBuffer,"0")    ==0 ||   
     strstr(evalBuffer,"DISABLE")!=NULL)
  {
        // set FALSE when explicitly told to
        resultBoolean = 0 ;
  } else {
      // set FALSE conservatively
        resultBoolean = 0 ; 
  }
  
 return resultBoolean ;
} // end:: l4c_TestBoolean()

        
// --------------------------------------------------------------------
// l4c_calc_vpd()
// Purpose:  -- compute vapor pressure deficit (in Pa units) from 
//           MERRA's QV2M specific humidity and PS (surface pressure) in Pa
// Inputs: 
//   tavg -- daily average temperature in degC (if in Kelvin, must convert first)
//   qv2m -- specific humidity, in MERRA units
//   ps   -- surface pressure, in Pa (in MERRA units)
// Returns: vpd, in Pa units
float32 l4c_calc_vpd(const float32 tavg,const float32 qv2m,const float32 ps)
{
 float32 tavg_degC = tavg - 273.15; // Convert from K to deg C
 float32 vpd = 0.0; // vapor pressure deficit
 float32 esat= 0.0; // saturated vapor pressure
 float32 avp = 0.0; // ambient (actual) vapor pressure
 

    // <Yonghong>: used this logic in MERRA_hdf_process.c (as shown here)
    avp = (qv2m * ps) / (0.622 + (0.378 * qv2m)); // here, in Pa  units
    
    // Note: replaced orig:  avp = avp * 1000.0 ;  to convert kpa to pa 
    esat =  610.7 * exp(17.38 * tavg_degC/(239.0 + tavg_degC));
    
    vpd = (esat - avp);
    // bound it...
    vpd = (vpd < 0.0)? 0.0 : esat - avp;
  // return VPD in Pa units 
 return vpd ;
} // end:: calc_vpd()


// --------------------------------------------------------------------
// Purpose: boolean tests if a (IEEE double precison real) number 'a' 
// is essentially equal to 'b' and returns IsTrue if they are 
// within EQ_TOLERANCE of each other.
int l4c_are_doubles_equal(const double a,const double b)
{
 int    boolResult = 0;
 double F64_TOLERANCE = 0.0000000001 ;

     if(fabs(a-b) <= F64_TOLERANCE)
         boolResult = 1;
 // return boolean result if 'a' is 'very close to equal' with 'b'
 return boolResult ;
} // end::l4c_are_doubles_equal()


// --------------------------------------------------------------------
// Purpose: boolean tests if a (IEEE single precision real) number 
//   'a' is essentially equal to 'b' and returns IsTrue if they 
//   are within EQ_TOLERANCE of each other.
int l4c_are_floats_equal(const float a,const float b)
{
 int   boolResult = 0;
 float F32_TOLERANCE = 0.000001f ;

     if(fabs((double)a - (double)b) <= F32_TOLERANCE)
         boolResult = 1;
 // return boolean result if 'a' is 'very close to equal' with 'b'
 return boolResult ; 
} // end::l4c_are_floats_equal()


// ------------------------------------------------------------
// function : l4c_sizOfDatatype()
// inputs   : const int hdf4Datatype
// returns  : size_t, size in bytes of scalar instance of indicated type
// purpose  : Returns the size in bytes of one scalar of indicated type. 
//            This is a generic substitute for HDF4's DFKNTsize() to allow 
//            for small apps to not have to link against HDF4 libs
// notes    : n/a
size_t l4c_sizOfDatatype(const int hdf4Datatype)
{
 size_t siz = 0 ;
 
    switch(hdf4Datatype)
    {
    case IsChar    :siz= sizeof(char)  ; break;
    case IsUint8   :siz= sizeof(uint8) ; break;
    case IsInt16   :siz= sizeof(int16) ; break;
    case IsUint16  :siz= sizeof(uint16); break;
    case IsInt32   :siz= sizeof(int32)  ; break;
    case IsUint32  :siz= sizeof(uint32) ; break;
    case IsInt64   :siz= sizeof(int64)  ; break;
    case IsUint64  :siz= sizeof(uint64) ; break;
    case IsFloat32 :siz= sizeof(float)  ; break;
    case IsFloat64 :siz= sizeof(double) ; break;
    default : 
        break;
    }
    return siz ;
} // end:: l4c_sizOfDatatype()


// -----------------------------------------------------------------------------
// function : l4c_test_Byteorder()
// purpose  : evaluates byte ordering on current platform and returns
//            0 if BigEndian, 1 if Little endian.
// Note     : the emum defining IsLSBfirst etc is located in the UM Framework API, 'frm_core.h"
// Notes    : the enum order of symbols is CRITICAL here, so these MUST conform to this ordering:
//            enum EndianStates { isMSBfirst, isLSBfirst, isVAXEndian, unknownEndian } ; 
//            char *endianLabels[] = {"big","little"} ;
uint8 l4c_testByteorder(void)
{
 uint8 rtn = 0;
 union {
   int i ;
   char c[sizeof(int)] ;
 } foo ;
 foo.i = 1 ;

 if(foo.c[0] == 1)
   rtn = isLSBfirst ;
 else
   rtn = isMSBfirst ;
 return rtn;
} // end::l4c_testByteorder() 


// -----------------------------------------------------------------------------
// function: l4c_GetBits()
// inputs     :  int16 startPos --the base0 starting bit index of the (data) to retrieve
//            :  int16 endPos -- the base0 ending bit index of the (data) to retrieve
//            :  void *data -- pointer to the data (word) to retrieve
//            :  size_t dataSize -- the size in bytes of the data (word) to retrieve, e.g. sizeof(uint8) etc.
//               where dataSize is in the supported range of 1 (byte) to 4 (uint32)
// purpose : from an integer 'data' of size 'dataSize', retrieve bits from 
//           base0 bit positions 'startPos' to 'endPos' inclusive, returning the result
//           as a 32 bit unsigned integer, which may itself be down-cast to a smaller
//           result scalar such as a byte,int16, or uint16
// Note that the enum defining IsMSBfirst and IsLSBFirst is located in frm_core.h
// Notes: (Petr Votava): All byte orderings are normalized to a standard
//           big endian (most significant byte first) ordering.
//
uint32 l4c_GetBits(const int16 startPos,const int16 endPos, void *data,size_t dataSize)
{
  int16  byteOrder = l4c_testByteorder(); // freshen retrieval of platforms byte-ordering
  uint32  localData = 0;
  uint8   *localDataPtr = (uint8 *)&localData;

  if(byteOrder == isMSBfirst)
   memcpy(localDataPtr  + (sizeof(uint32) - dataSize), data, dataSize);
  else
   memcpy(localDataPtr, data, dataSize);

  return((uint32)((localData >> startPos) & L4C_BIT32[endPos - startPos]));
  
} // end::l4c_GetBits()


// ---------------------------------------------------------------------
// function   : l4c_SetBits()
// inputs     :  int16 startPos --the base0 starting bit index of the (data) to modify
//            :  int16 endPos -- the base0 ending bit index of the (data) to modify
//            :  void *data -- pointer to the data (word) to modify
//            :  size_t dataSize -- the size in bytes of the data (word) to modify, e.g. sizeof(uint8) etc.
//               where dataSize is in the supported range of 1 (byte) to 4 (uint32)
// purpose    : sets a range of bits explicitly to '1' in 'data', starting at base0 bit 
//              position 'startPos' and ending at bit position endPos.
//              This function is NOT the best one to use when you need to
//              insert a multi-bit QA integer into a QA word. Use AssignBits()
//              when inserting a multi-bit QA integer (e.g. one with a value
//              >1) into a word.
// invocation : l4c_SetBits(strtBit,endingBit, &aUintWord,sizeof(aUintWord));
// history    : from Petr Votava, Joe Glassy from MUM API (MODIS code)
// Note       : The enum defining IsMSBfirst and IsLSBFirst is located in frm_core.h
// notes      : <JMG> 2014-08-21T20:02, in l4c_PostProc_fltVar9km() when
//              this l4c_SetBits() is called for nee_mean, for NDVI inputs only,
//              this sometimes triggers a SEGFAULT --the cause is under investigation.
// notes:     : <JMG> 2014-10-01T12:22, in l4c_PostProc_fltVar9km() when
//             l4c_SetBits() is caleld for nee_mean, for FPAR 8DAY, this once again
//             is triggering a FATAL Program received signal SIGSEGV, Segmentation fault
//             when memcpy is called in: memcpy(localDataPtr,dataToModify,datasize)
//             This is now again under investigation. The same problem 8/21/2014 was
//             noted and appeared to go away after a re-compile and refresh of input
//             data, but obviously, something deeper is going on.
uint32 l4c_SetBits(int16 startPos, int16 endPos, void *dataToModify, size_t dataSize)
{
  char fxName[256] = {0} ;
  
  int16  byteOrder = l4c_testByteorder(); // renamed from mumTestByteorder()...
  uint32 localData = 0;
  uint8  *localDataPtr = (uint8 *)&localData;

    // strcpy(fxName,__func__);
    if(byteOrder == isMSBfirst)
        memcpy(localDataPtr  + (sizeof(uint32) - dataSize), dataToModify, dataSize);
    else
        memcpy(localDataPtr, dataToModify, dataSize);

    localData = localData | (L4C_BIT32[endPos - startPos] << startPos);

  return(localData);
  
} // end::l4c_SetBits() 



// -----------------------------------------------------------------------------
// function: l4c_AssignBits()
// purpose : This function is typically used to ASSIGN (INSERT) an integer 
//           value (occupying multiple bits) into a specified bit position
//           range within a fixed sized destination integer variable.  The
//           origin (multiple-bit) integer value is as supplied via an integer
//          'newValueMask' of size 'dataSize'. 
// return     : uint32 -- final modified value of destination word after the insertion.
// inputs     :  int16 startPos --the base0 starting bit index of the (data) to retrieve
//            :  int16 endPos -- the base0 ending bit index of the (data) to retrieve
//            :  void *destData -- pointer to the data (word) to retrieve
//            :  size_t dataSize -- the size in bytes of the data (word) to 
//               retrieve, e.g. sizeof(uint8) etc.
//               where dataSize is in the supported range of 1 (byte) to 4 (uint32)
//            : void *newValueMask -- the pointer to the NEW integer (multiple-bit)
//              value to INSERT (e.g. ASSIGN) into the destination word (void *destData)
//            : size_t maskSize -- the size of the new integer variable containing
//              the new value to insert into the word.
// Notes      : adapted from mumMaskBits().
// Notes      : the enum defining IsMSBfirst and IsLSBFirst is located in frm_core.h
//
// EXAMPLE: from MODIS pge33, Petr Votava_ 8/15/02 - Yet new QQ schema:
//  isQaBest = 0,
//  isQaGood = 1,
//  isQaFair_BadGeom = 2,
//  isQaPoor = 3,
//  isQaNotPossible = 4
//  The bits are now {5..7}
//
//   NOTES: the SOURCE of the integer is uint8 "lai_obj_p->q_montana"
//          the DESTINATION (word) is qa_field variable (a uint8) 
//          the resulting value will occupy 3-bits, at base0 indices {5,6,7}
//          within the 8-bit result variable named 'qa_field'.
//
//   MUST retrieve the final result (modified word) via "qa_field" as 
//   shown in the example below, as returned to 'qa_field'
//
//    qa_field = mumMaskBits( 5,7, &qa_field, sizeof( uint8 ),
//               &lai_obj_p->q_montana, sizeof( uint8 ) );
//
uint32 l4c_AssignBits(const int16 startPos,const int16 endPos, void *destData, size_t dataSize,
                      void *newValueMask, size_t maskSize)
{
  int16  byteOrder = l4c_testByteorder();
  uint32  localData = 0;
  uint8   *localDataPtr = (uint8 *)&localData; // set pointer to addressable entity
  uint32  localMask = 0;
  uint8   *localMaskPtr = (uint8 *)&localMask; // set pointer to addressable entity

  if(byteOrder == isMSBfirst)
  {
    // offset is (sizeof(uint32) minus the size of the destData variable
    memcpy(localDataPtr  + (sizeof(uint32) - dataSize), destData, dataSize);
    memcpy(localMaskPtr  + (sizeof(uint32) - maskSize), newValueMask, maskSize);
  } else {
    memcpy(localDataPtr, destData, dataSize);  
    memcpy(localMaskPtr, newValueMask, maskSize);
  }  

    // first set 0's to the fields that we want to set, using logical AND  
    localData = localData & (~(L4C_BIT32[endPos - startPos] << startPos));

    // now mask it, using logical OR  
    localData = localData | ((localMask & L4C_BIT32[endPos - startPos]) << startPos);
  
    // Retrieve the new modified (destination) value
    // as a returned uint32 integer.  Cast to appropriate type if not a uint32.
    return(localData);
  
} // end:: l4c_AssignBits()


// ---------------------------------------------------------------------
// function   : l4c_GetISO8601_datetime()
// inputs     : time_t usersSec -- optional, 0, or user provided seconds =time(0L)
//            : char *loc_dtStr_p -- resulting localtime string YYYY-MM-DDThh:mm:ss
//              char *utc_dtStr_p -- resulting UTC string YYYY-MM-DDThh:mm:ssZ
// returns    : int, timezone difference in hours between UTC and local time (not adjusted for sign)
// purpose    : Retrieve an ISO8601 conformal datetime string, YYYY-MM-DDThh:mm:ssZ for UTC
//              position 'startPos' and ending at bit position endPos.
// revised    : 2014-05-13T18:02:00,jmg
// history    : n/a
// notes      : Joe A wants this in form 2014-09-12T09:53:00.000Z not 2014-09-12T09:53:00Z!
int l4c_GetISO8601_datetime(time_t usersSec,char *loc_dtStr_p,char *utc_dtStr_p)
{
 char fmtSpecStr[64]={0} ;
 char tzBuf1[16] = {0} ;
 char tzBuf2[16] = {0} ;
 
 time_t dtSecUsed ;
 struct tm *loc_tmPtr = NULL ; 
 struct tm  *gm_tmPtr = NULL ;
 int    tz_offset_hrdiff = 0;
  
    dtSecUsed = (usersSec <= 0)? time(0L): usersSec;
    
    // capture both the local time and UTC, GMT
    gm_tmPtr  = gmtime(&dtSecUsed);
    // <JMG> 2014-09-12T09:57, GMAP needs datetime stamp in 
    // exactly this format: 2014-09-12T09:57:00.000Z, presume
    // by rounding the 's' param to nearest even sec to result
    // in stamp ending with ".000Z" but we hard code in the .000Z for now
    strcpy(fmtSpecStr,"%Y-%2m-%2dT%H:%M:%S.000Z" );
    strftime(utc_dtStr_p, 64-1,fmtSpecStr,gm_tmPtr);
    strftime(tzBuf1,15,"%H", gm_tmPtr);
    // printf("\ngetISO8601datetime UTC [%s] tzBuf %s\n",utc_dtStr_p,tzBuf1);

    loc_tmPtr = localtime(&dtSecUsed);  
    strcpy(fmtSpecStr,"%Y-%2m-%2dT%H:%M:%S" );
    strftime(loc_dtStr_p,64-1,fmtSpecStr,loc_tmPtr);
    strftime(tzBuf2,15,"%H", loc_tmPtr);
    // printf("\ngetISO8601datetime MST [%s] tzBuf %s\n", loc_dtStr_p, tzBuf2);
    
    tz_offset_hrdiff = abs(atoi(tzBuf2) - atoi(tzBuf1))+1;
    // NOTE: need to correct time-zone hour "sign" to negative for West longitudes?
    // printf("\ntzBuf1 (%s) tzBuf2(%s) %d\n",tzBuf1,tzBuf2,tz_offset_hrdiff);
    
 return tz_offset_hrdiff ;
 
} // end:: l4c_getISO8601_datetime()
