// ---------------------------------------------------------------
// module       : l4c_DAILY_ses.c
// purpose      : Implement Session management and initialization functionality
//                the NASA SMAP l4c_DAILY.ex
// project      : SMAP L4C science development.
// dependencies : HDF4 4.2.x, HDF5 v1.8.x or later libraries, L4C FRAMEWORK APIs
// revised      : 2017-06-05T17:10, added display of L4C_BPLUT_UUID,L4C_BPLUT_MD5 if present in RIP
// revised      : 2016-03-02T17:12 <JMG>  added diagnostic output of all BPLUT info read
//                  sent to log file with a @D tag l4c_BPLUT_Report(stdout, nBplutRecords);
// revised      : 2015-10-14T16:35:00, changed open, close of log file to use
//                internal default, for added robustness
// history      : 2015-09-06T17:31:00, changed log file YD form to YDddd instead of YD_ddd.
// history      : 2015-07-29T13:41:00, v7 NEW FORK, added support for new GEOS5 derived FT 3km
//                added, new default, as: sesCtl_p->ftState = DoUseFT_GEOS5 ;
// history      : 2014-11-03, formalize initialization of runDayOfYear,runDayStepInterval=1,
//                and for SMAP Ops at GMAO, always set runToStopDay==runDayOfYear
//                and sesCtl_p->runDayStepInterval=1.
// history      : 2014-10-06T16:43, added L4C_ISO_MetaElements structure and access fx
//                l4c_Get_ISO_MetaElements()
// history      : 2014-09-04T09:42, consolidate all output into single _mdl_ granule file
// history      : 2014-07-23T17:29, added QcStateLimitAdt processing in l4c_SesCollectNVPs with Lucas.
// history      : 2014-03-11T10:10, changed to open all HDF5 files once (here), to avoid having
//                to close+re-open multiple times.
// history      : 2014-03-05T14:17, added parsing L4C_GPP_SOURCE to set Fpar vs. NDVI switch from RIP file
//                instead of relying on {TRUE|FALSE} flags in the BPLUT.
// history      : 2014-02-20T13:19:00, migrated CONT.taskId,taskIdStr, nElem1km, elem1kmOffset all to
//              : sesCtl from CONT, too many conflicts and risks of using the wrong ones..
// history      : 2014-02-19T21:28:00, patched in RunParamAdt in place of SesCtl for
//                all non-control RIP NVPs.
// history      : 2014-02-16T16:06:00, re-integrate NSB NVP mgr, a FRM_API client API
// history      : 2014-02-13T09:41:00, patched in frm_utilGetCtime().
// history      : 2014-01-27T12:17:00-07:00
// history      : 2013-11-25T17:38:00-07:00 --migration pending to use l4c_sll_CreateNode() or equiv.
//              : 2013-11-18T09:36:00, corrected bug where OUTPUT root dir not prepended to output path
//              : 2013-11-08T08:32:00, initial version delivered to NCCS
// author       : jglassy
// notes        : Reminder: session log facility is now in l4c_log.c, as xLogMsg()
// ---------------------------------------------------------------

#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"

// ------------- Session Control Structure (new) --------------
static SesCtl          sesCtl ;
// QcStateLimitAdt master, private instance
static QcStateLimitAdt qaStateLimits ;

// L4C_ISO_MetaElements master, private instance
// added 2014-10-06T16:27:00
static L4C_ISO_MetaElements l4cISO_MetaMaster ;

// <JMG> 2014-11-18, split out paths from now-too-diffuse SesCtl
// into more narrowly focused ADTs
static  SesPkgPathAdt      sesPkgPathAdtMaster    ;
static  SesPkgHdf5IdAdt    sesPkgHdf5IdAdtMaster  ;
static  RunParamAdt        sesPkgRunParamAdtMaster;

// CONT was a global, now accessed via l4c_GetLegacyCONT()
// control_struct legacyCONT;
AppControlAdt  legacyCONT ;  // Global program control structure

// ---------------------------------------------------------------------
// function : l4c_GetExecutablVersionId()
// purpose  : Retrieve public ptr to executables datetime stamped version ID
// inputs   : void
// returns  : char *, static (long-lived) string containing executables version id
// notes    : Implemented across all components as a macro {L4_C_|L4_C_MODIS_|L4C_MET_}
char *l4c_GetExecutableVersionId(void)
{
 static char execSwVersion[128]= {0} ;

    // technically, this assignment only needs to be done once
    strcpy(execSwVersion,(char *)L4C_VERS_TAG);

        // return pointer to long-lived version identifier
    return execSwVersion ;

} // end:: l4c_GetExecutableVersionId()


// ---------------------------------------------------------------------
// function : l4c_GetPathAdtOverseer()
// purpose  : return public pointer to private, master sesPkgHdf5IdAdt container
// inputs   : void
// revision : 2014-11-18T10:39:00
// author   : joe glassy
// returns  : SesPkgHdfIdAd *p --public pointer to private
//            master SesPkgHdf5IdAdt instance
// notes    : n/a
SesPkgHdf5IdAdt *l4c_GetHdf5IdOverseer(void)
{

    return &sesPkgHdf5IdAdtMaster ;

} // end:: l4c_GetHdf5IdOverseer()


// ---------------------------------------------------------------------
// function : l4c_GetPathAdtOverseer()
// purpose  : return public pointer to private, master SesPathAdt "paths" container
// inputs   : void
// returns  : SesPkgPathAdt *p --public pointer to private master SesPathAdt instance
// revision : 2014-11-18T10:39:00
// author   : joe glassy
// notes    : Replaces all session path references previously in SesCtl, to
//            migrate towards more narrowly focused ADTs instead of overally
//            broad SesCtl.
SesPkgPathAdt *l4c_GetPathAdtOverseer(void)
{
    // return public pointer to private, master SesPathAdt "paths" container
    return &sesPkgPathAdtMaster;
} // end:: l4c_GetPathAdtOverseer()


// ---------------------------------------------------------------
// l4c_GetLegacyCONT()
// purpose : return a pointer to container of additional ISO metadata
//           elements, e.g. DomainConsistency, CompletenessOmission
// revision: 2013-10-17T15:40:00, added as per Joe A request
// history : n/a
// note    : TBD, method to set values from pixel level QA
L4C_ISO_MetaElements *l4c_GetISO_MetaElements(void)
{

  return &l4cISO_MetaMaster ;

} // end:: l4c_Get_ISO_MetaElements()


// ---------------------------------------------------------------
// l4c_GetLegacyCONT()
// purpose: return a pointer to legacy control
// revision: 2013-11-04T17:48:00
// history : 2013-10-17T15:40:00, original refactoring
// note    : soon to deprecate
AppControlAdt *l4c_GetLegacyCONT(void)
{
 return &legacyCONT;
} // end:: l4c_GetLegacyCONT()


// ---------------------------------------------------------------
// l4c_GetSesCtl()
// purpose: return a pointer to the master (private) session control structure
// revision: 2013-11-04T14:55
// history : 2013-10-17T15:40:00, original refactoring
// note    : in progress...
SesCtl *l4c_GetSesCtl(void)
{
 return &sesCtl  ;

} // end:: l4c_GetSesCtl()


// ---------------------------------------------------------------------
// function : l4c_GetQcStateLimits()
// Purpose  : provides public access to master QcStateLimitAdt instance
// Inputs   : void
// returns  : Q
// notes    : TBD, wrap in a mutex lock to make thread safe
QcStateLimitAdt *l4c_GetQcStateLimits(void)
{
  return &qaStateLimits ;

} // end::l4c_GetQcStateLimits()


// ---------------------------------------------------------------
// function : l4c_GetRunParam()
// purpose  : return a pointer to the master (private) RunParamAdt
// revision : 2014-07-23T17:02:00, <JMG> added.
// history  : n/a
// note     : n/a
RunParamAdt *l4c_GetRunParam(void)
{
 return &sesCtl.runParamObj ;

} // end:: l4c_GetRunParam()



// ---------------------------------------------------------------
// l4c_GetNsbOverseer()
// purpose: return a pointer to the master NSB name-value-pair
//          in-memory database manager (overseer)
// revision: 2014-09-12T15:03, added basic integrity check
// history : 2014-05-19T12:59
// history : n/a
// note    : n/a
NsbObj *l4c_GetNsbOverseer(void)
{
 NsbObj *result_p = NULL;

    result_p = sesCtl.sesNsbDb_p ;
    // perform first minimal integrity check prior to return
    assert(result_p != NULL) ;

 // return pointer to
 return result_p ;

} // end:: l4c_GetNsbOverseer()


// ---------------------------------------------------------------
// l4c_GetPkgMgrOverseer()
// purpose: return a pointer to the master PkgMgr
//          scoping manager (overseer)
// revision: 2014-05-19T12:59
// history : n/a
// note    : n/a
PkgMgr *l4c_GetPkgMgrOverseer(void)
{
 PkgMgr *result_p = NULL;

    // set to  point to master PkgMgr instance
    result_p = sesCtl.l4cPkgMgr_p ;
    // perform first minimal integrity verification prior to return..
    assert(result_p != NULL);

 // return pointer to master PkgMgr
 return result_p;

} // end:: l4c_GetPkgMgrOverseer()


// ---------------------------------------------------------------
// l4c_GetSesControls()
// purpose : Enforce a "gateway" authentication enabled wrapper for
//           l4c_GetSesCtl() until all calls can be upgraded to
//           l4c_GetSesControls(). This is like a BOUNCER at a bar,
//           -- if he doesn't know you, or if your name is NOT on the
//           list of cool people, sorry you DON'T get in!
//
// inputs  : const int requesterCategory, const int requesterId
//
// revision: 2014-02-25T14:45, enable checks on WHO called this,
//           so we can verify if 'keys to the kingdom' should
//           given out. This is especially for multi-threading
//           (thread-safety) to enable smart MUTEXT locking,unlocking
// history : 2013-10-17T15:40:00, original refactoring
// note    : This stub logic is NOT yet fully implemented.
// note    : Wrapper in progress, will enable authentication enforcement
//           soon via requesterId, so depending which ID is submitted,
//           we can return a specific sub-instance. This is for thread
//           safe ops later.
SesCtl *l4c_GetSesControl(const int requesterCategory, const int requesterId)
{
 int tbd_Cat = requesterCategory;
 int tbd_Id  = requesterId ;

 // just a wrapper for now, until all calls to l4c_GetSesCtl() are converted
 // to authentication enabled calls to l4c_GetSesControls()
 xLogMsg("@D DIAG(%s) requesterCat (%d) requeserId(%d)~\n",tbd_Cat,tbd_Id);

 return l4c_GetSesCtl();

} // end:: l4c_GetSesControls()

// ---------------------------------------------------------------
// function   : int l4c_SesVerifyExistance(void)
// Purpose    : Pre-qualifies all critical input files, verifying they
//             actual exist where indicated
// Inputs     : void
// Revised    : 2014-11-11T14:04 added more specific context to error msg
// History    : 2014-08-08T11:34:00, added new.
// Returns    : int, status
// Notes      :  n/a
// Author     : joe glassy
int l4c_SesVerifyExistance(void)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();
 char fxName[256];
 char examplePath[FILENAME_MAX];
 int chkCount = 0 ;

    strcpy(fxName,__func__);

    if( !frm_ioDoesFileExist(sesCtl_p->inputBplutPath))
    {
        xLogMsg("@F In(%s) Mandatory BPLUT (%s) does not yet exist where stated! ~\n",fxName,
            sesCtl_p->inputBplutPath);
    }

    if( !frm_ioDoesFileExist(sesCtl_p->inputAncPath))
    {
        xLogMsg("@F In(%s) Mandatory ANC (%s) does not yet exist where stated! ~\n",fxName,
            sesCtl_p->inputAncPath);
    }

    if( !frm_ioDoesFileExist(sesCtl_p->inputSOCPath ))
    {
        xLogMsg("@F In(%s) Mandatory SOC (%s) does not yet exist where stated! ~\n",fxName,
            sesCtl_p->inputSOCPath);
    }

    if( !frm_ioDoesFileExist(sesCtl_p->inputMetPath))
    {
        xLogMsg("@F In(%s) Mandatory MET (%s) does not yet exist where stated! ~\n",fxName,
            sesCtl_p->inputMetPath);
    }

    // <JMG> 2014-11-11T13:45, added condition for when L4C_GPP_SOURCE = CLIM to allow
    // Ops to omit or skip specification of inputFparPath
    if( strstr(sesCtl_p->gppMethodStr,"CLIM")!= NULL && !frm_ioDoesFileExist(sesCtl_p->inputFparPath))
    {
        // Issue a soft advisory reminding user that omitting inputFparPath IS allowed since
        // we're forcing the use of FparClimatology instead
        xLogMsg("@A In(%s) Allowing optional omission of FPAR 8DAY(%s) since FparClimatology use is forced ~\n",fxName,
            sesCtl_p->inputFparPath);
    }

    // However, when gppMethodStr= {FPAR|NDVI} indicated, if inputFparPath does not exist, this triggers
    // a fatal error of omission...
    if( strstr(sesCtl_p->gppMethodStr,"CLIM") == NULL && !frm_ioDoesFileExist(sesCtl_p->inputFparPath))
    {
            xLogMsg("@F In(%s) Mandatory (FPAR|NDVI) [%s] does not yet exist where stated! ~\n",fxName,
                sesCtl_p->inputFparPath);
    } // end:: block where L4C_GPP_SOURCE = CLIM only option was NOT specified

    // In all cases, we must specify FparClimatology regardless...
    if( !frm_ioDoesFileExist(sesCtl_p->inputFparClimatologyPath))
    {
        xLogMsg("@F In(%s) Mandatory FPAR CLIMATOLOGY (%s) does not yet exist where stated! ~\n",fxName,
            sesCtl_p->inputFparClimatologyPath);
    }

    // <KAE> 2024-01-03 If the litterfall allocation scheme is to follow a schedule...
    if (strstr(sesCtl_p->litterfallScheme, "SCHEDULED") != NULL)
    {
      // Check to see if (at least) the first allocation schedule (for first 8-day period) exists
      snprintf(examplePath, FILENAME_MAX, sesCtl_p->litterfallFileTemplate, 1);
      if (!frm_ioDoesFileExist(examplePath))
      {
        xLogMsg("@F In(%s) Mandatory LITTERFALL FILE TEMPLATE (%s) for first 8-day period does not yet exist where stated! ~\n",
          fxName, examplePath);
      }
    }

    l4c_EventCompletionMsg(fxName);

    return chkCount;
} // end:: l4c_SesVerifyExistance()


// ---------------------------------------------------------------
// function   : l4c_SesInitAncillary()
// Purpose    : Opens, reads ancillary data from SMAP_L4_C_Ancillary.h5
//              to include lat,lon coord var layers and "MCD12Q1_M01land_lc_dom_uint8"
//              and related globally needed ancillary data.
// Inputs     : void
// Invocation : ./l4c_DAILY.ex <./l4c_DAILY.rip>
// Revised    : 2014-05-16T14:47, enabled alloc and read of PixFilter
//              "L4C_M01land_PixFilter_uint8"
// Returns    : int, status. Is >= 0 if ok, or -1 if failed.
// Notes      :  n/a
// Author     : joe glassy
int l4c_SesInitAncillary(void)
{
 SesCtl         *sesCtl_p = &sesCtl ;
 AncImagePkgAdt *ancImg_p = l4c_GetAncImagePkgAdt();

 // set  pointer to PkgMgr (overseer) for session
 PkgMgr   *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
 PkgTheme *pThme_p = NULL;
 char timeBuffer[32];
 char fxName[256];
 int status = 0 ;


    strcpy(fxName,(char *)__func__);

    xLogMsg("\n");
    xLogMsg("@A In(%s) started at %s\n",fxName,frm_utilGetCtime(timeBuffer,0));

    ancImg_p->nM01land_vecpix = EGv2_LDOM1KM;   // count of pixels in M01land domain
    ancImg_p->nM03land_vecpix = EGv2_LDOM3KM ;
    ancImg_p->nM09land_vecpix = EGv2_LDOM9KM ;

    // fetch the Theme holding all Ancillary references
    if((pThme_p = pkgMgr_GetThemeByName(pkgMgr_p,"SMAP_L4_C_Ancillary" ))==NULL)
    {
        xLogMsg("@F ~l4c_SesInitAncillary: Theme(%s) is NOT YET available!\n",
            "SMAP_L4_C_Ancillary");
    } else {
        xLogMsg("@A ~l4c_SesInitAncillary: Theme(%s) found OK with %d PkgObj\n",
            "SMAP_L4_C_Ancillary", pThme_p->pkThemeObjCount);
    }

    // ALLOCATE for MCD12Q1_M01land_lc_dom_uint8, EGv2_LDOM1KM   134787240
    ancImg_p->imgPftM01land_dom_p = memCalloc(EGv2_LDOM1KM, sizeof(uint8));
    memObjSetName(ancImg_p->imgPftM01land_dom_p, "MCD12Q1_M01land_lc_dom_uint8");

    // Bind (register to index) populated object "MCD12Q1_M01land_lc_dom_uint8" for
    // later access by other callers.Re-verify which THEME "MCD12Q1_M01land_lc_dom_uint8" is in
    // then set pThme_p pointer to it...

    pThme_p = pkgMgr_GetThemeByName(pkgMgr_p,"SMAP_L4_C_Ancillary");
    // Alloc and READ, Pixel filter, from DATASET "L4C_M01land_PixFilter_uint8"
    ancImg_p->imgM01land_PixFilter_p = memCalloc(EGv2_LDOM1KM, sizeof(uint8));
    memObjSetName(ancImg_p->imgM01land_PixFilter_p,"ancImg_p->imgM01land_PixFilter_p");

    memObjSetName(ancImg_p->imgM01land_PixFilter_p, "L4C_M01land_PixFilter_uint8");

    if((status= pkgMgr_BindObjByName(pThme_p, "L4C_M01land_PixFilter_uint8",
                ancImg_p->imgM01land_PixFilter_p))< 0)
    {
        xLogMsg("@E ERROR l4c_SesInitAncillary: PkgObj(%s) not available!~\n",
            "L4C_M01land_PixFilter_uint8");
    }

    if((status = pkgMgr_BindObjByName(pThme_p, "MCD12Q1_M01land_lc_dom_uint8",
                ancImg_p->imgPftM01land_dom_p )) < 0)
    {
        xLogMsg("@E ERROR l4c_SesInitAncillary: PkgObj(%s) not available!~\n",
            "MCD12Q1_M01land_lc_dom_uint8");
        return -1 ;
    }

    xLogMsg("@A ~l4c_SesInitAncillary: Reading PFT 1km classifier from (%s)\n",
        sesCtl_p->inputAncPath) ;

    // uses prev opened SMAP_L4_C_Ancillary.h5 handle

    // READ the PFT 1km M01land image in
    status = frm_hdf5_ReadData(sesCtl_p->anc_file_id, H5T_NATIVE_UINT8,"MCD12Q1_M01land_lc_dom_uint8",
                ancImg_p->imgPftM01land_dom_p, "l4c_SesInitAncillary");

    xLogMsg("@A In(%s) Reading Pixel Filter dataset: %s from (%s)~ \n",
        "L4C_M01land_PixFilter_uint8", fxName,sesCtl_p->inputAncPath) ;

    status = frm_hdf5_ReadData(sesCtl_p->anc_file_id, H5T_NATIVE_UINT8,"L4C_M01land_PixFilter_uint8",
                ancImg_p->imgM01land_PixFilter_p, "l4c_SesInitAncillary");

    l4c_EventCompletionMsg(fxName);

    xLogMsg("@A In(%s) Completed for DOY %d ~ \n",fxName,sesCtl_p->runDayOfYear);

 // placeholder return
 status = sesCtl_p->runDayOfYear ;

 return status;

} // end:: l4c_SesInitAncillary()


// ---------------------------------------------------------------------
// function   : l4c_SesInitPrePostReusableImages()
// Purpose    : Session (once-only) init and allocation of "re-usable"
//              and special-case arrays. Done here to avoid need to
//              do over each day, to support (next) multi-day mode.
// Inputs     : void
// Returns    : int status
// revision   : 2014-11-03T18:48, not yet fully enabled. do this next..
// Author     : joe glassy, Lupine Logic Inc.
// Notes      : Implemented fx but not yet enabled or called.
int l4c_SesInitPrePostReusableImages(void)
{
  int status =0 ;
  // fetch container of "re-usable" and special-case images
  PrePostImgPkgAdt *prePostImgAdt_p = l4c_GetPrePostImagePkgAdt();

  // Establish sizes (n. elem) of each array in the "re-usable" container
  prePostImgAdt_p->nElemPreProcM01         = EGv2_GLOBAL_1KM_PIXELS;
  prePostImgAdt_p->nElemPost_StateOutput2D = EGv2_GLOBAL_9KM_PIXELS;
  prePostImgAdt_p->nElemFrozenArea2D       = EGv2_GLOBAL_9KM_PIXELS;
  prePostImgAdt_p->nElemVPD_M09land        = EGv2_LDOM9KM ;

  // Allocate each pre-and-post working "re-usable" image
  prePostImgAdt_p->preProc_M01_p       =(float32 *)memCalloc(prePostImgAdt_p->nElemPreProcM01,sizeof(float32));
  prePostImgAdt_p->postProc_stateOutput2D_p= (float32 *)memCalloc(prePostImgAdt_p->nElemPost_StateOutput2D,sizeof(float32));
  prePostImgAdt_p->vpd_M09land_p       =(float32 *)memCalloc(prePostImgAdt_p->nElemVPD_M09land,sizeof(float32));
  prePostImgAdt_p->frozenArea_M09_2D_p =(float32 *)memCalloc(prePostImgAdt_p->nElemFrozenArea2D,sizeof(float32));

  return status;

} // end:: l4c_SesInitPrePostReusableImages()


// ---------------------------------------------------------------------
// function   : l4c_SesReleasePrePostReusableImages()
// Purpose    : Session (once-only) release (free)  "re-usable"
//              and special-case arrays.
// Inputs     : void
// Returns    : int status
// Notes      :  n/a
// Author     : joe glassy, Lupine Logic Inc.
int l4c_SesReleasePrePostReusableImages(void)
{
  int status =0;

  // fetch container of "re-usable" and special-case images
  PrePostImgPkgAdt *prePostImgAdt_p = l4c_GetPrePostImagePkgAdt();

  memFree( prePostImgAdt_p->preProc_M01_p );
  memFree( prePostImgAdt_p->postProc_stateOutput2D_p);
  memFree( prePostImgAdt_p->vpd_M09land_p);
  memFree( prePostImgAdt_p->frozenArea_M09_2D_p);

  return status ;

} // end:: l4c_SesReleasePrePostReusableImages()


// ---------------------------------------------------------------
// function   : l4c_FunctionCompleteMsg()
// Purpose    : Announces a function 'thisFunc_p' has
//              completed, with a snapshot of current time,
// Inputs     : const char *thisFunc_p
// Returns    : double, cumulative sec since session began
// Notes      :  n/a
// Author     : joe glassy
double l4c_EventCompletionMsg(const char *thisFunction_p)
{
 SesCtl *sesCtl_p  = &sesCtl ;
 time_t secNow = time(0);
 char timeStr[32] = {0};
 // snapshot cumulative running time
 double cumulativeSec = difftime(secNow,sesCtl_p->sesStrtSec) ;
 double decMinutes = cumulativeSec / 60.0 ;

    // <JMG> 4/28/2014, only display completion-message if enabled
    // to reduce clutter...
    if(sesCtl_p->enableCompleteMsg == 1)
        xLogMsg("@A <%s> COMPLETED %s %.2f elap sec. [%.2f min]~ \n",
            thisFunction_p,  frm_utilGetCtime(timeStr,0),
            cumulativeSec, decMinutes );

    return cumulativeSec;

} // end:: l4c_EventCompletionMsg()


// ---------------------------------------------------------------
// function   : l4c_EventHighWaterMarkMsg()
// Purpose    : Announces the point at which current
//              increment build (high-water-mark) testing has
//              been completed to.
// Inputs     : const char *thisFunc_p
// Returns    : double, cumulative sec since session began
// Notes      :  n/a
// Author     : joe glassy
double l4c_EventHighWaterMarkMsg(const char *thisFunction_p)
{
 SesCtl *sesCtl_p  = &sesCtl ;
 char timeStr[32] = {0};

 time_t secNow = time(0);
 // snapshot cumulative running time
 double cumulativeSec = difftime(secNow,sesCtl_p->sesStrtSec) ;

    xLogMsg("@A <%s> HIGH-WATER-MARK-COMPLETION %s (%.2f elap sec)~\n",
        thisFunction_p,  frm_utilGetCtime(timeStr,0), cumulativeSec );

    // PERFORM INTENTIONAL, ORDERLY EARLY SHUTDOWN OF SESSION
    l4c_CloseSession();

    // logic will never get to here but this is shown for completeness
    return cumulativeSec ;

} // end:: l4c_EventHighWaterMarkMsg()


// ---------------------------------------------------------------
// function :l4c_Ses_LogLabelTime()
// inputs   : const time_t finalSec, char *finalLogPath
// returns  : char *, ptr to final log path
// revised  : 2015-09-06T17:31, log prefix now SMAP_L4_C_MDL_
// Purpose  : Using initial log path, construct a FINAL datetime stamped
//            log file path that includes the datetime stamp, and return it.
// notes    : since initial log path (l4c_DAILY.log) is fixed, we don't pass
//            it in here explicitly.
char *l4c_Ses_LogLabelTime(const time_t finalSec,char *finalLogPath)
{
 SesCtl *sesCtl_p = &sesCtl ; // <JMG> 2015-08-20T10:10 import added...
 // char timeBuffer[64] = {0} ;
 // buffer for copying log file to final destination
 char logWorkingPath[FILENAME_MAX] = {0};
 char tzQualifier[16] = {0} ; // time zone qualifier
 struct tm workingTime  ;
 int curYear,curMonth,curDayOfMonth, curHour,curMin, curSec;

    // ONLY AFTER working log is closed, datestamp it and copy it to final log location
    //  Function: struct tm * gmtime_r (const time_t *time, struct tm *resultp)
    // using thread-safe GMT time fx, for final labelling of log file currently
    // NOTE: check if this SHOULD be 'UTC' instead of GMT
#ifdef L4_C_USE_GMT
    gmtime_r(&finalSec, &workingTime);
    strcpy(tzQualifier,"UTC");
#else
    localtime_r(&finalSec, &workingTime);
    strcpy(tzQualifier,"LCL"); // local time zone used...
#endif
    // de-reference the struct tm, noting subtle ways the
    // curYear and curMonth are actually defined as units SINCE a ref mark
    curYear       = workingTime.tm_year + 1900 ;
    curMonth      = workingTime.tm_mon  + 1;
    curDayOfMonth = workingTime.tm_mday;
    curHour       = workingTime.tm_hour ;
    curMin        = workingTime.tm_min  ;
    curSec        = workingTime.tm_sec  ;

    // NOTE: as an optional strategy, we could use the BASE NAME of the
    // current days L4_C OUTPUT _mdl_ FILE and append a .log extension to
    // it: sesCtl_p->output_L4C_DAILY_Path  (substitute extension ".h5" with ".log"
    // <JMG> 2014-09-28 log prefix now alined with _MDL_ output...

    // <JMG> 2015-09-0610:19 added actual DAY OF YEAR doy_Str, to result
    // in a form of "SMAP_L4_C_MDL_YDddd_YYYY-MM-ddThh-mm-ss.log" where
    // the Year Day (of day-of-year) index is always 3 digits (left padded with zeros)
    sprintf(logWorkingPath,"SMAP_L4_C_MDL_YD%s_%4.4d-%2.2d-%2.2dT%-2.2d%-2.2d%-2.2d_%s.log",
               sesCtl_p->doy_Str, curYear, curMonth, curDayOfMonth, curHour,curMin,curSec, tzQualifier);

    strcpy(finalLogPath, logWorkingPath);

 return finalLogPath;
} // end:: l4c_Ses_LogLabelTime()


// ---------------------------------------------------------------
// function :l4c_Ses_DatestampLog()
// inputs   : const time_t finalSec, char *finalLogPath
// returns  : char *, ptr to final log path
// revised  : 2015-08-20T10:10, added code to include actual
//            DayOfYear we're actually processing, since a datetime
//            stamp (e.g. PRODUCTION-DATETIME) does not alone
//            allow us to associate THIS log file back to the run it came from.
// revised  : 2014-02-13T09:41:00, patched in frm_utilGetCtime().
// Purpose  : Using initial log path, produce a datetime stamped log
//            file, using l4c_Ses_LogLabelTime(), by copying the current
//            fixed name log to this new datetime stamped log filed
//            as located in the proper user-defined final log destination dir.
// notes    : logWorkingPath is typically like: "SMAP_L4_C_MDL.log"
// notes    : CALL-LOGIC: this function calls l4c_Ses_LogLabelTime(), and is called by
//            l4c_CloseSession().
// notes    : The sesCtl_p->logf MUST be flushed and closed prior to calling this.
int l4c_Ses_DatestampLog(SesCtl *sesCtl_p)
{
 char logWorkingPath[FILENAME_MAX]  = {0} ;
 char sysCmdLogCopy[FILENAME_MAX*2] = {0} ;

    strcpy(logWorkingPath, sesCtl_p->logInitialPath);

    // ONLY AFTER working log is closed, datetimestamp it, and then
    // copy it to final filesystem location where the log file resides.
    //  Function: struct tm * gmtime_r (const time_t *time, struct tm *resultp)
    // using thread-safe GMT time fx, for final labelling of log file currently

    l4c_Ses_LogLabelTime(sesCtl_p->sesEndiSec, logWorkingPath);

    // <JMG> 2015-08-10T10:10, add DOY string, and preface with logRootDir here
    sprintf(sesCtl_p->logFinalPath,"%s/%s",sesCtl_p->logRootDir, logWorkingPath);

    // note: could perform a 'mv' here instead of 'cp' but for now, to be safe, do a cp
    sprintf(sysCmdLogCopy,"cp -p %s %s\n" ,sesCtl_p->logInitialPath, sesCtl_p->logFinalPath);

    // Perform system OS (shell) call to copy the final file
    // Note: later transition this to use a silent pipe instead of noisy system()
    system(sysCmdLogCopy);

    fprintf(stdout,"\n@A FINAL-LOG(%s)~\n",sesCtl_p->logFinalPath);

    return 0;

} // end:: l4c_Ses_DatestampLog()


// ---------------------------------------------------------------
// function   : l4c_l4c_InitSession()
// Purpose    : Initializes session for a given task executable invocation.
//                Actions include: opens the session logfile, logs initial
//                information, initializes the local memory manager/garbage collector,
//                snapshots starting time of session in seconds
// Inputs     : int argc,char *argv[],SesCtl *sesCtl_p
// Invocation : ./l4c_DAILY.ex <./l4c_DAILY.rip>
// Revision   : 2016-03-02T17:12 <JMG>  added diagnostic output of all BPLUT info read
//              sent to log file with a @D tag l4c_BPLUT_Report(stdout, nBplutRecords);
// Revision   : 2015-07-29T13:41:00, added support for new GEOS5 derived FT 9km
//              added, new default, as: sesCtl_p->ftState = DoUseFT_GEOS5 ;
// Revision   : 2014-06-24T12:50:00, moved UUID generation into here as
//              a one-time-only operation performed.
// Returns    : time_t --starting time in seconds.
// Notes      :  n/a
// Author     : joe glassy
//
time_t l4c_InitSession(int argc,char *argv[])
{
 // FETCH PTR TO (module level static) session controls
 SesCtl *sesCtl_p = &sesCtl ;
 NsbObj *nsbMgr_p = NULL; // changed to nsbMgr_p
 PkgMgr *pkgMgr_p = NULL; // ptr to master PkgMgr overseer instance
 PkgTheme *pThme_p =NULL; // ptr to a given PkgMgr theme
 PkgObj    *pObj_p= NULL; // ptr to a given PkgTheme's child PkgObj

 FILE *ripf = NULL; // handle for runtime input parameter (RIP) text file
 char timeBuffer[64]        = {0} ;
 char versTagBuffer[64] = {0} ;
 char fxName[256] ={0};
 char ripPath[FILENAME_MAX] = {0}; // path name for runtime input parameter (RIP) text file from command line arg 1
 char taskUUID[64] = {0}; // buffer storing latest UUID generated, one time only
 char nvpBuf[FILENAME_MAX] = {0} ;
 char enabledState[32] = {0}; // "is" or "" depending on if a switch is enabled

 char localDatetimeStr[64] = {0};
 char utcDatetimeStr[64] = {0} ;

 int nBplutRecords = 0; // verification BPLUT record check count returned by l4c_BPLUT_Read()
 int p =0;

 pid_t thisPID = 0 ;
 pid_t sesParentPID = 0 ;

    strcpy(fxName,(char *)__func__);

    if(argc <= 1)
    {
        fprintf(stderr,"@F ~In(%s) FAILED ON MISSING runtime input parameters (RIP) file!~\n",
                    fxName);
        exit(-1);
    } else {
        strcpy(ripPath,argv[1]);
        // store for provenence
        strcpy(sesCtl_p->inputRipPath, ripPath);
    }

    if( !frm_ioDoesFileExist(ripPath) )
    {
        xLogMsg("@F In(%s) Mandatory file(%s) does not yet exist where stated!~\n",fxName,
            ripPath);
    }

    // Snapshot start of session
    sesCtl_p->sesStrtSec  = time(0L);
    sesCtl_p->sesEndiSec  = time(0L);

    // ----------------------------------------------------------------
    // SET HARD DEFAULTs in session object. Each may be overriden in RIP

    // <JMG> 5/16/2014, 9km pixel geospatial step interval is now
    //  parameterized, to allow for systematic skip-sampling. Works in
    //  association with the pixel level task filter raster for custom
    //  job definitions.
    sesCtl_p->runPixStepInterval = 1 ;
    sesCtl_p->compressLev        = 6 ; // default, overridden by RIP param
    // Default: try to use FPAR first
    // all options: enum {DoUseFpar|DoUseNDVI|DoUseFparClim}
    sesCtl_p->fparState = DoUseFpar ; // other choice: DoUseNDVI

    // force an appropriate, explicit assignment of  sesCtl_p->runDayOfYear
    sesCtl_p->runDayOfYear = -1;
    sesCtl_p->runToStopDay = -1; // same as runDayOfYear for GMAO Ops
    sesCtl_p->runDayStepInterval=1; // always set to 1 for GMAP Ops
    // <JMG> 2015-07-29 added, new default. Also consider new DoUseFT_Passive.
    // DoUseFT_ActiveRadar, DoUseFT_ActiveTsurf, DoUseFT_PassiveRadiom,DoUseFT_PassiveTsurf, DoUseFT_GEOS5, DoUse_NO_FT
    sesCtl_p->ftState = DoUseFT_GEOS5 ;

    sesCtl_p->runCompositePeriod = -1;

    sesCtl_p->enableCompleteMsg  = 0 ;
    sesCtl_p->enableGranuleMetrics = 0 ;
    // <JMG> 2014-09-27 added (boolean) option to disable SOC updates
    // for diag or devel mode. For GMAO Ops use, set to 1
    sesCtl_p->enableSOCUpdate = 1 ;
    // <JMG> 2016-01-07, now default enabling RMSE , set 1
    sesCtl_p->enableRMSE         = 1 ;
    // <JMG> 2016-03-03T17:19, now enable write observed min,max,mean
    // as a dataset attribute by default. RIP equiv not yet supported
    // but will be L4C_ENABLE_WRITE_OBS_MINMAX = {YES|TRUE, or NO|FALSE}
    sesCtl_p->enableWriteObsMinMax =1;

    // fetch version numbers.
    strcpy(sesCtl_p->swVersion    ,l4c_GetExecutableVersionId());
    strcpy(sesCtl_p->frmApiVersion, frm_GetApiVersion(nvpBuf));
    strcpy(sesCtl_p->pkgApiVersion, pkgMgr_GetVersion(nvpBuf));
    strcpy(sesCtl_p->nsbApiVersion, nsbGetVersion(nvpBuf));

    // Initialize the local memory manager and garbage collector
    memObjInit(stdout);

    // register function to call atexit..
    atexit(memObjGarbageCollect);

    // NOTE: Critical to FIRST create the RunParamAdt instance inside SesCtl
    // <JMG> 3/13/2014, convert runParamObj to a scalar, streamline
    // sesCtl_p->runParamObj ;

    // Open session (initial) log with fixed name, will likely close,
    // rename as end of session in order to datetime-stamp, etc.
    strcpy(sesCtl_p->logInitialPath, "SMAP_L4_C_MDL.log");

    // <JMG> 2015-10-14, NEW usage to migrate to. Uses the internal
    // default session log file handle. Not for parallel processing.
    xLogOpenfile(sesCtl_p->logInitialPath);

    // Once (internal default) log file is open, redirect memObj's log to this one
    memObjRegisterLog(xLogGetFilehandle());

    // Capture name of user controlling terminal session. This will
    // fail if process is daemonized, but l4c_DAILY.ex is not normally
    // run as a daemon
    getlogin_r(sesCtl_p->thisUsername,128);
    // Capture name of host we're running on
    memset(sesCtl_p->thisHost,'\0',256);
    gethostname(sesCtl_p->thisHost,255) ;

    // Capture PID and any other diagnostic info
    thisPID = frm_utilGetPid(&sesParentPID);

    // <JMG> 2014-06-24 Assign unique task ID via UUID one time only
     // UUID displayMode 0 generates sys (libuuid.a) "packed" UUID identifier
    // UUID displayMode 1 generates sys (libuuid.a) a hyphenated UUID identifer
    // UUID displayMode 2 generates internal pseudo "packed" UUID identifer
    // UUID displayMode 3 generates internal (pseudo) a hyphenated UUID identifer
    frm_utilGenerateUUID(taskUUID,1);
    strcpy(sesCtl_p->l4c_Task_UUID, taskUUID);

    // Disable HDF5 default error stack handler. Can enable
    // situationally later if needed. Stores 0 when off, 1 when on.
    sesCtl_p->hdf5ErrStackState = frm_hdf5_ErrorStackDisable() ;

    // ----------------------------------------------------------------
    // post start up messages
    xLogMsg("\n~");
    xLogMsg("=====================================================================~\n");
    xLogMsg("@A L4_C TASK started %s at pid %zu  with parents-pid %zu ~\n",
            frm_utilGetCtime(timeBuffer, sesCtl_p->sesStrtSec),
            (size_t)thisPID,(size_t) sesParentPID);

    l4c_GetISO8601_datetime(sesCtl_p->sesStrtSec,localDatetimeStr,utcDatetimeStr);
    xLogMsg("@A TASK-START UTC(%s) LocalTimezone(%s) \n",utcDatetimeStr,localDatetimeStr);

    xLogMsg("@A L4_C running on host [%s] controlled by user [%s] ~\n",
        sesCtl_p->thisHost, sesCtl_p->thisUsername);

    xLogMsg("@A University of Montana UM Framework API version: %s~\n",
            frm_GetApiVersion(versTagBuffer));

    xLogMsg("@A size of SesCtl is: %d ~\n", (int)sizeof(SesCtl));

    // ----------------------------------------------------------------
    // SET UP and CREATE NSB (NVP manager) and PkgMgr (scoping indexer) facilities
    // for session to manage all name-value-pair info from RIP
    sesCtl_p->sesNsbDb_p = nsbCreateInstance("L4_C_NSB_Manager");

    // ----------------------------------------------------------------
    // CREATE PkgMgr with K themes. Ideally call this AFTER collecting all RIP nvps
    sesCtl_p->nPkgThemes = L4C_PKGMGR_THEME_COUNT ;

    // ----------------------------------------------------------------
    // OPEN and PARSE the main RIP file, retrieve ALL runtime inputs
    if((ripf=fopen(ripPath,"rt"))==NULL)
    {
        xLogMsg("@F In(%s) missing required RUNTIME INPUTS (RIP) FILE (%s)~\n",
            fxName, ripPath);

    } else {

    // --------------- PARSE RIP FILE ----------------------------------
        // Fetch set of name-value-pairs from the RIP file, store in a NSB stringblock
        //   database, where values are retrieved using access function
        //   nsbGetValueByKey(nsb_p,ky,valueStr)...

        // Note: the "" for delimiter string indicates to use the defaults of "\t= "
        // so a '=' is one of those supported here.
        sesCtl_p->sesNvpCount = nsbParseIntoStringblock(sesCtl_p->sesNsbDb_p,"",ripf);

        xLogMsg("@A l4c_SesInit: %d name-value-pairs stored from RIP (%s)~\n",
                sesCtl_p->sesNvpCount, ripPath);

        // Call this only if we need to verify that all keywords have been stored etc.
        // nsbExportDbToText(sesCtl_p->sesNsbDb_p,"L4C_RIP_KeyVerification.txt",' ');

        // CLOSE the runtime input parameters text file
        fclose(ripf);
    } // end:: block if RIP file available and parsed


    // Create all PkgMgr themes, now that RIP NVPs are available
    // Ideally call this AFTER collecting all RIP nvps
    xLogMsg("@A ~L4_C PkgMgr to be initialized with %d themes\n",sesCtl_p->nPkgThemes);

    l4c_PkgMgr_Init( sesCtl_p->nPkgThemes );
    pkgMgr_p = l4c_GetPkgMgrOverseer();

    // verify NVPs all collected, DIAG only, test was passed
    // l4c_SesVerifyNVPs(sesCtl_p->sesNsbDb_p) ;


    // COLLECT all name-value-pairs (NVPs) using expected keywords
    nsbMgr_p = l4c_SesCollectCoreNvps(sesCtl_p); // first get standard set of CORE nvps

    // Next, get application specific name-value-pairs (nvps), catching the
    // populated nsbMgr_p in case we need to direct-fetch any NVP value via its key
    // NOTE: nsbMgr_p here is simply a transient local pointer to sesCtl_p->sesNsb_p.
    nsbMgr_p = l4c_SesCollectNVPs() ;

    // SPECIAL CASE: explicitly set this seasonally specific dataset name into the
    // theme for FparClimatology. Reference by index (4) here is dangerous
    // since if this index changes this will be wrong. TBD: improve this
    pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "PRE_PROC_DEFLATE_1KM" );
    pkgMgr_SetThemes_ChildName(pThme_p, 2, sesCtl_p->fparClimDatasetname);

    //SPECIAL HANDLING FOR current periods FPAR climatology dataset
    // needed -- Assign the late-bound Fpar climatology dataset name
    // from RIP file into the primary theme "SMAP_L4_C_FparClimatology"
    pThme_p = pkgMgr_GetThemeByName(pkgMgr_p, "SMAP_L4_C_FparClimatology");
    pkgMgr_SetThemes_ChildName(pThme_p, 0, sesCtl_p->fparClimDatasetname);

    // full-circle test confirmation, TBD --test return value of pObj_p
    pObj_p  = pkgMgr_GetObjByName(pThme_p,sesCtl_p->fparClimDatasetname);

    // pre-qualify existance of all key input files
    l4c_SesVerifyExistance() ;

    // Note: swRelease is written OUT as file level attrib "L4_C_SW_DATETIME"
    xLogMsg("\n@A Executable %s version ID %s ~ \n",argv[0],sesCtl_p->swRelease);
    xLogMsg("@A API: UM Framework version    (%s)~ \n", sesCtl_p->frmApiVersion);
    xLogMsg("@A API: PkgMgr       version    (%s)~ \n", sesCtl_p->pkgApiVersion);
    xLogMsg("@A API: NSB          version    (%s)~ \n", sesCtl_p->nsbApiVersion);

    // POST initial session messages
    xLogMsg("@A In(%s) appAction             (%s)~\n", fxName, sesCtl_p->appAction);
    xLogMsg("@A L4_C Dataset version         (%s)~\n", sesCtl_p->datasetVersion);
    xLogMsg("@A L4_C Subsystem Release tag   (%s)~\n", sesCtl_p->SubsysReleaseTag);

    xLogMsg("@A L4_C RIP registered platform (%s)~\n", sesCtl_p->l4c_Platform);
    xLogMsg("@A L4_C RIP parameter count      %d retrieved from (%s)~\n",
                sesCtl_p->sesNvpCount, ripPath);

    xLogMsg("@A Production level             (%s)~\n", sesCtl_p->productionLevel);
    xLogMsg("@A Production task-run tag      (%s)~\n", sesCtl_p->l4c_Task_Label);
    xLogMsg("@A Production task-run UUID     (%s)~\n", sesCtl_p->l4c_Task_UUID );

    xLogMsg("@A L4_C session log             (%s)~\n", sesCtl_p->logInitialPath);
    xLogMsg("@A Designated log rootdir       (%s)~\n", sesCtl_p->logRootDir);

    xLogMsg("@A L4_C Mode (action)               [%s]~ \n" ,sesCtl_p->appAction);
    xLogMsg("@A L4_C Model Year (%d)             [%s]~ \n" ,sesCtl_p->runModelYear,sesCtl_p->year_Str);
    xLogMsg("@A L4_C Model Composite Period      [%d] ~ \n",sesCtl_p->runCompositePeriod);
    xLogMsg("@A L4_C Model Day of Year (%d)      [%s]~ \n" ,sesCtl_p->runDayOfYear,sesCtl_p->doy_Str);

    xLogMsg("@A L4_C Model GPP Method            [%d] %s ~ \n",sesCtl_p->fparState,sesCtl_p->gppMethodStr);
    xLogMsg("@A L4_C Model Freeze Thaw Source    [%d] %s ~ \n",sesCtl_p->ftState, sesCtl_p->ftMethodString);

    xLogMsg("@A L4_C Model SOC calendar mode     [%d]~ \n",sesCtl_p->socOpMode);
    xLogMsg("@A L4_C Compression level           [%d]~ \n",sesCtl_p->compressLev);
    // output step interval just for local documentation
    xLogMsg("@A L4_C Model Spatial step interval [%zu]~ \n",sesCtl_p->runPixStepInterval);

    xLogMsg("@A L4_C Dataset count L4C_HDF5_TOTAL_DATASETS [%d] ~\n",
      (int)L4C_HDF5_TOTAL_DATASETS);

    strcpy(enabledState,"is NOT");
    if(sesCtl_p->enableRMSE) strcpy(enabledState,"is");
    xLogMsg("@A In(%s) L4_C QA RMSE generation %s enabled ~ \n",fxName,enabledState);

    strcpy(enabledState,"are NOT");
    if(sesCtl_p->enableSOCUpdate) strcpy(enabledState,"are");
    xLogMsg("@A In(%s) L4_C SOC updates %s enabled ~ \n",fxName,enabledState);

    // Define ALL relevant DATASET properties for OUTPUT L4C HDF5
    // datasets, including: datatype,rank,fillvalue, etc and be sure
    // to do this BEFORE CreateOutput and writes are called.
    l4c_PrepOutput_HDF5_Datasets() ;

    // <JMG> 3/11/2014, open all primary INPUT HDF5 files once, so that if any
    //   fail to be opened, we find out VERY EARLY in the process
    // Note: must open ancillary file prior to l4c_Hdf5_CreateOutput() call since
    //   creating output requires access to Ancillary (lat,lon) vars...
    sesCtl_p->anc_file_id  = frm_hdf5_Openfile( sesCtl_p->inputAncPath ,fxName);
    sesCtl_p->met_file_id  = frm_hdf5_Openfile( sesCtl_p->inputMetPath ,fxName);

    if(strstr(sesCtl_p->gppMethodStr,"CLIM")!= NULL)
    {
        xLogMsg("@W In(%s) GPP_METHOD(%s) forced-use -- omitting opening of FPAR 8DAY file ~ \n",
            fxName, sesCtl_p->gppMethodStr);

    } else {
        sesCtl_p->fpar_file_id = frm_hdf5_Openfile( sesCtl_p->inputFparPath,fxName);

        // <KAE> 2021-11-01 Read the "Source" attribute of the input fPAR/ L4CMOD granule;
        //    this allows for "VIIRS" or "MODIS" to be specified in output carbon_model_bitflag;
        frm_hdf5_ReadAttributeString(sesCtl_p->fpar_file_id, "Source", sesCtl_p->fparSourceStr);

        // based on sesCtl_p->gppMethodStr, verify correct dataset is in place
        if(strcmp(sesCtl_p->gppMethodStr,"FPAR")==0)
        {
            if( frm_hdf5_DatasetExists(sesCtl_p->fpar_file_id,"fpar_M01") != 1 )
            {
                xLogMsg("@W In(%s) GPP_METHOD(%s) cannot find required (%s) dataset ~ \n",
                    fxName, sesCtl_p->gppMethodStr, sesCtl_p->soc_group_id);
            }
        } else {

            xLogMsg("@A In(%s) GPP_METHOD(%s) -- using NDVI or other ~ \n",
                fxName, sesCtl_p->gppMethodStr);

            if( frm_hdf5_DatasetExists(sesCtl_p->fpar_file_id,"ndvi_M01") != 1 )
            {
                xLogMsg("@E In(%s) GPP_METHOD(%s) cannot find required (%s) dataset ~ \n",
                    fxName, sesCtl_p->gppMethodStr, sesCtl_p->soc_group_id);
            }
        }
    } // end:: block if L4C_GPP_SOURCE = FPAR or NDVI

    // <JMG> added 4/22/2014, re-enable FparClimatology processing
    sesCtl_p->fparClima_file_id = frm_hdf5_Openfile( sesCtl_p->inputFparClimatologyPath,fxName);

    // <JMG> added 9/27/2014, if enableSOCUpdate, open in append mode, otherwise
    // open in read only mode.  TBD: investigate use of "CORE" in-memory driver
    // for SOC to put-off actual disk I/O and improve performance.
    if(sesCtl_p->enableSOCUpdate)
        sesCtl_p->soc_file_id  = frm_hdf5_OpenfileAppend(sesCtl_p->inputSOCPath, fxName);
    else
        sesCtl_p->soc_file_id = frm_hdf5_Openfile(sesCtl_p->inputSOCPath, fxName);

    // SPECIAL HANDLING required for nested groups, save specific SOC group id
    // for when we need to test if a fully qualified SOC dataset exists

    sesCtl_p->soc_group_id = H5Gopen(sesCtl_p->soc_file_id,"/SOC",H5P_DEFAULT);
    if(sesCtl_p->soc_group_id < 0)
    {
          xLogMsg("@E In(%s) L4_C FAIL to get SOC group id  [%d]~ \n",
                        fxName,sesCtl_p->soc_group_id);
    } else {
          xLogMsg("@A In(%s) L4_C SOC group id [%d]~ \n",fxName, sesCtl_p->soc_group_id);
    }

    // CREATE EMPTY output file(s), SMAP_L4_C_xxx.h5, writes latitude,longitude coord vars,
    // and creates all GROUPS, and creates all initially empty
    // DATASET definitions, and, in addition to creating empty datasets,
    // <JMG> 2014-0-24, added writes of all metadata GROUPS
    // and formal ATTRIBUTE values
    l4c_Hdf5_CreateOutput();

    // Since files are now open, we can now safely set
    // all INPUT HDF5 FILE level properties, in their respective themes
    l4c_PkgInit_SetFileProperties();

    // <JMG> 2017-06-05 added "soft verification" of BPLUT UUID and MD5 if present
    // via display in log file only. Future: may make a mis-match trip a fatal error.
    if(nsbDoesKeyExist(nsbMgr_p,"L4C_BPLUT_UUID"))
        xLogMsg("@A In (%s) BPLUT UUID (%s)~\n", fxName,sesCtl_p->l4c_BPLUT_UUID);

    if(nsbDoesKeyExist(nsbMgr_p,"L4C_BPLUT_MD5"))
        xLogMsg("@A In (%s) BPLUT  MD5 (%s)~\n",fxName,sesCtl_p->l4c_BPLUT_MD5);

    // <JMG> 2017-06-05 changed call interface slightly to pass in
    //  sesCtl_p, to facilitate runtime BPLUT version instance check via a
    //  match verification on BPLUT s internal UUID and or MD5 as needed
    //   BPLUT *L4C_Read_BPLUT(const char *bpLutPath,int *nBplutRecords)
    l4c_BPLUT_Read(sesCtl_p, &nBplutRecords);

    // <JMG> 2016-03-02T17:12 added diagnostic output of all BPLUT info read
    //       sent to log file with a @D tag
    // TESTED OK: l4c_BPLUT_Report(stdout, nBplutRecords);

    // Open and read key 2D rasters from SMAP_L4_C_Ancillary
    l4c_SesInitAncillary();

#ifdef DIAG_ONLY
    // DIAG BLOCK testing RMSE additions
    pThme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 8);
    // xLogMsg("@D In(%s) L4C_OUTPUTS PkgObj CHILD COUNT %d ~\n",fxName, pkgMgr_GetThemeObjectCount(pThme_p));

    for(p=0; p < pkgMgr_GetThemeObjectCount(pThme_p) ; p++)
    {
        xLogMsg("@D L4C_OUTPUTS %2d %s ~\n",p, pkgMgr_GetThemes_ChildName(pThme_p,p));
    }
    // <JMG> 2016-01-07 diag only
    // l4c_SesCtl_Display(sesCtl_p,stdout) ;
#endif

    xLogMsg("@A In(%s) initialized for DOY %d YEAR %d~ \n\n",fxName,
                 sesCtl_p->runDayOfYear, sesCtl_p->runModelYear);

    l4c_EventCompletionMsg(fxName);

    // IF FOR DIAG TESTING, perform early shutdown,  l4c_CloseSession() would be called here

 // return start time in unix seconds
 return sesCtl_p->sesStrtSec;

} // end :: l4c_SesInit()


// ---------------------------------------------------------------
// function :l4c_CloseSession()
// inputs   : SesCtl *sesCtl_p
// revised  : 2014-02-13T09:41:00, patched in frm_utilGetCtime().
// Purpose  : Closes session for a given task executable invocation. Stops the runtime clock, logs timing, and
//            program exit information and closes out the logfile.
// Notes    : if a @F fatal error trapped, l4c_SesClose() is preemptively called..
//          : tested OK to work on stout,wheat
void l4c_CloseSession(void)
{
 SesCtl         *sesCtl_p = &sesCtl ; // ptr to session controls, direct attach
 AncImagePkgAdt *ancImagePkgObj_p = l4c_GetAncImagePkgAdt();
 Hdf5GridPropAdt *hGridProp_p = l4c_GetHdf5GridProp();

 // <JMG> 2014-08-21, check counts, changed name 2015-10-15...
 PostInvalid     *pInvalidCounts_p = l4c_GetInvalidEventCounts();

 char timeBuffer[64] = {0} ;
 char logWorkingPath[FILENAME_MAX]= {0} ;
 char fxName[256] = {0} ;
 size_t highestSlot= 0;
 size_t totalBytesTracked=0;
 size_t hwmMem =0;
 herr_t ret = -1 ;


    strcpy(fxName,__func__);

    xLogMsg("~\n");
    xLogMsg("@A In(%s) Commencing orderly sesson shutdown~ \n",fxName);

    xLogMsg("@A In(%s) Selected Exceptions on nee_mean and gpp~ \n",fxName);
    xLogMsg("@A In(%s) GPP mean issues %zu ~ \n",fxName,pInvalidCounts_p->invalidGpp);
    xLogMsg("@A In(%s) GPP mean Min    %zu ~ \n",fxName,pInvalidCounts_p->invalidGppMin);
    xLogMsg("@A In(%s) GPP mean Max    %zu ~ \n",fxName,pInvalidCounts_p->invalidGppMax);
    xLogMsg("@A In(%s) NEE mean issues %zu ~ \n",fxName,pInvalidCounts_p->invalidNee);
    xLogMsg("@A In(%s) NEE mean Min    %zu ~ \n",fxName,pInvalidCounts_p->invalidNeeMin);
    xLogMsg("@A In(%s) NEE mean Max    %zu ~ \n",fxName,pInvalidCounts_p->invalidNeeMax);
    // added 2015-10-15T14:04, added to check SOC/soc_mean
    xLogMsg("@A In(%s) SOC mean issues %zu ~ \n",fxName,pInvalidCounts_p->invalidSocMean);

    // added 2014-11-11, for NaN, and Inf tracking
    xLogMsg("@A In(%s) NaN occurrences suppressed %zu ~ \n",fxName,pInvalidCounts_p->invalidHitNaN);
    xLogMsg("@A In(%s) Inf occurrences suppressed %zu ~ \n",fxName,pInvalidCounts_p->invalidHitInf);

    xLogMsg("@A In(%s) closing all HDF5 files and datasets~\n",fxName);

    // close out the dcpls, and dspaces prior to final file close
    ret = H5Pclose(hGridProp_p->dcpl_1km);
    ret = H5Pclose(hGridProp_p->dcpl_3km);
    ret = H5Pclose(hGridProp_p->dcpl_9km);

    // QA, SPECIAL CASE: need a dcpl for UINT8 for 255 fillvalue etc.
    ret = H5Pclose(hGridProp_p->dcpl_9km_uint8);
    ret = H5Pclose(hGridProp_p->dcpl_9km_uint16);

    ret = H5Sclose(hGridProp_p->dspace_1km);
    ret = H5Sclose(hGridProp_p->dspace_3km);
    ret = H5Sclose(hGridProp_p->dspace_9km);

    // close the Ancillary file
    ret = H5Fclose(sesCtl_p->anc_file_id);
    // close the MET file
    ret = H5Fclose(sesCtl_p->met_file_id);

    // <JMG> 2014-11-11T14:19:00, Explicitly close the FPAR 8DAY file,
    //   ONLY if L4C_GPP_SOURCE was NOT "CLIM" noting that, in worse
    //   case, if this close call is omitted, the "read-only file will
    //   typically still be closed via exit of process (verify this TBD)
    //   and note, an improvement in a future frm_hdf5_xxx wrapper
    //   for H5Fclose() should first verify that 'file_id' was valid
    //   before attempting to close it...
    if( strstr(sesCtl_p->gppMethodStr,"CLIM")==NULL)
        ret = H5Fclose(sesCtl_p->fpar_file_id);

    // <JMG> 2014-08-08, close the special nested SOC group id
    // and moved this 'close' ahead of file's close 9/4/2014...
    ret = H5Gclose(sesCtl_p->soc_group_id);
    // close the SOC intermediate file
    ret = H5Fclose(sesCtl_p->soc_file_id);

    // FINALLY, close out the main "_mdl_" .h5 output file
    ret = H5Fclose(sesCtl_p->l4cOutput_file_id);  // HDF5 file_id of SMAP_L4_C_OUTPUT.h5

    //  redirect memObj's log to this one
    memObjRegisterLog(stdout);

    xLogMsg("@A \n");
        memObjGetMetrics(&highestSlot,&totalBytesTracked,&hwmMem);
    xLogMsg("@A In(%s): session memory resource: high-slot %zu total-bytes %zu high-water-mark %zu (%.1f Gb)~\n",
            fxName, highestSlot, totalBytesTracked, hwmMem, ((float)hwmMem/1024000000.0) );

    // Verify possible bug in DNA viewer before enabling!
    // memObjReportDna(sesCtl_p->logf,"l4c_CloseSession: memory DNA report");
    // memObjReportDna(stdout, "l4c_CloseSession: memory DNA report");
    // fflush(sesCtl_p->logf);

    // Route detailed memory report to log file
    // memObjReport(stdout, "l4c_CloseSession: l4c_DAILY memory usage -- detail report");
    memObjReport(sesCtl_p->logf, "l4c_CloseSession: l4c_DAILY memory usage -- detail report");
    fflush(sesCtl_p->logf);

    // release other arrays not yet free'd whose names we DO know
    xLogMsg("@A In(%s): releasing other residual arrays~ \n",fxName);
    memObjFree(sesCtl_p->fparClimDatasetname,NULL);
    memObjFree("T2M_M09_MAX",NULL);
    memObjFree("PREPROC_1KM_BUFFER",NULL);

    if(sesCtl_p->fparState == DoUseFpar)
    {
        memObjFree("fpar_M01",NULL);
        memObjFree("fpar_qc_M01",NULL);
    } else if(sesCtl_p->fparState == DoUseNDVI) {
        // TBD : verify these names are right...
        memObjFree("ndvi_M01"   ,NULL);
        memObjFree("ndvi_qc_M01",NULL);
    }

    // free row,col LUT memory...
    memFree(ancImagePkgObj_p->imgM09land_row_uint16_p);
    memFree(ancImagePkgObj_p->imgM09land_col_uint16_p);

    // <JMG> 2014-05-16, free the main ancImg_p->imgPftM01land_dom_p PFT classifier image
    memFree(ancImagePkgObj_p->imgPftM01land_dom_p);
    // <JMG> 2014-05-16, free the 1D sparse vector PixFilter
    memFree(ancImagePkgObj_p->imgM01land_PixFilter_p);

    // FREE memory for the BPLUT lut, by name. The NULL here is the optional pointer
    // if one is known and available
    memObjFree("L4_C_BPLUT", NULL);

    // RELEASE all memory (3 parts) owned by this NsbObj *nsb_p instance
    xLogMsg("@A In(%s) closing out NSB (nvp-stringblock database)~ \n", fxName);
    nsbReleaseInstance(sesCtl_p->sesNsbDb_p);

    // RELEASE all PkgMgr memory
    xLogMsg("@A In(%s) closing out PkgMgr indexing facility~ \n",fxName);
    pkgMgr_ReleaseMgr(sesCtl_p->l4cPkgMgr_p);

    // Trigger final final garbage collection, to release all memory not otherwise
    //  explicitly free'd
    memObjGarbageCollect();

    // LAST, close and flush the log file (keep this as the LAST thing we
    //  do prior to exit) . TBD: extend xCloseLog as xCloseLog(logf,DestPath)
    //  to include the date-stamping and copy to destpath inside xCloseLog()
    // snapshot session level ending time and elapsed times in sec

    sesCtl_p->sesEndiSec = time(0L);
    // snapshot session level ending time and elapsed times in sec
    sesCtl_p->sesElapSec = difftime(sesCtl_p->sesEndiSec,sesCtl_p->sesStrtSec);

    xLogMsg("@A In(%s) task [%s] version [%s]~\n",
        fxName, (char *)L4C_APP_TAG,(char *)L4C_VERS_TAG  );
    xLogMsg("@A In(%s): completed at [%s] requiring %.2lf sec (%.2f min)~\n",
        fxName, frm_utilGetCtime(timeBuffer,sesCtl_p->sesEndiSec),
          sesCtl_p->sesElapSec,  ((float32)sesCtl_p->sesElapSec/60.0) );

    xLogMsg("@A In(%s) closing session log~ \n",fxName);

    // Flush and Close (internal default) file assoc with
    // path sesCtl_p->logInitialPath
    // NEW USAGE to migrate to
    xCloseLog(xGetLogfileHandle());

    // Finally, copy the (flushed,closed) log file to a datetime
    // stamped instance of it in the final log location
    l4c_Ses_DatestampLog(sesCtl_p) ;

} // end :: l4c_CloseSession()
