
// ----------------------------------------------------
/*
 module       : // l4c_DAILY_spinup_logic.c
 purpose      : Driver functions for performing model simulations and intialization.  These functions manage
                spatial and temporal looping and tracking of relevant model states in space and time.  Model simulation
		step logic invoked in the innermost loops.  Dynamic inputs loaded at various looping levels depending on
		spatial and temporal resolution.
 project      : SMAP L4C science development.
 dependencies : tcfModFunc.c, tcfMem.c
 revision     : 2014-02-15t13:08:00,jmg, for lead fork , ADT namespace updates
 revision     : Initial reformating -- laj 09/10/13
 author       : Lucas A. Jones; lucas@ntsg.umt.edu
 notes        : n/a
*/

// tcf_inc.h and many other headers have now been integrated into l4c_DAILY.h
// #include "tcf_inc.h"

#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"


// ----------------------------------------------------
// tcfModDriverGrid_anaPrecond()

// Purpose: Drives analytical preconditioning for numerical spinup SOC initialization. 
//
// Notes:  First, computes annual sums of kmult and NPP, then uses these sums in the analytical steady-state 
//         equation.  This provides an approximate 'annual average' SOC steady-state, that then must be 
//         numerically spun-up to the precise steady state with associate seasonal cycle.  Analytical 
//         preconditioning significantly reduces the number of spin-up iterations required to reach steady-state, 
//         particularly for the SOC pool with the slowest decay dynamics.
//
// Inputs: Dynamic daily met data (MetArrayObj), static pft map (PFT_ARRAY), and static model parameters 
//         (GLOB_PARAMS, BPLUT).    
//
// Outputs: Preconditioned C pools 1km grid arrays returned in the STATE_ARRAY structure. Annual NPP sum for 
//          litterfall calculations also returned.

void tcfModDriverGrid_anaPrecond(PftAdt PFT_ARRAY, MetArrayAdt MetArrayObj, EcoStateArrayAdt *STATE_ARRAY)
{
  SesCtl *sesCtl_p = l4c_GetSesCtl();
  
  uint8     pft_index;
  uint32       doy;
  uint32    t, t8day;
  uint32    rn3km, cn3km, rn1km, cn1km;
  uint32    s3km_nested9km, s1km_nested3km, s1km_nested9km;
  uint32    s1km, s9km;
  MetAdt    MET; /*for local, inner-most-loop state calculations, not requiring large grid arrays*/
  EcoStateAdt  STATE; /*for local, inner-most-loop state calculations, not requiring large grid arrays*/
  BPLUT         *bp_p = NULL;
  int       status;
  int       nDaysToProcess = NDAYS;
 
	//  Begin time loop through NDAYS to process
	t8day = 0;
  
	for (t=0; t < nDaysToProcess; t++)
	{
    
      doy = t + 1;
    
      xLogMsg( "@A Task %s: day %d ...\n", CONT.taskIdStr, doy);

      // Load Daily Met files
      status = tcfMem_loadDailyMet(doy, &t8day, &MetArrayObj);

      // Begin 9km land vector loop
      for (s9km=0; s9km < sesCtl_p->nElem9kmPerTask; s9km++)
	  {
	
		// Unpack input arrays for step call interface
		MET.tmin = (float64) MetArrayObj.tmin[s9km]; /* [K] */
		MET.vpd = (float64) MetArrayObj.vpd[s9km];   /* [Pa] */
		MET.par = (float64) MetArrayObj.par[s9km] * RAD2PAR;   /* convert to [MJ m-2] par */

		// Rescale root zone soil moisture
		MET.smrz = tcfModMult_rescalSmrz(MetArrayObj.smrz[s9km], MetArrayObj.smrz_mn[s9km], MetArrayObj.smrz_mx[s9km]);
	
		MET.tsoil = (float64) MetArrayObj.tsoil[s9km]; /* [K]*/
		MET.smtop = (float64) MetArrayObj.smtop[s9km] * 100.0; /*convert from [dim] to [%]*/

		// Begin nested 3km grid loop
		for (rn3km=0; rn3km < NESTED_3KM_IN_9KM; rn3km++)
		{
			for (cn3km=0; cn3km < NESTED_3KM_IN_9KM; cn3km++)
			{
				//  Unpack input 3-km FT arrays for step call interface
				//  serial array index for 3km grid
				s3km_nested9km = M_2D_B0(rn3km, cn3km, NESTED_3KM_IN_9KM); //range {0..9}
				
				//  Begin nested 1km 2D grid loops
				for (rn1km=0; rn1km < NESTED_1KM_IN_3KM; rn1km++)
				{
					for (cn1km=0; cn1km < NESTED_1KM_IN_3KM; cn1km++)
					{
	      
						// Serial array index for 1km grid
						s1km_nested3km = M_2D_B0(rn1km, cn1km, NESTED_1KM_IN_3KM); //range {0..9}
						s1km_nested9km = M_2D_B0(s3km_nested9km, s1km_nested3km, NESTED_1KM_IN_9KM); //range {0..81}
						s1km = (uint32) M_2D_B0(s9km, s1km_nested9km, NESTED_1KM_IN_9KM_SQ); //range {0..LDOM1KM}
		
						/*Unpack pft*/
						pft_index = PFT_ARRAY.dom_1km[s1km];

						//Retrieve bplut look-up parameters based on pft
						bp_p = L4C_GetBPLUT_ByIndex(pft_index, NPFT);
	
						/*Unpack fpar data*/
						MET.fpar = (float64) MetArrayObj.fpar[s1km] * GLOB_PARAMS.fparGain + GLOB_PARAMS.fparOffset;

						/*NPP sum must always be initialized for states to be available*/
						STATE.lit = (float64) STATE_ARRAY->npp_sum.dom_1km[s1km] / NDAYS;

						/*Check if cell is in 1-km processing domain (only do veg pfts with non-missing input)*/
						if (tcfModFunc_isInCell(pft_index, MET, STATE) != 0)
						{

							/*Get gpp and kmult*/
							tcfModFunc_gpp(bp_p, MET, &STATE);
							
							tcfModFunc_kmult(bp_p, MET, &STATE);
	  
							/*Must do tracking on the grid*/
							/*(because the time loop is outside of the grid loops)*/
							STATE_ARRAY->npp_sum.dom_1km[s1km] += (float32) STATE.npp;
							STATE_ARRAY->kmult_sum.dom_1km[s1km] += (float32) STATE.rh_mult.k;

						} else  {

							STATE_ARRAY->npp_sum.dom_1km[s1km] = FILVAL_FLT;
                            STATE_ARRAY->kmult_sum.dom_1km[s1km] = FILVAL_FLT;

						} // end :: 1km in-domain check 
			
					} // end :: 1km COLUMN nested 3km loop 
					
				} // end :: 1km ROW nested 3km loop 
		
			} // end :: 3km COLUMN nested 9km loop 
		  
		} // end :: 3km ROW nested 9km loop 
	
	  } // end :: 9km land vector loop
      
	} // end :: time loop
  
  	//  DO ANALYTICAL CALCS, Now that the time loop is complete.
	//  Must loop back over the grid domain
	tcfModDriverGrid_anaPrecond_stst(PFT_ARRAY, MetArrayObj, STATE_ARRAY);  
  
  // done

} // end :: tcfModDriverGrid_anaPrecond()


 
// -------------------------------------------------------------------------------------------------------------
// tcfModDriverGrid_anaPrecond_stst()

// Purpose: Does the spatial loop for analytical steady-state once kmult and npp sums have been calculated.
//
// Notes: The '_stst()' string in the function name indicates that this procedure is wrapped by a higher-level
//        calling function i.e., '_anaPrecond()'.  
//
// Inputs: Kmult and NPP annual sums (STATE_ARRAY), static pft map (PFT_ARRAY), static model parameters 
//         (GLOB_PARAMS, BPLUT).
//
// Outputs: Initialized C pool 2D 1km grid arrays returned in the STATE_ARRAY structure.


void tcfModDriverGrid_anaPrecond_stst(PftAdt PFT_ARRAY, MetArrayAdt MetArrayObj, EcoStateArrayAdt *STATE_ARRAY)
{
  int8      pft_index;

  uint32    l;
  uint32    rn3km, cn3km, rn1km, cn1km;
  uint32    s3km_nested9km, s1km_nested3km, s1km_nested9km;
  uint32    s1km, s9km;
 
  MetAdt    MET; /*for local, inner-most-loop state calculations, not requiring large grid arrays*/
  EcoStateAdt  STATE; /*for local, inner-most-loop state calculations, not requiring large 2D grid arrays*/
  BPLUT                  *bp_p = NULL;


  // Begin 9km 2D grid loops 
  for (s9km=0; s9km < sesCtl_p->nElem9kmPerTask; s9km++)
    {

      /*Unpack input arrays for step call interface*/
      //NOTE: tsoil required for domain check because GMAO land domain definition differs from ours
      MET.tsoil = (float64) MetArrayObj.tsoil[s9km]; /* [K]*/
    
      /*Begin nested 3km grid loop*/
      for (rn3km=0; rn3km < NESTED_3KM_IN_9KM; rn3km++)
	{
	  for (cn3km=0; cn3km < NESTED_3KM_IN_9KM; cn3km++)
	    {

	      // serial array index for 3km grid
	      s3km_nested9km = M_2D_B0(rn3km, cn3km, NESTED_3KM_IN_9KM); //range {0..9}
	
	      // Begin nested 1km 2D grid loops
	      for (rn1km=0; rn1km<NESTED_1KM_IN_3KM; rn1km++)
		{
		  for (cn1km=0; cn1km<NESTED_1KM_IN_3KM; cn1km++)
		    {

		      /*serial array index for 1km grid*/
		      s1km_nested3km = M_2D_B0(rn1km, cn1km, NESTED_1KM_IN_3KM); //range {0..9}
		      s1km_nested9km = M_2D_B0(s3km_nested9km, s1km_nested3km, NESTED_1KM_IN_9KM); //range {0..81}
		      s1km = (uint32) M_2D_B0(s9km, s1km_nested9km, NESTED_1KM_IN_9KM_SQ); //range {0..LDOM1KM}

		      //Initialize state array fill values
		      for (l=0; l<NPOOL; l++)
			STATE_ARRAY->C[l].dom_1km[s1km] = FILVAL_FLT;		      


		      /*Unpack pft*/
		      pft_index = PFT_ARRAY.dom_1km[s1km];
		      
		      //Retrieve bplut look-up parameters based on pft		      
		      bp_p = L4C_GetBPLUT_ByIndex(pft_index, NPFT);

		      /*This must always be initialized for fill values*/
		      STATE.lit = (float64) STATE_ARRAY->npp_sum.dom_1km[s1km] / NDAYS;

		      /*Check if cell is in 1-km processing domain (only do veg pfts with non-missing input)*/
		      if (tcfModFunc_isInCell(pft_index, MET, STATE) != 0)
			{
	  
			  /*Unpack summed arrays*/
			  STATE.npp_sum = (float64) STATE_ARRAY->npp_sum.dom_1km[s1km];
			  STATE.kmult_sum = (float64) STATE_ARRAY->kmult_sum.dom_1km[s1km];

			  /*Analytical calc*/
			  tcfModFunc_anaStst(bp_p, &STATE);
			  
			  /*Package SOC for next timestep*/
			  for (l=0; l<NPOOL; l++)
			    STATE_ARRAY->C[l].dom_1km[s1km] = (float32) STATE.C[l];
			  
			} // end :: 1km in-domain check
		    } // end :: 1km row nested 3km loop
		} // end :: 1km col nested 3km loop
	      
	    } // end :: 3km row nested 9km loop
	} // end :: 3km col nested 9km loop
      
    } // end :: 9km land vector loop
  
  
} // end :: tcfModDriverGrid_anaPrecond_stst()                                                                   


 
// -------------------------------------------------------------------------------------------------------------
// tcfModDriverGrid_numSpin()

// Purpose: Drives numerical spin-up SOC intialization. 

// Notes: Each iteration runs once through the input climatology and iterations stop when all grid cells have
//        met a pre-specified tolerance (abs(NEE annual sum)) or max number of iterations. Tolerances are tracked  
//        for all grid cells and overall maximum tolerance is determined.  Once a grid cell meets the tolerance
//        threshold, the cell is skipped on subsequent iterations.  Iterations terminate if the overall maximum 
//        tolerance for the grid mets the tolerance threshold, OR if the pre-specified max number of iterations 
//        has been reached. For multiple, parallel executable task invocations, the tolerance/max-iterations are 
//        determined ONLY for the portion of the array processed by the individual task (this scheme should
//        produce the same overall results regardless of the number of individual tasks).  A table of iteration 
//        information is produced in the session log-file, with rows prefaced by the '#T' character string. 

//        Spin-up can proceed from any initial SOC state (including zeros), but is much more efficient when 
//        pre-conditioned with an approximate analytical steady-state solution. This is especially true for 
//        pools with slow dynamics which require many iterations to reach steady-state if initialized from zero.  

// Inputs: Dynamic daily met data (MetArrayObj), static pft map (PFT_ARRAY), and static models parameters 
//         (GLOB_PARAMS, BPLUT).
//
// Outputs: Initialized C pool 1km grid arrays returned in the STATE_ARRAY structure.
//          Also returns spinup tolerance and iteration count arrays as 1km grids.


void tcfModDriverGrid_numSpin(PftAdt PFT_ARRAY, MetArrayAdt MetArrayObj, EcoStateArrayAdt *STATE_ARRAY)
{
  SesCtl *sesCtl_p = l4c_GetSesCtl();
  uint32    doyStop;
  uint32    cells_left_cnt;
  uint32    spinup_glob_iter_cnt;

  float32   current_max_tol = 0.0;


  /*Initialize global spin-up iteration count, should mirror max_tol_iter_cnt*/
  spinup_glob_iter_cnt = 0;
  cells_left_cnt = sesCtl_p->nElem1kmPerTask;

  //Spin-up iteration table logfile headers (#T token denotes table fields)
  xLogMsg( "@A #T Task %s:  BEGIN   MaxIter %d  StopTol %f\n",
	       CONT.taskIdStr, GLOB_PARAMS.spinup_max_iter, GLOB_PARAMS.spinup_stop_tol);
 
  xLogMsg( "@A #T %-15s%-15s%-15s%-15s%-15s%-15s\n",
	       "Iter", "MaxIter", "MaxTol", "StopTol", "CellsRemain", "TotCells");

  xLogMsg( " @A #T ------------------------------------------------------------------------------------\n");


  /*Begin spin-up loop*/
  /*Check global spin-up tolerance: continue iterating if any grid cells have not met stopping criteria*/

  /*Ensure at least one iteration, check tolerances in subsequent iterations*/
  while (cells_left_cnt > 0 && spinup_glob_iter_cnt < GLOB_PARAMS.spinup_max_iter)
    {

      /*Do model run through entire year of climatology*/
      doyStop = (uint32) NDAYS;
      
      tcfModDriverGrid_forwardSpin(doyStop, spinup_glob_iter_cnt, PFT_ARRAY, MetArrayObj, STATE_ARRAY);
      
      /*Loop back through space and check tolerance criteria, find global max tolerance to
	determine if additional global spin-up iterations are necessary*/
      cells_left_cnt = tcfModDriverGrid_numSpin_tols(&current_max_tol, PFT_ARRAY, MetArrayObj, STATE_ARRAY);


      xLogMsg( "@A #T %-15d%-15d%-15.4f%-15.4f%-15d%-15d\n",
		   spinup_glob_iter_cnt+1, GLOB_PARAMS.spinup_max_iter, current_max_tol, GLOB_PARAMS.spinup_stop_tol,
		   cells_left_cnt, sesCtl_p->nElem1kmPerTask);
      
      spinup_glob_iter_cnt++;
    
    } // end :: spin-up while loop


  /* Check final spin-up tolerance */
  xLogMsg( "@A #T ------------------------------------------------------------------------------------\n");
  xLogMsg( "@A #T Task %s: FINAL, Iter %d of %d tol thres %f  max tol %f;   %d of %d cells do not meet criteria\n",
	       CONT.taskIdStr, spinup_glob_iter_cnt, GLOB_PARAMS.spinup_max_iter, GLOB_PARAMS.spinup_stop_tol, current_max_tol,
	       cells_left_cnt, sesCtl_p->nElem1kmPerTask);

} // end :: tcfModDriverGrid_numSpin()


// -------------------------------------------------------------------------------------------------------------
// tcfModDriverGrid_forwardSpin()

// Purpose:  Similar to '_forwardRun()' but streamlined with minimal state tracking for numerical spin-up and 
//           maintains the ability to stop at a specified day of the year.
//
// Notes:   Could be combined with '_forwardRun()' in future code versions.
//
// Inputs:  Initial SOC, passed in the STATE_ARRAY structure, dynamic daily met data (MetArrayObj),
//          static pft map (PFT_ARRAY), and static model parameters (GLOB_PARAM, BPLUT).
//
// Outputs: Model state (SOC) propagated through one year passed in STATE_ARRAY structure.


void tcfModDriverGrid_forwardSpin(uint32 doyStop, uint32 spin_iter, PftAdt PFT_ARRAY, MetArrayAdt MetArrayObj, EcoStateArrayAdt *STATE_ARRAY)
{
SesCtl *sesCtl_p = l4c_GetSesCtl();
  int       status;
  uint8     pft_index;

  uint32    doy;
  uint32    t, t8day, l;

  uint32    rn3km, cn3km, rn1km, cn1km;
  uint32    s3km_nested9km, s1km_nested3km, s1km_nested9km;
  uint32    s1km, s9km;

  MetAdt    MET; /*for local, inner-most-loop state calculations, not requiring large 2D grid arrays*/
  EcoStateAdt  STATE; /*for local, inner-most-loop state calculations, not requiring large 2D grid arrays*/

  BPLUT        *bp_p = NULL;


  /*Begin time loop (run up to doyStop, so if doyStop=365, do entire year)*/
  t8day = 0;

  for (t=0; t<doyStop; t++)
    {

      doy = t+1;

      xLogMsg( "@A Task %s: Spin iter %d, day %d ...\n", CONT.taskIdStr, spin_iter, doy);

      //Load daily met data
      status = tcfMem_loadDailyMet(doy, &t8day, &MetArrayObj);
      
      /*Begin 9km land vector loop*/
      for (s9km=0; s9km<sesCtl_p->nElem9kmPerTask; s9km++)
	{

	  /*Unpack input arrays for step call interface*/
	  MET.tmin = (float64) MetArrayObj.tmin[s9km]; /* [K] */
	  MET.vpd = (float64) MetArrayObj.vpd[s9km];   /* [Pa] */
	  MET.par = (float64) MetArrayObj.par[s9km] * RAD2PAR;   /* convert from [W m-2] rad to [MJ m-2] par */

	  // Rescale root zone soil moisture
	  MET.smrz = tcfModMult_rescalSmrz(MetArrayObj.smrz[s9km], MetArrayObj.smrz_mn[s9km], MetArrayObj.smrz_mx[s9km]);
	  
	  MET.tsoil = (float64) MetArrayObj.tsoil[s9km]; /* [K]*/
	  MET.smtop = (float64) MetArrayObj.smtop[s9km] * 100.0; /*convert from [dim] to [%]*/
	  
	  /*Begin nested 3km grid loop*/
	  for (rn3km=0; rn3km<NESTED_3KM_IN_9KM; rn3km++)
	    {
	      for (cn3km=0; cn3km<NESTED_3KM_IN_9KM; cn3km++)
		{

		  /*Unpack input 3-km FT arrays for step call interface*/

		  /*serial array index for 3km grid*/
		  s3km_nested9km = M_2D_B0(rn3km, cn3km, NESTED_3KM_IN_9KM); //range {0..9}
		  
		  /*Begin nested 1km 2D grid loops*/
		  for (rn1km=0; rn1km<NESTED_1KM_IN_3KM; rn1km++)
		    {
		      for (cn1km=0; cn1km<NESTED_1KM_IN_3KM; cn1km++)
			{

			  /*serial array index for 1km grid*/
			  s1km_nested3km = M_2D_B0(rn1km, cn1km, NESTED_1KM_IN_3KM); //range {0..9}
			  s1km_nested9km = M_2D_B0(s3km_nested9km, s1km_nested3km, NESTED_1KM_IN_9KM); //range {0..81}
			  s1km = (uint32) M_2D_B0(s9km, s1km_nested9km, NESTED_1KM_IN_9KM_SQ); //range {0..LDOM1KM}

	  
			  /*Process only cells which have NOT met spin-up CRITERIA*/
			  if (STATE_ARRAY->spinup_stop[s1km] != 1)
 			    {

			      /*Unpack pft*/
			      pft_index = PFT_ARRAY.dom_1km[s1km];
				  
			      //Retrieve bplut look-up parameters based on pft
			      bp_p = L4C_GetBPLUT_ByIndex(pft_index, NPFT);
				  
			      /*Unpack fpar data*/
			      MET.fpar = (float64) MetArrayObj.fpar[s1km] * GLOB_PARAMS.fparGain + GLOB_PARAMS.fparOffset;

			      /*NPP sum must always be initialized for state estimates to be available*/
			      STATE.lit = (float64) STATE_ARRAY->npp_sum.dom_1km[s1km] / NDAYS;
			      
			      /*Check if cell is in 1-km processing domain (only do veg pfts with non-missing input)*/
			      if (tcfModFunc_isInCell(pft_index, MET, STATE) != 0)
				{
				  
				  /*Initialize SOC*/
				  /*SOC from whatever previous state*/
				  /*(from model run or spin-up)*/
				  for (l=0; l<NPOOL; l++)
				    STATE.C[l] = (float64) STATE_ARRAY->C[l].dom_1km[s1km];
				  
				  //If first timestep, always Initialize nee sum to zero 
				  if (t == 0)
				    STATE_ARRAY->nee_sum.dom_1km[s1km] = 0.0;
				  
				  /*Do tcf forward step to propagate SOC states from last timesetp to current timestep*/
				  tcfModFunc_forwardStep(bp_p, MET, &STATE);
				  
				  /*Package SOC for next timestep*/
 				  for (l=0; l<NPOOL; l++)
				    STATE_ARRAY->C[l].dom_1km[s1km] = (float32) STATE.C[l];
				  
				  /*Sum state arrays*/
				  STATE_ARRAY->nee_sum.dom_1km[s1km] += (float32) STATE.nee;
				  
				} // end :: 1km in-domain check 

			      else
				{
				  
				  //Assign state fill values
				  STATE_ARRAY->nee_sum.dom_1km[s1km] = FILVAL_FLT;

 				  for (l=0; l<NPOOL; l++)
				    STATE_ARRAY->C[l].dom_1km[s1km] = FILVAL_FLT;

				} // end :: 1km in-domain check 
			      
			    } // end :: spinup criteria check
			  
			} // end :: 1km row nested 3km loop
		    } // end :: 1km col nested 3km loop
		  
		} // end :: 3km row nested 9km loop
	    } // end :: 3km col nested 9km loop
	  
	} // end :: 9km land vector loop
      
    } // end :: time loop

  
} // end :: tcfModDriverGrid_forwardSpin()


// -------------------------------------------------------------------------------------------------------------
// tcfModDriverGrid_numSpin_tols()

// Purpose:  Does spatial loop for checking individual grid cell numerical spinup tolerances at each iteration 
//           once sums have been calculated.
//
// Notes:  
//
// Inputs:  Static pft map (PFT_ARRAY), and static model parameters (GLOB_PARAMS).
//
// Outputs: Global, overall max tolerance for each iteration and a boolean (0 or 1) array of grid cells        
//          where 1 means the cell has met the tolerance and should subsequently be skipped and 0 means the 
//          cell has not met the tolerance and will be included in the next iteration.

uint32 tcfModDriverGrid_numSpin_tols(float32 *max_tol, PftAdt PFT_ARRAY, MetArrayAdt MetArrayObj, EcoStateArrayAdt *STATE_ARRAY)
{
SesCtl *sesCtl_p = l4c_GetSesCtl();
  uint8     pft_index;

  uint32    rn3km, cn3km, rn1km, cn1km;
  uint32    s3km_nested9km, s1km_nested3km, s1km_nested9km;
  uint32    s1km, s9km;

  uint32    current_cell_iter;
  uint32    cells_left_cnt;
  float32   current_max_tol, current_cell_tol;

  // for local, inner-most-loop state calculations, not requiring large grid arrays
  MetAdt      MET;    
  EcoStateAdt    STATE;

  // Initialize current max tol for the grid
  current_max_tol = 0.0;
  current_cell_tol = 0.0;
  current_cell_iter = 0;
  cells_left_cnt = 0;

  /*Begin 9km land vector loop*/
  for (s9km=0; s9km < sesCtl_p->nElem9kmPerTask; s9km++)
    {

      /*Unpack input arrays for step call interface*/
      //NOTE: tsoil required for domain check because GMAO land domain definition differs from ours
      MET.tsoil = (float64) MetArrayObj.tsoil[s9km]; /* [K]*/
    
      /*Begin nested 3km grid loop*/
      for (rn3km=0; rn3km<NESTED_3KM_IN_9KM; rn3km++)
	{
	  for (cn3km=0; cn3km<NESTED_3KM_IN_9KM; cn3km++)
	    {
	
	      /*serial array index for 3km grid*/
	      s3km_nested9km = M_2D_B0(rn3km, cn3km, NESTED_3KM_IN_9KM); //range {0..9}

	      /*Begin nested 1km 2D grid loops*/
	      for (rn1km=0; rn1km<NESTED_1KM_IN_3KM; rn1km++)
		{
		  for (cn1km=0; cn1km<NESTED_1KM_IN_3KM; cn1km++)
		    {

		      /*serial array index for 1km grid*/
		      s1km_nested3km = M_2D_B0(rn1km, cn1km, NESTED_1KM_IN_3KM); //range {0..9}
		      s1km_nested9km = M_2D_B0(s3km_nested9km, s1km_nested3km, NESTED_1KM_IN_9KM); //range {0..81}
		      s1km = (uint32) M_2D_B0(s9km, s1km_nested9km, NESTED_1KM_IN_9KM_SQ); //range {0..LDOM1KM}
		      
		      /*Retrieve bplut parameters based on PFT*/
		      pft_index = PFT_ARRAY.dom_1km[s1km];

		      /*NPP sum must always be initialized for state estimates to be available*/
		      STATE.lit = (float64) STATE_ARRAY->npp_sum.dom_1km[s1km] / NDAYS;


		      /*Check if cell is in 1-km processing domain (only do veg pfts with non-missing input)*/
		      if (tcfModFunc_isInCell(pft_index, MET, STATE) != 0)
			{

			  /*Check cell's PREVIOUS iteration CRITERIA
			   * if cell's critera HAVE been met, skip the cell.                                      *                             
			   * If NOT, flag cell for another spin-up iteration (leave as 0)                         *
			   * and include in the search for the highest tolerance                                  */
			  if (STATE_ARRAY->spinup_stop[s1km] != 1)
			    {
      
			      /*Find grid maximum current tolerance (do for all veg cells)*/
			      current_cell_tol = fabs(STATE_ARRAY->nee_sum.dom_1km[s1km]);
			      current_cell_iter = STATE_ARRAY->spinup_iter[s1km];
			  
			      /*track the highest current tolerance (and its associated iter) for           *
			       * entire domain to determine if additional spinup iterations are necessary   *
			       * NOTE: the max tol is only complete once the time loop exits and the sums   *
			       * are complete.
			       */
			      if (current_cell_tol > current_max_tol)
				{
				  
				  current_max_tol = current_cell_tol;
				  
				} // end :: highest tolerance check/assignment
 
			      //increment cell's iteration count (all veg cells get at least one spin-up iteration)
			      STATE_ARRAY->spinup_iter[s1km]++;
			      
			      /*Check if cell has met CURRENT iteration CRITERIA, if not set nee_sum=0.0 and 
			        increment cell count. */
			      if (current_cell_tol > GLOB_PARAMS.spinup_stop_tol)
				{
				  
				  cells_left_cnt++; //increment remaining cell count
				  
				}
			      /*If criteria HAVE CURRENTLY been met, flag the cell to skip on subsequent spin-up iterations*/
			      else
				{
				
				  STATE_ARRAY->spinup_stop[s1km] = 1;
			  			    			    
				} // end :: CURRENT iteration CRITERIA check*/
	      
			    } // end :: PREVIOUS iteration CRITERIA check*/

			} // end :: 1km in-domain check 
		      
		    } // end :: 1km row nested 3km loop 
		} // end :: 1km col nested 3km loop 
	      
	    } // end :: 3km row nested 9km loop 
	} // end :: 3km col nested 9km loop 
      
    } // end :: 9km land vector loop
  
  //update max tolerance
  *max_tol = current_max_tol;
  
  //return count of remaining cells
  return(cells_left_cnt);

} // end :: tcfModDriverGrid_numSpin_tols()



// -------------------------------------------------------------------------------------------------------------
// tcfModDriverGrid_forwardRun()

// Purpose:  Drives forward run for one entire simulation year.
//
// Notes: Does not contain full suite of PSD output state fields as in '_forwardStep()'.  Primary purpose of this
//        routine is to generate timeseries for model checking.
//         
// Inputs:  Initial SOC pools and litterfall, passed in the STATE_ARRAY structure, dynamic daily met data 
//          (MetArrayObj), static pft map (PFT_ARRAY), and static model parameters (GLOB_PARAM, BPLUT).
//
// Outputs: Daily states, including final SOC pools, in 1km grid arrays in the STATE_ARRAY structure.
//          

void tcfModDriverGrid_forwardRun(PftAdt PFT_ARRAY, MetArrayAdt MetArrayObj, EcoStateArrayAdt *STATE_ARRAY)
{
SesCtl *sesCtl_p = l4c_GetSesCtl();
  char      doyStr[MXSTRLEN]={0};
  char      *fpath = NULL;

  int       status;
  uint8     pft_index;

  uint32    doy;
  uint32    t, t8day, l;

  uint32    rn3km, cn3km, rn1km, cn1km;  // 3km and 1km nested relative row and column indicies 
  uint32    s3km_nested9km, s1km_nested3km, s1km_nested9km;  // 3km and 1km nested relative serial indicies
  uint32    s1km, s3km, s9km; // 3km and 1km absolute serial indicies into respective global arrays
  uint32    nelem;

  MetAdt    MET; /*for local, inner-most-loop state calculations, not requiring large 2D grid arrays*/
  EcoStateAdt  STATE; /*for local, inner-most-loop state calculations, not requiring large 2D grid arrays*/

  BPLUT                  *bp_p = NULL;



  /*Begin time loop (run up to doyStop, so if doyStop=365, do entire year)*/
  t8day = 0;

  for (t=0; t < NDAYS; t++)
    {

      doy = t+1;

      fprintf(stdout, "%d\n", doy);

      //Load daily met data
      status = tcfMem_loadDailyMet(doy, &t8day, &MetArrayObj);
      
      /*Begin 9km land vector loop*/
      for (s9km=0; s9km<sesCtl_p->nElem9kmPerTask; s9km++)
	{

	  /*Unpack input arrays for step call interface*/
	  MET.tmin = (float64) MetArrayObj.tmin[s9km]; /* [K] */
	  MET.vpd = (float64) MetArrayObj.vpd[s9km];   /* [Pa] */
	  MET.par = (float64) MetArrayObj.par[s9km] * RAD2PAR;   /* convert from [W m-2] rad to [MJ m-2] par */

	  // Rescale root zone soil moisture
	  MET.smrz = tcfModMult_rescalSmrz(MetArrayObj.smrz[s9km], MetArrayObj.smrz_mn[s9km], MetArrayObj.smrz_mx[s9km]);
	  
	  MET.tsoil = (float64) MetArrayObj.tsoil[s9km]; /* [K]*/
	  MET.smtop = (float64) MetArrayObj.smtop[s9km] * 100.0; /*convert from [dim] to [%]*/
	  
	  /*Begin nested 3km grid loop*/
	  for(rn3km=0; rn3km < NESTED_3KM_IN_9KM; rn3km++)
	    {
	      for(cn3km=0; cn3km < NESTED_3KM_IN_9KM; cn3km++)
		{
	  
		  /*Unpack input 3-km FT arrays for step call interface*/
	  
		  /*serial array index for 3km grid*/
		  s3km_nested9km = M_2D_B0(rn3km, cn3km, NESTED_3KM_IN_9KM); //range {0..9}
		  
		  // Absolute serial reference into 3km global array
		  s3km = (uint32) M_2D_B0(s9km, s3km_nested9km, NESTED_3KM_IN_9KM_SQ); //range {0..LDOM3KM}

		  /*Begin nested 1km 2D grid loops*/
		  for(rn1km=0; rn1km < NESTED_1KM_IN_3KM; rn1km++)
		    {
		      for(cn1km=0; cn1km < NESTED_1KM_IN_3KM; cn1km++)
			{
			  
			  // serial array index nesting for 1km grid
			  s1km_nested3km = M_2D_B0(rn1km, cn1km, NESTED_1KM_IN_3KM); //range {0..9}
			  s1km_nested9km = M_2D_B0(s3km_nested9km, s1km_nested3km, NESTED_1KM_IN_9KM); //range {0..81}

			  // Absolute serial reference into 1km global array
			  s1km = (uint32) M_2D_B0(s9km, s1km_nested9km, NESTED_1KM_IN_9KM_SQ); //range {0..LDOM1KM}

			  // Unpack pft
			  pft_index = PFT_ARRAY.dom_1km[s1km];

			  // Retrieve bplut look-up parameters based on pft		
			  bp_p = L4C_GetBPLUT_ByIndex(pft_index, NPFT);

			  // Unpack fpar data
			  MET.fpar = (float64) MetArrayObj.fpar[s1km] * GLOB_PARAMS.fparGain + GLOB_PARAMS.fparOffset;

			  // NPP sum must always be initialized for state estimates to be available
			  STATE.lit = (float64) STATE_ARRAY->npp_sum.dom_1km[s1km] / NDAYS;
		  
		      
		// Check if cell is in 1-km processing domain
		//			  --only do veg pfts with non-missing input
			  if (tcfModFunc_isInCell(pft_index, MET, STATE) != 0)
			    {
			      
			      /*Initialize SOC*/
			      /*SOC from whatever previous state*/
			      /*(from model run or spin-up)*/
			      for(l=0; l < NPOOL; l++)
				STATE.C[l] = (float64) STATE_ARRAY->C[l].dom_1km[s1km];
			      
					// Perform tcf forward step to propagate SOC states
					// from last timesetp to current timestep
					tcfModFunc_forwardStep(bp_p, MET, &STATE);
			      
					// Required: Package SOC for next timestep
					for(l=0; l < NPOOL; l++)
					STATE_ARRAY->C[l].dom_1km[s1km] = (float32) STATE.C[l];
					
					// Sum state arrays/
					STATE_ARRAY->nee_sum.dom_1km[s1km] += (float32) STATE.nee;
					STATE_ARRAY->rhtot_sum.dom_1km[s1km] += (float32) STATE.rhtot;
					STATE_ARRAY->gpp_sum.dom_1km[s1km] += (float32) STATE.gpp;
					
					/*Optional:*/
					STATE_ARRAY->gpp.dom_1km[s1km] = (float32) STATE.gpp;
					STATE_ARRAY->rhtot.dom_1km[s1km] = (float32) STATE.rhtot;
					STATE_ARRAY->nee.dom_1km[s1km] = (float32) STATE.nee;
						// end :: 1km in-domain check
					} else  {
						// State tracking
						for (l=0; l<NPOOL; l++)
							STATE_ARRAY->C[l].dom_1km[s1km] = FILVAL_FLT;
			  
						//Sum state arrays
						STATE_ARRAY->nee_sum.dom_1km[s1km] = FILVAL_FLT;
						STATE_ARRAY->rhtot_sum.dom_1km[s1km] = FILVAL_FLT;
					STATE_ARRAY->gpp_sum.dom_1km[s1km] = FILVAL_FLT;
			      
					/*Optional:*/
					STATE_ARRAY->gpp.dom_1km[s1km] = FILVAL_FLT;
					STATE_ARRAY->rhtot.dom_1km[s1km] = FILVAL_FLT;
					STATE_ARRAY->nee.dom_1km[s1km] = FILVAL_FLT;
				
					} // end :: fill values
			  
				} // end :: 1km COLUMN nested 3km loop 
			
		    } // end :: 1km ROW nested 3km loop 
		  
		} // end :: 3km COLUMN nested 9km loop 
		
	    } // end :: 3km ROW nested 9km loop 
	  
	} // end :: 9km land vector loop
      
      
      //WRITE DAILY STATE OUTPUT
      
      // Prepare doy string
      sprintf(doyStr,"%s%03d", "day", doy); doyStr[strlen(doyStr)] = '\0' ;
      
      nelem = sesCtl_p->nElem1kmPerTask;
      
      /* fpath = fpath_cat(5, GLOB_PARAMS.output_basepath, "tcf_nee_",doyStr ,"_M01land_task",CONT.taskIdStr); */
      /* status = write_flt32(fpath, STATE_ARRAY->nee.dom_1km, nelem); */
      
      /* fpath = fpath_cat(5, GLOB_PARAMS.output_basepath, "tcf_gpp_",doyStr ,"_M01land_task",CONT.taskIdStr); */
      /* status = write_flt32(fpath, STATE_ARRAY->gpp.dom_1km, nelem); */
      
      /* fpath = fpath_cat(5, GLOB_PARAMS.output_basepath, "tcf_rhtot_",doyStr ,"_M01land_task",CONT.taskIdStr); */
      /* status = write_flt32(fpath, STATE_ARRAY->rhtot.dom_1km, nelem); */
      
    }// end :: time loop

	
  //  WRITE ANNUAL STATE OUTPUT
  nelem = sesCtl_p->nElem1kmPerTask;

  fpath = fpath_cat(3, GLOB_PARAMS.output_basepath, "tcf_gpp_sum_M01land_task",CONT.taskIdStr);
  status = write_flt32(fpath, STATE_ARRAY->gpp_sum.dom_1km, nelem);
  
  fpath = fpath_cat(3, GLOB_PARAMS.output_basepath, "tcf_rhtot_sum_M01land_task",CONT.taskIdStr);
  status = write_flt32(fpath, STATE_ARRAY->rhtot_sum.dom_1km, nelem);
  
  fpath = fpath_cat(3, GLOB_PARAMS.output_basepath, "tcf_nee_sum_M01land_task",CONT.taskIdStr);
  status = write_flt32(fpath, STATE_ARRAY->nee_sum.dom_1km, nelem);

} // end :: tcfModDriverGrid_forwardRun()

 
