// --------------------------------------------------------------
// module       : l4c_DAILY_tcf.h (was orig tcf_struct.h)
// purpose      : Contains data structure definitions for 'tcf_spin' program.
// project      : SMAP L4C science development.
// dependencies : HDF4, FRM_API
// revision     : v7.1.x, 2016-02-17, module and header names harmonized to use
// l4c_DAILY_xxx prefix revision     : v7 2015-07-30, <JMG>, changed
// "ft_in_swath_3km" to "ft_in_swath" to emphasize this
//                is no longer a strictly resolution dependent conditional

// revision     : v7 2015-07-29, <JMG> added support for F/T from GEOS5 or
// Passive radiometer L3_SM_P at 3KM revision     : v5.x-2014-11-10T11:07,
// renamed (4-state) metObj.gppMethod to metObj.gppInputUsed
//                to better distiguish it from RIP files (3-state)
//                L4C_GPP_SOURCE
// history      : v5.0.1-2014-08-07T13:41:00, removed surflag_3km[] array, and
// added
//                F/T term to GPP calculation, requiring adding (6) fields to
//                new BPLUT etc. and <JMG> added ft_state_um_M03, as replacement
//                for *surface_flag where ft_state_um_M03 carries (4) bits
//                {ft_tsurf,ft_radar,ft_radar_qc,ft_in_swath}
// history      : v5.0.1-2014-05-16T13:30, added viirsNdvi_qc, and note
//                TBD, planned change: change all fpar_qa to "fpar_qc" to match
//                actual legacy name in MODIS HDF4, to avoid endless confusion
//                between "fpar_qa" and "fpar_qc" etc
// history      : v5.0.1-2014-04-16, Lucas added fpar_clim, with other
//                minor changes possibility still pending(?)
// history      : v4.3.0-2014-02-27T15:44:00,MILESTONE
// history      : v3.x-2014-02-24T15:57, joe added MET.h5's qv2m,t2m_avg,ps_avg
// to consolidate. history      : v3.x-2014-02-23T20:02:00, Joe added missing
// *ftRollingQA_3km; history      : v3.2-2014-02-15T13:23:00 (ADT namespace pass
// 1 completed) author       : joe glassy , Lupine Logic Inc. notes        : The
// "lead_fork" configuration of tcf_spin_v3.2 implements new
//                ADT namespaces, consolidates all C and .h headers in one
//                directory and serves as the middle (transitional) codebase
//                betweeen the ORIGINAL code and the final L4_DAILY production
//                codebase.
//
// notes        : NAMESPACE CONVENTIONS summary:
//                -- all structure typedef names start with a capital letter,
//                and end in "Adt" to allow the programmer reading this code to
//                quickly distinguish them from instances
//                -- all "instances" of ADTs start with a lower case letter and
//                typically end in "Obj", to allow the programmer reading this
//                code to quickly distinguish instances from class (parent)
//                -- "instances" that are pointers are given a suffix of "_p" to
//                allow the reader to confirm they are pointers, as opposed to
//                non pointers.
//                -- where appropriate, "ARRAYS" are distinguished from non
//                arrays by use of a "Ary" suffix.
/*
 module       : tcf_struct.h
 purpose      : Contains data structure definitions for 'tcf_spin' program.
 project      : SMAP L4C science development.
 dependencies :
 history     : v3.1, Initial reformatting -- laj 09/10/13
 author       : Lucas A. Jones; lucas@ntsg.umt.edu
 notes        :

*/

#ifndef TCF_STRUCT_DEF
#define TCF_STRUCT_DEF 1

// local ADTs needed for DFNT_FLOAT32, 'float32' style datatype definitions
#include "hdf.h"
#include "mfhdf.h"

// PARAMETER STRUCTURES-------------------------------------------------------

// include BPLUT ADTs and protos

typedef struct tagBPLUT {
  // key for lookups...
  char lutKey[32];     // landcover or PFT abbreviated unique labels, e.g. EBF,
                       // ENF,...
  int lutIntegerIndex; // added for base0 index retrievals

  /*  parameters for NDVI-FPAR conversion  */
  char useNDVItoFPAR[32];   // TRUE: use linear model between NDVI and FPAR;
                            // Otherwise: use FPAR with standard gain and offset
  float64 NDVItoFPAR_scale; // linear model scale coefficient used to convert
                            // NDVI to FPAR
  float64 NDVItoFPAR_offset; // linear model offset coefficient used to convert
                             // NDVI to FPAR

  /* parameters for MODIS GPP/NPP algorithm */
  float64 LUEmax; /* [g C m-2 d-1 MJ-1] Maximum light use efficiency for GPP */

  /* meteorological constraint parameters */
  float64 Tmin_min; // [K] Minimum value where minimum daily air temperature
                    // constrains GPP
  float64 Tmin_max; // [K] Maximum value where minimum daily air temperature
                    // constrains GPP
  float64 VPD_min;  // [Pa] Minimum value where average daily vapor pressure
                    // deficit constrains GPP
  float64 VPD_max;  // [Pa] Maximum value where average daily vapor pressure
                    // deficit constrains GPP
  float64 SMrz_min; // [% Sat.] Minimum value where root zone soil moisture
                    // constrains GPP
  float64 SMrz_max; // [% Sat.] Maximum value where root zone soil moisture
                    // constrains GPP

  // <LAJ> 08/06/2014, Addded FT_min/max fields for incorporating FT into model
  // forward step and QA calculations
  float64 FT_min; // [dim.] Minimum value where freeze-thaw constrains GPP
                  // (Value if Frozen, FT==1)
  float64 FT_max; // [dim.] Maximum value where freeze-thaw constrains GPP
                  // (Value if non-frozen, FT==0)

  float64 SMtop_min;   // [% Sat.] Minimum value where surface soil moisture
                       // constrains Rh
  float64 SMtop_max;   // [% Sat.] Maximum value where surface soil moisture
                       // constrains Rh
  float64 Tsoil_beta0; // [dim.] Arrhenius temperature [K] constraint curve
                       // shape parameter for Rh
  float64 Tsoil_beta1; // [dim.] Arrhenius temperature [K] constraint curve
                       // shape parameter for Rh
  float64 Tsoil_beta2; // [dim.] Arrhenius temperature [K] constraint curve
                       // shape parameter for Rh

  /* Fractionation parameters for the ecosystem respiration model component */
  float64 fraut; // [dim.] Fraction of GPP respired as autotrophic respiration
  float64 fmet;  // [dim.] Fraction of annual NPP allocated to soil organic C
                 // metabolic pool
  float64 fstr;  // [dim.] Fraction of structural soil organic C pool allocated
                 // to slow soil organic C pool

  // Rate constants for soil decompoisiton model component
  float64 kopt; // [g C m-2 d-1] Optimal decomposition rate for metabolic soil
                // organic C pool
  float64 kstr; // [dim.] Structural soil organic C pool decomposition optimal
                // rate as a fraction of the metabolic optimal rate
  float64 kslw; // [dim.] Slow soil organic C pool decomposition optimal rate as
                // a fraction of the metabolic optimal rate

  // QA parameter fields
  uint32 Nee_QA_Rank_min;   // [dim.] Integer rank representing highest quality
                            // (lowest error level) Net Ecosystem CO_2 Exchange
  uint32 Nee_QA_Rank_max;   // [dim.] Integer rank representing lowest quality
                            // (highest error level) Net Ecosystem CO_2 Exchange
  float64 Nee_QA_Error_min; // [g C m-2 d-1] Floating point lowest error
                            // corresponding to highest quality rank Net
                            // Ecosystem CO_2 Exchange
  float64 Nee_QA_Error_max; // [g C m-2 d-1] Floating point highest error
                            // corresponding to lowest quality rank Net
                            // Ecosystem CO_2 Exchange

  uint32 Fpar_QA_Rank_min;   // [dim.] Integer rank representing highest quality
                             // (lowest error level) Fpar input
  uint32 Fpar_QA_Rank_max;   // [dim.] Integer rank representing lowest quality
                             // (highest error level) Fpar input
  float64 Fpar_QA_Error_min; // [dim.] Floating point lowest error corresponding
                             // to highest quality rank Fpar input
  float64 Fpar_QA_Error_max; // [dim.] Floating point highest error
                             // corresponding to lowest quality rank Fpar input

  // <LAJ> 08/06/2014, Addded FtAge fields for incorporating FT into QA
  // calculation
  float64 FtMethod_QA_mult; // [dim.] Multiplier representing additional error
                            // for Radar vs. Tsurf FT input (TsurfError =
                            // RadarError + RadarError * FtMethod_QA_mult)
  uint32 FtAge_QA_Rank_min; // [dim.] Integer rank representing highest quality
                            // (lowest error level) FT Age input
  uint32 FtAge_QA_Rank_max; // [dim.] Integer rank representing lowest quality
                            // (highest error level) FT Age input
  float64 FtAge_QA_Error_min; // [dim.] Floating point lowest error
                              // corresponding to highest quality FT Age input
  float64 FtAge_QA_Error_max; // [dim.] Floating point highest error
                              // corresponding to lowest quality FT Age input

  float64 Par_QA_Error;  // [MJ m-2 d-1] Model input static photosynthetically
                         // active radiation error level
  float64 Tmin_QA_Error; // [K] Model input static minimum daily air temperature
                         // error level
  float64 Tsoil_QA_Error; // [K] Model input static rootzone soil moisture error
                          // level
  float64 Vpd_QA_Error;  // [Pa] Model input static average daily vapor pressure
                         // deficit error level
  float64 Smrz_QA_Error; // [% Sat.] Model input static rootzone soil moisture
                         // error level
  float64 Smtop_QA_Error; // [% Sat.] Model input static surface soil moisture
                          // error level

} BPLUT;

// Structure for log file
// PRIOR: log_struct  NEW NS: LogSesAdt
typedef struct tagLogSesAdt {
  char *fname;
  FILE *fid;
} LogSesAdt; // instance would be: logSesObj

// Global Parameter Structure
// PRIOR: param_struct, NEW namespace: ParamAdt
typedef struct tagParamAdt {

  char basedir[MXSTRLEN]; // Base directory for entire program

  char subset_name[MXSTRLEN]; // Name of subset (not used for grid runs)
  uint32 nsite;               // Number of sites in subset
  // <JMG> most if not all of these path specs are obsolete, superceded by
  // SesCtl's instances
  char met_basepath[MXSTRLEN];  // Directory and prefix for input met datasets
  char fpar_basepath[MXSTRLEN]; // Directory and prefix for input fpar datasets
  char fpar_qa_basepath[MXSTRLEN]; // Directory and prefix for input fpar qa
                                   // datasets
  int fpar_qa_year; // Year for example fpar qa (for the 'step_forward' action)
  char pft_basepath[MXSTRLEN];    // Directory and prefix for input pft datasets
  char output_basepath[MXSTRLEN]; // Directory and prefix for model output task
                                  // results
  char output_basepath_land[MXSTRLEN]; // Directory and prefix for model land
                                       // domain output results (usually read
                                       // SOC/lit from here)
  char log_basepath[MXSTRLEN]; // Directory and prefix for model log files
  char bpLutPath[MXSTRLEN];    // Full path to BPLUT

  uint32
      prog_cnt_div; // Number of divisions (NDAYS/div = %finished) for progress
                    // counter for numerical spin-up and forward runs
  uint32 spinup_max_iter;  // Max number of times meteorolgy is cycled for
                           // numerical spin-up cycles
  float64 spinup_stop_tol; // [g m-2 y-1] Nee stopping tolerance for spinup

  float64 fparGain; // MODIS Fpar gain to convert from uint8 to floating point
  float64
      fparOffset; // MODIS Fpar offset to convert from uint8 to floating point

} ParamAdt; // instance would be: paramObj  or paramObj_p if a pointer

// <LAJ>  07/18/2014: Added QcStateLimitAdt.
// Contains min/max limit parameters used to determine qc bit values.
// State variable values are checked against these limits and corresponding qc
// bit set to 1 if limits exceeded.  Values to populate this structure are
// specified in the RIP file. Data stored in this structure will be accessed
// using an accessor function.

typedef struct tagQcStateLimitAdt {

  // Max and min values for flaging state variables as out-of-range in
  // 'carbon_model_bitflag'
  float32 neeLimitMin;
  float32 neeLimitMax;

  float32 gppLimitMin;
  float32 gppLimitMax;

  float32 rhLimitMin;
  float32 rhLimitMax;

  float32 socLimitMin;
  float32 socLimitMax;

  float32 useFparClimThres; // Set bit for GPP_Method above this threshold (as a
                            // percentage of valid 1km grid cells where Fpar
                            // clim was used)
  float32
      useNdviThres; // Set bit for NDVI_Method above this threshold (as a
                    // percentage of valid 1km grid cells where Ndvi was used)

} QcStateLimitAdt; // instance would be: QcStateLimitObj or QcStateLimitAdt_p if
                   // a pointer

// Holds variables for controlling program behavior. This structure is
// in the process of being deprecated, as all variables representing
// its functionality are being migrated to SesCtl or RunParamAdt
//  PRIOR name : control_struct
//  NEW   name : AppControlAdt
//  Instance(s): are usually "CONT" or "CONT_p"
typedef struct tagAppControlAdt {

  // Parameters read in from text file
  // char        action[MXSTRLEN];             // Specifies mode:  whether to do
  // grid or subset

  uint32 doyStep;  // (deprecated) Specifies Doy (1-365) to run step simulation
  uint32 doy8step; // (deprecated) Specifies Doy8 (1-46) fpar interval to run
                   // step simulation

  char glob_param_fpath[MXSTRLEN]; // File path for global parameter file

  // Universal timing and logging information used by program
  time_t strtSec;  // Begin time
  time_t endiSec;  // End time
  float64 elapSec; // Elapsed time
  char dtBuf[32];  // Message buffer

  // in final l4c_DAILY, FRM_API's session log facility is used
  LogSesAdt logSesObj; // instance of LogSesAdt is logSesObj, structure for
                       // storing logfile

} AppControlAdt; // instance would be: appControlObj   or  appControlAdt_p if a
                 // pointer

/*INPUT DATA
 * STRUCTURES---------------------------------------------------------------------------*/

// DEPRECATED: type_struct" (now: PftClassAdt)
// Contains summary arrays and scalars for each state variable

//  <LAJ> 3/11/14 - Moved from l4c_PostProc9km.h. The structure here now used
//  within l4c_PostProc9km.c.

typedef struct tagPftClassAggr9kmAdt {
  uint8 *cnt;
  float32 *avg;
  float32 sum;
  float32 sum2;
  uint8 *rank;
} PftClassAggr9kmAdt;

/* Base structure to contain various state variable summaries*/
// PRIOR: var_float32_struct,   NEW: VarFloat32Adt
// WARNING: tcf_proc9km_v1.0 ALSO defines same-named "var_float32_struct"
//  as this one here below, but defines its contents differently. It should
//  have been named something different so as not to trigger a namespace
//  conflict.
//  <LAJ> 3/11/14 - Moved fields from cooresponding structure in
//  l4c_PostProc9km.h. The structure here now used within
//  l4c_DAILY_ForwardStep.c AND l4c_PostProc9km.c.
typedef struct tagVarFloat32Adt {

  PftClassAggr9kmAdt
      pft_9km[NPFT]; // average for each cover type within 9km grid cell

  uint8 domPft_9km; // dominant landcover for 9km grid cell (written immediately
                    // into 'carbon_model_bitflag' so does not require array)

  float32 *dom_1km; // Fundamental model spatial unit containing global 1km data
                    // array (NOTE: only used for SOC)
  float32 *nested_1km; // Fundamental model spatial unit containing 1km data (81
                       // total grid cells) nested within each 9km grid cell

  uint8 *cnt_9km;   // count of 1km grid cells within overlying 9km grid cell
  float32 *std_9km; // count of 1km grid cells within overlying 9km grid cell
  float32 *avg_9km; // spatial average of 1km grid cells within overlying 9km
                    // grid cell
  uint8 *rank_9km;  // qa rank
  uint8 *cnt_useFparClim_9km; // Count of instances of 1km FPAR climatology
                              // usage for GPP method
  uint8 *cnt_useNdvi_9km; // Count of instances of 1km NDVI usage for GPP method
  uint8
      *cnt_useRadarFT_9km; // Count of instances of 1km NDVI usage for FT method
  uint8 *cnt_useTsurfFT_9km; // Count of instances of 1km NDVI usage for Tsurf
                             // method
  // <LAJ> 7/21/14 : Changed name to 'carbon_model_bitflag' from 'qcScore' to
  // match PSD.
  uint16 *carbon_model_bitflag;
  // Intermediate variables for average and stdev. calculations
  float32 sum_9km; // sum of 1km grid cells within overlying 9km grid cell
  float32
      sum2_9km; // sum of squares 1km grid cells within overlying 9km grid cell
  float32 var_9km; // variance 1km grid cells within overlying 9km grid cell

  uint8 limitExceedCnt_9km; // Count of nested 1km grid cells exceeding expected
                            // reasonable threshold within each 9km grid cell
  float32 limitMin; // Maximum expected reasonable value for state variable
  float32 limitMax; // Minimum expected reasonable value for state variable

  // <LAJ> 10/29/14 : Added these grand sums for each grid
  size_t domainConsistencySum_1km;    // Sum of 1km field over ENTIRE global 1-D
                                      // sparse vector
  size_t completenessOmissionSum_1km; // Sum of 1km field over ENTIRE global 1-D
                                      // sparse vector

} VarFloat32Adt; // instance would be: varFloat32Obj

// Plant Function Type Data Structure
// PRIOR: pft_struct  NEW NS: PftAdt
//
// WARNING: tcf_proc9km_v1.0 ALSO DEFINES same-named "pft_struct" HOWEVER
//  it is not identical to this one below. It should have been named something
//  different, so as not to conflict with this namespace!!!
//  <LAJ> 3/11/14 - Now accepts PftClassAggr9kmAdt data type as
//  defined above.  Fields other than 'dom_1km' moved here from
//  l4c_PostProc9km.h.
typedef struct tagPftAdt {
  uint8 *dom_1km;    // Dominant cover type for each 1km grid cell
  uint8 *nested_1km; // Holds nested 1km array within each 9km grid cell (81
                     // elements)

  /* The following derived types are calculated from 1km data (rather than 300m,
   * the other c5 option)*/
  /* This is done because 1km is the fundemental resolution of the Gpp data*/
  uint8 *dom_9km; /*Dominant cover type for each 9km grid cell (from 1km data)*/

  // recall that NLC is __ so is NOT equal to NPFT
  PftClassAggr9kmAdt lc_type[NLC]; // was: 'type_struct', but now with
                                   // namespace: PftClassAggr9kmAdt

  uint8 *veg_tot;     /*total number of cells with veg pft*/
  uint8 *margveg_tot; /*total number of cells with marginal-veg pft that might
                         be excluded or included*/
  uint8 *nonveg_tot;  /*total number of cells with non-veg pft*/
  uint8 *msnglc_tot;  /*total number of cells with msng pft*/

} PftAdt; // instance would be: pftObj  or pftObj_p

// Meteorological data structure for input arrays
// PRIOR: met_array_struct    NEW NS : MetArrayAdt
// Planned changed: change all fpar_qa to "fpar_qc" to match actual legacy name
// in MODIS HDF4...
typedef struct MetArrayAdt {

  uint8 *viirsNdvi;    // [dim.] Optional VIIRS NDVI input (alternative to FPAR)
  uint8 *viirsNdvi_qc; // [bit-field] NDVI's qc field, analog to "fpar_qc" <JMG>
                       // added 5/16/2014
  uint8 *fpar;         // [dim.] FPAR 8-day composite
  uint8 *fpar_clim;    // [dim.] FPAR 8-day composite climatology <LAJ> added
                       // 04/16/2014

  // <JMG> TBD, later carefully change all "fpar_qa" to "fpar_qc" to match MODIS
  // HDF4 legacy...
  uint8 *fpar_qa; // [bit field] FPAR 8-day composite qa bit field

  // uint16     *surflag_3km;       // [bit field] L3_SM_A surface flags
  // contains L3 SM A Freeze-Thaw flag (bit 6, base 0) <JMG> 2014-08-07T12:16,
  // added for update to new (uint8*)ft_state_um_M03 replacement for
  // *surface_flag from HDF5 uint8 FT_STATE_UM_M03, contains 4 bits extracted
  // from L3_SM_A uint16 surface_flag <JMG> 2015-07-29, added support for  F/T
  // from either GEOS5 TS derivation, or L3_SM_P radiometer, still at 3KM
  uint8 *ft_state_um_3km; // {ft_tsurf,ft_radar,ft_radar_qc,ft_in_swath}

  float32 *tmin; // [K] Minimum daily 2m air temperature
  float32 *tmax; // [K] Maximum daily 2m air temperature
  float32 *vpd;  // [Pa] Average daily vapor pressure deficit
  float32 *par;  // [MJ m-2 d-1] Downwelling photosynthetically-active radiation

  float32 *tsoil; // [K] Soil temperature top layer (0-5 cm)
  float32 *smtop; // [% Sat.] Soil moisture surface layer (0-5 cm)
  float32 *smrz;  // [% Sat.] Soil moisture root zone (0-100 cm)

  float32
      *smrz_mn; // [% Sat.] Soil moisture root zone record minimum for rescaling
  float32
      *smrz_mx; // [% Sat.] Soil moisture root zone record maximum for rescaling
  // <JMG> added 2/24/2014 to consolidate...
  float32
      *qv2m; // specific humidity, kg/kg, from SMAP_L4_C_MET.h5, for vpd calc
  float32 *t2m_avg; // 2 meter air temperatures, K, from SMAP_L4_C_MET.h5, for
                    // vpd calc
  float32 *ps_avg;  // surface pressure, Pa, from SMAP_L4_C_MET.h5, for vpd calc

} MetArrayAdt; // instance would be: metArrayObj or metArrayObj_p

// Meteorological Data Structure for internal calculations
// PRIOR: met_struct  NEW NS: MetAdt
typedef struct tagMetAdt {

  float64 fpar;  // [dim.] FPAR 8-day composite
  uint8 fpar_qa; // [bit field] FPAR 8-day composite qa bit field

  // <JMG> 2015-07-29, v7 FORK, added support for new 9KM F/T from GEOS5 FP TS,
  // and/or Passive L3_SM_P 36KM TBD: need to finalize how to handle {In|Out}
  // swath independently of resolution?

  // DEPRECATED 2014-08-07, (uint16) surflag_3km, now use (uint8)
  // ft_state_um_3km uint16     surflag_3km;      // [bit field] L3_SM_A surface
  // flags contains L3 SM A Freeze-Thaw flag (bit 6, base 0)
  uint8 ft_state_um_3km; // NEW, from L3_SM_A, stores (4) bits
                         // {ft_tsurf,ft_radar,ft_radar_qc,ft_in_swath}
  // <JMG> 2015-08-26T16:11, ft_in_swath now changed BACK to ft_in_swath_3km for
  // compatibility with v6 code for comparison testing. <JMG> 2015-07-30, when
  // GEOS5 FP "TS" or "T2M" source for F/T is used, set to ALWAYS in swath note
  // that prior 'ft_in_swath_3km' is now changed to 'ft_in_swath' to emphasize
  // this is not a resolution dependent switch now.
  uint8 ft_in_swath_3km; // [bool] Boolean value from L3_SM_A Surface in-swath
                         // flag (1=in swath; 0=not in swath)

  // <LAJ> 7/25/14, added ft_Radar_3km and ft_Tsurf_3km
  uint8 ft_Radar_3km; // from L3_SM_A Surface Flag RADAR Freeze-Thaw (1=Thawed;
                      // 0=Frozen)
  // <JMG> 2014-08-07 added for completeness, ft_Radar_Qc_3km not yet used?
  uint8 ft_Radar_Qc_3km;
  // <LAJ> added...
  uint8 ft_Tsurf_3km; // from L3_SM_A Surface Flag Tsurf Freeze-Thaw (1=Thawed;
                      // 0=Frozen)

  float64 tmin; // [K] Minimum daily 2m air temperature
  float64 tmax; // [K] Maximum daily 2m air temperature
  float64 vpd;  // [Pa] Average daily vapor pressure deficit
  // <JMG> ..read swgdn into this 'par' field...
  float64 par; // [MJ m-2 d-1] Downwelling photosynthetically-active radiation

  float64 tsoil; // [K] Soil temperature top layer (0-5 cm)
  float64 smtop; // [% Sat.] Soil moisture surface layer (0-5 cm)
  float64 smrz;  // [% Sat.] Soil moisture root zone (0-100 cm)

  uint8 ft;    // [bool] Boolean freeze-thaw (1=Thawed; 0=Frozen) assigned from
               // ftRolling 3km to use in L4C model step
  uint8 ftAge; // [days] Age of freeze-thaw information contained in ftRolling
               // 3km (NOTE: older data outside of current swath are retained
               // from prior swaths)

  // <JMG> added 2/24/2014 to consolidate...
  float64 qv2m; // specific humidity, kg/kg, from SMAP_L4_C_MET.h5, for vpd calc
  float64 t2m_avg; // 2 meter air temperatures, K, from SMAP_L4_C_MET.h5, for
                   // vpd calc
  float64 ps_avg;  // surface pressure, Pa, from SMAP_L4_C_MET.h5, for vpd calc

  // <LAJ> added 8/6/2014 to make these fields available wherever metObj is
  // passed <JMG> 2014-11-10, renamed prior (4-state) gppMethod to gppInputUsed
  // to better distinguish this from the RIP file's
  // (3-state:FPAR8DAY|NDVI|FPARCLIM) L4C_GPP_SOURCE.
  int gppInputUsed; // Whether FPAR or NDVI was used in the GPP calculation

  // <JMG> 2015-07-29 pending, check all usages and modify to support new 3km
  // GEOS5 derived ft modes:{DoUseFT_ActiveRadar, DoUseFT_ActiveTsurf,
  // DoUseFT_PassiveRadiom,DoUseFT_PassiveTsurf, DoUseFT_GEOS5}
  int ftMethod; // Whether Radar FT or Turf FT was used in the Frozen Area %
                // calculation

} MetAdt; // instance would be: metObj or metObj_p

// MODEL STATE VARIABLE DATA
// STRUCTURES-------------------------------------------------------------

// <LAJ> 3/12/14 REMOVED EcoStateArrayQaAdt.  NO LONGER NEEDED

// Structure for output state variable arrays
// PRIOR: state_array_struct  NEW NS: EcoStateArrayAdt
typedef struct tagEcoStateArrayAdt {

  VarFloat32Adt nee; // [g C m-2 d-1] Net Ecosystem CO_2 Exchange (NEE)
  VarFloat32Adt
      nee_qa; // [g C m-2 d-1] NEE Psuedo-Error for QA rank classification

  VarFloat32Adt gpp;       // [g C m-2 d-1] Gross Primary Productivity (GPP)
  VarFloat32Adt npp;       // [g C m-2 d-1] Net Primary Productivity (NPP)
  VarFloat32Adt reco;      // [g C m-2 d-1] Ecosystem Respiration (Reco)
  VarFloat32Adt rh[NPOOL]; // [g C m-2 d-1] Heterotrophic Respiration (Rh)
  VarFloat32Adt rhtot;     // [g C m-2 d-1] Total Heterotrophic Respiration
  VarFloat32Adt C[NPOOL];  // [g C] Soil Organic Carbon Pools (SOC)
  VarFloat32Adt Ctot;      // [g C] Total Soil Organic Carbon
  VarFloat32Adt
      dCtot; // [g C d-1] (diagnostic) Daily change in Total Soil Organic Carbon

  VarFloat32Adt
      Emult; // [dim.] Bulk environmental constraint multiplier for GPP
  VarFloat32Adt Tmult; // [dim.] Soil temperature constraint multiplier for Rh
  VarFloat32Adt Wmult; // [dim.] Soil moisture constraint multiplier for Rh

  // NOTE: this is needed for analytical stst AND litterfall calcs. litterfall
  // == npp_sum

  // <KAE> 2024-01-03 To support litterfall allocation
  float32 *litter_rate;     // [dim.] Fraction of available NPP that enters SOC pools
  VarFloat32Adt npp_sum;   // [g C m-2 y-1] Annual climatological sum of NPP
  VarFloat32Adt kmult_sum; // [dim. x days] Annual climatological sum of Rh
                           // environmental constraints

  VarFloat32Adt nee_sum;  // [g C m-2 y-1] Annual climatological sum of NEE
  VarFloat32Adt gpp_sum;  // [g C m-2 y-1] Annual climatological sum of GPP
  VarFloat32Adt reco_sum; // [g C m-2 y-1] Annual climatological sum of Reco
  VarFloat32Adt
      rhtot_sum; // [g C m-2 y-1] Annual climatological sum of Total Rh

  // moved new members to bottom
  uint32 *spinup_iter; // [count] Number of spin-up iterations completed
  uint8 *spinup_stop;  // [g C m-2 y-1] Spin up tolerance when spin-up completes

  float32 *FrozenArea_9km; // [%] Percent of subgrid Frozen Area within 9-km
                           // grid cell
  uint8 *ftRolling_3km;    // [0,1] Contains current 3-km FT state updated daily
  uint8 *ftRollingAge_3km; // [d] Contains age (in days) of current rolling FT
                           // state
  uint8 *ftRollingQA_3km;  // <JMG> may deprecate this, as ftRollingAge_3km
                           // serving this role

} EcoStateArrayAdt; // instance would be:

// Structure for holding multiplier data
// PRIOR: mult_struct   NEW NS: MultAdt
typedef struct tagMultCoefAdt {
  // state fields
  float64 k; // [g C m-2 d-1] Effective rate for Rh

  float64 tsoil; // [dim.] Soil temperature multiplier for Rh
  float64 tmin;  // [dim.] Minimum daily air temperature multiplier for GPP
  float64 vpd; // [dim.] Average daily vapor pressure deficit multiplier for GPP
  float64 smtop; // [dim.] Surface soil moisture multiplier for Rh
  float64 smrz;  // [dim.] Root zone soil moisture multiplier for GPP
  float64 ft;    // [dim.] Freeze-Thaw multiplier for GPP

  // state partial derivative fields (for QA)
  float64 d_tsoil; // [g C m-2 d-1 K-1] NEE derivative with respect to soil
                   // temperature
  float64 d_tmin;  // [g C m-2 d-1 K-1] NEE derivative with respect to minimum
                   // daily air temperature
  float64 d_vpd;   // [g C m-2 d-1 Pa-1] NEE derivative with respect to average
                   // daily vapor deficit
  float64 d_smtop; // [g C m-2 d-1 %Sat.-1] NEE derivative with respect to
                   // surface soil moisture
  float64 d_smrz;  // [g C m-2 d-1 %Sat.-1] NEE derivative with respect to root
                   // zone soil moisture
  float64
      d_ft; // [g C m-2 d-1 dim.-1] NEE derivative with respect to freeze-thaw

} MultCoefAdt;

// State Variable Structure for qa calculations
// PRIOR: state_qa_struct  NEW NS: EcoStateQaAdt
typedef struct tagEcoStateQaAdt {

  float64 jac[NMET]; // Jacobian (array of partial derivatives) for propagation
                     // error from model inputs to outputs
  float64 met_error[NMET]; // Input error array

  float64
      nee_error; // Output error for Net Ecosystem CO_2 Exchange state variable

} EcoStateQaAdt; // instance would be:  ecoStateQaObj  or ecoStateqaObj_p

// State Variable Structure for calculations
// PRIOR: state_struct  NEW NS: EcoStateAdt
typedef struct tagEcoStateAdt {

  float64 nee;       // [g C m-2 d-1] Net Ecosystem CO_2 Exchange (NEE) [OUTPUT]
  float64 gpp;       // [g C m-2 d-1] Gross Primary Productivity (GPP)  [OUTPUT]
  float64 npp;       // [g C m-2 d-1] Net Primary Productivity (NPP)    [OUTPUT]
  float64 reco;      // [g C m-2 d-1] Ecosystem Respiration (Reco)
  float64 rh[NPOOL]; // [g C m-2 d-1] Heterotrophic Respiration (Rh)    [OUTPUT]
  float64 rhtot;     // [g C m-2 d-1] Total Heterotrophic Respiration   [OUTPUT]
  float64 C[NPOOL];  // [g C] Soil Organic Carbon Pools (SOC)
  float64 dC[NPOOL]; // [g C d-1] Daily change in Soil Organic Carbon   [OUTPUT]
  float64 Ctot;      // [g C] Total Soil Organic Carbon
  float64 dCtot;     // [g C d-1] Daily change in Total Soil Organic Carbon

  // NOTE: needed for analytical stst AND litterfall calcs. litterfall ==
  // npp_sum

  float64 lit;       // [g C m-2 d-1] Daily litterfall calculated from npp_sum/365
  float64 npp_sum;   // [g C m-2 y-1] Annual climatological sum of NPP
  float64 kmult_sum; // [dim. x days] Annual climatological sum of Rh
                     // environmental constraints

  MultCoefAdt rh_mult; // [dim.] Bulk environmental constraint multiplier for Rh
  MultCoefAdt gpp_mult; // [dim.] Bulk environmental constraint multiplier for
                        // GPP (equivalent to Emult)

  EcoStateQaAdt ecoStateQaObj; // was "qa", now 'ecoStateQaObj',  Subfield to
                               // hold qa-related data

  float32 ft_sum_9km; // [count] Sum of 3-km grid cells with frozen state for
                      // 9-km frozen area [%] calculation

} EcoStateAdt; // instance would be:  ecoStateObj   or ecoStateObj_p

#endif
