// ----------------------------------------------------------------
// module       : l4c_tcfMem.c
// purpose      : encapsulates tcfMem.c
// author       : Lucas A. Jones; lucas@ntsg.umt.edu
// author       : Joe Glassy
// project      : SMAP L4C science development.
// dependencies : tcfModDriverGrid.c, tcfMem.c
// revision     : 2014-08-07 <JMG> removed surflag_3km[], replaced with ft_state_um_3km[]
// history      : 2014-0516T14:11, add both alloc AND free of ndvi --
//                {viirsNdvi,viirsNdvi_qc} 
// revision     : Initial reformatting -- laj 09/09/13		
// notes        : uses memObj memory management now
// ----------------------------------------------------------------

// #include "tcf_inc.h"
#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"

// ----------------------------------------------------
/*
 module       : tcfMem.c
 purpose      : Memory management and high level input-output functions where convenient.
 project      : SMAP L4C science development.
 dependencies : array_io.c
 revision     : Initial reformating -- laj 09/10/13
 author       : Lucas A. Jones; lucas@ntsg.umt.edu
 notes        : 
*/

// ----------------------------------------------------
// function :tcfMem_metArrayAlloc()
// Purpose  : Memory allocation for input daily met grid arrays. 
// Inputs   : Array element lengths for 9km and 1km land grids.
// Outputs  : Allocated MetArrayObj structure fields.
// revised  : 2014-05-16T14:10:00, added viirsNdvi,virrsNdvi_qc alloc
// history  : 2014-02-24T15:51, added qv2m,t2m_avg,ps_avg to consolidate
// notes    : TBD: carefully review which arrays originate in the
//            pre-processor and thus do NOT want to be re-allocated
//            here.
void tcfMem_metArrayAlloc(uint32 nelem9km)
{
 SesCtl       *sesCtl_p = l4c_GetSesCtl();
 char fxName[256] = {0} ;
 
 MetArrayAdt  *metArrayObj_p  = l4c_GetMetArrayAdt();
 uint32  nelem1km =0, nelem3km =0;

	strcpy(fxName,__func__);
	
	xLogMsg("@A In(%s) Allocate 9KM (%zu) tmin,vpd,par,tsoil,smtop,smrz,qv2m,tm2_avg,ps_avg ~ \n",
		fxName, nelem9km );

  // met data arrays:
  metArrayObj_p->tmin    = (float32*)memCalloc(nelem9km, sizeof(float32));
	memObjSetName(metArrayObj_p->tmin,"metArrayObj_p->tmin");
	
  // VERIFY! Double check if vpd already allocated in l4c_PreProc1km_func.c
  // in function: l4c_PreProc_Calculate_VPD() line 801!
  // TEST: fetch name set earlier via memObjSetName and if equal "metArrayObj_p->vpd"
  // then DO NOT re-allocate this since it is already allocated!
  metArrayObj_p->vpd     = (float32*)memCalloc(nelem9km, sizeof(float32));
	memObjSetName(metArrayObj_p->vpd,"metArrayObj_p->vpd");
	
  metArrayObj_p->par     = (float32*)memCalloc(nelem9km, sizeof(float32));
	memObjSetName(metArrayObj_p->par,"metArrayObj_p->par");
  metArrayObj_p->tsoil   = (float32*)memCalloc(nelem9km, sizeof(float32));
	memObjSetName(metArrayObj_p->tsoil,"metArrayObj_p->tsoil");
  metArrayObj_p->smtop   = (float32*)memCalloc(nelem9km, sizeof(float32));
	memObjSetName(metArrayObj_p->smtop,"metArrayObj_p->smtop");
  metArrayObj_p->smrz    = (float32*)memCalloc(nelem9km, sizeof(float32));
	memObjSetName(metArrayObj_p->smrz,"metArrayObj_p->smrz");
	
  // <JMG> 2/24/2014,added (3) from SMAP_L4_C_MET.h5, to consolidate
  metArrayObj_p->qv2m    =(float32*)memCalloc(nelem9km, sizeof(float32));
	memObjSetName(metArrayObj_p->qv2m,"metArrayObj_p->qv2m");
  metArrayObj_p->t2m_avg =(float32*)memCalloc(nelem9km, sizeof(float32));
	memObjSetName(metArrayObj_p->t2m_avg,"metArrayObj_p->t2m_avg");
  metArrayObj_p->ps_avg  =(float32*)memCalloc(nelem9km, sizeof(float32));
	memObjSetName(metArrayObj_p->ps_avg,"metArrayObj_p->ps_avg");

  nelem1km = nelem9km * NESTED_1KM_IN_9KM_SQ;
  // nelem1km is ________ here (verify: should be 1D-sparse vec 1km count
  metArrayObj_p->fpar = (uint8*)memCalloc(nelem1km, sizeof(uint8));
	memObjSetName(metArrayObj_p->fpar,"metArrayObj_p->fpar");
	
  // Allocate space for fpar qc and L3_SM_A_FT for forward step only
  if (strcmp("step_forward",sesCtl_p->appAction) == 0)
  {
    nelem1km = nelem9km * NESTED_1KM_IN_9KM_SQ;

	xLogMsg("@A In(%s) Allocating 1KM (%zu) fpar,fpar_qa,viirsndvi,viirsNdvi_qc~ \n",
		fxName, nelem1km );

    metArrayObj_p->fpar_qa = (uint8*)memCalloc(nelem1km, sizeof(uint8));
		memObjSetName(metArrayObj_p->fpar_qa,"metArrayObj_p->fpar_qa");

    // <JMG> 2014-05-16, verify: where is 'viirsNdvi' and 'viirsNdvi_qc' allocated?
    // most likely this is the place to add these allocations
    //  uint8      *viirsNdvi;     // [dim.] Optional VIIRS NDVI input (alternative to FPAR) 
    // uint8      *viirsNdvi_qc;  // [bit-field] NDVI's qc field, analog to "fpar_qc" 
    metArrayObj_p->viirsNdvi    = (uint8*)memCalloc(nelem1km, sizeof(uint8));
		memObjSetName(metArrayObj_p->viirsNdvi,"metArrayObj_p->viirsNdvi");
    metArrayObj_p->viirsNdvi_qc = (uint8*)memCalloc(nelem1km, sizeof(uint8));
		memObjSetName(metArrayObj_p->viirsNdvi_qc,"metArrayObj_p->viirsNdvi_qc");
    
    nelem3km = nelem9km * NESTED_3KM_IN_9KM_SQ;
    
	// <JMG> 2014-08-07, replaced surflag_3km[] with uint8 ft_state_um_3km[] array
	// reminder: ft_state_um_3km is a read-only array; it is not written
	// back out.
    metArrayObj_p->ft_state_um_3km = (uint8*)memCalloc(nelem3km, sizeof(uint8));
		memObjSetName(metArrayObj_p->ft_state_um_3km,"metArrayObj_p->ft_state_um_3km");
  }

  // removed REDUNDANT memory allocations (memCalloc) on smrz_mn, smrz_mx

  l4c_EventCompletionMsg(__func__);

} // end :: tcfMem_metArrayAlloc()


// ----------------------------------------------------  
// tcfMem_stateArrayAlloc()
// Purpose: Memory allocation for essential model state 
//          arrays required by all model modes. This fx only needed
//          or allowed where old design of separate "mkgrid" is used.
//          It CANNOT be used in new design where the L4_C pre-processor
//          allocates and populates as 1D sparse vectors these same
//          variables. 
// Limitations: this routine assumes 'array-parallel' style of segmented
//          problem partitioning, which is NOT YET completed in the 
//          production L4C codes.
// Notes  : This fx allocates only the linear subset (or sub region)
//          portion of each global array as required by an individual 
//          subtask, referenced now in SesCtl 
//
// Inputs : Array element length for 1km land grid.
//
// Outputs: Allocated STATE_ARRAY structure SOC and litterfall fields.
  
void tcfMem_stateArrayAlloc(uint32 nelem1km)
{
  EcoStateArrayAdt  *ecoStateArrayObj_p = l4c_GetEcoStateArrayAdt();
  int    l = 0;
  char fxName[256] = {0} ;
    
  // Initialization state data arrays:
  ecoStateArrayObj_p->npp_sum.dom_1km = (float32*)memCalloc(nelem1km, sizeof(float32));
  for (l=0; l < NPOOL; l++)
  {
    ecoStateArrayObj_p->C[l].dom_1km = (float32*)memCalloc(nelem1km, sizeof(float32));
  }
  
  l4c_EventCompletionMsg(__func__);

} // end :: tcfMem_stateArrayAlloc()


// ----------------------------------------------------
// functon: tcfMem_tcfMem_stateArrayFree()
// Purpose: Free memory for fundemental state array fields (SOC and litterfall).
// Notes: 
// Inputs: Previously allocated STATE_ARRAY structure fields pointers.
// Outputs: De-allocated STATE_ARRAY structure fields pointers.
void tcfMem_stateArrayFree(EcoStateArrayAdt  *ecoStateArrayObj_p)
{
  int     l;
  
  // FREE : ecoStateArrayObj_p->npp_sum.dom_1km 
  memFree(ecoStateArrayObj_p->npp_sum.dom_1km);

  for (l=0; l < NPOOL;l++)
    memFree( ecoStateArrayObj_p->C[l].dom_1km );
    
  l4c_EventCompletionMsg(__func__);  
  
} // end :: tcfMem_stateArrayFree()


// ----------------------------------------------------
// tcfMem_tcfMem_metArrayFree()
// Purpose: Free memory for input met array fields.
// Notes: 
// Inputs: Previously allocated MetArrayObj structure fields pointers.
// Outputs: De-allocated MetArrayObj structure fields pointers.
// Revised: 2014-0516T14:11, add free of viirsNdvi,viirsNdvi_qc to 
//          match new allocs of these.
void tcfMem_metArrayFree(MetArrayAdt  *metArrayObj_p)
{
  SesCtl *sesCtl_p = l4c_GetSesCtl();
 
  memFree(metArrayObj_p->tmin);
  memFree(metArrayObj_p->vpd);
  memFree(metArrayObj_p->par);
  
  memFree(metArrayObj_p->tsoil);
  memFree(metArrayObj_p->smtop);
  memFree(metArrayObj_p->smrz);

  memFree(metArrayObj_p->smrz_mn);
  memFree(metArrayObj_p->smrz_mx);
  // <JMG> 2/24/2014, added..
  memFree(metArrayObj_p->qv2m);
  memFree(metArrayObj_p->t2m_avg);
  memFree(metArrayObj_p->ps_avg);
  
  // Free heap memory assoc with fpar qa,  for forward step only
  if (strcmp("step_forward",sesCtl_p->appAction) == 0)
  {
    memFree(metArrayObj_p->ft_state_um_3km); // now uses ft_state_um_3km[]

#ifdef LET_GARBAGE_COLLECTOR_FREE_THESE
	memFree(metArrayObj_p->fpar);	
	// <JMG> TBD: verify if fpar_qa should be free'd here
    memFree(metArrayObj_p->fpar_qa);
    // <JMG> 2014-05-16, must free, to match new alloc of viirsNdvi etc
    memFree(metArrayObj_p->viirsNdvi);
    memFree(metArrayObj_p->viirsNdvi_qc);
#endif
    
  } // end:: step_forward()

  l4c_EventCompletionMsg(__func__);

} // end :: tcfMem_tcfMem_metArrayFree()
