// ---------------------------------------------------------------
// function     : l4c_DAILY_tcfModDriverSubset.c
// purpose      : Wraps tcfModDriverSubset.c for further evolution
// author       : joe glassy, Lupine Logic Inc.
// version      : see L4C_VERS_TAG , as 4.1.5 or later
// revised      : v4.1.6-2014-02-05T11:242:00-07:00,wheat,jmg"
// dependencies : HDF4, HDF5 libraries and L4C FRAMEWORK API.
// history      : n/a
// notes        : Based on and adapted from tcf_spin v3.2, this
//                module contains setup logic for just the NON DAILY 
//                SPINUP (actions) 
// ---------------------------------------------------------------

// #include "tcf_inc.h" now re-factored into l4c_DAILY.h
#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"

 
// MODEL DRIVER FUNCTIONS 

/*TCF_MOD_DRIVER_SUBSET_FORWARD_RUN()                                                              *
 *                                                                                                 *
 *    Drives forward run simulation.                                                               *
 *                                                                                                 *
 *    RETURNS:  Daily states, including final C pools, in 2D 1km grid arrays returned              * 
 *              in the STATE_ARRAY structure.                                                      *
 *                                                                                                 *
 *    REQUIRES: Initial C pools, passed in the STATE structure, dynamic daily met data,            *
 *              static pft map, and static models parameters.                                      *                                                   *                                                                                                 *
 */ 

void tcfModDriverSubset_forwardRun(PftAdt PFT, MetArrayAdt MetArrayObj, EcoStateArrayAdt *STATE_ARRAY)
{
  SesCtl *sesCtl_p = l4c_GetSesCtl(); // now, fetch ptr to L4_C high level session controls
  RunParamAdt *runParamObj_p = &sesCtl_p->runParamObj;
  int       pft_index;

  uint8     pft_now;
  int s;

  size_t    p, t, t8day, l, rn, cn;
  size_t    rc_1km_nested, rc_1km, rc_9km;

  size_t    rct8day_1km, rct_1km, rct_9km;

  size_t    nelem;

  MetAdt    MET;      // for local state calculations, not requiring large 2D grid arrays
  EcoStateAdt  STATE; // for local state calculations, not requiring large 2D grid arrays

  BPLUT *bp_p = NULL;

  /* Begin PFT loop */
  for (p=0; p < NPFT; p++)
  {

    pft_now = p+1; /*veg. pft from offset index*/
    
    /* Retrieve bplut parameters based on PFT*/
    pft_index = p+1;
    bp_p = l4c_GetBPLUT_ByIndex(pft_index, NPFT);

    /* Begin site loop*/
    for (s=0; s < runParamObj_p->nsite; s++)
    {

      rc_9km = s;


      // Begin nested 1km 2D grid loops
      
      for (rn=0; rn < NESTED_1KM_IN_9KM; rn++)
      {
	for (cn=0; cn <NESTED_1KM_IN_9KM; cn++)
	{

	  rc_1km_nested = M_2D_B0(rn,cn,NESTED_1KM_IN_9KM);
	  rc_1km = M_2D_B0(rc_1km_nested, rc_9km, runParamObj_p->nsite);

	  //	  printf("%d  %d  %d\n", rc_1km_nested, PFT.dom_1km[rc_1km_nested], PFT.dom_1km[rc_1km]); 

	  // do only specified pft
	  if (PFT.dom_1km[rc_1km] == pft_now)
	  {

	    // Initialize SOC
	    // SOC from spin-up
	    for (l=0;  l < NPOOL; l++)
	    {

	      STATE.C[l] = (float64) STATE_ARRAY->C[l].dom_1km[rc_1km];
	      //	      STATE.C[l] = 0.0;
	    }

	    //  Initialize npp sum for litterfall
	    // NPP sum from analytical steady-state SOC calc.
	    // Divide by number of days in the year, for constant daily litterfall
	    STATE.npp_sum = (float64) STATE_ARRAY->npp_sum.dom_1km[rc_1km] / NDAYS; 


	    /* printf("%d C[0] = %f  C[1] = %f  C[2] = %f  npp_sum = %f\n",  */
            /*        rc_1km, STATE.C[0], STATE.C[1], STATE.C[2], STATE.npp_sum); */


	    //  Begin time loop
		t8day = 0;
		for (t=0; t < NDAYS; t++)
	    {

	      nelem = runParamObj_p->nsite * NESTED_1KM_IN_9KM_SQ;
	      rct_1km = M_2D_B0(t,rc_1km, nelem);
	      rct_9km = M_2D_B0(t,rc_9km, runParamObj_p->nsite);

	      // Unpack input arrays for call interface
	      MET.tmin = (float64) MetArrayObj.tmin[rct_9km]; /* [K] */
	      MET.vpd = (float64) MetArrayObj.vpd[rct_9km];   /* [Pa] */
	      MET.par = (float64) MetArrayObj.par[rct_9km] * RAD2PAR;   /* convert to [MJ m-2] par */
	      MET.smrz = (float64) MetArrayObj.smrz[rct_9km] * 100.0; /*convert from [dim] to [%]*/

	      MET.tsoil = (float64) MetArrayObj.tsoil[rct_9km]; /* [K]*/
	      MET.smtop = (float64) MetArrayObj.smtop[rct_9km] * 100.0; /*convert from [dim] to [%]*/ 
	       
	      
	      /*  Read in fpar data when available every 8-days*/
	      /*  Otherwise, hold fpar constant at whatever was last read*/
	      /*  NOTE: this works for subsets, because the arrays are indexed to zero*/
	      if (t == DAY8_to_DOY(t8day))
	      {
	
				nelem = runParamObj_p->nsite * NESTED_1KM_IN_9KM_SQ;
				rct8day_1km = M_2D_B0(t8day,rc_1km,nelem);
				// <JMG> FPAR_scale,FPAR_offset have been moved... to?? find out!
				// <JMG> NOTE: sesCtl_p->gppFpar_or_NDVI is an int option switch set in l4c_DAILY_ses.c
				// in l4c_SesInit(), using an enum choice: DoUseFpar or DoUseNDVI set inside the RIP file
				MET.fpar = (float64) MetArrayObj.fpar[rct8day_1km] * runParamObj_p->FPAR_scale + runParamObj_p->FPAR_offset;
				t8day++;
	      } // end:: t test on DOY


	      // Do tcf forward step to propagate SOC states from last
	      // timesetp to current timestep
	      tcfModFunc_forwardStep(bp_p, MET, &STATE);

	      // Package state-tracking arrays
	      // Original 1km output states
	      STATE_ARRAY->gpp.dom_1km[rct_1km] = (float32) STATE.gpp;
	      STATE_ARRAY->rhtot.dom_1km[rct_1km] = (float32) STATE.rhtot;
	      STATE_ARRAY->nee.dom_1km[rct_1km] = (float32) STATE.nee;


	      /* printf("%d   %d  %f   %f   %f   %f   %f  %f  %f\n", p, t, MET.tmin, MET.vpd, MET.smrz, MET.par, MET.fpar, STATE.gpp, */
	      /* 	     STATE_ARRAY->gpp.dom_1km[rct_1km]); */

	      //	      printf("%d   %f   %f \n", rct_1km, MET.fpar, STATE_ARRAY->gpp.dom_1km[rct_1km]);


	      /*running average, completes when the 1km nested loop finishes*/
	      /* STATE_ARRAY->gpp.pft_9km[p].cnt[rct_9km] ++; /\*only need to track cnt for one state variable*\/ */
	      /* STATE_ARRAY->gpp.pft_9km[p].avg[rct_9km] += (float32) STATE.gpp / STATE_ARRAY->gpp.pft_9km[p].cnt[rct_9km]; */
	     
	      /*running grand average for all pfts in 9km grid cell,
	        completes when outer pft loop finishes*/
	      /* STATE_ARRAY->gpp.cnt_9km[rct_9km] ++; */
	      /* STATE_ARRAY->gpp.avg_9km[rct_9km] += (float32) STATE.gpp / STATE_ARRAY->gpp.cnt_9km[rct_9km]; */

	    } // End time loop

	    //  Package final SOC for future initialization
	    for (l=0; l < NPOOL; l++)
	      STATE_ARRAY->C[l].dom_1km[rc_1km] = (float32) STATE.C[l]; 
	    
 	    /* printf("%d C[0] = %f  C[1] = %f  C[2] = %f \n",  */
            /*        rc_1km, STATE_ARRAY->C[0].dom_1km[rc_1km], STATE_ARRAY->C[1].dom_1km[rc_1km],STATE_ARRAY->C[2].dom_1km[rc_1km]); */


	  } // end:: pft check

	}
      }  // End nested 1km 2D grid loops

    }
    /*End 9km location loop*/

  }
  // End PFT loop


} // END FUNCTION  TCF_DO_FORWARD_RUN()   



/*TCF_MOD_DRIVER_SUBSET_NUMSPIN_RUN()                                                              *
 *                                                                                                 *
 *    Drives numerical spinup.                                                                     *
 *                                                                                                 *
 *    RETURNS:  Initialized C pool 2D 1km grid arrays returned in the STATE structure.             *
 *              Also returns spinup tolerance and iteration count arrays (as 2D 1km grids)         *
 *              returned in the AUX structure.                                                     *
 *                                                                                                 *
 *    REQUIRES: Dynamic daily met data, static pft map, and static models parameters.              *
 *                                                                                                 *
 */

void tcfModDriverSubset_numSpin(PftAdt PFT, MetArrayAdt MetArrayObj, EcoStateArrayAdt *STATE_ARRAY)
{
  SesCtl *sesCtl_p = l4c_GetSesCtl(); // now, fetch ptr to L4_C high level session controls
  RunParamAdt *runParamObj_p = &sesCtl_p->runParamObj;
  int       pft_index;
  int s=0;
  
  uint8     pft_now;

  uint32    p, t, t8day, l, rn, cn;
  uint32    rc_1km_nested, rc_1km, rc_9km;

  uint32    rct8day_1km, rct_1km, rct_9km;

  uint32    nelem;

  uint32    spinup_iter_cnt;
  float64   spinup_current_iter_tol;

  MetAdt    MET;      // for local state calculations, not requiring large 2D grid arrays*/
  EcoStateAdt  STATE; // for local state calculations, not requiring large 2D grid arrays*/

  BPLUT                  *bp_p = NULL;

  // Begin PFT loop
  
  for (p=0; p < NPFT; p++)
  {

    pft_now = p+1; /*veg. pft from offset index*/
    
    /*Retrieve bplut parameters based on PFT*/
    pft_index = p+1;
    bp_p = l4c_GetBPLUT_ByIndex(pft_index, NPFT);

    /*Begin site loop*/
    for (s=0; s < runParamObj_p->nsite; s++)
    {

      rc_9km = s;
 
      /*Begin nested 1km 2D grid loops*/
      for (rn=0; rn < NESTED_1KM_IN_9KM; rn++)
      {
	for (cn=0; cn < NESTED_1KM_IN_9KM; cn++)
	{

	  rc_1km_nested = M_2D_B0(rn,cn,NESTED_1KM_IN_9KM);
	  rc_1km = M_2D_B0(rc_1km_nested, rc_9km, runParamObj_p->nsite);

	  //	  printf("%d  %d  %d\n", rc_1km_nested, PFT.dom_1km[rc_1km_nested], PFT.dom_1km[rc_1km]); 

	  //  do only specified pft
	  if (PFT.dom_1km[rc_1km] == pft_now)
	  {

	    /*Initialize SOC*/
	    /*SOC from spin-up*/
	    for (l=0; l < NPOOL; l++)
	    {

	      STATE.C[l] = (float64) STATE_ARRAY->C[l].dom_1km[rc_1km];
	      // STATE.C[l] = 0.0;
	    }

	    /*Initialize npp sum for litterfall*/
	    /*NPP sum from analytical steady-state SOC calc.*/
	    /*Divide by number of days in the year, for constant daily litterfall*/
	    STATE.npp_sum = (float64) STATE_ARRAY->npp_sum.dom_1km[rc_1km] / NDAYS; 


	    /* printf("%d C[0] = %f  C[1] = %f  C[2] = %f  npp_sum = %f\n",  */
            /*        rc_1km, STATE.C[0], STATE.C[1], STATE.C[2], STATE.npp_sum); */

	    /*Initialize spinup iteration counter and tolerance*/
	    spinup_iter_cnt = 0;


	    /* set spinup tolerance to pre-defined high level:*/
	    spinup_current_iter_tol = STARTING_SPINUP_TOL; 
	    /* Some arbitrarily high floating point number, must be greater than stopping tolerance*/
	    
	    /* Tolerances should be specified as absolute (unsigned) values. */

	    /* Begin spin-up loop */
	    while(fabs(spinup_current_iter_tol) > runParamObj_p->spinup_stop_tol && 
	           spinup_iter_cnt < runParamObj_p->spinup_max_iter)
	    {	    

	      /*Initialize tolerance (Nee sum) for current spinup iteration*/
	      spinup_current_iter_tol = 0.0;

	      /* Begin time loop */
	      t8day = 0;
	      for (t=0; t < NDAYS; t++)
	      {

		nelem = runParamObj_p->nsite * NESTED_1KM_IN_9KM_SQ;
		rct_1km = M_2D_B0(t,rc_1km, nelem);
		rct_9km = M_2D_B0(t,rc_9km, runParamObj_p->nsite);

		/*Unpack input arrays for call interface*/
		MET.tmin = (float64) MetArrayObj.tmin[rct_9km]; /* [K] */
		MET.vpd = (float64) MetArrayObj.vpd[rct_9km];   /* [Pa] */
		MET.par = (float64) MetArrayObj.par[rct_9km] * RAD2PAR;   /* [MJ m-2] par */
		MET.smrz = (float64) MetArrayObj.smrz[rct_9km] * 100.0; /*convert from [dim] to [%]*/

		MET.tsoil = (float64) MetArrayObj.tsoil[rct_9km]; /* [K]*/
		MET.smtop = (float64) MetArrayObj.smtop[rct_9km] * 100.0; /*convert from [dim] to [%]*/ 
		     
		/*Read in fpar data when available every 8-days*/
		/*Otherwise, hold fpar constant at whatever was last read*/
		/*NOTE: this works for subsets, because the arrays are indexed to zero*/
		
		if (t == DAY8_to_DOY(t8day))
		{
		  
		  nelem = runParamObj_p->nsite * NESTED_1KM_IN_9KM_SQ;
		  rct8day_1km = M_2D_B0(t8day,rc_1km,nelem);
		  
			// FPAR_scale was in BPLUT, now in sesCtl_p as read from RIP file, which has 
			// inherited all of the glob_grid_params NVPs.
		  MET.fpar = (float64) MetArrayObj.fpar[rct8day_1km] * runParamObj_p->FPAR_scale + runParamObj_p->FPAR_offset;
		  t8day++;
		} // end:: t test on DOY


		/*Do tcf forward step to propagate SOC states from last timesetp to current timestep*/
		tcfModFunc_forwardStep(bp_p, MET, &STATE);

		/*Track info for spinup tolerance*/
		spinup_current_iter_tol += STATE.nee;
		
	      }
	      /*End time loop*/

	      /*Increment spin-up iteration counter (tolerance checked in while statement)*/
	      spinup_iter_cnt++;

	    }
	    /*End spin-up while-loop iterations*/

	    /*Package final tolerance info*/
 	    STATE_ARRAY->nee_sum.dom_1km[rc_1km] = (float32) spinup_current_iter_tol; 
	    STATE_ARRAY->spinup_iter[rc_1km] = (uint32) spinup_iter_cnt; 

	    /*Package final, spun-up SOC for future initialization*/
	    for (l=0; l<NPOOL; l++)
	      STATE_ARRAY->C[l].dom_1km[rc_1km] = (float32) STATE.C[l]; 

 	    /* printf("%d C[0] = %f  C[1] = %f  C[2] = %f  iter tol = %f  inter cnt = %d \n", */
            /*        rc_1km, STATE.C[0], STATE.C[1], STATE.C[2], spinup_current_iter_tol, spinup_iter_cnt); */
	    

	  } // end pft check

	}
      } // End nested 1km 2D grid loops

    } //  End 9km location loop

  }  // End PFT loop


} // END FUNCTION TCF_MOD_DRIVER_NUMSPIN_RUN() 




/*TCF_MOD_DRIVER_ANAPRECOND()                                                                      *
 *                                                                                                 *
 *    Drives analytical preconditioning for numerical spinup SOC initialization.                   *
 *                                                                                                 *
 *    RETURNS:  Initialized C pool 2D 1km grid arrays returned in the STATE structure.             *
 *              Also returns spinup tolerance and iteration count arrays (as 2D 1km grids)         *
 *              returned in the AUX structure.                                                     *
 *                                                                                                 *
 *    REQUIRES: Dynamic daily met data, static pft map, and static models parameters.              *                                                  
 *                                                                                                 *
 */
void tcfModDriverSubset_anaPrecond(PftAdt PFT, MetArrayAdt MetArrayObj, EcoStateArrayAdt *STATE_ARRAY)
{
  SesCtl *sesCtl_p = l4c_GetSesCtl(); // now, fetch ptr to L4_C high level session controls
  RunParamAdt *runParamObj_p = &sesCtl_p->runParamObj;  
  int       pft_index;
  int s;
  uint8     pft_now;

  uint32    p, t, t8day, l, rn, cn;
  uint32    rc_1km_nested, rc_1km, rc_9km;

  uint32    rct8day_1km, rct_1km, rct_9km;

  uint32    nelem;

  MetAdt    MET; // for local state calculations, not requiring large 2D grid arrays
  EcoStateAdt  STATE; // for local state calculations, not requiring large 2D grid arrays

  BPLUT                  *bp_p = NULL;

  // Begin PFT loop
  for (p=0; p < NPFT; p++)
  {

	// veg. pft from offset index 
    pft_now = p+1; 
    
    //  Retrieve bplut parameters based on PFT 
    pft_index = p+1;
    bp_p = l4c_GetBPLUT_ByIndex(pft_index, NPFT);

    //  Begin site loop
    for (s=0; s < runParamObj_p->nsite; s++)
    {

      rc_9km = s;
 
      // Begin nested 1km 2D grid loops
      for (rn=0; rn < NESTED_1KM_IN_9KM; rn++)
      {
		  
	for (cn=0; cn < NESTED_1KM_IN_9KM; cn++)
	{

	  rc_1km_nested = M_2D_B0(rn,cn,NESTED_1KM_IN_9KM);
	  rc_1km = M_2D_B0(rc_1km_nested, rc_9km, runParamObj_p->nsite);

	  //	  printf("%d  %d  %d\n", rc_1km_nested, PFT.dom_1km[rc_1km_nested], PFT.dom_1km[rc_1km]); 

	  /*do only specified pft*/
	  if (PFT.dom_1km[rc_1km] == pft_now)
	  {

	    /*initialize npp and kmult sums for grid cell*/
	    STATE.npp_sum = 0.0;  
	    STATE.kmult_sum = 0.0;  

	    /* Begin time loop*/
	    t8day = 0;
	    for (t=0; t < NDAYS; t++)
	    {

	      nelem = runParamObj_p->nsite * NESTED_1KM_IN_9KM_SQ;
	      rct_1km = M_2D_B0(t,rc_1km, nelem);
	      rct_9km = M_2D_B0(t,rc_9km, runParamObj_p->nsite);
	      
	      /*Unpack input arrays for call interface*/
	      MET.tmin = (float64) MetArrayObj.tmin[rct_9km]; /* [K] */
	      MET.vpd = (float64) MetArrayObj.vpd[rct_9km];   /* [Pa] */
	      MET.par = (float64) MetArrayObj.par[rct_9km] * RAD2PAR;   /* [MJ m-2] par */
	      MET.smrz = (float64) MetArrayObj.smrz[rct_9km] * 100.0; /*convert from [dim] to [%]*/

	      MET.tsoil = (float64) MetArrayObj.tsoil[rct_9km]; /* [K]*/
	      MET.smtop = (float64) MetArrayObj.smtop[rct_9km] * 100.0; /*convert from [dim] to [%]*/ 

	      /*Read in fpar data when available every 8-days*/
	      /*Otherwise, hold fpar constant at whatever was last read*/
	      /*NOTE: this works for subsets, because the arrays are indexed to zero*/
	      if (t == DAY8_to_DOY(t8day))
	      {
	
		nelem = runParamObj_p->nsite * NESTED_1KM_IN_9KM_SQ;

		rct8day_1km = M_2D_B0(t8day,rc_1km,nelem);

		// <JMG> fixed, now FPAR_scale (from glob_grid_params) is in sesCtl from RIP file that has
		// inherited all glob_grid_params name-value-pairs
        MET.fpar = (float64) MetArrayObj.fpar[rct8day_1km] * runParamObj_p->FPAR_scale + runParamObj_p->FPAR_offset;

		t8day++;
	
	      }

	      /*Get gpp and kmult*/
	      tcfModFunc_gpp(bp_p, MET, &STATE);
	      tcfModFunc_kmult(bp_p, MET, &STATE);

	      STATE.npp_sum += STATE.npp;  
	      STATE.kmult_sum += STATE.rh_mult.k;  

	    }
	    /*End time loop*/
	  
	    /*Calculate analytical steady-state (requires npp sum and kmult sums across time)*/
	    tcfModFunc_anaStst(bp_p, &STATE);

	    /*Now npp and kmult sums are complete for the 1km grid cell*/
	    STATE_ARRAY->npp_sum.dom_1km[rc_1km] = (float32) STATE.npp_sum;
	    STATE_ARRAY->kmult_sum.dom_1km[rc_1km] = (float32) STATE.kmult_sum;


	    /*Package final SOC for future initialization*/
	    for (l=0; l < NPOOL; l++)
	      STATE_ARRAY->C[l].dom_1km[rc_1km] = (float32) STATE.C[l];

	    if (s == 11 )
	    { 
	      printf("%d  %d  C[0] = %f  C[1] = %f  C[2] = %f  npp_sum = %f  kmult_sum = %f \n",
	       p,  s, STATE.C[0], STATE.C[1], STATE.C[2], STATE.npp_sum, STATE.kmult_sum);
	    } // end:: s is 11 check 

		} // end:: pft check

		}
	
		} // end:: nested 1km 2D grid loops

   
    }  // end:: location loop

  } // end:: PFT loop


} // end:: TCF_MOD_DRIVER_ANAPRECOND()   
