// ---------------------------------------------------------------
// module       : l4c_DAILY_ModQc.c
// purpose      : Encapsulates tcfModQc.c
// author       : Lucas A. Jones; lucas@ntsg.umt.edu
// author       : joe glassy, Lupine Logic Inc.
// version      : see L4C_VERS_TAG , as 4.1.5 or later
// revised      : 2016-03-02T11:00 <LAJ> Checked against science code and found discrepancies
//              :   Fixed ordering of error array to agree with science code. Commented-out and set
//              :   all freeze-thaw related error propagation to zero.  Freeze-thaw related error 
//              :   propagation will require more thorough vetting before re-implementation
//              :   because it does not currently have a counterpart in the science prototype code.
//              : 2014-10-16T12:59, updated code doc only, no logic changes.
// history      : 2014-08-07T11:47:00, <JMG> changed non-existent
//                bp_p->FparAge_QA_Error_max to FtAge_QA_Error_max
// history      : 2014-02-05T11:242:00-07:00,wheat,jmg"
// dependencies : HDF4, HDF5 libraries and L4C FRAMEWORK API.
// history      : n/a
// notes        : Based on and adapted from tcf_spin v3.1/3.2
// ---------------------------------------------------------------

// #include "tcf_inc.h"
#include "l4c_DAILY.h"
#include "l4c_DAILY_prot.h"

// --------------------------------------------------------------
//  module       : tcfModQa.c
//  purpose      : Core model quality assessment (QA) and control (QA) 
//                 logic. These functions parallel model state computations.
//  project      : SMAP L4C science development.
//  dependencies : tcfModMult.c, fpar_qa_bin_util.c
//  revision     : 2014-02-15T12:57:00, for lead fork with new ADT namespaces
//  revision     : Initial reformating -- laj 09/10/13
//  author       : Lucas A. Jones; lucas@ntsg.umt.edu
//  notes        : n/a
// 


// --------------------------------------------------------------
// tcfModQa_jac()

// Purpose: Performs jacobian calculations for model simulation step.  
//
// Notes: Currently, NEE is the only state variable in the jacobian, so it's a 1 x M vector, where M is the number
//        of input-variable partial derivatives.
//
// Inputs: State scalars (STATE), Dynamic daily met scalar data (MET), and static model parameters (GLOB_PARAMS, BPLUT).    
//
// Outputs: Model partial derivatives of NEE with respect to input variables (jacobian) stored as a 1 x M array.   
//          
//          
//          

void tcfModQa_jac(BPLUT *bp_p, MetAdt MET, EcoStateAdt *STATE)
{

  uint32     i;
  uint32     offset;

  float64    tmp;

  // Partial derivatives for NEE, GPP-related term of NEE sum
  STATE->ecoStateQaObj.jac[0] = MET.par * STATE->gpp_mult.k; // dnee/dfpar
  STATE->ecoStateQaObj.jac[1] = MET.fpar * STATE->gpp_mult.k; // dnee/dpar
  STATE->ecoStateQaObj.jac[2] = MET.fpar * MET.par * STATE->gpp_mult.vpd * STATE->gpp_mult.smrz * STATE->gpp_mult.ft * STATE->gpp_mult.d_tmin; // dnee/dtmin
  STATE->ecoStateQaObj.jac[3] = MET.fpar * MET.par * STATE->gpp_mult.tmin * STATE->gpp_mult.smrz  * STATE->gpp_mult.ft * STATE->gpp_mult.d_vpd; // dnee/dvpd
  STATE->ecoStateQaObj.jac[4] = MET.fpar * MET.par * STATE->gpp_mult.tmin * STATE->gpp_mult.vpd * STATE->gpp_mult.ft * STATE->gpp_mult.d_smrz; // dnee/dsmrz

  // <LAJ>  2016-03-02T11:00 Set FT-related jacobian to zero
  STATE->ecoStateQaObj.jac[5] = 0.0;
  
// <LAJ>  2016-03-03T11:55 Included FT-related error complier directive 
//         USE_FT_ERROR_PROP.
  
#ifdef USE_FT_ERROR_PROP

  STATE->ecoStateQaObj.jac[5] = MET.fpar * MET.par * STATE->gpp_mult.tmin * STATE->gpp_mult.smrz * STATE->gpp_mult.vpd * STATE->gpp_mult.d_ft; // dnee/dft

#endif


  // Apply GPP common factor (do now to keep above calculations clean and concise) 
  for (i=0; i< NMET_GPP; i++)
    STATE->ecoStateQaObj.jac[i] *= bp_p->LUEmax * (bp_p->fraut - 1.0);

  // Partial derivatives for NEE, Rh-related term of NEE sum
  offset = NMET_GPP;
  STATE->ecoStateQaObj.jac[offset] = STATE->rh_mult.smtop * STATE->rh_mult.d_tsoil;  // dnee/dtsoil

  offset = NMET_GPP+1;
  STATE->ecoStateQaObj.jac[offset] = STATE->rh_mult.tsoil * STATE->rh_mult.d_smtop;  // dnee/dsmtop
  
  // Apply Rh common factor (do now to keep above calculations clean and concise) 
  for (i=NMET_GPP; i<NMET; i++)
  {

      tmp = bp_p->kopt * (STATE->C[0] + (1.0 - bp_p->fmet) * bp_p->kstr * STATE->C[1] + bp_p->kslw * STATE->C[2]);
      
      STATE->ecoStateQaObj.jac[i] *= tmp;
    
  } // end:: i=NMET_GPP...

} // end :: tcfModQa_jac()



// -------------------------------------------------------------------------------------------------------------
// function :tcfModQa_step()
// purpose  : Performs QA calculations for one model simulation step.  
// notes: Currently, propagates error from FPAR only.  In the future, will include error from other input fields.
// onputs: State scalars (STATE), Dynamic daily met scalar data (MET), and static model parameters (GLOB_PARAMS, BPLUT).    
// outputs: Floating point output error fields for NEE and GPP fluxes are eturned in the STATE structure. Integer QA ranks 
//          reclassified from floating point are also currently returned, but this reclassification should really 
//          be done in post-processing for the 9km x pft grid, which requires averaging 1km grid cells (requiring
//          floating point) data.
// notes   : 
void tcfModQa_step(BPLUT *bp_p, MetAdt MET, EcoStateAdt *STATE)
{

  int i = 0;

  float64    tmp = 0.0;
  float64    neeError2 = 0.0;

  // Rescale integer classification (1-4) to input fpar quasi-error (float)
  if (MET.fpar_qa > 0)
    {
      STATE->ecoStateQaObj.met_error[0] = tcfModQa_rescal(MET.fpar_qa, bp_p->Fpar_QA_Rank_min, bp_p->Fpar_QA_Rank_max,
                       bp_p->Fpar_QA_Error_min, bp_p->Fpar_QA_Error_max);
    } else {

      STATE->ecoStateQaObj.met_error[0] = 0.0;

    }

  // Input met error fields QA-related parameters 
  // (NOTE: Must be in exactly the same order as the corresponding
  // Jacobian fields from tcfModQa_jac())
  
  // Percent Error (NOTE: must be careful with % Error with Fpar, 
  // because cloud-contaminated Fpar data has generally lower values)
  STATE->ecoStateQaObj.met_error[0] *= MET.fpar; 
  STATE->ecoStateQaObj.met_error[1] = bp_p->Par_QA_Error * MET.par; // Percent Error
  STATE->ecoStateQaObj.met_error[2] = bp_p->Tmin_QA_Error;
  STATE->ecoStateQaObj.met_error[3] = bp_p->Vpd_QA_Error * MET.vpd; // Percent Error
  STATE->ecoStateQaObj.met_error[4] = bp_p->Smrz_QA_Error;

  // <LAJ>  2016-03-02T11:00 Set FT-related error to zero and commented-out 
  //         above related codes.
  STATE->ecoStateQaObj.met_error[5] = 0.0;

  // Rescale freeze-thaw age (NOTE: Older freeze-thaw age cooresponds
  // to higher error)
  // <JMG> 2014-08-07T11:49, changed non-existent
  // bp_p->FparAge_QA_Error_max to FtAge_QA_Error_max

// <LAJ>  2016-03-03T11:55 Included FT-related error complier directive 
//         USE_FT_ERROR_PROP.
  
#ifdef USE_FT_ERROR_PROP
  STATE->ecoStateQaObj.met_error[5] = tcfModQa_rescal(MET.ftAge, 
    bp_p->FtAge_QA_Rank_min, 
    bp_p->FtAge_QA_Rank_max,
    bp_p->FtAge_QA_Error_min, 
    bp_p->FtAge_QA_Error_max);
  
  // Apply additional error if using Tsurf FT
  if (MET.ftMethod == 1) 
    { 
      STATE->ecoStateQaObj.met_error[5] *= (1.0 + bp_p->FtMethod_QA_mult);
    }
#endif

  // <LAJ>  2016-03-02T14:00 Revised ordering to agree with science code.
  //        Previously Smtop was incorrectly listed first.
  STATE->ecoStateQaObj.met_error[6] = bp_p->Tsoil_QA_Error;
  STATE->ecoStateQaObj.met_error[7] = bp_p->Smtop_QA_Error;  
  
  // Compute Jacobian
  tcfModQa_jac(bp_p, MET, STATE);

  // Error propagation (i.e. maps 'input' error to 'output' error)
  // NOTE: Un-squared (std. dev.) error units are [g C m-2 d-1].
  
  // Compute output error variance (input errors assumed independent, so can compute scalar multiplication and sum across squares)  
  neeError2 = 0.0;
  for (i=0; i < NMET; i++)
  {
      tmp = pow(STATE->ecoStateQaObj.met_error[i], 2.0);
      tmp *= pow(STATE->ecoStateQaObj.jac[i], 2.0);
      neeError2 += tmp;
  }

  // Compute nee error standard deviation
  STATE->ecoStateQaObj.nee_error = sqrt(neeError2);

} // end :: tcfModQa_step()


// -------------------------------------------------------------------------------------------------------------
// function : tcfModQa_rescal()
// Purpose  : Generic utility function to rescale an integer value with an
//            integer range into a float value  with specified float range.
// Notes    : Required for initial assignment of input error float ranges 
//            from input QA integer rank fields.  If input 
//            QA is provided in a bitfield (as for MOD15A2 FPAR), the 
//            bitfield will be reclassified to integer rank and then 
//            rescaled to a float range. 
//
// Inputs: Integer rank scalar, and pre-specified integer rank extrema and float extrema.    
//
// Outputs: Rescaled float scalar.

float64 tcfModQa_rescal(uint8 qa_rank, uint32 qa_rank_mn, uint32 qa_rank_mx, float64 error_mn, float64 error_mx)
{

  float64  qa_rank_flt = 0.0;
  float64  error= 0.0;

  // Step 1: Rescale integer rank as float (float from 0 to 1)
  qa_rank_flt = (float64) (qa_rank - qa_rank_mn) / (qa_rank_mx - qa_rank_mn);

  //Step 2: Rescale rank floats (float [0,1]) to state error units (float, [state units]) 
  error = qa_rank_flt * (error_mx - error_mn) + error_mn;

  return(error);

} // end:: tcfModQa_rescal()



// ---------------------------------------------------------------------
// function : tcfModQa_reclas()
// Purpose  : Generic utility function to reclassify float value  with 
//            specified float range into an integer value
//            with a specified integer range. e.g. Reclassify a float
//            value with a float range into an integer value with a 
//            specified integer range
// Notes     : Required for assignment of output QA integer rank fields, 
//             which will likely be done in post-processing,   
//             but retain the capability here in case 1km QA fields are desired.
//
// Inputs: Float scalar, and pre-specified float extrema and integer rank extrema.    
//
// Outputs: Reclassified float scalar.

uint8 tcfModQa_reclas(float64 error, float64 error_mn, float64 error_mx, uint32 qa_rank_mn, uint32 qa_rank_mx)
{

  uint8      qa_rank;

  float64    tmp;
  float64    qa_rank_flt;

  // <LAJ, 10/22/2014> Ensure error does not exceed specified bounds
  error = (error > error_mx) ? error_mx : error;
  error = (error < error_mn) ? error_mn : error;

  // Step 1: Rescale float range from 0 to 1
  qa_rank_flt = (error - error_mn) / (error_mx - error_mn);

  // Step 2: Rescale rank floats (float [0,1]) to state error units (float, [state units]) 
  tmp = (float64) qa_rank_flt * (qa_rank_mx - qa_rank_mn) + qa_rank_mn;

  // <JMG> round state error units to nearest integer using FRM API's round
  qa_rank = (uint8)FRM_API_ROUND(tmp); 

  return(qa_rank);

} // end :: tcfModQa_reclas()
